/*
Copyright © 2020-2023 IoT.bzh Company
Contact: https://www.iot.bzh/licensing

This file is part of the redpesk-cli project, the redpesk command line tool.

$RP_BEGIN_LICENSE$
Commercial License Usage
 Licensees holding valid commercial IoT.bzh licenses may use this file in
 accordance with the commercial license agreement provided with the
 Software or, alternatively, in accordance with the terms contained in
 a written agreement between you and The IoT.bzh Company. For licensing terms
 and conditions see https://www.iot.bzh/terms-conditions. For further
 information use the contact form at https://www.iot.bzh/contact.

GNU General Public License Usage
 Alternatively, this file may be used under the terms of the GNU General
 Public license version 3. This license is as published by the Free Software
 Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
 of this file. Please review the following information to ensure the GNU
 General Public License requirements will be met
 https://www.gnu.org/licenses/gpl-3.0.html.
$RP_END_LICENSE$
*/

package cmd

import (
	"errors"
	"fmt"

	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/printer"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/utils"
	"git.ovh.iot/redpesk/redpesk-infra/rp-webserver/actions/rpapiv1"
	"github.com/spf13/cobra"
	"github.com/spf13/viper"
)

func initAdminCmd(rootCmd *cobra.Command) {

	// adminCmd represents the admin command
	var adminCmd = &cobra.Command{
		Use:    adminCommand,
		Short:  adminCmdShortDesc,
		Hidden: !isUserAdmin,
	}

	// initCmd represents the "admin init" command
	var initCmd = &cobra.Command{
		Use:   adminInitCmd,
		Short: adminInitCmdShortDesc,
		RunE:  checkAndInitAdmin,
	}

	// importPrjCmd represents the "admin import-project" command
	var importPrjCmd = &cobra.Command{
		Use:     importPrjCmd,
		Short:   importPrjCmdShortDesc,
		Long:    importPrjCmdLongDesc,
		PreRunE: initEventProject,
		RunE:    importProject,
		Annotations: map[string]string{
			"__flag_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":        "TRUE"},
	}

	// importImgCmd represents the "admin import-project" command
	var importImgCmd = &cobra.Command{
		Use:     importImgCmd,
		Short:   importImgCmdShortDesc,
		Long:    importImgCmdLongDesc,
		PreRunE: initEventImages,
		RunE:    importImage,
		Annotations: map[string]string{
			"__flag_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":        "TRUE"},
	}

	// Here you will define your flags and configuration settings.

	// Persistent Flags -----------------------------------------
	// ----------------------------------------------------------

	// Local Flags ----------------------------------------------
	// Define the flags specific to "import-project" command
	importPrjCmd.Flags().StringP(importPrjCmdIDFlag, importPrjCmdIDFlagAlias, "", importPrjCmdIDFlagUsage)
	importPrjCmd.Flags().BoolP(importPrjCmdRefreshFlag, importPrjCmdRefreshFlagAlias, false, importPrjCmdRefreshFlagUsage)
	importPrjCmd.Flags().StringP(importPrjCmdVersionFlag, importPrjCmdVersionFlagAlias, "", importPrjCmdVersionFlagUsage)
	importPrjCmd.Flags().String(importPrjCmdSubVersionFlag, "", importPrjCmdSubVersionFlagUsage)
	importPrjCmd.Flags().StringP(importPrjCmdTagFlag, importPrjCmdTagFlagAlias, "", importPrjCmdTagFlagUsage)
	importPrjCmd.Flags().StringP(importPrjCmdDescFlag, importPrjCmdDescFlagAlias, "", importPrjCmdDescFlagUsage)
	// Define the flags specific to "import-image" command
	importImgCmd.Flags().StringP(imagesCmdNameFlag, imagesCmdNameFlagAlias, "", imagesCmdNameFlagUsage)
	importImgCmd.Flags().StringP(imagesCmdAliasFlag, imagesCmdAliasFlagAlias, "", imagesCmdAliasFlagUsage)
	importImgCmd.Flags().StringP(imagesCmdDistroFlag, imagesCmdDistroFlagAlias, "", imagesCmdDistroFlagUsage)
	importImgCmd.Flags().StringP(imagesCmdDistroVersionFlag, imagesCmdDistroVersionFlagAlias, "", imagesCmdDistroVersionFlagUsage)
	importImgCmd.Flags().String(imagesCmdArchFlag, "", imagesCmdArchFlagUsage)
	importImgCmd.Flags().StringP(imagesCmdSecurityFlag, imagesCmdSecurityFlagAlias, "", imagesCmdSecurityFlagUsage)
	importImgCmd.Flags().String(imagesCmdKickstartURLFlag, "", imagesCmdKickstartURLFlagUsage)
	importImgCmd.Flags().StringP(imagesCmdKickstartNameFlag, imagesCmdKickstartNameFlagAlias, "", imagesCmdKickstartNameFlagUsage)
	importImgCmd.Flags().StringSliceP(imagesCmdAddRepoFlag, imagesCmdAddRepoFlagAlias, []string{}, imagesCmdAddRepoFlagUsage)
	importImgCmd.Flags().StringP(imagesCmdPrjFlag, imagesCmdPrjFlagAlias, "", imagesCmdPrjFlagUsage)
	importImgCmd.Flags().StringP(imagesCmdDlURLFlag, imagesCmdDlURLFlagAlias, "", imagesCmdDlURLFlagUsage)
	importImgCmd.Flags().StringP(imagesCmdBoardModelFlag, imagesCmdBoardModelFlagAlias, "", imagesCmdBoardModelFlagUsage)
	importImgCmd.Flags().BoolP(imagesCmdForceFlag, imagesCmdForceFlagAlias, false, imagesCmdForceFlagUsage)
	// ----------------------------------------------------------

	// Add some flags as required -------------------------------
	importPrjCmd.MarkFlagRequired(importPrjCmdIDFlag)
	importImgCmd.MarkFlagRequired(imagesCmdNameFlag)
	importImgCmd.MarkFlagRequired(imagesCmdDistroFlag)
	importImgCmd.MarkFlagRequired(imagesCmdArchFlag)
	// ----------------------------------------------------------

	// Specify dynamic flag completion
	importPrjCmd.RegisterFlagCompletionFunc(importPrjCmdIDFlag, getListExtPrjsAsTab)
	importImgCmd.RegisterFlagCompletionFunc(imagesCmdArchFlag, getListArchAsTab)
	importImgCmd.RegisterFlagCompletionFunc(imagesCmdDistroFlag, getListDistAsTab)
	importImgCmd.RegisterFlagCompletionFunc(imagesCmdPrjFlag, getListPrjsAsTab)
	importImgCmd.RegisterFlagCompletionFunc(imagesCmdSecurityFlag, getSecuModelsAsTab)

	// Add the group commands "admin" as subcommand of rootCmd
	rootCmd.AddCommand(adminCmd)
	// Add the command "init" as a subcommand of "admin"
	adminCmd.AddCommand(initCmd)
	// Add the command "import-project" as a subcommand of "admin"
	adminCmd.AddCommand(importPrjCmd)
	// Add the command "import-image" as a subcommand of "admin"
	adminCmd.AddCommand(importImgCmd)
}

// --------------------------------------------------------------
// Completion functions -----------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Command functions --------------------------------------------
// --------------------------------------------------------------

// checkAndInitAdmin checks that the user is admin on the
// current factory and update the connections file according to it
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func checkAndInitAdmin(cmd *cobra.Command, args []string) error {

	// Print the first line for this command
	RPPrinter.CustomPrintLn(adminInitCmdAnsLine0)

	// Request the backend in order to see if the user is admin on the current factory
	userIsAdmin, err := isCurrentUserAdmin("")
	if err != nil {
		RPPrinter.CustomColorPrintF(printer.ColorRed, adminInitCmdAnsLine1, globalKOLabel)
		return err
	}

	// Get the current server alias
	serverAlias := viper.GetString(serverAliasFlag)

	if !userIsAdmin {
		// The user is not an admin, return an error
		RPPrinter.CustomColorPrintF(printer.ColorRed, adminInitCmdAnsLine1, globalKOLabel)
		errorMsg := fmt.Sprintf(userIsNotAdminMsg, serverAlias)
		return errors.New(errorMsg)
	}

	// Get the current connection parameters
	connecParams, err := getConnectionParams(serverAlias, accessTokenFile)
	if err != nil {
		RPPrinter.CustomColorPrintF(printer.ColorRed, adminInitCmdAnsLine1, globalKOLabel)
		return err
	}

	// If the user is admin, modify it in the access tokens file (if necessary)
	if userIsAdmin && !connecParams.IsAdmin {
		connecParams.IsAdmin = true
		err = appendOrOverwriteInTokenFile(accessTokenFile, connecParams)
		if err != nil {
			RPPrinter.CustomColorPrintF(printer.ColorRed, adminInitCmdAnsLine1, globalKOLabel)
			return err
		}
	}

	RPPrinter.CustomColorPrintF(printer.ColorGreen, adminInitCmdAnsLine1, globalOKLabel)
	return nil
}

// importProject allows an administrator to import a project
// existing in Koji on the remote factory
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func importProject(cmd *cobra.Command, args []string) error {

	Log.Warnln("[DEPRECATED] This command is deprecated and soon wil be deleted, please use 'rp-cli distributions add' or 'rp-cli distributions update' instead")

	// Get the "--refresh" flag value
	isRefresh, err := cmd.Flags().GetBool(importPrjCmdRefreshFlag)
	if err != nil {
		logFlagError(importPrjCmdRefreshFlag)
		return err
	}

	// Print the first line for this command
	if isRefresh {
		RPPrinter.CustomPrintLn(importPrjCmdAnsLine1_0)
	} else {
		RPPrinter.CustomPrintLn(importPrjCmdAnsLine0_0)
	}

	// Get the value of the name flag
	projectName, err := cmd.Flags().GetString(importPrjCmdIDFlag)
	if err != nil {
		logFlagError(importPrjCmdIDFlag)
		return err
	}

	if projectName == "" {
		return errors.New(emptyPrjNameMsg)
	}

	// Get the value of the version flag
	prjVersion, err := cmd.Flags().GetString(importPrjCmdVersionFlag)
	if err != nil {
		logFlagError(importPrjCmdVersionFlag)
		return err
	}

	// Get the value of the sub-version flag
	prjSubVersion, err := cmd.Flags().GetString(importPrjCmdSubVersionFlag)
	if err != nil {
		logFlagError(importPrjCmdSubVersionFlag)
		return err
	}

	// Get the tag flag
	prjTag, err := cmd.Flags().GetString(importPrjCmdTagFlag)
	if err != nil {
		logFlagError(importPrjCmdTagFlag)
		return err
	}

	// Description
	impDistroDesc, err := cmd.Flags().GetString(importPrjCmdDescFlag)
	if err != nil {
		logFlagError(importPrjCmdDescFlag)
		return err
	}

	prjAns := rpapiv1.RPProjectFull{}

	body := rpapiv1.RPProjectFull{
		RPProject: rpapiv1.RPProject{
			Description: impDistroDesc,
		},
	}
	if isRefresh {
		// Get the project
		project, err := getProjectFromIdentifier(projectName, false)
		if err != nil {
			return err
		}
		// Request the backend to refresh the imported project
		RPPrinter.CustomPrintF(importPrjCmdAnsLine1_1, projectName)
		if err := bckAPI.ProjectMngtRefresh(project.ID.String(), &prjAns); err != nil {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
			return err
		}
	} else {
		// Request the backend to import the project
		RPPrinter.CustomPrintF(importPrjCmdAnsLine0_1, projectName)
		if err := bckAPI.ProjectMngtImport(projectName, prjVersion, prjSubVersion, prjTag, &body, &prjAns); err != nil {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
			return err
		}
	}

	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	// Display something to the user if in raw mode
	RPPrinter.CustomPrintObject(prjAns, false, displayNothing)

	// Check from viper the non-blocking flag to see if we wait
	nonBlocking := viper.GetBool(nonBlockingFlag)
	if nonBlocking {
		return nil
	}

	// If we reach here, the user wants the command to be blocking
	if isRefresh {
		RPPrinter.CustomPrintF(importPrjCmdAnsLine1_2, projectName+" "+prjVersion)
	} else {
		RPPrinter.CustomPrintF(importPrjCmdAnsLine0_2, projectName+" "+prjVersion)
	}
	RPPrinter.ForceFlush()

	// Block until we receive a corresponding websocket event
	for {
		rEvent := <-cmdEventChannel

		// Check that the received event is for a project state changed
		if rEvent.Type == rpapiv1.EVTProjectStateChange {
			// Decode the received project in event
			rProject, err := rEvent.DecodeProject()
			if err != nil {
				RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
				return err
			}

			if utils.UUIDEqual(rProject.ID, prjAns.ID) && (rProject.State == rpapiv1.RPPrjStableState) {
				// Display OK if not in raw mode
				RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
				// Display something to the user if in raw mode
				RPPrinter.CustomPrintObject(rProject, false, displayNothing)
				break
			}
		}

		// Check if we get an error event during the operation
		if (rEvent.Type == rpapiv1.EVTError) || (rEvent.Type == rpapiv1.EVTSessionExpired) {
			// Function handling the events
			err = handleErrorEvent(prjAns.ID, rEvent, true)
			if err != nil {
				return err
			}
		}
	}

	return nil
}

// importImage allows an administrator to import an image
// in the backend
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func importImage(cmd *cobra.Command, args []string) error {
	// Display the first line to the user
	RPPrinter.CustomPrintLn(importImgCmdAnsLine0)

	// Create the structure to fill
	img := rpapiv1.RPImageFull{}

	// Fill the structure with the right values (given in command line)
	if err := prepareImageStruct(cmd, args, &img, true, true); err != nil {
		return err
	}

	// Tell the user that a request is sent to the backend
	RPPrinter.CustomPrintF(importImgCmdAnsLine1, img.Name)

	// Request the import of the image
	imgAns := rpapiv1.RPImageFull{}
	if err := bckAPI.ImageMngtImport(&img, &imgAns); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	// Display [OK]
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	// Display something to the user if in raw mode
	RPPrinter.CustomPrintObject(imgAns, false, displayNothing)

	// // Check from viper the non-blocking flag to see if we wait
	// nonBlocking := viper.GetBool(nonBlockingFlag)
	// if nonBlocking {
	// 	// The user do not want blocking functions
	// 	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
	// 	return nil
	// }

	// // Block until we receive a corresponding websocket event
	// for {
	// 	rEvent := <-cmdEventChannel

	// 	// Check that the received event is for a image addition
	// 	if rEvent.Type == rpapiv1.EVTImageAdd {
	// 		// Function handling the events
	// 		// DecodeImage needs to be done!
	// 		rImage, err := rEvent.DecodeImage()
	// 		if err != nil {
	// 			return err
	// 		}

	// 		if utils.UUIDEqual(rImage.ID, imgAns.ID) {
	// 			// Display OK if not in raw mode
	// 			RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
	// 			// Display something to the user if in raw mode
	// 			RPPrinter.CustomPrintObject(rImage, false, displayNothing)
	// 			break
	// 		}
	// 	}
	//  // TODO: MANAGE THE ERROR CASE
	// }

	return nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Utilities functions ------------------------------------------
// --------------------------------------------------------------

// isUserAdminOnAnyStack is a private function allowing to know if
// the user is administrator on, at least, one factory
//
// :returns: (bool, error) -- true if the user is admin and an err or if a problem occurs
func isUserAdminOnAnyStack() (bool, error) {
	// Initialize the result variable
	isUserAdmin := false

	// Get all the connection parameters that exist
	allConnectionParams, err := getAllConnectionsParams(accessTokenFile)
	if err != nil {
		return isUserAdmin, err
	}

	// Check if at least the user is Admin on one connection
	for _, connection := range allConnectionParams {
		if connection.IsAdmin {
			isUserAdmin = true
			break
		}
	}

	return isUserAdmin, nil
}

// isUserAdminOnCurrentStack is a private function allowing to know if
// the user is administrator on the current factory. The user would have needed
// to run `rp-cli admin init` to validate that
//
// :returns: (bool, error) -- true if the user is admin and an err or if a problem occurs
func isUserAdminOnCurrentStack() (bool, error) {
	// Get the current server alias
	serverAlias := viper.GetString(serverAliasFlag)

	// Get the current connection parameters
	connecParams, err := getConnectionParams(serverAlias, accessTokenFile)
	if err != nil {
		return false, err
	}

	return connecParams.IsAdmin, nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Display functions --------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// --------------------------------------------------------------
