/*
Copyright © 2020-2024 IoT.bzh Company
Contact: https://www.iot.bzh/licensing

This file is part of the redpesk-cli project, the redpesk command line tool.

$RP_BEGIN_LICENSE$
# Commercial License Usage

Licensees holding valid commercial IoT.bzh licenses may use this file in
accordance with the commercial license agreement provided with the
Software or, alternatively, in accordance with the terms contained in
a written agreement between you and The IoT.bzh Company. For licensing terms
and conditions see https://www.iot.bzh/terms-conditions. For further
information use the contact form at https://www.iot.bzh/contact.

# GNU General Public License Usage

Alternatively, this file may be used under the terms of the GNU General
Public license version 3. This license is as published by the Free Software
Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
of this file. Please review the following information to ensure the GNU
General Public License requirements will be met
https://www.gnu.org/licenses/gpl-3.0.html.

$RP_END_LICENSE$
*/
package cmd

import (
	"strings"
	"testing"

	"git.ovh.iot/redpesk/redpesk-infra/rp-webserver/actions/rpapiv1"
	"github.com/stretchr/testify/assert"
)

// TestAdminInitUserAdmin is the function allowing to test if
// the command "rp-cli admin init" is working correctly when the user
// requesting it is an "user admin"
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAdminInitUserAdmin(test *testing.T) {
	// Create a fake access token file (with all IsAdmin of connections to false)
	accessTokenFilePath, err := createFalseTokenFile(nil)
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false file created at the beginning
	defer deleteFalseFile(accessTokenFilePath)

	// Create a fake configuration file (serveralias set to 'mytestalias')
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Run the `rp-cli admin init` command and check that all is [OK]
	checkActionCmdAns(test, adminCommand, "init")

	// Check that the 'mytestalias' connection has 'isAdmin' to true
	connecParams, err := getConnectionParams("mytestalias", accessTokenFile)
	if err != nil {
		test.Fatal(err)
	}
	assert.True(test, connecParams.IsAdmin)
}

// TestAdminInitSuperAdmin is the function allowing to test if
// the command "rp-cli admin init" is working correctly when the user
// requesting it is a "superAdmin"
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAdminInitSuperAdmin(test *testing.T) {
	// Create a fake access token file (with all IsAdmin of connections to false)
	accessTokenFilePath, err := createFalseTokenFile(nil)
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false file created at the beginning
	defer deleteFalseFile(accessTokenFilePath)

	// Create a fake configuration file (serveralias set to 'mytestalias')
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Tell the mock to return the "superAdmin" user
	superAdminRoles := rpapiv1.RPRoles{
		Roles: []rpapiv1.RPRole{
			{
				Name:   "superAdmin of system",
				Label:  "superAdmin",
				Object: "system",
			},
		},
	}
	MockedRPBackend.SetCurrentUserRole(superAdminRoles, false)

	// Run the `rp-cli admin init` command and check that all is [OK]
	checkActionCmdAns(test, adminCommand, "init")

	// Put the mock to its default value
	MockedRPBackend.SetCurrentUserRole(rpapiv1.RPRoles{}, true)

	// Check that the 'mytestalias' connection has 'isAdmin' to true
	connecParams, err := getConnectionParams("mytestalias", accessTokenFile)
	if err != nil {
		test.Fatal(err)
	}
	assert.True(test, connecParams.IsAdmin)
}

// TestAdminCompletion is the function allowing to test if
// the completion on "rp-cli admin" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAdminCompletion(test *testing.T) {
	// Create a fake access token file (with all IsAdmin of connections to false)
	accessTokenFilePath, err := createFalseTokenFile(nil)
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false file created at the beginning
	defer deleteFalseFile(accessTokenFilePath)

	// Create a fake configuration file (serveralias set to 'mytestalias')
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Check that the completion is not done on 'admin' subcommand
	cmdResults, err := runCompletionCmd()
	if err != nil {
		test.Fatal(err)
	}
	for _, result := range cmdResults {
		assert.NotRegexp(test, "^admin", result)
	}

	// Run the `rp-cli admin init` command and check that all is [OK]
	checkActionCmdAns(test, adminCommand, "init")

	// Check that the completion now contains the admin sub-command
	adminSubcmdIncluded := false
	cmdResults, err = runCompletionCmd()
	if err != nil {
		test.Fatal(err)
	}
	for _, result := range cmdResults {
		if strings.HasPrefix(result, "admin") {
			adminSubcmdIncluded = true
			break
		}
	}
	assert.True(test, adminSubcmdIncluded)
}

// TestAdminImportImage is the function allowing to test if
// the command "rp-cli admin import-image ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAdminImportImage(test *testing.T) {
	// Create the command
	cmdToRun := getMainImportImgCmd("Xilinx aarch64 image", "xilinx-aarch64-image", "aarch64", "redpesk-devel-arz", "1.0", "smack",
		"-u", "http://download-image.bzh",
		"-k", "Redpesk-core-aarch64-selinux.ks",
		"--ks-url", "http://git.ovh.iot/redpesk/redpesk-infra/imager-configuration.git?kickstarts/RedPesk#02d883f4",
		"-r", "https://download.redpesk.bzh/redpesk-devel/releases/arz/Everything/aarch64/os/",
		"-r", "https://download.redpesk.bzh/redpesk-devel/releases/arz/AGL/aarch64/os/")

	// Create the image if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, imagesCmdName, "xilinx-aarch64-image", cmdToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// The image has been created, check it is now in backend

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID":           `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Slug":         `^xilinx-aarch64-image$`,
		"Name":         `^Xilinx aarch64 image$`,
		"Architecture": `^aarch64$`,
		"Type":         `^iotbzh \(external\)$`,
		"Disabled":     `^false$`,
	}

	// Run the command and the tests (no verbose and verbose, since this is the same result)
	checkCmdAnsKeysRegexpCol(test, imgDisplayNoVerboseColNb, regexpMap, imagesCmdName, "get", "xilinx-aarch64-image")
}

// TestAdminImportProject is the function allowing to test if
// the command "rp-cli admin import-project ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAdminImportProject(test *testing.T) {
	// Create the projects if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, prjsCommand, "redpesk-devel-cezembre", "admin", "import-project", "-i", "redpesk-devel-cezembre",
		"-v", "1.0", "--sub-version", "44", "-d", "description of imported distro")
	if err != nil {
		test.Fatal(err)
	}

	// Check that the project exists

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":               `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":              `^redpesk-devel-cezembre$`,
		"Slug":              `^redpesk-devel-cezembre$`,
		"Version":           `^1\.0$`,
		"SubVersion":        `^44$`,
		"Type":              `^distro \(external\)-applications$`,
		"Supported Objects": `^applications$`,
		"Description":       `description of imported distro`,
		"Distributions":     `.*`,
		"Architectures":     `.*`,
		"Creation time":     `.+`,
		"Update time":       `.+`,
		"Bookmarked":        `^(true|false)$`,
	}

	// Run the command and the tests
	checkCmdAnswerListKeysRegexp(test, projectDisplayVerboseLightLineNb, "", regexpMap, prjsCommand, "get", "redpesk-devel-cezembre", "-v")
}

// TestAdminRefreshImportedProject is the function allowing to test if
// the command "rp-cli admin import-project ..." is working correctly
// when the flag "--refresh" is used
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAdminRefreshImportedProject(test *testing.T) {
	// Create the projects if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, prjsCommand, "redpesk-devel-hoedic", "admin", "import-project", "-i", "redpesk-devel-hoedic", "-v", "1.5")
	if err != nil {
		test.Fatal(err)
	}

	// Ask for the refresh of the project
	checkActionCmdAns(test, "admin", "import-project", "-i", "redpesk-devel-hoedic", "--refresh")

	// Check that the project exists

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":               `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":              `^redpesk-devel-hoedic$`,
		"Slug":              `^redpesk-devel-hoedic$`,
		"Version":           `^1\.5-refreshed$`,
		"Type":              `^distro \(external\)-applications$`,
		"Supported Objects": `^applications$`,
		"Description":       `.+`,
		"Distributions":     `.*`,
		"Architectures":     `.*`,
		"Creation time":     `.+`,
		"Update time":       `.+`,
		"Bookmarked":        `^(true|false)$`,
	}

	// Run the command and the tests
	checkCmdAnswerListKeysRegexp(test, projectDisplayVerboseLightLineNb, "", regexpMap, prjsCommand, "get", "redpesk-devel-hoedic", "-v")
}

// --------------------------------------------------------------
// Utilities functions ------------------------------------------
// --------------------------------------------------------------

// getMainImportImgCmd is an initializer function to get the mainly
// used command to import an image
//
// :param name: Name of the image to create
// :type name: string
// :param alias: Alias of the image to create
// :type alias: string
// :param arch: Architecture of the image to create
// :type arch: string
// :param distro: Distribution of the image to create
// :type distro: string
// :param distVersion: Distribution version of the image to create
// :type distVersion: string
// :param secu: Security model of the image to create
// :type secu: string
// :param args: Other args to pass as command to create the project
// :type args: ...string
// :returns: []string -- the value of the wanted constant
func getMainImportImgCmd(name, alias, arch, distro, distVersion, secu string, args ...string) []string {
	cmdToReturn := []string{
		adminCommand,
		importImgCmd,
		"-n",
		name,
		"-a",
		alias,
		"--arch",
		arch,
		"-d",
		distro,
		"-v",
		distVersion,
		"-s",
		secu,
	}

	// Add other args to the command
	cmdToReturn = append(cmdToReturn, args...)

	return cmdToReturn
}

// --------------------------------------------------------------
// --------------------------------------------------------------
