/*
Copyright © 2020-2023 IoT.bzh Company
Contact: https://www.iot.bzh/licensing

This file is part of the redpesk-cli project, the redpesk command line tool.

$RP_BEGIN_LICENSE$
Commercial License Usage
 Licensees holding valid commercial IoT.bzh licenses may use this file in
 accordance with the commercial license agreement provided with the
 Software or, alternatively, in accordance with the terms contained in
 a written agreement between you and The IoT.bzh Company. For licensing terms
 and conditions see https://www.iot.bzh/terms-conditions. For further
 information use the contact form at https://www.iot.bzh/contact.

GNU General Public License Usage
 Alternatively, this file may be used under the terms of the GNU General
 Public license version 3. This license is as published by the Free Software
 Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
 of this file. Please review the following information to ensure the GNU
 General Public License requirements will be met
 https://www.gnu.org/licenses/gpl-3.0.html.
$RP_END_LICENSE$
*/

package cmd

import (
	"encoding/json"
	"errors"
	"fmt"
	"os"
	"path"
	"slices"
	"sort"
	"strconv"
	"strings"
	"text/tabwriter"
	"time"

	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/printer"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/rpxxxapi/bckapi"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/utils"
	"git.ovh.iot/redpesk/redpesk-infra/rp-webserver/actions/rpapiv1"
	"github.com/gofrs/uuid"
	"github.com/jinzhu/copier"
	"github.com/spf13/cobra"
	"github.com/spf13/pflag"
	"github.com/spf13/viper"
)

// -------------------------------------------------
// Types definition --------------------------------
// -------------------------------------------------

// -------------------------------------------------
// -------------------------------------------------

// -------------------------------------------------
// Global variable (inside this package) -----------
// -------------------------------------------------

// -------------------------------------------------
// -------------------------------------------------

// --------------------------------------------------------------
// Initialization functions -------------------------------------
// --------------------------------------------------------------

// Initialization function for the applications commands
func initAppCmd(rootCmd *cobra.Command) {
	// appsCmd represents the applications command
	var appsCmd = &cobra.Command{
		Use:     appsCommand,
		Aliases: []string{appsCommandAlias},
		Short:   appsCommandShortUsage,
	}

	// listAppsCmd represents the "applications list" command
	var listAppsCmd = &cobra.Command{
		Use:     appsListCmd,
		Aliases: []string{appsListCmdAlias},
		Short:   appsListCmdShortUsage,
		RunE:    listApplications,
		Annotations: map[string]string{
			"__flag_need_bck_connect__": "TRUE",
			"__read_object_command__":   "TRUE"},
	}

	// getAppsCmd represents the "applications get" command
	var getAppsCmd = &cobra.Command{
		Use:               appsGetCmd,
		Short:             appsGetCmdShortUsage,
		RunE:              getApplication,
		ValidArgsFunction: getFullListAppsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__read_object_command__":  "TRUE"},
	}

	// addAppsCmd represents the "applications add" command
	var addAppsCmd = &cobra.Command{
		Use:     appsAddCmd,
		Aliases: []string{appsAddAlias},
		Short:   appsAddCmdShortUsage,
		PreRunE: initEventApp,
		RunE:    createApplication,
		Annotations: map[string]string{
			"__flag_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":        "TRUE"},
	}

	// addAppsFromSampleCmd represents the "applications add-from-sample" command
	var addAppsFromSampleCmd = &cobra.Command{
		Use:               appsAddFromSampleCmd,
		Short:             appsAddFromSampleCmdShortUsage,
		Long:              appsAddFromSampleCmdLongUsage,
		PreRunE:           initEventApp,
		RunE:              createApplicationFromSample,
		ValidArgsFunction: getListAppsSamplesAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// updateAppsCmd represents the "applications update" command
	var updateAppsCmd = &cobra.Command{
		Use:               appsUpdateCmd,
		Aliases:           []string{appsUpdateAlias},
		Short:             appsUpdateCmdShortUsage,
		PreRunE:           initEventApp,
		RunE:              updateApplication,
		ValidArgsFunction: getListAppsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// delAppsCmd represents the "applications delete" command
	var delAppsCmd = &cobra.Command{
		Use:               appsDelCmd,
		Aliases:           []string{appsDelAlias},
		Short:             appsDelCmdShortUsage,
		PreRunE:           initEventApp,
		RunE:              deleteApplication,
		ValidArgsFunction: getListAppsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// bookmarkAppCmd represents the "applications bookmark" command
	var bookmarkAppCmd = &cobra.Command{
		Use:               appsBookmarkCmd,
		Aliases:           []string{appsBookmarkAlias},
		Short:             appsBookmarkCmdShortUsage,
		PreRunE:           initEventApp,
		RunE:              bookmarkApplication,
		ValidArgsFunction: getFullListAppsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__read_object_command__":  "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// filesCmd represents the "applications files" command
	var filesCmd = &cobra.Command{
		Use:   appsFilesCmd,
		Short: appsFilesCmdShortUsage,
	}

	// contentAppsCmd represents the "applications content" command
	var contentAppsCmd = &cobra.Command{
		Use:               appsFilesListCmd,
		Aliases:           []string{appsFilesListAlias1, appsFilesListAlias2},
		Short:             appsFilesListCmdShortUsage,
		RunE:              contentSpecApplication,
		ValidArgsFunction: getListAppsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE"},
	}

	// removeAppsCmd represents the "applications remove" command
	var removeAppsCmd = &cobra.Command{
		Use:               appsRemoveCmd,
		Aliases:           []string{appsRemoveAlias},
		Short:             appsRemoveCmdShortUsage,
		RunE:              removeFilesApplication,
		ValidArgsFunction: getListAppsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE"},
	}

	// downloadAppsCmd represents the "applications download" command
	var downloadAppsCmd = &cobra.Command{
		Use:               appsDownloadCmd,
		Aliases:           []string{appsDownloadAlias},
		Short:             appsDownloadCmdShortUsage,
		RunE:              downloadFilesApplication,
		ValidArgsFunction: getListAppsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE"},
	}

	// uploadAppsCmd represents the "applications upload" command
	var uploadAppsCmd = &cobra.Command{
		Use:               appsUploadCmd,
		Aliases:           []string{appsUploadAlias},
		Short:             appsUploadCmdShortUsage,
		PreRunE:           initEventApp,
		RunE:              uploadFilesApplication,
		ValidArgsFunction: getListAppsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// buildAppsCmd represents the "applications build" command
	var buildAppsCmd = &cobra.Command{
		Use:               fmt.Sprintf(genericBuildCmd, "app-ID"),
		Aliases:           []string{genericBuildAlias},
		Short:             fmt.Sprintf(genericBuildCmdShortUsage, "application"),
		Long:              fmt.Sprintf(genericBuildCmdLongUsage, "application", "application", "applications"),
		PreRunE:           initEventAppBuild,
		RunE:              buildApplication,
		ValidArgsFunction: getListAppsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// buildsCmd represents the "applications builds" command
	var buildsCmd = &cobra.Command{
		Use:   genericBuildsCmd,
		Short: fmt.Sprintf(genericBuildsCmdShortUsage, "application"),
	}

	// buildsListAppsCmd represents the "applications builds list" command
	var buildsListAppsCmd = &cobra.Command{
		Use:               fmt.Sprintf(genericSubBuildsListCmd, "app-ID"),
		Aliases:           []string{genericSubBuildsListAlias},
		Short:             fmt.Sprintf(genericSubBuildsListCmdShortUsage, "application"),
		RunE:              buildsListApplication,
		ValidArgsFunction: getListAppsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// buildsGetAppsCmd represents the "applications builds get" command
	var buildsGetAppsCmd = &cobra.Command{
		Use:               fmt.Sprintf(genericBuildsGetCmd, "app-ID"),
		Short:             genericBuildsGetCmdShortUsage,
		Long:              fmt.Sprintf(genericBuildsGetCmdLongUsage, "application", "application", "app"),
		RunE:              appBuildsGetAndDisplay,
		ValidArgsFunction: getListAppsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// buildsCancelAppsCmd represents the "applications builds cancel" command
	var buildsCancelAppsCmd = &cobra.Command{
		Use:               fmt.Sprintf(genericBuildsCancelCmd, "app-ID"),
		Short:             genericBuildsCancelCmdShortUsage,
		Long:              fmt.Sprintf(genericBuildsCancelCmdLongUsage, "application", "application", "app"),
		RunE:              buildCancelApplication,
		ValidArgsFunction: getListAppsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// buildsLogsAppsCmd represents the "applications builds logs" command
	var buildsLogsAppsCmd = &cobra.Command{
		Use:               fmt.Sprintf(genericBuildLogsCmd, "app-ID"),
		Short:             fmt.Sprintf(genericBuildLogsCmdShortUsage, "application"),
		Long:              fmt.Sprintf(genericBuildLogsCmdLongUsage, "application", "application", "application", "app"),
		RunE:              buildGetLogsApplication,
		ValidArgsFunction: getListAppsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// buildsStatsCmd represents the "applications builds stats" command
	var buildsStatsCmd = &cobra.Command{
		Use:               appsBuildsStatsCmd,
		Aliases:           []string{appsBuildsStatsAlias},
		Short:             appsBuildsStatsCmdShortUsage,
		SilenceUsage:      true,
		RunE:              getAppBuildsStats,
		ValidArgsFunction: getListAppsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	var seriousCmd = &cobra.Command{
		Use:          "serious",
		SilenceUsage: true,
		Hidden:       true,
		RunE:         seriousCmdHandler,
	}

	// testAppsCmd represents the "applications test" command
	var testAppsCmd = &cobra.Command{
		Use:               fmt.Sprintf(genericTestCmd, "app-ID"),
		Aliases:           []string{genericTestAlias},
		Short:             fmt.Sprintf(genericTestCmdShortUsage, "application"),
		Long:              fmt.Sprintf(genericTestCmdLongUsage, "application", "application", "application", "--force-app"),
		PreRunE:           initEventTests,
		RunE:              testGeneric,
		ValidArgsFunction: getListAppsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// testsCmd represents the "applications tests" command
	var testsCmd = &cobra.Command{
		Use:   genericTestsCmd,
		Short: fmt.Sprintf(genericTestsCmdShortUsage, "application"),
	}

	// testsListAppsCmd represents the "applications tests list" command
	var testsListAppsCmd = &cobra.Command{
		Use:               fmt.Sprintf(genericTestsListCmd, "app-ID"),
		Aliases:           []string{genericTestsListAlias},
		Short:             fmt.Sprintf(genericTestsListCmdShortUsage, "application"),
		RunE:              testsListGeneric,
		ValidArgsFunction: getListAppsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// testsGetAppsCmd represents the "applications tests get" command
	var testsGetAppsCmd = &cobra.Command{
		Use:               fmt.Sprintf(genericTestsGetCmd, "app-ID"),
		Short:             genericTestsGetCmdShortUsage,
		Long:              fmt.Sprintf(genericTestsGetCmdLongUsage, "application", "application", "--force-app"),
		RunE:              testsGetGeneric,
		ValidArgsFunction: getListAppsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// testsLogsAppsCmd represents the "applications tests logs" command
	var testsLogsAppsCmd = &cobra.Command{
		Use:               fmt.Sprintf(genericTestLogsCmd, "app-ID"),
		Short:             fmt.Sprintf(genericTestLogsCmdShortUsage, "application"),
		Long:              fmt.Sprintf(genericTestLogsCmdLongUsage, "application", "application", "application", "--force-app"),
		RunE:              testGetLogsGeneric,
		ValidArgsFunction: getListAppsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// testCancelAppsCmd represents the "applications tests cancel" command
	var testsCancelAppsCmd = &cobra.Command{
		Use:               fmt.Sprintf(genericTestCancelCmd, "app-ID"),
		Short:             genericTestCancelCmdShortUsage,
		Long:              fmt.Sprintf(genericTestCancelCmdLongUsage, "application", "application", "--force-app"),
		RunE:              testCancelGeneric,
		ValidArgsFunction: getListAppsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// testsStatsCmd represents the "applications tests stats" command
	var testsStatsCmd = &cobra.Command{
		Use:               fmt.Sprintf(genericTestsStatsCmd, "app-ID"),
		Aliases:           []string{genericTestsStatsAlias},
		Short:             fmt.Sprintf(genericTestsStatsCmdShortUsage, "application"),
		SilenceUsage:      true,
		RunE:              getAppTestsStats,
		ValidArgsFunction: getListAppsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// servicesCmd represents the "applications services" command
	var servicesCmd = &cobra.Command{
		Use:   appsServicesCmd,
		Short: appsServicesCmdShortDesc,
		RunE:  rpServices,
	}

	// auditAppsCmd represents the "applications audit" command
	var auditAppsCmd = &cobra.Command{
		Use:               appsAuditCmd,
		Aliases:           []string{appsAuditAlias},
		Short:             appsAuditCmdShortUsage,
		PreRunE:           initEventAuditApp,
		RunE:              auditApplication,
		ValidArgsFunction: getListAppsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// auditsCmd represents the "applications audits" command
	var auditsCmd = &cobra.Command{
		Use:   appsAuditsCmd,
		Short: appsAuditsCmdShortUsage,
	}

	// auditsListToolsCmd represents the "applications audits tools" command
	var auditsListToolsCmd = &cobra.Command{
		Use:     appsAuditsListToolsCmd,
		Aliases: []string{appsAuditsListToolsAlias},
		Short:   appsAuditsListToolsCmdShortUsage,
		RunE:    auditsListTools,
	}

	// auditsListCmd represents the "applications audits list" command
	var auditsListCmd = &cobra.Command{
		Use:               appsAuditsListCmd,
		Aliases:           []string{appsAuditsListAlias},
		Short:             appsAuditsListCmdShortUsage,
		RunE:              auditsList,
		ValidArgsFunction: getListAppsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// auditsGetCmd represents the "applications audits get" command
	var auditsGetCmd = &cobra.Command{
		Use:               appsAuditsGetCmd,
		Short:             appsAuditsGetCmdShortUsage,
		Long:              appsAuditsGetCmdLongUsage,
		RunE:              auditsGetAndDisplay,
		ValidArgsFunction: getListAppsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// auditsLogsCmd represents the "applications audits logs" command
	var auditsLogsCmd = &cobra.Command{
		Use:         appsAuditsLogsCmd,
		Short:       appsAuditsLogsCmdShortUsage,
		Long:        appsAuditsLogsCmdLongUsage,
		RunE:        auditGetLogs,
		Annotations: map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// branchCmd represents the "applications branch" command
	var branchCmd = &cobra.Command{
		Use:               appsBranchCmd,
		Short:             appsBranchCmdShortDesc,
		Long:              appsBranchCmdLongDesc,
		RunE:              appCreateOrSwitchBranch,
		ValidArgsFunction: getListAppsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// branchesCmd represents the "applications branches" command
	var branchesCmd = &cobra.Command{
		Use:   appsBranchesCmd,
		Short: appsBranchesCmdShortUsage,
	}

	// branchesListCmd represents the "applications branches list" command
	var branchesListCmd = &cobra.Command{
		Use:               appsBranchesListCmd,
		Aliases:           []string{appsBranchesListAlias},
		Short:             appsBranchesListCmdShortUsage,
		RunE:              appBranchesList,
		ValidArgsFunction: getListAppsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// branchesDeleteCmd represents the "applications branches delete" command
	var branchesDeleteCmd = &cobra.Command{
		Use:               appsBranchesDelCmd,
		Aliases:           []string{appsBranchesDelCmdAlias},
		Short:             appsBranchesDelCmdShortUsage,
		RunE:              appBranchesDelete,
		ValidArgsFunction: getListAppsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// copyCmd represents the "applications copy" command
	var copyAppCmd = &cobra.Command{
		Use:               appsCopyCmd,
		Aliases:           []string{appsCopyAlias},
		Short:             appsCopyCmdShortDesc,
		Long:              appsCopyCmdLongDesc,
		RunE:              appCopyToProject,
		ValidArgsFunction: getListAppsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// refreshAppCmd represents the "applications refresh-files-pkg" command
	var refreshAppCmd = &cobra.Command{
		Use:               appsRefreshFilesPkgCmd,
		Aliases:           []string{appsRefreshFilesPkgAlias},
		Short:             appsRefreshFilesPkgShortDesc,
		PreRunE:           initEventApp,
		RunE:              appRefreshGitpkgFiles,
		ValidArgsFunction: getListAppsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// Here you will define your flags and configuration settings.

	// Persistent Flags -----------------------------------------
	// ----------------------------------------------------------

	// Local Flags ----------------------------------------------
	// Define the flags specific to "list" command
	listAppsCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	listAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	listAppsCmd.Flags().BoolP(appsListCmdBookmarkedFlag, appsListCmdBookmarkedFlagAlias, appsListCmdBookmarkedFlagDefault, appsListCmdBookmarkedFlagUsage)
	// Define the flags specific to "get" command
	getAppsCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	getAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	getAppsCmd.Flags().StringSlice(appsListCmdTypeFlag, []string{"standard"}, appsListCmdTypeFlagUsage)
	// Define the flags specific to "build" command
	buildAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	buildAppsCmd.Flags().StringSliceP(appsBuildCmdDistroFlag, appsBuildCmdDistroFlagAlias, []string{}, appsBuildCmdDistroFlagUsage)
	buildAppsCmd.Flags().StringP(genericBuildCmdCommentFlag, genericBuildCmdCommentFlagAlias, "", genericBuildCmdCommentFlagUsage)
	buildAppsCmd.Flags().BoolP(appsBuildCmdSrpmFlag, appsBuildCmdSrpmFlagAlias, false, appsBuildCmdSrpmFlagUsage)
	buildAppsCmd.Flags().BoolP(appsBuildCmdCheckDepsFlag, appsBuildCmdCheckDepsFlagAlias, false, appsBuildCmdCheckDepsFlagUsage)
	// Define the flags specific to "test" command
	testAppsCmd.Flags().BoolP(appsForceAppFlag, appsForceAppFlagAlias, false, appsForceAppFlagUsage)
	testAppsCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	testAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	testAppsCmd.Flags().StringP(genericTestCmdCommentFlag, genericTestCmdCommentFlagAlias, "", genericTestCmdCommentFlagUsage)
	// Define the flags specific to "builds list" command
	buildsListAppsCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	buildsListAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	buildsListAppsCmd.Flags().BoolP(appsSubBuildsListCmdSRpmFlag, appsSubBuildsListCmdSRpmFlagAlias, false, appsSubBuildsListCmdSRpmFlagUsage)
	// Define the flags specific to "builds get" command
	buildsGetAppsCmd.Flags().BoolP(appsForceAppFlag, appsForceAppFlagAlias, false, appsForceAppFlagUsage)
	buildsGetAppsCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	buildsGetAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	// Define the flags specific to "builds logs" command
	buildsLogsAppsCmd.Flags().BoolP(appsForceAppFlag, appsForceAppFlagAlias, false, appsForceAppFlagUsage)
	buildsLogsAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	buildsLogsAppsCmd.Flags().StringP(genericBuildLogsDirFlag, genericBuildLogsDirFlagAlias, genericBuildLogsDirFlagDefault, genericBuildLogsDirFlagUsage)
	buildsLogsAppsCmd.Flags().BoolP(genericBuildZipFlag, genericBuildZipFlagAlias, false, genericBuildZipFlagUsage)
	// Define the flags specific to "builds cancel" command
	buildsCancelAppsCmd.Flags().BoolP(appsForceAppFlag, appsForceAppFlagAlias, false, appsForceAppFlagUsage)
	buildsCancelAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	// Define the flags specific to "builds stats" command
	buildsStatsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	buildsStatsCmd.Flags().Int(appsBuildsStatsCmdLastFlag, 0, appsBuildsStatsCmdLastFlagUsage)
	buildsStatsCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	// Define the flags specific to "tests list" command
	testsListAppsCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	testsListAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	// Define the flags specific to "tests get" command
	testsGetAppsCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	testsGetAppsCmd.Flags().BoolP(appsForceAppFlag, appsForceAppFlagAlias, false, appsForceAppFlagUsage)
	testsGetAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	// Define the flags specific to "tests logs" command
	testsLogsAppsCmd.Flags().BoolP(appsForceAppFlag, appsForceAppFlagAlias, false, appsForceAppFlagUsage)
	testsLogsAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	testsLogsAppsCmd.Flags().StringP(genericBuildLogsDirFlag, genericBuildLogsDirFlagAlias, genericBuildLogsDirFlagDefault, genericBuildLogsDirFlagUsage)
	testsLogsAppsCmd.Flags().BoolP(genericBuildZipFlag, genericBuildZipFlagAlias, false, genericBuildZipFlagUsage)
	// Define the flags specific to "tests cancel" command
	testsCancelAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	testsCancelAppsCmd.Flags().BoolP(appsForceAppFlag, appsForceAppFlagAlias, false, appsForceAppFlagUsage)
	// Define the flags specific to "tests stats" command
	testsStatsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	testsStatsCmd.Flags().Int(genericTestsStatsCmdLastFlag, 0, genericTestsStatsCmdLastFlagUsage)
	testsStatsCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	// Define the flags specific to "delete" command
	delAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	// Define the flags specific to "bookmark" command
	bookmarkAppCmd.Flags().BoolP(appsBookmarkCmdResetFlag, appsBookmarkCmdResetFlagAlias, false, appsBookmarkCmdResetFlagUsage)
	bookmarkAppCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)

	// Slice flags needed for application list
	listAppsCmd.Flags().StringSlice(appsListCmdTypeFlag, []string{"standard"}, appsListCmdTypeFlagUsage)

	// Slice flags needed for application creation/update
	// Creation
	addAppsCmd.Flags().StringSlice(appsAddCmdMandDistFlag, []string{}, appsAddCmdMandDistFlagUsage)
	addAppsCmd.Flags().StringSlice(appsAddCmdOptDistFlag, []string{}, appsAddCmdOptDistFlagUsage)
	// Update
	updateAppsCmd.Flags().StringSlice(appsAddCmdMandDistFlag, []string{flagUpdateKeepCurrentValue}, appsAddCmdMandDistFlagUsage)
	updateAppsCmd.Flags().StringSlice(appsAddCmdOptDistFlag, []string{flagUpdateKeepCurrentValue}, appsAddCmdOptDistFlagUsage)

	// Other flags needed for application creation/update
	// Creation
	addAppsCmd.Flags().StringP(appsAddCmdNameFlag, appsAddCmdNameFlagAlias, "", appsAddCmdNameFlagUsage)
	addAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsAddCmdPrjFlagUsage)
	addAppsCmd.Flags().StringP(appsAddCmdAliasFlag, appsAddCmdAliasFlagAlias, "", appsAddCmdAliasFlagUsage)
	addAppsCmd.Flags().StringP(appsAddCmdDescFlag, appsAddCmdDescFlagAlias, "", appsAddCmdDescFlagUsage)
	addAppsCmd.Flags().String(appsAddCmdPkgFlag, "", appsAddCmdPkgFlagUsage)
	addAppsCmd.Flags().String(appsAddCmdSrcURLFlag, "", appsAddCmdSrcURLFlagUsage)
	addAppsCmd.Flags().String(appsAddCmdSrcRevFlag, "", appsAddCmdSrcRevFlagUsage)
	addAppsCmd.Flags().String(appsAddCmdSpecTypeFlag, string(rpapiv1.RPAppSpecTypeInternal), appsAddCmdSpecTypeFlagUsage)
	addAppsCmd.Flags().String(appsAddCmdSpecLocationFlag, "", appsAddCmdSpecLocationFlagUsage)
	addAppsCmd.Flags().Bool(appsAddCmdAutoTestFlag, false, appsAddCmdAutoTestFlagUsage)
	addAppsCmd.Flags().StringP(appsAddCmdBuildTypeFlag, appsAddCmdBuildTypeFlagAlias, appsAddCmdBuildTypeFlagDefault, appsAddCmdBuildTypeFlagUsage)
	addAppsCmd.Flags().Bool(appsAddCmdAutoAuditFlag, false, appsAddCmdAutoAuditFlagUsage)
	addAppsCmd.Flags().StringSlice(appsAddCmdDefAuditFlag, []string{}, appsAddCmdDefAuditFlagUsage)
	addAppsCmd.Flags().StringP(appsAddCmdActiveBranchFlag, appsAddCmdActiveBranchFlagAlias, "", appsAddCmdActiveBranchFlagUsage)
	addAppsCmd.Flags().Bool(appsAddCmdIgnoreCheckDepsResultFlag, false, appsAddCmdIgnoreCheckDepsResultFlagUsage)
	addAppsCmd.Flags().Bool(appsAddCmdIgnoreMissingRedTestRpmFlag, false, appsAddCmdIgnoreMissingRedTestRpmFlagUsage)

	// App auto srpm rebuilds
	addAppsCmd.Flags().StringSlice(appsAddCmdAutoSRpmBuildFlag, []string{}, appsAddCmdAutoSRpmBuildFlagUsage)
	// Set/unset all auto srpm build options
	addAppsCmd.Flags().String(appsAddCmdAutoSRpmBuildAllFlag, "true", appsAddCmdAutoSRpmBuildAllFlagUsage)

	// App services
	addAppsCmd.Flags().StringArrayP(appsAddCmdServiceFlag, appsAddCmdServicesFlagAlias, []string{}, appsAddCmdServicesFlagUsage)

	seriousCmd.Flags().Bool("why", false, "")
	// Creation from sample
	addAppsFromSampleCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsAddCmdPrjFlagUsage)
	// Update
	updateAppsCmd.Flags().StringP(appsAddCmdNameFlag, appsAddCmdNameFlagAlias, flagUpdateKeepCurrentValue, appsAddCmdNameFlagUsage)
	updateAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	updateAppsCmd.Flags().StringP(appsAddCmdDescFlag, appsAddCmdDescFlagAlias, flagUpdateKeepCurrentValue, appsAddCmdDescFlagUsage)
	updateAppsCmd.Flags().String(appsAddCmdSrcRevFlag, flagUpdateKeepCurrentValue, appsAddCmdSrcRevFlagUsage)
	updateAppsCmd.Flags().String(appsAddCmdSrcURLFlag, flagUpdateKeepCurrentValue, appsUpdateCmdSrcURLFlagUsage)
	updateAppsCmd.Flags().String(appsAddCmdSpecTypeFlag, flagUpdateKeepCurrentValue, appsAddCmdSpecTypeFlagUsage)
	updateAppsCmd.Flags().String(appsAddCmdSpecLocationFlag, flagUpdateKeepCurrentValue, appsAddCmdSpecLocationFlagUsage)
	updateAppsCmd.Flags().String(appsAddCmdAutoTestFlag, flagUpdateKeepCurrentValue, appsAddCmdAutoTestFlagUsage)
	updateAppsCmd.Flags().StringP(appsAddCmdBuildTypeFlag, appsAddCmdBuildTypeFlagAlias, flagUpdateKeepCurrentValue, appsAddCmdBuildTypeFlagUsage)
	updateAppsCmd.Flags().String(appsAddCmdAutoAuditFlag, flagUpdateKeepCurrentValue, appsAddCmdAutoAuditFlagUsage)
	updateAppsCmd.Flags().StringSlice(appsAddCmdDefAuditFlag, []string{flagUpdateKeepCurrentValue}, appsAddCmdDefAuditFlagUsage)
	updateAppsCmd.Flags().String(appsAddCmdIgnoreCheckDepsResultFlag, flagUpdateKeepCurrentValue, appsAddCmdIgnoreCheckDepsResultFlagUsage)
	updateAppsCmd.Flags().String(appsAddCmdIgnoreMissingRedTestRpmFlag, flagUpdateKeepCurrentValue, appsAddCmdIgnoreMissingRedTestRpmFlagUsage)
	updateAppsCmd.Flags().StringP(appsAddCmdActiveBranchFlag, appsAddCmdActiveBranchFlagAlias, flagUpdateKeepCurrentValue, appsUpdateCmdActiveBranchFlagUsage)

	// Update app services
	updateAppsCmd.Flags().StringArrayP(appsAddCmdServiceFlag, appsAddCmdServicesFlagAlias, []string{flagUpdateKeepCurrentValue}, appsAddCmdServicesFlagUsage)
	updateAppsCmd.Flags().StringSlice(appsUpdateCmdDisableServiceFlag, []string{}, appsUpdateCmdDisableServicesFlagUsage)

	// Update app auto srpm rebuilds
	updateAppsCmd.Flags().StringSlice(appsAddCmdAutoSRpmBuildFlag, []string{}, appsAddCmdAutoSRpmBuildFlagUsage)
	// Update app all auto srpm rebuilds
	updateAppsCmd.Flags().String(appsAddCmdAutoSRpmBuildAllFlag, "", appsAddCmdAutoSRpmBuildAllFlagUsage)

	// Hidden updates flags
	updateAppsCmd.Flags().Bool(appsForceStateFlag, false, appsForceStateFlagUsage)
	updateAppsCmd.Flags().MarkHidden(appsForceStateFlag)

	// Flag for content list
	contentAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	contentAppsCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)

	// Flag for remove file
	removeAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)

	// Flag for download file
	downloadAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)

	// Flag needed for spec file upload
	uploadAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)

	// Flag for the test command
	testAppsCmd.Flags().StringP(appsTestCmdImageFlag, appsTestCmdImageFlagAlias, "", appsTestCmdImageFlagUsage)
	testAppsCmd.Flags().StringP(genericTestCmdBoardModelFlag, genericTestCmdBoardModelFlagAlias, "", genericTestCmdBoardModelFlagUsage)
	testAppsCmd.Flags().Int(genericTestCmdTimeOutFlag, 5, genericTestCmdTimeOutFlagUsage)
	testAppsCmd.Flags().Bool(genericTestCmdProgressFlag, genericTestCmdProgressFlagDefault, genericTestCmdProgressFlagUsage)

	// Flag for the available services command
	servicesCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)

	// Flag for the audit command
	auditAppsCmd.Flags().StringSlice(appsAuditToolFlag, []string{}, appsAuditToolFlagUsage)
	auditAppsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)

	// Flag for the audits tools list command
	auditsListToolsCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)

	// Flag for the audits list command
	auditsListCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	auditsListCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)

	// Flags for the audits get command
	auditsGetCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	auditsGetCmd.Flags().BoolP(appsForceAppFlag, appsForceAppFlagAlias, false, appsForceAppFlagUsage)
	auditsGetCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)

	// Flags for the audits logs command
	auditsLogsCmd.Flags().StringP(genericBuildLogsDirFlag, genericBuildLogsDirFlagAlias, genericBuildLogsDirFlagDefault, genericBuildLogsDirFlagUsage)
	auditsLogsCmd.Flags().BoolP(appsForceAppFlag, appsForceAppFlagAlias, false, appsForceAppFlagUsage)
	auditsLogsCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)

	// Flags for the branch command
	branchCmd.Flags().StringP(appsBranchSourceFlag, appsBranchSourceFlagAlias, "", appsBranchSourceFlagUsage)
	branchCmd.Flags().StringP(appsBranchDestinationFlag, appsBranchDestinationFlagAlias, "", appsBranchDestinationFlagUsage)
	branchCmd.Flags().Bool(appsBranchDefaultFlag, false, appsBranchDefaultFlagUsage)
	branchCmd.Flags().Bool(appsBranchNotActiveFlag, false, appsBranchNotActiveFlagUsage)
	branchCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	branchCmd.Flags().StringSlice(appsBranchDistrosFlag, []string{}, appsBranchDistrosFlagUsage)
	branchCmd.Flags().BoolP(appsBranchResetDistrosFlag, appsBranchResetDistrosFlagAlias, false, appsBranchResetDistrosFlagUsage)

	// Flags for the branch list command
	branchesListCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	branchesListCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)

	// Flags for the branch delete command
	branchesDeleteCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	branchesDeleteCmd.Flags().StringP(appsBranchFlag, appsBranchFlagAlias, "", appsBranchFlagUsage)

	// Flags for the copy app command
	copyAppCmd.Flags().StringP(appCopyCmdDestinationPrjFlag, appCopyCmdDestinationPrjAlias, "", appCopyCmdDestinationPrjFlagUsage)
	copyAppCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)

	// Flags for the refresh gitpkgs files command
	refreshAppCmd.Flags().StringP(appsCmdPrjFlag, appsCmdPrjFlagAlias, appsCmdPrjFlagDefault, appsCmdPrjFlagUsage)
	// ----------------------------------------------------------

	// Add some flags as required
	addAppsCmd.MarkFlagRequired(appsAddCmdPkgFlag)
	addAppsCmd.MarkFlagRequired(appsAddCmdNameFlag)

	// Specify dynamic flag completion
	addAppsCmd.RegisterFlagCompletionFunc(appsCmdPrjFlag, getListPrjsAsTab)
	addAppsFromSampleCmd.RegisterFlagCompletionFunc(appsCmdPrjFlag, getListPrjsAsTab)
	listAppsCmd.RegisterFlagCompletionFunc(appsListCmdTypeFlag, getListTypeAsTab)
	listAppsCmd.RegisterFlagCompletionFunc(appsCmdPrjFlag, getFullListPrjsAsTab)
	updateAppsCmd.RegisterFlagCompletionFunc(appsCmdPrjFlag, getListPrjsAsTab)
	contentAppsCmd.RegisterFlagCompletionFunc(appsCmdPrjFlag, getListPrjsAsTab)
	uploadAppsCmd.RegisterFlagCompletionFunc(appsCmdPrjFlag, getListPrjsAsTab)
	getAppsCmd.RegisterFlagCompletionFunc(appsCmdPrjFlag, getFullListPrjsAsTab)
	getAppsCmd.RegisterFlagCompletionFunc(appsListCmdTypeFlag, getListTypeAsTab)
	buildAppsCmd.RegisterFlagCompletionFunc(appsCmdPrjFlag, getListPrjsAsTab)
	buildAppsCmd.RegisterFlagCompletionFunc(appsBuildCmdDistroFlag, getListDistAsTab)
	buildsStatsCmd.RegisterFlagCompletionFunc(appsCmdPrjFlag, getListPrjsAsTab)
	testAppsCmd.RegisterFlagCompletionFunc(appsCmdPrjFlag, getListPrjsAsTab)
	testAppsCmd.RegisterFlagCompletionFunc(appsTestCmdImageFlag, getListImagesAsTab)
	testAppsCmd.RegisterFlagCompletionFunc(genericTestCmdBoardModelFlag, getListRTMBoardModelsAsTab)
	buildsListAppsCmd.RegisterFlagCompletionFunc(appsCmdPrjFlag, getListPrjsAsTab)
	buildsLogsAppsCmd.RegisterFlagCompletionFunc(appsCmdPrjFlag, getListPrjsAsTab)
	buildsCancelAppsCmd.RegisterFlagCompletionFunc(appsCmdPrjFlag, getListPrjsAsTab)
	testsListAppsCmd.RegisterFlagCompletionFunc(appsCmdPrjFlag, getListPrjsAsTab)
	testsLogsAppsCmd.RegisterFlagCompletionFunc(appsCmdPrjFlag, getListPrjsAsTab)
	testsCancelAppsCmd.RegisterFlagCompletionFunc(appsCmdPrjFlag, getListPrjsAsTab)
	testsStatsCmd.RegisterFlagCompletionFunc(appsCmdPrjFlag, getListPrjsAsTab)
	delAppsCmd.RegisterFlagCompletionFunc(appsCmdPrjFlag, getListPrjsAsTab)
	// Creation
	addAppsCmd.RegisterFlagCompletionFunc(appsAddCmdMandDistFlag, getListDistAsTab)
	addAppsCmd.RegisterFlagCompletionFunc(appsAddCmdOptDistFlag, getListDistAsTab)
	addAppsCmd.RegisterFlagCompletionFunc(appsAddCmdDefAuditFlag, getAuditToolsAsTab)
	addAppsCmd.RegisterFlagCompletionFunc(appsAddCmdBuildTypeFlag, getBuildTypesAsTab)
	addAppsCmd.RegisterFlagCompletionFunc(appsAddCmdSpecTypeFlag, getSpecTypeList)
	// Add app services
	addAppsCmd.RegisterFlagCompletionFunc(appsAddCmdServiceFlag, getServicesNamesAsTab)
	// Add app srpm auto rebuild appsAddCmdAutoSRpmBuildFlag
	addAppsCmd.RegisterFlagCompletionFunc(appsAddCmdAutoSRpmBuildFlag, getSrpmRebuildOptionsNamesAsTab)

	// Update
	updateAppsCmd.RegisterFlagCompletionFunc(appsAddCmdMandDistFlag, getListDistAsTab)
	updateAppsCmd.RegisterFlagCompletionFunc(appsAddCmdOptDistFlag, getListDistAsTab)
	updateAppsCmd.RegisterFlagCompletionFunc(appsAddCmdDefAuditFlag, getAuditToolsAsTab)
	updateAppsCmd.RegisterFlagCompletionFunc(appsAddCmdBuildTypeFlag, getBuildTypesAsTab)
	updateAppsCmd.RegisterFlagCompletionFunc(appsAddCmdSpecTypeFlag, getSpecTypeList)
	updateAppsCmd.RegisterFlagCompletionFunc(appsAddCmdActiveBranchFlag, getAppBranchesAsTab)
	// Update app services
	updateAppsCmd.RegisterFlagCompletionFunc(appsAddCmdServiceFlag, getServicesNamesAsTab)
	updateAppsCmd.RegisterFlagCompletionFunc(appsUpdateCmdDisableServiceFlag, getServicesNamesAsTab)
	// Update app srpm auto rebuild options
	updateAppsCmd.RegisterFlagCompletionFunc(appsAddCmdAutoSRpmBuildFlag, getSrpmRebuildOptionsNamesAsTab)
	updateAppsCmd.RegisterFlagCompletionFunc(appsAddCmdAutoSRpmBuildAllFlag, getGlobalSrpmRebuildOptionsValuesAsTab)

	// Audit
	auditAppsCmd.RegisterFlagCompletionFunc(appsAuditToolFlag, getAuditToolsAsTab)
	// Branches
	branchCmd.RegisterFlagCompletionFunc(appsBranchSourceFlag, getAppBranchesAsTab)
	branchCmd.RegisterFlagCompletionFunc(appsBranchDestinationFlag, getAppBranchesAsTab)
	branchCmd.RegisterFlagCompletionFunc(appsBranchDistrosFlag, getListDistAsTab)
	branchesDeleteCmd.RegisterFlagCompletionFunc(appsBranchFlag, getAppBranchesAsTab)

	// App Copy
	copyAppCmd.RegisterFlagCompletionFunc(appCopyCmdDestinationPrjFlag, getListPrjsAsTab)
	copyAppCmd.RegisterFlagCompletionFunc(appsCmdPrjFlag, getListPrjsAsTab)

	// App refresh
	refreshAppCmd.RegisterFlagCompletionFunc(appsCmdPrjFlag, getListPrjsAsTab)

	// Add the group commands "applications" as subcommand of rootCmd
	rootCmd.AddCommand(appsCmd)
	rootCmd.AddCommand(seriousCmd)

	// Add the command "list" as a subcommand of appsCmd
	appsCmd.AddCommand(listAppsCmd)
	// Add the command "get" as a subcommand of appsCmd
	appsCmd.AddCommand(getAppsCmd)
	// Add the command "add" as a subcommand of appsCmd
	appsCmd.AddCommand(addAppsCmd)
	// Add the command "add-from-sample" as a subcommand of appsCmd
	appsCmd.AddCommand(addAppsFromSampleCmd)
	// Add the command "update" as a subcommand of appsCmd
	appsCmd.AddCommand(updateAppsCmd)
	// Add the command "del" as a subcommand of appsCmd
	appsCmd.AddCommand(delAppsCmd)
	// Add the command "bookmark" as a subcommand of appsCmd
	appsCmd.AddCommand(bookmarkAppCmd)
	// Add the command "files" as a subcommand of appsCmd
	appsCmd.AddCommand(filesCmd)
	// Add the command "list" as a subcommand of filesCmd
	filesCmd.AddCommand(contentAppsCmd)
	// Add the command "remove" as a subcommand of filesCmd
	filesCmd.AddCommand(removeAppsCmd)
	// Add the command "download" as a subcommand of filesCmd
	filesCmd.AddCommand(downloadAppsCmd)
	// Add the command "upload" as a subcommand of filesCmd
	filesCmd.AddCommand(uploadAppsCmd)

	// Add the command "build" as a subcommand of appsCmd
	appsCmd.AddCommand(buildAppsCmd)
	// Add the command "builds" as a subcommand of appsCmd
	appsCmd.AddCommand(buildsCmd)
	// Add the command "list" as a subcommand of buildsCmd
	buildsCmd.AddCommand(buildsListAppsCmd)
	// Add the command "get" as a subcommand of buildsCmd
	buildsCmd.AddCommand(buildsGetAppsCmd)
	// Add the command "logs" as a subcommand of appsCmd
	buildsCmd.AddCommand(buildsLogsAppsCmd)
	// Add the command "cancel" as a subcommand of appsCmd
	buildsCmd.AddCommand(buildsCancelAppsCmd)
	// Add the command "stats" as a subcommand of appsCmd
	buildsCmd.AddCommand(buildsStatsCmd)

	// Add the command "test" as a subcommand of appsCmd
	appsCmd.AddCommand(testAppsCmd)
	// Add the command "tests" as a subcommand of appsCmd
	appsCmd.AddCommand(testsCmd)
	// Add the command "list" as a subcommand of testsCmd
	testsCmd.AddCommand(testsListAppsCmd)
	// Add the command "get" as a subcommand of testsCmd
	testsCmd.AddCommand(testsGetAppsCmd)
	// Add the command "logs" as a subcommand of testsCmd
	testsCmd.AddCommand(testsLogsAppsCmd)
	// Add the command "cancel" as a subcommand of testsCmd
	testsCmd.AddCommand(testsCancelAppsCmd)
	// Add the command "stats" as a subcommand of testsCmd
	testsCmd.AddCommand(testsStatsCmd)

	// Add the command "services" as a subcommand of appsCmd
	appsCmd.AddCommand(servicesCmd)

	// Add the command "audit" as a subcommand of appsCmd
	appsCmd.AddCommand(auditAppsCmd)
	// Add the command "audits" as a subcommand of appsCmd
	appsCmd.AddCommand(auditsCmd)
	// Add the command "tools" as a subcommand of auditsCmd
	auditsCmd.AddCommand(auditsListToolsCmd)
	// Add the command "list" as a subcommand of auditsCmd
	auditsCmd.AddCommand(auditsListCmd)
	// Add the command "get" as a subcommand of auditsCmd
	auditsCmd.AddCommand(auditsGetCmd)
	// Add the command "logs" as a subcommand of auditsCmd
	auditsCmd.AddCommand(auditsLogsCmd)

	// Add the command "branch" as a subcommand of appsCmd
	appsCmd.AddCommand(branchCmd)
	// Add the command "branches" as a subcommand of appsCmd
	appsCmd.AddCommand(branchesCmd)
	// Add the command "list" as a subcommand of branchesCmd
	branchesCmd.AddCommand(branchesListCmd)
	// Add the command "delete" as a subcommand of branchesCmd
	branchesCmd.AddCommand(branchesDeleteCmd)

	// Add the command "copy" as a subcommand of appsCmd
	appsCmd.AddCommand(copyAppCmd)

	// Add the command "refresh" as a subcommand of appsCmd
	appsCmd.AddCommand(refreshAppCmd)
}

// initEventApp inits everything that is needed for the application add
// command to run
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func initEventApp(cmd *cobra.Command, args []string) error {

	// Init the events list
	eventsList := []string{
		rpapiv1.EvtAppAdd,
		rpapiv1.EvtAppDelete,
		rpapiv1.EvtAppStateChange,
	}

	return CLIRegisterToMultipleEvents(eventsList)
}

// initEventAppBuild inits everything that is needed for the application
// build command
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func initEventAppBuild(cmd *cobra.Command, args []string) error {

	// Init the events list
	eventsList := []string{
		rpapiv1.EvtAppBuildInfo,
		rpapiv1.EvtAppBuildSubTaskInfo,
	}

	return CLIRegisterToMultipleEvents(eventsList)
}

// initEventAuditApp inits everything that is needed for the application audit
// command to run
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func initEventAuditApp(cmd *cobra.Command, args []string) error {

	// Init the events list
	eventsList := []string{
		rpapiv1.EvtAuditAdd,
		rpapiv1.EvtAuditDelete,
		rpapiv1.EvtAuditStateChange,
	}

	return CLIRegisterToMultipleEvents(eventsList)
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Completion functions -----------------------------------------
// --------------------------------------------------------------

// getListAppsAsTab gets all the applications available in the Redpesk
// instance and return it as an array of strings
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getListAppsAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	appsList := []string{}
	typesList := []string{"standard", "released"}

	// Depending if the user is admin, get all types of projects or just "standard" and "released"
	userAdmin, _ := isUserAdminOnCurrentStack()
	if userAdmin {
		typesList = append(typesList, "external")
	}

	// Get applications list
	apps := rpapiv1.RPApplications{}
	// Add a smart filter to avoid asking for all the apps
	filter := fmt.Sprintf("slug~=%s*", toComplete)
	// Request the apps using the filter
	if err := bckAPI.ApplicationsListGetOfTypes(&apps, "", typesList, filter); err != nil {
		// No error logs in autocompletion functions!
		return appsList, cobra.ShellCompDirectiveNoFileComp
	}

	// We got correctly the projects list, fill the prjsList tab
	for _, application := range apps.Apps {
		// Add the slug of the application in the suggestions
		appsList = append(appsList, application.Slug)
	}

	return appsList, cobra.ShellCompDirectiveNoFileComp
}

// getFullListAppsAsTab gets all (even external) the applications available in the Redpesk
// instance and return it as an array of strings
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getFullListAppsAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	appsList := []string{}
	typesList := []string{"standard", "released", "external"}

	// Get applications list
	apps := rpapiv1.RPApplications{}
	// Add a smart filter to avoid asking for all the apps
	filter := fmt.Sprintf("slug~=%s*", toComplete)
	// Request the apps using the filter
	if err := bckAPI.ApplicationsListGetOfTypes(&apps, "", typesList, filter); err != nil {
		// No error logs in autocompletion functions!
		return appsList, cobra.ShellCompDirectiveNoFileComp
	}

	// We got correctly the projects list, fill the prjsList tab
	for _, application := range apps.Apps {
		// Add the slug of the application in the suggestions
		appsList = append(appsList, application.Slug)
	}

	return appsList, cobra.ShellCompDirectiveNoFileComp
}

// getListAppsSamplesAsTab gets all the sample applications available in the Redpesk
// instance and return it as an array of strings
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getListAppsSamplesAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	appsList := []string{}

	// Get sample applications list
	apps := []rpapiv1.RPSampleApplication{}
	if err := bckAPI.ApplicationsSamplesListGet(&apps); err != nil {
		// No error logs in autocompletion functions!
		return appsList, cobra.ShellCompDirectiveNoFileComp
	}

	// We got correctly the apps samples list, fill the prjsList tab
	for _, appSample := range apps {
		// Add the package-name of the application in the suggestions
		appsList = append(appsList, appSample.PackageName)
	}

	return appsList, cobra.ShellCompDirectiveNoFileComp
}

// getSpecTypeList returns the list of possible specfile type
// available. It returns a static array.
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getSpecTypeList(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	return []string{"external_url", "internal", "relative_path"}, cobra.ShellCompDirectiveNoFileComp
}

// getSrpmRebuildOptionsNamesAsTab returns the list of possible auto srpm rebuild options. It returns a static array.
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getSrpmRebuildOptionsNamesAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	return []string{SrpmPackageFileChangeTrueShort, SrpmMetaFileChangeTrueShort, SrpmDistroBranchesManagementTrueShort, SrpmNoValidSrpmBuildTrueShort, SrpmPackageFileChangeFalseShort, SrpmMetaFileChangeFalseShort, SrpmDistroBranchesManagementFalseShort, SrpmNoValidSrpmBuildFalseShort}, cobra.ShellCompDirectiveNoFileComp
}

// getSrpmRebuildOptionsNames returns the list of possible auto srpm rebuild options. It returns a static array.
//
// :returns: []string
func getSrpmRebuildOptionsNames() []string {
	return []string{SrpmPackageFileChangeShort, SrpmMetaFileChangeShort, SrpmDistroBranchesManagementShort, SrpmNoValidSrpmBuildShort}
}

// getGlobalSrpmRebuildOptionsValuesAsTab returns the list of possible global auto srpm rebuild options. It returns a static array.
//
// :returns: []string
func getGlobalSrpmRebuildOptionsValuesAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	return []string{GlobalAutoSrpmRebuildTrue, GlobalAutoSrpmRebuildFalse}, cobra.ShellCompDirectiveNoFileComp
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Command functions --------------------------------------------
// --------------------------------------------------------------

// listApplications gets all the applications available in the Redpesk
// instance and prints it. If a project is specified, only the applications
// from this project is retrieved.
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func listApplications(cmd *cobra.Command, args []string) error {

	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Get the project flag
	projectID, err := cmd.Flags().GetString(appsCmdPrjFlag)
	if err != nil {
		logFlagError(appsCmdPrjFlag)
		return err
	}

	// Get the bookmarked flag
	bookmarkOnly, err := cmd.Flags().GetBool(appsListCmdBookmarkedFlag)
	if err != nil {
		logFlagError(appsListCmdBookmarkedFlag)
		return err
	}

	// Get the type flag
	typesSlice, err := getAppTypesFromFlag(cmd)
	if err != nil {
		return err
	}

	var prjUUIDStr string
	if projectID != "" {
		// Retrieve the true project
		project, err := getProjectFromIdentifier(projectID, false)
		if err != nil {
			return err
		}
		prjUUIDStr = project.ID.String()
	} else {
		prjUUIDStr = ""
	}

	// Prepare the function needed by the printer to get Applications
	getAppsCB := func(pageNb, perPage int) (interface{}, printer.RPPrinterPaging, error) {
		// Get applications list
		apps := rpapiv1.RPApplications{}
		if err := bckAPI.ApplicationsListByTypePage(&apps, prjUUIDStr, typesSlice, pageNb, perPage, bookmarkOnly); err != nil {
			return apps.Apps, printer.RPPrinterPaging{}, err
		}
		// Get the paging interface from the structure retrieved
		pagingStruct, _ := printer.GetPagingFromInterface(&apps)

		return apps.Apps, pagingStruct, nil
	}

	// Prepare the structure to manage paging
	pagingStruct := printer.RPPagingToolsStruct{
		Verbose:       verbose,
		VerboseLineNb: appDisplayVerboseLightLineNb,
		GetObjCB:      getAppsCB,
		PrintFancyCB:  displayApps,
	}

	return RPPrinter.InteractivelyPrintPages(&pagingStruct, viper.GetBool(NoHeaderFlag))
}

// getApplication gets the application information available in the Redpesk
// instance and prints it
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func getApplication(cmd *cobra.Command, args []string) error {

	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Get the application that we are looking for
	appStructure, err := genericGetAppFromArgs(cmd, args, true)
	if err != nil {
		return err
	}

	// Display the application obtained
	err = RPPrinter.CustomPrintObject(appStructure, verbose, displayOneAppFull)

	return err
}

// buildsListApplication gets all the builds available for the given application
// and prints it.
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func buildsListApplication(cmd *cobra.Command, args []string) error {
	return buildsList(cmd, args, "application")
}

// buildGetLogsApplication allows to retrieve logs from an application build.
// If the build ID is not specified, the logs of the latest build are retrieved
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func buildGetLogsApplication(cmd *cobra.Command, args []string) error {

	RPPrinter.CustomPrintLn(genericBuildLogsCmdAnsLine0)

	// Retrieve the build ID from the args given by the user
	buildID, err := getBuildIDFromArgs(cmd, args, appsForceAppFlag, "application", getAppUUIDCallBack)
	if err != nil {
		return err
	}

	return buildGetLogs(cmd, buildID, "application")
}

// buildCancelApplication allows to cancel a build.
// If the build ID is not specified, the latest build is cancelled
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func buildCancelApplication(cmd *cobra.Command, args []string) error {

	// Retrieve the build ID from the args given by the user
	buildID, err := getBuildIDFromArgs(cmd, args, appsForceAppFlag, "application", getAppUUIDCallBack)
	if err != nil {
		return err
	}

	return buildCancel(buildID)
}

// getAppBuildsStats display the builds statistics for the given
// application
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func getAppBuildsStats(cmd *cobra.Command, args []string) error {
	// Get the application that we are looking for
	appStructure, err := genericGetAppFromArgs(cmd, args, false)
	if err != nil {
		return err
	}

	// Now request and display builds stats for this project
	return getAndDisplayBuildsStats(cmd, appStructure.Parent.ProjectID, &appStructure)
}

// getAppTestsStats display the tests statistics for the given
// application
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func getAppTestsStats(cmd *cobra.Command, args []string) error {
	// Get the application that we are looking for
	appStructure, err := genericGetAppFromArgs(cmd, args, false)
	if err != nil {
		return err
	}

	// Now request and display builds stats for this project
	return getAndDisplayTestsStats(cmd, appStructure.Parent.ProjectID, appStructure.ID, uuid.Nil)
}

// uploadFilesApplication uploads files for the given application
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func uploadFilesApplication(cmd *cobra.Command, args []string) error {

	// Communication to user
	RPPrinter.CustomPrintLn(appsUploadCmdAnsLine0)

	// Retrieve the application from UUID or its slug
	app, err := genericGetAppFromArgs(cmd, args, false)
	if err != nil {
		return err
	}

	// Get the path to the specfile
	files := []string{}
	if len(args) > 1 {
		files = args[1:]
	}

	// Uploading files
	return uploadFileUtility(files, app.ID, false)
}

// contentSpecApplication content of the given application
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func contentSpecApplication(cmd *cobra.Command, args []string) error {

	// Retrieve the application from UUID or its slug
	app, errargs := genericGetAppFromArgs(cmd, args, false)
	if errargs != nil {
		return errargs
	}

	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// ask for the content (the list of files)
	content := rpapiv1.RPFileList{}
	errreq := bckAPI.ApplicationGetListFiles(app.ID.String(), &content)
	if errreq != nil {
		return errreq
	}

	// display the content
	errprt := RPPrinter.CustomPrintObject(content, verbose, displayContent)
	return errprt
}

// displayContent display the list of files
//
// :param content: List of services to display
// :type content: interface{}
// :param verbose: Tell if the print needs to be verbose or not
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func displayContent(content interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Convert to the right type
	list := content.(rpapiv1.RPFileList)

	// Display each service
	for _, item := range list.ListFiles {
		if verbose {
			fmt.Fprintf(writer, "%s\t(%d bytes)\n", item.Name, item.Size)
		} else {
			fmt.Fprintln(writer, item.Name)
		}
	}
	return nil
}

// removeFilesApplication content of the given application
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func removeFilesApplication(cmd *cobra.Command, args []string) error {

	// Retrieve the application from UUID or its slug
	app, errargs := genericGetAppFromArgs(cmd, args, false)
	if errargs != nil {
		return errargs
	}

	RPPrinter.CustomPrintLn(appsRemoveCmdAnsLine0)
	for _, name := range args[1:] {
		// Communication to user
		RPPrinter.CustomPrint(appsRemoveCmdAnsLine1, name, " ")
		err := bckAPI.ApplicationRemoveFile(app.ID.String(), name)
		if err != nil {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
			Log.Error("error when removing ", name, ": ", err)
		} else {
			RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
		}
	}
	return nil
}

// downloadApplication content of the given application
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func downloadFilesApplication(cmd *cobra.Command, args []string) error {

	// Retrieve the application from UUID or its slug
	app, errargs := genericGetAppFromArgs(cmd, args, false)
	if errargs != nil {
		return errargs
	}

	files := []string{}
	if len(args) > 1 {
		files = args[1:]
	} else {
		// ask for the content (the list of files)
		content := rpapiv1.RPFileList{}
		errreq := bckAPI.ApplicationGetListFiles(app.ID.String(), &content)
		if errreq != nil {
			return errreq
		}
		for _, item := range content.ListFiles {
			files = append(files, item.Name)
		}
	}

	return downloadGitPkgsAppFiles(app.ID, "", files, true)
}

// prepareAppStruct prepare a application structure to send to the backend
// to create or update an application. It returns the local path to a local specfile
// to upload if it is necessary
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param app: Application structure to fill
// :type app: *rpapiv1.RPApplicationFull
// :param parentPrj: Parent project of the application
// :type parentPrj: *rpapiv1.RPProjectFull
// :param isCreation: Indicate if we are creating the application
// :type isCreation: bool
// :returns: string, error -- the path to a local specfile to upload if needed and an error if a problem occurs
func prepareAppStruct(cmd *cobra.Command, app *rpapiv1.RPApplicationFull, parentPrj *rpapiv1.RPProjectFull, isCreation bool) (string, error) {

	// Get all the attributes available only during creation
	if isCreation {
		// Get the alias of the application wanted by the user (optional)
		appAlias, err := cmd.Flags().GetString(appsAddCmdAliasFlag)
		if err != nil {
			logFlagError(appsAddCmdAliasFlag)
			return "", err
		}

		// Get the package name wanted by the user
		appPkgName, err := cmd.Flags().GetString(appsAddCmdPkgFlag)
		if err != nil {
			logFlagError(appsAddCmdPkgFlag)
			return "", err
		}

		// Get the optional active branch to set
		activeBranch, err := cmd.Flags().GetString(appsAddCmdActiveBranchFlag)
		if err != nil {
			logFlagError(appsAddCmdActiveBranchFlag)
			return "", err
		}

		// Set the attributes
		app.Slug = appAlias
		app.PackageName = appPkgName
		app.PackageActiveBranch = activeBranch
	}

	// Get the name of the application to create
	appName, err := cmd.Flags().GetString(appsAddCmdNameFlag)
	if err != nil {
		logFlagError(appsAddCmdNameFlag)
		return "", err
	}

	// Get the description of the application wanted by the user (optional)
	appDesc, err := cmd.Flags().GetString(appsAddCmdDescFlag)
	if err != nil {
		logFlagError(appsAddCmdDescFlag)
		return "", err
	}

	// Get the source url wanted by the user
	appSrcURL, err := cmd.Flags().GetString(appsAddCmdSrcURLFlag)
	if err != nil {
		logFlagError(appsAddCmdSrcURLFlag)
		return "", err
	}

	// Get the source revision wanted by the user
	appSrcRev, err := cmd.Flags().GetString(appsAddCmdSrcRevFlag)
	if err != nil {
		logFlagError(appsAddCmdSrcRevFlag)
		return "", err
	}

	// Get the build type wanted by the user
	buildType, err := cmd.Flags().GetString(appsAddCmdBuildTypeFlag)
	if err != nil {
		logFlagError(appsAddCmdBuildTypeFlag)
		return "", err
	}

	// Set these attributes now
	if isCreation {
		app.Name = appName
		app.Description = appDesc
		app.Metadata.Version = appsMetadataVersion
		app.Metadata.Data.SourcesURI = appSrcURL
		app.Metadata.Data.Revision = appSrcRev
		app.Metadata.Data.BuildType = rpapiv1.RPAppBuildMethod(buildType)

	} else {
		if appName != flagUpdateKeepCurrentValue {
			app.Name = appName
		}
		if appDesc != flagUpdateKeepCurrentValue {
			app.Description = appDesc
		}
		if appSrcRev != flagUpdateKeepCurrentValue {
			app.Metadata.Data.Revision = appSrcRev
		}
		if appSrcURL != flagUpdateKeepCurrentValue {
			app.Metadata.Data.SourcesURI = appSrcURL
		}
		if buildType != flagUpdateKeepCurrentValue {
			app.Metadata.Data.BuildType = rpapiv1.RPAppBuildMethod(buildType)
		}
	}

	// Get the auto-test, auto-audit, ... values wanted by the user
	boolSettings := map[string]*bool{
		appsAddCmdAutoTestFlag:                &app.AutoTests.Set,
		appsAddCmdAutoAuditFlag:               &app.AutoAudits.Set,
		appsAddCmdIgnoreCheckDepsResultFlag:   &app.IgnoreCheckDepsResult,
		appsAddCmdIgnoreMissingRedTestRpmFlag: &app.IgnoreMissingRedTestRpm,
	}

	for optName, value := range boolSettings {
		*value, err = genericGetBoolFlagValue(cmd, isCreation, *value, optName,
			fmt.Sprintf("%s \"%s\" flag", pbFlagMainPrefix, optName))
		if err != nil {
			return "", err
		}
	}

	// Get the specfile type specified by the user
	appSpecType, err := cmd.Flags().GetString(appsAddCmdSpecTypeFlag)
	if err != nil {
		logFlagError(appsAddCmdSpecTypeFlag)
		return "", err
	}

	// Check that the spectype given is authorized
	possibleSpecType := []string{
		string(rpapiv1.RPAppSpecTypeInternal),
		string(rpapiv1.RPAppSpecTypeExternal),
		string(rpapiv1.RPAppSpecTypeRelative),
	}
	if !isCreation {
		possibleSpecType = append(possibleSpecType, flagUpdateKeepCurrentValue)
	}
	if !utils.ContainsString(possibleSpecType, appSpecType) {
		// The value given by the user is not authorized
		return "", errors.New(appSpecTypeNotAuthMsg)
	}

	// Get the specfile path specified by the user
	appSpecLocation, err := cmd.Flags().GetString(appsAddCmdSpecLocationFlag)
	if err != nil {
		logFlagError(appsAddCmdSpecLocationFlag)
		return "", err
	}

	// Convert the type given by the user unless we have the "keep current value" string
	if appSpecType != flagUpdateKeepCurrentValue {
		app.Metadata.Data.SpecfileType = rpapiv1.RPAppSpecType(appSpecType)
	}

	localSpecfilePath := ""
	// If the specfile type is internal the location is always empty
	if app.Metadata.Data.SpecfileType == rpapiv1.RPAppSpecTypeInternal {
		app.Metadata.Data.SpecfileLocation = ""

		// In this case, if the user has given a location, it is a local one!
		if appSpecLocation != flagUpdateKeepCurrentValue {
			localSpecfilePath = appSpecLocation
		}

		// Get the location given by the user unless we have the "keep current value" string
	} else if appSpecLocation != flagUpdateKeepCurrentValue {
		app.Metadata.Data.SpecfileLocation = appSpecLocation
	}

	// Get the services given by the user
	inputServices, err := cmd.Flags().GetStringArray(appsAddCmdServiceFlag)
	if err != nil {
		logFlagError(appsAddCmdServiceFlag)
		return localSpecfilePath, err
	}

	currentSrpmReBuilds, _ := intToAutoSrpmRebuildOptions(*app)

	// Get global srpm rebuild value if given by user
	srpmAutoRebuild, err := cmd.Flags().GetString(appsAddCmdAutoSRpmBuildAllFlag)
	if err != nil {
		logFlagError(appsAddCmdAutoSRpmBuildAllFlag)
		return localSpecfilePath, err
	}

	// Get the srpm rebuilds given by the user
	inputRebuildSrpmSettings, err := cmd.Flags().GetStringSlice(appsAddCmdAutoSRpmBuildFlag)
	if err != nil {
		logFlagError(appsAddCmdServiceFlag)
		return localSpecfilePath, err
	}

	// Ignore auto-srpm-build parameters if global auto-srpm flag is set
	if len(srpmAutoRebuild) == 0 {
		for _, value := range inputRebuildSrpmSettings {
			// Split the flag value given by the user using ":"
			inputSrpmSettingsSplit := strings.Split(value, "=")
			if len(inputSrpmSettingsSplit) != 2 {
				return "", fmt.Errorf("invalid auto srpm setting definition %s", value)
			}

			// Get key value and clean it up
			autoSrpmSettingName := inputSrpmSettingsSplit[0]
			autoSrpmSettingName = strings.Trim(autoSrpmSettingName, "\"")
			autoSrpmSettingName = strings.Trim(autoSrpmSettingName, "'")

			// Check if srpm rebuild option from cmd line exists in available options and if not found - return an error
			if !checkIfSrpmRebuildCmdParamExists(autoSrpmSettingName) {
				logFlagError(appsAddCmdServiceFlag)
				errMsg := fmt.Sprintf(appAutoSrpmParamNotExist, autoSrpmSettingName)
				return "", errors.New(errMsg)
			}

			// Get the boolean value of the setting
			value, err := strconv.ParseBool(inputSrpmSettingsSplit[1])
			if err != nil {
				logFlagError(appsAddCmdAutoSRpmBuildFlag)
				return "", err
			}

			// Set received from input value for the relative auto srpm build setting
			currentSrpmReBuilds[appendPositionToKey(autoSrpmSettingName)] = value
		}
	}

	autoSrpmBuild := convertAutoSrpmBuildSettingsToInt(currentSrpmReBuilds)

	// Parse auto_srpm flags value if true set to 15(1111) if false set to 0 (0000), if any other value log a flag error
	if len(srpmAutoRebuild) > 0 {
		autoSrpmRebuildValue, err := strconv.ParseBool(srpmAutoRebuild)
		if err != nil {
			logFlagError(appsAddCmdAutoSRpmBuildAllFlag)
			errMsg := fmt.Sprint(appAutoSrpmRebuildNotValid)
			return "", errors.New(errMsg)
		}

		if autoSrpmRebuildValue {
			autoSrpmBuild = autoSrpmOptionsAllTrue
		} else {
			autoSrpmBuild = 0
		}
	}

	// Display warning if both flags (auto-srpm and auto-srpm-build) has been set
	if len(srpmAutoRebuild) > 0 && len(inputRebuildSrpmSettings) > 0 {
		Log.Warnf(autoSrpmBuildIncompatibleFlags)
	}

	// if any autoSrpm rebuild parameter is set to "false" display a warning message
	if autoSrpmBuild != autoSrpmOptionsAllTrue {
		Log.Warnf(autoSrpmBuildParameterFalse)
	}

	// Copy auto srpm build value as an int into
	err = copier.Copy(&app.AutoRSrpmBuild, &autoSrpmBuild)
	if err != nil {
		return localSpecfilePath, err
	}

	bckServices := bckapi.RPBCKServices{}
	needToModifyAppService := true
	servicesToDisable := []string{}
	if isCreation {
		// Get the list of services available on redpesk factory
		err = bckAPI.RPServicesGet(&bckServices)
		if err != nil {
			return localSpecfilePath, err
		}
	} else {
		// Try to get possible services to disable
		servicesToDisable, err = cmd.Flags().GetStringSlice(appsUpdateCmdDisableServiceFlag)
		if err != nil {
			logFlagError(appsUpdateCmdDisableServiceFlag)
			return localSpecfilePath, err
		}

		// Delete flagUpdateKeepCurrentValue from inputServices
		utils.DeleteStringFromSlice(&inputServices, flagUpdateKeepCurrentValue)

		// Check if user gave service flag
		if len(inputServices) > 0 || len(servicesToDisable) > 0 {
			// Check if the Services structure is nil
			if app.State == rpapiv1.RPAppLockState {
				return localSpecfilePath, errors.New("application currently locked: update impossible")
			}

			// Convert from map[string]interfaces to RPBCKServices
			// decode first map of interface{}, so it can be encode into RPBCKServices
			jsonbody, err := json.Marshal(app.Services)
			if err != nil {
				return localSpecfilePath, err
			}

			// then convert from map[string]interfaces to RPBCKServices
			if err := json.Unmarshal(jsonbody, &bckServices); err != nil {
				return localSpecfilePath, err
			}
		} else {
			needToModifyAppService = false
		}
	}

	if needToModifyAppService {
		// Modify the services according to what the user gave
		newServices, err := decodeServiceFromCmdOptions(inputServices, bckServices)
		if err != nil {
			return localSpecfilePath, err
		}

		// If we are in "update" mode, the user can ask to disable services
		if !isCreation {
			// Loop on services and disable the ones requested by the user
			for _, service := range servicesToDisable {
				if serviceStruct, ok := newServices[service]; ok {
					// Disable the service
					serviceStruct.Used = false
				} else {
					// Print a warning saying that the service does not exist
					Log.Warnf(appUpdateDisableServiceNotExisting, service)
				}
			}
		}

		// Copy the services as a map[string]interface{} in the application
		err = copier.Copy(&app.Services, &newServices)
		if err != nil {
			return localSpecfilePath, err
		}
	}

	// Manage the default audits requested by the user
	defAuditsWanted, err := cmd.Flags().GetStringSlice(appsAddCmdDefAuditFlag)
	if err != nil {
		logFlagError(appsAddCmdDefAuditFlag)
		return localSpecfilePath, err
	}
	// Check if we need to update the value in the application structure
	if isCreation || !utils.ContainsString(defAuditsWanted, flagUpdateKeepCurrentValue) {

		// In the case where the user gave only "" or "reset", gives an empty slice
		if (len(defAuditsWanted) == 1) && (utils.ContainsString(defAuditsWanted, "") || utils.ContainsString(defAuditsWanted, "reset")) {
			defAuditsWanted = []string{}
		}

		// Prepare the structure
		auditsStruct, err := checkAndPrepareToolsRequest(defAuditsWanted)
		if err != nil {
			return localSpecfilePath, err
		}
		// Convert the structure to a string
		bytesAudits, err := json.Marshal(auditsStruct)
		if err != nil {
			return localSpecfilePath, err
		}
		// Save in the app structure
		app.AuditsConf = string(bytesAudits)
	}

	// Get the list of all the mandatory distributions
	appMandDistros, err := cmd.Flags().GetStringSlice(appsAddCmdMandDistFlag)
	if err != nil {
		logFlagError(appsAddCmdMandDistFlag)
		return localSpecfilePath, err
	}
	if !isCreation {
		if utils.ContainsString(appMandDistros, flagUpdateKeepCurrentValue) {
			// Retrieve the distros already existing in application
			appMandDistros = []string{}
			for _, distro := range app.Distributions {
				if distro.Set && distro.Mandatory {
					appMandDistros = append(appMandDistros, distro.Slug)
				}
			}
		} else {
			// Reset the mandatory distro for the apps and let the end of the function do its work
			app.Distributions = []rpapiv1.RPDistro{}
		}
	}

	// Get the list of all the optional distributions
	appOptDistros, err := cmd.Flags().GetStringSlice(appsAddCmdOptDistFlag)
	if err != nil {
		logFlagError(appsAddCmdOptDistFlag)
		return localSpecfilePath, err
	}
	if !isCreation {
		if utils.ContainsString(appOptDistros, flagUpdateKeepCurrentValue) {
			// Retrieve the distros already existing in application
			appOptDistros = []string{}
			for _, distro := range app.Distributions {
				if distro.Set && !distro.Mandatory {
					appOptDistros = append(appOptDistros, distro.Slug)
				}
			}
		} else {
			// Reset the mandatory distro for the apps and let the end of the function do its work
			app.Distributions = []rpapiv1.RPDistro{}
		}
	}

	// Fill the last attributes
	app.AutoTests.ID = "Automatic tests"
	app.AutoAudits.ID = "Automatic audits"
	app.Parent.ProjectID = parentPrj.ID

	// Fill the distributions infos for the app structure
	RPPrinter.CustomPrint(appsAddCmdAnsLine2)
	err = fillDistInfo(appMandDistros, appOptDistros, parentPrj, app)
	if err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return localSpecfilePath, err
	}
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	return localSpecfilePath, nil
}

// decodeServiceFromCmdOptions decode --service option values
//
// :param inputServices:
// :type inputServices:
// :returns: error -- an error if something went wrong, nil in the other case
func decodeServiceFromCmdOptions(inputServices []string, defaultBckServices bckapi.RPBCKServices) (bckapi.RPBCKServices, error) {

	// Create the resulting map
	res := make(bckapi.RPBCKServices)

	// Copy the default services inside the result
	copier.Copy(&res, &defaultBckServices)

	for _, values := range inputServices {

		// Split the flag value given by the user using "="
		serviceSplitted := strings.Split(values, "=")
		if len(serviceSplitted) != 2 {
			return res, fmt.Errorf("invalid service definition %s", values)
		}

		serviceName := serviceSplitted[0]

		// Replace simple quote by escaped double quotes to be able to unmarshal the json string
		serviceValue := strings.ReplaceAll(serviceSplitted[1], "'", "\"")

		// Convert the json string to a map
		settedParams := make(map[string]interface{})
		err := json.Unmarshal([]byte(serviceValue), &settedParams)
		if err != nil {
			return res, fmt.Errorf("invalid definition services values %s : error = %v", serviceValue, err)
		}

		// Check that the service name passed is existing in the available services
		if bckSrvStruct, ok := defaultBckServices[serviceName]; ok {
			// For each parameter given by the user, take the value and set it in the result map
			for paramName, paramValue := range settedParams {
				// Check that the parameter exists for this service in the available service
				if _, ok := bckSrvStruct.Parameters[paramName]; ok {
					// Modify the value as given by the user
					tmpParam := bckapi.RPBCKServiceParam{
						Type:        res[serviceName].Parameters[paramName].Type,
						Description: res[serviceName].Parameters[paramName].Description,
						Value:       paramValue,
					}
					// Assign new parameter value
					res[serviceName].Parameters[paramName] = tmpParam
					res[serviceName].Used = true
				} else {
					return res, fmt.Errorf("param '%s' not found for service %s", paramName, serviceName)
				}
			}
		} else {
			fmt.Println(bckSrvStruct, "\n", serviceName)
			return res, fmt.Errorf("unsupported service name %s", serviceName)
		}
	}

	return res, nil
}

// createApplication creates an application as requested from the user
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func createApplication(cmd *cobra.Command, args []string) error {

	RPPrinter.CustomPrintLn(appsAddCmdAnsLine0)

	// Get the name of the project where the application needs to be created
	appPrjName, err := cmd.Flags().GetString(appsCmdPrjFlag)
	if err != nil {
		logFlagError(appsCmdPrjFlag)
		return err
	}
	if appPrjName == "" {
		Log.Error(appPrjFlagNotSetCreateMsg)
		return errors.New(appPrjFlagRequiredCreateMsg)
	}

	// Retrieve the project from UUID or its slug
	project, err := getProjectFromIdentifier(appPrjName, true)
	if err != nil {
		return err
	}

	// Create the structure to fill
	app := rpapiv1.RPApplicationFull{}

	// Prepare the application structure to send to the backend
	appLocalSpecPath, err := prepareAppStruct(cmd, &app, &project, true)
	if err != nil {
		return err
	}

	// Request the creation of the application to the backend
	if err = requestAppCreation(&app, project.ID); err != nil {
		return err
	}

	if appLocalSpecPath != "" {
		// Upload the specfile to the redpesk factory
		RPPrinter.CustomPrintLn()
		RPPrinter.ForceFlush()
		RPPrinter.CustomPrintLn(appsUploadCmdAnsLine0)

		err = uploadFileUtility([]string{appLocalSpecPath}, app.ID, false)
		if err != nil {
			return err
		}
	}

	return nil
}

// createApplicationFromSample creates an application from a sample
// as requested from the user
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func createApplicationFromSample(cmd *cobra.Command, args []string) error {

	RPPrinter.CustomPrintLn(appsAddCmdAnsLine0)

	// Get the application ID and return an error if there is nothing
	appID, err := getFirstArg(args, fmt.Sprintf(objectIDNotSetMsg, "application"), fmt.Sprintf(objectIDRequiredMsg, "application"))
	if err != nil {
		return err
	}

	// Get the available samples from the backend
	apps := []rpapiv1.RPSampleApplication{}
	if err := bckAPI.ApplicationsSamplesListGet(&apps); err != nil {
		return err
	}

	// Check that specified app exists in the samples furnished by the backend
	appIndex := -1
	for index, appSample := range apps {
		if appSample.PackageName == appID {
			appIndex = index
			break
		}
	}

	if appIndex == -1 {
		errMsg := fmt.Sprintf(appSampleNotFoundMsg, appID)
		return errors.New(errMsg)
	}

	// Get the name of the project where the application needs to be created
	appPrjName, err := cmd.Flags().GetString(appsCmdPrjFlag)
	if err != nil {
		logFlagError(appsCmdPrjFlag)
		return err
	}
	if appPrjName == "" {
		Log.Error(appPrjFlagNotSetCreateMsg)
		return errors.New(appPrjFlagRequiredCreateMsg)
	}

	// Retrieve the project from UUID or its slug
	project, err := getProjectFromIdentifier(appPrjName, true)
	if err != nil {
		return err
	}

	// Cast rpapiv1.RPSampleApplication to rpapiv1.RPApplication using json as "interface"
	castedApp := rpapiv1.RPApplicationFull{}
	byteArrayApp, err := json.Marshal(apps[appIndex])
	if err != nil {
		return err
	}
	err = json.Unmarshal(byteArrayApp, &castedApp)
	if err != nil {
		return err
	}

	// Fill the parent project UUID in the structure
	castedApp.Parent.ProjectID = project.ID

	// Fill the distributions infos for the app structure
	RPPrinter.CustomPrint(appsAddCmdAnsLine2)
	err = fillDistInfo([]string{}, []string{}, &project, &castedApp)
	if err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	// Request the creation of the application from the sample
	if err = requestAppCreation(&castedApp, project.ID); err != nil {
		return err
	}

	return nil
}

// updateApplication update an application as requested from the user
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func updateApplication(cmd *cobra.Command, args []string) error {
	// Declare some necessary variables
	var appLocalSpecPath string

	RPPrinter.CustomPrintLn(appsUpdateCmdAnsLine0)

	// Retrieve the application from UUID or its slug
	app, err := genericGetAppFromArgs(cmd, args, true)
	if err != nil {
		return err
	}

	// Get the project structure
	parentPrj := rpapiv1.RPProjectFull{}
	err = bckAPI.ProjectGet(app.Parent.ProjectID.String(), &parentPrj)
	if err != nil {
		return err
	}

	// Check if it is only a pkg branch change or a full update
	classicUpdate, branchUpdate := appGetUpdatesNeeded(cmd)

	Log.Infof(appUpdateInfoMsg, app.Name)

	if classicUpdate {
		// Prepare the application structure to send to the backend
		appLocalSpecPath, err = prepareAppStruct(cmd, &app, &parentPrj, false)
		if err != nil {
			return err
		}

		// Get the "force-stable-state" flag
		forceStable, err := cmd.Flags().GetBool(appsForceStateFlag)
		if err != nil {
			return err
		}
		// If the user wants it, force the app state at "stable"
		if forceStable {
			app.State = rpapiv1.RPAppStableState
		}

		Log.Debugf(appUpdateDebugMsg, app)

		// Request the update of this application to the backend
		appAns := rpapiv1.RPApplicationFull{}
		if err := bckAPI.ApplicationUpdate(app.ID.String(), &app, &appAns); err != nil {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
			return err
		}

		// Display something to the user if in raw mode
		RPPrinter.CustomPrintObject(appAns, false, displayNothing)
	}

	// Tell the user that the application is currently being updated
	RPPrinter.CustomPrintF(appsUpdateCmdAnsLine1, app.Name)

	// Check from viper the non-blocking flag to see if we wait
	nonBlocking := viper.GetBool(nonBlockingFlag)

	// Block until we receive a corresponding websocket event
	for classicUpdate && !nonBlocking {
		rEvent := <-cmdEventChannel

		// Check that the received event is for an app update
		if rEvent.Type == rpapiv1.EvtAppStateChange {
			// Decode the application
			rApp, err := rEvent.DecodeApplication()
			if err != nil {
				RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
				return err
			}

			if utils.UUIDEqual(rApp.ID, app.ID) && (rApp.State == rpapiv1.RPAppStableState) {
				RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
				// Display something to the user if in raw mode
				RPPrinter.CustomPrintObject(rApp, false, displayNothing)
				break
			}
		}

		// Check if we get an error event during the operation
		if (rEvent.Type == rpapiv1.EVTError) || (rEvent.Type == rpapiv1.EVTSessionExpired) {
			// Function handling the events
			err = handleErrorEvent(app.ID, rEvent, true)
			if err != nil {
				return err
			}
		}
	}

	if nonBlocking {
		// The user do not want blocking functions
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
	}

	if branchUpdate {
		// Change the active branch of this application
		RPPrinter.CustomPrintLn()
		RPPrinter.ForceFlush()

		// Get the branch wanted by the user
		newActiveBranch, err := cmd.Flags().GetString(appsAddCmdActiveBranchFlag)
		if err != nil {
			logFlagError(appsAddCmdActiveBranchFlag)
			return err
		}

		// Get the current distributions for the destination branch
		distrosSlice, err := getBranchDistrosList(app.ID, newActiveBranch)
		if err != nil {
			return err
		}

		// Request the backend to create the branch or to switch to an existing one
		rpBranch := rpapiv1.RPBranch{
			Source:        "",
			Destination:   newActiveBranch,
			Distributions: distrosSlice,
			Default:       false,
			Active:        true,
		}
		if err := bckAPI.ApplicationCreateOrSwitchBranch(app.ID.String(), &rpBranch); err != nil {
			return err
		}
		RPPrinter.CustomPrintF(appsBranchCmdAnsLine0, newActiveBranch, app.Slug)
	}

	if appLocalSpecPath != "" {
		// Upload the specfile to the redpesk factory
		RPPrinter.CustomPrintLn()
		RPPrinter.ForceFlush()
		RPPrinter.CustomPrintLn(appsUploadCmdAnsLine0)

		err = uploadFileUtility([]string{appLocalSpecPath}, app.ID, false)
		if err != nil {
			return err
		}
	}

	return nil
}

// deleteApplication deletes an existing application as requested from the user
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func deleteApplication(cmd *cobra.Command, args []string) error {

	RPPrinter.CustomPrintLn(appsDelCmdAnsLine0)

	// Retrieve the application from UUID or its slug
	app, err := genericGetAppFromArgs(cmd, args, false)
	if err != nil {
		return err
	}

	// Tell the user that the application is currently being deleted
	RPPrinter.CustomPrintF(appsDelCmdAnsLine1, app.Slug)

	// Ask the backend for deletion
	appAns := rpapiv1.RPApplicationFull{}
	if err := bckAPI.ApplicationDelete(app.ID.String(), &appAns); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	// Display something to the user if in raw mode
	RPPrinter.CustomPrintObject(appAns, false, displayNothing)

	// Check from viper the non-blocking flag to see if we wait
	nonBlocking := viper.GetBool(nonBlockingFlag)
	if nonBlocking {
		// The user do not want blocking functions
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
		return nil
	}

	// Block until we receive a corresponding websocket event
	for {
		rEvent := <-cmdEventChannel

		// Check that the received event is for a application deletion addition
		if rEvent.Type == rpapiv1.EvtAppDelete {
			// Function handling the events
			rApp, err := rEvent.DecodeApplication()
			if err != nil {
				return err
			}

			if utils.UUIDEqual(rApp.ID, appAns.ID) {
				// Display something to the user if in raw mode
				RPPrinter.CustomPrintObject(rApp, false, displayNothing)
				RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
				break
			}
		}

		// Check if we get an error event during the operation
		if (rEvent.Type == rpapiv1.EVTError) || (rEvent.Type == rpapiv1.EVTSessionExpired) {
			// Function handling the events
			err = handleErrorEvent(appAns.ID, rEvent, true)
			if err != nil {
				return err
			}
		}
	}

	return nil
}

// bookmarkApplication bookmarks an existing application as requested from the user
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func bookmarkApplication(cmd *cobra.Command, args []string) error {

	RPPrinter.CustomPrintLn(appsBookmarkCmdAnsLine0)

	// Get the reset flag
	reset, err := cmd.Flags().GetBool(appsBookmarkCmdResetFlag)
	if err != nil {
		logFlagError(appsBookmarkCmdResetFlag)
		return err
	}

	// Get the application given by the user
	appStructure, err := genericGetAppFromArgs(cmd, args, false)
	if err != nil {
		return err
	}

	// Do the appropriate action according to the variables
	resetDone := false
	setDone := false
	bookmarkObj := rpapiv1.RPApplicationBookmark{}
	switch {
	case (reset && !appStructure.Bookmarked):
		// Reset is requested but the app is not bookmarked, nothing to do
		RPPrinter.CustomPrintF(appsBookmarkCmdAnsLine3, appStructure.Slug)
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	case (reset && appStructure.Bookmarked):
		// Reset is requested and the app is bookmarked
		RPPrinter.CustomPrintF(appsBookmarkCmdAnsLine2, appStructure.Slug)
		if err := bckAPI.ApplicationBookmark(appStructure.ID.String(), &bookmarkObj); err != nil {
			return err
		}
		resetDone = true

	case (!reset && appStructure.Bookmarked):
		// Bookmark is requested but the app is already bookmarked, nothing to do
		RPPrinter.CustomPrintF(appsBookmarkCmdAnsLine4, appStructure.Slug)
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	case (!reset && !appStructure.Bookmarked):
		// Bookmark is requested and the app is not bookmarked
		RPPrinter.CustomPrintF(appsBookmarkCmdAnsLine1, appStructure.Slug)
		if err := bckAPI.ApplicationBookmark(appStructure.ID.String(), &bookmarkObj); err != nil {
			return err
		}
		setDone = true
	}

	if setDone || resetDone {

		// Display something to the user if in raw mode
		RPPrinter.CustomPrintObject(bookmarkObj, false, displayNothing)

		// Check from viper the non-blocking flag to see if we wait
		nonBlocking := viper.GetBool(nonBlockingFlag)
		if nonBlocking {
			// The user do not want blocking functions
			RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
			return nil
		}

		// Block until we receive a corresponding websocket event
		for {
			rEvent := <-cmdEventChannel

			// Check that the received event is for a project addition
			if rEvent.Type == rpapiv1.EvtAppStateChange {
				// Function handling the events
				rApp, err := rEvent.DecodeApplication()
				if err != nil {
					return err
				}

				if utils.UUIDEqual(rApp.ID, bookmarkObj.ApplicationID) {
					if rApp.Bookmarked != reset {
						// Display OK if not in raw mode
						RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
						// Display something to the user if in raw mode
						RPPrinter.CustomPrintObject(rApp, false, displayNothing)
						break
					}
				}
			}

			// Check if we get an error event during the operation
			if (rEvent.Type == rpapiv1.EVTError) || (rEvent.Type == rpapiv1.EVTSessionExpired) {
				// Function handling the events
				err = handleErrorEvent(appStructure.ID, rEvent, true)
				if err != nil {
					return err
				}
			}
		}
	}
	return nil
}

// buildApplication launch a build on the existing application
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func buildApplication(cmd *cobra.Command, args []string) error {

	RPPrinter.CustomPrintLn(fmt.Sprintf(genericBuildCmdAnsLine0, "Application"))

	// Get the application given by the user
	app, err := genericGetAppFromArgs(cmd, args, false)
	if err != nil {
		return err
	}

	// Get the additional information (comment) to save with the build
	comment, err := cmd.Flags().GetString(genericBuildCmdCommentFlag)
	if err != nil {
		logFlagError(genericBuildCmdCommentFlag)
		return err
	}

	// Get the value of the srpm flag
	bldType := rpapiv1.RPBuildApp
	isSrpmBld, err := cmd.Flags().GetBool(appsBuildCmdSrpmFlag)
	if err != nil {
		logFlagError(appsBuildCmdSrpmFlag)
		return err
	}

	// Get the value of the check-deps flag
	isCheckDepsBld, err := cmd.Flags().GetBool(appsBuildCmdCheckDepsFlag)
	if err != nil {
		logFlagError(appsBuildCmdCheckDepsFlag)
		return err
	}

	if isSrpmBld {
		bldType = rpapiv1.RPBuildSrpm
	} else if isCheckDepsBld {
		bldType = rpapiv1.RPBuildDeps
	}

	// Manage the distribution to build
	startBuildObj, err := createDistroToBuild(cmd, &app, nil)
	if err != nil {
		return err
	}

	// Update the comment field of the start build object
	startBuildObj.Comment = comment

	// Tell the user that we start the build of the application
	RPPrinter.CustomPrintF(genericBuildCmdAnsLine1, app.Slug, "application")

	// Ask the backend to start the build
	buildsAns := rpapiv1.RPBuilds{}
	if err := bckAPI.ApplicationBuild(app.ID.String(), startBuildObj, bldType, &buildsAns); err != nil {
		return err
	}

	return monitorBuilds(&buildsAns, app.ID, rpapiv1.RPBuildApp)
}

// appCopyToProject copy the application to another project
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func appCopyToProject(cmd *cobra.Command, args []string) error {

	// Get the application given by the user
	app, err := genericGetAppFromArgs(cmd, args, false)
	if err != nil {
		return err
	}

	//  Get the name of the project where the application will be copied
	destProject, err := cmd.Flags().GetString(appCopyCmdDestinationPrjFlag)
	if err != nil {
		logFlagError(appCopyCmdDestinationPrjFlag)
		return err
	}

	// Check that the user gave a correct value for the destination project
	if destProject == "" {
		return errors.New(copyAppDestPrjEmptyMsg)
	}
	// Retrieve the project from UUID or its slug
	project, err := getProjectFromIdentifier(destProject, false)
	if err != nil {
		return err
	}

	// Construct the right structure and send the request to the backend
	rpApplicationCopy := rpapiv1.RPApplicationCopy{Destination: project.ID}
	if err := bckAPI.ApplicationCopy(app.ID.String(), &rpApplicationCopy); err != nil {
		return err
	}

	// Tell the user that everything went well
	RPPrinter.CustomPrintF(appsCopyCmdAnsLine0, app.Slug, destProject)
	return nil
}

// appRefreshGitpkgFiles requests the backend to refresh the gitpkg files for the
// given application
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func appRefreshGitpkgFiles(cmd *cobra.Command, args []string) error {

	RPPrinter.CustomPrintLn(appsRefreshFilesPkgAnsLine0)

	// Get the application given by the user
	app, err := genericGetAppFromArgs(cmd, args, false)
	if err != nil {
		return err
	}

	RPPrinter.CustomPrintF(appsRefreshFilesPkgAnsLine1, app.Slug)

	// Construct the right structure and send the request to the backend
	if err := bckAPI.ApplicationRefreshFiles(app.ID.String()); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	// Check from viper the non-blocking flag to see if we wait
	nonBlocking := viper.GetBool(nonBlockingFlag)
	if nonBlocking {
		// The user do not want blocking functions
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
		return nil
	}

	// Block until we receive a corresponding websocket event
	for {
		rEvent := <-cmdEventChannel

		// Check that the received event is for a project addition
		if rEvent.Type == rpapiv1.EvtAppStateChange {
			// Decode the application
			rApp, err := rEvent.DecodeApplication()
			if err != nil {
				RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
				return err
			}

			if utils.UUIDEqual(rApp.ID, app.ID) && (rApp.State == rpapiv1.RPAppStableState) {
				RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
				break
			}
		}

		// Check if we get an error event during the operation
		if (rEvent.Type == rpapiv1.EVTError) || (rEvent.Type == rpapiv1.EVTSessionExpired) {
			// Function handling the events
			err = handleErrorEvent(app.ID, rEvent, true)
			if err != nil {
				return err
			}
		}
	}

	return nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Utilities functions ------------------------------------------
// --------------------------------------------------------------

// getAppUUIDCallBack is a function specifically designed to be call by the
// `getBuildIDFromArgs()` function to retrieve the UUID of the build parent.
// Most of the time, it is better to call the `genericGetAppFromArgs()`
// since you get the app structure.
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: uuid.UUID, error -- the app UUID and an error if something went wrong
func getAppUUIDCallBack(cmd *cobra.Command, args []string) (uuid.UUID, error) {
	app, err := genericGetAppFromArgs(cmd, args, false)
	if err != nil {
		return uuid.Nil, err
	}
	return app.ID, nil
}

// genericGetAppFromArgs allows a function to get RPApplicationFull from one
// of the identifiers (slug or UUID)
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param getFullApp: if set, the caller request to have a truly full app (so a GET request on /apps/<app_uuid>) will be done
// :type getFullApp: bool
// :returns: rpapiv1.RPApplicationFull, error -- the app and an error if something went wrong
func genericGetAppFromArgs(cmd *cobra.Command, args []string, getFullApp bool) (rpapiv1.RPApplicationFull, error) {

	// Get the application ID and return an error if there is nothing
	appID, err := getFirstArg(args, fmt.Sprintf(objectIDNotSetMsg, "application"), fmt.Sprintf(objectIDRequiredMsg, "application"))
	if err != nil {
		return rpapiv1.RPApplicationFull{}, err
	}

	// Get the name of the project if the user set it
	appPrjID, err := cmd.Flags().GetString(appsCmdPrjFlag)
	if err != nil {
		logFlagError(appsCmdPrjFlag)
		return rpapiv1.RPApplicationFull{}, err
	}

	// Get the type flag
	typesSlice, err := getAppTypesFromFlag(cmd)
	if err != nil {
		return rpapiv1.RPApplicationFull{}, err
	}

	return getAppFromIdentifiers(appID, appPrjID, typesSlice, getFullApp)
}

// getAppFromIdentifiers is a private function allowing to retrieve
// an app from a given identifier.
// This identifier can be the UUID itself as a string representation or
// the slug of the app
//
// :param appIdentifier: the identifier of the app
// :type appIdentifier: string
// :param prjIdentifier: the identifier of the parent project [optional]
// :type prjIdentifier: string
// :param appTypes: application types to ask for [optional]
// :type appTypes: []string
// :param getFullApp: if set, the caller request to have a truly full app (so a GET request on /apps/<app_uuid>) will be done
// :type getFullApp: bool
// :returns: (rpapiv1.RPApplicationFull, error) -- the app and an error if something went bad
func getAppFromIdentifiers(appIdentifier, prjIdentifier string, appTypes []string, getFullApp bool) (rpapiv1.RPApplicationFull, error) {
	// First try to see if the appIdentifier is a valid UUID
	isValidUUID := utils.IsValidUUID(appIdentifier)

	// If it is the case, call directly the backend with this UUID to see if it works
	if isValidUUID {
		app := rpapiv1.RPApplicationFull{}
		if err := bckAPI.ApplicationGet(appIdentifier, &app); err == nil {
			// If everything went well, this means that the identifier was truly the UUID of the app!
			// In the other case keep going and try with slug...
			return app, nil
		}
	}

	// So, try with slug now (using filter feature)
	tmpList := &rpapiv1.RPApplications{}
	foundIndex := -1
	// Create a slice of string to use as filter
	filterValues := []string{
		fmt.Sprintf("slug=%s", appIdentifier),
		fmt.Sprintf("package_name=%s", appIdentifier),
		fmt.Sprintf("name=%s", appIdentifier),
	}
	// Try each filter until finding the right one
	for index, filter := range filterValues {

		// Call the API with the current filter
		if err := bckAPI.ApplicationsListGetOfTypes(tmpList, "", appTypes, filter); err != nil {
			return rpapiv1.RPApplicationFull{}, err
		}

		// If we got one match, and that we don't have a project to check, bingo!
		if len(tmpList.Apps) == 1 && (prjIdentifier == "") {
			foundIndex = 0
			break
		}

		// If we got no match, try the other filter
		if (len(tmpList.Apps) < 1) && (runMode != "testing") {
			Log.Warnf(appIDNotFoundOnTryX, appIdentifier, index)
			continue
		}

		// If we got too many match and that no project identifier has been given, try the other filter
		if (len(tmpList.Apps) > 1) && (prjIdentifier == "") && (runMode != "testing") {
			Log.Warnf(tooManyAppIDFoundOnTryX, index)
			continue
		}

		// If we got too many match but a project identifier, try to get it
		if (len(tmpList.Apps) >= 1) && (prjIdentifier != "") {

			parentPrj, err := getProjectFromIdentifier(prjIdentifier, false)
			if err != nil {
				return rpapiv1.RPApplicationFull{}, err
			}

			for tmpIndex, app := range tmpList.Apps {
				if utils.UUIDEqual(app.Parent.ProjectID, parentPrj.ID) {
					foundIndex = tmpIndex
					break
				}
			}

			if foundIndex != -1 {
				break
			} else {
				Log.Warnf(appExistsButNotWithThisPrj, appIdentifier)
			}
		}
	}

	if foundIndex == -1 {
		// Error: no app found with this identifier
		return rpapiv1.RPApplicationFull{}, fmt.Errorf(appIdentifierNotFound, appIdentifier)
	}

	// Now we have the right app index in any case

	// Check if the caller wants a real full app
	if getFullApp && tmpList.Apps[foundIndex].Parent.Type != rpapiv1.RPProjectTypeExternal {
		appToReturn := rpapiv1.RPApplicationFull{}
		err := bckAPI.ApplicationGet(tmpList.Apps[foundIndex].ID.String(), &appToReturn)
		return appToReturn, err
	}

	// The caller does not need a full app: just set the light one inside an "empty" shell
	appToReturn := rpapiv1.RPApplicationFull{
		RPApplication: tmpList.Apps[foundIndex],
	}
	return appToReturn, nil
}

// getAppTypesFromFlag checks if the flag exists in the get command and if it exists, it returns the value given
// by the user
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :returns: []string, error -- the types slice and an error if something went wrong
func getAppTypesFromFlag(cmd *cobra.Command) ([]string, error) {
	defaultFlags := []string{"standard", "external"}

	// Check that the right flag exists
	if cmd.Flags().Lookup(appsListCmdTypeFlag) != nil {
		typesSlice, err := cmd.Flags().GetStringSlice(appsListCmdTypeFlag)
		if err != nil {
			logFlagError(appsListCmdTypeFlag)
		}

		// Convert "distro" to "external", and remove duplicated values
		utils.ReplaceInStringSlice(&typesSlice, "distro", "external")
		typesSlice = utils.RemoveDuplicateInStrSlice(typesSlice)

		return typesSlice, err
	}

	return defaultFlags, nil
}

// fillDistInfo can be used to fill the Distributions part of the
// application structure from what the user wants and from the project ID
//
// :param mandDistWanted: tab of string containing the mandatory distributions
// wanted by the user
// :type mandDistWanted: []string
// :param optDistWanted: tab of string containing the optional distributions
// wanted by the user
// :type optDistWanted: []string
// :param prj: pointer to the project structure
// :type prj: *rpapiv1.RPProjectFull
// :param pApp: Pointer to the structure to fill
// :type pApp: *rpapiv1.RPApplicationFull
// :returns: error -- an error if something went wrong, nil in the other case
func fillDistInfo(mandDistWanted []string, optDistWanted []string, prj *rpapiv1.RPProjectFull, pApp *rpapiv1.RPApplicationFull) error {

	atLeastOne := false

	// Convert the tabs of string containing the distributions wanted to tab of uuid.UUID
	mandDistWantedUUID, err := getTabOfDistrosUUID(mandDistWanted)
	if err != nil {
		return err
	}
	optDistWantedUUID, err := getTabOfDistrosUUID(optDistWanted)
	if err != nil {
		return err
	}

	// If the user gave nothing, use the project distributions
	if len(mandDistWanted)+len(optDistWanted) <= 0 {
		Log.Infof(appNoDistByUser, prj.Slug)
		for _, tmpDist := range prj.Distributions {
			if tmpDist.Set {
				pApp.Distributions = append(pApp.Distributions, tmpDist)
				atLeastOne = true
			}
		}
	} else {
		// If the user has specified some distributions, check that they match the project one
		// and if they does, add it to the structure
		for _, tmpDist := range prj.Distributions {

			if tmpDist.Set && tmpDist.Mandatory {
				// Set a log saying that we are adding it anyway
				Log.Infof(appMandDistMsg, tmpDist.ID, prj.Slug)
				pApp.Distributions = append(pApp.Distributions, tmpDist)
				atLeastOne = true

			} else if tmpDist.Set && !tmpDist.Mandatory {
				// Check if the user wants this architecture as optional
				if utils.ContainsString(optDistWantedUUID, tmpDist.ID.String()) {
					pApp.Distributions = append(pApp.Distributions, tmpDist)
					atLeastOne = true
				}
				// Check if the user wants this architecture as mandatory
				if utils.ContainsString(mandDistWantedUUID, tmpDist.ID.String()) {
					tmpDist.Mandatory = true
					pApp.Distributions = append(pApp.Distributions, tmpDist)
					atLeastOne = true
				}
			}
		}
	}

	// Check if we have at least one architecture set
	if !atLeastOne {
		return errors.New(appDistNotSetMsg)
	}

	return nil
}

// appGetUpdatesNeeded is a function allowing to check what kind of update is needed
// for the application: either a classic update, or a pkg branch switch, or both
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :returns: bool, bool -- classicUpdateNeeded, pkGBranchUpdateNeeded
func appGetUpdatesNeeded(cmd *cobra.Command) (bool, bool) {
	classicUpdateNeeded := false
	pkGBranchUpdateNeeded := false

	// Function saving flag set
	activatedFlags := []string{}
	saveFlag := func(f *pflag.Flag) {
		activatedFlags = append(activatedFlags, f.Name)
	}

	// Get the flags activated by the user
	cmd.Flags().Visit(saveFlag)

	// Loop over the args
	for _, flag := range activatedFlags {
		if flag == appsAddCmdActiveBranchFlag {
			pkGBranchUpdateNeeded = true
		} else {
			classicUpdateNeeded = true
		}
	}

	return classicUpdateNeeded, pkGBranchUpdateNeeded
}

// uploadFileUtility is a private function allowing to upload given files to the
// Redpesk factory (for the application given in parameters)
//
// :param filePaths: paths to the files to upload
// :type filePaths: []string
// :param appUUID: ID of the application concerned by the file
// :type appUUID: uuid.UUID
// :param quiet: tell the function to not print user message
// :type quiet: bool
// :returns: error -- an error if something went bad
func uploadFileUtility(filePaths []string, appUUID uuid.UUID, quiet bool) error {

	// If empty directory, does nothing
	if len(filePaths) < 1 {
		return nil
	}

	// Check that the files exists
	fileBaseNames := []string{}
	for _, filePath := range filePaths {
		if _, err := os.Stat(filePath); errors.Is(err, os.ErrNotExist) {
			return fmt.Errorf(appUploadFileDoesNotExists, filePath)
		}
		fileBaseNames = append(fileBaseNames, path.Base(filePath))
	}

	// Communication to user
	pathsToPrint := strings.Join(fileBaseNames, ", ")
	if !quiet {
		RPPrinter.CustomPrint(appsUploadCmdAnsLine2, pathsToPrint, " ")
		RPPrinter.ForceFlush()
	}

	// Send the request to the backend
	appAns := rpapiv1.RPApplicationFull{}
	if err := bckAPI.ApplicationUploadFiles(appUUID.String(), filePaths, &appAns); err != nil {
		if !quiet {
			// Communication to user
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		}
		return err
	}

	if !quiet {
		// The user do not want blocking functions
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
	}

	return nil
}

// manageAppCreationEvents is a function allowing to manage the events received during the
// application creation. The boolean returned allows the caller of the function to know when
// all the needed events have been received.
//
// :param event: event received through the websocket
// :type event: rpapiv1.EventMsg
// :param appAnsUUID: UUID of the application that is being created
// :type appAnsUUID: uuid.UUID
// :returns: bool, error -- true if the app is "stable" or "created" and an error if something went wrong
func manageAppCreationEvents(event rpapiv1.EventMsg, appAnsUUID uuid.UUID) (bool, error) {
	// Initialize the result
	result := false

	// Check that the received event is for a application addition
	if event.Type == rpapiv1.EvtAppAdd || event.Type == rpapiv1.EvtAppStateChange {
		// Decode the application
		rApp, err := event.DecodeApplication()
		if err != nil {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
			return result, err
		}

		if utils.UUIDEqual(rApp.ID, appAnsUUID) {
			// The application slug corresponds to the one we have created
			if event.Type == rpapiv1.EvtAppAdd {
				// Display something to the user if in raw mode
				RPPrinter.CustomPrintObject(rApp, false, displayNothing)
				// Write that we are waiting for confirmation
				RPPrinter.CustomPrint(appsAddCmdAnsLine4)
			} else if (event.Type == rpapiv1.EvtAppStateChange) && (rApp.State == rpapiv1.RPAppStableState) {
				// Display something to the user if in raw mode
				RPPrinter.CustomPrintObject(rApp, false, displayNothing)
				// The application is completely created
				RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
				result = true
			}
		}
	}
	return result, nil
}

// requestAppCreation is a function allowing to request the creation of an application
// to the backend. It manages also all the prints done during the process.
//
// :param application: pointer to the application to create
// :type application: *rpapiv1.RPApplicationFull
// :param projectUUID: UUID of the parent project
// :type projectUUID: uuid.UUID
// :returns: error -- an error if something went wrong
func requestAppCreation(application *rpapiv1.RPApplicationFull, projectUUID uuid.UUID) error {

	// Few logs
	Log.Infof(appCreationInfoMsg, application.Name)
	Log.Debugf(appCreationDebugMsg, application)

	// Tell the user that the project is currently being created
	RPPrinter.CustomPrintF(appsAddCmdAnsLine3, application.Name)

	// Request the creation of this project to the backend
	appAns := rpapiv1.RPApplicationFull{}
	if err := bckAPI.ApplicationCreate(projectUUID.String(), application, &appAns); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	// Set the right UUID in our application (for caller)
	application.ID = appAns.ID

	// The creation request has been taken into account by the backend
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	// Display something to the user if in raw mode
	RPPrinter.CustomPrintObject(appAns, false, displayNothing)

	// Create the sentence to display when the object will be added
	slugSentence := fmt.Sprintf(objectSlugDisplayMsg, "application", appAns.Slug)

	// Check from viper the non-blocking flag to see if we wait
	nonBlocking := viper.GetBool(nonBlockingFlag)
	if nonBlocking {
		// The user do not want blocking functions
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
		// Display the slug of the created object
		RPPrinter.CustomPrintLn(slugSentence)
		return nil
	}

	// Block until we receive all the websocket events we need to confirm creation
	for {
		rEvent := <-cmdEventChannel

		// Check that the received event is for a application addition
		isEndOfWait, err := manageAppCreationEvents(rEvent, appAns.ID)
		if err != nil {
			return err
		}

		// If this is the end of waiting time, escape!
		if isEndOfWait {
			// Display the slug of the created object
			RPPrinter.CustomPrintLn(slugSentence)
			break
		}

		// Check if we get an error event during the operation
		if (rEvent.Type == rpapiv1.EVTError) || (rEvent.Type == rpapiv1.EVTSessionExpired) {
			// Function handling the events
			err = handleErrorEvent(appAns.ID, rEvent, true)
			if err != nil {
				return err
			}
		}
	}

	return nil
}

// manageDownloadDir manages the creation of a download directory
// where the user specifies it. In order to do that, it retrieves the
// value of the "--directory" flag to do so.
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param subDirName: name of the subdirectory to create
// :type subDirName: string
// :param objectID: ID of the concerned object (will be used as another subdirectory)
// :type objectID: string
// :returns: string, error -- path to the created directory and an error if something went wrong
func manageDownloadDir(cmd *cobra.Command, subDirName, objectID string) (string, error) {
	// Get the directory given by the user
	dir, err := cmd.Flags().GetString(genericBuildLogsDirFlag)
	if err != nil {
		logFlagError(genericBuildLogsDirFlag)
		return "", err
	}

	// Get the absolute path from what the user gave us
	cleanDir, err := utils.SanitizePath(dir)
	if err != nil {
		return "", err
	}

	// Create the full path of our directory
	fullPathDir := path.Join(cleanDir, subDirName, objectID)

	// Create the directory
	err = utils.CreateDirIfNotExists(fullPathDir, true)
	return fullPathDir, err
}

// createDistroToBuild is the function managing the distribution to pass to the backend
// when we request to build an application or a project
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param app: App to build
// :type app: *rpapiv1.RPApplicationFull
// :param project: Project to build
// :type project: *rpapiv1.RPProjectFull
// :returns: *rpapiv1.RPDistrosToBuild, error -- pointer to the distro to build and an error if something went wrong
func createDistroToBuild(cmd *cobra.Command, app *rpapiv1.RPApplicationFull, project *rpapiv1.RPProjectFull) (*rpapiv1.RPBuildStart, error) {
	distrosResult := &rpapiv1.RPBuildStart{
		Applications:  []uuid.UUID{},
		Distributions: []rpapiv1.RPDistro{},
		Comment:       "",
	}

	// Get the distributions given by the user
	userDistros, err := cmd.Flags().GetStringSlice(appsBuildCmdDistroFlag)
	if err != nil {
		logFlagError(appsBuildCmdDistroFlag)
		return distrosResult, err
	}
	emptyUserList := (len(userDistros) == 0)

	// Get the distributions available in the project
	var genericDistrosList []rpapiv1.RPDistro
	var genericSlug string

	if app != nil {
		genericDistrosList = app.Distributions
		genericSlug = app.Slug

	} else if project != nil {
		genericDistrosList = project.Distributions
		genericSlug = project.Slug
	} else {
		return distrosResult, errors.New(appUUIDOrPrjUUIDNeedToBeSet)
	}

	// Fill a map containing the distribution slugs as key (easier for next steps)
	distroMap := make(map[string]rpapiv1.RPDistro)
	distroMapUUID := make(map[string]rpapiv1.RPDistro)
	for _, distro := range genericDistrosList {
		if distro.Set {
			distroMapUUID[distro.ID.String()] = distro
			distroMap[distro.Slug] = distro
			// If the user did not give any distro set them all
			if emptyUserList {
				userDistros = append(userDistros, distro.Slug)
			}
		}
	}

	// Now compare the available distro with what the user gave us
	for _, usrDistro := range userDistros {
		// Check that the distro is available for this project
		distroValUUID, okUUID := distroMapUUID[usrDistro]
		distroVal, ok := distroMap[usrDistro]
		if !ok && !okUUID {
			errMsg := fmt.Sprintf(distroNotAvailableMsg, usrDistro, genericSlug)
			return distrosResult, errors.New(errMsg)
		} else if ok {
			// Add the distribution in the result
			distrosResult.Distributions = append(distrosResult.Distributions, rpapiv1.RPDistro{
				ID:        distroVal.ID,
				Slug:      distroVal.Slug,
				Name:      distroVal.Name,
				Version:   distroVal.Version,
				Set:       distroVal.Set,
				Mandatory: distroVal.Mandatory,
			})
		} else if okUUID {
			// Add the distribution in the result
			distrosResult.Distributions = append(distrosResult.Distributions, rpapiv1.RPDistro{
				ID:        distroValUUID.ID,
				Slug:      distroValUUID.Slug,
				Name:      distroValUUID.Name,
				Version:   distroValUUID.Version,
				Set:       distroValUUID.Set,
				Mandatory: distroValUUID.Mandatory,
			})
		}
	}

	if len(distrosResult.Distributions) == 0 {
		distrosResult = nil
	}

	return distrosResult, nil
}

// downloadGitPkgsAppFiles is the function allowing to download a list of files from an application gitpkg.
// A download directory can be specified. If the caller wants to print user messages, the 'printMsg' argument
// needs to be set to true
//
// :param appUUID: Unique identifier of the application containing the files
// :type appUUID: uuid.UUID
// :param dlDirectory: Download directory, if it is the current one that is wanted, give empty string
// :type dlDirectory: string
// :param filesList: List of files to download
// :type filesList: []string
// :param printMsg: Set to true to display messaged in the command line during downloads
// :type printMsg: bool
// :returns: error -- an error if something went wrong
func downloadGitPkgsAppFiles(appUUID uuid.UUID, dlDirectory string, filesList []string, printMsg bool) error {
	// Initialize the variable telling if we got one error
	atLeastOneErr := false

	// Custom function allowing to print OK or KO if needed
	funcPrintOKKOIfNeeded := func(color string, a ...interface{}) {
		if printMsg {
			RPPrinter.CustomColorPrint(color, a...)
		}
	}

	// Print a title message if wanted
	if printMsg {
		RPPrinter.CustomPrintLn(appsDownloadCmdAnsLine0)
	}

	// Now iterate over the list of files
	for _, name := range filesList {
		// Communication to user
		if printMsg {
			RPPrinter.CustomPrint(appsDownloadCmdAnsLine1, name, " ")
			RPPrinter.ForceFlush()
		}
		// Download the current file
		err := bckAPI.ApplicationDownloadFile(appUUID.String(), dlDirectory, name)
		if err != nil {
			funcPrintOKKOIfNeeded(printer.ColorRed, globalKOLabel)
			atLeastOneErr = true
		} else {
			funcPrintOKKOIfNeeded(printer.ColorGreen, globalOKLabel)
		}
	}

	if atLeastOneErr {
		return errors.New("at least one file download fails")
	}

	return nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Display functions --------------------------------------------
// --------------------------------------------------------------

// displayApps is a private function allowing to display
// a slice of applications in the console
//
// :param applications: tab of applications to display
// :type applications: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func displayApps(applications interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Convert to the right type
	apps := applications.([]rpapiv1.RPApplication)

	// Display result
	first := true
	for _, application := range apps {
		if verbose {
			if !first {
				fmt.Fprintln(writer)
			}

			// Print this app
			err := _printOneApp(application, verbose, false, false, writer)
			if err != nil {
				return err
			}

		} else {
			if first {
				manageHeaderPrint(appsListCmdHeader, writer)
			}
			// Print this app
			err := _printOneApp(application, verbose, false, false, writer)
			if err != nil {
				return err
			}
		}
		first = false
	}

	return nil
}

// displayOneApp is a private function allowing to display
// one application in the console
//
// :param application: Application to display
// :type application: interface{}
// :param verbose: Verbose flag passed in the command line
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func displayOneAppFull(application interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Print the app
	err := _printOneApp(application, verbose, true, true, writer)
	if err != nil {
		return err
	}
	return nil
}

// _printOneApp is a private function allowing to print one application in the console
//
// :param app: application to print (can be rpapiv1.RPApplicationFull or rpapiv1.RPApplication)
// :type app: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param isAlone: True if there is only one project to display in the context where this function is called
// :type isAlone: bool
// :param isFull: True if we need to display the Full object, false in the other case
// :type isFull: bool
// :param writer: Writer to used to print the project information
// :type writer: *tabwriter.Writer
func _printOneApp(appInterface interface{}, verbose, isAlone, isFull bool, writer *tabwriter.Writer) error {
	var typeStr string
	app := rpapiv1.RPApplicationFull{}

	// Cast to the full type in all case but use "full" parameters as indicated by the user
	// We have to pass through the Marshal/Unmarshal tricks to not have problem during conversion from
	// RPApplication -> RPApplicationFull at runtime
	dataBytes, _ := json.Marshal(appInterface)
	json.Unmarshal(dataBytes, &app)

	// Create the string for type
	typeToDisplay := app.Parent.Type
	if typeToDisplay == "external" {
		typeToDisplay = "distro (external)"
	}

	if verbose {
		// Create the string containing all the distributions
		distributions := ""
		for _, distro := range app.Distributions {
			if distro.Set {
				if distributions != "" {
					distributions += ", "
				}
				if distro.Mandatory {
					typeStr = " [Mand]"
				} else {
					typeStr = " [Opt]"
				}
				distributions += (distro.Slug + typeStr)
			}
		}

		// Create the string containing the results
		fmt.Fprintln(writer, appsListCmdVerboseLine0, app.ID.String())
		fmt.Fprintln(writer, appsListCmdVerboseLine1, app.Name)
		fmt.Fprintln(writer, appsListCmdVerboseLine8, app.PackageName)
		fmt.Fprintln(writer, appsListCmdVerboseLine2, app.Slug)
		fmt.Fprintln(writer, appsListCmdVerboseLine15, typeToDisplay)
		fmt.Fprintln(writer, appsListCmdVerboseLine3, app.Parent.ProjectSlug)
		fmt.Fprintln(writer, appsListCmdVerboseLine9, utils.OptionNotSetConv(app.Description))
		fmt.Fprintln(writer, appsListCmdVerboseLine6, distributions)
		fmt.Fprintln(writer, appsListCmdVerboseLine13, utils.OptionMandatoryOptionalConv(app.AutoTests.Set, app.AutoTests.Mandatory))
		fmt.Fprintln(writer, appsListCmdVerboseLine23, utils.OptionOffOnConv(app.IgnoreMissingRedTestRpm))
		fmt.Fprintln(writer, appsListCmdVerboseLine19, utils.OptionMandatoryOptionalConv(app.AutoAudits.Set, app.AutoAudits.Mandatory))
		fmt.Fprintln(writer, appsListCmdVerboseLine20, getAutoSRPMStrFromInt(app))

		fmt.Fprintln(writer, appsListCmdVerboseLine21, utils.OptionNotSetConv(app.AuditsConf))
		fmt.Fprintln(writer, appsListCmdVerboseLine24, utils.OptionOffOnConv(app.IgnoreCheckDepsResult))
		fmt.Fprintln(writer, appsListCmdVerboseLine4, app.CreatedAt.String())
		fmt.Fprintln(writer, appsListCmdVerboseLine5, app.UpdatedAt.String())
		fmt.Fprintln(writer, appsListCmdVerboseLine14, app.Bookmarked)

		if isFull {
			// Print the metadata relative to this app
			extSpec := mainNotSetMsg
			if app.Metadata.Data.SpecfileLocation != "" {
				extSpec = app.Metadata.Data.SpecfileLocation
			}
			specType := string(app.Metadata.Data.SpecfileType)
			fmt.Fprintln(writer, appsListCmdVerboseLine18, app.Metadata.Data.BuildType)
			fmt.Fprintln(writer, appsListCmdVerboseLine10, app.Metadata.Data.SourcesURI)
			fmt.Fprintln(writer, appsListCmdVerboseLine11, app.Metadata.Data.Revision)
			fmt.Fprintln(writer, appsListCmdVerboseLine22, specType)
			fmt.Fprintln(writer, appsListCmdVerboseLine12, extSpec)

			// Print eventual overrides
			if len(app.Overrides) > 0 {
				fmt.Fprintln(writer, appsListCmdVerboseLine16, app.Overrides)
			}

			fmt.Fprintln(writer, appsListCmdVerboseLine17)
			fmt.Fprintln(writer)

			// Convert from map[string]interfaces to RPBCKServices
			// decode first map of interface{}, so it can be encode into RPBCKServices
			jsonbody, err := json.Marshal(app.Services)
			if err != nil {
				return err
			}

			// then convert from map[string]interfaces to RPBCKServices
			services := bckapi.RPBCKServices{}
			if err := json.Unmarshal(jsonbody, &services); err != nil {
				return err
			}

			// Display the services
			if err = displayAppServices(services, verbose, writer); err != nil {
				return err
			}
		}

	} else {
		if isAlone {
			fmt.Fprintln(writer, appsListCmdHeader)
		}
		fmt.Fprintln(writer, app.ID.String(), "\t", app.Name, "\t", app.Slug, "\t", typeToDisplay, "\t", app.Parent.ProjectSlug)
	}

	return nil
}

// displayAppServices display the list of the current services available
// for the specified application
//
// :param services: List of services to display
// :type services: interface{}
// :param verbose: Tell if the print needs to be verbose or not
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func displayAppServices(services interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Convert to the right type
	servicesToPrint := services.(bckapi.RPBCKServices)

	// Display each service
	for serviceName, serviceStruct := range servicesToPrint {
		// Do not display a service that is not used
		if !serviceStruct.Used {
			continue
		}

		fmt.Fprintln(writer, "\t", "-", appsServicesAnsLine1, serviceName)
		fmt.Fprintln(writer, "\t", "-", appsServicesAnsLine2)
		for paramName, param := range serviceStruct.Parameters {
			paramFullString := "    " + paramName + ":\t "

			paramValueAsStr := fmt.Sprintf("%v", param.Value)
			if paramValueAsStr == "" {
				paramValueAsStr = "\"\""
			}

			paramFullString += paramValueAsStr
			fmt.Fprintln(writer, "\t"+paramFullString)
		}
		fmt.Fprintln(writer)
	}

	return nil
}

// rpServices gets the list of all the available services on
// the Redpesk backend
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func rpServices(cmd *cobra.Command, args []string) error {

	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Get the list of every available services
	services := bckapi.RPBCKServices{}
	if err = bckAPI.RPServicesGet(&services); err != nil {
		return err
	}

	// Here, for all services set "Used" to true in order to display it with "printServices()" function
	for _, service := range services {
		service.Used = true
	}

	// Print the result to the user
	err = RPPrinter.CustomPrintObject(services, verbose, printServices)

	return err
}

// printServices display the list of the current services available
// in the current Redpesk instance
//
// :param services: List of services to display
// :type services: interface{}
// :param verbose: Tell if the print needs to be verbose or not
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func printServices(services interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Convert to the right type
	servicesToPrint := services.(bckapi.RPBCKServices)

	if verbose {
		// Display each service
		fmt.Fprintln(writer, appsServicesAnsLine0)
		fmt.Fprintln(writer)
		for serviceName, serviceStruct := range servicesToPrint {
			// Do not display anything if the service is not used
			if !serviceStruct.Used {
				continue
			}

			fmt.Fprintln(writer, appsServicesAnsLine1, serviceName)
			fmt.Fprintln(writer, appsServicesAnsLine4, serviceStruct.Version)
			fmt.Fprintln(writer, appsServicesAnsLine3, serviceStruct.Description)
			fmt.Fprintln(writer, appsServicesAnsLine2)
			for paramName, param := range serviceStruct.Parameters {
				paramFullString := "  " + paramName + "\t [" + param.Type + "]" + "\t => " + param.Description
				fmt.Fprintln(writer, paramFullString)
			}
			fmt.Fprintln(writer)
		}
	} else {
		// Print the header for services
		fmt.Fprintln(writer, appsServicesAnsHeader)
		// Print each service
		for serviceName, serviceStruct := range servicesToPrint {

			// Do not display anything if the service is not used
			if !serviceStruct.Used {
				continue
			}

			// Aggregate all the names of parameters (max 3)
			parametersSlice := []string{}
			index := 0
			for paramName := range serviceStruct.Parameters {
				index++
				if index > 3 {
					break
				}
				parametersSlice = append(parametersSlice, paramName)
			}
			if index > 3 {
				parametersSlice = append(parametersSlice, "etc.")
			}
			parametersStr := strings.Join(parametersSlice, ", ")

			fmt.Fprintln(writer, serviceName, "\t", serviceStruct.Summary, "\t", serviceStruct.Version, "\t", parametersStr)
		}
	}

	return nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Banano's Touch -----------------------------------------------
// --------------------------------------------------------------

// seriousCmdHandler is a serious command
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func seriousCmdHandler(cmd *cobra.Command, args []string) error {

	// Get the flag
	why, _ := cmd.Flags().GetBool("why")
	if !why {
		RPPrinter.CustomPrintLn("Yes I am. But the real question is WHY are you so serious?")
	} else {
		// Split the string by line
		toDisplay := strings.Split(seriousASCIIArt, "\n")
		for _, line := range toDisplay {
			RPPrinter.CustomPrintLn(line)
			if runMode == "testing" {
				time.Sleep(2 * time.Millisecond)
			} else {
				time.Sleep(250 * time.Millisecond)
			}
			RPPrinter.ForceFlush()
		}
		RPPrinter.CustomPrintLn()
	}

	return nil
}

// check if flags's bit is set in auto_srpm_build app's setting
func ShouldAutoRSrpmBuild(appli rpapiv1.RPApplicationFull, flag rpapiv1.AutoSRPMFieldType) bool {
	return appli.AutoRSrpmBuild&flag == flag
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// intToAutoSrpmRebuildOptions converts application's AutoSrpmBuild of type int into a map of type bckapi.RPBCSAutoRebuildSrpmOptions ({option_name: value}) and a map of type {index_option_name: value})
//
// :param app: application for which we are defining/updating auto_srpm_build options
// :returns:autoRebuildSrpmOptionsWithIndex, autoRebuildSrpmOptions :bckapi.RPBCSAutoRebuildSrpmOptions --maps of key-value pairs
func intToAutoSrpmRebuildOptions(app rpapiv1.RPApplicationFull) (bckapi.RPBCSAutoRebuildSrpmOptions, bckapi.RPBCSAutoRebuildSrpmOptions) {

	autoRebuildSrpmOptionsWithIndex := map[string]bool{

		appendPositionToKey(SrpmPackageFileChangeShort):        ShouldAutoRSrpmBuild(app, rpapiv1.AppAutoSRPMBuildPkgChange),
		appendPositionToKey(SrpmMetaFileChangeShort):           ShouldAutoRSrpmBuild(app, rpapiv1.AppAutoSRPMBuildMetadataChange),
		appendPositionToKey(SrpmDistroBranchesManagementShort): ShouldAutoRSrpmBuild(app, rpapiv1.AppAutoSRPMBuildBranchChange),
		appendPositionToKey(SrpmNoValidSrpmBuildShort):         ShouldAutoRSrpmBuild(app, rpapiv1.AppAutoSRPMBuildNoSrpm),
	}

	autoRebuildSrpmOptions := map[string]bool{

		SrpmPackageFileChangeShort:        ShouldAutoRSrpmBuild(app, rpapiv1.AppAutoSRPMBuildPkgChange),
		SrpmMetaFileChangeShort:           ShouldAutoRSrpmBuild(app, rpapiv1.AppAutoSRPMBuildMetadataChange),
		SrpmDistroBranchesManagementShort: ShouldAutoRSrpmBuild(app, rpapiv1.AppAutoSRPMBuildBranchChange),
		SrpmNoValidSrpmBuildShort:         ShouldAutoRSrpmBuild(app, rpapiv1.AppAutoSRPMBuildNoSrpm),
	}

	return autoRebuildSrpmOptionsWithIndex, autoRebuildSrpmOptions

}

// checkIfSrpmRebuildCmdParamExists checks if given auto srpm rebuild option exists in predefined slice of available options
//
// :param option: value to check if it exists in slice
// :returns: true if found in slice and false if not
func checkIfSrpmRebuildCmdParamExists(option string) bool {
	existingSrpmOptions := getSrpmRebuildOptionsNames()
	return slices.Contains(existingSrpmOptions, option)
}

// convertAutoSrpmBuildSettingsToInt converts a map of auto srpm parms of type map of key(string) -> value(bool) pairs into
// an int based on maps values
//
// :param srpmSettings: map of srpm settings key->value pairs
// :return: an int which in binary form represents settings
func convertAutoSrpmBuildSettingsToInt(srpmSettings bckapi.RPBCSAutoRebuildSrpmOptions) int {
	boolSlice := make([]bool, 0, len(srpmSettings))
	result := 0

	// Extract the keys into a slice
	keys := make([]string, 0, len(srpmSettings))
	for key := range srpmSettings {
		keys = append(keys, key)
	}

	// Sort the keys
	sort.Strings(keys)

	// create an array of settings boolean values
	if srpmSettings != nil {
		// Iterate over the sorted keys and access the map
		for _, key := range keys {
			boolSlice = append(boolSlice, srpmSettings[key])
		}
	}

	// convert an array of boolean values into an int
	for i, b := range boolSlice {
		if b {
			result |= 1 << i
		}
	}
	return result
}

// appendPositionToKey adds position to the srpm auto rebuild flag in order to be able to create a map with key-values
// which can always be sorted in the same order. (Existence of the key has been checked already)
//
// :param key: string srpm auto rebuild parameter name
// :return: int number for relative key name
func appendPositionToKey(key string) string {
	switch key {
	case SrpmPackageFileChangeShort:
		return "1_" + key
	case SrpmMetaFileChangeShort:
		return "2_" + key
	case SrpmDistroBranchesManagementShort:
		return "3_" + key
	default:
		return "4_" + key
	}
}

// getAutoSRPMStrFromInt converts an int into a string to display in app verbose mode
//
// :param app: rpapiv1.RPApplicationFull application data
// :return: string comma separated map's key-value pairs
func getAutoSRPMStrFromInt(app rpapiv1.RPApplicationFull) string {
	_, autoSrpmOptionsMap := intToAutoSrpmRebuildOptions(app)
	mJson, _ := json.Marshal(autoSrpmOptionsMap)
	str := string(mJson)
	str = strings.Trim(str, "\"")
	str = strings.Trim(str, "{")
	str = strings.Trim(str, "}")

	return string(str)
}
