/*
Copyright © 2020-2023 IoT.bzh Company
Contact: https://www.iot.bzh/licensing

This file is part of the redpesk-cli project, the redpesk command line tool.

$RP_BEGIN_LICENSE$
Commercial License Usage

Licensees holding valid commercial IoT.bzh licenses may use this file in
accordance with the commercial license agreement provided with the
Software or, alternatively, in accordance with the terms contained in
a written agreement between you and The IoT.bzh Company. For licensing terms
and conditions see https://www.iot.bzh/terms-conditions. For further
information use the contact form at https://www.iot.bzh/contact.

# GNU General Public License Usage

Alternatively, this file may be used under the terms of the GNU General
Public license version 3. This license is as published by the Free Software
Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
of this file. Please review the following information to ensure the GNU
General Public License requirements will be met
https://www.gnu.org/licenses/gpl-3.0.html.

$RP_END_LICENSE$
*/
package cmd

import (
	"errors"
	"fmt"
	"math/rand"
	"os"
	"path"
	"regexp"
	"strconv"
	"strings"
	"testing"
	"time"

	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/rpbckmock"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/utils"
	"git.ovh.iot/redpesk/redpesk-infra/rp-webserver/actions/rpapiv1"
	"github.com/stretchr/testify/assert"
)

const (
	testsLogsDirectory  = "/tmp/rp-tests-logs/"
	buildsLogsDirectory = "/tmp/rp-application-builds-logs/"

	// Applications
	appDisplayNoVerboseColNb                = 5
	appDisplayVerboseFullLineNb             = 23
	appDisplayVerboseFullLineWithServicesNb = 28

	// Application builds stats
	appBuildsStatsNoVerboseLineNb = 7
	appBuildsStatsVerboseColNb    = 3

	// Application tests stats
	appTestsStatsNoVerboseLineNb = 7
	appTestsStatsVerboseColNb    = 3
)

// TestAppsAddNoBlock is the function allowing to test if
// the command "rp-cli applications add --nonblocking ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsAddNoBlock(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("SNCF", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "sncf", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the command
	cmdToRun2 := getMainAddAppCmd("TGV", "train-grande-vitesse", "sncf")
	// Add the non-blocking flag
	cmdToRun2 = append(cmdToRun2, "--nonblocking")

	// Create the application if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, appsCommand, "tgv", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// The application has been created, check it is now in backend

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID":      `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":    `^TGV$`,
		"Slug":    `^tgv$`,
		"Type":    `^standard$`,
		"Project": `^sncf$`,
	}

	// Run the command and the tests (no verbose and verbose, since this is the same result)
	checkCmdAnsKeysRegexpCol(test, appDisplayNoVerboseColNb, regexpMap, appsCommand, "get", "tgv")
}

// TestAppsAddBlock is the function allowing to test if
// the command "rp-cli applications add ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsAddBlock(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("SNCF", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "sncf", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the command
	cmdToRun2 := getMainAddAppCmd("TER", "train-express-regional", "sncf")

	// Create the application if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, appsCommand, "ter", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// The application has been created, check it is now in backend

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID":      `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":    `^TER$`,
		"Slug":    `^ter$`,
		"Type":    `^standard$`,
		"Project": `^sncf$`,
	}

	// Run the command and the tests (no verbose and verbose, since this is the same result)
	checkCmdAnsKeysRegexpCol(test, appDisplayNoVerboseColNb, regexpMap, appsCommand, "get", "ter")
}

// TestAppsAddBlockChangeDist is the function allowing to test if
// the command "rp-cli applications add ..." is working correctly
// when the user specify other distribution than the one used by project
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsAddBlockChangeDist(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("SNCF", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "sncf", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the command
	cmdToRun2 := getMainAddAppCmd("Wagon Bar", "wagon-bar", "sncf")
	cmdToRun2 = append(cmdToRun2, "--optional-distro", "redpesk-devel-arz")

	// Create the application if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, appsCommand, "wagon-bar", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// The application has been created, check it is now in backend

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID":      `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":    `^Wagon Bar$`,
		"Slug":    `^wagon-bar$`,
		"Type":    `^standard$`,
		"Project": `^sncf$`,
	}

	// Run the command and the tests (no verbose and verbose, since this is the same result)
	checkCmdAnsKeysRegexpCol(test, appDisplayNoVerboseColNb, regexpMap, appsCommand, "get", "wagon-bar")
}

// TestAppsAddFromSamplesBlock is the function allowing to test if
// the command "rp-cli applications add-from-sample ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsAddFromSamplesBlock(test *testing.T) {

	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("SNCF", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "sncf", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the command
	cmdToRun2 := getAddFromSampleAppCmd("helloworld-binding", "sncf")

	// Create the application if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, appsCommand, "helloworld-binding", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// The application has been created, check it is now in backend

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID":      `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":    `^Helloworld binding$`,
		"Slug":    `^helloworld-binding$`,
		"Type":    `^standard$`,
		"Project": `^sncf$`,
	}

	// Run the command and the tests (no verbose and verbose, since this is the same result)
	checkCmdAnsKeysRegexpCol(test, appDisplayNoVerboseColNb, regexpMap, appsCommand, "get", "helloworld-binding")
}

// TestAppsAddSetActiveBranch is the function allowing to test if
// the command "rp-cli applications add ..." is working correctly
// when the user specifies an active branch at creation with '--active-branch'
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsAddSetActiveBranch(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("SNCF", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "sncf", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the command
	cmdToRun2 := getMainAddAppCmd("Legende Vivante", "legende-vivante", "sncf")
	// Update the command with the active-branch flag
	cmdToRun2 = append(cmdToRun2, "--active-branch", "master")

	// Create the application if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, appsCommand, "sncf", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// The application has been created, check that the branch has been taken into account
	testActiveAndDefaultBranch(test, "legende-vivante", "sncf", "master", "master")
}

// TestAppsAddWithLocalSpecfile is the function allowing to test if
// the command "rp-cli applications add ..." is working correctly
// when the user specifies a local specfile to send to the factory
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsAddWithLocalSpecfile(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Local Spec File Prj", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "local-spec-file-prj", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create a fake specfile
	err = utils.WriteToFile("/tmp/local-spec-file-app.spec", "COUCOU TU VEUX VOIR MA VERSION?")
	if err != nil {
		test.Fatal(err)
	}
	defer os.Remove("/tmp/local-spec-file-app.spec")

	// Create the command
	cmdToRun2 := getMainAddAppCmd("local-spec-file-app", "local-spec-file-app", "local-spec-file-prj")
	// Update the command with the active-branch flag
	cmdToRun2 = append(cmdToRun2, "--specfile-location", "/tmp/local-spec-file-app.spec")

	// Create the application if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, appsCommand, "local-spec-file-prj", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}
}

// TestAppsList is the function allowing to test if
// the command "rp-cli applications list ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsList(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("SNCF", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "sncf", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the commands
	cmdToRun2 := getMainAddAppCmd("TGV", "train-grande-vitesse", "sncf")
	cmdToRun3 := getMainAddAppCmd("TER", "train-express-regional", "sncf")
	cmdToRun4 := getMainAddAppCmd("Train Corail", "train-corail", "sncf")

	// Create the applications if they do not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, appsCommand, "tgv", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, appsCommand, "ter", cmdToRun3...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, appsCommand, "train-corail", cmdToRun4...)
	if err != nil {
		test.Fatal(err)
	}

	// The applications have been created, check it is now in backend

	// Not verbose mode
	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID":      `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":    `.+`,
		"Slug":    `^[a-z0-9-]+$`,
		"Type":    `^standard$`,
		"Project": `^sncf$`,
	}

	// Run the command and the tests (no verbose and verbose, since this is the same result)
	checkCmdAnsKeysRegexpCol(test, appDisplayNoVerboseColNb, regexpMap, appsCommand, "list", "-p", "sncf")

	// Verbose mode
	// Create the map containing the regexp we want to match in the answer
	regexpMap2 := map[string]string{
		"ID:":                        `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":                       `.+`,
		"Package Name":               `^[a-z0-9-]+$`,
		"Slug":                       `^[a-z0-9-]+$`,
		"Type":                       `^standard$`,
		"Parent project":             `^[a-z0-9-]+$`,
		"Description":                `.+`,
		"Distributions":              `^redpesk-devel-arz.*`,
		"Auto on board tests":        `^(ON|OFF) \[Opt\]$`,
		"Auto security audits tests": `^(ON|OFF) \[Opt\]$`,
		"AutoRSrpmBuild":             `dist_branch":true,"meta_changes":true,"no_srpm_bld":true,"pkg_changes":true`,
		"Audits settings":            `(\[Not set\]|\{"audit":null\})`,
		"Creation time":              `.+`,
		"Update time":                `.+`,
		"Bookmarked":                 `^(true|false)$`,
	}

	// Run the command and the tests
	checkCmdAnswerListKeysRegexp(test, appDisplayVerboseLightLineNb, "", regexpMap2, appsCommand, "list", "-v", "-p", "sncf")
}

// TestAppsListTypeSpecified is the function allowing to test if
// the command "rp-cli applications list ..." is working correctly
// when the flag `--type` is specified
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsListTypeSpecified(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("GESLEIR", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "gesleir", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the commands
	cmdToRun2 := getMainAddAppCmd("Tres Disco", "tres-disco", "gesleir")
	cmdToRun3 := getMainAddAppCmd("Paradis", "paradis", "gesleir")
	cmdToRun4 := getMainAddAppCmd("Saoul", "saoul", "gesleir")

	// Create the applications if they do not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, appsCommand, "tres-disco", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, appsCommand, "paradis", cmdToRun3...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, appsCommand, "saoul", cmdToRun4...)
	if err != nil {
		test.Fatal(err)
	}

	// FIXME: TO REWORK WITH NEW TESTS ON PROJECT RELEASE (ARMAND)

	// // The applications have been created, now release the project
	// checkActionCmdAns(test, adminCommand, "release-project", "gesleir")

	// // Not verbose mode
	// // Create the map containing the regexp we want to match in the answer
	// regexpMap := map[string]string{
	// 	"ID":      `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
	// 	"Name":    `.+`,
	// 	"Slug":    `^[a-z0-9-]+$`,
	// 	"Type":    `^released$`,
	// 	"Project": `^[a-z0-9-]+$`,
	// }

	// // Run the command and the tests
	// checkCmdAnsKeysRegexpCol(test, appDisplayNoVerboseColNb, regexpMap, appsCommand, "list", "-p", "gesleir", "--type", "released")
	// checkCmdAnsKeysRegexpCol(test, appDisplayNoVerboseColNb, regexpMap, appsCommand, "list", "--type", "released")

	// // Verbose mode
	// // Create the map containing the regexp we want to match in the answer
	// regexpMap2 := map[string]string{
	// 	"ID:":               `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
	// 	"Name":              `.+`,
	// 	"Package Name":      `^[a-z0-9-]+$`,
	// 	"Slug":              `^[a-z0-9-]+$`,
	// 	"Type":              `^released$`,
	// 	"Parent project":    `^[a-z0-9-]+$`,
	// 	"Description":       `.+`,
	// 	"Source URL":        `http(s)*://`,
	// 	"Source Revision":   `^[a-z0-9-]+$`,
	// 	"External Specfile": `(\[Not set\]|(http(s)*://))`,
	// 	"Distributions":     `^redpesk-devel-arz.*`,
	// 		"Auto on board tests": `^(ON|OFF) \[Opt\]$`,
	// 	"Build type":    `^(standard_build_cross|emulated_build|accel_build_cross)$`,
	// 	"Creation time":     `.+`,
	// 	"Update time":       `.+`,
	// 	"Bookmarked":        `^(true|false)$`,
	// }

	// // Run the command and the tests
	// checkCmdAnswerListKeysRegexp(test, appDisplayVerboseLightLineNb, "", regexpMap2, appsCommand, "list", "-v", "--type", "released")

	// // Now request type "standard" and "released" and check that we have it all
	// regexpMap["Type"] = `^(released|standard)$`
	// checkCmdAnsKeysRegexpCol(test, appDisplayNoVerboseColNb, regexpMap, appsCommand, "list", "--type", "released", "--type", "standard")
}

// TestAppsGet is the function allowing to test if
// the command "rp-cli teams get ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsGet(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Airbus", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "airbus", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the commands
	cmdToRun2 := getMainAddAppCmd("Avion A320", "a320", "airbus")

	// Create the application if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, appsCommand, "avion-a320", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Verbose only, since the no verbose option is tested in almost all other tests

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":                 `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":                `^Avion A320$`,
		"Package Name":        `^a320$`,
		"Slug":                `^avion-a320$`,
		"Type":                `^standard$`,
		"Parent project":      `^airbus$`,
		"Description":         `.+`,
		"Source URL":          `http(s)*://`,
		"Source Revision":     `^[a-z0-9-]+$`,
		"External Specfile":   `(\[Not set\]|(http(s)*://))`,
		"Specfile type":       `(internal|relative_path|external_url)`,
		"Distributions":       `^redpesk-devel-arz.*`,
		"Auto on board tests": `^(ON|OFF) \[Opt\]$`,
		"Build type":          `^(standard_build_cross|emulated_build|accel_build_cross)$`,
		"Creation time":       `.+`,
		"Update time":         `.+`,
		"Bookmarked":          `^false$`,
		"Services":            `.*`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, appDisplayVerboseFullLineNb, regexpMap, appsCommand, "get", "avion-a320", "-v")
}

// TestAppsGetFromPkgName is the function allowing to test if
// the command "rp-cli apps get ..." is working correctly
// when the user gives a package name
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsGetFromPkgName(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Get From Pkg Name Project", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "get-from-pkg-name-project", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the commands
	cmdToRun2 := getMainAddAppCmd("Pkg Name App", "pkg-name-app-pkgname", "get-from-pkg-name-project")

	// Create the application if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, appsCommand, "pkg-name-app", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":                 `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":                `^Pkg Name App$`,
		"Package Name":        `^pkg-name-app-pkgname$`,
		"Slug":                `^pkg-name-app$`,
		"Type":                `^standard$`,
		"Parent project":      `^get-from-pkg-name-project$`,
		"Description":         `.+`,
		"Source URL":          `http(s)*://`,
		"Source Revision":     `^[a-z0-9-]+$`,
		"External Specfile":   `(\[Not set\]|(http(s)*://))`,
		"Specfile type":       `(internal|relative_path|external_url)`,
		"Distributions":       `^redpesk-devel-arz.*`,
		"Auto on board tests": `^(ON|OFF) \[Opt\]$`,
		"Build type":          `^(standard_build_cross|emulated_build|accel_build_cross)$`,
		"Creation time":       `.+`,
		"Update time":         `.+`,
		"Bookmarked":          `^false$`,
		"Services":            `.*`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, appDisplayVerboseFullLineNb, regexpMap, appsCommand, "get", "pkg-name-app-pkgname", "-v")
}

// TestAppsGetFromName is the function allowing to test if
// the command "rp-cli apps get ..." is working correctly
// when the user gives a name and not a slug
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsGetFromName(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Get From Name Project", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "get-from-name-project", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the commands
	cmdToRun2 := getMainAddAppCmd("Name App", "name-app-name", "get-from-name-project")

	// Create the application if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, appsCommand, "name-app", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":                 `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":                `^Name App$`,
		"Package Name":        `^name-app-name$`,
		"Slug":                `^name-app$`,
		"Type":                `^standard$`,
		"Parent project":      `^get-from-name-project$`,
		"Description":         `.+`,
		"Source URL":          `http(s)*://`,
		"Source Revision":     `^[a-z0-9-]+$`,
		"External Specfile":   `(\[Not set\]|(http(s)*://))`,
		"Specfile type":       `(internal|relative_path|external_url)`,
		"Distributions":       `^redpesk-devel-arz.*`,
		"Auto on board tests": `^(ON|OFF) \[Opt\]$`,
		"Build type":          `^(standard_build_cross|emulated_build|accel_build_cross)$`,
		"Creation time":       `.+`,
		"Update time":         `.+`,
		"Bookmarked":          `^false$`,
		"Services":            `.*`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, appDisplayVerboseFullLineNb, regexpMap, appsCommand, "get", "Name App", "-v")
}

// TestAppsUpdateDesc is the function allowing to test if
// the command "rp-cli application update ..." is working correctly
// when the update is done on the description
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsUpdateDesc(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Airbus", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "airbus", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the commands
	cmdToRun2 := getMainAddAppCmd("Avion A380", "a380", "airbus")

	// Create the application if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, appsCommand, "avion-a380", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the update command
	updateCmd := getUpdateAppCmd("avion-a380", "-d", "Nouvelle description")

	// Run the update command and check that all is [OK]
	checkActionCmdAns(test, updateCmd...)

	// Check that the values have been updated

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":                 `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":                `^Avion A380$`,
		"Package Name":        `^a380$`,
		"Slug":                `^avion-a380$`,
		"Type":                `^standard$`,
		"Parent project":      `^airbus$`,
		"Description":         `^Nouvelle`,
		"Source URL":          `http(s)*://`,
		"Source Revision":     `^master$`,
		"External Specfile":   `(\[Not set\]|(http(s)*://))`,
		"Specfile type":       `(internal|relative_path|external_url)`,
		"Distributions":       `^redpesk-devel-arz.*`,
		"Auto on board tests": `^(ON|OFF) \[Opt\]$`,
		"Build type":          `^(standard_build_cross|emulated_build|accel_build_cross)$`,
		"Creation time":       `.+`,
		"Update time":         `.+`,
		"Bookmarked":          `^false$`,
		"Services":            `.*`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, appDisplayVerboseFullLineNb, regexpMap, appsCommand, "get", "avion-a380", "-v")
}

// TestAppsUpdateName is the function allowing to test if
// the command "rp-cli application update ..." is working correctly
// when the update is done on the name
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsUpdateName(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Update App Name Prj", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "update-app-name-prj", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the commands
	cmdToRun2 := getMainAddAppCmd("App test name update", "app-test-name-update", "update-app-name-prj")

	// Create the application if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, appsCommand, "app-test-name-update", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the update command
	updateCmd := getUpdateAppCmd("app-test-name-update", "-n", "New name for app")

	// Run the update command and check that all is [OK]
	checkActionCmdAns(test, updateCmd...)

	// Check that the values have been updated

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":                 `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":                `^New name for app$`,
		"Package Name":        `^app-test-name-update$`,
		"Slug":                `^app-test-name-update$`,
		"Type":                `^standard$`,
		"Parent project":      `^update-app-name-prj$`,
		"Description":         `^This is a testing application`,
		"Source URL":          `http(s)*://`,
		"Source Revision":     `^master$`,
		"External Specfile":   `(\[Not set\]|(http(s)*://))`,
		"Specfile type":       `(internal|relative_path|external_url)`,
		"Distributions":       `^redpesk-devel-arz.*`,
		"Auto on board tests": `^(ON|OFF) \[Opt\]$`,
		"Build type":          `^(standard_build_cross|emulated_build|accel_build_cross)$`,
		"Creation time":       `.+`,
		"Update time":         `.+`,
		"Bookmarked":          `^false$`,
		"Services":            `.*`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, appDisplayVerboseFullLineNb, regexpMap, appsCommand, "get", "app-test-name-update", "-v")
}

// TestAppsUpdateExtSpecFile is the function allowing to test if
// the command "rp-cli application update ..." is working correctly
// when the update is done on the external specfile URL
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsUpdateExtSpecFile(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Update App Prj", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "update-app-prj", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the commands
	cmdToRun2 := getMainAddAppCmd("App test spec url update", "app-test-spec-url-update", "update-app-prj")

	// Create the application if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, appsCommand, "app-test-spec-url-update", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the update command
	updateCmd := getUpdateAppCmd("app-test-spec-url-update", "--specfile-type", "external_url")
	updateCmd = append(updateCmd, "--specfile-location", "https://new-ext-spec-url.fr/")

	// Run the update command and check that all is [OK]
	checkActionCmdAns(test, updateCmd...)

	// Check that the values have been updated

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":                 `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":                `^App test spec url update$`,
		"Package Name":        `^app-test-spec-url-update$`,
		"Slug":                `^app-test-spec-url-update$`,
		"Type":                `^standard$`,
		"Parent project":      `^update-app-prj$`,
		"Description":         `^This is a testing application`,
		"Source URL":          `http(s)*://`,
		"Source Revision":     `^master$`,
		"External Specfile":   `^https://new-ext-spec-url.fr/$`,
		"Specfile type":       `(internal|relative_path|external_url)`,
		"Distributions":       `^redpesk-devel-arz.*`,
		"Auto on board tests": `^(ON|OFF) \[Opt\]$`,
		"Build type":          `^(standard_build_cross|emulated_build|accel_build_cross)$`,
		"Creation time":       `.+`,
		"Update time":         `.+`,
		"Bookmarked":          `^false$`,
		"Services":            `.*`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, appDisplayVerboseFullLineNb, regexpMap, appsCommand, "get", "app-test-spec-url-update", "-v")
}

// TestAppsUpdateSources is the function allowing to test if
// the command "rp-cli application update ..." is working correctly
// when the update is done on the sources (revision and URL)
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsUpdateSources(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Update App Prj", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "update-app-prj", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the commands
	cmdToRun2 := getMainAddAppCmd("App test src rev update", "app-test-src-rev-update", "update-app-prj")

	// Create the application if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, appsCommand, "app-test-src-rev-update", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the update command
	updateCmd := getUpdateAppCmd("app-test-src-rev-update", "--source-rev", "master-next")
	updateCmd = append(updateCmd, "--source-url", "http://git.ovh.iot/newurl/test-application.git")

	// Run the update command and check that all is [OK]
	checkActionCmdAns(test, updateCmd...)

	// Check that the values have been updated

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":                 `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":                `^App test src rev update$`,
		"Package Name":        `^app-test-src-rev-update$`,
		"Slug":                `^app-test-src-rev-update$`,
		"Type":                `^standard$`,
		"Parent project":      `^update-app-prj$`,
		"Description":         `^This is a testing application`,
		"Source URL":          `http://git\.ovh\.iot/newurl/test-application\.git`,
		"Source Revision":     `^master-next$`,
		"External Specfile":   `(\[Not set\]|(http(s)*://))`,
		"Specfile type":       `(internal|relative_path|external_url)`,
		"Distributions":       `^redpesk-devel-arz.*`,
		"Auto on board tests": `^(ON|OFF) \[Opt\]$`,
		"Build type":          `^(standard_build_cross|emulated_build|accel_build_cross)$`,
		"Creation time":       `.+`,
		"Update time":         `.+`,
		"Bookmarked":          `^false$`,
		"Services":            `.*`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, appDisplayVerboseFullLineNb, regexpMap, appsCommand, "get", "app-test-src-rev-update", "-v")
}

// TestAppsUpdateFailsOnLock is the function allowing to test if
// the command "rp-cli application update ..." has the right behavior
// when the application is locked
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsUpdateFailsOnLock(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Destination Lonely", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "destination-lonely", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the commands
	cmdToRun2 := getMainAddAppCmd("Je men vais", "je-men-vais", "destination-lonely")

	// Create the application if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, appsCommand, "je-men-vais", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Request the mock to put the application in lock mode
	rpbckmock.ChangeAppState(rpapiv1.RPAppLockState, "je-men-vais")
	defer rpbckmock.ChangeAppState(rpapiv1.RPAppStableState, "je-men-vais")

	// Try to update the service of the app
	updateCmd := getUpdateAppCmd("je-men-vais", "--service", "setverrel={'version':false, 'release':true}")

	// First, let's run the wanted command
	_, err = executeCommand(updateCmd...)

	// Check that the answer contains the right error
	assert.Error(test, err, "application currently locked: update impossible")
}

// TestAppsUpdateDeleteDistro is the function allowing to test if
// the command "rp-cli application update ..." has the right behavior
// when the user tries to delete a distribution
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsUpdateDeleteDistro(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getAddPrjCmdOptDistro("update-app-prj-delete-distro", "redpesk-devel-arz,redpesk-devel-batz,redpesk-devel-batz-update")
	err := addObjectIfNotExists(test, prjsCommand, "update-app-prj-delete-distro", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the commands
	cmdToRun2 := getMainAddAppCmd("App update delete distro", "app-update-delete-distro", "update-app-prj-delete-distro")

	// Create the application if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, appsCommand, "app-update-delete-distro", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Check that the distributions have been correctly taken into account

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":                 `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":                `^App update delete distro$`,
		"Package Name":        `^app-update-delete-distro$`,
		"Slug":                `^app-update-delete-distro$`,
		"Type":                `^standard$`,
		"Parent project":      `^update-app-prj-delete-distro$`,
		"Description":         `^This is a testing application`,
		"Source URL":          `http(s)*://`,
		"Source Revision":     `^master$`,
		"External Specfile":   `(\[Not set\]|(http(s)*://))`,
		"Specfile type":       `(internal|relative_path|external_url)`,
		"Distributions":       `^redpesk-devel-arz \[Opt\], redpesk-devel-batz \[Opt\], redpesk-devel-batz-update \[Opt\]$`,
		"Auto on board tests": `^(ON|OFF) \[Opt\]$`,
		"Build type":          `^(standard_build_cross|emulated_build|accel_build_cross)$`,
		"Creation time":       `.+`,
		"Update time":         `.+`,
		"Bookmarked":          `^false$`,
		"Services":            `.*`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, appDisplayVerboseFullLineNb, regexpMap, appsCommand, "get", "app-update-delete-distro", "-v")

	// Now delete one optional distribution
	updateCmd := getUpdateAppCmd("app-update-delete-distro", "--optional-distro", "redpesk-devel-arz,redpesk-devel-batz")

	// Run the update command and check that all is [OK]
	checkActionCmdAns(test, updateCmd...)

	// Check that the right value has been updated
	regexpMap["Distributions"] = `^redpesk-devel-arz \[Opt\], redpesk-devel-batz \[Opt\]$`

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, appDisplayVerboseFullLineNb, regexpMap, appsCommand, "get", "app-update-delete-distro", "-v")
}

// TestAppsUpdateChangeBranch is the function allowing to test if
// the command "rp-cli application update ..." has the right behavior
// when the user tries to change the current active banch
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsUpdateChangeBranch(test *testing.T) {
	// Retrieve the project and app names to use for this test
	localPrj, localApp := getPrjAndAppNames()

	// Create an application and a project in order to start testing the branches
	err := createPrjAndApp(test, localPrj, "redpesk-lts-arz", localApp, localApp)
	if err != nil {
		test.Fatal(err)
	}

	// Create a new branch on this app
	output, err := createBranchOnApp(test, localPrj, localApp, "main", "first-new-branch", false, false, []string{})
	if err != nil {
		test.Fatal(err)
	}

	// Check that the answer is the expected one
	assert.Contains(test, output, fmt.Sprintf(appsBranchCmdAnsLine2, "first-new-branch", localApp))

	// Ask rp-cli to update the application to set the new branch as the active one
	updateCmd := getUpdateAppCmd(localApp, "-b", "first-new-branch")
	output, err = executeCommand(updateCmd...)
	if err != nil {
		test.Fatal(err)
	}

	// Check that the answer is the expected one
	assert.Contains(test, output, fmt.Sprintf(appsBranchCmdAnsLine0, "first-new-branch", localApp))
}

// TestAppsDelete is the function allowing to test if
// the command "rp-cli applications delete ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsDelete(test *testing.T) {

	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("SNCF", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "sncf", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the commands
	cmdToRun2 := getMainAddAppCmd("TGV", "train-grande-vitesse", "sncf")
	cmdToRun3 := getMainAddAppCmd("TER", "train-express-regional", "sncf")
	cmdToRun4 := getMainAddAppCmd("Train Corail", "train-corail", "sncf")

	// Create the applications if they do not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, appsCommand, "tgv", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, appsCommand, "ter", cmdToRun3...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, appsCommand, "train-corail", cmdToRun4...)
	if err != nil {
		test.Fatal(err)
	}

	// The applications have been created, now try to delete one

	// Run the delete command and check that all is [OK]
	checkActionCmdAns(test, appsCommand, "delete", "train-corail")

	// Check that the application deleted does not exist anymore
	_, err = executeCommand(appsCommand, "get", "train-corail")
	assert.NotNil(test, err)
}

// TestAppsBookmark is the function allowing to test if
// the command "rp-cli applications bookmark ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsBookmark(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Airbus", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "airbus", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Then, creates an application if it does not exists
	cmdToRun2 := getMainAddAppCmd("Avion A320", "a320", "airbus")
	err = addObjectIfNotExists(test, appsCommand, "avion-a320", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Run the bookmark command and check that all is [OK]
	checkActionCmdAns(test, appsCommand, "bookmark", "avion-a320")

	// Check that the values have been updated

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":                 `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":                `^Avion A320$`,
		"Package Name":        `^a320$`,
		"Slug":                `^avion-a320$`,
		"Type":                `^standard$`,
		"Parent project":      `^airbus$`,
		"Description":         `.*`,
		"Source URL":          `http(s)*://`,
		"Source Revision":     `^master.*$`,
		"External Specfile":   `(\[Not set\]|(http(s)*://))`,
		"Specfile type":       `(internal|relative_path|external_url)`,
		"Distributions":       `^redpesk-devel-arz.*`,
		"Auto on board tests": `^(ON|OFF) \[Opt\]$`,
		"Build type":          `^(standard_build_cross|emulated_build|accel_build_cross)$`,
		"Creation time":       `.+`,
		"Update time":         `.+`,
		"Bookmarked":          `^true$`,
		"Services":            `.*`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, appDisplayVerboseFullLineNb, regexpMap, appsCommand, "get", "avion-a320", "-v")

	// Run the reset bookmark command and check that all is [OK]
	checkActionCmdAns(test, appsCommand, "bookmark", "avion-a320", "--reset")

	// Change slightly the regexpMap in order to check if the bookmark has been reset
	regexpMap["Bookmarked"] = `^false$`

	// Check if the last command has been taken into account
	checkCmdAnswerKeysRegexp(test, appDisplayVerboseFullLineNb, regexpMap, appsCommand, "get", "avion-a320", "-v")
}

// TestAppsDisplayBookmarked is the function allowing to test if
// the command "rp-cli applications list --bookmarked" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsDisplayBookmarked(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Airbus", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "airbus", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	cmdToRun2 := getMainAddAppCmd("Avion A320", "a320", "airbus")
	cmdToRun3 := getMainAddAppCmd("Avion A380", "a380", "airbus")
	cmdToRun4 := getMainAddAppCmd("Avion A350", "a350", "airbus")

	// Create the applications if they do not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, appsCommand, "avion-a320", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, appsCommand, "avion-a380", cmdToRun3...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, appsCommand, "avion-a350", cmdToRun4...)
	if err != nil {
		test.Fatal(err)
	}

	// Bookmark 2 out of 3
	checkActionCmdAns(test, appsCommand, "bookmark", "avion-a380")
	checkActionCmdAns(test, appsCommand, "bookmark", "avion-a350")

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":                        `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":                       `.+`,
		"Package Name":               `^a3[58]0$`,
		"Slug":                       `^avion-a3[58]0$`,
		"Type":                       `^standard$`,
		"Parent project":             `^[a-z0-9-]+$`,
		"Description":                `.+`,
		"Distributions":              `^redpesk-devel-arz.*`,
		"Auto on board tests":        `^(ON|OFF) \[Opt\]$`,
		"Auto security audits tests": `^(ON|OFF) \[Opt\]$`,
		"AutoRSrpmBuild":             `dist_branch":true,"meta_changes":true,"no_srpm_bld":true,"pkg_changes":true`,
		"Audits settings":            `(\[Not set\]|\{"audit":null\})`,
		"Creation time":              `.+`,
		"Update time":                `.+`,
		"Bookmarked":                 `^true$`,
	}

	// Run the command and tests that only bookmarked applications are displayed
	checkCmdAnswerListKeysRegexp(test, appDisplayVerboseLightLineNb, "", regexpMap, appsCommand, "list", "-v", "-b")

	// Reset the bookmarks on the applications
	checkActionCmdAns(test, appsCommand, "bookmark", "avion-a380", "--reset")
	checkActionCmdAns(test, appsCommand, "bookmark", "avion-a350", "--reset")
}

// TestAppsBuild is the function allowing to test if
// the command "rp-cli applications build ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsBuildNotVerbose(test *testing.T) {

	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("SNCF", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "sncf", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Then, creates an application if it does not exists
	cmdToRun2 := getMainAddAppCmd("TGV", "train-grande-vitesse", "sncf")
	err = addObjectIfNotExists(test, appsCommand, "tgv", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// The applications have been created, now try to build it
	outputCmd, err := executeCommand(appsCommand, "build", "tgv")
	if err != nil {
		test.Fatal(err)
	}

	// Check the output of the build
	checkNotVerboseBuildOutput(test, outputCmd)

	if bckIsReal {
		// Create a false spec file
		falseSpecFile, err := createFakeSpecfile("train-grande-vitesse.spec")
		if err != nil {
			test.Fatal(err)
		}

		// Delete the false file
		defer deleteFalseFile(falseSpecFile)

		// In the case of real backend, we need to upload a new spec file each time
		checkActionCmdAns(test, appsCommand, "upload", "tgv", "--file-path", falseSpecFile)

		// Wait a bit
		time.Sleep(500 * time.Millisecond)
	}

	// Try to build it when specifying a project
	outputCmd, err = executeCommand(appsCommand, "build", "tgv", "-p", "sncf")
	if err != nil {
		test.Fatal(err)
	}

	// Check the output of the build
	checkNotVerboseBuildOutput(test, outputCmd)
}

// TestAppsBuildVerbose is the function allowing to test if
// the command "rp-cli applications build <appID> -v" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsBuildVerbose(test *testing.T) {

	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Airbus", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "airbus", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Then, creates an application if it does not exists
	cmdToRun2 := getMainAddAppCmd("Avion A320", "a320", "airbus")
	err = addObjectIfNotExists(test, appsCommand, "avion-a320", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// The applications have been created, now try to build it

	// Check that the application deleted does not exist anymore
	outputCmd, err := executeCommand(appsCommand, "build", "avion-a320")
	if err != nil {
		test.Fatal(err)
	}

	// Check the output of the build
	checkNotVerboseBuildOutput(test, outputCmd)
}

// TestAppsBuildSpecifyDistro is the function allowing to test if
// the command "rp-cli applications build <appID> -d <distroID>" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsBuildSpecifyDistro(test *testing.T) {

	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Peugeot", "redpesk-lts-arz")
	err := addObjectIfNotExists(test, prjsCommand, "peugeot", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Then, creates an application if it does not exists
	cmdToRun2 := getMainAddAppCmd("308 break", "308-break", "peugeot")
	err = addObjectIfNotExists(test, appsCommand, "308-break", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Update the project to add a distribution that won't be available for the app
	cmdToRun3 := getUpdatePrjCmd("peugeot", "--mandatory-distro", "redpesk-lts-arz")
	cmdToRun3 = append(cmdToRun3, "--optional-distro", "redpesk-devel-batz")

	// Run the update command and check that all is [OK]
	checkActionCmdAns(test, cmdToRun3...)

	// Try to build with the wrong distribution (to check log message)
	_, err = executeCommand(appsCommand, "build", "308-break", "-d", "redpesk-devel-arz")
	if assert.Error(test, err) {
		errMsgToHave := fmt.Sprintf(distroNotAvailableMsg, "redpesk-devel-arz", "308-break")
		assert.Contains(test, err.Error(), errMsgToHave)
	}

	// Try to build with the wrong distribution (to check log message)
	_, err = executeCommand(appsCommand, "build", "308-break", "-d", "redpesk-devel-batz")
	if assert.Error(test, err) {
		errMsgToHave := fmt.Sprintf(distroNotAvailableMsg, "redpesk-devel-batz", "308-break")
		assert.Contains(test, err.Error(), errMsgToHave)
	}

	// The applications have been created, now try to build it
	outputCmd, err := executeCommand(appsCommand, "build", "308-break", "-d", "redpesk-lts-arz")
	if err != nil {
		test.Fatal(err)
	}

	// Check the output of the build
	checkNotVerboseBuildOutput(test, outputCmd)

	// Now try to build it with the distribution UUID
	outputCmd, err = executeCommand(appsCommand, "build", "308-break", "-d", "d2264369-29df-4475-b352-7d8a6df9a760")
	if err != nil {
		test.Fatal(err)
	}

	// Check the output of the build
	checkNotVerboseBuildOutput(test, outputCmd)
}

// TestAppsBuildNotVerboseSRpm is the function allowing to test if
// the command "rp-cli applications build ..." is working correctly
// with the flag 'srpm'
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsBuildNotVerboseSRpm(test *testing.T) {

	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Thales", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "thales", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Then, creates an application if it does not exists
	cmdToRun2 := getMainAddAppCmd("C4ISTAR", "c4istar", "thales")
	err = addObjectIfNotExists(test, appsCommand, "c4istar", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// The applications have been created, now try to build it

	// Check that the application deleted does not exist anymore
	outputCmd, err := executeCommand(appsCommand, "build", "c4istar", "-p", "thales", "--srpm")
	if err != nil {
		test.Fatal(err)
	}

	// Check the output of the build
	checkNotVerboseBuildOutput(test, outputCmd)
}

// TestAppsBuildNotVerboseCheckDeps is the function allowing to test if
// the command "rp-cli applications build ..." is working correctly
// with the flag 'check-deps'
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsBuildNotVerboseCheckDeps(test *testing.T) {

	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Texas Instrument", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "texas-instrument", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Then, creates an application if it does not exists
	cmdToRun2 := getMainAddAppCmd("AM62X", "am62x", "texas-instrument")
	err = addObjectIfNotExists(test, appsCommand, "am62x", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// The applications have been created, now try to build it

	// Check that the application deleted does not exist anymore
	outputCmd, err := executeCommand(appsCommand, "build", "am62x", "-p", "texas-instrument", "-e")
	if err != nil {
		test.Fatal(err)
	}

	// Check the output of the build
	checkNotVerboseBuildOutput(test, outputCmd)
}

// TestAppsBuildsList is the function allowing to test if
// the command "rp-cli applications builds list <appID> ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsBuildsList(test *testing.T) {

	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("SNCF", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "sncf", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Then, creates an application if it does not exists
	cmdToRun2 := getMainAddAppCmd("TGV", "train-grande-vitesse", "sncf")
	err = addObjectIfNotExists(test, appsCommand, "tgv", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	var falseSpecFile string
	if bckIsReal {
		// Create a false spec file
		falseSpecFile, err = createFakeSpecfile("train-grande-vitesse.spec")
		if err != nil {
			test.Fatal(err)
		}

		// Delete the false file
		defer deleteFalseFile(falseSpecFile)

		// In the case of real backend, we need to upload a new spec file each time
		checkActionCmdAns(test, appsCommand, "upload", "tgv", "--file-path", falseSpecFile)

		// Wait a bit
		time.Sleep(500 * time.Millisecond)
	}

	// Make sure to have at least three builds for this application (in backend mocked mode)
	// If we are with a real backend, just rebuild the application once
	_, err = executeCommand(appsCommand, "build", "tgv", "-p", "sncf")
	if err != nil {
		test.Fatal(err)
	}
	if !bckIsReal {
		_, err = executeCommand(appsCommand, "build", "tgv", "-p", "sncf")
		if err != nil {
			test.Fatal(err)
		}

		_, err = executeCommand(appsCommand, "build", "tgv", "-p", "sncf")
		if err != nil {
			test.Fatal(err)
		}
	}

	// Run the "builds list" command an check the output
	checkNotVerboseListBuildsOutput(test, appsCommand, "builds", "list", "tgv")

	// Run the "builds list -v" command an check the output
	checkVerboseListBuildsOutput(test, rpapiv1.RPBuildApp, appsCommand, "builds", "list", "tgv", "-v")

	// Run the "builds list -p ..." command an check the output
	checkNotVerboseListBuildsOutput(test, appsCommand, "builds", "list", "tgv", "-p", "sncf")
}

// TestAppsBuildsSRpmList is the function allowing to test if
// the command "rp-cli applications builds list <appID> --srpm ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsBuildsSRpmList(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Opel", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "opel", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Then create an application in this project
	cmdToRun := getMainAddAppCmd("Opel Astra Tourer", "opel-astra-tourer", "opel")
	err = addObjectIfNotExists(test, appsCommand, "opel-astra-tourer", cmdToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// Run the "builds list" command an check the output
	checkNotVerboseListBuildsOutput(test, appsCommand, "builds", "list", "opel-astra-tourer", "--srpm")

	// Run the "builds list -v" command an check the output
	checkVerboseListBuildsOutput(test, rpapiv1.RPBuildSrpm, appsCommand, "builds", "list", "opel-astra-tourer", "--srpm", "-v")
}

// TestAppsBuildsLogs is the function allowing to test if
// the command "rp-cli applications logs-logs <appID> ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsBuildsLogs(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Vendredi sur Mer", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "vendredi-sur-mer", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Then, creates an application if it does not exists
	cmdToRun2 := getMainAddAppCmd("Mort-Fine", "mort-fine", "vendredi-sur-mer")
	err = addObjectIfNotExists(test, appsCommand, "mort-fine", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Firstly, check if the "mort-fine" application exists and if it has builds
	buildID, err := checkIfBuildsExist("mort-fine", appsCommand)
	if err != nil {
		test.Fatal(err)
	}

	// Check if at least one build exists
	if buildID == "" {
		// No build exists, creates one
		_, err = executeCommand(appsCommand, "build", "mort-fine", "-p", "vendredi-sur-mer")
		if err != nil {
			test.Fatal(err)
		}
	}

	// Now try to download build-logs from the server
	cmdOutput, err := executeCommand(appsCommand, "builds", "logs", "mort-fine", "-p", "vendredi-sur-mer", "-d", "/tmp")
	if err != nil {
		test.Fatal(err)
	}

	// Check that rp-cli got the latest build
	latestBuildRetrieved := assert.Contains(test, cmdOutput, "The build logs for build \"")

	ourBuildID := ""
	if latestBuildRetrieved {
		// Get the buildID used
		lineWanted := utils.ExtractLineWithPrefix(cmdOutput, "The build logs for build \"")
		splitedLine := strings.Split(lineWanted, "\"")
		ourBuildID = strings.TrimSpace(splitedLine[1])
	} else {
		test.Fatal()
	}

	// Check that the output contains the needed element
	assert.Contains(test, cmdOutput, "Downloading logs for subtask \"x86_64\"...")
	assert.Contains(test, cmdOutput, "Downloading logs for subtask \"aarch64\"...")

	// Check that both files (build.log and root.log) are available
	assert.FileExists(test, path.Join("/tmp", "rp-application-builds-logs", ourBuildID, "x86_64", "build.log"))
	assert.FileExists(test, path.Join("/tmp", "rp-application-builds-logs", ourBuildID, "x86_64", "root.log"))
	assert.FileExists(test, path.Join("/tmp", "rp-application-builds-logs", ourBuildID, "aarch64", "build.log"))
	assert.FileExists(test, path.Join("/tmp", "rp-application-builds-logs", ourBuildID, "aarch64", "root.log"))

	// Now try to download build-logs from the server but ask for zip logs
	_, err = executeCommand(appsCommand, "builds", "logs", "mort-fine", "-p", "vendredi-sur-mer", "-d", "/tmp", "--zip")
	if err != nil {
		test.Fatal(err)
	}

	// Add to list the zip file to be deleted at teardown time
	zipName := path.Join("/tmp", "rp-application-builds-logs-"+ourBuildID+".zip")
	filesListToDel = append(filesListToDel, zipName)

	// Check that a zip file has been correctly created
	assert.FileExists(test, zipName)
}

// TestAppsBuildsSRpmLogs is the function allowing to test if
// the command "rp-cli applications logs-logs <appID> ..." is working correctly
// with SRPM builds
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsBuildsSRpmLogs(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Opel", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "opel", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Then create an application in this project
	cmdToRun := getMainAddAppCmd("Opel Vectra", "opel-vectra", "opel")
	err = addObjectIfNotExists(test, appsCommand, "opel-vectra", cmdToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// Extract the latest build ID from the build list
	lowestID, err := getLowestIDFromOutput("Build ID", appsCommand, "builds", "list", "opel-vectra", "--srpm")
	if err != nil {
		test.Fatal(err)
	}

	// Now try to download build-logs from the server
	cmdOutput, err := executeCommand(appsCommand, "builds", "logs", lowestID, "-d", "/tmp")
	if err != nil {
		test.Fatal(err)
	}

	// Check that rp-cli got the latest build
	latestBuildRetrieved := assert.Contains(test, cmdOutput, "The build logs for build \"")

	ourBuildID := ""
	if latestBuildRetrieved {
		// Get the buildID used
		lineWanted := utils.ExtractLineWithPrefix(cmdOutput, "The build logs for build \"")
		splitedLine := strings.Split(lineWanted, "\"")
		ourBuildID = strings.TrimSpace(splitedLine[1])
	} else {
		test.Fatal()
	}

	// Check that the output contains the needed element
	assert.Contains(test, cmdOutput, "Downloading logs for subtask \"srpm\"...")

	// Check that both files (build.log and root.log) are available
	assert.FileExists(test, path.Join("/tmp", "rp-application-builds-logs", ourBuildID, "srpm", "build.log"))
	assert.FileExists(test, path.Join("/tmp", "rp-application-builds-logs", ourBuildID, "srpm", "root.log"))
}

// TestAppsBuildsGet is the function allowing to test if
// the command "rp-cli applications builds get <buildID> ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsBuildsGet(test *testing.T) {

	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("SNCF", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "sncf", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Then, creates an application if it does not exists
	cmdToRun2 := getMainAddAppCmd("TGV", "train-grande-vitesse", "sncf")
	err = addObjectIfNotExists(test, appsCommand, "tgv", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	var falseSpecFile string
	if bckIsReal {
		// Create a false spec file
		falseSpecFile, err = createFakeSpecfile("train-grande-vitesse.spec")
		if err != nil {
			test.Fatal(err)
		}

		// Delete the false file
		defer deleteFalseFile(falseSpecFile)

		// In the case of real backend, we need to upload a new spec file each time
		checkActionCmdAns(test, appsCommand, "upload", "tgv", "--file-path", falseSpecFile)

		// Wait a bit
		time.Sleep(500 * time.Millisecond)
	}

	// Make sure to have at least three builds for this application (in backend mocked mode)
	// If we are with a real backend, just rebuild the application once
	_, err = executeCommand(appsCommand, "build", "tgv", "-p", "sncf", "-c", "Build comment n°1")
	if err != nil {
		test.Fatal(err)
	}
	if !bckIsReal {
		_, err = executeCommand(appsCommand, "build", "tgv", "-p", "sncf", "-c", "Build comment n°2")
		if err != nil {
			test.Fatal(err)
		}

		_, err = executeCommand(appsCommand, "build", "tgv", "-p", "sncf", "-c", "Build comment n°3")
		if err != nil {
			test.Fatal(err)
		}
	}

	// Extract the latest build ID from the build list
	highestID, err := getHighestIDFromOutput("Build ID", appsCommand, "builds", "list", "tgv")
	if err != nil {
		test.Fatal(err)
	}

	// Create the regexp map used to check (not verbose mode)
	regexpMap := map[string]string{
		"Build ID:":               `[0-9]{4}`,
		"Status:":                 `^(done|failed)$`,
		"Subtask x86_64 status:":  `^(closed|failed)$`,
		"Subtask aarch64 status:": `^(closed|failed)$`,
	}

	// Run the "builds get" command an check the output
	checkCmdAnswerKeysRegexp(test, 5, regexpMap, appsCommand, "builds", "get", highestID)

	// Create the regexp map used to check (verbose mode)
	regexpMap2 := map[string]string{
		"Build ID:":                   `[0-9]{4}`,
		"Build type":                  `application build`,
		"Application ID:":             `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Application slug:":           `^tgv$`,
		"Distribution:":               `^redpesk-devel-arz$`,
		"Built by":                    `.*`,
		"Status:":                     `^(done|failed)$`,
		"Subtask:":                    `^(srpm|x86_64|aarch64)$`,
		"Subtask state:":              `^(closed|failed)$`,
		"Subtask method:":             `^(buildSRPMFromSCM|buildArch|buildArchCross)$`,
		"Subtask method description:": `.+`,
		"Comment:":                    `^Build comment n°[0-9]{1}`,
	}

	// Run the "builds get" command an check the output
	output := checkCmdAnswerKeysRegexp(test, 37, regexpMap2, appsCommand, "builds", "get", highestID, "-v")

	// Assert that the "tag" for the log files and rpms are existing
	assert.Contains(test, output, "[Log files]")
	assert.Contains(test, output, "[RPM files]")
}

// TestAppsBuildsSRpmGet is the function allowing to test if
// the command "rp-cli applications builds get <buildID> ..." is working correctly
// with a SRpm build
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsBuildsSRpmGet(test *testing.T) {

	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Opel", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "opel", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Then create an application in this project
	cmdToRun := getMainAddAppCmd("Opel Astra", "opel-astra", "opel")
	err = addObjectIfNotExists(test, appsCommand, "opel-astra", cmdToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// Extract the latest build ID from the build list
	lowestID, err := getLowestIDFromOutput("Build ID", appsCommand, "builds", "list", "opel-astra", "--srpm")
	if err != nil {
		test.Fatal(err)
	}

	// Create the regexp map used to check (not verbose mode)
	regexpMap := map[string]string{
		"Build ID:":            `[0-9]{4}`,
		"Status:":              `^(done|failed)$`,
		"Subtask srpm status:": `^(closed|failed)$`,
		"Subtask deps status:": `^(closed|failed)$`,
	}

	// Run the "builds get" command an check the output
	checkCmdAnswerKeysRegexp(test, 6, regexpMap, appsCommand, "builds", "get", lowestID)

	// Create the regexp map used to check (verbose mode)
	regexpMap2 := map[string]string{
		"Build ID:":                   `[0-9]{4}`,
		"Build type":                  `^application SRPM build$`,
		"Application ID:":             `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Application slug:":           `^opel-astra$`,
		"Distribution:":               `^redpesk-devel-arz$`,
		"Built by":                    `.*`,
		"Status:":                     `^(done|failed)$`,
		"Subtask:":                    `^(srpm|deps|)$`,
		"Subtask state:":              `^(closed|failed)$`,
		"Subtask method:":             `^(buildSRPMFromSCM|check-deps|waitrepo)$`,
		"Subtask method description:": `.+`,
		"Comment:":                    `^\[Not set\]$`,
	}

	// Run the "builds get" command an check the output
	output := checkCmdAnswerKeysRegexp(test, 28, regexpMap2, appsCommand, "builds", "get", lowestID, "-v")

	// Assert that the "tag" for the log files and rpms are existing
	assert.Contains(test, output, "[Log files]")
	assert.Contains(test, output, "[SRPM files]")
}

// TestAppsBuildsCancel is the function allowing to test if
// the command "rp-cli applications builds cancel <appID> ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsBuildsCancel(test *testing.T) {

	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Dombrance", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "dombrance", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Then, creates an application if it does not exists
	cmdToRun2 := getMainAddAppCmd("Giscard", "giscard", "dombrance")
	err = addObjectIfNotExists(test, appsCommand, "giscard", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Say to the mocked backend that we will cancel the build
	rpbckmock.RequestNextBuildCanceled()

	// Start a build without waiting for the end
	_, err = executeCommand(appsCommand, "build", "giscard", "-p", "dombrance", "--nonblocking")
	if err != nil {
		test.Fatal(err)
	}

	// Right after request for the build to be canceled
	checkActionCmdAns(test, appsCommand, genericBuildsCmd, "cancel", "giscard", "-p", "dombrance")

	// Wait a bit for the cancel request got processed
	time.Sleep(40 * time.Millisecond)

	// Extract the latest build ID from the build list
	highestID, err := getHighestIDFromOutput("Build ID", appsCommand, "builds", "list", "giscard")
	if err != nil {
		test.Fatal(err)
	}

	// Create the regexp map used to check (not verbose mode)
	regexpMap := map[string]string{
		"Build ID:":               `[0-9]{4}`,
		"Status:":                 `^canceled$`,
		"Subtask x86_64 status:":  `^(closed|failed|canceled)$`,
		"Subtask aarch64 status:": `^(closed|failed|canceled)$`,
	}

	// Run the "builds get" command to check that the build has been correctly canceled
	checkCmdAnswerKeysRegexp(test, 5, regexpMap, appsCommand, "builds", "get", highestID)
}

// TestSeriousCmd is the function allowing to test if
// the command "rp-cli serious" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestSeriousCmd(test *testing.T) {
	executeCommand("serious")
	executeCommand("serious", "--why")
}

// TestAppsTestNotVerbose is the function allowing to test if
// the command "rp-cli applications test ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsTestNotVerbose(test *testing.T) {

	// Firstly, check if the "tgv" application exists and if it has builds
	buildID, err := checkIfBuildsExist("tgv", appsCommand)
	if err != nil {
		test.Fatal(err)
	}

	// Check if a buildID have been found
	if buildID == "" {
		// No build ID have been found, run a build test to have some builds to test
		TestAppsBuildNotVerbose(test)
	}

	// Run a test
	cmdOutput, err := executeCommand(appsCommand, "test", "tgv")
	if err != nil {
		test.Fatal(err)
	}

	// Check the output
	checkTestOutput(test, cmdOutput, "application", false)

	// Re-run a test, with the project flag
	cmdOutput, err = executeCommand(appsCommand, "test", "tgv", "-p", "sncf")
	if err != nil {
		test.Fatal(err)
	}

	// Check the output
	checkTestOutput(test, cmdOutput, "application", false)
}

// TestAppsTestVerbose is the function allowing to test if
// the command "rp-cli applications test -v ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsTestVerbose(test *testing.T) {

	buildID := ""
	var err error
	for buildID == "" {
		// Firstly, check if the "avion-a320" application exists and if it has builds
		buildID, err = checkIfBuildsExist("avion-a320", appsCommand)
		if err != nil {
			test.Fatal(err)
		}
		TestAppsBuildVerbose(test)
	}

	// Run a test
	cmdOutput, err := executeCommand(appsCommand, "test", "avion-a320", "-v")
	if err != nil {
		test.Fatal(err)
	}

	// Check the output
	checkTestOutput(test, cmdOutput, "application", true)

	// Re-run a test, with the build flag
	buildID, err = checkIfBuildsExist("avion-a320", appsCommand)
	if err != nil {
		test.Fatal(err)
	}
	cmdOutput, err = executeCommand(appsCommand, "test", buildID)
	if err != nil {
		test.Fatal(err)
	}

	// Check the output
	checkTestOutput(test, cmdOutput, "application", false)
}

// TestAppsTestsList is the function allowing to test if
// the command "rp-cli applications tests-list <appID> ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsTestsList(test *testing.T) {

	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("SNCF", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "sncf", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Then, creates an application if it does not exists
	cmdToRun2 := getMainAddAppCmd("TGV", "train-grande-vitesse", "sncf")
	err = addObjectIfNotExists(test, appsCommand, "tgv", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Firstly, check if the "avion-a320" application exists and if it has builds
	buildID, err := checkIfBuildsExist("tgv", appsCommand)
	if err != nil {
		test.Fatal(err)
	}

	// Check if at least one build exists
	if buildID == "" {
		// No build exists, creates one
		_, err = executeCommand(appsCommand, "build", "tgv", "-p", "sncf")
		if err != nil {
			test.Fatal(err)
		}
	}

	// Make sure to have at least three tests for this application
	_, err = executeCommand(appsCommand, "test", "tgv", "-b", "qemu-system-x86_64")
	if err != nil {
		test.Fatal(err)
	}
	_, err = executeCommand(appsCommand, "test", "tgv", "-b", "qemu-system-x86_64")
	if err != nil {
		test.Fatal(err)
	}
	_, err = executeCommand(appsCommand, "test", "tgv", "-b", "qemu-system-x86_64")
	if err != nil {
		test.Fatal(err)
	}

	// Run the "tests list" command an check the output
	checkNotVerboseListTestsOutput(test, appsCommand, "tests", "list", "tgv")

	// Run the "tests list -v" command an check the output
	checkVerboseListTestsOutput(test, false, appsCommand, "tests", "list", "tgv", "-v")

	// Run the "tests list -p ..." command an check the output
	checkNotVerboseListTestsOutput(test, appsCommand, "tests", "list", "tgv", "-p", "sncf")
}

// TestAppsTestsLogs is the function allowing to test if
// the command "rp-cli applications test-logs <appID> ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsTestsLogs(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Vendredi sur Mer", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "vendredi-sur-mer", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Then, creates an application if it does not exists
	cmdToRun2 := getMainAddAppCmd("Mort-Fine", "mort-fine", "vendredi-sur-mer")
	err = addObjectIfNotExists(test, appsCommand, "mort-fine", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Firstly, check if the "mort-fine" application exists and if it has builds
	buildID, err := checkIfBuildsExist("mort-fine", appsCommand)
	if err != nil {
		test.Fatal(err)
	}

	// Check if at least one build exists
	if buildID == "" {
		// No build exists, creates one
		_, err = executeCommand(appsCommand, "build", "mort-fine", "-p", "vendredi-sur-mer")
		if err != nil {
			test.Fatal(err)
		}
	}

	// Make sure to have at least three tests for this application
	_, err = executeCommand(appsCommand, "test", "mort-fine")
	if err != nil {
		test.Fatal(err)
	}
	_, err = executeCommand(appsCommand, "test", "mort-fine")
	if err != nil {
		test.Fatal(err)
	}
	_, err = executeCommand(appsCommand, "test", "mort-fine")
	if err != nil {
		test.Fatal(err)
	}

	// Now try to download test-logs from the server
	cmdOutput, err := executeCommand(appsCommand, "tests", "logs", "mort-fine", "-p", "vendredi-sur-mer", "-d", "/tmp")
	if err != nil {
		test.Fatal(err)
	}

	// Check that rp-cli got the latest test
	latestTestRetrieved := assert.Contains(test, cmdOutput, "Downloading logs for test")

	ourTestID := ""
	if latestTestRetrieved {
		// Get the testID used
		lineWanted := utils.ExtractLineWithPrefix(cmdOutput, "Downloading logs for test")
		splitedLine := strings.Split(lineWanted, "\"")
		ourTestID = strings.TrimSpace(splitedLine[1])
	} else {
		test.Fatal()
	}

	// Check that the output contains the needed element
	assert.Contains(test, cmdOutput, "The test logs for test \""+ourTestID+"\" are available in:")

	// Check that both files (.tap and .zip) are available
	assert.FileExists(test, path.Join("/tmp", "rp-tests-logs", ourTestID, "tests_logs.zip"))
	assert.FileExists(test, path.Join("/tmp", "rp-tests-logs", ourTestID, "all_tests.tap"))
	assert.FileExists(test, path.Join("/tmp", "rp-tests-logs", ourTestID, "boot.log"))
	assert.FileExists(test, path.Join("/tmp", "rp-tests-logs", ourTestID, "console.log"))

	// Now try to download test-logs from the server but ask for zip logs
	_, err = executeCommand(appsCommand, "tests", "logs", "mort-fine", "-p", "vendredi-sur-mer", "-d", "/tmp", "--zip")
	if err != nil {
		test.Fatal(err)
	}

	// Add to list the zip file to be deleted at teardown time
	zipName := path.Join("/tmp", "rp-tests-logs-"+ourTestID+".zip")
	filesListToDel = append(filesListToDel, zipName)

	// Check that a zip file has been correctly created
	assert.FileExists(test, zipName)
}

// TestAppsTestsGet is the function allowing to test if
// the command "rp-cli applications tests get <buildID> ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsTestsGet(test *testing.T) {

	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("SNCF", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "sncf", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Then, creates an application if it does not exists
	cmdToRun2 := getMainAddAppCmd("TGV", "train-grande-vitesse", "sncf")
	err = addObjectIfNotExists(test, appsCommand, "tgv", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Firstly, check if the "avion-a320" application exists and if it has builds
	buildID, err := checkIfBuildsExist("tgv", appsCommand)
	if err != nil {
		test.Fatal(err)
	}

	// Check if at least one build exists
	if buildID == "" {
		// No build exists, creates one
		_, err = executeCommand(appsCommand, "build", "tgv", "-p", "sncf")
		if err != nil {
			test.Fatal(err)
		}
	}

	// TODO: make sure that the comment is handled correctly !

	// Make sure to have at least three tests for this application
	_, err = executeCommand(appsCommand, "test", "tgv", "-c", "Test comment n°1")
	if err != nil {
		test.Fatal(err)
	}
	_, err = executeCommand(appsCommand, "test", "tgv", "-c", "Test comment n°2")
	if err != nil {
		test.Fatal(err)
	}
	_, err = executeCommand(appsCommand, "test", "tgv", "-c", "Test comment n°3")
	if err != nil {
		test.Fatal(err)
	}

	// Extract the latest test ID from the build list
	highestID, err := getHighestIDFromOutput("ID", appsCommand, "tests", "list", "tgv")
	if err != nil {
		test.Fatal(err)
	}

	// Run the "builds get" command and check the output
	checkNotVerboseListTestsOutput(test, appsCommand, "tests", "get", highestID)

	// Create the regexp map used to check (verbose mode)
	regexpMap2 := map[string]string{
		"Test ID":                `^[0-9]+$`,
		"Start time":             `.+`,
		"Application ID":         `^tgv$`,
		"Build ID":               `^[0-9]{4}$`,
		"Image ID":               `^(\[Not set\]|[a-z0-9-]+)$`,
		"Board model":            `^qemu-system-x86_64$`,
		"Status":                 `^(success|failed)$`,
		"Succeeded tests number": `[0-9]+`,
		"Failed tests number":    `[0-9]+`,
		"Skipped tests number":   `[0-9]+`,
		"Total tests number":     `[0-9]+`,
		"Comment:":               `^Test comment n°[0-9]{1}`,
	}

	// Run the "builds get" command an check the output
	checkCmdAnswerKeysRegexp(test, appTestsDisplayVerboseLineNb, regexpMap2, appsCommand, "tests", "get", highestID, "-v")
}

// TestAppsTestsWithBoard is the function allowing to test if
// the command "rp-cli applications tests ... --board <another_board>" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsTestsWithBoard(test *testing.T) {

	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Fatals Picards", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "fatals-picards", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Then, creates an application if it does not exists
	cmdToRun2 := getMainAddAppCmd("Le Retour a la terre", "le-retour-a-la-terre", "fatals-picards")
	err = addObjectIfNotExists(test, appsCommand, "le-retour-a-la-terre", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Firstly, check if the "le-retour-a-la-terre" application exists and if it has builds
	buildID, err := checkIfBuildsExist("le-retour-a-la-terre", appsCommand)
	if err != nil {
		test.Fatal(err)
	}

	// Check if at least one build exists
	if buildID == "" {
		// No build exists, creates one
		_, err = executeCommand(appsCommand, "build", "le-retour-a-la-terre", "-p", "fatals-picards")
		if err != nil {
			test.Fatal(err)
		}
	}

	// Run the test on the application
	_, err = executeCommand(appsCommand, "test", "le-retour-a-la-terre", "-p", "fatals-picards", "-b", "r8a7795-h3ulcb-kf")
	if err != nil {
		test.Fatal(err)
	}

	// Extract the latest test ID from the build list
	highestID, err := getHighestIDFromOutput("Test ID", appsCommand, "tests", "list", "le-retour-a-la-terre")
	if err != nil {
		test.Fatal(err)
	}

	// Create the regexp map used to check (verbose mode)
	regexpMap2 := map[string]string{
		"Test ID":                `^[0-9]+$`,
		"Start time":             `.+`,
		"Application ID":         `^le-retour-a-la-terre$`,
		"Build ID":               `^[0-9]{4}$`,
		"Image ID":               `^(\[Not set\]|[a-z0-9-]+)$`,
		"Board model":            `^r8a7795-h3ulcb-kf$`,
		"Status":                 `^(success|failed)$`,
		"Succeeded tests number": `[0-9]+`,
		"Failed tests number":    `[0-9]+`,
		"Skipped tests number":   `[0-9]+`,
		"Total tests number":     `[0-9]+`,
		"Comment:":               `.+`,
	}

	// Run the "builds get" command an check the output
	checkCmdAnswerKeysRegexp(test, appTestsDisplayVerboseLineNb, regexpMap2, appsCommand, "tests", "get", highestID, "-v")
}

// TestAppsTestsCancelOK is the function allowing to test if
// the command "rp-cli applications test-cancel <appID> ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsTestsCancelOK(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Vendredi sur Mer", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "vendredi-sur-mer", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Then, creates an application if it does not exists
	cmdToRun2 := getMainAddAppCmd("Mort-Fine", "mort-fine", "vendredi-sur-mer")
	err = addObjectIfNotExists(test, appsCommand, "mort-fine", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Firstly, check if the "mort-fine" application exists and if it has builds
	buildID, err := checkIfBuildsExist("mort-fine", appsCommand)
	if err != nil {
		test.Fatal(err)
	}

	// Check if at least one build exists
	if buildID == "" {
		// No build exists, creates one
		_, err = executeCommand(appsCommand, "build", "mort-fine", "-p", "vendredi-sur-mer")
		if err != nil {
			test.Fatal(err)
		}
	}

	// Start a test in non-blocking mode
	_, err = executeCommand(appsCommand, "test", "mort-fine", "--nonblocking")
	if err != nil {
		test.Fatal(err)
	}

	// Now try to cancel the test
	cmdOutput, err := executeCommand(appsCommand, "tests", "cancel", "mort-fine", "-p", "vendredi-sur-mer")
	if err != nil {
		test.Fatal(err)
	}

	// Test that everything went OK
	checkLineEndWith(test, cmdOutput, "Cancelling test with ID", "[OK]")

	// Extract the latest test ID from the build list
	highestID, err := getHighestIDFromOutput("Test ID", appsCommand, "tests", "list", "mort-fine", "-p", "vendredi-sur-mer")
	if err != nil {
		test.Fatal(err)
	}

	// Create the regexp map used to check (not verbose mode)
	regexpMap := map[string]string{
		"ID":     `^[0-9]+$`,
		"Status": `^aborted$`,
	}

	// Run the "tests get" command an check the output
	time.Sleep(50 * time.Millisecond)
	checkCmdAnsKeysRegexpCol(test, 2, regexpMap, appsCommand, "tests", "get", highestID)
}

// TestAppsTestsCancelKO is the function allowing to test if
// the command "rp-cli applications test-cancel <appID> ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsTestsCancelKO(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Vendredi sur Mer", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "vendredi-sur-mer", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Then, creates an application if it does not exists
	cmdToRun2 := getMainAddAppCmd("Mort-Fine", "mort-fine", "vendredi-sur-mer")
	err = addObjectIfNotExists(test, appsCommand, "mort-fine", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Firstly, check if the "mort-fine" application exists and if it has builds
	buildID, err := checkIfBuildsExist("mort-fine", appsCommand)
	if err != nil {
		test.Fatal(err)
	}

	// Check if at least one build exists
	if buildID == "" {
		// No build exists, creates one
		_, err = executeCommand(appsCommand, "build", "mort-fine", "-p", "vendredi-sur-mer")
		if err != nil {
			test.Fatal(err)
		}
	}

	// Start a test in blocking mode to make sure to have finished it
	_, err = executeCommand(appsCommand, "test", "mort-fine")
	if err != nil {
		test.Fatal(err)
	}

	// Now try to cancel the test (that is already finished)
	_, err = executeCommand(appsCommand, "tests", "cancel", "mort-fine", "-p", "vendredi-sur-mer")
	if assert.Error(test, err) {
		assert.Contains(test, err.Error(), "cannot cancel tests when tests' status is success")
	}
}

// TestAppsTestsTimeOut is the function allowing to test if
// rp-cli is behaving well when a test is in time-out
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsTestsTimeOut(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Rone", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "rone", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Then, creates an application if it does not exists
	cmdToRun2 := getMainAddAppCmd("Bora vocal", "bora-vocal", "rone")
	err = addObjectIfNotExists(test, appsCommand, "bora-vocal", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Firstly, check if the "bora-vocal" application exists and if it has builds
	buildID, err := checkIfBuildsExist("bora-vocal", appsCommand)
	if err != nil {
		test.Fatal(err)
	}

	// Check if at least one build exists
	if buildID == "" {
		// No build exists, creates one
		_, err = executeCommand(appsCommand, "build", "bora-vocal", "-p", "rone")
		if err != nil {
			test.Fatal(err)
		}
	}

	// Request the next test to be "time-out"
	rpbckmock.RequestTimedOutTest()

	// Start a test in non-blocking mode
	_, err = executeCommand(appsCommand, "test", "bora-vocal", "-p", "rone", "-v")

	// Test that we have an error in return
	assert.EqualError(test, err, appTestsTimedOutMsg)

	// Extract the latest test ID from the build list
	highestID, err := getHighestIDFromOutput("Test ID", appsCommand, "tests", "list", "bora-vocal", "-p", "rone")
	if err != nil {
		test.Fatal(err)
	}

	// Create the regexp map used to check (not verbose mode)
	regexpMap := map[string]string{
		"ID":     `^[0-9]+$`,
		"Status": `^timed-out$`,
	}

	// Run the "tests get" command an check the output
	time.Sleep(50 * time.Millisecond)
	checkCmdAnsKeysRegexpCol(test, 2, regexpMap, appsCommand, "tests", "get", highestID)
}

// TestAppsFilesUpload is the function allowing to test if
// the command "rp-cli applications files upload <appID> ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsFilesUpload(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Airbus", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "airbus", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create an application if it does not exists,
	cmdToRun2 := getMainAddAppCmd("Avion A320", "a320", "airbus")
	err = addObjectIfNotExists(test, appsCommand, "avion-a320", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Create a false spec file
	falseSpecFile, err := createFakeSpecfile("avion-a320.spec")
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false file
	defer deleteFalseFile(falseSpecFile)

	// Try to upload the file and check that everything is at [OK]
	checkActionCmdAns(test, appsCommand, appsFilesCmd, "upload", "avion-a320", falseSpecFile)

	// Try the same thing but specify the project
	checkActionCmdAns(test, appsCommand, appsFilesCmd, "upload", "avion-a320", falseSpecFile, "-p", "airbus")
}

// TestAppsFilesList is the function allowing to test if
// the command "rp-cli applications files list <appID> ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsFilesList(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Airbus", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "airbus", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create an application if it does not exists,
	cmdToRun2 := getMainAddAppCmd("Avion A350", "a350", "airbus")
	err = addObjectIfNotExists(test, appsCommand, "avion-a350", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"autogen::service.json":   `.*`,
		"avion-a350-1.0.0.tar.gz": `.*`,
		"avion-a350.spec":         `.*`,
	}

	// Run the command and the tests (no verbose)
	checkCmdAnswerKeysRegexp(test, 3, regexpMap, appsCommand, appsFilesCmd, "list", "avion-a350")

	// Test in verbose mode
	// Create the map containing the regexp we want to match in the answer
	regexpMap2 := map[string]string{
		"autogen::service.json":   `^\(657 bytes\)$`,
		"avion-a350-1.0.0.tar.gz": `^\(837 bytes\)$`,
		"avion-a350.spec":         `^\(507 bytes\)$`,
	}

	// Execute the verbose command
	outputCmd, err := executeCommand(appsCommand, appsFilesCmd, "content", "avion-a350", "-v")
	if err != nil {
		test.Fatal(err)
	}
	outputTrimmed := strings.TrimSpace(outputCmd)
	outputLines := strings.Split(outputTrimmed, "\n")

	// Test that we have the right number of line
	assert.Equal(test, len(regexpMap2), len(outputLines))

	// Test that each line matches its regexp
	for _, line := range outputLines {
		columns := getEachColumnFromLine(line)
		if assert.Equal(test, len(columns), 2) {
			assert.Regexp(test, regexp.MustCompile(regexpMap2[columns[0]]), columns[1])
		}
	}
}

// TestAppsFilesRemove is the function allowing to test if
// the command "rp-cli applications files remove <appID> files..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsFilesRemove(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Airbus", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "airbus", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create an application if it does not exists,
	cmdToRun2 := getMainAddAppCmd("Avion A340", "a340", "airbus")
	err = addObjectIfNotExists(test, appsCommand, "avion-a340", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Try to remove one file and check that everything is [OK]
	checkActionCmdAns(test, appsCommand, appsFilesCmd, "remove", "avion-a340", "avion-a340.spec")
}

// TestAppsFilesDownload is the function allowing to test if
// the command "rp-cli applications files download <appID> files..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsFilesDownload(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Airbus", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "airbus", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create an application if it does not exists,
	cmdToRun2 := getMainAddAppCmd("Avion A350", "a350", "airbus")
	err = addObjectIfNotExists(test, appsCommand, "avion-a350", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Get curent working directory
	oldWorkDir, err := os.Getwd()
	if err != nil {
		test.Fatal(err)
	}

	// Go to temporary directory
	err = os.Chdir("/tmp")
	if err != nil {
		test.Fatal(err)
	}

	defer os.Chdir(oldWorkDir)

	// Download two files from applications and check that everything went well
	checkActionCmdAns(test, appsCommand, appsFilesCmd, "download", "avion-a350", "avion-a350.spec", ".redpesk/autogen::service.json")

	// Check that the files exists in the /tmp directory
	specExists := assert.FileExists(test, "/tmp/avion-a350.spec")
	serviceFileExists := assert.FileExists(test, "/tmp/.redpesk/autogen::service.json")

	// Remove the files if they exists
	if specExists {
		err = os.Remove("/tmp/avion-a350.spec")
		if err != nil {
			test.Fatal(err)
		}
	}

	if serviceFileExists {
		err = os.Remove("/tmp/.redpesk/autogen::service.json")
		if err != nil {
			test.Fatal(err)
		}
		err = os.Remove("/tmp/.redpesk")
		if err != nil {
			test.Fatal(err)
		}
	}
}

// TestAppsFilesDownloadFullList is the function allowing to test if
// the command "rp-cli applications files download <appID>" is working correctly
// when no files name is given, i.e. when rp-cli tries to download the full list
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsFilesDownloadFullList(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Airbus", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "airbus", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create an application if it does not exists,
	cmdToRun2 := getMainAddAppCmd("Avion A350", "a350", "airbus")
	err = addObjectIfNotExists(test, appsCommand, "avion-a350", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Create a temporary directory
	utils.CreateDirIfNotExists("/tmp/TestAppsFilesDownloadFullList", true)

	// Get curent working directory
	oldWorkDir, err := os.Getwd()
	if err != nil {
		test.Fatal(err)
	}

	// Go to temporary directory
	err = os.Chdir("/tmp/TestAppsFilesDownloadFullList")
	if err != nil {
		test.Fatal(err)
	}

	defer os.Chdir(oldWorkDir)

	// Download two files from applications and check that everything went well
	checkActionCmdAns(test, appsCommand, appsFilesCmd, "download", "avion-a350")

	// Check that the files exists in the /tmp directory
	assert.FileExists(test, "/tmp/TestAppsFilesDownloadFullList/avion-a350.spec")
	assert.FileExists(test, "/tmp/TestAppsFilesDownloadFullList/avion-a350-1.0.0.tar.gz")

	// Remove the files if they exists
	defer os.RemoveAll("/tmp/TestAppsFilesDownloadFullList")
}

// TestAppsServices is the function allowing to test if
// the command "rp-cli applications services" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsServices(test *testing.T) {

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"Name":       `^[a-z_]+$`,
		"Summary":    `^[a-zA-Z- ()]+$`,
		"Version":    `^[0-9.]+$`,
		"Parameters": `^[a-z, ]+$`,
	}

	// Run the command and the tests (no verbose and verbose, since this is the same result)
	checkCmdAnsKeysRegexpCol(test, 4, regexpMap, appsCommand, appsServicesCmd)
}

// TestAppsServicesVerbose is the function allowing to test if
// the command "rp-cli applications services -v" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T

func TestAppsServicesVerbose(test *testing.T) {

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		appsServicesAnsLine0: `.*`,
		"Name":               `^[a-z_]+$`,
		"Version":            `^[0-9.]+$`,
		"Description":        `^[a-zA-Z- ().,]+$`,
		"Parameters":         `.*`,
	}

	// First, let's run the wanted command
	cmdOutput, err := executeCommand(appsCommand, appsServicesCmd, "-v")
	if err != nil {
		test.Log("A problem occurred while executing the following command: ", appsCommand, appsServicesCmd, "-v")
		test.Fatal(err)
	}
	cmdOutputTrimmed := strings.TrimSpace(cmdOutput)

	// Check the object obtained
	checkObjectKeysRegexp(test, cmdOutputTrimmed, 27, regexpMap, true)

	// Check the type of service parameters in answer
	checkServicesParamsTypesValues(test, cmdOutputTrimmed, "download", `^\[boolean\]$`, `^=> Download sources$`)
	checkServicesParamsTypesValues(test, cmdOutputTrimmed, "force", `^\[boolean\]$`, `^=> Overwrite sources$`)
	checkServicesParamsTypesValues(test, cmdOutputTrimmed, "arches", `^\[string_list\]$`, `^=> Array of architecture$`)
	checkServicesParamsTypesValues(test, cmdOutputTrimmed, "ref", `^\[string\]$`, `^=> Git reference$`)
	checkServicesParamsTypesValues(test, cmdOutputTrimmed, "url", `^\[string\]$`, `^=> Gitpkg URL$`)
	checkServicesParamsTypesValues(test, cmdOutputTrimmed, "auto", `^\[boolean\]$`, `^=> Set automatically version & release$`)

}

// ************ APP SERVICES *************
// TestAppsAddWithServicesVerbose is the function allowing to test if
// the command "rp-cli applications get ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsAddWithServicesVerbose(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("my-awesome-project", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "my-awesome-project", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the commands
	cmdToRun2 := getMainAddAppWithServiceCmd("perfect-app", "perfect-app", "my-awesome-project")

	// Create the application if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, appsCommand, "perfect-app", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":                        `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":                       `^perfect-app$`,
		"Package Name":               `^perfect-app$`,
		"Slug":                       `^perfect-app$`,
		"Type":                       `^standard$`,
		"Parent project":             `^my-awesome-project$`,
		"Description":                `.+`,
		"Source URL":                 `http(s)*://`,
		"Source Revision":            `^[a-z0-9-]+$`,
		"External Specfile":          `(\[Not set\]|(http(s)*://))`,
		"Distributions":              `^redpesk-devel-arz.*`,
		"Auto security audits tests": `^(ON|OFF) \[Opt\]$`,
		"AutoRSrpmBuild":             `dist_branch":true,"meta_changes":true,"no_srpm_bld":true,"pkg_changes":true`,
		"Build type":                 `^(standard_build_cross|emulated_build|accel_build_cross)$`,
		"Creation time":              `.+`,
		"Update time":                `.+`,
		"Bookmarked":                 `^false$`,
		"Services":                   `.*`,
		"- Name":                     `^[a-z_]+$`,
		"Parameters":                 `.*`,
	}

	regexpMapServices := map[string]string{}

	regexpMapServices["Services"] = regexpMap["Services"]
	regexpMapServices["- Name"] = regexpMap["- Name"]
	regexpMapServices["Parameters"] = regexpMap["Parameters"]

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, appDisplayVerboseFullLineWithServicesNb, regexpMap, appsCommand, "get", "perfect-app", "-v")

	// First, let's run the wanted command
	cmdOutput, err := executeCommand(appsCommand, "get", "perfect-app", "-v")
	if err != nil {
		test.Log("A problem occurred while executing the following command: ", appsCommand, appsGetCmd, "-v")
		test.Fatal(err)
	}

	cmdOutputTrimmed := strings.TrimSpace(cmdOutput)

	// Check the object obtained
	checkObjectKeysRegexp(test, cmdOutputTrimmed, appDisplayVerboseFullLineWithServicesNb, regexpMapServices, true)

	// Check the type of service parameters in answer
	checkServicesParamsTypesValues(test, cmdOutputTrimmed, "download", "", `^true$`)
	checkServicesParamsTypesValues(test, cmdOutputTrimmed, "force", "", `^true$`)
}

// TestAppsUpdateServices is the function allowing to test if
// the command "rp-cli application update --service ..." is working correctly
// when the update is done on services
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsUpdateServices(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Awesome project updated", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "awesome-project-updated", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the commands
	cmdToRun2 := getMainAddAppWithServiceCmd("Cool app updated", "cool-app-updated", "awesome-project-updated")

	// Create the application if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, appsCommand, "cool-app-updated", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the update command
	updateCmd := getUpdateAppCmd("cool-app-updated", "--service", "download_sources={'download':false,'force':false}")

	// Run the update command and check that all is [OK]
	checkActionCmdAns(test, updateCmd...)

	// Check that the values have been updated

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":                 `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":                `^Cool app updated$`,
		"Package Name":        `^cool-app-updated$`,
		"Slug":                `^cool-app-updated$`,
		"Type":                `^standard$`,
		"Parent project":      `^awesome-project-updated$`,
		"Description":         `^This is a testing application`,
		"Source URL":          `http(s)*://`,
		"Source Revision":     `^master$`,
		"External Specfile":   `(\[Not set\]|(http(s)*://))`,
		"Distributions":       `^redpesk-devel-arz.*`,
		"Auto on board tests": `^(ON|OFF) \[Opt\]$`,
		"Build type":          `^(standard_build_cross|emulated_build|accel_build_cross)$`,
		"Creation time":       `.+`,
		"Update time":         `.+`,
		"Bookmarked":          `^false$`,
		"Services":            `.*`,
		"- Name":              `^[a-z_]+$`,
		"Parameters":          `.*`,
	}

	regexpMapServices := map[string]string{}

	regexpMapServices["Services"] = regexpMap["Services"]
	regexpMapServices["- Name"] = regexpMap["- Name"]
	regexpMapServices["Parameters"] = regexpMap["Parameters"]

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, appDisplayVerboseFullLineWithServicesNb, regexpMap, appsCommand, "get", "cool-app-updated", "-v")

	// First, let's run the wanted command
	cmdOutput, err := executeCommand(appsCommand, "get", "cool-app-updated", "-v")
	if err != nil {
		test.Log("A problem occurred while executing the following command: ", appsCommand, appsGetCmd, "-v")
		test.Fatal(err)
	}

	cmdOutputTrimmed := strings.TrimSpace(cmdOutput)

	// Check the object obtained
	checkObjectKeysRegexp(test, cmdOutputTrimmed, appDisplayVerboseFullLineWithServicesNb, regexpMapServices, true)

	// Check the type of service parameters in answer
	checkServicesParamsTypesValues(test, cmdOutputTrimmed, "download", "", `^false$`)
	checkServicesParamsTypesValues(test, cmdOutputTrimmed, "force", "", `^false$`)
}

// TestAppsDisableServices is the function allowing to test if
// the command "rp-cli application update --disable-services ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsDisableServices(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Awesome project disable", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "awesome-project-disable", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the commands
	cmdToRun2 := getMainAddAppWithServiceCmd("Cool app disable", "cool-app-disable", "awesome-project-disable")

	// Create the application if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, appsCommand, "cool-app-disable", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the update command
	updateCmd := getUpdateAppCmd("cool-app-disable", "--disable-services", "download_sources")

	// Run the update command and check that all is [OK]
	checkActionCmdAns(test, updateCmd...)

	// Check that the values have been updated
	regexpMap := map[string]string{
		"ID:":                 `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":                `^Cool app disable$`,
		"Package Name":        `^cool-app-disable$`,
		"Slug":                `^cool-app-disable$`,
		"Type":                `^standard$`,
		"Parent project":      `^awesome-project-disable$`,
		"Description":         `^This is a testing application`,
		"Source URL":          `http(s)*://`,
		"Source Revision":     `^master$`,
		"External Specfile":   `(\[Not set\]|(http(s)*://))`,
		"Specfile type":       `(internal|relative_path|external_url)`,
		"Distributions":       `^redpesk-devel-arz.*`,
		"Auto on board tests": `^(ON|OFF) \[Opt\]$`,
		"Build type":          `^(standard_build_cross|emulated_build|accel_build_cross)$`,
		"Creation time":       `.+`,
		"Update time":         `.+`,
		"Bookmarked":          `^false$`,
		"Services":            `.*`,
	}

	// Run the command and the tests
	output := checkCmdAnswerKeysRegexp(test, appDisplayVerboseFullLineNb, regexpMap, appsCommand, "get", "cool-app-disable", "-v")

	// Check that the service name does not appear anymore in the result
	assert.NotContains(test, output, "download_sources")
}

// TestAppsUpdateEmulatedField is the function allowing to test if
// the command "rp-cli application update ... --emulated" is working correctly
// when the update is done on the description
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsUpdateEmulatedField(test *testing.T) {

	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("black dog prj", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "black-dog-prj", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the commands
	cmdToRun2 := getMainAddAppWithEmulatedCmd("Test app", "test-app", "black-dog-prj")

	// Create the application if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, appsCommand, "black-dog-prj", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the update command
	updateCmd := getUpdateAppCmd("test-app", "--build-type", "standard_build_cross")

	// Run the update command and check that all is [OK]
	checkActionCmdAns(test, updateCmd...)

	// Check that the values have been updated

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":                 `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":                `^Test app$`,
		"Package Name":        `^test-app$`,
		"Slug":                `^test-app$`,
		"Type":                `^standard$`,
		"Parent project":      `^black-dog-prj$`,
		"Description":         `^This is a testing application$`,
		"Source URL":          `http(s)*://`,
		"Source Revision":     `^master$`,
		"External Specfile":   `(\[Not set\]|(http(s)*://))`,
		"Specfile type":       `(internal|relative_path|external_url)`,
		"Distributions":       `^redpesk-devel-arz.*`,
		"Auto on board tests": `^(ON|OFF) \[Opt\]$`,
		"Build type":          `^standard_build_cross$`,
		"Creation time":       `.+`,
		"Update time":         `.+`,
		"Bookmarked":          `^false$`,
		"Services":            `.*`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, appDisplayVerboseFullLineNb, regexpMap, appsCommand, "get", "test-app", "-v")
}

// TestAppsUpdateGlobalSRPMRebuild is the function allowing to test if
// the command "rp-cli application update --auto-srpm ..." is working correctly
// when the update is done on global auto srpm rebuild
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsUpdateGlobalSRPMRebuild(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Awesome project updated", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "awesome-project-updated", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the commands
	cmdToRun2 := getMainAddAppWithServiceCmd("Srpm rebuild app updated", "srpm-rebuild-app-updated", "awesome-project-updated")

	// Create the application if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, appsCommand, "srpm-rebuild-app-updated", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the update command
	updateCmd := getUpdateAppCmd("srpm-rebuild-app-updated", "--auto-srpm", "false")

	// Run the update command and check that all is [OK]
	checkActionCmdAns(test, updateCmd...)

	// Check that the values have been updated

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":                 `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":                `^Srpm rebuild app updated$`,
		"Package Name":        `^srpm-rebuild-app-updated$`,
		"Slug":                `^srpm-rebuild-app-updated$`,
		"Type":                `^standard$`,
		"Parent project":      `^awesome-project-updated$`,
		"Description":         `^This is a testing application`,
		"Source URL":          `http(s)*://`,
		"Source Revision":     `^master$`,
		"External Specfile":   `(\[Not set\]|(http(s)*://))`,
		"Distributions":       `^redpesk-devel-arz.*`,
		"Auto on board tests": `^(ON|OFF) \[Opt\]$`,
		"AutoRSrpmBuild":      `dist_branch":false,"meta_changes":false,"no_srpm_bld":false,"pkg_changes":false`,
		"Build type":          `^(standard_build_cross|emulated_build|accel_build_cross)$`,
		"Creation time":       `.+`,
		"Update time":         `.+`,
		"Bookmarked":          `^false$`,
		"Services":            `.*`,
		"- Name":              `^[a-z_]+$`,
		"Parameters":          `.*`,
	}

	regexpMapServices := map[string]string{}

	regexpMapServices["Services"] = regexpMap["Services"]
	regexpMapServices["- Name"] = regexpMap["- Name"]
	regexpMapServices["Parameters"] = regexpMap["Parameters"]

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, appDisplayVerboseFullLineWithServicesNb, regexpMap, appsCommand, "get", "srpm-rebuild-app-updated", "-v")

	// First, let's run the wanted command
	cmdOutput, err := executeCommand(appsCommand, "get", "srpm-rebuild-app-updated", "-v")
	if err != nil {
		test.Log("A problem occurred while executing the following command: ", appsCommand, appsGetCmd, "-v")
		test.Fatal(err)
	}

	cmdOutputTrimmed := strings.TrimSpace(cmdOutput)

	// Check the object obtained
	checkObjectKeysRegexp(test, cmdOutputTrimmed, appDisplayVerboseFullLineWithServicesNb, regexpMap, true)
}

// TestAppsUpdateSRPMAutoRebuildSingle is the function allowing to test if
// the command "rp-cli application update --auto-srpm-build ..." is working correctly
// when the update is done on one setting
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsUpdateSRPMAutoRebuildSingle(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Awesome project updated", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "awesome-project-updated", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the commands
	cmdToRun2 := getMainAddAppWithServiceCmd("Srpm rebuild app updated single", "srpm-rebuild-app-updated-single", "awesome-project-updated")

	// Create the application if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, appsCommand, "srpm-rebuild-app-updated-single", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the update command
	updateCmd := getUpdateAppCmd("srpm-rebuild-app-updated-single", "--auto-srpm-build", "pkg_changes=false")

	// Run the update command and check that all is [OK]
	checkActionCmdAns(test, updateCmd...)

	// Check that the values have been updated

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":                 `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":                `^Srpm rebuild app updated single$`,
		"Package Name":        `^srpm-rebuild-app-updated-single$`,
		"Slug":                `^srpm-rebuild-app-updated-single$`,
		"Type":                `^standard$`,
		"Parent project":      `^awesome-project-updated$`,
		"Description":         `^This is a testing application`,
		"Source URL":          `http(s)*://`,
		"Source Revision":     `^master$`,
		"External Specfile":   `(\[Not set\]|(http(s)*://))`,
		"Distributions":       `^redpesk-devel-arz.*`,
		"Auto on board tests": `^(ON|OFF) \[Opt\]$`,
		"AutoRSrpmBuild":      `dist_branch":true,"meta_changes":true,"no_srpm_bld":true,"pkg_changes":false`,
		"Build type":          `^(standard_build_cross|emulated_build|accel_build_cross)$`,
		"Creation time":       `.+`,
		"Update time":         `.+`,
		"Bookmarked":          `^false$`,
		"Services":            `.*`,
		"- Name":              `^[a-z_]+$`,
		"Parameters":          `.*`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, appDisplayVerboseFullLineWithServicesNb, regexpMap, appsCommand, "get", "srpm-rebuild-app-updated-single", "-v")

	// First, let's run the wanted command
	cmdOutput, err := executeCommand(appsCommand, "get", "srpm-rebuild-app-updated-single", "-v")
	if err != nil {
		test.Log("A problem occurred while executing the following command: ", appsCommand, appsGetCmd, "-v")
		test.Fatal(err)
	}

	cmdOutputTrimmed := strings.TrimSpace(cmdOutput)

	// Check the object obtained
	checkObjectKeysRegexp(test, cmdOutputTrimmed, appDisplayVerboseFullLineWithServicesNb, regexpMap, true)
}

// TestAppsUpdateSRPMAutoRebuildMultiple is the function allowing to test if
// the command "rp-cli application update --auto-srpm-build ..." is working correctly
// when the update is done on multiple settings
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsUpdateSRPMAutoRebuildMultiple(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Awesome project updated", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "awesome-project-updated", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the commands
	cmdToRun2 := getMainAddAppWithServiceCmd("Srpm rebuild app updated multi", "srpm-rebuild-app-updated-multi", "awesome-project-updated")

	// Create the application if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, appsCommand, "srpm-rebuild-app-updated-multi", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the update command for one parameter
	updateCmd := getUpdateAppCmd("srpm-rebuild-app-updated-multi", "--auto-srpm-build", "pkg_changes=false")

	// Run the update command and check that all is [OK]
	checkActionCmdAns(test, updateCmd...)

	// Create an update command for the second parameter
	updateCmd1 := getUpdateAppCmd("srpm-rebuild-app-updated-multi", "--auto-srpm-build", "dist_branch=false")

	// Run the update command and check that all is [OK]
	checkActionCmdAns(test, updateCmd1...)

	// Check that the values have been updated

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":                 `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":                `^Srpm rebuild app updated multi$`,
		"Package Name":        `^srpm-rebuild-app-updated-multi$`,
		"Slug":                `^srpm-rebuild-app-updated-multi$`,
		"Type":                `^standard$`,
		"Parent project":      `^awesome-project-updated$`,
		"Description":         `^This is a testing application`,
		"Source URL":          `http(s)*://`,
		"Source Revision":     `^master$`,
		"External Specfile":   `(\[Not set\]|(http(s)*://))`,
		"Distributions":       `^redpesk-devel-arz.*`,
		"Auto on board tests": `^(ON|OFF) \[Opt\]$`,
		"AutoRSrpmBuild":      `dist_branch":false,"meta_changes":true,"no_srpm_bld":true,"pkg_changes":false`,
		"Build type":          `^(standard_build_cross|emulated_build|accel_build_cross)$`,
		"Creation time":       `.+`,
		"Update time":         `.+`,
		"Bookmarked":          `^false$`,
		"Services":            `.*`,
		"- Name":              `^[a-z_]+$`,
		"Parameters":          `.*`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, appDisplayVerboseFullLineWithServicesNb, regexpMap, appsCommand, "get", "srpm-rebuild-app-updated-multi", "-v")

	// First, let's run the wanted command
	cmdOutput, err := executeCommand(appsCommand, "get", "srpm-rebuild-app-updated-multi", "-v")
	if err != nil {
		test.Log("A problem occurred while executing the following command: ", appsCommand, appsGetCmd, "-v")
		test.Fatal(err)
	}

	cmdOutputTrimmed := strings.TrimSpace(cmdOutput)

	// Check the object obtained
	checkObjectKeysRegexp(test, cmdOutputTrimmed, appDisplayVerboseFullLineWithServicesNb, regexpMap, true)
}

// TestAppsUpdateSRPMAutoRebuildIncompatibleFlags is the function allowing to test if
// the command "rp-cli application update --auto-srpm.... --auto-srpm-build ..." ignors --auto-srpm-build flag and takes into account only global flag
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsUpdateSRPMAutoRebuildIncompatibleFlags(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Awesome project updated", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "awesome-project-updated", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the commands
	cmdToRun2 := getMainAddAppWithServiceCmd("Srpm rebuild app updated", "srpm-rebuild-app-updated", "awesome-project-updated")

	// Create the application if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, appsCommand, "srpm-rebuild-app-updated", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the update command for one parameter
	updateCmd := getUpdateAppCmd("srpm-rebuild-app-updated", "--auto-srpm-build", "pkg_changes=true")

	// Run the update command and check that all is [OK]
	checkActionCmdAns(test, updateCmd...)

	// Create an update command for the second parameter
	updateCmd1 := getUpdateAppCmd("srpm-rebuild-app-updated", "--auto-srpm", "false")

	// Run the update command and check that all is [OK]
	checkActionCmdAns(test, updateCmd1...)

	// Check that the values have been updated

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":                 `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":                `^Srpm rebuild app updated$`,
		"Package Name":        `^srpm-rebuild-app-updated$`,
		"Slug":                `^srpm-rebuild-app-updated$`,
		"Type":                `^standard$`,
		"Parent project":      `^awesome-project-updated$`,
		"Description":         `^This is a testing application`,
		"Source URL":          `http(s)*://`,
		"Source Revision":     `^master$`,
		"External Specfile":   `(\[Not set\]|(http(s)*://))`,
		"Distributions":       `^redpesk-devel-arz.*`,
		"Auto on board tests": `^(ON|OFF) \[Opt\]$`,
		"AutoRSrpmBuild":      `dist_branch":false,"meta_changes":false,"no_srpm_bld":false,"pkg_changes":false`,
		"Build type":          `^(standard_build_cross|emulated_build|accel_build_cross)$`,
		"Creation time":       `.+`,
		"Update time":         `.+`,
		"Bookmarked":          `^false$`,
		"Services":            `.*`,
		"- Name":              `^[a-z_]+$`,
		"Parameters":          `.*`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, appDisplayVerboseFullLineWithServicesNb, regexpMap, appsCommand, "get", "srpm-rebuild-app-updated", "-v")

	// First, let's run the wanted command
	cmdOutput, err := executeCommand(appsCommand, "get", "srpm-rebuild-app-updated", "-v")
	if err != nil {
		test.Log("A problem occurred while executing the following command: ", appsCommand, appsGetCmd, "-v")
		test.Fatal(err)
	}

	cmdOutputTrimmed := strings.TrimSpace(cmdOutput)

	// Check the object obtained
	checkObjectKeysRegexp(test, cmdOutputTrimmed, appDisplayVerboseFullLineWithServicesNb, regexpMap, true)
}

// TestAppsBuildsStats is the function allowing to test if
// the command "rp-cli applications builds stats <appID>" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsBuildsStats(test *testing.T) {
	nbApps := 5
	nbBuilds := 3

	// First create all the projects, apps and builds
	err := createFullPrjAppsBuildsTests(test, "Lolo", "redpesk-lts-arz", "Honore", nbApps, nbBuilds, 0)
	if err != nil {
		test.Fatal(err)
	}

	// Check the content of the "not verbose" output
	regexpMap := map[string]string{
		"Application ID:":   `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Application slug:": `^honore-1$`,
		"Package:":          `^honore-1$`,
		"Done:":             `^[0-3]+$`,
		"Failed:":           `^[0-3]+$`,
		"Cancelled:":        `^[0-3]+$`,
		"Total:":            `^3$`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, appBuildsStatsNoVerboseLineNb, regexpMap, appsCommand, genericBuildsCmd, "stats", "honore-1", "-p", "lolo")

	// Now, let's go for the verbose mode
	verboseAns := checkCmdAnswerKeysRegexp(test, appBuildsStatsNoVerboseLineNb+3+nbBuilds+1, regexpMap, appsCommand, genericBuildsCmd, "stats", "honore-1", "-p", "lolo", "-v")

	// Create the regexp map we will use
	regexpMapCol := map[string]string{
		"ID":           `^[0-9]+$`,
		"Distribution": `^redpesk-lts-arz$`,
		"Status":       `^(done|failed)$`,
	}

	// Get the second part of the answer
	columnsAns := strings.Split(verboseAns, appsBuildsStatsCmdAnsLine7)[1]
	// Trim it
	columnsAns = strings.TrimSpace(columnsAns)

	// Check the answer of the command
	checkObjectKeysRegexpCol(test, columnsAns, appBuildsStatsVerboseColNb, regexpMapCol)
}

// TestAppsTestsStats is the function allowing to test if
// the command "rp-cli applications tests stats <appID>" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsTestsStats(test *testing.T) {
	nbApps := 5
	nbBuilds := 3
	nbTests := 5

	// First create all the projects, apps and builds
	err := createFullPrjAppsBuildsTests(test, "Disiz", "redpesk-lts-arz", "All in", nbApps, nbBuilds, nbTests)
	if err != nil {
		test.Fatal(err)
	}

	// Check the content of the "not verbose" output
	regexpMap := map[string]string{
		"Application ID:":   `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Application slug:": `^all-in-1$`,
		"Package:":          `^all-in-1$`,
		"Succeeded:":        `^[0-5]+$`,
		"Failed:":           `^[0-5]+$`,
		"Cancelled:":        `^[0-5]+$`,
		"Total:":            `^5$`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, appTestsStatsNoVerboseLineNb, regexpMap, appsCommand, genericTestsCmd, "stats", "all-in-1", "-p", "disiz")

	// Now, let's go for the verbose mode
	verboseAns := checkCmdAnswerKeysRegexp(test, appTestsStatsNoVerboseLineNb+4+nbTests, regexpMap, appsCommand, genericTestsCmd, "stats", "all-in-1", "-p", "disiz", "-v")

	// Create the regexp map we will use
	regexpMapCol := map[string]string{
		"ID":     `^[0-9]+$`,
		"Type":   `^(CI_test|audit)$`,
		"Status": `^(success|failed)$`,
	}

	// Get the second part of the answer
	columnsAns := strings.Split(verboseAns, genericTestsStatsCmdAnsLine7)[1]
	// Trim it
	columnsAns = strings.TrimSpace(columnsAns)

	// Check the answer of the command
	checkObjectKeysRegexpCol(test, columnsAns, appTestsStatsVerboseColNb, regexpMapCol)
}

// TestAppCopy is the function allowing to test if
// the command "rp-cli applications copy <app-ID> --destination-project <destination-project>" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppCopy(test *testing.T) {

	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Project Japon", "redpesk-devel-arz")
	// Create a first project
	err := addObjectIfNotExists(test, prjsCommand, "project-japon", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the command
	cmdToRun2 := getMainAddAppCmd("Origami", "origami", "project-japon")
	// then create the application if it does not exists
	err = addObjectIfNotExists(test, appsCommand, "origami", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the command
	cmdToRun3 := getMainAddPrjCmd("Project France", "redpesk-devel-arz")
	// Create a second project where the application will be copied
	err = addObjectIfNotExists(test, prjsCommand, "project-france", cmdToRun3...)
	if err != nil {
		test.Fatal(err)
	}

	// Request to copy the application to another project
	cmdToRun4 := []string{appsCommand, "copy", "origami", "-d", "project-france", "-p", "project-japon"}

	// Run the command to get the copye the application
	output, err := executeCommand(cmdToRun4...)
	if err != nil {
		test.Fatal(err)
	}

	// Check that the answer is the expected one
	assert.Contains(test, output, fmt.Sprintf(appsCopyCmdAnsLine0, "origami", "project-france"))
}

// TestAppRefresh is the function allowing to test if
// the command "rp-cli applications refresh-files-pkg <app-ID>" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppRefresh(test *testing.T) {

	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Project Opel", "redpesk-devel-arz")
	// Create a first project
	err := addObjectIfNotExists(test, prjsCommand, "project-opel", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the command
	cmdToRun2 := getMainAddAppCmd("app-tourer", "app-tourer", "project-opel")
	// then create the application if it does not exists
	err = addObjectIfNotExists(test, appsCommand, "app-tourer", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Request to refresh the application files
	cmdToRun3 := []string{appsCommand, "refresh-files-pkg", "app-tourer"}

	// Run the command to get the copye the application
	output, err := executeCommand(cmdToRun3...)
	if err != nil {
		test.Fatal(err)
	}

	// Check that the answer is the expected one
	assert.Contains(test, output, fmt.Sprintf(appsRefreshFilesPkgAnsLine1, "app-tourer"))
	assert.Contains(test, output, "[OK]")
}

// --------------------------------------------------------------
// Utilities functions ------------------------------------------
// --------------------------------------------------------------

// getMainAddAppCmd is an initializer function to get the mainly
// used command to create an application
//
// :param name: Name of the application to create
// :type name: string
// :param pkgName: Name of the package created by the application
// :type pkgName: string
// :param project: Project parent of the app to create
// :type project: string
// :returns: []string -- the value of the wanted constant
func getMainAddAppCmd(name string, pkgName string, project string) []string {
	return []string{
		appsCommand,
		appsAddCmd,
		"-n",
		name,
		"--pkg-name",
		pkgName,
		"-p",
		project,
		"--source-url",
		"http://git.ovh.iot/redpesk/redpesk-samples/simple-helloworld.git",
		"-d",
		"This is a testing application",
	}
}

// getAddFromSampleAppCmd is an initializer function to get the
// command to create an application from sample
//
// :param pkgName: Name of the package created by the application
// :type pkgName: string
// :param project: Project parent of the app to create
// :type project: string
// :returns: []string -- the value of the wanted constant
func getAddFromSampleAppCmd(pkgName string, project string) []string {
	return []string{
		appsCommand,
		"add-from-sample",
		pkgName,
		"-p",
		project,
	}
}

// getMainAddAppWithServiceCmd is an initializer function to get the mainly used command to create an application with a service
//
// :param name: Name of the application to create
// :type name: string
// :param pkgName: Name of the package created by the application
// :type pkgName: string
// :param project: Project parent of the app to create
// :type project: string
// :returns: []string -- the value of the wanted constant
func getMainAddAppWithServiceCmd(name string, pkgName string, project string) []string {
	return []string{
		appsCommand,
		appsAddCmd,
		"-n",
		name,
		"--pkg-name",
		pkgName,
		"-p",
		project,
		"--source-url",
		"http://git.ovh.iot/redpesk/redpesk-samples/test-application.git",
		"-d",
		"This is a testing application",
		"--service",
		"download_sources={'download':true,'force':true}",
	}
}

// getUpdateAppCmd is an initializer function to get the mainly
// used command to create a project
//
// :param slug: Slug of the application to update
// :type slug: string
// :param flag: Flag corresponding to the parameter we want to change in app
// :type flag: string
// :param value: New value to set for the parameter we want to change
// :type value: string
// :returns: []string -- the value of the wanted constant
func getUpdateAppCmd(slug string, flag string, value string) []string {
	return []string{
		appsCommand,
		"update",
		slug,
		flag,
		value,
	}
}

// getMainAddAppWithEmulatedCmd is an initializer function to get the mainly
// used command to create an application with --emulated
//
// :param name: Name of the application to create
// :type name: string
// :param pkgName: Name of the package created by the application
// :type pkgName: string
// :param project: Project parent of the app to create
// :type project: string
// :returns: []string -- the value of the wanted constant
func getMainAddAppWithEmulatedCmd(name string, pkgName string, project string) []string {
	return []string{
		appsCommand,
		appsAddCmd,
		"-n",
		name,
		"--pkg-name",
		pkgName,
		"-p",
		project,
		"--source-url",
		"http://git.ovh.iot/redpesk/redpesk-samples/test-application.git",
		"-d",
		"This is a testing application",
		"--build-type",
		"emulated_build",
		"true",
	}
}

// checkNotVerboseBuildOutput is function allowing to process the output of
// a "applications build" command, and checks each object of the list
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param outputCmd: Output to process
// :type outputCmd: string
func checkNotVerboseBuildOutput(test *testing.T, outputCmd string) {

	trimmedOutput := strings.TrimSpace(outputCmd)

	// Firstly, transform the raw output to a slice of lines
	linesSlice := strings.Split(trimmedOutput, "\n")

	// Get rid of the first two lines since it is just headers
	linesSliceToProc := linesSlice[2:]

	// Now process each line
	for _, line := range linesSliceToProc {
		switch {
		case strings.HasPrefix(line, genericBuildLiveLabelStatusChange):
			// Check that the line match the right regexp
			assert.Regexp(test, `^\[STATUS CHANGE\] Build: [0-9]+ \[[0-9a-z- ]+\]\[[a-z- ]+\] -- (open|done|failed) +(-- Subtasks:( *(livemedia x86_64|x86_64|aarch64) \[(open|closed|failed) +\],*)*)*`, line)

		case strings.HasPrefix(line, genericBuildLiveLabelSuccess):
			// Check that the line match the right regexp
			assert.Regexp(test, `^\[RESULT\]\[SUCCESS \] Build: [0-9]+ \[[0-9a-z- ]+\]\[[a-z- ]+\] -- done +(-- Subtasks:( *(livemedia x86_64|x86_64|aarch64) \[closed +\],*)*)*`, line)

		case strings.HasPrefix(line, genericBuildLiveLabelFailed):
			if strings.HasPrefix(line, genericBuildLiveLabelFailed+" Build:") {
				// Check that the line match the right regexp
				assert.Regexp(test, `^\[RESULT\]\[FAILED  \] Build: [0-9]+ \[[0-9a-z- ]+\]\[[a-z- ]+\] -- failed +(-- Subtasks:( *(livemedia x86_64|x86_64|aarch64) \[(closed|failed) +\],*)*)*`, line)
			} else {
				// Check that the line match the right regexp
				assert.Regexp(test, `^\[RESULT\]\[FAILED  \] (Error code: +[0-9]+|Error message: +'.+')`, line)
			}

		case strings.HasPrefix(line, genericBuildLiveLabelCanceled):
			assert.Regexp(test, `^\[RESULT\]\[CANCELED\] Build: [0-9]+ \[[0-9a-z- ]+\]\[[a-z- ]+\] -- canceled +(-- Subtasks:( *(livemedia x86_64|x86_64|aarch64) \[(open|closed|failed) +\],*)*)*`, line)

		default:
			err := errors.New("this prefix is not allowed in the build output...")
			test.Fatal(err)
		}
	}
}

// checkVerboseListBuildsOutput is function allowing to run the
// "builds-list -v" command, and checks each object of the list
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param buildType: Build type to check
// :type buildType: rpapiv1.RPBuildType
// :param args: Args to pass as command (sub-command, flags, etc.)
// :type args: ...string
func checkVerboseListBuildsOutput(test *testing.T, buildType rpapiv1.RPBuildType, args ...string) {

	// Create the regexp map used to check
	regexpMap := map[string]string{
		"Build ID:":     `[0-9]{4}`,
		"Build type:":   `(application)|(image) build`,
		"Distribution:": `^[a-z0-9-]+$`,
		"Built by":      `.*`,
		"Status:":       `^(open|done|failed)$`,
		"Comment:":      `.+`,
	}

	if buildType == rpapiv1.RPBuildApp {
		regexpMap["Application ID:"] = `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`
		regexpMap["Application slug:"] = `^[a-z0-9-]+$`
		regexpMap["SRPM build ID:"] = `^[a-z0-9-]+$`
	} else if buildType == rpapiv1.RPBuildSrpm {
		regexpMap["Application ID:"] = `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`
		regexpMap["Application slug:"] = `^[a-z0-9-]+$`
		regexpMap["Build type:"] = `^application SRPM build$`
	} else if buildType == rpapiv1.RPBuildImage {
		regexpMap["Image ID:"] = `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`
		regexpMap["Image slug:"] = `^[a-z0-9-]+$`
	}

	checkCmdAnswerListKeysRegexp(test, len(regexpMap), genericBuildCmdAnsLine4, regexpMap, args...)
}

// checkNotVerboseListBuildsOutput is function allowing to run the
// "applications builds-list" command, and checks each object of the list
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param args: Args to pass as command (sub-command, flags, etc.)
// :type args: ...string
func checkNotVerboseListBuildsOutput(test *testing.T, args ...string) {

	// Create the regexp map used to check
	regexpMap := map[string]string{
		"ID":     `[0-9]{4}`,
		"Status": `^(open|done|failed)$`,
	}

	checkCmdAnsKeysRegexpCol(test, 2, regexpMap, args...)
}

// getHighestIDFromOutput is function allowing to get the highest ID for a given key
// in the output coming from rp-cli (typically builds list and tests list)
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param args: Args to pass as command (sub-command, flags, etc.)
// :type args: ...string
func getHighestIDFromOutput(key string, command ...string) (string, error) {
	// Initialize the result
	currentHighestID := "0"

	// First, let's run the wanted command
	cmdOutput, err := executeCommand(command...)
	if err != nil {
		return "", err
	}
	cmdOutputTrimmed := strings.TrimSpace(cmdOutput)

	// Split the answer line by line
	outputLines := strings.Split(cmdOutputTrimmed, "\n")

	// Loop on the lines and get the highest ID
	for _, line := range outputLines[1:] {

		// Get the ID available on the line
		splittedLine := strings.Fields(line)
		currentID := strings.TrimSpace(splittedLine[0])

		// Convert the different IDs to int
		currentIDInt, err := strconv.Atoi(currentID)
		if err != nil {
			return "", err
		}
		currentHighestIDInt, err := strconv.Atoi(currentHighestID)
		if err != nil {
			return "", err
		}

		if currentIDInt > currentHighestIDInt {
			currentHighestID = currentID
		}
	}

	if currentHighestID == "0" {
		return "", errors.New("No ID found in the returned output")
	}

	return currentHighestID, nil
}

// getLowestIDFromOutput is function allowing to get the lowest ID for a given key
// in the output coming from rp-cli (typically builds list and tests list)
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param args: Args to pass as command (sub-command, flags, etc.)
// :type args: ...string
func getLowestIDFromOutput(key string, command ...string) (string, error) {
	// Initialize the result
	currentLowestID := "9999999999"

	// First, let's run the wanted command
	cmdOutput, err := executeCommand(command...)
	if err != nil {
		return "", err
	}
	cmdOutputTrimmed := strings.TrimSpace(cmdOutput)

	// Split the answer line by line
	outputLines := strings.Split(cmdOutputTrimmed, "\n")

	// Loop on the lines and get the highest ID
	for _, line := range outputLines[1:] {
		// Get the ID available on the line
		splittedLine := strings.Fields(line)
		currentID := strings.TrimSpace(splittedLine[0])

		// Convert the different IDs to int
		currentIDInt, err := strconv.Atoi(currentID)
		if err != nil {
			return "", err
		}
		currentHighestIDInt, err := strconv.Atoi(currentLowestID)
		if err != nil {
			return "", err
		}

		if currentIDInt < currentHighestIDInt {
			currentLowestID = currentID
		}
	}

	if currentLowestID == "9999999999" {
		return "", errors.New("No ID found in the returned output")
	}

	return currentLowestID, nil
}

var specFileRelease = 1

// createFakeSpecfile is function allowing to create a fake spec
// file from a given name
//
// :param name: Name of the spec file to create
// :type name: string
// :returns: string, error -- The full path to the spec file and an error if a problem has occurred
func createFakeSpecfile(name string) (string, error) {

	// Increment the release
	specFileRelease++

	// Create a false spec file
	// Get the current directory
	currDir, err := os.Getwd()
	if err != nil {
		return "", err
	}

	fullSpecPath := currDir + "/" + name
	updatedReleaseSpec := fmt.Sprintf(fakeSpecReleaseLine, specFileRelease)
	fullSpecFile := fakeSpecFileBegin + "\n" + updatedReleaseSpec + "\n" + fakeSpecFileEnd
	err = createFalseFile(fullSpecPath, fullSpecFile)
	if err != nil {
		return "", err
	}

	return fullSpecPath, nil
}

// createFakeSpecfile is function allowing to the full tree structure from project to test
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param prjName: Name of the project
// :type prjName: string
// :param distro: Name of the distro for the project
// :type distro: string
// :param appsPrefixName: Prefix to use for the application name
// :type appsPrefixName: string
// :param appNb: Number of apps to create
// :type appNb: int
// :param buildNb: Number of builds to run
// :type buildNb: int
// :param testNb: Number of tests to run
// :type testNb: int
// :returns: error -- an error if a problem has occurred
func createFullPrjAppsBuildsTests(test *testing.T, prjName, distro, appsPrefixName string, appNb, buildNb, testNb int) error {
	// From the name, create the slug
	prjSlug := strings.ReplaceAll(strings.ToLower(prjName), " ", "-")

	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd(prjName, distro)
	err := addObjectIfNotExists(test, prjsCommand, prjSlug, cmdToRun1...)
	if err != nil {
		return err
	}

	// Create the number of application(s) wanted
	for i := 0; i < appNb; i++ {
		// Create the name with i
		tmpAppName := appsPrefixName + " " + strconv.Itoa(i)
		// Create the corresponding slug
		tmpAppSlug := strings.ReplaceAll(strings.ToLower(tmpAppName), " ", "-")

		// Create the command
		cmdToRun2 := getMainAddAppCmd(tmpAppName, tmpAppSlug, prjSlug)

		// Create the application if it does not exists, and check that all steps are OK
		err = addObjectIfNotExists(test, appsCommand, tmpAppSlug, cmdToRun2...)
		if err != nil {
			return err
		}

		// Now create the number of builds wanted
		// Get a random number for the build index that will fail
		randNb := 0
		if buildNb > 0 {
			randNb = rand.Intn(buildNb)
		}
		for j := 0; j < buildNb; j++ {
			cmdToPlay := []string{appsCommand, "build", tmpAppSlug, "-p", prjSlug}

			// Manage the fail case
			if j == randNb {
				rpbckmock.RequestNextBuildFailed()
				cmdToPlay = append(cmdToPlay, "--noerror")
			}

			_, err = executeCommand(cmdToPlay...)
			if err != nil {
				return err
			}
		}

		// Now create the number of tests wanted
		for k := 0; k < testNb; k++ {
			_, err := executeCommand(appsCommand, "test", tmpAppSlug, "-p", prjSlug)
			if err != nil {
				test.Fatal(err)
			}
		}
	}

	return nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------
