/*
Copyright © 2022-2023 IoT.bzh Company
Contact: https://www.iot.bzh/licensing

This file is part of the redpesk-cli project, the redpesk command line tool.

$RP_BEGIN_LICENSE$
Commercial License Usage
 Licensees holding valid commercial IoT.bzh licenses may use this file in
 accordance with the commercial license agreement provided with the
 Software or, alternatively, in accordance with the terms contained in
 a written agreement between you and The IoT.bzh Company. For licensing terms
 and conditions see https://www.iot.bzh/terms-conditions. For further
 information use the contact form at https://www.iot.bzh/contact.

GNU General Public License Usage
 Alternatively, this file may be used under the terms of the GNU General
 Public license version 3. This license is as published by the Free Software
 Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
 of this file. Please review the following information to ensure the GNU
 General Public License requirements will be met
 https://www.gnu.org/licenses/gpl-3.0.html.
$RP_END_LICENSE$
*/

package cmd

import (
	"encoding/json"
	"errors"
	"fmt"
	"strconv"
	"text/tabwriter"

	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/printer"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/rpxxxapi/bckapi"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/utils"
	"git.ovh.iot/redpesk/redpesk-infra/rp-webserver/actions/rpapiv1"
	"github.com/gofrs/uuid"
	"github.com/spf13/cobra"
	"github.com/spf13/viper"
)

// -------------------------------------------------
// Types definition --------------------------------
// -------------------------------------------------

// flatCodeAuditTool type allow to exchange audit tool definition
// between function and it allows to have the definition of an audit
// tool in a "flat" structure
type flatCodeAuditTool struct {
	AuditType     string
	AuditCategory string
	Language      string
	Name          string
	Summary       string
	Version       string
}

// -------------------------------------------------
// -------------------------------------------------

// --------------------------------------------------------------
// Completion functions -----------------------------------------
// --------------------------------------------------------------

// getAuditToolsAsTab gets all the audit tools available to audit
// an application
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getAuditToolsAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	toolsList := []string{}

	// Retrieve the list of tools
	tools := rpapiv1.AuditStruct{}
	err := bckAPI.RPAuditToolsList(&tools)
	if err != nil {
		return toolsList, cobra.ShellCompDirectiveNoFileComp
	}

	// Convert to a structure easier to handle
	flatTools := convertToFlatAuditStruct(tools)

	// Now, save the names in the result
	for _, tool := range flatTools {
		// Add the  tool name in the suggestions
		toolsList = append(toolsList, tool.Name)
	}

	return toolsList, cobra.ShellCompDirectiveNoFileComp
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Command functions --------------------------------------------
// --------------------------------------------------------------
// auditApplication launch an audit on the existing application
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func auditApplication(cmd *cobra.Command, args []string) error {

	RPPrinter.CustomPrintLn(appsAuditCmdAnsLine0)

	// Get the application given by the user
	app, err := genericGetAppFromArgs(cmd, args, false)
	if err != nil {
		return err
	}

	RPPrinter.CustomPrint(appsAuditCmdAnsLine2)

	// Get the tools requested by the user
	toolsStrSlice, err := cmd.Flags().GetStringSlice(appsAuditToolFlag)
	if err != nil {
		logFlagError(appsAuditToolFlag)
		return err
	}

	// If the tool slice is empty, retrieve the default ones
	if len(toolsStrSlice) == 0 {
		// Finally, fill the tools tab with the default ones
		toolsStrSlice, err = convertAppAuditsToolsToStrSlice(app.AuditsConf)
		if err != nil {
			return err
		}
	}

	toolsForRequest, err := checkAndPrepareToolsRequest(toolsStrSlice)
	if err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	// Tell the user that we start the audit of the application
	RPPrinter.CustomPrintF(appsAuditCmdAnsLine1, app.Slug)

	// Ask the backend to start the audit
	auditAns := rpapiv1.RPAudit{}
	if err := bckAPI.ApplicationAudit(app.ID.String(), &toolsForRequest, &auditAns); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	// Display something to the user if in raw mode
	RPPrinter.CustomPrintObject(auditAns, false, displayNothing)

	// Check from viper the non-blocking flag to see if we wait
	nonBlocking := viper.GetBool(nonBlockingFlag)
	if nonBlocking {
		// The user do not want blocking functions
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
		return nil
	}

Wait_loop:
	// Block until we receive a corresponding websocket event
	for {
		// Wait for an event reception
		rEvent := <-cmdEventChannel

		switch rEvent.Type {
		case rpapiv1.EvtAuditStateChange:
			// Decode the event data
			rAudit, err := rEvent.DecodeAudit()
			if err != nil {
				RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
				return err
			}
			// Check if it is the audit we are monitoring and if it is stable
			if (rAudit.ID == auditAns.ID) && (rAudit.Status == rpapiv1.RPAuditStatusStable) {
				RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
				break Wait_loop
			}
		case rpapiv1.EVTError, rpapiv1.EVTSessionExpired:
			// Function handling the error
			err = handleErrorEvent(uuid.Nil, rEvent, true)
			if err != nil {
				return err
			}
		}
	}

	return nil
}

// auditsListTools list the available audit tools in the factory
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func auditsListTools(cmd *cobra.Command, args []string) error {
	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Retrieve the list of tools
	tools := rpapiv1.AuditStruct{}
	if err = bckAPI.RPAuditToolsList(&tools); err != nil {
		return err
	}

	// Display the list of tools
	err = RPPrinter.CustomPrintObject(tools, verbose, printAuditTools)

	return err
}

// auditsList list the run audits for an application
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func auditsList(cmd *cobra.Command, args []string) error {

	// Get the application given by the user
	app, err := genericGetAppFromArgs(cmd, args, false)
	if err != nil {
		return err
	}

	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Prepare the function needed by the printer to get the audits
	getAuditsCB := func(pageNb, perPage int) (interface{}, printer.RPPrinterPaging, error) {
		// Get audits list
		audits := rpapiv1.RPAudits{}
		if err := bckAPI.ApplicationsAuditsListGetByPage(app.ID.String(), &audits, pageNb, perPage, ""); err != nil {
			return audits.Audits, printer.RPPrinterPaging{}, err
		}
		// Get the paging interface from the structure retrieved
		pagingStruct, _ := printer.GetPagingFromInterface(&audits)

		return audits.Audits, pagingStruct, nil
	}

	// Prepare the structure to manage paging
	pagingStruct := printer.RPPagingToolsStruct{
		Verbose:       verbose,
		VerboseLineNb: auditDisplayVerboseLineNb,
		GetObjCB:      getAuditsCB,
		PrintFancyCB:  displayAudits,
	}

	return RPPrinter.InteractivelyPrintPages(&pagingStruct, viper.GetBool(NoHeaderFlag))
}

// auditsGetAndDisplay get an audit run from its id and display it
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func auditsGetAndDisplay(cmd *cobra.Command, args []string) error {

	// Retrieve the audit ID from the args given by the user
	auditID, err := getAuditIDFromArgs(cmd, args, appsForceAppFlag, "application", getAppUUIDCallBack)
	if err != nil {
		return err
	}
	auditStr := strconv.Itoa(auditID)

	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Get the properties of this audit
	audit := rpapiv1.RPAuditFull{}
	if err := bckAPI.RPAuditGet(auditStr, &audit); err != nil {
		return err
	}

	// Display the resulting tests
	err = RPPrinter.CustomPrintObject(audit, verbose, displayOneAudit)

	return err
}

// auditGetLogs allow to download the logs for a given audit
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func auditGetLogs(cmd *cobra.Command, args []string) error {

	RPPrinter.CustomPrintLn(appsAuditsLogsCmdAnsLine0)

	// Retrieve the audit ID from the args given by the user
	auditID, err := getAuditIDFromArgs(cmd, args, appsForceAppFlag, "application", getAppUUIDCallBack)
	if err != nil {
		return err
	}
	auditStr := strconv.Itoa(auditID)

	// Get the directory flag
	fullDirPath, err := manageDownloadDir(cmd, "rp-audits-logs", auditStr)
	if err != nil {
		return err
	}

	RPPrinter.CustomPrintF(appsAuditsLogsCmdAnsLine1, auditID)

	// Download the logs
	downloadedFile, err := bckAPI.RPAuditGetLogs(auditStr, fullDirPath)
	if err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	// Say to the user that the file has been downloaded
	RPPrinter.CustomColorPrintF(printer.ColorCyan, appsAuditsLogsCmdAnsLine2, auditID, downloadedFile)

	return nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Utilities functions ------------------------------------------
// --------------------------------------------------------------

// convertToFlatAuditStruct allows to convert a AuditStruct object
// to a slice of flatCodeAuditTool that is easier to manipulate
// afterwards.
//
// :param audit: Structure containing the audit tools available
// :type audit: rpapiv1.AuditStruct
// :returns: []flatCodeAuditTool -- a slice of flat structure, easier to use
func convertToFlatAuditStruct(audit rpapiv1.AuditStruct) []flatCodeAuditTool {
	result := []flatCodeAuditTool{}

	for _, serviceStruct := range audit.Audit {

		// Convert to a structure that we know
		entry := bckapi.RPAuditEntry{}
		dataBytes, _ := json.Marshal(serviceStruct.Entry)
		json.Unmarshal(dataBytes, &entry)

		for _, tools := range entry.Tools {
			for _, tool := range tools.Tools {
				result = append(result, flatCodeAuditTool{
					AuditType:     string(serviceStruct.Type),
					AuditCategory: entry.Category,
					Language:      tools.Language,
					Name:          tool.Name,
					Summary:       tool.Summary,
					Version:       tool.Version,
				})
			}
		}
	}
	return result
}

// convertAppAuditsToolsToStrSlice allows to convert a AuditStruct contained as string
// in an application structure (field audits_conf), into a slice of string, corresponding
// to the default tools
//
// :param auditConf: Audit configuration available in the application
// :type auditConf: string
// :returns: []string, error -- a slice containing all the tool names and an error if a problem occurred
func convertAppAuditsToolsToStrSlice(auditConf string) ([]string, error) {
	result := []string{}

	// Convert the audit configuration to a structure we know
	auditStruct := rpapiv1.AuditStruct{}
	err := json.Unmarshal([]byte(auditConf), &auditStruct)
	if err != nil {
		return result, err
	}

	for _, serviceStruct := range auditStruct.Audit {

		// Convert to a structure that we know
		entry := make(map[string][]string)
		dataBytes, _ := json.Marshal(serviceStruct.Entry)
		err = json.Unmarshal(dataBytes, &entry)
		if err != nil {
			return result, err
		}

		toolsTab, ok := entry["tools"]
		if ok {
			result = append(result, toolsTab...)
		}
	}
	return result, nil
}

// checkAndPrepareToolsRequest is a function allowing to check that all the tools specified
// by the user are available. It also creates the RPAuditStruct structure to pass to the request
// that will be sent to trigger an audit
//
// :param audit: Structure containing the audit tools available
// :type audit: rpapiv1.AuditStruct
// :returns: []flatCodeAuditTool -- a slice of flat structure, easier to use
func checkAndPrepareToolsRequest(userTools []string) (rpapiv1.AuditStruct, error) {
	// Retrieve the list of tools
	tools := rpapiv1.AuditStruct{}
	err := bckAPI.RPAuditToolsList(&tools)
	if err != nil {
		return tools, err
	}

	// Get a slice of flat structure, it is easier to use
	flatAuditTools := convertToFlatAuditStruct(tools)

	// Fill a slice containing only the tools name
	// Fill also a map containing the tools sort by their type
	availableToolNames := []string{}
	toolsAndTypeMap := make(map[string][]string)
	for _, tool := range flatAuditTools {
		// Complete the available tools list
		availableToolNames = append(availableToolNames, tool.Name)

		// Add an entry in our map if it has been specified in the user slice
		if utils.ContainsString(userTools, tool.Name) {
			if ttm, ok := toolsAndTypeMap[tool.AuditType]; ok {
				ttm = append(ttm, tool.Name)
				toolsAndTypeMap[tool.AuditType] = ttm
			} else {
				toolsAndTypeMap[tool.AuditType] = []string{tool.Name}
			}
		}
	}

	// Check that the tools given by the user are available
	for _, toolItem := range userTools {
		if !utils.ContainsString(availableToolNames, toolItem) {
			errMsg := fmt.Sprintf(auditToolNotAvailable, toolItem)
			return tools, errors.New(errMsg)
		}
	}

	// Now that we are sure that all the tools given by the user are available, we can construct the resulting structure
	resultTools := rpapiv1.AuditStruct{}
	for auditType, toolsList := range toolsAndTypeMap {
		resultTools.Audit = append(resultTools.Audit, rpapiv1.AuditBaseStruct{
			Type:  rpapiv1.RPAuditType(auditType),
			Entry: map[string][]string{"tools": append([]string{}, toolsList...)},
		})
	}

	return resultTools, nil
}

// getLatestAudit is a private function allowing to get the latest audit of an application.
//
// :param appID: The application ID for which the audits are wanted
// :type appID: string
// :returns: int, error -- the last test ID and an error if something went wrong, nil in the other case
func getLatestAudit(appID string) (int, error) {

	// Get the first page of audits for this app in descending sorting
	audits := rpapiv1.RPAudits{}
	if err := bckAPI.ApplicationsAuditsListGetByPage(appID, &audits, 1, 1, "desc"); err != nil {
		return -1, err
	}

	if len(audits.Audits) == 0 {
		return -1, errors.New(noAuditFoundInAppMsg)
	}

	return audits.Audits[0].ID, nil
}

// getAuditIDFromArgs allows to retrieve the audit ID wanted from what the
// user gave to rp-cli. The user could either give an app slug/UUID or
// directly an audit ID.
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param forceParentFlag: Flag allowing the user to force the arg to be interpreted as parent ID
// :type forceParentFlag: string
// :param parentType: Type of the parent (image/app)
// :type parentType: string
// :param ObjGetUuidCb: Function allowing to get the UUID of the "parent" object (image or app) if necessary
// :type ObjGetUuidCb: func(*cobra.Command, []string) (uuid.UUID, error)
// :returns: int, error -- the build ID and an error if something went wrong, nil in the other case
func getAuditIDFromArgs(cmd *cobra.Command, args []string, forceParentFlag, parentType string, ObjGetUuidCb func(*cobra.Command, []string) (uuid.UUID, error)) (int, error) {

	// Try to get the "force-parent" flag
	forceParent, err := cmd.Flags().GetBool(forceParentFlag)
	if err != nil {
		logFlagError(forceParentFlag)
		return -1, err
	}

	// Get the first argument of the command line
	argID, err := getFirstArg(args, fmt.Sprintf(objectIDNotSetMsg, "app/audit"), fmt.Sprintf(objectIDRequiredMsg, "app/audit"))
	if err != nil {
		return -1, err
	}

	// Try to convert the argument given to an int
	auditIDInt, err := strconv.Atoi(argID)
	isInteger := (err == nil)

	// Act according to what we have

	// If the arg given is an int and that the user do not force the arg to be a parent object
	// return this integer directly
	if isInteger && !forceParent {
		return auditIDInt, nil
	}

	// In the other case, the user argument is the parent slug, use it to retrieve the latest test

	// Retrieve the UUID of the parent
	parentUUID, err := ObjGetUuidCb(cmd, args)
	if err != nil {
		return -1, err
	}
	tmpTestID, err := getLatestAudit(parentUUID.String())
	if err != nil {
		return -1, err
	}

	return tmpTestID, nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Display functions --------------------------------------------
// --------------------------------------------------------------

// printAuditTools display the list of the current audit tools
// available in the current Redpesk instance
//
// :param services: List of services to display
// :type services: interface{}
// :param verbose: Tell if the print needs to be verbose or not
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func printAuditTools(tools interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Convert to the right type
	toolsToPrint := tools.(rpapiv1.AuditStruct)

	// Get a slice of flat structure, it is easier to use
	flatAuditTools := convertToFlatAuditStruct(toolsToPrint)

	if verbose {
		for index, tool := range flatAuditTools {
			if index > 0 {
				fmt.Fprintln(writer)
			}
			fmt.Fprintln(writer, appsAuditsListToolsAnsLine3, tool.Name)
			fmt.Fprintln(writer, appsAuditsListToolsAnsLine5, tool.Version)
			fmt.Fprintln(writer, appsAuditsListToolsAnsLine4, tool.Summary)
			fmt.Fprintln(writer, appsAuditsListToolsAnsLine2, tool.Language)
			fmt.Fprintln(writer, appsAuditsListToolsAnsLine0, tool.AuditType)
			fmt.Fprintln(writer, appsAuditsListToolsAnsLine1, tool.AuditCategory)
		}
	} else {
		// Print the header for services
		fmt.Fprintln(writer, appsAuditsListToolsHeader)

		for _, tool := range flatAuditTools {
			fmt.Fprintln(writer, tool.Name, "\t", tool.Version, "\t", tool.AuditType, "\t", tool.Language)
		}
	}

	return nil
}

// displayAudits is a private function allowing to display
// a slice of audits in the console
//
// :param audits: tab of audits to display
// :type audits: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func displayAudits(audits interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Convert to the right type
	auditsToPrint := audits.([]rpapiv1.RPAudit)

	// Display result
	for index, audit := range auditsToPrint {
		// Eventually print something before
		if verbose && (index > 0) {
			fmt.Fprintln(writer)
		} else if !verbose && (index == 0) {
			manageHeaderPrint(appsAuditsListCmdHeader, writer)
		}

		// Print this audit
		err := _printOneAudit(audit, verbose, false, writer)
		if err != nil {
			return err
		}
	}

	return nil
}

// displayOneAudit is a private function allowing to display
// one audit in the console
//
// :param audit: Audit to display
// :type audit: interface{}
// :param verbose: Verbose flag passed in the command line
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func displayOneAudit(audit interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Print the app
	err := _printOneAudit(audit, verbose, true, writer)
	if err != nil {
		return err
	}
	return nil
}

// _printOneAudit is a private function allowing to print one audit object in the console
//
// :param auditInterface: audit to print (can be rpapiv1.RPAuditFull or rpapiv1.RPAudit)
// :type auditInterface: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param isAlone: True if there is only one audit to display in the context where this function is called
// :type isAlone: bool
// :param writer: Writer to used to print the project information
// :type writer: *tabwriter.Writer
func _printOneAudit(auditInterface interface{}, verbose, isAlone bool, writer *tabwriter.Writer) error {
	audit := rpapiv1.RPAuditFull{}

	// Cast to the full type in all case but use "full" parameters as indicated by the user
	// We have to pass through the Marshal/Unmarshal tricks to not have problem during conversion from
	// RPApplication -> RPApplicationFull at runtime
	dataBytes, _ := json.Marshal(auditInterface)
	json.Unmarshal(dataBytes, &audit)

	if verbose {
		// Create the string containing the results
		fmt.Fprintln(writer, appsAuditsListCmdVerboseLine0, fmt.Sprint(audit.ID))
		fmt.Fprintln(writer, appsAuditsListCmdVerboseLine1, audit.Parent.AppSlug)
		fmt.Fprintln(writer, appsAuditsListCmdVerboseLine2, audit.Parent.ProjectSlug)
		fmt.Fprintln(writer, appsAuditsListCmdVerboseLine3, audit.Status)
		fmt.Fprintln(writer, appsAuditsListCmdVerboseLine4, fmt.Sprint(audit.SuccessNb))
		fmt.Fprintln(writer, appsAuditsListCmdVerboseLine5, fmt.Sprint(audit.FailedNb))
		fmt.Fprintln(writer, appsAuditsListCmdVerboseLine6, fmt.Sprint(audit.SkippedNb))
		fmt.Fprintln(writer, appsAuditsListCmdVerboseLine7, fmt.Sprint(audit.TotalNb))
		fmt.Fprintln(writer, appsAuditsListCmdVerboseLine8, audit.CreatedAt.String())
		fmt.Fprintln(writer, appsAuditsListCmdVerboseLine9, audit.UpdatedAt.String())
		fmt.Fprintln(writer, appsAuditsListCmdVerboseLine10, audit.UserLauncherPseudo)

		if audit.FullData {
			fmt.Fprintln(writer)
			fmt.Fprintln(writer, appsAuditsListCmdVerboseLine11)
			fmt.Fprintln(writer)

			for _, result := range audit.Result.Audit {
				// Convert the result to something understandable for us
				results := bckapi.RPAuditResults{}
				dataBytes, err := json.Marshal(result.Entry)
				if err != nil {
					return err
				}
				err = json.Unmarshal(dataBytes, &results)
				if err != nil {
					return err
				}

				fmt.Fprintln(writer, appsAuditsListToolsAnsLine0, result.Type)
				fmt.Fprintln(writer, appsAuditsListCmdVerboseLine19)
				_printCodeAuditResults(results.TotalReport, writer)

				for _, toolResult := range results.Analyse {
					fmt.Fprintln(writer)
					fmt.Fprintln(writer, appsAuditsListCmdVerboseLine20, toolResult.Tool)
					_printCodeAuditResults(toolResult.Report, writer)
				}
			}
		}
	} else {
		if isAlone {
			fmt.Fprintln(writer, appsAuditsListCmdHeader)
		}
		fmt.Fprintln(writer, fmt.Sprint(audit.ID), "\t", audit.Parent.AppSlug, "\t", audit.Status, "\t", fmt.Sprintf("%d/%d/%d/%d", audit.SuccessNb, audit.FailedNb, audit.SkippedNb, audit.TotalNb))
	}

	return nil
}

// _printCodeAuditResults allows to print the detailed results of a code audit
//
// :param report: structure containing the results to print
// :type report: RPAuditCodeReport
// :param writer: Writer to used to print the project information
// :type writer: *tabwriter.Writer
func _printCodeAuditResults(report bckapi.RPAuditCodeReport, writer *tabwriter.Writer) {
	fmt.Fprintln(writer, appsAuditsListCmdVerboseLine12, report.Info)
	fmt.Fprintln(writer, appsAuditsListCmdVerboseLine13, report.Unknown)
	fmt.Fprintln(writer, appsAuditsListCmdVerboseLine14, report.Low)
	fmt.Fprintln(writer, appsAuditsListCmdVerboseLine15, report.Medium)
	fmt.Fprintln(writer, appsAuditsListCmdVerboseLine16, report.High)
	fmt.Fprintln(writer, appsAuditsListCmdVerboseLine17, report.Critical)
	fmt.Fprintln(writer, appsAuditsListCmdVerboseLine18, report.Total)
}

// --------------------------------------------------------------
// --------------------------------------------------------------
