/*
Copyright © 2020-2023 IoT.bzh Company
Contact: https://www.iot.bzh/licensing

This file is part of the redpesk-cli project, the redpesk command line tool.

$RP_BEGIN_LICENSE$
Commercial License Usage

	Licensees holding valid commercial IoT.bzh licenses may use this file in
	accordance with the commercial license agreement provided with the
	Software or, alternatively, in accordance with the terms contained in
	a written agreement between you and The IoT.bzh Company. For licensing terms
	and conditions see https://www.iot.bzh/terms-conditions. For further
	information use the contact form at https://www.iot.bzh/contact.

GNU General Public License Usage

	Alternatively, this file may be used under the terms of the GNU General
	Public license version 3. This license is as published by the Free Software
	Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
	of this file. Please review the following information to ensure the GNU
	General Public License requirements will be met
	https://www.gnu.org/licenses/gpl-3.0.html.

$RP_END_LICENSE$
*/
package cmd

import (
	"path"
	"strings"
	"testing"

	"github.com/stretchr/testify/assert"
)

const (
	auditDisplayVerboseFullBaseLineNb = 23
	auditLineNbPerToolResults         = 9
	auditDisplayNoVerboseColNb        = 4
)

// TestListAuditTools is the function allowing to test if
// the command "rp-cli applications audits tools" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestListAuditTools(test *testing.T) {

	// Not verbose

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"Name":     `^[a-z_]+$`,
		"Version":  `^[0-9.]+$`,
		"Type":     `^code$`,
		"Language": `^[a-zA-Z- ()/+]+$`,
	}

	// Run the command and the tests (no verbose and verbose, since this is the same result)
	checkCmdAnsKeysRegexpCol(test, 4, regexpMap, appsCommand, appsAuditsCmd, appsAuditsListToolsCmd)

	// Verbose

	// Create the map containing the regexp we want to match in the answer
	regexpMap = map[string]string{
		"Name":     `^[a-z_]+$`,
		"Version":  `^[0-9.]+$`,
		"Summary":  `^[a-zA-Z- ()/+]+$`,
		"Language": `^[a-zA-Z- ()/+]+$`,
		"Type":     `^code$`,
		"Category": `^[a-zA-Z- ()/+]+$`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, (2*6)+1, regexpMap, appsCommand, appsAuditsCmd, appsAuditsListToolsCmd, "-v")
}

// TestAppsAddWithAuditVerbose is the function allowing to test if
// the command "rp-cli applications add ..." is working correctly when
// code audit tools are specified
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsAddWithAuditVerbose(test *testing.T) {

	// Create an application with default audit tools set-up!
	err := createPrjAndAppWithAudits(test, "audit-test-project-0", "redpesk-lts-arz", "audit-test-app", "audit-test-app", []string{"flawfinder"})
	if err != nil {
		test.Fatal(err)
	}

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":                        `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":                       `^audit-test-app$`,
		"Package Name":               `^audit-test-app$`,
		"Slug":                       `^audit-test-app$`,
		"Type":                       `^standard$`,
		"Parent project":             `^audit-test-project-0$`,
		"Description":                `.+`,
		"Source URL":                 `http(s)*://`,
		"Source Revision":            `^[a-z0-9-]+$`,
		"External Specfile":          `(\[Not set\]|(http(s)*://))`,
		"Specfile type":              `(internal|relative_path|external_url)`,
		"Distributions":              `^redpesk-lts-arz.*`,
		"Auto on board tests":        `^(ON|OFF) \[Opt\]$`,
		"Ignore missing redtest RPM": `^(ON|OFF)$`,
		"Auto security audits tests": `^(ON|OFF) \[Opt\]$`,
		"Audits settings":            `\{"audit":\[\{"type":"code","entry":\{"tools":\["flawfinder"\]\}\}\]\}`,
		"Build type":                 `^(standard_build_cross|emulated_build|accel_build_cross)$`,
		"Ignore check deps result":   `^(ON|OFF)$`,
		"Creation time":              `.+`,
		"Update time":                `.+`,
		"Bookmarked":                 `^false$`,
		"Services":                   `.*`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, appDisplayVerboseFullLineNb, regexpMap, appsCommand, "get", "audit-test-app", "-v")
}

// TestAppsUpdateAuditTools is the function allowing to test if
// the command "rp-cli application update --audits ..." is working correctly
// when the update is done on audit tools
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppsUpdateAuditTools(test *testing.T) {

	// Create an application with default audit tools set-up!
	err := createPrjAndAppWithAudits(test, "audit-test-project-1", "redpesk-lts-arz", "audit-test-update-app", "audit-test-update-app", []string{"flawfinder"})
	if err != nil {
		test.Fatal(err)
	}

	// Create the update command
	updateCmd := getUpdateAppCmd("audit-test-update-app", "--audits", "gosec")

	// Run the update command and check that all is [OK]
	checkActionCmdAns(test, updateCmd...)

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":                        `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":                       `^audit-test-update-app$`,
		"Package Name":               `^audit-test-update-app$`,
		"Slug":                       `^audit-test-update-app$`,
		"Type":                       `^standard$`,
		"Parent project":             `^audit-test-project-1$`,
		"Description":                `.+`,
		"Source URL":                 `http(s)*://`,
		"Source Revision":            `^[a-z0-9-]+$`,
		"External Specfile":          `(\[Not set\]|(http(s)*://))`,
		"Specfile type":              `(internal|relative_path|external_url)`,
		"Distributions":              `^redpesk-lts-arz.*`,
		"Auto on board tests":        `^(ON|OFF) \[Opt\]$`,
		"Ignore missing redtest RPM": `^(ON|OFF)$`,
		"Auto security audits tests": `^(ON|OFF) \[Opt\]$`,
		"Audits settings":            `\{"audit":\[\{"type":"code","entry":\{"tools":\["gosec"\]\}\}\]\}`,
		"Build type":                 `^(standard_build_cross|emulated_build|accel_build_cross)$`,
		"Ignore check deps result":   `^(ON|OFF)$`,
		"Creation time":              `.+`,
		"Update time":                `.+`,
		"Bookmarked":                 `^false$`,
		"Services":                   `.*`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, appDisplayVerboseFullLineNb, regexpMap, appsCommand, "get", "audit-test-update-app", "-v")
}

// TestAppAudit is the function allowing to test if
// the command "rp-cli applications audit" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppAudit(test *testing.T) {

	// Create an application with default audit tools set-up!
	err := createPrjAndAppWithAudits(test, "audit-test-project-2", "redpesk-lts-arz", "run-audit-app", "run-audit-app", []string{"gosec"})
	if err != nil {
		test.Fatal(err)
	}

	// Run the audit command and check that all is [OK]
	checkActionCmdAns(test, appsCommand, "audit", "run-audit-app")
}

// TestAppAudit is the function allowing to test if
// the command "rp-cli applications audits get" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppAuditGet(test *testing.T) {

	// Create an application with default audit tools set-up!
	err := createPrjAndAppWithAudits(test, "audit-test-project-3", "redpesk-lts-arz", "get-audit-app", "get-audit-app", []string{"flawfinder"})
	if err != nil {
		test.Fatal(err)
	}

	// Run the audit command and check that all is [OK]
	checkActionCmdAns(test, appsCommand, "audit", "get-audit-app")

	// Run another time
	checkActionCmdAns(test, appsCommand, "audit", "get-audit-app")

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":           `^[0-9]+$`,
		"Application":   `^get-audit-app$`,
		"Project":       `^audit-test-project-3$`,
		"Status":        `^stable$`,
		"Success":       `^[0-9]+$`,
		"Fail":          `^[0-9]+$`,
		"Skip":          `^[0-9]+$`,
		"Total":         `^[0-9]+$`,
		"Creation time": `.+`,
		"Update time":   `.+`,
		"User launcher": `.*`,
	}

	// Run the command and the tests
	cmdOutput := checkCmdAnswerKeysRegexp(test, auditDisplayVerboseFullBaseLineNb+1*auditLineNbPerToolResults, regexpMap, appsCommand, appsAuditsCmd, "get", "get-audit-app", "-v")

	// Extract the detailed results
	splitedRes := strings.Split(cmdOutput, "[Detailed results]")
	detailRes := ""
	if assert.Equal(test, 2, len(splitedRes)) {
		detailRes = strings.TrimSpace(splitedRes[1])
	} else {
		test.Fatal(err)
	}

	// Create the map containing the regexp we want to match in the answer
	mapToCheck := map[string]string{
		"Type":            `^code$`,
		"Overall results": `.*`,
		"Tool":            `^(flawfinder|gosec)$`,
		"Info":            `^[0-9]+$`,
		"Unknown":         `^[0-9]+$`,
		"Low":             `^[0-9]+$`,
		"Medium":          `^[0-9]+$`,
		"High":            `^[0-9]+$`,
		"Critical":        `^[0-9]+$`,
		"Total":           `^[0-9]+$`,
	}

	// Check the detailed result
	checkObjectKeysRegexp(test, detailRes, 2*auditLineNbPerToolResults, mapToCheck, true)

	// Get the highest ID for the audit
	highestID, err := getHighestIDFromOutput("ID", appsCommand, appsAuditsCmd, "list", "get-audit-app")
	if err != nil {
		test.Fatal(err)
	}

	// Create the map containing the regexp we want to match in the answer
	regexpMapNotVerbose := map[string]string{
		"ID":                      `^[0-9]+$`,
		"App.":                    `^get-audit-app$`,
		"Status":                  `^stable$`,
		"Success/Fail/Skip/Total": `^([0-9]+/){3}[0-9]+$`,
	}

	checkCmdAnsKeysRegexpCol(test, auditDisplayNoVerboseColNb, regexpMapNotVerbose, appsCommand, appsAuditsCmd, "get", highestID)
}

// TestAppAuditsList is the function allowing to test if
// the command "rp-cli applications audits list" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppAuditsList(test *testing.T) {
	// Create an application with default audit tools set-up!
	err := createPrjAndAppWithAudits(test, "audit-test-project-4", "redpesk-lts-arz", "list-audit-app", "list-audit-app", []string{"flawfinder", "gosec"})
	if err != nil {
		test.Fatal(err)
	}

	// Run the audit command and check that all is [OK]
	checkActionCmdAns(test, appsCommand, "audit", "list-audit-app")
	// Run another time
	checkActionCmdAns(test, appsCommand, "audit", "list-audit-app")
	// And another time
	checkActionCmdAns(test, appsCommand, "audit", "list-audit-app")

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":           `^[0-9]+$`,
		"Application":   `^list-audit-app$`,
		"Project":       `^audit-test-project-4$`,
		"Status":        `^stable$`,
		"Success":       `^[0-9]+$`,
		"Fail":          `^[0-9]+$`,
		"Skip":          `^[0-9]+$`,
		"Total":         `^[0-9]+$`,
		"Creation time": `.+`,
		"Update time":   `.+`,
		"User launcher": `.*`,
	}

	// Run the command and the tests
	checkCmdAnswerListKeysRegexp(test, auditDisplayVerboseLineNb, "", regexpMap, appsCommand, appsAuditsCmd, "list", "list-audit-app", "-v")

	// Create the map containing the regexp we want to match in the answer
	regexpMapNotVerbose := map[string]string{
		"ID":                      `^[0-9]+$`,
		"App.":                    `^list-audit-app$`,
		"Status":                  `^stable$`,
		"Success/Fail/Skip/Total": `^([0-9]+/){3}[0-9]+$`,
	}

	checkCmdAnsKeysRegexpCol(test, auditDisplayNoVerboseColNb, regexpMapNotVerbose, appsCommand, appsAuditsCmd, "list", "list-audit-app")
}

// TestAppAuditsLogs is the function allowing to test if
// the command "rp-cli applications audits logs" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppAuditsLogs(test *testing.T) {
	// Create an application with default audit tools set-up!
	err := createPrjAndAppWithAudits(test, "audit-test-project-5", "redpesk-lts-arz", "logs-audit-app", "logs-audit-app", []string{"gosec"})
	if err != nil {
		test.Fatal(err)
	}

	// Run the audit command and check that all is [OK]
	checkActionCmdAns(test, appsCommand, "audit", "logs-audit-app")

	// Now try to download audit logs from the server
	cmdOutput, err := executeCommand(appsCommand, appsAuditsCmd, "logs", "logs-audit-app", "-d", "/tmp")
	if err != nil {
		test.Fatal(err)
	}

	// Check that the output contains the needed element
	assert.Contains(test, cmdOutput, appsAuditsLogsCmdAnsLine0)
	assert.Contains(test, cmdOutput, "Downloading logs for audit")
	assert.Contains(test, cmdOutput, "The audit logs for audit")
	assert.Contains(test, cmdOutput, "[OK]")
	assert.NotContains(test, cmdOutput, "[KO]")

	// Get the highest ID for the audit
	highestID, err := getHighestIDFromOutput("ID", appsCommand, appsAuditsCmd, "list", "logs-audit-app")
	if err != nil {
		test.Fatal(err)
	}

	// Check that the file containing the logs has been downloaded
	assert.FileExists(test, path.Join("/tmp", "rp-audits-logs", highestID, "mocked_audit_logs.zip"))
}

// --------------------------------------------------------------
// Utilities functions ------------------------------------------
// --------------------------------------------------------------

// createPrjAndAppWithAudits is an initializer function to create a project and an application with
// some audits tools initialized
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param prjName: Name of the project to create
// :type prjName: string
// :param distro: Name of the distro to set for the project
// :type distro: string
// :param appName: Name of the application to create
// :type appName: string
// :param pkgName: Package name for the application to create
// :type pkgName: string
// :param auditTools: List of audit tools to set as default
// :type auditTools: []string
// :returns: []string -- the value of the wanted constant
func createPrjAndAppWithAudits(test *testing.T, prjName, distro, appName, pkgName string, auditTools []string) error {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd(prjName, distro)
	err := addObjectIfNotExists(test, prjsCommand, prjName, cmdToRun1...)
	if err != nil {
		return err
	}

	// Create the add application command
	cmdToRun2 := getMainAddAppWithAuditsCmd(appName, pkgName, prjName, auditTools)

	// Create the application if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, appsCommand, appName, cmdToRun2...)

	return err
}

// getMainAddAppWithAuditsCmd is an initializer function to get the mainly used command to create an application with
// default audit tools
//
// :param name: Name of the application to create
// :type name: string
// :param pkgName: Name of the package created by the application
// :type pkgName: string
// :param project: Project parent of the app to create
// :type project: string
// :param auditTools: List of audit tools to set as default
// :type auditTools: []string
// :returns: []string -- the value of the wanted constant
func getMainAddAppWithAuditsCmd(name, pkgName, project string, auditTools []string) []string {
	result := []string{
		appsCommand,
		appsAddCmd,
		"-n",
		name,
		"--pkg-name",
		pkgName,
		"-p",
		project,
		"--source-url",
		"http://git.ovh.iot/redpesk/redpesk-samples/test-application.git",
		"-d",
		"This is a testing application",
	}

	for _, tool := range auditTools {
		result = append(result, "--audits", tool)
	}

	return result
}

// --------------------------------------------------------------
// --------------------------------------------------------------
