/*
Copyright © 2020-2023 IoT.bzh Company
Contact: https://www.iot.bzh/licensing

This file is part of the redpesk-cli project, the redpesk command line tool.

$RP_BEGIN_LICENSE$
Commercial License Usage

	Licensees holding valid commercial IoT.bzh licenses may use this file in
	accordance with the commercial license agreement provided with the
	Software or, alternatively, in accordance with the terms contained in
	a written agreement between you and The IoT.bzh Company. For licensing terms
	and conditions see https://www.iot.bzh/terms-conditions. For further
	information use the contact form at https://www.iot.bzh/contact.

GNU General Public License Usage

	Alternatively, this file may be used under the terms of the GNU General
	Public license version 3. This license is as published by the Free Software
	Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
	of this file. Please review the following information to ensure the GNU
	General Public License requirements will be met
	https://www.gnu.org/licenses/gpl-3.0.html.

$RP_END_LICENSE$
*/
package cmd

import (
	"fmt"
	"strconv"
	"text/tabwriter"

	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/printer"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/utils"
	"git.ovh.iot/redpesk/redpesk-infra/rp-webserver/actions/rpapiv1"
	"github.com/gofrs/uuid"
	"github.com/spf13/cobra"
	"github.com/spf13/viper"
)

func initBoardModelsCmd(rootCmd *cobra.Command) {

	// boardModelsCmd represents the board-models command
	var boardModelsCmd = &cobra.Command{
		Use:   boardModelsCommand,
		Short: boardModelsCmdShortDesc,
	}

	// listCmd represents the list command
	var listCmd = &cobra.Command{
		Use:   boardModelsListCmd,
		Short: boardModelsListCmdShortDesc,
		Long:  boardModelsListCmdLongDesc,
		RunE:  listBoardModels,
	}

	// getCmd represents the get command
	var getCmd = &cobra.Command{
		Use:               boardModelsGetCmd,
		Short:             boardModelsGetCmdShortUsage,
		RunE:              getBoardModel,
		ValidArgsFunction: getFullListBoardModelsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// Here you will define your flags and configuration settings.

	// Persistent Flags -----------------------------------------
	// ----------------------------------------------------------

	// Local Flags ----------------------------------------------
	listCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	getCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	// ----------------------------------------------------------

	// Add some flags as required -------------------------------
	// ----------------------------------------------------------

	// Add the group commands "board-models" as subcommand of rootCmd
	rootCmd.AddCommand(boardModelsCmd)
	// Add the command "list" as a subcommand of boardModelsCmd
	boardModelsCmd.AddCommand(listCmd)
	// Add the command "get" as a subcommand of boardModelsCmd
	boardModelsCmd.AddCommand(getCmd)
}

// --------------------------------------------------------------
// Completion functions -----------------------------------------
// --------------------------------------------------------------

// getListBoardModelsAsTab gets all the board models available in the Redpesk
// instance and return it as an array of strings
//
// :param categories: List of categories that interest us
// :type categories: []string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getBoardModelsOfCategoriesAsTab(categories []string) ([]string, cobra.ShellCompDirective) {
	boardsList := []string{}

	// Get the list of every available boards
	boards := []rpapiv1.RPModelBoard{}
	if err := bckAPI.RPBoardModelsGet(&boards, ""); err != nil {
		return boardsList, cobra.ShellCompDirectiveNoFileComp
	}

	// We got correctly the boards list, fill the boardsList tab
	for _, board := range boards {
		if utils.ContainsString(categories, string(board.Category)) {
			boardsList = append(boardsList, board.Name)
		}
	}

	return boardsList, cobra.ShellCompDirectiveNoFileComp
}

// getListRTMBoardModelsAsTab gets all the RTM board models available in the Redpesk
// instance and return it as an array of strings
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getListRTMBoardModelsAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	return getBoardModelsOfCategoriesAsTab([]string{string(rpapiv1.RPBoardTypeRTM)})
}

// getFullListBoardModelsAsTab gets all the board models available in the Redpesk
// instance and return it as an array of strings
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getFullListBoardModelsAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	return getBoardModelsOfCategoriesAsTab([]string{string(rpapiv1.RPBoardTypeRTM), string(rpapiv1.RPBoardTypeOTA), string(rpapiv1.RPBoardTypeDEV)})
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Command functions --------------------------------------------
// --------------------------------------------------------------

// listBoardModels gets and prints the list of all the available board
// models on the Redpesk backend
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func listBoardModels(cmd *cobra.Command, args []string) error {
	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Prepare the function needed by the printer to get Released Projects
	getBoardModelsCB := func(pageNb, perPage int) (interface{}, printer.RPPrinterPaging, error) {
		// Get models list
		models := rpapiv1.RPModelBoards{}
		if err := bckAPI.RPBoardModelsListByPage(&models, pageNb, perPage); err != nil {
			return models.ModelBoards, printer.RPPrinterPaging{}, err
		}
		// Get the paging interface from the structure retrieved
		pagingStruct, _ := printer.GetPagingFromInterface(&models)

		return models.ModelBoards, pagingStruct, nil
	}

	// Prepare the structure to manage paging
	pagingStruct := printer.RPPagingToolsStruct{
		Verbose:       verbose,
		VerboseLineNb: boardModelsDisplayListLineNb,
		GetObjCB:      getBoardModelsCB,
		PrintFancyCB:  displayBoardModels,
	}

	return RPPrinter.InteractivelyPrintPages(&pagingStruct, viper.GetBool(NoHeaderFlag))
}

// getBoardModel gets and prints the given board model
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func getBoardModel(cmd *cobra.Command, args []string) error {

	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Get the board ID wanted by the user
	brdMdlID, err := getFirstArg(args, fmt.Sprintf(objectIDNotSetMsg, "board model"), fmt.Sprintf(objectIDRequiredMsg, "board model"))
	if err != nil {
		return err
	}

	boardModel, err := getModelBoardStruct(brdMdlID)
	if err != nil {
		return err
	}

	// Print the result to the user
	return RPPrinter.CustomPrintObject(boardModel, verbose, displayOneBoardModel)
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Utilities functions ------------------------------------------
// --------------------------------------------------------------

// getModelBoardStruct allows to retrieve the full board structure
// from its name or UUID
//
// :param boardModelID: Name or UUID of the board model that interests us
// :type boardModelID: string
// :returns: rpapiv1.RPModelBoard, error -- the model board structure and an error if a problem appends
func getModelBoardStruct(boardModelID string) (rpapiv1.RPModelBoard, error) {
	// If empty string, returns an empty struct
	if boardModelID == "" {
		return rpapiv1.RPModelBoard{ID: uuid.Nil}, nil
	}

	// Check if the ID given is a valid UUID
	isValidUUID := utils.IsValidUUID(boardModelID)

	filterValue := ""
	if !isValidUUID {
		filterValue = "name=" + boardModelID
	}

	// Get the list of available boards with the right filter
	boardModels := []rpapiv1.RPModelBoard{}
	if err := bckAPI.RPBoardModelsGet(&boardModels, filterValue); err != nil {
		return rpapiv1.RPModelBoard{}, err
	}

	wantedIndex := -1
	// If the ID was directly an UUID, look for the right inside the list
	if isValidUUID {
		boardUUID, _ := uuid.FromString(boardModelID)
		for index, brdModel := range boardModels {
			if utils.UUIDEqual(brdModel.ID, boardUUID) {
				wantedIndex = index
				break
			}
		}
	} else {
		wantedIndex = 0
	}

	if len(boardModels) == 0 || wantedIndex == -1 {
		return rpapiv1.RPModelBoard{}, fmt.Errorf(boardModelIdentifierNotFound, boardModelID)
	}

	// We got correctly the board
	return boardModels[wantedIndex], nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Display functions --------------------------------------------
// --------------------------------------------------------------

// displayBoardModels is a private function allowing to display
// a slice of board models in the console
//
// :param boardModels: tab of board models to display
// :type boardModels: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func displayBoardModels(boardModels interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Convert to the right type
	brdModels := boardModels.([]rpapiv1.RPModelBoard)

	// Display result
	for index, model := range brdModels {
		if verbose && (index > 0) {
			fmt.Fprintln(writer)
		} else if !verbose && index == 0 {
			manageHeaderPrint(boardModelsListAnsHeader, writer)
		}
		// Print this project
		_printOneBoardModel(model, verbose, false, writer)
	}
	return nil
}

// displayOneBoardModel is a private function allowing to display
// one board model in the console
//
// :param boardModel: Board model to display
// :type boardModel: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :returns: error -- an error if something went bad
func displayOneBoardModel(boardModel interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Print the project
	_printOneBoardModel(boardModel, verbose, true, writer)
	return nil
}

// _printOneBoardModel display one board model available in the
// current Redpesk instance
//
// :param boardModel: Board model to print
// :type boardModel: interface{}
// :param verbose: Tell if the print needs to be verbose or not
// :type verbose: bool
// :param isAlone: Tell if the object is alone or not
// :type isAlone: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func _printOneBoardModel(boardModel interface{}, verbose, isAlone bool, writer *tabwriter.Writer) error {

	// Convert to the right type
	boardToPrint := boardModel.(rpapiv1.RPModelBoard)

	if verbose {
		cpuToPrint := boardToPrint.CPU
		if cpuToPrint == "" {
			cpuToPrint = mainNotSetMsg
		}
		fmt.Fprintln(writer, boardModelsListAnsLine0, boardToPrint.ID.String())
		fmt.Fprintln(writer, boardModelsListAnsLine1, boardToPrint.Name)
		fmt.Fprintln(writer, boardModelsListAnsLine2, boardToPrint.CreatedAt.String())
		fmt.Fprintln(writer, boardModelsListAnsLine3, boardToPrint.UpdatedAt.String())
		fmt.Fprintln(writer, boardModelsListAnsLine4, string(boardToPrint.Category))
		fmt.Fprintln(writer, boardModelsListAnsLine5, boardToPrint.Archi)
		fmt.Fprintln(writer, boardModelsListAnsLine6, cpuToPrint)
		fmt.Fprintln(writer, boardModelsListAnsLine7, strconv.FormatBool(boardToPrint.Virtual))
	} else {
		if isAlone {
			fmt.Fprintln(writer, boardModelsListAnsHeader)
		}
		fmt.Fprintln(writer, boardToPrint.ID.String(), "\t", boardToPrint.Name, "\t", boardToPrint.Category, "\t", boardToPrint.Archi, "\t", strconv.FormatBool(boardToPrint.Virtual))
	}

	return nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------
