/*
Copyright © 2020-2023 IoT.bzh Company
Contact: https://www.iot.bzh/licensing

This file is part of the redpesk-cli project, the redpesk command line tool.

$RP_BEGIN_LICENSE$
Commercial License Usage

	Licensees holding valid commercial IoT.bzh licenses may use this file in
	accordance with the commercial license agreement provided with the
	Software or, alternatively, in accordance with the terms contained in
	a written agreement between you and The IoT.bzh Company. For licensing terms
	and conditions see https://www.iot.bzh/terms-conditions. For further
	information use the contact form at https://www.iot.bzh/contact.

GNU General Public License Usage

	Alternatively, this file may be used under the terms of the GNU General
	Public license version 3. This license is as published by the Free Software
	Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
	of this file. Please review the following information to ensure the GNU
	General Public License requirements will be met
	https://www.gnu.org/licenses/gpl-3.0.html.

$RP_END_LICENSE$
*/
package cmd

import (
	"errors"
	"fmt"
	"strings"
	"testing"

	"github.com/stretchr/testify/assert"
)

// Constants definition
const (
	baseBranchProjectName = "branches-test-project-%d"
	baseBranchAppName     = "branches-test-app-%d"
)

// Variable definition
var branchesTestsCounter = 0

// --------------------------------------------------------------
// Initialization functions -------------------------------------
// --------------------------------------------------------------

// getPrjAndAppNames is an initialization function allowing to
// retrieve the project name and application name to use for the
// test to run
//
// :returns: string, string -- project name and app name to use for test
func getPrjAndAppNames() (string, string) {
	projectName := fmt.Sprintf(baseBranchProjectName, branchesTestsCounter)
	appName := fmt.Sprintf(baseBranchAppName, branchesTestsCounter)
	branchesTestsCounter++
	return projectName, appName
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// TestAppBranchCreate is the function allowing to test if
// the command "rp-cli applications branch" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppBranchCreate(test *testing.T) {
	// Retrieve the project and app names to use for this test
	localPrj, localApp := getPrjAndAppNames()

	// Create an application and a project in order to start testing the branches
	err := createPrjAndApp(test, localPrj, "redpesk-lts-arz", localApp, localApp)
	if err != nil {
		test.Fatal(err)
	}

	// Create a new branch on this app
	output, err := createBranchOnApp(test, localPrj, localApp, "main", "first-new-branch", false, true, []string{})
	if err != nil {
		test.Fatal(err)
	}

	// Check that the answer is the expected one
	assert.Contains(test, output, fmt.Sprintf(appsBranchCmdAnsLine1, "first-new-branch", localApp))

	// Check that the active branch is "first-new-branch" and default branch is "main"
	testActiveAndDefaultBranch(test, localApp, localPrj, "first-new-branch", "main")
}

// TestAppBranchCreateNoSwitch is the function allowing to test if
// the command "rp-cli applications branch --not-active" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppBranchCreateNoSwitch(test *testing.T) {
	// Retrieve the project and app names to use for this test
	localPrj, localApp := getPrjAndAppNames()

	// Create an application and a project in order to start testing the branches
	err := createPrjAndApp(test, localPrj, "redpesk-lts-arz", localApp, localApp)
	if err != nil {
		test.Fatal(err)
	}

	// Create a new branch on this app
	output, err := createBranchOnApp(test, localPrj, localApp, "main", "first-new-branch", false, false, []string{})
	if err != nil {
		test.Fatal(err)
	}

	// Check that the answer is the expected one
	assert.Contains(test, output, fmt.Sprintf(appsBranchCmdAnsLine2, "first-new-branch", localApp))

	// Check that the active branch is "main" and default branch is "main"
	testActiveAndDefaultBranch(test, localApp, localPrj, "main", "main")
}

// TestAppBranchCreateDefault is the function allowing to test if
// the command "rp-cli applications branch --default" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppBranchCreateDefault(test *testing.T) {
	// Retrieve the project and app names to use for this test
	localPrj, localApp := getPrjAndAppNames()

	// Create an application and a project in order to start testing the branches
	err := createPrjAndApp(test, localPrj, "redpesk-lts-arz", localApp, localApp)
	if err != nil {
		test.Fatal(err)
	}

	// Create a new branch on this app
	output, err := createBranchOnApp(test, localPrj, localApp, "main", "first-new-branch", true, true, []string{})
	if err != nil {
		test.Fatal(err)
	}

	// Check that the answer is the expected one
	assert.Contains(test, output, fmt.Sprintf(appsBranchCmdAnsLine1, "first-new-branch", localApp))

	// Check that the active branch is "first-new-branch" and default branch is "first-new-branch"
	testActiveAndDefaultBranch(test, localApp, localPrj, "first-new-branch", "first-new-branch")
}

// TestAppBranchCreateDistros is the function allowing to test if
// the command "rp-cli applications branch --distributions" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppBranchCreateDistros(test *testing.T) {
	// Retrieve the project and app names to use for this test
	localPrj, localApp := getPrjAndAppNames()

	// Create an application and a project in order to start testing the branches
	err := createPrjAndApp(test, localPrj, "redpesk-lts-arz", localApp, localApp)
	if err != nil {
		test.Fatal(err)
	}

	// Create a new branch on this app
	output, err := createBranchOnApp(test, localPrj, localApp, "main", "first-new-branch", true, true, []string{"redpesk-lts-arz"})
	if err != nil {
		test.Fatal(err)
	}

	// Check that the answer is the expected one
	assert.Contains(test, output, fmt.Sprintf(appsBranchCmdAnsLine1, "first-new-branch", localApp))

	// Check that the distribution appears in the display of the branches
	distrosTab, err := getDistrosForBranch(localApp, localPrj, "first-new-branch")
	if err != nil {
		test.Fatal(err)
	}
	assert.Contains(test, distrosTab, "redpesk-lts-arz")
}

// TestAppBranchSwitch is the function allowing to test if
// the command "rp-cli applications branch" is working correctly
// when only a switch is done, and not a creation
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppBranchSwitch(test *testing.T) {
	// Retrieve the project and app names to use for this test
	localPrj, localApp := getPrjAndAppNames()

	// Create an application and a project in order to start testing the branches
	err := createPrjAndApp(test, localPrj, "redpesk-lts-arz", localApp, localApp)
	if err != nil {
		test.Fatal(err)
	}

	// Create a new branch on this app
	output, err := createBranchOnApp(test, localPrj, localApp, "main", "first-new-branch", false, false, []string{})
	if err != nil {
		test.Fatal(err)
	}

	// Check that the answer is the expected one
	assert.Contains(test, output, fmt.Sprintf(appsBranchCmdAnsLine2, "first-new-branch", localApp))

	// Now, switch only on the branch previously created
	output, err = createBranchOnApp(test, localPrj, localApp, "", "first-new-branch", false, true, []string{})
	if err != nil {
		test.Fatal(err)
	}

	// Check that the answer is the expected one
	assert.Contains(test, output, fmt.Sprintf(appsBranchCmdAnsLine0, "first-new-branch", localApp))

	// Check that the active branch is "first-new-branch" and default branch is "main"
	testActiveAndDefaultBranch(test, localApp, localPrj, "first-new-branch", "main")
}

// TestAppBranchSwitchDefault is the function allowing to test if
// the command "rp-cli applications branch --default" is working correctly
// when only a switch is done, and not a creation
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppBranchSwitchDefault(test *testing.T) {
	// Retrieve the project and app names to use for this test
	localPrj, localApp := getPrjAndAppNames()

	// Create an application and a project in order to start testing the branches
	err := createPrjAndApp(test, localPrj, "redpesk-lts-arz", localApp, localApp)
	if err != nil {
		test.Fatal(err)
	}

	// Create a new branch on this app
	output, err := createBranchOnApp(test, localPrj, localApp, "main", "first-new-branch", false, false, []string{})
	if err != nil {
		test.Fatal(err)
	}

	// Check that the answer is the expected one
	assert.Contains(test, output, fmt.Sprintf(appsBranchCmdAnsLine2, "first-new-branch", localApp))

	// Now, switch only on the branch previously created
	output, err = createBranchOnApp(test, localPrj, localApp, "", "first-new-branch", true, true, []string{})
	if err != nil {
		test.Fatal(err)
	}

	// Check that the answer is the expected one
	assert.Contains(test, output, fmt.Sprintf(appsBranchCmdAnsLine0, "first-new-branch", localApp))

	// Check that the active branch is "first-new-branch" and default branch is "main"
	testActiveAndDefaultBranch(test, localApp, localPrj, "first-new-branch", "first-new-branch")
}

// TestAppBranchSwitchDistro is the function allowing to test if
// the command "rp-cli applications branch" is working correctly
// when a switch is done but with distributions modification
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppBranchSwitchDistro(test *testing.T) {
	// Retrieve the project and app names to use for this test
	localPrj, localApp := getPrjAndAppNames()

	// Create an application and a project in order to start testing the branches
	err := createPrjAndApp(test, localPrj, "redpesk-lts-arz", localApp, localApp)
	if err != nil {
		test.Fatal(err)
	}

	// Create a new branch on this app
	output, err := createBranchOnApp(test, localPrj, localApp, "main", "first-new-branch", false, false, []string{"redpesk-lts-arz"})
	if err != nil {
		test.Fatal(err)
	}

	// Check that the answer is the expected one
	assert.Contains(test, output, fmt.Sprintf(appsBranchCmdAnsLine2, "first-new-branch", localApp))

	// Now, switch only on the branch previously created
	output, err = createBranchOnApp(test, localPrj, localApp, "", "first-new-branch", false, true, []string{"redpesk-lts-arz", "redpesk-lts-brehat"})
	if err != nil {
		test.Fatal(err)
	}

	// Check that the answer is the expected one
	assert.Contains(test, output, fmt.Sprintf(appsBranchCmdAnsLine0, "first-new-branch", localApp))

	// Check that the active branch is "first-new-branch" and default branch is "main"
	testActiveAndDefaultBranch(test, localApp, localPrj, "first-new-branch", "main")

	// Check that the distributions appears in the display of the branches
	distrosTab, err := getDistrosForBranch(localApp, localPrj, "first-new-branch")
	if err != nil {
		test.Fatal(err)
	}
	assert.Contains(test, distrosTab, "redpesk-lts-arz")
	assert.Contains(test, distrosTab, "redpesk-lts-brehat")
}

// TestAppBranchesList is the function allowing to test if
// the command "rp-cli applications branches list" is working correctly
// in verbose, and not verbose
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppBranchesList(test *testing.T) {
	// Retrieve the project and app names to use for this test
	localPrj, localApp := getPrjAndAppNames()

	// Create an application and a project in order to start testing the branches
	err := createPrjAndApp(test, localPrj, "redpesk-lts-arz", localApp, localApp)
	if err != nil {
		test.Fatal(err)
	}

	// Create a new branch on this app
	_, err = createBranchOnApp(test, localPrj, localApp, "main", "first-new-branch", false, false, []string{})
	if err != nil {
		test.Fatal(err)
	}

	// Create a new branch on this app
	_, err = createBranchOnApp(test, localPrj, localApp, "main", "second-new-branch", false, false, []string{"redpesk-lts-arz"})
	if err != nil {
		test.Fatal(err)
	}

	// Create a new branch on this app
	_, err = createBranchOnApp(test, localPrj, localApp, "main", "third-new-branch", false, false, []string{"redpesk-lts-brehat"})
	if err != nil {
		test.Fatal(err)
	}

	// Test NOT VERBOSE

	// Request the list of branch for the given application
	cmdToRun := []string{
		appsCommand,
		appsBranchesCmd,
		"list",
		localApp,
		"-p",
		localPrj,
	}

	// Run the command to get the list of branches
	output, err := executeCommand(cmdToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// Check that the answer contains everything needed
	assert.Contains(test, output, "* main")
	assert.Contains(test, output, "first-new-branch")
	assert.Contains(test, output, "second-new-branch")
	assert.Contains(test, output, "third-new-branch")
	assert.NotContains(test, output, "[ACTIVE]")
	assert.NotContains(test, output, "[DEFAULT]")

	// Test VERBOSE

	// Add the verbose option
	cmdToRun = append(cmdToRun, "-v")

	// Run the command to get the list of branches
	output, err = executeCommand(cmdToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// Check that the answer contains everything needed
	assert.Contains(test, output, "* main")
	assert.Contains(test, output, "first-new-branch")
	assert.Contains(test, output, "second-new-branch")
	assert.Contains(test, output, "third-new-branch")
	assert.Contains(test, output, "[ACTIVE]")
	assert.Contains(test, output, "[DEFAULT]")

	// Check that the distributions appears in the display of the branches
	distrosTab, err := getDistrosForBranch(localApp, localPrj, "second-new-branch")
	if err != nil {
		test.Fatal(err)
	}
	assert.Contains(test, distrosTab, "redpesk-lts-arz")

	// Check that the distributions appears in the display of the branches
	distrosTab, err = getDistrosForBranch(localApp, localPrj, "third-new-branch")
	if err != nil {
		test.Fatal(err)
	}
	assert.Contains(test, distrosTab, "redpesk-lts-brehat")
}

// TestAppBranchesDelete is the function allowing to test if
// the command "rp-cli applications branches delete" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestAppBranchesDelete(test *testing.T) {
	// Retrieve the project and app names to use for this test
	localPrj, localApp := getPrjAndAppNames()

	// Create an application and a project in order to start testing the branches
	err := createPrjAndApp(test, localPrj, "redpesk-lts-arz", localApp, localApp)
	if err != nil {
		test.Fatal(err)
	}

	// Create a new branch on this app
	_, err = createBranchOnApp(test, localPrj, localApp, "main", "first-new-branch", false, false, []string{})
	if err != nil {
		test.Fatal(err)
	}

	// Create a new branch on this app
	_, err = createBranchOnApp(test, localPrj, localApp, "main", "second-new-branch", false, false, []string{})
	if err != nil {
		test.Fatal(err)
	}

	// Request to delete a branch
	cmdToRun := []string{
		appsCommand,
		appsBranchesCmd,
		"delete",
		localApp,
		"-p",
		localPrj,
		"-b",
		"first-new-branch",
	}

	// Run the command to get the list of branches
	output, err := executeCommand(cmdToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// Check that the answer is the expected one
	assert.Contains(test, output, fmt.Sprintf(appsBranchesDelCmdAnsLine0, "first-new-branch", localApp))

	// Request the list of branch for the given application
	cmdToRun2 := []string{
		appsCommand,
		appsBranchesCmd,
		"list",
		localApp,
		"-p",
		localPrj,
	}

	// Run the command to get the list of branches
	output, err = executeCommand(cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Check that the answer contains everything needed
	assert.Contains(test, output, "* main")
	assert.NotContains(test, output, "first-new-branch")
	assert.Contains(test, output, "second-new-branch")
}

// --------------------------------------------------------------
// Utilities functions ------------------------------------------
// --------------------------------------------------------------

// createPrjAndApp is an initializer function to create a project and an application
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param prjName: Name of the project to create
// :type prjName: string
// :param distro: Name of the distro to set for the project
// :type distro: string
// :param appName: Name of the application to create
// :type appName: string
// :param pkgName: Package name for the application to create
// :type pkgName: string
// :returns: error -- an error if a problem occurred
func createPrjAndApp(test *testing.T, prjName, distro, appName, pkgName string) error {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd(prjName, distro)
	err := addObjectIfNotExists(test, prjsCommand, prjName, cmdToRun1...)
	if err != nil {
		return err
	}

	// Create the add application command
	cmdToRun2 := []string{
		appsCommand,
		appsAddCmd,
		"-n",
		appName,
		"--pkg-name",
		pkgName,
		"-p",
		prjName,
		"--source-url",
		"http://git.ovh.iot/redpesk/redpesk-samples/test-application.git",
		"-d",
		"This is a testing application",
	}

	// Create the application if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, appsCommand, appName, cmdToRun2...)

	return err
}

// createBranchOnApp is a function allowing to create a new branch in an application
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param prjName: Name of the parent project (opt.)
// :type prjName: string
// :param appName: Name of the parent application (mand.)
// :type appName: string
// :param src: Name of the source branch
// :type src: string
// :param dst: Name of the destination branch
// :type dst: string
// :param def: Set the branch to create or to switch as default
// :type def: bool
// :param active: Set the branch to create or to switch as active
// :type active: bool
// :param distros: List of distributions linked with the branch
// :type distros: []string
// :returns: string, error -- the output of the command an error if a problem occurred
func createBranchOnApp(test *testing.T, prjName, appName, src, dest string, def, active bool, distros []string) (string, error) {
	// Create the basic command tab
	cmdToRun := []string{
		appsCommand,
		"branch",
		appName,
		"-d",
		dest,
	}

	// Add the project argument if it is specified
	if prjName != "" {
		cmdToRun = append(cmdToRun, "-p", prjName)
	}
	// Add the source argument if specified
	if src != "" {
		cmdToRun = append(cmdToRun, "-s", src)
	}
	if def {
		cmdToRun = append(cmdToRun, "--default")
	}
	if !active {
		cmdToRun = append(cmdToRun, "--not-active")
	}
	// Add the linked distributions if needed
	if len(distros) > 0 {
		distroListStr := strings.Join(distros, ",")
		cmdToRun = append(cmdToRun, "--distributions", distroListStr)
	}

	// Run the command and return the corresponding values
	return executeCommand(cmdToRun...)

}

// testActiveAndDefaultBranch is a function allowing to test the values
// of the active branch and the default branch
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param appName: Name of the parent application
// :type appName: string
// :param prjName: Name of the parent project
// :type prjName: string
// :param activeWanted: Name of the active branch wanted
// :type activeWanted: string
// :param defWanted: Name of the default branch wanted
// :type defWanted: string
func testActiveAndDefaultBranch(test *testing.T, appName, prjName, activeWanted, defWanted string) {
	// Active
	activeBranch, err := getActiveBranchFromLogs(appName, prjName)
	if err != nil {
		test.Fatal(err)
	}
	assert.Equal(test, activeWanted, activeBranch)

	// Default
	defBranch, err := getDefaultBranchFromLogs(appName, prjName)
	if err != nil {
		test.Fatal(err)
	}
	assert.Equal(test, defWanted, defBranch)
}

// getActiveBranch is a function allowing to get the active branch of
// a given application
//
// :param appName: Name of the parent application (mand.)
// :type appName: string
// :param prjName: Name of the parent project (opt.)
// :type prjName: string
// :returns: string, error -- the active branch and an error if a problem occurred
func getActiveBranchFromLogs(appName, prjName string) (string, error) {
	return getRequestedTypeBranch(appName, prjName, "[ACTIVE]")
}

// getDefaultBranch is a function allowing to get the default branch of
// a given application
//
// :param appName: Name of the parent application (mand.)
// :type appName: string
// :param prjName: Name of the parent project (opt.)
// :type prjName: string
// :returns: string, error -- the active branch and an error if a problem occurred
func getDefaultBranchFromLogs(appName, prjName string) (string, error) {
	return getRequestedTypeBranch(appName, prjName, "[DEFAULT]")
}

// getRequestedBranch is a function allowing to get the branch of
// a given application
//
// :param appName: Name of the parent application
// :type appName: string
// :param prjName: Name of the parent project
// :type prjName: string
// :param typeStr: Type string to look for in the command output
// :type typeStr: string
// :returns: string, error -- the active branch and an error if a problem occurred
func getRequestedTypeBranch(appName, prjName, typeStr string) (string, error) {
	// Create the value to return
	retValue := ""

	// Create the command to run
	cmdToRun := []string{
		appsCommand,
		appsBranchesCmd,
		"list",
		appName,
		"-p",
		prjName,
		"-v",
	}

	// Run the command to get the list of branches
	output, err := executeCommand(cmdToRun...)
	if err != nil {
		return retValue, err
	}

	// Now look for the typeStr
	outputLine := strings.Split(output, "\n")
	for _, line := range outputLine {
		// Get rid of the * and trim the line
		reworkedLine := strings.TrimSpace(strings.ReplaceAll((strings.ReplaceAll(line, "*", "")), "\t", " "))
		if strings.Contains(reworkedLine, typeStr) {
			// The typeStr is in the current line, get the name of the branch
			tmpTab := strings.Split(reworkedLine, " ")
			retValue = tmpTab[0]

			if retValue == reworkedLine {
				return "", errors.New("A problem occurred while extracting the branch name")
			}

			break
		}
	}

	return retValue, nil
}

// getDistrosForBranch is a function allowing to get the distros linked
// to a given branch inside an application
//
// :param appName: Name of the parent application
// :type appName: string
// :param prjName: Name of the parent project
// :type prjName: string
// :param branchName: Name of the branch
// :type branchName: string
// :returns: []string, error -- the active branch and an error if a problem occurred
func getDistrosForBranch(appName, prjName, branchName string) ([]string, error) {
	// Create the value to return
	retValue := []string{}

	// Create the command to run
	cmdToRun := []string{
		appsCommand,
		appsBranchesCmd,
		"list",
		appName,
		"-p",
		prjName,
		"-v",
	}

	// Run the command to get the list of branches
	output, err := executeCommand(cmdToRun...)
	if err != nil {
		return retValue, err
	}

	// Now look for the typeStr
	outputLine := strings.Split(output, "\n")
	for _, line := range outputLine {
		// Get rid of the * and trim the line
		reworkedLine := strings.TrimSpace(strings.ReplaceAll((strings.ReplaceAll(line, "*", "")), "\t", " "))
		if strings.HasPrefix(reworkedLine, branchName) {
			// The current line is for the branch we are looking for, now retrieve only the strings between parenthesis
			splitLine1 := strings.Split(reworkedLine, "(")
			splitLine2 := strings.Split(splitLine1[1], ")")

			if len(splitLine2) > 0 {
				retValue = strings.Split(splitLine2[0], ",")
			}
			break
		}
	}

	return retValue, nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------
