/*
Copyright © 2022-2023 IoT.bzh Company
Contact: https://www.iot.bzh/licensing

This file is part of the redpesk-cli project, the redpesk command line tool.

$RP_BEGIN_LICENSE$
Commercial License Usage
 Licensees holding valid commercial IoT.bzh licenses may use this file in
 accordance with the commercial license agreement provided with the
 Software or, alternatively, in accordance with the terms contained in
 a written agreement between you and The IoT.bzh Company. For licensing terms
 and conditions see https://www.iot.bzh/terms-conditions. For further
 information use the contact form at https://www.iot.bzh/contact.

GNU General Public License Usage
 Alternatively, this file may be used under the terms of the GNU General
 Public license version 3. This license is as published by the Free Software
 Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
 of this file. Please review the following information to ensure the GNU
 General Public License requirements will be met
 https://www.gnu.org/licenses/gpl-3.0.html.
$RP_END_LICENSE$
*/

package cmd

import (
	"encoding/json"
	"errors"
	"fmt"
	"net/url"
	"os"
	"path"
	"regexp"
	"strconv"
	"strings"
	"text/tabwriter"
	"time"

	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/printer"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/utils"
	"git.ovh.iot/redpesk/redpesk-infra/rp-webserver/actions/rpapiv1"
	"github.com/gofrs/uuid"
	"github.com/spf13/cobra"
	"github.com/spf13/viper"
)

// -------------------------------------------------
// Global variable (inside this package) -----------
// -------------------------------------------------

// Format string used to display live build status
var liveBuildFormatLine = (genericBuildLiveCmdLinePart0 + "%-12v][%-12v" + genericBuildLiveCmdLinePart1)

// -------------------------------------------------
// -------------------------------------------------

// --------------------------------------------------------------
// Completion functions -----------------------------------------
// --------------------------------------------------------------

// getBuildTypesAsTab returns the available build types available for the
// cross-compilation
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getBuildTypesAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	return []string{"emulated_build", "accel_build_cross", "standard_build_cross"}, cobra.ShellCompDirectiveNoFileComp
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Command functions --------------------------------------------
// --------------------------------------------------------------

// buildsGetAndDisplay gets all the properties of a given build (specified by the user with an ID)
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param forceParentFlag: Flag allowing the user to force the arg to be interpreted as parent ID
// :type forceParentFlag: string
// :param parentType: Type of the parent (image/app)
// :type parentType: string
// :param ObjGetUuidCb: Function allowing to get the UUID of the "parent" object (image or app) if necessary
// :type ObjGetUuidCb: func(*cobra.Command, []string) (uuid.UUID, error)
// :returns: error -- an error if something went wrong, nil in the other case
func buildsGetAndDisplay(cmd *cobra.Command, args []string, forceParentFlag, parentType string, ObjGetUuidCb func(*cobra.Command, []string) (uuid.UUID, error)) error {

	// Retrieve the build ID from the args given by the user
	buildID, err := getBuildIDFromArgs(cmd, args, forceParentFlag, parentType, ObjGetUuidCb)
	if err != nil {
		return err
	}

	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Get the properties of this build
	build := rpapiv1.RPBuildFull{}
	if err := bckAPI.BuildGet(&build, strconv.Itoa(buildID)); err != nil {
		return err
	}

	// Display the resulting tests
	err = RPPrinter.CustomPrintObject(build, verbose, displayOneBuild)

	return err
}

// appBuildsGetAndDisplay gets all the properties of a given application build
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func appBuildsGetAndDisplay(cmd *cobra.Command, args []string) error {
	return buildsGetAndDisplay(cmd, args, appsForceAppFlag, "application", getAppUUIDCallBack)
}

// imgBuildsGetAndDisplay gets all the properties of a given image build
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func imgBuildsGetAndDisplay(cmd *cobra.Command, args []string) error {
	return buildsGetAndDisplay(cmd, args, imgsForceImgFlag, "image", getImgUUIDCallBack)
}

// buildsList gets all the builds available for the given object
// and prints it.
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param objType: Type of object for which a list of builds is wanted
// :type objType: string
// :returns: error -- an error if something went wrong, nil in the other case
func buildsList(cmd *cobra.Command, args []string, objType string) error {

	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Prepare the variable that we will fill according to the case (app build or image build)
	var getBuildsCB func(pageNb, perPage int) (interface{}, printer.RPPrinterPaging, error)
	var verboseLineNb int
	if strings.Contains(objType, "app") {
		// Get the application given by the user
		app, err := genericGetAppFromArgs(cmd, args, false)
		if err != nil {
			return err
		}

		// Get the srpm flag
		srpmBld, err := cmd.Flags().GetBool(appsSubBuildsListCmdSRpmFlag)
		if err != nil {
			logFlagError(appsSubBuildsListCmdSRpmFlag)
			return err
		}

		// According to the srpm flag, change the build type wanted
		bldType := rpapiv1.RPBuildApp
		if srpmBld {
			bldType = rpapiv1.RPBuildSrpm
		}

		// Prepare the function needed by the printer to get Builds
		getBuildsCB = func(pageNb, perPage int) (interface{}, printer.RPPrinterPaging, error) {
			// Get builds list
			blds := rpapiv1.RPBuilds{}
			if err := bckAPI.ApplicationsBuildsListGetByPage(&blds, app.ID.String(), bldType, pageNb, perPage, ""); err != nil {
				return blds.Builds, printer.RPPrinterPaging{}, err
			}
			// Get the paging interface from the structure retrieved
			pagingStruct, _ := printer.GetPagingFromInterface(&blds)

			return blds.Builds, pagingStruct, nil
		}
		verboseLineNb = 8

	} else {
		// Get the image given by the user
		img, err := genericGetImgFromArgs(args, false)
		if err != nil {
			return err
		}

		// Prepare the function needed by the printer to get the builds lisrt
		getBuildsCB = func(pageNb, perPage int) (interface{}, printer.RPPrinterPaging, error) {
			// Get builds list
			blds := rpapiv1.RPBuilds{}
			if err := bckAPI.ImagesBuildsListByPage(&blds, img.ID.String(), pageNb, perPage, ""); err != nil {
				return blds.Builds, printer.RPPrinterPaging{}, err
			}
			// Get the paging interface from the structure retrieved
			pagingStruct, _ := printer.GetPagingFromInterface(&blds)

			return blds.Builds, pagingStruct, nil
		}
		verboseLineNb = 7
	}

	// Prepare the structure to manage paging
	pagingStruct := printer.RPPagingToolsStruct{
		Verbose:       verbose,
		VerboseLineNb: verboseLineNb,
		GetObjCB:      getBuildsCB,
		PrintFancyCB:  displayBuilds,
	}

	return RPPrinter.InteractivelyPrintPages(&pagingStruct, viper.GetBool(NoHeaderFlag))
}

// buildGetLogs allows to retrieve logs from a build.
// If the build ID is not specified, the logs of the latest build for the given object
// are retrieved
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param objID: ID of the object for which we want the build logs
// :type objID: string
// :param parentObjType: Type of parent object (image, app, etc.)
// :type parentObjType: string
// :returns: error -- an error if something went wrong, nil in the other case
func buildGetLogs(cmd *cobra.Command, buildID int, parentObjType string) error {

	// Convert to string
	buildIDStr := strconv.Itoa(buildID)

	// Get the build object that is interesting us
	fullBuild := rpapiv1.RPBuildFull{}
	if err := bckAPI.BuildGet(&fullBuild, buildIDStr); err != nil {
		return err
	}

	// Create the download directory specified by the user
	buildLogsDir, err := manageDownloadDir(cmd, fmt.Sprintf("rp-%s-builds-logs", parentObjType), buildIDStr)
	if err != nil {
		return err
	}

	// Get the "zip" flag from the user
	zipRequested, err := cmd.Flags().GetBool(genericBuildZipFlag)
	if err != nil {
		logFlagError(genericBuildZipFlag)
		return err
	}

	baseURL, err := url.Parse(fullBuild.Task.BaseURL)
	if err != nil {
		return err
	}

	// Loop on build subtask and create a directory containing the logs for each one
	for _, subtask := range fullBuild.Task.Subtasks {
		// Do something only if the logs slice is not empty
		if len(subtask.Results.Logs) > 0 {

			// Create the subtask directory
			subtaskDir := path.Join(buildLogsDir, subtask.Label)
			err = utils.CreateDirIfNotExists(subtaskDir, false)
			if err != nil {
				return nil
			}

			// Print a message to the user saying that we are downloading the logs
			RPPrinter.CustomPrintF(genericBuildLogsCmdAnsLine1, subtask.Label)

			// Download, in this directory all the logs available
			for _, logURL := range subtask.Results.Logs {
				// Create the full URL
				fullURL := strings.Join([]string{baseURL.String(), logURL}, "/")

				// Download the file
				filename := path.Base(fullURL)
				fullFilename := path.Join(subtaskDir, filename)
				_, err = bckAPI.HTTPCli.DownloadFile(fullURL, fullFilename, true, "")
				if err != nil {
					RPPrinter.CustomColorPrintF(printer.ColorRed, genericBuildLogsCmdAnsLine3, subtask.Label, err)
				}
			}

		} else {
			// Display a message to the user
			RPPrinter.CustomColorPrintF(printer.ColorYellow, genericBuildLogsCmdAnsLine2, subtask.Label)
		}
	}

	// Initialize the name we will output to the user
	dirOrFileInOutput := buildLogsDir

	if zipRequested {
		dirOrFileInOutput, err = zipLogs(buildLogsDir)
		if err != nil {
			return nil
		}
	}

	// Display a message to the user
	RPPrinter.CustomColorPrintF(printer.ColorCyan, genericBuildLogsCmdAnsLine4, buildID, dirOrFileInOutput)

	return nil
}

// buildCancel allows to cancel a build
//
// :param buildID: ID of the build to cancel
// :type buildID: string
// :returns: error -- an error if something went wrong, nil in the other case
func buildCancel(buildID int) error {
	// Print things to the user
	RPPrinter.CustomPrintLn(genericBuildsCancelCmdAnsLine0)
	RPPrinter.CustomPrintF(genericBuildsCancelCmdAnsLine1, buildID)

	// Request the backend to cancel the build
	buildIDStr := strconv.Itoa(buildID)
	if err := bckAPI.BuildCancel(buildIDStr, nil); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	return nil
}

// getAndDisplayBuildsStats gets and display builds statistics from an
// application or a project
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param projectID: UUID of the concerned project, as a string
// :type projectID: uuid.UUID
// :param app: pointer to the concerned application structure
// :type app: *rpapiv1.RPApplicationFull
// :returns: error -- an error if something went wrong, nil in the other case
func getAndDisplayBuildsStats(cmd *cobra.Command, projectID uuid.UUID, app *rpapiv1.RPApplicationFull) error {
	var err error

	// Get the value of "last" flag
	// Default value at 10 for project to take care of the backend
	lastNb := 10
	if app != nil {
		lastNb, err = cmd.Flags().GetInt(appsBuildsStatsCmdLastFlag)
		if err != nil {
			logFlagError(appsBuildsStatsCmdLastFlag)
			return err
		}
	}

	// Get the value of the "verbose" flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Now request the backend for the projects stats
	projectBuildsStats := rpapiv1.RPBuildStatsProject{}
	if err := bckAPI.ProjectGetBuildsStats(projectID.String(), lastNb, &projectBuildsStats); err != nil {
		return err
	}

	// Now display either the full project builds stats or uniquely the app, according
	// to what the caller gave us
	if app == nil {
		return RPPrinter.CustomPrintObject(projectBuildsStats, verbose, displayPrjBuildsStats)
	} else {
		for _, appStats := range projectBuildsStats.AppStats {
			if utils.UUIDEqual(appStats.AppID, app.ID) {
				return RPPrinter.CustomPrintObject(appStats, verbose, displayAppBuildsStats)
			}
		}
	}

	// If we reach here, the application has not been found
	// It means that no build exists for it, therefore, display an empty structure
	statsToDisplay := rpapiv1.RPBuildStatsApp{
		AppID:       app.ID,
		AppSlug:     app.Slug,
		PackageName: app.PackageName,
	}
	return RPPrinter.CustomPrintObject(statsToDisplay, verbose, displayAppBuildsStats)
}

// getAndDisplayBuildsMatrix is a function allowing to retrieve all the latest builds
// for a project and to display them as a matrix
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param project: Pointer to the project wanted
// :type project: *rpapiv1.RPProjectFull
// :returns: error -- an error if something went wrong, nil in the other case
func getAndDisplayBuildsMatrix(cmd *cobra.Command, project *rpapiv1.RPProjectFull) error {

	// Get the value of the "fail-only" flag
	failOnly, err := cmd.Flags().GetBool(prjsSubBuildsMatrixFailOnlyFlag)
	if err != nil {
		logFlagError(prjsSubBuildsMatrixFailOnlyFlag)
		return err
	}

	// Request the backend for the project builds stats
	projectBuildsStats := rpapiv1.RPBuildStatsProject{}
	if err := bckAPI.ProjectGetBuildsStats(project.ID.String(), 0, &projectBuildsStats); err != nil {
		return err
	}

	// Get the list of all the "set" distributions
	distroList := []string{}
	for _, tmpDistro := range project.Distributions {
		if tmpDistro.Set {
			distroList = append(distroList, tmpDistro.Slug)
		}
	}

	// Now fill the structure we will use to display the matrix
	matrixStruct := make(map[string]map[string]string)
	for _, appStats := range projectBuildsStats.AppStats {
		matrixStruct[appStats.PackageName] = make(map[string]string)
		// Now add an entry in the sub-map for each distro
		for _, distro := range distroList {
			// Get the status for the given distro
			matrixStruct[appStats.PackageName][distro] = getLatestBuildStatus(appStats.LastBuilds, distro)
		}
	}

	// Now print the matrix
	// Create the header
	matrixHeaderEnd := strings.Join(distroList, "\t ")
	fmt.Fprintf(RPPrinter.Writer, prjsSubBuildsMatrixCmdAnsHeader, matrixHeaderEnd)
	for appPkg, lineMap := range matrixStruct {
		// Create the line
		currentLine := appPkg
		for _, distro := range distroList {
			currentLine = currentLine + "\t " + lineMap[distro]
		}
		// Print the line (if the fail-only flag is set, print it uniquely if it contains "failed")
		if !failOnly || (failOnly && strings.Contains(currentLine, string(rpapiv1.RPBuildTaskStatusFailed))) {
			fmt.Fprintln(RPPrinter.Writer, currentLine)
		}
	}

	return nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Utilities functions ------------------------------------------
// --------------------------------------------------------------

// getBuildIDFromArgs allows to retrieve the build ID wanted from what the
// user gave to rp-cli. The user could either give an object slug/UUID (image or app) or
// directly a build ID.
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param forceParentFlag: Flag allowing the user to force the arg to be interpreted as parent ID
// :type forceParentFlag: string
// :param parentType: Type of the parent (image/app)
// :type parentType: string
// :param ObjGetUuidCb: Function allowing to get the UUID of the "parent" object (image or app) if necessary
// :type ObjGetUuidCb: func(*cobra.Command, []string) (uuid.UUID, error)
// :returns: int, error -- the build ID and an error if something went wrong, nil in the other case
func getBuildIDFromArgs(cmd *cobra.Command, args []string, forceParentFlag, parentType string, ObjGetUuidCb func(*cobra.Command, []string) (uuid.UUID, error)) (int, error) {

	// Try to get the "force-parent" flag
	forceParent, err := cmd.Flags().GetBool(forceParentFlag)
	if err != nil {
		logFlagError(forceParentFlag)
		return -1, err
	}

	// Get the first argument of the command line
	argID, err := getFirstArg(args, fmt.Sprintf(objectIDNotSetMsg, "app/build/image"), fmt.Sprintf(objectIDRequiredMsg, "app/build/image"))
	if err != nil {
		return -1, err
	}

	// Try to convert the argument given to an int
	buildIDInt, err := strconv.Atoi(argID)
	isInteger := (err == nil)

	// Act according to what we have

	// If the arg given is an int and that the user do not force the arg to be a parent object
	// return this integer directly
	if isInteger && !forceParent {
		return buildIDInt, nil
	}

	// In the other case, the user argument is the application slug, use it to retrieve the latest build

	// Retrieve the UUID of the parent
	parentUUID, err := ObjGetUuidCb(cmd, args)
	if err != nil {
		return -1, err
	}
	tmpBuildID, err := getLatestBuild(parentUUID.String(), parentType)
	if err != nil {
		return -1, err
	}

	return tmpBuildID, nil
}

// getLatestBuild is a private function allowing to get
// the latest build of an object. This object can be either an application
// or an image.
//
// :param objID: The object ID for which the builds are wanted
// :type objID: string
// :param objType: The object type for which the builds are wanted (app or image)
// :type objType: string
// :returns: int, error -- the last build ID and an error if something went wrong, nil in the other case
func getLatestBuild(objID, objType string) (int, error) {
	// First get all the builds from the object
	builds := rpapiv1.RPBuilds{}
	var err error

	// Use the right API call according to what is needed
	if strings.Contains(objType, "app") {
		err = bckAPI.ApplicationsBuildsListGetByPage(&builds, objID, rpapiv1.RPBuildApp, 1, 1, "desc")
	} else {
		err = bckAPI.ImagesBuildsListByPage(&builds, objID, 1, 1, "desc")
	}

	if err != nil {
		return -1, err
	}

	if len(builds.Builds) == 0 {
		return -1, fmt.Errorf(genericNoBuildFoundErr, objID)
	}

	return strconv.Atoi(builds.Builds[0].ID)
}

// getLatestBuildStatus is a private function allowing to get
// the latest build status for a given distribution.
// If no status is found for the given distribution, "N/A" is returned.
//
// :param lastBuilds: List of last builds for the concerned application
// :type lastBuilds: []rpapiv1.RPBuild
// :param distroSlug: Slug of the distribution we want
// :type distroSlug: string
// :returns: int, error -- the last build ID and an error if something went wrong, nil in the other case
func getLatestBuildStatus(lastBuilds []rpapiv1.RPBuild, distroSlug string) string {
	result := "N/A"

	// Look for the right build in the latest builds
	for _, build := range lastBuilds {
		if build.DistributionSlug == distroSlug {
			result = string(build.Status)
			break
		}
	}
	return result
}

// zipLogs is a private function allowing to zip any kind of logs inside a ZIP file
//
// :param logsDir: Directory containing the logs
// :type logsDir: string
// :returns: string, error -- full path to the zip archive and an error if something went wrong, nil in the other case
func zipLogs(logsDir string) (string, error) {
	// Rework zip path
	// Get the build ID
	build := path.Base(logsDir)
	// Get the directory name
	parentDir := path.Dir(logsDir)
	// Get directory name
	zipName := parentDir + "-" + build + ".zip"

	// Create zip file
	err := utils.CreateZipFromDir(logsDir, zipName)
	if err != nil {
		return "", err
	}

	// Delete corresponding directory
	err = os.RemoveAll(parentDir)
	if err != nil {
		return "", err
	}

	return zipName, nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Display functions --------------------------------------------
// --------------------------------------------------------------

// displayBuilds is a private function allowing to display
// a slice of builds in the console
//
// :param builds: tab of builds to display
// :type builds: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func displayBuilds(builds interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Convert to the right type
	blds := builds.([]rpapiv1.RPBuild)

	// Display result
	first := true
	for _, build := range blds {
		if !verbose && first {
			manageHeaderPrint(genericBuildListHeader, writer)
		} else if verbose && !first {
			fmt.Fprintln(writer, genericBuildCmdAnsLine4)
		}

		// Print this build
		printOneBuild(build, verbose, false, writer)

		first = false
	}

	return nil
}

// displayOneBuild is a private function allowing to display
// one build in the console
//
// :param build: Build to display
// :type build: interface{}
// :param verbose: Verbose flag passed in the command line
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func displayOneBuild(build interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Print the build
	printOneBuild(build, verbose, true, writer)
	return nil
}

// displayPrjBuildsStats is a private function allowing to display
// the builds statistics for a full project
//
// :param buildsStatsInterface: Stats to display
// :type buildsStatsInterface: interface{}
// :param verbose: Verbose flag passed in the command line
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func displayPrjBuildsStats(buildsStatsInterface interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Convert to the right type
	buildsStats := buildsStatsInterface.(rpapiv1.RPBuildStatsProject)
	// Print the builds stats
	return printPrjBuildsStats(buildsStats, verbose, writer)
}

// displayAppBuildsStats is a private function allowing to display
// the builds statistics for an app
//
// :param buildsStatsInterface: Stats to display
// :type buildsStatsInterface: interface{}
// :param verbose: Verbose flag passed in the command line
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func displayAppBuildsStats(buildsStatsInterface interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Convert to the right type
	buildsStats := buildsStatsInterface.(rpapiv1.RPBuildStatsApp)
	// Print the builds stats
	return printAppBuildsStats(buildsStats, verbose, writer)
}

// printOneBuild is a private function allowing to print one build in the console
//
// :param build: pointer to the build to print
// :type build: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param isFull: True if the user wants to display a full build
// :type isFull: bool
// :param writer: Writer to used to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func printOneBuild(build interface{}, verbose, isFull bool, writer *tabwriter.Writer) error {
	bld := rpapiv1.RPBuildFull{}

	// Cast to the full type in all case but use "full" parameters as indicated by the user
	// We have to pass through the Marshal/Unmarshal tricks to not have problem during conversion from
	// RPBuild -> RPBuildFull at runtime
	dataBytes, _ := json.Marshal(build)
	json.Unmarshal(dataBytes, &bld)

	if verbose {

		// Create the string containing the results
		fmt.Fprintln(writer, genericBuildCmdVerboseLine0, bld.ID)

		bType := "unknown"
		switch bld.BuildType {
		case rpapiv1.RPBuildSrpm:
			bType = "application SRPM build"
		case rpapiv1.RPBuildDeps:
			bType = "application dependencies check"
		case rpapiv1.RPBuildApp:
			bType = "application build/cross-build"
		case rpapiv1.RPBuildImage:
			bType = "image build"
		}
		fmt.Fprintln(writer, genericBuildCmdVerboseLine13, bType)

		if bld.Parent.ApplicationID != uuid.Nil {
			fmt.Fprintln(writer, appsBuildCmdVerboseLine1, bld.Parent.ApplicationID.String())
			fmt.Fprintln(writer, appsBuildCmdVerboseLine8, bld.Parent.ApplicationSlug)
		} else {
			fmt.Fprintln(writer, imgsBuildCmdVerboseLine1, bld.Parent.ImageID.String())
			fmt.Fprintln(writer, imgsBuildCmdVerboseLine2, bld.Parent.ImageSlug)
		}

		fmt.Fprintln(writer, genericBuildCmdVerboseLine7, bld.DistributionSlug)

		fmt.Fprintln(writer, genericBuildCmdVerboseLine14, bld.UserLauncherPseudo)

		fmt.Fprintln(writer, genericBuildCmdVerboseLine2, bld.Status)

		// Print comment for the build
		commentToPrint := mainNotSetMsg
		if bld.Comment != "" {
			commentToPrint = bld.Comment
		}
		fmt.Fprintln(writer, genericBuildCmdVerboseLine9, commentToPrint)

		if bld.BuildType == rpapiv1.RPBuildApp {
			fmt.Fprintln(writer, genericBuildCmdVerboseLine12, bld.Parent.BuildSrpmID)
		}

		if isFull {

			for _, subtask := range bld.Task.Subtasks {
				fmt.Fprintln(writer)

				// Subtask description
				fmt.Fprintln(writer, genericBuildCmdVerboseLine3, subtask.Label)
				fmt.Fprintln(writer, genericBuildCmdVerboseLine4, subtask.State)
				fmt.Fprintln(writer, genericBuildCmdVerboseLine5, subtask.Method)
				fmt.Fprintln(writer, genericBuildCmdVerboseLine6, subtask.MethodName)

				// Print logs available
				if len(subtask.Results.Logs) > 0 {
					fmt.Fprintln(writer, genericBuildCmdVerboseLogFileTag)
				}
				for _, logURL := range subtask.Results.Logs {
					fullLogURL := strings.TrimRight(bld.Task.BaseURL, "/") + "/" + logURL
					fmt.Fprintln(writer, " ", fullLogURL)
				}

				// Print srpms available
				if len(subtask.Results.SRPMS) > 0 {
					fmt.Fprintln(writer, genericBuildCmdVerboseSRPMFileTag)
				}
				for _, srpmURL := range subtask.Results.SRPMS {
					fullRPMURL := strings.TrimRight(bld.Task.BaseURL, "/") + "/" + srpmURL
					fmt.Fprintln(writer, " ", fullRPMURL)
				}

				// Print rpms available
				if len(subtask.Results.RPMS) > 0 {
					fmt.Fprintln(writer, genericBuildCmdVerboseRPMFileTag)
				}
				for _, rpmURL := range subtask.Results.RPMS {
					fullRPMURL := strings.TrimRight(bld.Task.BaseURL, "/") + "/" + rpmURL
					fmt.Fprintln(writer, " ", fullRPMURL)
				}

				// Print files available
				if len(subtask.Results.Files) > 0 {
					fmt.Fprintln(writer, genericBuildCmdVerboseFilesTag)
				}
				for _, fileURL := range subtask.Results.Files {
					fullFileURL := strings.TrimRight(bld.Task.BaseURL, "/") + "/" + fileURL
					fmt.Fprintln(writer, " ", fullFileURL)
				}
			}

			// Show output of error fault code & string
			if bld.Task.Error.FaultString != "" {
				fmt.Fprintln(writer)
				fmt.Fprintln(writer, genericBuildCmdVerboseErrorFaultCode, bld.Task.Error.FaultCode)
				fmt.Fprintln(writer, genericBuildCmdVerboseErrorFaultMsg, bld.Task.Error.FaultString)
			}
		}

	} else if isFull {
		// Create the string containing the results
		fmt.Fprintln(writer, genericBuildCmdLine0, bld.ID)
		fmt.Fprintln(writer, genericBuildCmdLine1, bld.Status)

		fmt.Fprintln(writer)
		for _, subtask := range bld.Task.Subtasks {
			fmt.Fprintf(writer, genericBuildCmdLine2, subtask.Label, subtask.State)
		}

		// Show output of error fault code & string
		if bld.Task.Error.FaultString != "" {
			fmt.Fprintln(writer)
			fmt.Fprintln(writer, genericBuildCmdVerboseErrorFaultCode, bld.Task.Error.FaultCode)
			fmt.Fprintln(writer, genericBuildCmdVerboseErrorFaultMsg, bld.Task.Error.FaultString)
		}
	} else {
		// List mode, simply print 2 columns with buildID and the status
		fmt.Fprintln(writer, bld.ID, "\t", bld.Status)
	}
	return nil
}

// printOneBuildLive is a private function allowing to print one build in the console
// in live while the build is currently running
//
// :param build: pointer to the build to print
// :type build: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param writer: Writer to used to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func printOneBuildLive(build interface{}, verbose bool, writer *tabwriter.Writer) error {

	// Get the line to print
	lineToPrint := getLiveBuildLine(build, genericBuildLiveLabelStatusChange)

	// Print the line
	RPPrinter.CustomColorPrintLn(printer.ColorWhite, lineToPrint)

	return nil
}

// getLiveBuildLine is a private function allowing to obtain a line describing the build
// given as argument
//
// :param build: pointer to the build to print
// :type build: interface{}
// :param startLabel: Label between square bracket at the beginning of line
// :type startLabel: string
// :returns: error -- an error if something went bad
func getLiveBuildLine(build interface{}, startLabel string) string {
	// Convert to the right type
	bld := build.(rpapiv1.RPBuildFull)

	// Get the object slug
	objSlug := ""
	switch bld.ParentType {
	case rpapiv1.RPBuildAppParentType:
		objSlug = bld.Parent.ApplicationSlug
	case rpapiv1.RPBuildImgParentType:
		objSlug = bld.Parent.ImageSlug
	case rpapiv1.RPBuildPrjParentType:
		objSlug = bld.Parent.ProjectSlug
	}

	strToReturn := fmt.Sprintf(liveBuildFormatLine, startLabel, bld.ID, objSlug, bld.DistributionSlug, bld.Status)

	if len(bld.Task.Subtasks) > 0 {
		subTaskTab := []string{}
		for _, subtask := range bld.Task.Subtasks {
			subTaskTab = append(subTaskTab, fmt.Sprintf("%s [%-8v]", subtask.Label, subtask.State))
		}
		strToReturn += (genericBuildLiveCmdLineComplement + strings.Join(subTaskTab, ", "))
	}

	return strToReturn
}

// printLiveBuildResults is a private function allowing to print
// all the result lines at the end of a live build. It returns an error
// if the build has failed or has been canceled.
//
// :param resultsTab: tab containing the results
// :type resultsTab: []string
// :param errorMsg: Error message to display if the build is considered has failed
// :type errorMsg: string
// :returns: error -- an error if something went bad
func printLiveBuildResults(resultsTab []string, errorMsg string) error {
	cancelOrFailRes := false

	resultsToPrint := utils.RemoveDuplicateInStrSlice(resultsTab)

	// Print the results
	for _, line := range resultsToPrint {
		switch {
		case strings.HasPrefix(line, genericBuildLiveLabelSuccess):
			// Green print
			RPPrinter.CustomColorPrintLn(printer.ColorGreen, line)

		case strings.HasPrefix(line, genericBuildLiveLabelCanceled):
			cancelOrFailRes = true
			// Yellow print
			RPPrinter.CustomColorPrintLn(printer.ColorYellow, line)

		case strings.HasPrefix(line, genericBuildLiveLabelFailed):
			cancelOrFailRes = true
			// Red print
			RPPrinter.CustomColorPrintLn(printer.ColorRed, line)

		default:
			// White print
			RPPrinter.CustomColorPrintLn(printer.ColorWhite, line)
		}
	}

	if cancelOrFailRes {
		RPPrinter.ForceFlush()
		time.Sleep(50 * time.Millisecond)
		return errors.New(errorMsg)
	}

	return nil
}

// printPrjBuildsStats is a private function allowing to print the statistics of
// builds inside a project
//
// :param buildsStats: structure containing the stats to print
// :type buildsStats: rpapiv1.RPBuildStatsProject
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param writer: Writer to used to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func printPrjBuildsStats(buildsStats rpapiv1.RPBuildStatsProject, verbose bool, writer *tabwriter.Writer) error {
	// Firstly print basics information
	fmt.Fprintln(writer, prjsSubBuildsStatsCmdAnsLine0, buildsStats.ProjectID.String())
	fmt.Fprintln(writer, prjsSubBuildsStatsCmdAnsLine1, buildsStats.NbDone)
	fmt.Fprintln(writer, prjsSubBuildsStatsCmdAnsLine2, buildsStats.NbFailed)
	fmt.Fprintln(writer, prjsSubBuildsStatsCmdAnsLine3, buildsStats.NbCanceled)
	fmt.Fprintln(writer, prjsSubBuildsStatsCmdAnsLine4, buildsStats.NbTotal)

	// If verbose is requested, give stats for every applications
	if verbose {
		fmt.Fprintln(writer)
		fmt.Fprintln(writer, prjsSubBuildsStatsCmdAnsLine5)
		fmt.Fprintln(writer)

		// Display a header
		fmt.Fprintln(writer, prjsSubBuildsStatsCmdAnsAppsHeader)
		for _, appStats := range buildsStats.AppStats {
			fmt.Fprintln(writer, appStats.AppID.String(), "\t", appStats.AppSlug, "\t", appStats.PackageName, "\t", appStats.NbDone, "\t", appStats.NbFailed, "\t", appStats.NbCanceled, "\t", appStats.NbTotal)
		}
	}

	return nil
}

// printAppBuildsStats is a private function allowing to print the builds statistics
// of an application
//
// :param buildsStats: structure containing the stats to print
// :type buildsStats: rpapiv1.RPBuildStatsProject
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param writer: Writer to used to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func printAppBuildsStats(buildsStats rpapiv1.RPBuildStatsApp, verbose bool, writer *tabwriter.Writer) error {
	// Print the information wanted
	fmt.Fprintln(writer, appsBuildsStatsCmdAnsLine0, buildsStats.AppID.String())
	fmt.Fprintln(writer, appsBuildsStatsCmdAnsLine1, buildsStats.AppSlug)
	fmt.Fprintln(writer, appsBuildsStatsCmdAnsLine2, buildsStats.PackageName)
	fmt.Fprintln(writer, appsBuildsStatsCmdAnsLine3, buildsStats.NbDone)
	fmt.Fprintln(writer, appsBuildsStatsCmdAnsLine4, buildsStats.NbFailed)
	fmt.Fprintln(writer, appsBuildsStatsCmdAnsLine5, buildsStats.NbCanceled)
	fmt.Fprintln(writer, appsBuildsStatsCmdAnsLine6, buildsStats.NbTotal)

	// If verbose is requested, give stats for every build
	if verbose {
		fmt.Fprintln(writer)
		fmt.Fprintln(writer, appsBuildsStatsCmdAnsLine7)
		fmt.Fprintln(writer)

		// Display a header
		fmt.Fprintln(writer, appsBuildsStatsCmdAnsBuildsHeader)
		for _, build := range buildsStats.LastBuilds {
			fmt.Fprintln(writer, build.ID, "\t", build.DistributionSlug, "\t", build.Status)
		}
	}

	return nil
}

// monitorBuilds monitors several builds at once after that these builds
// have been requested to the backend (through the project build)
//
// :param buildsToMonitor: Pointer to object containing the builds to monitor
// :type buildsToMonitor: *rpapiv1.RPBuilds
// :param objUUID: UUID of the source object of the build (app, image  or project)
// :type objUUID: uuid.UUID
// :param buildType: type of builds to monitor (app or image)
// :type buildType: rpapiv1.RPBuildType
// :returns: error -- an error if something went bad
func monitorBuilds(buildsToMonitor *rpapiv1.RPBuilds, objUUID uuid.UUID, buildType rpapiv1.RPBuildType) error {
	// Display something to the user if in raw mode
	RPPrinter.CustomPrintObject(*buildsToMonitor, false, displayNothing)

	// Check from viper the non-blocking flag to see if we wait
	nonBlocking := viper.GetBool(nonBlockingFlag)
	if nonBlocking {
		// The user do not want blocking functions
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
		return nil
	}

	// Determine if need to monitor app or image build
	appBuild := buildType == rpapiv1.RPBuildApp

	// Create a map for all the builds we are monitoring
	buildsMap := make(map[string]string)
	longestDistroSlugValue := 0
	longestObjectSlugValue := 0
	for _, bld := range buildsToMonitor.Builds {
		buildsMap[bld.ID] = "building"

		// Get the longest distribution slug
		if len(bld.DistributionSlug) > longestDistroSlugValue {
			longestDistroSlugValue = len(bld.DistributionSlug)
		}

		if appBuild {
			// Get the longest application slug
			if len(bld.Parent.ApplicationSlug) > longestObjectSlugValue {
				longestObjectSlugValue = len(bld.Parent.ApplicationSlug)
			}
		} else {
			// Get the longest image slug
			if len(bld.Parent.ImageSlug) > longestObjectSlugValue {
				longestObjectSlugValue = len(bld.Parent.ImageSlug)
			}
		}
	}

	// Create the format string with the right length for distribution
	lengthDistroStr := strconv.Itoa(longestDistroSlugValue)
	lengthObjStr := strconv.Itoa(longestObjectSlugValue)
	// Change the format line used in builds.go in getLiveBuildLine()
	liveBuildFormatLine = genericBuildLiveCmdLinePart0 + "%-" + lengthObjStr + "v]" + "[%-" + lengthDistroStr + "v" + genericBuildLiveCmdLinePart1

	// Create a tab that will contains the results we want to display at the end
	resultsTab := []string{}

	// Block until we receive a corresponding websocket event
	for {
		rEvent := <-cmdEventChannel

		if appBuild {
			Log.Tracef("buildApplication: receive event=%+v", rEvent)
		} else {
			Log.Tracef("buildImage: receive event=%+v", rEvent)
		}

		// Check that the received event is for build
		if rEvent.Type == rpapiv1.EvtAppBuildInfo || rEvent.Type == rpapiv1.EvtAppBuildSubTaskInfo || rEvent.Type == rpapiv1.EvtImageBuildInfo {
			// Function handling the events data
			rBuild, err := rEvent.DecodeBuild()
			if err != nil {
				return err
			}

			// Check if we are monitoring this build ID
			_, ok := buildsMap[rBuild.ID]
			if ok {
				// Print the build in the output
				RPPrinter.CustomPrintObject(rBuild, true, printOneBuildLive)

				// Check if the build is done
				if rBuild.Status == rpapiv1.RPBuildTaskStatusDone {

					// This build is finished and does not need to be monitored anymore
					buildsMap[rBuild.ID] = "done"

					// Add a result line for this build
					resultsTab = append(resultsTab, getLiveBuildLine(rBuild, genericBuildLiveLabelSuccess))

				} else if rBuild.Status == rpapiv1.RPBuildTaskStatusCanceled {
					buildsMap[rBuild.ID] = "canceled"

					// Add a result line for this build
					resultsTab = append(resultsTab, getLiveBuildLine(rBuild, genericBuildLiveLabelCanceled))

					// Display the comment if there is one
					if rBuild.Comment != "" {
						resultsTab = append(resultsTab, fmt.Sprintf("%s Cancellation comment: '%s'", genericBuildLiveLabelCanceled, rBuild.Comment))

						// If it fit the regexp, tell the user to do a 'build get' to have more details about the error/cancellation
						regularExp := regexp.MustCompile(`build id ([0-9]+) status is failed`)
						matchGroups := regularExp.FindStringSubmatch(rBuild.Comment)
						if len(matchGroups) >= 2 {
							if appBuild {
								resultsTab = append(resultsTab, fmt.Sprintf("%s Please run 'rp-cli applications builds get %s' for more info about cancellation", genericBuildLiveLabelCanceled, matchGroups[1]))
							} else {
								resultsTab = append(resultsTab, fmt.Sprintf("%s Please run 'rp-cli images builds get %s' for more info about cancellation", genericBuildLiveLabelCanceled, matchGroups[1]))
							}
						}
					}
				}

				// Check if the build is failed
				if rBuild.Status == "failed" {
					// This build is finished and does not need to be monitored anymore
					buildsMap[rBuild.ID] = "failed"

					// Add a result line for this build
					resultsTab = append(resultsTab, getLiveBuildLine(rBuild, genericBuildLiveLabelFailed))

					// Save output of error fault code and string if there are some
					if rBuild.Task.Error.FaultString != "" {
						resultsTab = append(resultsTab, fmt.Sprintf("%s %s%d", genericBuildLiveLabelFailed, genericBuildCmdVerboseErrorFaultCode, rBuild.Task.Error.FaultCode))
						resultsTab = append(resultsTab, fmt.Sprintf("%s %s'%s'", genericBuildLiveLabelFailed, genericBuildCmdVerboseErrorFaultMsg, rBuild.Task.Error.FaultString))
					}
				}

				// Check if there are still builds to monitor
				if !utils.MapContainsString(buildsMap, "building") {
					if appBuild {
						return printLiveBuildResults(resultsTab, appBuildErrorMsg)
					}
					return printLiveBuildResults(resultsTab, imgBuildFailedMsg)
				}
			}
		}

		// Check if we get an error event during the operation
		if (rEvent.Type == rpapiv1.EVTError) || (rEvent.Type == rpapiv1.EVTSessionExpired) {
			// Function handling the events
			err := handleErrorEvent(objUUID, rEvent, false)
			if err != nil {
				return err
			}
		}
	}
}

// --------------------------------------------------------------
// --------------------------------------------------------------
