/*
Copyright © 2022-2023 IoT.bzh Company
Contact: https://www.iot.bzh/licensing

This file is part of the redpesk-cli project, the redpesk command line tool.

$RP_BEGIN_LICENSE$
Commercial License Usage
 Licensees holding valid commercial IoT.bzh licenses may use this file in
 accordance with the commercial license agreement provided with the
 Software or, alternatively, in accordance with the terms contained in
 a written agreement between you and The IoT.bzh Company. For licensing terms
 and conditions see https://www.iot.bzh/terms-conditions. For further
 information use the contact form at https://www.iot.bzh/contact.

GNU General Public License Usage
 Alternatively, this file may be used under the terms of the GNU General
 Public license version 3. This license is as published by the Free Software
 Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
 of this file. Please review the following information to ensure the GNU
 General Public License requirements will be met
 https://www.gnu.org/licenses/gpl-3.0.html.
$RP_END_LICENSE$
*/

package cmd

import (
	"errors"
	"fmt"
	"path"
	"strconv"
	"strings"
	"text/tabwriter"

	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/printer"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/utils"
	"git.ovh.iot/redpesk/redpesk-infra/rp-webserver/actions/rpapiv1"
	"github.com/gofrs/uuid"
	"github.com/spf13/cobra"
	"github.com/spf13/viper"
	"golang.org/x/text/cases"
	"golang.org/x/text/language"
)

// --------------------------------------------------------------
// Initialization functions -------------------------------------
// --------------------------------------------------------------

// initEventTests inits everything that is needed to launch the
// test of an object
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func initEventTests(cmd *cobra.Command, args []string) error {

	// Init the events list
	eventsList := []string{
		rpapiv1.EvtTestAdd,
		rpapiv1.EvtTestStateChange,
		rpapiv1.EvtTestProgress,
		rpapiv1.EvtTestConsole,
	}

	return CLIRegisterToMultipleEvents(eventsList)
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Command functions --------------------------------------------
// --------------------------------------------------------------

// testsListGeneric gets all the tests available for the given object (linked to the sub-command)
// and prints it.
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func testsListGeneric(cmd *cobra.Command, args []string) error {

	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Get the great-parent command to see if we start a build on an app or an image
	greatParentCmd := cmd.Parent().Parent()

	// Prepare the variable that we will fill according to the case (app test or image test)
	var getTestsCB func(pageNb, perPage int) (interface{}, printer.RPPrinterPaging, error)
	if greatParentCmd.Use == appsCommand {
		// APP CASE

		// Get the application given by the user
		app, err := genericGetAppFromArgs(cmd, args, false)
		if err != nil {
			return err
		}

		// Prepare the function needed by the printer to get Builds
		getTestsCB = func(pageNb, perPage int) (interface{}, printer.RPPrinterPaging, error) {
			// Get tests list
			testsList, err := bckAPI.ApplicationsTestsListGetByPage(app.Parent.ProjectID.String(), app.ID.String(), pageNb, perPage, (runMode == "testing"))
			if err != nil {
				return testsList.Tests, printer.RPPrinterPaging{}, err
			}

			// Get the paging interface from the structure retrieved
			pagingStruct, _ := printer.GetPagingFromInterface(&testsList)

			return testsList.Tests, pagingStruct, nil
		}

	} else {
		// IMAGE CASE

		// Get the image given by the user
		img, err := genericGetImgFromArgs(args, false)
		if err != nil {
			return err
		}

		// Prepare the function needed by the printer to get Builds
		getTestsCB = func(pageNb, perPage int) (interface{}, printer.RPPrinterPaging, error) {
			// Get tests list
			testsList, err := bckAPI.ImagesTestsListGetByPage(img.ParentProjectID.String(), img.ID.String(), pageNb, perPage, (runMode == "testing"))
			if err != nil {
				return testsList.Tests, printer.RPPrinterPaging{}, err
			}

			// Get the paging interface from the structure retrieved
			pagingStruct, _ := printer.GetPagingFromInterface(&testsList)

			return testsList.Tests, pagingStruct, nil
		}
	}

	// Prepare the structure to manage paging
	pagingStruct := printer.RPPagingToolsStruct{
		Verbose:       verbose,
		VerboseLineNb: appTestsDisplayVerboseLineNb,
		GetObjCB:      getTestsCB,
		PrintFancyCB:  displayTests,
	}

	return RPPrinter.InteractivelyPrintPages(&pagingStruct, viper.GetBool(NoHeaderFlag))
}

// testsGetGeneric gets all the properties of a given tests (specified by the user with an ID)
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func testsGetGeneric(cmd *cobra.Command, args []string) error {

	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Get the right test ID as a string
	testIDStr, err := getTestIDForAllObjects(cmd, args)
	if err != nil {
		return err
	}

	// Get the properties of this application
	test := rpapiv1.RPTest{}
	if err := bckAPI.GenericTestGet(&test, testIDStr); err != nil {
		return err
	}

	// Display the resulting tests
	err = RPPrinter.CustomPrintObject(test, verbose, displayOneTest)

	return err
}

// testGetLogsGeneric allows to retrieve logs from a test.
// If the test ID is not specified, the logs of the latest test is retrieved
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func testGetLogsGeneric(cmd *cobra.Command, args []string) error {

	RPPrinter.CustomPrintLn(genericTestLogsCmdAnsLine0)

	// Get the right test ID as a string
	testIDStr, err := getTestIDForAllObjects(cmd, args)
	if err != nil {
		return err
	}

	// Get the list of logs available for this test
	filesList := rpapiv1.RPFileList{}
	if err := bckAPI.GenericTestGetLogsFilesList(testIDStr, &filesList); err != nil {
		return err
	}

	// Manage the creation of the directory specified by the user
	testLogsDir, err := manageDownloadDir(cmd, "rp-tests-logs", testIDStr)
	if err != nil {
		return err
	}

	// Get the "zip" flag from the user
	zipRequested, err := cmd.Flags().GetBool(genericBuildZipFlag)
	if err != nil {
		logFlagError(genericBuildZipFlag)
		return err
	}

	RPPrinter.CustomPrintF(genericTestLogsCmdAnsLine1, testIDStr)

	// Now for each logs file, download it
	for _, logsFile := range filesList.ListFiles {
		// Download the file
		tmpFileStr, err := bckAPI.GenericTestGetLogsFile(testIDStr, logsFile.Name)
		if err != nil {
			RPPrinter.CustomColorPrintF(printer.ColorRed, genericTestLogsCmdAnsLine3, logsFile.Name, err)
			return nil
		}
		// Write the file
		err = utils.WriteToFile(path.Join(testLogsDir, logsFile.Name), tmpFileStr)
		if err != nil {
			RPPrinter.CustomColorPrintF(printer.ColorRed, genericTestLogsCmdAnsLine3, logsFile.Name, err)
			return nil
		}
	}

	if len(filesList.ListFiles) <= 0 {
		RPPrinter.CustomColorPrintF(printer.ColorYellow, genericTestLogsCmdAnsLine2, testIDStr)
	} else {
		dirOrFileInOutput := testLogsDir
		// Write logs to zip if asked
		if zipRequested {
			dirOrFileInOutput, err = zipLogs(testLogsDir)
			if err != nil {
				return nil
			}
		}

		RPPrinter.CustomColorPrintF(printer.ColorCyan, genericTestLogsCmdAnsLine4, testIDStr, dirOrFileInOutput)
	}

	return nil
}

// testCancelGeneric allows to cancel a test currently running.
// If the test ID is not specified, the latest test is cancelled.
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func testCancelGeneric(cmd *cobra.Command, args []string) error {

	RPPrinter.CustomPrintLn(genericTestCancelCmdAnsLine0)

	// Get the right test ID as a string
	testIDStr, err := getTestIDForAllObjects(cmd, args)
	if err != nil {
		return err
	}

	// Request the backend to cancel the test
	RPPrinter.CustomPrintF(genericTestCancelCmdAnsLine1, testIDStr)
	err = bckAPI.GenericTestCancel(testIDStr, nil)

	if err != nil {
		// Display KO
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		RPPrinter.ForceFlush()
		return err
	}

	// Display OK
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	return nil
}

// testGeneric launch a test for either the current application or the current image
// If the build ID is not specified, launch the test on the latest
// build.
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func testGeneric(cmd *cobra.Command, args []string) error {

	// Get the parent command to see if we start a build on an app or an image
	parentCmd := cmd.Parent()
	isAppTest := (parentCmd.Use == appsCommand)

	// Now, according to the case, execute specific tasks
	var image, appPrjID, appUUIDStr, parentForceFlag, parentStr string
	var testedObjUUID uuid.UUID
	var getUUIDFunc func(cmd *cobra.Command, args []string) (uuid.UUID, error)
	var err error
	if isAppTest {
		// APP TEST CASE

		// Set the variable for an app object
		parentForceFlag = appsForceAppFlag
		parentStr = "application"
		getUUIDFunc = getAppUUIDCallBack

		// Get the image flag
		image, err = cmd.Flags().GetString(appsTestCmdImageFlag)
		if err != nil {
			logFlagError(appsTestCmdImageFlag)
			return err
		}
	} else {
		// IMAGE TEST CASE

		// Set the variable for an image object
		parentForceFlag = imgsForceImgFlag
		parentStr = "image"
		getUUIDFunc = getImgUUIDCallBack

		// Get the app flag
		appID, err := cmd.Flags().GetString(imgsTestCmdAppFlag)
		if err != nil {
			logFlagError(imgsTestCmdAppFlag)
			return err
		}

		// Get the app project flag
		appPrjID, err = cmd.Flags().GetString(imgsTestCmdAppProjectFlag)
		if err != nil {
			logFlagError(imgsTestCmdAppProjectFlag)
			return err
		}

		// Retrieve the true project ID from UUID or its slug
		project := rpapiv1.RPProjectFull{RPProject: rpapiv1.RPProject{ID: uuid.Nil}}
		if appPrjID != "" {
			// Get the project
			project, err = getProjectFromIdentifier(appPrjID, false)
			if err != nil {
				return err
			}
		}

		if appID != "" {
			// Retrieve the true application ID from UUID or its slug
			app, err := getAppFromIdentifiers(appID, project.ID.String(), []string{"standard"}, false)
			if err != nil {
				return err
			}
			appUUIDStr = app.ID.String()
		}
	}

	// Convert object name to a title
	parentStrUp := cases.Title(language.English).String(parentStr)
	RPPrinter.CustomPrintF(genericTestCmdAnsLine0, parentStrUp)

	// Retrieve the build ID from the args given by the user
	buildID, err := getBuildIDFromArgs(cmd, args, parentForceFlag, parentStr, getUUIDFunc)
	if err != nil {
		return err
	}

	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Get the progress flag
	progress, err := cmd.Flags().GetBool(genericTestCmdProgressFlag)
	if err != nil {
		logFlagError(genericTestCmdProgressFlag)
		return err
	}

	// Get the image UUID if set
	imageUUID := uuid.Nil
	if image != "" {
		img, err := getImgFromIdentifier(image, false)
		if err != nil {
			return err
		}
		imageUUID = img.ID
	}

	// Get the board model to use for the tests
	boardModel, err := cmd.Flags().GetString(genericTestCmdBoardModelFlag)
	if err != nil {
		logFlagError(genericTestCmdBoardModelFlag)
		return err
	}

	// Get the target model structure from its slug
	boardModelStruct, err := getModelBoardStruct(boardModel)
	if err != nil {
		return err
	}

	// Get the time-out specified by the user
	testTimeOut, err := cmd.Flags().GetInt(genericTestCmdTimeOutFlag)
	if err != nil {
		logFlagError(genericTestCmdTimeOutFlag)
		return err
	}

	// Get the additional information (comment) to save with the test
	comment, err := cmd.Flags().GetString(genericTestCmdCommentFlag)
	if err != nil {
		logFlagError(genericTestCmdCommentFlag)
		return err
	}

	// Create the structure needed to start the tests
	testReq := rpapiv1.RPTestStart{
		TargetModelID:  boardModelStruct.ID,
		ImageID:        imageUUID,
		AppForTestID:   uuid.FromStringOrNil(appUUIDStr),
		TestTimeoutMin: testTimeOut,
		Comment:        comment,
	}

	if parentStr == "image" {
		testedObjUUID = testReq.ImageID
	} else {
		testedObjUUID = testReq.AppForTestID
	}

	// Tell the user that we start the test of the application
	RPPrinter.CustomPrintF(genericTestCmdAnsLine1, parentStr, buildID)

	// Ask the backend to start the test
	testAns := rpapiv1.RPTest{}
	if err := bckAPI.GenericTest(strconv.Itoa(buildID), &testReq, &testAns); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	// Display something to the user if in raw mode
	RPPrinter.CustomPrintObject(testAns, false, displayNothing)

	// Detect specific case when test aborted due to lack of redtest rpm
	if testAns.Status == rpapiv1.RPTestStatusAborted {
		if testAns.TotalTestsNb == 1 && testAns.SkippedNb == 1 {
			RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
			RPPrinter.CustomPrintObject(testAns, verbose, printOneTest)
			return nil
		}
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		RPPrinter.CustomColorPrintF(printer.ColorRed, genericTestCmdAnsLine5, parentStrUp)
		if testAns.ErrorMsg != "" {
			return errors.New(testAns.ErrorMsg)
		}
		return errors.New(appTestsAbortedMsg)
	}

	// Check from viper the non-blocking flag to see if we wait
	nonBlocking := viper.GetBool(nonBlockingFlag)
	if nonBlocking {
		// The user do not want blocking functions
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
		return nil
	}

	RPPrinter.CustomPrintLn()
	RPPrinter.CustomPrintLn()

	// Block until we receive a corresponding websocket event
	for {
		rEvent := <-cmdEventChannel

		switch rEvent.Type {
		case rpapiv1.EvtTestAdd, rpapiv1.EvtTestStateChange:

			// Function handling the events data
			rTest, err := rEvent.DecodeTest()
			if err != nil {
				return err
			}

			if rTest.ID == testAns.ID {

				RPPrinter.CustomPrintObject(rTest, verbose, printOneTestLive)

				// Check if the test is done
				if rTest.Status == "success" {
					RPPrinter.CustomColorPrintF(printer.ColorGreen, genericTestCmdAnsLine2, parentStrUp)
					return nil
				}
				// Check if the test is failed
				if rTest.Status == "failed" {
					RPPrinter.CustomColorPrintF(printer.ColorRed, genericTestCmdAnsLine3, parentStrUp)
					return errors.New(appTestsFailedMsg)
				}
				// Check if the test is timed-out
				if rTest.Status == "timed-out" {
					RPPrinter.CustomColorPrintF(printer.ColorRed, genericTestCmdAnsLine4, parentStrUp)
					return errors.New(appTestsTimedOutMsg)
				}
				// Check if the test is aborted
				if rTest.Status == "aborted" {
					RPPrinter.CustomColorPrintF(printer.ColorRed, genericTestCmdAnsLine5, parentStrUp)
					return errors.New(appTestsAbortedMsg)
				}
			}

		case rpapiv1.EvtTestProgress:
			// No need to go further if we do not display progress
			if !progress {
				break
			}

			// Function handling the events data
			rEvtTest, err := rEvent.DecodeEventTest()
			if err != nil {
				return err
			}

			if rEvtTest.ID == testAns.ID {
				// This test is the one we are monitoring
				RPPrinter.CustomPrintObject(rEvtTest, verbose, printOneTestProgress)
			}

		case rpapiv1.EvtTestConsole:
			// Function handling the events data
			rEvtTestConsole, err := rEvent.DecodeEventTestConsole()
			if err != nil {
				return err
			}

			if rEvtTestConsole.ID == testAns.ID {
				// This test is the one we are monitoring
				RPPrinter.CustomPrintObject(rEvtTestConsole, verbose, printOneTestConsole)
			}

		case rpapiv1.EVTError, rpapiv1.EVTSessionExpired:
			// Function handling the error
			err = handleErrorEvent(testedObjUUID, rEvent, false)
			if err != nil {
				return err
			}
		}
	}
}

// getAndDisplayTestsStats gets and display tests statistics from an
// application, an image or a project
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param projectID: UUID of the concerned project
// :type projectID: uuid.UUID
// :param appID: UUID of the concerned application
// :type appID: uuid.UUID
// :param imgID: UUID of the concerned image
// :type imgID: uuid.UUID
// :returns: error -- an error if something went wrong, nil in the other case
func getAndDisplayTestsStats(cmd *cobra.Command, projectID, appID, imgID uuid.UUID) error {
	var err error

	// Get the value of "last" flag
	// Default value at 10 for project to take care of the backend
	lastNb := 10
	if (appID != uuid.Nil) || (imgID != uuid.Nil) {
		lastNb, err = cmd.Flags().GetInt(genericTestsStatsCmdLastFlag)
		if err != nil {
			logFlagError(genericTestsStatsCmdLastFlag)
			return err
		}
	}

	// Get the value of the "verbose" flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Now request the backend for the projects stats
	projectTestsStats := rpapiv1.RPTestStatsProject{}
	if err := bckAPI.ProjectGetTestsStats(projectID.String(), lastNb, &projectTestsStats); err != nil {
		return err
	}

	// Now display either the full project builds stats or uniquely the app, according
	// to what the caller gave us
	if (appID == uuid.Nil) && (imgID == uuid.Nil) {
		return RPPrinter.CustomPrintObject(projectTestsStats, verbose, displayPrjTestsStats)
	} else if appID != uuid.Nil {
		for _, appStats := range projectTestsStats.AppStats {
			if utils.UUIDEqual(appStats.AppID, appID) {
				return RPPrinter.CustomPrintObject(appStats, verbose, displayAppTestsStats)
			}
		}
	} else if imgID != uuid.Nil {
		for _, imgStats := range projectTestsStats.ImageStats {
			if utils.UUIDEqual(imgStats.ImageID, imgID) {
				return RPPrinter.CustomPrintObject(imgStats, verbose, displayImgTestsStats)
			}
		}
	}

	// If we reach here, the application or the image has not been found
	// It means that no build exists for it, therefore, display an empty structure
	if appID != uuid.Nil {
		statsToDisplay := rpapiv1.RPTestStatsApp{
			AppID: appID,
		}
		return RPPrinter.CustomPrintObject(statsToDisplay, verbose, displayAppTestsStats)
	}

	statsToDisplay := rpapiv1.RPTestStatsImg{
		ImageID: imgID,
	}
	return RPPrinter.CustomPrintObject(statsToDisplay, verbose, displayImgTestsStats)
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Utilities functions ------------------------------------------
// --------------------------------------------------------------

// getTestIDForAllObjects go through the whole process of finding the right test ID
// in an standalone way (it process the commands and sub-commands for that)
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func getTestIDForAllObjects(cmd *cobra.Command, args []string) (string, error) {
	// Get the great-parent command to see if we start a build on an app or an image
	greatParentCmd := cmd.Parent().Parent()

	var parentForceFlag string
	if greatParentCmd.Use == appsCommand {
		// APP CASE
		parentForceFlag = appsForceAppFlag
	} else {
		// IMAGE CASE
		parentForceFlag = imgsForceImgFlag
	}

	// Try to get the "force-parent" flag
	forceParent, err := cmd.Flags().GetBool(parentForceFlag)
	if err != nil {
		logFlagError(parentForceFlag)
		return "", err
	}

	// Get the first argument of the command line
	argID, err := getFirstArg(args, fmt.Sprintf(objectIDNotSetMsg, "app/image/test"), fmt.Sprintf(objectIDRequiredMsg, "app/image/test"))
	if err != nil {
		return "", err
	}

	// Try to convert the argument given to an int
	testIDInt, err := strconv.Atoi(argID)
	isInteger := (err == nil)

	// Act according to what we have

	// If the arg given is an int and that the user do not force the arg to be a parent object
	// return this integer directly
	if isInteger && !forceParent {
		return strconv.Itoa(testIDInt), nil
	}

	// In the other case, the user argument is a slug, use it to retrieve the parent object and then the latest test
	var object interface{}
	if greatParentCmd.Use == appsCommand {
		// APP CASE
		object, err = genericGetAppFromArgs(cmd, args, false)
	} else {
		// IMAGE CASE
		object, err = genericGetImgFromArgs(args, false)
	}
	if err != nil {
		return "", err
	}

	// Retrieve the test ID from the args given by the user
	testID, err := getLatestTest(object)
	if err != nil {
		return "", err
	}

	return strconv.Itoa(testID), nil
}

// getLatestTest is a private function allowing to get the latest test of an object.
//
// :param object: The object for which the builds are wanted
// :type object: interface{}
// :param objType: The object type for which the builds are wanted (app or image)
// :type objType: string
// :returns: int, error -- the last test ID and an error if something went wrong, nil in the other case
func getLatestTest(object interface{}) (int, error) {

	// Use the right API call according to what is needed
	var testsList []rpapiv1.RPTest
	var err error

	switch typedObj := object.(type) {
	case rpapiv1.RPApplicationFull:
		// Get the full tests list for this application
		testsList, err = bckAPI.ApplicationsTestsListGet(typedObj.Parent.ProjectID.String(), typedObj.ID.String())
	case rpapiv1.RPImageFull:
		// Get the full tests list for this image
		testsList, err = bckAPI.ImagesTestsListGet(typedObj.ParentProjectID.String(), typedObj.ID.String())
	}

	if err != nil {
		return -1, err
	}

	// Then, go through all the tests and get the highest ID
	highestID := 0
	for _, test := range testsList {
		tmpID, err := strconv.Atoi(test.ID)
		if err != nil {
			return -1, err
		}

		if tmpID > highestID {
			highestID = tmpID
		}
	}
	return highestID, nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Display functions --------------------------------------------
// --------------------------------------------------------------

// displayOneTest is a private function allowing to display
// one test in the console
//
// :param application: Application to display
// :type application: interface{}
// :param verbose: Verbose flag passed in the command line
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func displayOneTest(test interface{}, verbose bool, writer *tabwriter.Writer) error {
	if !verbose {
		fmt.Fprintln(writer, genericTestsListHeader)
	}
	// Print the test
	printOneTest(test, verbose, writer)
	return nil
}

// displayTests is a private function allowing to display
// a slice of tests in the console
//
// :param tsts: tab of tests to display
// :type tsts: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func displayTests(tsts interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Convert to the right type
	tests := tsts.([]rpapiv1.RPTest)

	// Display result
	first := true
	for _, test := range tests {
		if !verbose && first {
			manageHeaderPrint(genericTestsListHeader, writer)
		} else if verbose && !first {
			fmt.Fprintln(writer, genericBuildCmdAnsLine4)
		}

		// Print this test
		printOneTest(test, verbose, writer)

		first = false
	}

	return nil
}

// displayPrjTestsStats is a private function allowing to display
// the tests statistics for a full project
//
// :param testsStatsInterface: Stats to display
// :type testsStatsInterface: interface{}
// :param verbose: Verbose flag passed in the command line
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func displayPrjTestsStats(testsStatsInterface interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Convert to the right type
	testsStats := testsStatsInterface.(rpapiv1.RPTestStatsProject)
	// Print the builds stats
	return printPrjTestsStats(testsStats, verbose, writer)
}

// displayAppTestsStats is a private function allowing to display
// the tests statistics for an app
//
// :param testsStatsInterface: Stats to display
// :type testsStatsInterface: interface{}
// :param verbose: Verbose flag passed in the command line
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func displayAppTestsStats(testsStatsInterface interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Convert to the right type
	testsStats := testsStatsInterface.(rpapiv1.RPTestStatsApp)
	// Print the builds stats
	return printAppTestsStats(testsStats, verbose, writer)
}

// displayImgTestsStats is a private function allowing to display
// the tests statistics for an image
//
// :param testsStatsInterface: Stats to display
// :type testsStatsInterface: interface{}
// :param verbose: Verbose flag passed in the command line
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func displayImgTestsStats(testsStatsInterface interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Convert to the right type
	testsStats := testsStatsInterface.(rpapiv1.RPTestStatsImg)
	// Print the builds stats
	return printImgTestsStats(testsStats, verbose, writer)
}

// printOneTest is a private function allowing to print one test in the console
//
// :param test: pointer to the test to print
// :type test: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param writer: Writer to used to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func printOneTest(test interface{}, verbose bool, writer *tabwriter.Writer) error {

	// Convert to the right type
	tst := test.(rpapiv1.RPTest)

	if verbose {

		// Set-up the value of image ID for the test
		imageIDToDisplay := mainNotSetMsg
		if !utils.UUIDEqual(uuid.Nil, tst.ImageID) {
			tmpImage, err := getImgFromIdentifier(tst.ImageID.String(), false)
			if err == nil {
				imageIDToDisplay = tmpImage.Slug
			}
		}

		// Set-up the value of app ID for the test
		appIDToDisplay := "[Default image test app]"
		if tst.Parent.AppSlug != "" {
			appIDToDisplay = tst.Parent.AppSlug
		}

		// Create the string containing the results
		fmt.Fprintln(writer, genericTestCmdVerboseLine0, tst.ID)
		fmt.Fprintln(writer, genericTestCmdVerboseLine11, tst.Started.String())
		fmt.Fprintln(writer, genericTestCmdVerboseLine1, appIDToDisplay)
		fmt.Fprintln(writer, genericTestCmdVerboseLine3, tst.Parent.BuildID)
		fmt.Fprintln(writer, genericTestCmdVerboseLine8, imageIDToDisplay)
		fmt.Fprintln(writer, genericTestCmdVerboseLine9, tst.ModelBoardSlug)
		fmt.Fprintln(writer, genericTestCmdVerboseLine2, tst.Status)
		fmt.Fprintln(writer, genericTestCmdVerboseLine4, strconv.Itoa(tst.SucceededNb))
		fmt.Fprintln(writer, genericTestCmdVerboseLine5, strconv.Itoa(tst.FailedNb))
		fmt.Fprintln(writer, genericTestCmdVerboseLine6, strconv.Itoa(tst.SkippedNb))
		fmt.Fprintln(writer, genericTestCmdVerboseLine7, strconv.Itoa(tst.TotalTestsNb))

		// Print comment for the test
		commentToPrint := mainNotSetMsg
		if tst.Comment != "" {
			commentToPrint = tst.Comment
		}
		fmt.Fprintln(writer, genericTestCmdVerboseLine10, commentToPrint)

	} else {
		// Not verbose mode, simply print 2 columns with test ID and the status
		fmt.Fprintln(writer, tst.ID, "\t", tst.Status)
	}

	return nil
}

// printOneTestLive is a private function allowing to print one test in the console
// in live while the test is currently running
//
// :param test: pointer to the test to print
// :type test: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param writer: Writer to used to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func printOneTestLive(test interface{}, verbose bool, writer *tabwriter.Writer) error {

	// Convert to the right type
	tst := test.(rpapiv1.RPTest)

	if verbose {
		// Create the string to print
		strToPrint := fmt.Sprintf(genericTestLiveCmdVerbose1Line,
			tst.Status,
			tst.SucceededNb,
			tst.FailedNb,
			tst.SkippedNb)

		// Print the line
		RPPrinter.CustomColorPrintLn(printer.ColorWhite, strToPrint)

	} else {
		// Create the string to print
		strToPrint := fmt.Sprintf(genericTestLiveCmdNoVerbose1Line,
			tst.Status,
			tst.TotalTestsNb)

		// Print the line
		RPPrinter.CustomColorPrintLn(printer.ColorWhite, strToPrint)
	}

	return nil
}

// printOneTestProgress is a private function allowing to print one test progress event
// in the console
//
// :param testEvt: pointer to the test progress event to print
// :type testEvt: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param writer: Writer to used to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func printOneTestProgress(testEvt interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Convert to the right type
	tstEvt := testEvt.(rpapiv1.RPEventTest)

	if verbose {
		// Create the string to print
		strToPrint := fmt.Sprintf(genericTestProgressVerbose1Line,
			tstEvt.CurrentTestNb,
			tstEvt.CurrentTestStatus,
			tstEvt.CurrentTestsOKNb,
			tstEvt.CurrentTestsKONb,
			tstEvt.CurrentTestsSkippedNb)

		// Print the line
		RPPrinter.CustomColorPrintLn(printer.ColorBlue, strToPrint)

	} else {
		// Create the string containing the results
		strToPrint := fmt.Sprintf(genericTestProgNoVerboseLine0,
			tstEvt.CurrentTestNb,
			tstEvt.CurrentTestStatus,
			tstEvt.TotalTestsNb)

		// Print the line
		RPPrinter.CustomColorPrintLn(printer.ColorBlue, strToPrint)
	}

	return nil
}

// printOneTestConsole is a private function allowing to print one test console event
// in the console
//
// :param tstEvtCsl: pointer to the test console event to print
// :type tstEvtCsl: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param writer: Writer to used to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func printOneTestConsole(tstEvtCsl interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Convert to the right type
	tstEvt := tstEvtCsl.(rpapiv1.RPEventTestConsole)

	// Print the output only in verbose mode
	if verbose {
		stdoutLines := []string{}
		stderrLines := []string{}

		// First, split stdout and stderr per line
		if tstEvt.LineStdOut != "" {
			stdoutLines = strings.Split(strings.TrimSpace(tstEvt.LineStdOut), "\n")
		}
		if tstEvt.LineStdErr != "" {
			stderrLines = strings.Split(strings.TrimSpace(tstEvt.LineStdErr), "\n")
		}

		// Print first stdout
		for _, line := range stdoutLines {
			lineToPrint := fmt.Sprintf(genericTestConsVerbose1Line, tstEvt.Timestamp, line)
			RPPrinter.CustomColorPrintLn(printer.ColorGreen, lineToPrint)
		}

		// Then print stderr
		for _, line := range stderrLines {
			lineToPrint := fmt.Sprintf(genericTestConsVerbose1Line, tstEvt.Timestamp, line)
			RPPrinter.CustomColorPrintLn(printer.ColorGreen, lineToPrint)
		}
	}

	return nil
}

// printPrjTestsStats is a private function allowing to print the statistics of
// tests inside a project
//
// :param testsStats: structure containing the stats to print
// :type testsStats: rpapiv1.RPBuildStatsProject
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param writer: Writer to used to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func printPrjTestsStats(testsStats rpapiv1.RPTestStatsProject, verbose bool, writer *tabwriter.Writer) error {
	// Firstly print basics information
	fmt.Fprintln(writer, prjsTestsStatsCmdAnsLine0, testsStats.ProjectID.String())
	fmt.Fprintln(writer, prjsTestsStatsCmdAnsLine1, testsStats.NbSucceeded)
	fmt.Fprintln(writer, prjsTestsStatsCmdAnsLine2, testsStats.NbFailed)
	fmt.Fprintln(writer, prjsTestsStatsCmdAnsLine3, testsStats.NbCanceled)
	fmt.Fprintln(writer, prjsTestsStatsCmdAnsLine4, testsStats.NbTotal)

	// If verbose is requested, give stats for every object
	if verbose {
		// Check if there are app stats to show
		if len(testsStats.AppStats) > 0 {
			fmt.Fprintln(writer)
			fmt.Fprintln(writer, prjsTestsStatsCmdAnsLine5)
			fmt.Fprintln(writer)

			// Display a header
			fmt.Fprintln(writer, prjsTestsStatsCmdAnsAppsHeader)
			for _, appStats := range testsStats.AppStats {
				fmt.Fprintln(writer, appStats.AppID.String(), "\t", appStats.AppSlug, "\t", appStats.PackageName, "\t", appStats.NbSucceeded, "\t", appStats.NbFailed, "\t", appStats.NbCanceled, "\t", appStats.NbTotal)
			}
		}

		// Check if there are image stats to show
		if len(testsStats.ImageStats) > 0 {
			fmt.Fprintln(writer)
			fmt.Fprintln(writer, prjsTestsStatsCmdAnsLine6)
			fmt.Fprintln(writer)

			// Display a header
			fmt.Fprintln(writer, prjsTestsStatsCmdAnsImgsHeader)
			for _, imgStats := range testsStats.ImageStats {
				fmt.Fprintln(writer, imgStats.ImageID.String(), "\t", imgStats.ImageSlug, "\t", imgStats.NbSucceeded, "\t", imgStats.NbFailed, "\t", imgStats.NbCanceled, "\t", imgStats.NbTotal)
			}
		}
	}

	return nil
}

// printAppTestsStats is a private function allowing to print the tests statistics
// of an application
//
// :param testsStats: structure containing the stats to print
// :type testsStats: rpapiv1.RPTestStatsProject
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param writer: Writer to used to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func printAppTestsStats(testsStats rpapiv1.RPTestStatsApp, verbose bool, writer *tabwriter.Writer) error {
	// Print the information wanted
	fmt.Fprintln(writer, appsTestsStatsCmdAnsLine0, testsStats.AppID.String())
	fmt.Fprintln(writer, appsTestsStatsCmdAnsLine1, testsStats.AppSlug)
	fmt.Fprintln(writer, appsTestsStatsCmdAnsLine2, testsStats.PackageName)
	fmt.Fprintln(writer, genericTestsStatsCmdAnsLine3, testsStats.NbSucceeded)
	fmt.Fprintln(writer, genericTestsStatsCmdAnsLine4, testsStats.NbFailed)
	fmt.Fprintln(writer, genericTestsStatsCmdAnsLine5, testsStats.NbCanceled)
	fmt.Fprintln(writer, genericTestsStatsCmdAnsLine6, testsStats.NbTotal)

	// If verbose is requested, give stats for every build
	if verbose {
		fmt.Fprintln(writer)
		fmt.Fprintln(writer, genericTestsStatsCmdAnsLine7)
		fmt.Fprintln(writer)

		// Display a header
		fmt.Fprintln(writer, genericTestsStatsCmdAnsBuildsHeader)
		for _, test := range testsStats.LastCITests {
			fmt.Fprintln(writer, test.ID, "\t", "CI_test", "\t", test.Status)
		}
		for _, audit := range testsStats.LastAuditTests {
			fmt.Fprintln(writer, audit.ID, "\t", "audit", "\t", audit.Status)
		}
	}

	return nil
}

// printImgTestsStats is a private function allowing to print the tests statistics
// of an image
//
// :param testsStats: structure containing the stats to print
// :type testsStats: rpapiv1.RPTestStatsProject
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param writer: Writer to used to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func printImgTestsStats(testsStats rpapiv1.RPTestStatsImg, verbose bool, writer *tabwriter.Writer) error {
	// Print the information wanted
	fmt.Fprintln(writer, imgsTestsStatsCmdAnsLine0, testsStats.ImageID.String())
	fmt.Fprintln(writer, imgsTestsStatsCmdAnsLine1, testsStats.ImageSlug)
	fmt.Fprintln(writer, genericTestsStatsCmdAnsLine3, testsStats.NbSucceeded)
	fmt.Fprintln(writer, genericTestsStatsCmdAnsLine4, testsStats.NbFailed)
	fmt.Fprintln(writer, genericTestsStatsCmdAnsLine5, testsStats.NbCanceled)
	fmt.Fprintln(writer, genericTestsStatsCmdAnsLine6, testsStats.NbTotal)

	// If verbose is requested, give stats for every build
	if verbose {
		fmt.Fprintln(writer)
		fmt.Fprintln(writer, genericTestsStatsCmdAnsLine7)
		fmt.Fprintln(writer)

		// Display a header
		fmt.Fprintln(writer, genericTestsStatsCmdAnsBuildsHeader)
		for _, test := range testsStats.LastCITests {
			fmt.Fprintln(writer, test.ID, "\t", "CI_test", "\t", test.Status)
		}
	}

	return nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------
