/*
Copyright © 2021-2023 IoT.bzh Company
Contact: https://www.iot.bzh/licensing

This file is part of the redpesk-cli project, the redpesk command line tool.

$RP_BEGIN_LICENSE$
Commercial License Usage
 Licensees holding valid commercial IoT.bzh licenses may use this file in
 accordance with the commercial license agreement provided with the
 Software or, alternatively, in accordance with the terms contained in
 a written agreement between you and The IoT.bzh Company. For licensing terms
 and conditions see https://www.iot.bzh/terms-conditions. For further
 information use the contact form at https://www.iot.bzh/contact.

GNU General Public License Usage
 Alternatively, this file may be used under the terms of the GNU General
 Public license version 3. This license is as published by the Free Software
 Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
 of this file. Please review the following information to ensure the GNU
 General Public License requirements will be met
 https://www.gnu.org/licenses/gpl-3.0.html.
$RP_END_LICENSE$
*/

package cmd

import (
	"fmt"
	"strings"
	"testing"

	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/utils"
	"github.com/stretchr/testify/assert"
)

// --------------------------------------------------------------
// Utilities functions ------------------------------------------
// --------------------------------------------------------------

// checkIfBuildsExist is function allowing the tests functions to
// test if some builds exist already in the "database"
//
// :param appName: Name of the function to check out
// :type appName: string
// :param masterCommand: Name of the master command to use (applications, images, etc.)
// :type masterCommand: string
// :returns: string, error -- build ID if one exists and an error if a problem has occurred
func checkIfBuildsExist(objName, masterCommand string) (string, error) {

	// Run the "builds list" command an get the output
	cmdOutput, err := executeCommand(masterCommand, "builds", "list", objName)
	if err != nil {
		objType := strings.TrimRight(masterCommand, "s")
		if strings.Contains(err.Error(), fmt.Sprintf("no %s has been found", objType)) {
			return "", nil
		}
		return "", err
	}

	// The command went well, check if we have a build ID
	// (get the first one we found with a "done" status )
	buildID := ""
	outputLines := strings.Split(cmdOutput, "\n")
	if len(outputLines) <= 1 {
		return buildID, nil
	}

	for _, line := range outputLines[1:] {
		// Get the ID available on the line
		splitLine := strings.Fields(line)

		if utils.ContainsString(splitLine, "done") {
			buildID = strings.TrimSpace(splitLine[0])
			break
		}
	}
	return buildID, nil
}

// checkTestOutput is function allowing to process the output of
// a "(applications|images) test [-v]" command, and checks each object of the list
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param outputCmd: Output to process
// :type outputCmd: string
// :param testType: Type of test that has been run (app or image)
// :type testType: string
func checkTestOutput(test *testing.T, outputCmd, testType string, verbose bool) {

	trimmedOutput := strings.TrimSpace(outputCmd)

	// Firstly, transform the raw output to a slice of lines
	linesSlice := strings.Split(trimmedOutput, "\n")

	// Check that the first line corresponds to what we want
	firstLine := fmt.Sprintf(genericTestCmdAnsLine0, strings.Title(testType))
	assert.Contains(test, firstLine, linesSlice[0])

	// Check that the second line corresponds to what we want
	secondLine := fmt.Sprintf("Testing the %s on build", testType)
	assert.Contains(test, linesSlice[1], secondLine)

	// Go further on the slice to check
	linesSlice = linesSlice[2:]

	// Now process each line, accumulate object and check them
	containsStateChange := false
	for _, line := range linesSlice {
		if strings.Contains(line, "[STATE CHANGE]") && !verbose {
			containsStateChange = true
			assert.Regexp(test, `^\[STATE CHANGE\] (pending|deploying|started|success|failed|testing) +-- TOTAL => [0-9]+`, line)
		}
		if strings.Contains(line, "[STATE CHANGE]") && verbose {
			containsStateChange = true
			assert.Regexp(test, `^\[STATE CHANGE\] (pending|deploying|started|success|failed|testing) +-- TOTAL => Success: [0-9]* +- Fail: [0-9]* +- Skip: [0-9]*`, line)
		}
		if strings.Contains(line, "[OUTPUT]") {
			assert.Regexp(test, `^\[OUTPUT\] [0-9]{2}-[0-9]{2}-[0-9]{4} [0-9]{2}:[0-9]{2}:[0-9]{2} > .*`, line)
		}
	}

	// Check that at least one "[STATE CHANGE]" has been seen
	assert.True(test, containsStateChange)

	// Check that the last line is good
	successEnd := strings.ReplaceAll(fmt.Sprintf(genericTestCmdAnsLine2, strings.Title(testType)), "\n", "")
	failEnd := strings.ReplaceAll(fmt.Sprintf(genericTestCmdAnsLine3, strings.Title(testType)), "\n", "")
	availableEnds := []string{successEnd, failEnd}
	assert.Contains(test, availableEnds, linesSlice[len(linesSlice)-1])
}

// checkVerboseListTestsOutput is function allowing to run the
// "tests list -v" command, and checks each object of the list
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param isImgTest: To set to true if the object tested is an image
// :type isImgTest: bool
// :param args: Args to pass as command (sub-command, flags, etc.)
// :type args: ...string
func checkVerboseListTestsOutput(test *testing.T, isImgTest bool, args ...string) {

	// Create the regexp map used to check
	regexpMap := map[string]string{
		"Test ID:":               `^[0-9]+$`,
		"Start time":             `.+`,
		"Application ID:":        `^[a-z0-9-]+$`,
		"Build ID":               `[0-9]{4}`,
		"Image ID":               `^(\[Not set\]|[a-z0-9-]+)$`,
		"Board model":            `^qemu-system-x86_64$`,
		"Status":                 `^(pending|deploying|started|testing|success|failed)$`,
		"Succeeded tests number": `[0-9]+`,
		"Failed tests number":    `[0-9]+`,
		"Skipped tests number":   `[0-9]+`,
		"Total tests number":     `[0-9]+`,
		"Comment":                `.+`,
	}

	if isImgTest {
		regexpMap["Application ID:"] = `\[Default image test app\]`
	}

	checkCmdAnswerListKeysRegexp(test, appTestsDisplayVerboseLineNb, genericBuildCmdAnsLine4, regexpMap, args...)
}

// checkNotVerboseListTestsOutput is function allowing to run the
// "applications tests-list" command, and checks each object of the list
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param args: Args to pass as command (sub-command, flags, etc.)
// :type args: ...string
func checkNotVerboseListTestsOutput(test *testing.T, args ...string) {

	// Create the regexp map used to check
	regexpMap := map[string]string{
		"ID":     `^[0-9]+$`,
		"Status": `^(pending|deploying|started|testing|success|failed)$`,
	}

	checkCmdAnsKeysRegexpCol(test, 2, regexpMap, args...)
}

// --------------------------------------------------------------
// --------------------------------------------------------------
