/*
Copyright © 2020-2023 IoT.bzh Company
Contact: https://www.iot.bzh/licensing

This file is part of the redpesk-cli project, the redpesk command line tool.

$RP_BEGIN_LICENSE$
# Commercial License Usage

Licensees holding valid commercial IoT.bzh licenses may use this file in
accordance with the commercial license agreement provided with the
Software or, alternatively, in accordance with the terms contained in
a written agreement between you and The IoT.bzh Company. For licensing terms
and conditions see https://www.iot.bzh/terms-conditions. For further
information use the contact form at https://www.iot.bzh/contact.

# GNU General Public License Usage

Alternatively, this file may be used under the terms of the GNU General
Public license version 3. This license is as published by the Free Software
Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
of this file. Please review the following information to ensure the GNU
General Public License requirements will be met
https://www.gnu.org/licenses/gpl-3.0.html.

$RP_END_LICENSE$
*/
package cmd

import (
	"fmt"
	"strings"
	"testing"

	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/rpconnect"
	"github.com/stretchr/testify/assert"
)

// TestCompletionConnectAliasTab is the function allowing to test if
// the completion of connection's alias list is working correctly
// in rp-cli
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestCompletionConnectAliasTab(test *testing.T) {

	// Create a false token file with the values we want
	accessTokenFilePath, err := createFalseTokenFile([]string{"testAlias", "productionTest", "preprodTest"})
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false file created at the beginning
	defer deleteFalseFile(accessTokenFilePath)

	// Set-up the answer list
	answerList := []string{"testAlias", "productionTest", "preprodTest"}

	// Create the list of commands to run for test
	cmdList := [][]string{
		{prjsCommand, prjsAddCmd, "--serveralias"},
		{prjsCommand, prjsAddCmd, "--serveralias"},
		{miscCommand, miscVersionCmd, "--serveralias"},
		{miscCommand, miscArchCmd, "--serveralias"},
	}

	// Check the whole list of commands
	err = checkListOfCompletionCmd(test, cmdList, answerList)
	if err != nil {
		test.Fatal(err)
	}
}

// TestCompletionLBConnectAliasTab is the function allowing to test if
// the completion of connection's alias list is working correctly
// in rp-cli
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestCompletionLBConnectAliasTab(test *testing.T) {
	// Create a new local-builder connection object
	connectToCreate := rpconnect.RPCLIConnection{
		Alias:     "localbuilder-test-completion-0",
		ServerURL: "10.231.233.1",
	}
	localBuilderFilePath, err := addLocalConnectionForTest(test, &connectToCreate)

	// Delete the false file created at the beginning
	defer deleteFalseFile(localBuilderFilePath)
	if err != nil {
		test.Fatal(err)
	}

	// Create other connections
	for i := 1; i <= 3; i++ {
		connectToCreate.Alias = fmt.Sprintf("localbuilder-test-completion-%d", i)
		_, err = addLocalConnectionForTest(test, &connectToCreate)
		if err != nil {
			test.Fatal(err)
		}
	}

	// Set-up the answer list
	answerList := []string{"localbuilder-test-completion-0", "localbuilder-test-completion-1", "localbuilder-test-completion-2", "localbuilder-test-completion-3"}

	// Create the list of commands to run for test
	cmdList := [][]string{
		{localCommand, connectionsCommand, "get"},
		{localCommand, connectionsCommand, "update"},
		{localCommand, connectionsCommand, "delete"},
		{localCommand, connectionsCommand, "use"},
	}

	// Check the whole list of commands
	err = checkListOfCompletionCmd(test, cmdList, answerList)
	if err != nil {
		test.Fatal(err)
	}
}

// TestCompletionArchTab is the function allowing to test if
// the completion of architectures list is working correctly
// in rp-cli
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestCompletionArchTab(test *testing.T) {

	// Set-up the answer list
	answerList := []string{"x86_64", "aarch64"}

	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("SNCF", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "sncf", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Then, creates an application if it does not exists
	cmdToRun2 := getMainAddAppCmd("TGV", "train-grande-vitesse", "sncf")
	err = addObjectIfNotExists(test, appsCommand, "tgv", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the list of commands to run for test
	cmdList := [][]string{
		{prjsCommand, prjsAddCmd, "--mandatory-arch"},
		{prjsCommand, prjsAddCmd, "--optional-arch"},
		{prjsCommand, "update", "sncf", "--mandatory-arch"},
		{prjsCommand, "update", "sncf", "--optional-arch"},
	}

	// Check the whole list of commands
	err = checkListOfCompletionCmd(test, cmdList, answerList)
	if err != nil {
		test.Fatal(err)
	}
}

// TestCompletionListTypeTab is the function allowing to test if
// the completion of list type is working correctly
// in rp-cli
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestCompletionListTypeTab(test *testing.T) {

	// Check for projects and apps --------------
	// Set-up the answer list
	answerList := []string{"external", "standard", "distro"}

	// Create the list of commands to run for test
	cmdList := [][]string{
		{prjsCommand, prjsListCmd, "--type"},
		{appsCommand, appsListCmd, "--type"},
	}

	// Check the whole list of commands
	err := checkListOfCompletionCmd(test, cmdList, answerList)
	if err != nil {
		test.Fatal(err)
	}
	// ------------------------------------------

	// Check for images -------------------------
	// Set-up the answer list
	answerList = []string{"external", "standard", "iotbzh"}

	// Create the list of commands to run for test
	cmdList = [][]string{
		{imagesCmdName, imagesListCmd, "--type"},
	}

	// Check the whole list of commands
	err = checkListOfCompletionCmd(test, cmdList, answerList)
	if err != nil {
		test.Fatal(err)
	}
	// ------------------------------------------
}

// TestCompletionBoardModelsTab is the function allowing to test if
// the completion of boards list is working correctly
// in rp-cli
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestCompletionBoardModelsTab(test *testing.T) {

	// Set-up the answer list
	answerList := []string{"qemu-system-x86_64", "r8a7795-h3ulcb-kf", "x86-x5-z8350", "bcm2711-rpi-4-b", "nitrogen8mn"}

	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Guts", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "guts", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Then, creates an application if it does not exists
	cmdToRun2 := getMainAddAppCmd("Come closer", "come-closer", "guts")
	err = addObjectIfNotExists(test, appsCommand, "come-closer", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Check if the "come-closer" application exists and if it has builds
	buildID, err := checkIfBuildsExist("come-closer", appsCommand)
	if err != nil {
		test.Fatal(err)
	}

	// Check if at least one build exists
	if buildID == "" {
		// No build exists, creates one
		_, err = executeCommand(appsCommand, "build", "come-closer", "-p", "guts")
		if err != nil {
			test.Fatal(err)
		}
	}

	// Create the list of commands to run for test
	cmdList := [][]string{
		{appsCommand, "test", "--board-model"},
	}

	// Check the whole list of commands
	err = checkListOfCompletionCmd(test, cmdList, answerList)
	if err != nil {
		test.Fatal(err)
	}
	// ------------------------------------------
}

// TestCompletionAllBoardModelsTab is the function allowing to test if
// the completion of board models list is working correctly
// in rp-cli
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestCompletionAllBoardModelsTab(test *testing.T) {

	// Set-up the answer list
	answerList := []string{"qemu-system-x86_64",
		"r8a7795-h3ulcb-kf",
		"x86-x5-z8350",
		"bcm2711-rpi-4-b",
		"nitrogen8mn",
		"test_ota_x86_64",
		"test_ota_aarch64",
	}

	// Create the list of commands to run for test
	cmdList := [][]string{
		{boardModelsCommand, "get"},
	}

	// Check the whole list of commands
	err := checkListOfCompletionCmd(test, cmdList, answerList)
	if err != nil {
		test.Fatal(err)
	}
	// ------------------------------------------
}

// TestCompletionAllBoardsTab is the function allowing to test if
// the completion of boards list is working correctly
// in rp-cli
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestCompletionAllBoardsTab(test *testing.T) {

	// Set-up the answer list
	answerList := []string{
		"ota-redpesk-lts-arz-1.0-x86_64",
		"ota-redpesk-lts-arz-1.0-aarch64",
		"qemu-system-x86_64",
		"h3ulcb-ws3.0_25159",
		"nitrogen8mn01",
		"qemu-redpesk01",
		"rpi4",
		"solidrun-edge-gateways",
		"upboard04",
	}

	// Create the list of commands to run for test
	cmdList := [][]string{
		{boardsCommand, "get"},
	}

	// Check the whole list of commands
	err := checkListOfCompletionCmd(test, cmdList, answerList)
	if err != nil {
		test.Fatal(err)
	}
	// ------------------------------------------
}

// TestCompletionDistTab is the function allowing to test if
// the completion of distributions list is working correctly
// in rp-cli
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestCompletionDistTab(test *testing.T) {

	// Set-up the answer list
	answerList := []string{"redpesk-devel-33", "redpesk-devel-arz"}

	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("SNCF", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "sncf", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Then, creates an application if it does not exists
	cmdToRun2 := getMainAddAppCmd("TGV", "train-grande-vitesse", "sncf")
	err = addObjectIfNotExists(test, appsCommand, "tgv", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the list of commands to run for test
	cmdList := [][]string{
		{prjsCommand, prjsAddCmd, "--mandatory-distro"},
		{prjsCommand, prjsAddCmd, "--optional-distro"},
		{prjsCommand, "update", "sncf", "--mandatory-distro"},
		{prjsCommand, "update", "sncf", "--optional-distro"},
		{appsCommand, appsAddCmd, "-p", "sncf", "--mandatory-distro"},
		{appsCommand, appsAddCmd, "-p", "sncf", "--optional-distro"},
		{appsCommand, "update", "tgv", "--mandatory-distro"},
		{appsCommand, "update", "tgv", "--optional-distro"},
	}

	// Check the whole list of commands
	err = checkListOfCompletionCmd(test, cmdList, answerList)
	if err != nil {
		test.Fatal(err)
	}
}

// TestCompletionAppsTab is the function allowing to test if
// the completion of applications list is working correctly
// in rp-cli
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestCompletionAppsTab(test *testing.T) {

	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("SNCF", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "sncf", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the commands
	cmdToRun2 := getMainAddAppCmd("TGV", "train-grande-vitesse", "sncf")
	cmdToRun3 := getMainAddAppCmd("TER", "train-express-regional", "sncf")
	cmdToRun4 := getMainAddAppCmd("Train Corail", "train-corail", "sncf")

	// Create the applications if they do not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, appsCommand, "tgv", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, appsCommand, "ter", cmdToRun3...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, appsCommand, "train-corail", cmdToRun4...)
	if err != nil {
		test.Fatal(err)
	}

	// Set-up the answer list
	answerList := []string{"tgv", "ter", "train-corail"}

	// Create the list of commands to run for test
	cmdList := [][]string{
		{appsCommand, "get"},
		{appsCommand, "update"},
		{appsCommand, "delete"},
		{appsCommand, appsFilesCmd, "upload"},
		{appsCommand, "build"},
		{appsCommand, "builds", "list"},
		{appsCommand, "test"},
		{appsCommand, "tests", "list"},
	}

	// Check the whole list of commands
	err = checkListOfCompletionCmd(test, cmdList, answerList)
	if err != nil {
		test.Fatal(err)
	}
}

// TestCompletionAppsFromPrjTab is the function allowing to test if
// the completion of applications list is working correctly
// in rp-cli when a project is specified
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestCompletionAppsFromPrjTab(test *testing.T) {

	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("my-completion-project", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "my-completion-project", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the commands
	cmdToRun2 := getMainAddAppCmd("completion-app-1", "completion-app-1", "my-completion-project")
	cmdToRun3 := getMainAddAppCmd("completion-app-2", "completion-app-2", "my-completion-project")
	cmdToRun4 := getMainAddAppCmd("completion-app-3", "completion-app-3", "my-completion-project")

	// Create the applications if they do not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, appsCommand, "completion-app-1", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, appsCommand, "completion-app-2", cmdToRun3...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, appsCommand, "completion-app-3", cmdToRun4...)
	if err != nil {
		test.Fatal(err)
	}

	// Set-up the answer list
	answerList := []string{"completion-app-1", "completion-app-2", "completion-app-3"}

	// Create the list of commands to run for test
	cmdList := [][]string{
		{prjsCommand, "update", "my-completion-project", "--bld-type-apps"},
	}

	// Check the whole list of commands
	err = checkListOfCompletionCmd(test, cmdList, answerList)
	if err != nil {
		test.Fatal(err)
	}
}

// TestCompletionAppsSamplesTab is the function allowing to test if
// the completion of applications samples list is working correctly
// in rp-cli
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestCompletionAppsSamplesTab(test *testing.T) {

	// Set-up the answer list
	answerList := []string{"helloworld-binding"}

	// Create the list of commands to run for test
	cmdList := [][]string{
		{appsCommand, "add-from-sample"},
	}

	// Check the whole list of commands
	err := checkListOfCompletionCmd(test, cmdList, answerList)
	if err != nil {
		test.Fatal(err)
	}
}

// TestCompletionAppsServicesTab is the function allowing to test if
// the completion of applications services list is working correctly
// in rp-cli
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestCompletionAppsServicesTab(test *testing.T) {

	// Set-up the answer list
	answerList := []string{"import_gitpkg=", "download_sources=", "excludearch=", "setverrel="}

	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("SNCF", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "sncf", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Then, creates an application if it does not exists
	cmdToRun2 := getMainAddAppCmd("TGV", "train-grande-vitesse", "sncf")
	err = addObjectIfNotExists(test, appsCommand, "tgv", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the list of commands to run for test
	cmdList := [][]string{
		{appsCommand, appsAddCmd, "--service"},
		{appsCommand, "update", "tgv", "--service"},
	}

	// Check the whole list of commands
	err = checkListOfCompletionCmd(test, cmdList, answerList)
	if err != nil {
		test.Fatal(err)
	}
}

// TestCompletionAuditToolsTab is the function allowing to test if
// the completion of audit tools list is working correctly
// in rp-cli
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestCompletionAuditToolsTab(test *testing.T) {

	// Set-up the answer list
	answerList := []string{"gosec", "flawfinder"}

	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Marco the best", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "marco-the-best", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Then, creates an application if it does not exists
	cmdToRun2 := getMainAddAppCmd("Pictures", "pictures", "marco-the-best")
	err = addObjectIfNotExists(test, appsCommand, "pictures", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the list of commands to run for test
	cmdList := [][]string{
		{appsCommand, appsAddCmd, "--audits"},
		{appsCommand, "update", "pictures", "--audits"},
		{appsCommand, "audit", "pictures", "--tool"},
	}

	// Check the whole list of commands
	err = checkListOfCompletionCmd(test, cmdList, answerList)
	if err != nil {
		test.Fatal(err)
	}
}

// TestCompletionImagesTab is the function allowing to test if
// the completion of images list is working correctly
// in rp-cli
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestCompletionImagesTab(test *testing.T) {

	// Create a project and an image
	err := createProjectAndImage(test, "Images completion Project", "My image completion")
	if err != nil {
		test.Fatal(err)
	}

	// Set-up the answer list
	answerList := []string{"my-image-completion", "redpesk-min-arz-aarch64-smack", "redpesk-min-arz-x8664-smack"}

	// Create the list of commands to run for test
	cmdList := [][]string{
		{imagesCmdName, "get"},
		{imagesCmdName, "update"},
		{imagesCmdName, "delete"},
		{imagesCmdName, "localbuild"},
		{imagesCmdName, "build"},
		{imagesCmdName, "builds", "list"},
		{imagesCmdName, "builds", "get"},
		{imagesCmdName, "builds", "logs"},
		{imagesCmdName, "release"},
	}

	// Check the whole list of commands
	err = checkListOfCompletionCmd(test, cmdList, answerList)
	if err != nil {
		test.Fatal(err)
	}
}

// TestCompletionImagesFromPrjTab is the function allowing to test if
// the completion of images list is working correctly
// in rp-cli, but with a project specified
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestCompletionImagesFromPrjTab(test *testing.T) {
	// Create a project and an image
	err := createProjectAndImage(test, "Images completion Project 2", "My image")
	if err != nil {
		test.Fatal(err)
	}

	// Create another image in this project
	err = createProjectAndImage(test, "Images completion Project 2", "My image 2")
	if err != nil {
		test.Fatal(err)
	}

	// Set-up the answer list
	answerList := []string{"my-image", "my-image-2"}

	// Create the list of commands to run for test
	cmdList := [][]string{
		{prjsCommand, prjsBuildCmd, "images", "images-completion-project-2", "-i"},
		{prjsCommand, prjsBuildCmd, "images", "images-completion-project-2", "--images"},
	}

	// Check the whole list of commands
	err = checkListOfCompletionCmd(test, cmdList, answerList)
	if err != nil {
		test.Fatal(err)
	}

}

// TestCompletionReleasedImagesTab is the function allowing to test if
// the completion of released images list is working correctly
// in rp-cli
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestCompletionReleasedImagesTab(test *testing.T) {

	// Create a project, an image and released it
	err := createAndReleaseImage(test, "Images release project", "My released image comp", []string{}, []string{}, false)
	if err != nil {
		test.Fatal(err)
	}

	// Set-up the answer list
	answerList := []string{"my-released-image-comp-1.0"}

	// Create the list of commands to run for test
	cmdList := [][]string{
		{imagesCmdName, imageReleasesCmd, "get"},
		{imagesCmdName, imageReleasesCmd, "update"},
		{imagesCmdName, imageReleasesCmd, "delete"},
	}

	// Check the whole list of commands
	err = checkListOfCompletionCmd(test, cmdList, answerList)
	if err != nil {
		test.Fatal(err)
	}

	// Delete the previously released images
	checkActionCmdAns(test, imagesCmdName, "releases", "delete", "my-released-image-comp-1.0")
}

// TestCompletionSecurityTab is the function allowing to test if
// the completion of security model list is working correctly
// in rp-cli
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestCompletionSecurityTab(test *testing.T) {

	// Create a project and an image
	err := createProjectAndImage(test, "Images completion Project", "My image completion")
	if err != nil {
		test.Fatal(err)
	}

	// Set-up the answer list
	answerList := []string{"smack", "selinux"}

	// Create the list of commands to run for test
	cmdList := [][]string{
		{adminCommand, importImgCmd, "--security"},
		{imagesCmdName, "add", "--security"},
		{imagesCmdName, "update", "my-image-completion", "--security"},
	}

	// Check the whole list of commands
	err = checkListOfCompletionCmd(test, cmdList, answerList)
	if err != nil {
		test.Fatal(err)
	}
}

// TestCompletionPrjsTab is the function allowing to test if
// the completion of projects list is working correctly
// in rp-cli
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestCompletionPrjsTab(test *testing.T) {

	// Firstly, creates test projects if they do not exist
	cmdToRun1 := getMainAddPrjCmd("SNCF", "redpesk-devel-arz")
	cmdToRun2 := getMainAddPrjCmd("Airbus", "redpesk-devel-arz")
	cmdToRun3 := getMainAddPrjCmd("Air France", "redpesk-devel-arz")

	// Create the projects if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, prjsCommand, "sncf", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, prjsCommand, "airbus", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, prjsCommand, "air-france", cmdToRun3...)
	if err != nil {
		test.Fatal(err)
	}

	// Create one application if it does not already exists
	cmdToRun4 := getMainAddAppCmd("TGV", "train-grande-vitesse", "sncf")
	err = addObjectIfNotExists(test, appsCommand, "tgv", cmdToRun4...)
	if err != nil {
		test.Fatal(err)
	}

	// Set-up the answer list
	answerList := []string{"sncf", "airbus", "air-france"}

	// Create the list of commands to run for test
	cmdList := [][]string{
		{prjsCommand, "add", "--inherit"},
		{prjsCommand, "get"},
		{prjsCommand, "update"},
		{prjsCommand, "update", "air-france", "--inherit"},
		{prjsCommand, "delete"},
		{prjsCommand, "release"},
		{prjsCommand, "members", "add"},
		{prjsCommand, "members", "get"},
		{prjsCommand, "members", "update"},
		{prjsCommand, "members", "delete"},
		{appsCommand, "add", "-p"},
		{appsCommand, "list", "-p"},
		{appsCommand, "update", "tgv", "-p"},
		{appsCommand, appsFilesCmd, "upload", "tgv", "-p"},
		{appsCommand, "get", "tgv", "-p"},
		{appsCommand, "delete", "tgv", "-p"},
		{appsCommand, "build", "tgv", "-p"},
		{appsCommand, "builds", "list", "tgv", "-p"},
		{appsCommand, "test", "tgv", "-p"},
		{appsCommand, "tests", "list", "tgv", "-p"},
	}

	// Check the whole list of commands
	err = checkListOfCompletionCmd(test, cmdList, answerList)
	if err != nil {
		test.Fatal(err)
	}
}

// TestCompletionExtPrjsTab is the function allowing to test if
// the completion of external projects list is working correctly
// in rp-cli
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestCompletionExtPrjsTab(test *testing.T) {

	// Set-up the answer list
	answerList := []string{"redpesk-brehat", "redpesk-cezembre", "redpesk-devel-arz", "redpesk-lts-arz"}

	// Create the list of commands to run for test
	cmdList := [][]string{
		{adminCommand, importPrjCmd, "--id"},
	}

	// Check the whole list of commands
	err := checkListOfCompletionCmd(test, cmdList, answerList)
	if err != nil {
		test.Fatal(err)
	}
}

// TestCompletionUsersToInviteTab is the function allowing to test if
// the completion of the list of users to invite in a project is working correctly
// in rp-cli
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestCompletionUsersToInviteTab(test *testing.T) {

	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Test Project 3", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "test-project-3", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Then creates users if they do not exist
	cmdToRun2 := getMainAddUserCmd("tiramisu", "tiramisu@iot.bzh", "Ti", "Ramisu", "0666989945")
	cmdToRun3 := getMainAddUserCmd("kravmager", "krav@iot.bzh", "Imi", "Lichtenfeld", "0666999949")
	cmdToRun4 := getMainAddTeamCmd("Redpesk Team", "Poisson rouge!", true)
	err = addObjectIfNotExists(test, usersCommand, "tiramisu", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, usersCommand, "kravmager", cmdToRun3...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, teamsCommand, "redpesk-team", cmdToRun4...)
	if err != nil {
		test.Fatal(err)
	}

	// Set-up the answer list
	answerList := []string{"tiramisu", "kravmager", "redpesk-team"}

	// Create the list of commands to run for test
	cmdList := [][]string{
		{prjsCommand, "members", "add", "test-project-3", "-g"},
	}

	// Check the whole list of commands
	err = checkListOfCompletionCmd(test, cmdList, answerList)
	if err != nil {
		test.Fatal(err)
	}
}

// TestCompletionRolesTab is the function allowing to test if
// the completion of the list of roles for a project is working correctly
// in rp-cli
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestCompletionRolesTab(test *testing.T) {

	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Test Project 3", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "test-project-3", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Set-up the answer list
	answerList := []string{"manager", "tester", "developer"}

	// Create the list of commands to run for test
	cmdList := [][]string{
		{prjsCommand, "members", "add", "test-project-3", "-r"},
		{prjsCommand, "members", "update", "test-project-3", "-r"},
	}

	// Check the whole list of commands
	err = checkListOfCompletionCmd(test, cmdList, answerList)
	if err != nil {
		test.Fatal(err)
	}
}

// TestCompletionPrjMembersTab is the function allowing to test if
// the completion of the list of project's members is working correctly
// in rp-cli
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestCompletionPrjMembersTab(test *testing.T) {

	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Test Completion", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "test-completion", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Then creates users if they do not exist
	cmdToRun2 := getMainAddUserCmd("tiramisu", "tiramisu@iot.bzh", "Ti", "Ramisu", "0666989945")
	cmdToRun3 := getMainAddUserCmd("kravmager", "krav@iot.bzh", "Imi", "Lichtenfeld", "0666999949")
	cmdToRun4 := getMainAddTeamCmd("Redpesk Team", "Poisson rouge!", true)
	err = addObjectIfNotExists(test, usersCommand, "tiramisu", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, usersCommand, "kravmager", cmdToRun3...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, teamsCommand, "redpesk-team", cmdToRun4...)
	if err != nil {
		test.Fatal(err)
	}

	// Then add the members in the project
	cmdToRun5 := getMainAddMemberCmd("test-completion", "tiramisu", []string{"tester"})
	checkActionCmdAns(test, cmdToRun5...)
	cmdToRun6 := getMainAddMemberCmd("test-completion", "kravmager", []string{"developer", "tester"})
	checkActionCmdAns(test, cmdToRun6...)
	cmdToRun7 := getMainAddMemberCmd("test-completion", "redpesk-team", []string{"developer"})
	checkActionCmdAns(test, cmdToRun7...)

	// Set-up the answer list
	answerList := []string{"tiramisu", "kravmager", "redpesk-team"}

	// Create the list of commands to run for test
	cmdList := [][]string{
		{prjsCommand, "members", "update", "test-completion", "-m"},
		{prjsCommand, "members", "delete", "test-completion", "-m"},
	}

	// Check the whole list of commands
	err = checkListOfCompletionCmd(test, cmdList, answerList)
	if err != nil {
		test.Fatal(err)
	}
}

// TestCompletionTeamsTab is the function allowing to test if
// the completion of teams list is working correctly
// in rp-cli
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestCompletionTeamsTab(test *testing.T) {

	// Create the teams if they do not exists
	cmdToRun1 := getMainAddTeamCmd("Bruit", "La meilleure des teams", true)
	cmdToRun2 := getMainAddTeamCmd("Redpesk Team", "Poisson rouge!", false)
	cmdToRun3 := getMainAddTeamCmd("Among Us", "Bouh!", true)

	// Create the teams if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, teamsCommand, "bruit", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, teamsCommand, "redpesk-team", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, teamsCommand, "among-us", cmdToRun3...)
	if err != nil {
		test.Fatal(err)
	}

	// Set-up the answer list
	answerList := []string{"bruit", "redpesk-team", "among-us"}

	// Create the list of commands to run for test
	cmdList := [][]string{
		{teamsCommand, "get"},
		{teamsCommand, "update"},
		{teamsCommand, "invite"},
	}

	// Check the whole list of commands
	err = checkListOfCompletionCmd(test, cmdList, answerList)
	if err != nil {
		test.Fatal(err)
	}
}

// TestCompletionTeamMembersTab is the function allowing to test if
// the completion of team members list is working correctly
// in rp-cli
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestCompletionTeamMembersTab(test *testing.T) {

	// Firstly, creates a test team if it does not exists
	cmdToRun1 := getMainAddTeamCmd("Lincoln Barbers", "There are barbershops", true)
	err := addObjectIfNotExists(test, teamsCommand, "lincoln-barbers", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Then creates users if they do not exist
	cmdToRun2 := getMainAddUserCmd("tiramisu", "tiramisu@iot.bzh", "Ti", "Ramisu", "0666989945")
	cmdToRun3 := getMainAddUserCmd("kravmager", "krav@iot.bzh", "Imi", "Lichtenfeld", "0666999949")
	err = addObjectIfNotExists(test, usersCommand, "tiramisu", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, usersCommand, "kravmager", cmdToRun3...)
	if err != nil {
		test.Fatal(err)
	}

	// Invite all users in the team and check that everything is [OK]
	checkActionCmdAns(test, teamsCommand, "invite", "lincoln-barbers", "-g", "tiramisu")
	checkActionCmdAns(test, teamsCommand, "invite", "lincoln-barbers", "-g", "kravmager")

	// Set-up the completion answer list
	answerList := []string{"tiramisu", "kravmager"}

	// Create the list of commands to run for completion test
	cmdList := [][]string{
		{teamsCommand, "del-member", "lincoln-barbers", "--member"},
	}

	// Check the whole list of commands
	err = checkListOfCompletionCmd(test, cmdList, answerList)
	if err != nil {
		test.Fatal(err)
	}
}

// TestCompletionMembersToInviteTab is the function allowing to test if
// the completion of the list of users to invite in a team is working correctly
// in rp-cli
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestCompletionMembersToInviteTab(test *testing.T) {

	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddTeamCmd("Bruit", "La meilleure des teams", true)
	err := addObjectIfNotExists(test, teamsCommand, "bruit", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Then creates users if they do not exist
	cmdToRun2 := getMainAddUserCmd("tiramisu", "tiramisu@iot.bzh", "Ti", "Ramisu", "0666989945")
	cmdToRun3 := getMainAddUserCmd("kravmager", "krav@iot.bzh", "Imi", "Lichtenfeld", "0666999949")
	cmdToRun4 := getMainAddTeamCmd("Redpesk Team", "Poisson rouge!", true)
	err = addObjectIfNotExists(test, usersCommand, "tiramisu", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, usersCommand, "kravmager", cmdToRun3...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, teamsCommand, "redpesk-team", cmdToRun4...)
	if err != nil {
		test.Fatal(err)
	}

	// Set-up the answer list
	answerList := []string{"tiramisu", "kravmager", "redpesk-team"}

	// Create the list of commands to run for test
	cmdList := [][]string{
		{teamsCommand, "invite", "bruit", "-g"},
	}

	// Check the whole list of commands
	err = checkListOfCompletionCmd(test, cmdList, answerList)
	if err != nil {
		test.Fatal(err)
	}
}

// TestCompletionUsersTab is the function allowing to test if
// the completion of users list is working correctly in rp-cli
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestCompletionUsersTab(test *testing.T) {

	// Creates users if they do not exist already
	cmdToRun1 := getMainAddUserCmd("the_banano", "banano@iot.bzh", "Armand", "Bénéteau", "0666989945")
	cmdToRun2 := getMainAddUserCmd("tiramisu", "tiramisu@iot.bzh", "Ti", "Ramisu", "0666989945")
	cmdToRun3 := getMainAddUserCmd("kravmager", "krav@iot.bzh", "Imi", "Lichtenfeld", "0666999949")

	// Create the users if they do not exist, and check that all steps are OK
	err := addObjectIfNotExists(test, usersCommand, "the_banano", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, usersCommand, "tiramisu", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, usersCommand, "kravmager", cmdToRun3...)
	if err != nil {
		test.Fatal(err)
	}

	// Set-up the answer list
	answerList := []string{"tiramisu", "kravmager", "the_banano"}

	// Create the list of commands to run for test
	cmdList := [][]string{
		{usersCommand, "get"},
		{usersCommand, "update"},
		{usersCommand, "disable"},
		{usersCommand, "enable"},
	}

	// Check the whole list of commands
	err = checkListOfCompletionCmd(test, cmdList, answerList)
	if err != nil {
		test.Fatal(err)
	}
}

// TestPersistentFlagsTab is the function allowing to test if
// the completion of persistent flags list is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestPersistentFlagsTab(test *testing.T) {

	// Set-up the answer list
	answerList := []string{"config", "logfilename", "loglevel", "nonblocking", "pkgsdir", "rawoutput",
		"serveralias", "sourcesdir", "timestamp"}

	// Create the list of commands to run for test
	cmdList := [][]string{
		{miscCommand, miscConfigCmd, "get"},
		{miscCommand, miscConfigCmd, "set"},
	}

	// Check the whole list of commands
	err := checkListOfCompletionCmd(test, cmdList, answerList)
	if err != nil {
		test.Fatal(err)
	}
}

// TestCompletionBuildTypeTab is the function allowing to test if
// the completion of build type list is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestCompletionBuildTypeTab(test *testing.T) {
	// Set-up the answer list
	answerList := []string{"emulated_build", "accel_build_cross", "standard_build_cross"}

	// Create the list of commands to run for test
	cmdList := [][]string{
		{appsCommand, "add", "--build-type"},
	}

	// Check the whole list of commands
	err := checkListOfCompletionCmd(test, cmdList, answerList)
	if err != nil {
		test.Fatal(err)
	}
}

// TestCompletionSrpmRebuildOption is the function allowing to test if
// the completion of rebuild options list is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestCompletionSrpmRebuildOption(test *testing.T) {
	// Set-up the answer list
	answerList := []string{SrpmPackageFileChangeTrueShort, SrpmMetaFileChangeTrueShort, SrpmDistroBranchesManagementTrueShort,
		SrpmNoValidSrpmBuildTrueShort, SrpmPackageFileChangeFalseShort, SrpmMetaFileChangeFalseShort,
		SrpmDistroBranchesManagementFalseShort, SrpmNoValidSrpmBuildFalseShort}

	// Create the list of commands to run for test
	cmdList := [][]string{
		{appsCommand, "add", "--auto-srpm-build"},
		{appsCommand, "update", "--auto-srpm-build"},
	}

	// Check the whole list of commands
	err := checkListOfCompletionCmd(test, cmdList, answerList)
	if err != nil {
		test.Fatal(err)
	}
}

// TestCompletionGlobalSrpmRebuildOption is the function allowing to test if
// the completion of rebuild options list is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestCompletionGlobalSrpmRebuildOption(test *testing.T) {
	// Set-up the answer list
	answerList := []string{GlobalAutoSrpmRebuildTrue, GlobalAutoSrpmRebuildFalse}

	// Create the list of commands to run for test
	cmdList := [][]string{
		{appsCommand, "update", "--auto-srpm"},
	}

	// Check the whole list of commands
	err := checkListOfCompletionCmd(test, cmdList, answerList)
	if err != nil {
		test.Fatal(err)
	}
}

// TestCompletionOTABoardsTab is the function allowing to test if
// the completion of OTA boards list is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestCompletionOTABoardsTab(test *testing.T) {
	// Set-up the answer list
	answerList := []string{"ota-redpesk-lts-arz-1.0-x86_64", "ota-redpesk-lts-arz-1.0-aarch64"}

	// Create the list of commands to run for test
	cmdList := [][]string{
		{prjsCommand, prjsReleasesCmd, "deploy", "-b"},
	}

	// Check the whole list of commands
	err := checkListOfCompletionCmd(test, cmdList, answerList)
	if err != nil {
		test.Fatal(err)
	}
}

// TestCompletionAppBranchTab is the function allowing to test if
// the completion on app branches list is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestCompletionAppBranchTab(test *testing.T) {
	// Retrieve the project and app names to use for this test
	localPrj, localApp := getPrjAndAppNames()

	// Create an application and a project in order to start testing the branches
	err := createPrjAndApp(test, localPrj, "redpesk-lts-arz", localApp, localApp)
	if err != nil {
		test.Fatal(err)
	}

	// Create a new branch on this app
	_, err = createBranchOnApp(test, localPrj, localApp, "main", "first-new-branch", false, false, []string{})
	if err != nil {
		test.Fatal(err)
	}

	// Create a new branch on this app
	_, err = createBranchOnApp(test, localPrj, localApp, "main", "second-new-branch", false, false, []string{})
	if err != nil {
		test.Fatal(err)
	}

	// Set-up the answer list
	answerList := []string{"main", "first-new-branch", "second-new-branch"}

	// Create the list of commands to run for test
	cmdList := [][]string{
		{appsCommand, appsBranchesCmd, "delete", localApp, "-p", localPrj, "--branch"},
	}

	// Check the whole list of commands
	err = checkListOfCompletionCmd(test, cmdList, answerList)
	if err != nil {
		test.Fatal(err)
	}
}

// TestCompletionSpecTypeTab is the function allowing to test if
// the completion of specfile type list is working correctly
// in rp-cli
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestCompletionSpecTypeTab(test *testing.T) {

	// Set-up the answer list
	answerList := []string{"internal", "external_url", "relative_path"}

	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("SNCF", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "sncf", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Then, creates an application if it does not exists
	cmdToRun2 := getMainAddAppCmd("TGV", "train-grande-vitesse", "sncf")
	err = addObjectIfNotExists(test, appsCommand, "tgv", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the list of commands to run for test
	cmdList := [][]string{
		{appsCommand, appsAddCmd, "--specfile-type"},
		{appsCommand, "update", "tgv", "--specfile-type"},
	}

	// Check the whole list of commands
	err = checkListOfCompletionCmd(test, cmdList, answerList)
	if err != nil {
		test.Fatal(err)
	}
}

// TestCompletionAppFromLocalBuildCfg is the function allowing to test if
// the completion of the application slugs from the local configuration
// is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestCompletionFromLocalBuildCfg(test *testing.T) {
	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Firstly, set-up 3 fake local configurations
	// 1
	fakeCfg := RPLocalBuildConfig{
		ProjectSlug: "completion-prj-test",
		AppPkgName:  "completion-app-test-0-pkg",
		AppSlug:     "completion-app-test-0",
		ArchsList:   []string{"aarch64"},
		DistrosList: []string{"distro-0", "distro-1"},
	}
	if err := createFakeBuildConfig(fakeCfg); err != nil {
		test.Fatal(err)
	}
	// 2
	fakeCfg.AppPkgName = "completion-app-test-1-pkg"
	fakeCfg.AppSlug = "completion-app-test-1"
	fakeCfg.ArchsList = []string{"x86_64"}
	if err := createFakeBuildConfig(fakeCfg); err != nil {
		test.Fatal(err)
	}
	// 3
	fakeCfg.AppPkgName = "completion-app-test-2-pkg"
	fakeCfg.AppSlug = "completion-app-test-2"
	fakeCfg.DistrosList = []string{"distro-0", "distro-2"}
	if err := createFakeBuildConfig(fakeCfg); err != nil {
		test.Fatal(err)
	}

	// Set-up the answer list
	answerAppList := []string{"completion-app-test-0", "completion-app-test-1", "completion-app-test-2"}
	answerPrjList := []string{"completion-prj-test"}
	answerArchList := []string{"x86_64", "aarch64"}
	answerDistroList := []string{"distro-0", "distro-1", "distro-2"}

	// Create the list of commands to run for test
	cmdList := [][]string{
		{localCommand, localPkgCmdName, "status"},
		{localCommand, localPkgCmdName, "commit"},
		{localCommand, localPkgCmdName, "build"},
		{localCommand, localPkgCmdName, "service"},
		{localCommand, localPkgCmdName, "metadata"},
		{localCommand, localPkgCmdName, "shell"},
		{localCommand, localPkgCmdName, "clean"},
	}

	// Check the whole list of commands
	err = checkListOfCompletionCmd(test, cmdList, answerAppList)
	if err != nil {
		test.Fatal(err)
	}

	// Change the command list to run the test with project slug
	cmdList = [][]string{
		{localCommand, localPkgCmdName, "status", "-p"},
		{localCommand, localPkgCmdName, "commit", "-p"},
		{localCommand, localPkgCmdName, "build", "-p"},
		{localCommand, localPkgCmdName, "service", "-p"},
		{localCommand, localPkgCmdName, "metadata", "-p"},
		{localCommand, localPkgCmdName, "shell", "-p"},
		{localCommand, localPkgCmdName, "clean", "-p"},
	}

	// Check the whole list of commands
	err = checkListOfCompletionCmd(test, cmdList, answerPrjList)
	if err != nil {
		test.Fatal(err)
	}

	// Change the command list to run the test with architecture slug
	cmdList = [][]string{
		{localCommand, localPkgCmdName, "build", "--arch"},
		{localCommand, localPkgCmdName, "shell", "-a"},
	}

	// Check the whole list of commands
	err = checkListOfCompletionCmd(test, cmdList, answerArchList)
	if err != nil {
		test.Fatal(err)
	}

	// Change the command list to run the test with architecture slug
	cmdList = [][]string{
		{localCommand, localPkgCmdName, "build", "--distro"},
		{localCommand, localPkgCmdName, "shell", "-d"},
	}

	// Check the whole list of commands
	err = checkListOfCompletionCmd(test, cmdList, answerDistroList)
	if err != nil {
		test.Fatal(err)
	}
}

// --------------------------------------------------------------
// Utilities functions ------------------------------------------
// --------------------------------------------------------------

// runCompletionCmd is the function allowing to run a completion
// command and to return the result as a tab of strings
//
// :param args: Args to pass to the commands (sub-command, flags, etc.)
// :type args: ...string
// :returns: []string, error -- command results and an error if a problem has occurred
func runCompletionCmd(args ...string) ([]string, error) {

	results := []string{}

	// Prepare the full command -------
	fullCmd := []string{"__complete"}
	fullCmd = append(fullCmd, args...)
	fullCmd = append(fullCmd, "")
	// --------------------------------

	// Run the command ----------------
	cmdOutput, err := executeCommand(fullCmd...)
	if err != nil {
		return results, err
	}
	// --------------------------------

	// Process the output to get the results
	trimmedOutput := strings.TrimSpace(cmdOutput)
	outputLines := strings.Split(trimmedOutput, "\n")
	for _, line := range outputLines {
		// Check if the line start with ":", this is the sign
		// that the tab has been completely displayed
		if strings.HasPrefix(line, ":") {
			// End of the process
			break
		} else {
			results = append(results, line)
		}
	}
	// --------------------------------

	return results, nil
}

// checkListOfCompletionCmd is the function allowing to run a list
// of completion commands and to check if the answer contains the element
// of another list
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param cmdList: List of commands to play
// :type cmdList: [][]string
// :param answerList: List of answer to check for in the result of each command
// :type answerList: []string
// :returns: error -- An error if a problem has occurred
func checkListOfCompletionCmd(test *testing.T, cmdList [][]string, answerList []string) error {

	// Go through the commands list
	for _, command := range cmdList {
		// Run the completion command
		cmdResults, err := runCompletionCmd(command...)
		if err != nil {
			return err
		}

		// We have now the answer, check if it contains the answer we want
		for _, answer := range answerList {
			assert.Contains(test, cmdResults, answer)
		}
	}

	return nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------
