/*
Copyright © 2023 IoT.bzh Company
Contact: https://www.iot.bzh/licensing

This file is part of the redpesk-cli project, the redpesk command line tool.

$RP_BEGIN_LICENSE$
Commercial License Usage
 Licensees holding valid commercial IoT.bzh licenses may use this file in
 accordance with the commercial license agreement provided with the
 Software or, alternatively, in accordance with the terms contained in
 a written agreement between you and The IoT.bzh Company. For licensing terms
 and conditions see https://www.iot.bzh/terms-conditions. For further
 information use the contact form at https://www.iot.bzh/contact.

GNU General Public License Usage
 Alternatively, this file may be used under the terms of the GNU General
 Public license version 3. This license is as published by the Free Software
 Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
 of this file. Please review the following information to ensure the GNU
 General Public License requirements will be met
 https://www.gnu.org/licenses/gpl-3.0.html.
$RP_END_LICENSE$
*/

package cmd

import (
	"fmt"
	"net/http"
	"os"
	"strconv"
	"strings"
	"text/tabwriter"

	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/printer"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/rpconnect"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/utils"
	"github.com/spf13/cobra"
	"github.com/spf13/viper"
)

// -------------------------------------------------
// Types definition --------------------------------
// -------------------------------------------------

// ConnectionType type definition
type ConnectionType string

const (
	factoryConnection      ConnectionType = "factory"
	localbuilderConnection ConnectionType = "local-builder"
	unknownConnection      ConnectionType = "unknown"
)

// ConnectionToolsSuite type definition
type ConnectionToolsSuite struct {
	Type                    ConnectionType
	File                    string
	GlobalFlag              string
	OneObjDisplay           func(interface{}, bool, *tabwriter.Writer) error
	ListObjsDisplay         func(interface{}, bool, *tabwriter.Writer) error
	GetAvailableConnections func(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective)
}

// ConnectionListParam
type ConnectionListParam struct {
	AvailableConn []rpconnect.RPCLIConnection
	CurrentConn   string
}

// -------------------------------------------------
// -------------------------------------------------

// --------------------------------------------------------------
// Initialization function --------------------------------------
// --------------------------------------------------------------
func initConnectionsCmd(rootCmd *cobra.Command) {

	// connectionsCmd represents the connections command
	var connectionsCmd = &cobra.Command{
		Use:   connectionsCommand,
		Short: connectionsCmdShortDesc,
	}

	// listCmd represents the list command
	var listCmd = &cobra.Command{
		Use:         connectionsListCmd,
		Short:       fmt.Sprintf(connectionsListCmdShortDesc, "redpesk factories"),
		Long:        fmt.Sprintf(connectionsListCmdLongDesc, "redpesk factories"),
		RunE:        listConnections,
		Annotations: map[string]string{"__bck_connect__": "FALSE"},
	}

	// getCmd represents the get command
	var getCmd = &cobra.Command{
		Use:               connectionsGetCmd,
		Short:             connectionsGetCmdShortUsage,
		RunE:              getConnection,
		ValidArgsFunction: getRPFactoryAliases,
		Annotations:       map[string]string{"__bck_connect__": "FALSE"},
	}

	// currentCmd represents the get command
	var currentCmd = &cobra.Command{
		Use:         connectionsCurrentCmd,
		Short:       connectionsCurrentCmdShortUsage,
		RunE:        currentConnection,
		Annotations: map[string]string{"__bck_connect__": "FALSE"},
	}

	// addCmd represents the get command
	var addCmd = &cobra.Command{
		Use:         connectionsAddCmd,
		Short:       connectionsAddShortDesc,
		Long:        fmt.Sprintf(connectionsAddLongDesc, "Redpesk factories"),
		RunE:        addConnection,
		Annotations: map[string]string{"__bck_connect__": "FALSE"},
	}

	// updateCmd represents the update command
	var updateCmd = &cobra.Command{
		Use:               connectionsUpdateCmd,
		Short:             connectionsUpdateShortDesc,
		RunE:              updateConnection,
		ValidArgsFunction: getRPFactoryAliases,
		Annotations:       map[string]string{"__bck_connect__": "FALSE"},
	}

	// deleteCmd represents the delete command
	var deleteCmd = &cobra.Command{
		Use:               connectionsDeleteCmd,
		Short:             fmt.Sprintf(connectionsDeleteCmdShortUsage, "factory"),
		RunE:              deleteConnection,
		ValidArgsFunction: getRPFactoryAliases,
		Annotations:       map[string]string{"__bck_connect__": "FALSE"},
	}

	// useCmd represents the use command
	var useCmd = &cobra.Command{
		Use:               connectionsUseCmd,
		Short:             fmt.Sprintf(connectionsUseCmdShortDesc, "redpesk factory"),
		RunE:              useConnection,
		ValidArgsFunction: getRPFactoryAliases,
		Annotations:       map[string]string{"__bck_connect__": "FALSE"},
	}

	// Here you will define your flags and configuration settings.

	// Persistent Flags -----------------------------------------
	// ----------------------------------------------------------

	// Local Flags ----------------------------------------------
	listCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	getCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	currentCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	addCmd.Flags().StringP(connectionsAddURLFlag, connectionsAddURLFlagAlias, "", connectionsAddURLFlagUsage)
	addCmd.Flags().StringP(connectionsAddTokenFlag, connectionsAddTokenFlagAlias, "", connectionsAddTokenFlagUsage)
	addCmd.Flags().StringP(connectionsAddLocalBFlag, connectionsAddLocalBFlagAlias, "", connectionsAddLocalBFlagUsage)
	addCmd.Flags().StringP(connectionsAddGitpkgsDirFlag, connectionsAddGitpkgsDirFlagAlias, "", connectionsAddGitpkgsDirFlagUsage)
	addCmd.Flags().StringP(connectionsAddGitsrcsDirFlag, connectionsAddGitsrcsDirFlagAlias, "", connectionsAddGitsrcsDirFlagUsage)
	updateCmd.Flags().StringP(connectionsUpdateURLFlag, connectionsUpdateURLFlagAlias, flagUpdateKeepCurrentValue, connectionsUpdateURLFlagUsage)
	updateCmd.Flags().StringP(connectionsUpdateTokenFlag, connectionsUpdateTokenFlagAlias, flagUpdateKeepCurrentValue, connectionsUpdateTokenFlagUsage)
	updateCmd.Flags().StringP(connectionsAddLocalBFlag, connectionsAddLocalBFlagAlias, flagUpdateKeepCurrentValue, connectionsAddLocalBFlagUsage)
	updateCmd.Flags().StringP(connectionsAddGitpkgsDirFlag, connectionsAddGitpkgsDirFlagAlias, flagUpdateKeepCurrentValue, connectionsAddGitpkgsDirFlagUsage)
	updateCmd.Flags().StringP(connectionsAddGitsrcsDirFlag, connectionsAddGitsrcsDirFlagAlias, flagUpdateKeepCurrentValue, connectionsAddGitsrcsDirFlagUsage)
	// ----------------------------------------------------------

	// Add some flags as required -------------------------------
	// ----------------------------------------------------------

	// Add the group commands "connections" as subcommand of rootCmd
	rootCmd.AddCommand(connectionsCmd)
	// Add the command "list" as a subcommand of connectionsCmd
	connectionsCmd.AddCommand(listCmd)
	// Add the command "get" as a subcommand of connectionsCmd
	connectionsCmd.AddCommand(getCmd)
	// Add the command "current" as a subcommand of connectionsCmd
	connectionsCmd.AddCommand(currentCmd)
	// Add the command "add" as a subcommand of connectionsCmd
	connectionsCmd.AddCommand(addCmd)
	// Add the command "update" as a subcommand of connectionsCmd
	connectionsCmd.AddCommand(updateCmd)
	// Add the command "delete" as a subcommand of connectionsCmd
	connectionsCmd.AddCommand(deleteCmd)
	// Add the command "use" as a subcommand of connectionsCmd
	connectionsCmd.AddCommand(useCmd)
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Completion functions -----------------------------------------
// --------------------------------------------------------------

// getRPFactoryAliases gets all the factory aliases currently
// defined for rp-cli
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getRPFactoryAliases(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {

	// Get the full list of factory aliases
	factoryAliasesTab := []string{}

	// Get the list of available connections for rp-cli
	availableConnect, err := getAllConnectionsParams(accessTokenFile)
	if err != nil {
		return factoryAliasesTab, cobra.ShellCompDirectiveNoFileComp
	}

	// Fill the string slice
	for _, connection := range availableConnect {
		factoryAliasesTab = append(factoryAliasesTab, connection.Alias)
	}

	return factoryAliasesTab, cobra.ShellCompDirectiveNoFileComp
}

// getLocalBuilderAliases gets all the local builder aliases currently
// defined for rp-cli
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getLocalBuilderAliases(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {

	// Get the full list of factory aliases
	localBuilderAliasesTab := []string{}

	// Get the list of available connections for rp-cli
	availableConnect, err := getAllConnectionsParams(localBuilderFile)
	if err != nil {
		return localBuilderAliasesTab, cobra.ShellCompDirectiveNoFileComp
	}

	// Fill the string slice
	for _, connection := range availableConnect {
		localBuilderAliasesTab = append(localBuilderAliasesTab, connection.Alias)
	}

	return localBuilderAliasesTab, cobra.ShellCompDirectiveNoFileComp
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Command functions --------------------------------------------
// --------------------------------------------------------------

// listConnections gets and prints the list of all the available connections
// on the host
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func listConnections(cmd *cobra.Command, args []string) error {
	// Get the tools suite to manage the connections that is interesting us
	toolsSuite, err := getConnectionToolsSuite(cmd)
	if err != nil {
		return err
	}

	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Get the list of every available connections
	avConn, err := getAllConnectionsParams(toolsSuite.File)
	if err != nil {
		return err
	}

	// Init ConnectionListParam and retrieve the current the connection alias
	connParam := ConnectionListParam{
		AvailableConn: avConn,
		CurrentConn:   viper.GetString(toolsSuite.GlobalFlag),
	}

	// Print the result to the user
	err = RPPrinter.CustomPrintObject(connParam, verbose, toolsSuite.ListObjsDisplay)

	return err
}

// getConnection gets and prints the given connection
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func getConnection(cmd *cobra.Command, args []string) error {
	// Get the tools suite to manage the connections that is interesting us
	toolsSuite, err := getConnectionToolsSuite(cmd)
	if err != nil {
		return err
	}

	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Get the connection alias wanted by the user
	connectAlias, err := getFirstArg(args, connectionAliasNotSetMsg, connectionAliasRequiredMsg)
	if err != nil {
		return err
	}

	// Get the list of every available connections
	availableConnects, err := getAllConnectionsParams(toolsSuite.File)
	if err != nil {
		return err
	}

	// Get the connection that is interesting us
	connectFound := false
	for _, connect := range availableConnects {
		if connect.Alias == connectAlias {
			// Print the result to the user
			err = RPPrinter.CustomPrintObject(connect, verbose, toolsSuite.OneObjDisplay)
			connectFound = true
			break
		}
	}

	if !connectFound {
		err = fmt.Errorf(connectionAliasNotExistMsg, connectAlias)
	}

	return err
}

// currentConnection gets and prints the current connection
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func currentConnection(cmd *cobra.Command, args []string) error {
	// Get the tools suite to manage the connections that is interesting us
	toolsSuite, err := getConnectionToolsSuite(cmd)
	if err != nil {
		return err
	}

	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Get the alias of the remote connection
	serverAlias := viper.GetString(toolsSuite.GlobalFlag)

	// Get the connection parameters
	connecParams, err := getConnectionParams(serverAlias, toolsSuite.File)
	if err != nil {
		return err
	}

	return RPPrinter.CustomPrintObject(connecParams, verbose, toolsSuite.OneObjDisplay)
}

// addConnection adds a connection in the host by saving the token necessary
// to connect rp-cli to the Redpesk backend and the backend URL
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func addConnection(cmd *cobra.Command, args []string) error {
	// Get the tools suite to manage the connections that is interesting us
	toolsSuite, err := getConnectionToolsSuite(cmd)
	if err != nil {
		return err
	}

	// Print the first line
	RPPrinter.CustomPrintLn(connectionsAddAnsLine0)

	// Get the alias passed as argument
	alias, err := getFirstArg(args, connectionAliasNotSetMsg, connectionAliasRequiredMsg)
	if err != nil {
		return err
	}

	var serverURL, accessToken, localbuilderURL, pkgsDir, sourcesDir string
	var localbuilderIP string
	var connectParams rpconnect.RPCLIConnection
	if toolsSuite.Type == factoryConnection {
		// Get the value of url flag (mandatory)
		serverURL, err = cmd.Flags().GetString(connectionsAddURLFlag)
		if err != nil {
			logFlagError(connectionsAddURLFlag)
			return err
		}

		// Get the value of token flag (mandatory)
		accessToken, err = cmd.Flags().GetString(connectionsAddTokenFlag)
		if err != nil {
			logFlagError(connectionsAddTokenFlag)
			return err
		}

		// Get the value of the localbuilder URL flag (optional)
		localbuilderURL, err = cmd.Flags().GetString(connectionsAddLocalBFlag)
		if err != nil {
			logFlagError(connectionsAddLocalBFlag)
			return err
		}

		// Get the value of the pkgsdir flag (optional)
		pkgsDir, err = cmd.Flags().GetString(connectionsAddGitpkgsDirFlag)
		if err != nil {
			logFlagError(connectionsAddGitpkgsDirFlag)
			return err
		}

		// Get the value of the sourcesdir flag (optional)
		sourcesDir, err = cmd.Flags().GetString(connectionsAddGitsrcsDirFlag)
		if err != nil {
			logFlagError(connectionsAddGitsrcsDirFlag)
			return err
		}

		connectParams = rpconnect.RPCLIConnection{
			Alias:           alias,
			ServerURL:       serverURL,
			AccessToken:     accessToken,
			IsAdmin:         false,
			LocalBuilderURL: localbuilderURL,
			PkgsDir:         pkgsDir,
			SourcesDir:      sourcesDir,
		}

	} else {
		// Get the value of IP flag (mandatory)
		localbuilderIP, err = cmd.Flags().GetString(connectionsAddIPFlag)
		if err != nil {
			logFlagError(connectionsAddIPFlag)
			return err
		}

		// Get the value of localbuilder root directory (optional)
		localbuilderRootDir, err := cmd.Flags().GetString(connectionsLocalBldRootDirFlag)
		if err != nil {
			logFlagError(connectionsLocalBldRootDirFlag)
			return err
		}

		connectParams = rpconnect.RPCLIConnection{
			Alias:               alias,
			ServerURL:           localbuilderIP,
			LocalBuilderRootDir: localbuilderRootDir,
		}
	}

	// Tell the user we are adding the new connection
	RPPrinter.CustomPrint(connectionsAddAnsLine1)

	err = addOrUpdateConnection(connectParams, toolsSuite.File)
	if err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
	} else {
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
	}
	return err
}

// updateConnection update a connection already existing in the host
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func updateConnection(cmd *cobra.Command, args []string) error {
	// Get the tools suite to manage the connections that is interesting us
	toolsSuite, err := getConnectionToolsSuite(cmd)
	if err != nil {
		return err
	}

	// Print the first line
	RPPrinter.CustomPrintLn(connectionsUpdateAnsLine0)

	// Get the connection alias wanted by the user
	connectAlias, err := getFirstArg(args, connectionAliasNotSetMsg, connectionAliasRequiredMsg)
	if err != nil {
		return err
	}

	var serverURL, accessToken, localbuilderURL, pkgsDir, sourcesDir string
	var localbuilderIP, localbuilderRootDir string
	if toolsSuite.Type == factoryConnection {
		// Get the value of url flag (optional)
		serverURL, err = cmd.Flags().GetString(connectionsUpdateURLFlag)
		if err != nil {
			logFlagError(connectionsUpdateURLFlag)
			return err
		}

		// Get the value of token flag (optional)
		accessToken, err = cmd.Flags().GetString(connectionsUpdateTokenFlag)
		if err != nil {
			logFlagError(connectionsUpdateTokenFlag)
			return err
		}

		// Get the value of the localbuilder URL flag (optional)
		localbuilderURL, err = cmd.Flags().GetString(connectionsAddLocalBFlag)
		if err != nil {
			logFlagError(connectionsAddLocalBFlag)
			return err
		}

		// Get the value of the pkgsdir flag (optional)
		pkgsDir, err = cmd.Flags().GetString(connectionsAddGitpkgsDirFlag)
		if err != nil {
			logFlagError(connectionsAddGitpkgsDirFlag)
			return err
		}

		// Get the value of the sourcesdir flag (optional)
		sourcesDir, err = cmd.Flags().GetString(connectionsAddGitsrcsDirFlag)
		if err != nil {
			logFlagError(connectionsAddGitsrcsDirFlag)
			return err
		}
	} else {
		// Get the value of IP flag (mandatory)
		localbuilderIP, err = cmd.Flags().GetString(connectionsUpdateIPFlag)
		if err != nil {
			logFlagError(connectionsAddIPFlag)
			return err
		}

		// Get the value of localbuilder root directory (optional)
		localbuilderRootDir, err = cmd.Flags().GetString(connectionsLocalBldRootDirFlag)
		if err != nil {
			logFlagError(connectionsLocalBldRootDirFlag)
			return err
		}
	}

	// Get the list of every available connections
	availableConnects, err := getAllConnectionsParams(toolsSuite.File)
	if err != nil {
		return err
	}

	// Try to find the connection to update
	connectIndex := -1
	for index, connect := range availableConnects {
		if connect.Alias == connectAlias {
			// We found the connection to update
			connectIndex = index
			break
		}
	}

	// If we did not found the connection, return an error
	if connectIndex < 0 {
		return fmt.Errorf(connectionAliasNotExistMsg, connectAlias)
	}

	// Tell the user we are updating the connection
	RPPrinter.CustomPrintF(connectionsUpdateAnsLine1, connectAlias)

	// Now update the values
	if toolsSuite.Type == factoryConnection {
		if serverURL != flagUpdateKeepCurrentValue {
			availableConnects[connectIndex].ServerURL = serverURL
		}
		if accessToken != flagUpdateKeepCurrentValue {
			availableConnects[connectIndex].AccessToken = accessToken
		}
		if localbuilderURL != flagUpdateKeepCurrentValue {
			availableConnects[connectIndex].LocalBuilderURL = localbuilderURL
		}
		if pkgsDir != flagUpdateKeepCurrentValue {
			availableConnects[connectIndex].PkgsDir = pkgsDir
		}
		if sourcesDir != flagUpdateKeepCurrentValue {
			availableConnects[connectIndex].SourcesDir = sourcesDir
		}
	} else {
		if localbuilderIP != flagUpdateKeepCurrentValue {
			availableConnects[connectIndex].ServerURL = localbuilderIP
		}
		if localbuilderRootDir != flagUpdateKeepCurrentValue {
			availableConnects[connectIndex].LocalBuilderRootDir = localbuilderRootDir
		}
	}

	// Update the file
	err = addOrUpdateConnection(availableConnects[connectIndex], toolsSuite.File)
	if err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
	} else {
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
	}
	return err
}

// deleteConnection delete a connection existing in the host
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func deleteConnection(cmd *cobra.Command, args []string) error {
	// Get the tools suite to manage the connections that is interesting us
	toolsSuite, err := getConnectionToolsSuite(cmd)
	if err != nil {
		return err
	}

	// Print the first line
	RPPrinter.CustomPrintF(connectionsDeleteAnsLine0, string(toolsSuite.Type))

	// Get the connection alias specified by the user
	connectAlias, err := getFirstArg(args, connectionAliasNotSetMsg, connectionAliasRequiredMsg)
	if err != nil {
		return err
	}

	// Tell the user we are deleting the connection
	RPPrinter.CustomPrintF(connectionsDeleteAnsLine1, connectAlias)

	// Delete the connection in the file
	err = deleteInTokenFile(connectAlias, toolsSuite.File)
	if err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
	} else {
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
	}
	return err
}

// useConnection change the connection currently used by rp-cli
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func useConnection(cmd *cobra.Command, args []string) error {
	// Get the tools suite to manage the connections that is interesting us
	toolsSuite, err := getConnectionToolsSuite(cmd)
	if err != nil {
		return err
	}

	// Get the new alias to set for the current used connection
	newConnection, err := getFirstArg(args, connectionAliasNotSetMsg, connectionAliasRequiredMsg)
	if err != nil {
		return err
	}

	// Get the full list of available connections
	availableConnections, _ := toolsSuite.GetAvailableConnections(cmd, args, "")

	// Check that the alias given by the user exists
	doesConnectionExists := utils.ContainsString(availableConnections, newConnection)
	if !doesConnectionExists {
		return fmt.Errorf(connectionAliasNotExistMsg, newConnection)
	}

	// Update the configuration file
	return initOrUpdateDefCfgFile(toolsSuite.GlobalFlag, newConnection)
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Utilities functions ------------------------------------------
// --------------------------------------------------------------

// findOutConnectionType allows to find which kind of connections
// is concerned about the current command being run. Either factory
// connection or local-builder connection
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :returns: error -- an error if something went bad
func findOutConnectionType(cmd *cobra.Command) ConnectionType {
	// Get the grandparent command to see if the concerned connections
	// are local-builder connection or factory connections
	grandParentCmd := cmd.Parent().Parent()

	switch grandParentCmd.Use {
	case localCommand:
		return localbuilderConnection
	case applicationName:
		return factoryConnection
	default:
		return unknownConnection
	}
}

// getConnectionToolsSuite returns a structure containing the type of connection
// as well as both functions to use to display a single connection or a list of them
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :returns: *ConnectionToolsSuite, error -- a pointer to the tools suite and an error if something went bad
func getConnectionToolsSuite(cmd *cobra.Command) (*ConnectionToolsSuite, error) {
	// Initialize the return structures
	var err error
	toolsSuite := &ConnectionToolsSuite{
		Type: findOutConnectionType(cmd),
	}

	// Fill the field according to the connection type
	switch toolsSuite.Type {
	case factoryConnection:
		toolsSuite.OneObjDisplay = displayOneFactoryConnection
		toolsSuite.ListObjsDisplay = displayFactoryConnections
		toolsSuite.File = accessTokenFile
		toolsSuite.GlobalFlag = serverAliasFlag
		toolsSuite.GetAvailableConnections = getRPFactoryAliases
		err = nil
	case localbuilderConnection:
		toolsSuite.OneObjDisplay = displayOneLocalConnection
		toolsSuite.ListObjsDisplay = displayLocalConnections
		toolsSuite.File = localBuilderFile
		toolsSuite.GlobalFlag = localBuilderAliasFlag
		toolsSuite.GetAvailableConnections = getLocalBuilderAliases
		err = nil
	default:
		err = fmt.Errorf(connectionTypeUnknownMsg, cmd.Use)
	}

	return toolsSuite, err
}

// addOrUpdateConnection allows to properly add or update a connection
// in the host, it handles the cookies deletion if necessary and the
// writing in the connections file
//
// :param connection: Connection structure to use for update
// :type connection: rpconnect.RPCLIConnection
// :param file: File containing the connections to update
// :type file: string
// :returns: error -- an error if something went bad
func addOrUpdateConnection(connection rpconnect.RPCLIConnection, file string) error {
	// Get the configuration directory
	cfgDir, err := getConfigDir()
	if err != nil {
		return err
	}

	// Delete the cookie file if it exists
	if err = rpconnect.RPDelCookieFile(cfgDir, connection.Alias); err != nil {
		return err
	}

	return appendOrOverwriteInTokenFile(file, connection)
}

// appendOrOverwriteInTokenFile is a private function allowing to add in the
// token file a new set of connection parameters. If the alias already exists
// in the file, the parameters are overwritten.
//
// :param file: Filename of the connection file to update
// :type file: string
// :param params: Structure containing the set of parameters
// :type params: rpconnect.RPCLIConnection
// :returns: error -- an error if something went bad
func appendOrOverwriteInTokenFile(file string, params rpconnect.RPCLIConnection) error {
	connectionsMapsSlice := []rpconnect.RPCLIConnection{}

	// Get the configuration directory
	cfgDir, err := getConfigDir()
	if err != nil {
		return err
	}

	// Check that the file exists
	filePath := cfgDir + file
	if _, err := os.Stat(filePath); err == nil {
		// If the file exists, get the params already existing inside
		err := utils.ReadJsonFile(filePath, &connectionsMapsSlice)
		if err != nil {
			return err
		}
	}

	// Check if the given alias already exists in the file
	aliasIndex := -1
	for index, connection := range connectionsMapsSlice {
		if connection.Alias == params.Alias {
			aliasIndex = index
			break
		}
	}

	if aliasIndex > -1 {
		// The alias given already exists in the file, overwrite it!
		connectionsMapsSlice[aliasIndex].ServerURL = params.ServerURL
		connectionsMapsSlice[aliasIndex].AccessToken = params.AccessToken
		connectionsMapsSlice[aliasIndex].IsAdmin = params.IsAdmin
		connectionsMapsSlice[aliasIndex].LocalBuilderURL = params.LocalBuilderURL
		connectionsMapsSlice[aliasIndex].PkgsDir = params.PkgsDir
		connectionsMapsSlice[aliasIndex].SourcesDir = params.SourcesDir
	} else {
		// This alias does not exists, append it
		connectionsMapsSlice = append(connectionsMapsSlice, params)
	}

	// Overwrite the token file
	err = utils.WriteJsonFile(filePath, connectionsMapsSlice)

	return err
}

// deleteInTokenFile is a private function allowing to delete a connection from the
// token file on the host
//
// :param connectionsFile: File containing the connections managed
// :type connectionsFile: string
// :param aliasToDel: Structure containing the set of parameters
// :type aliasToDel: string
// :returns: error -- an error if something went bad
func deleteInTokenFile(aliasToDel, connectionsFile string) error {
	connectionsMapsSlice := []rpconnect.RPCLIConnection{}

	// Get the configuration directory
	cfgDir, err := getConfigDir()
	if err != nil {
		return err
	}

	// Check that the file exists
	filePath := cfgDir + connectionsFile
	if _, err := os.Stat(filePath); err == nil {
		// If the file exists, get the params already existing inside
		err := utils.ReadJsonFile(filePath, &connectionsMapsSlice)
		if err != nil {
			return err
		}
	}

	// Check if the given alias exists in the file
	aliasIndex := -1
	for index, connection := range connectionsMapsSlice {
		if connection.Alias == aliasToDel {
			aliasIndex = index
			break
		}
	}

	if aliasIndex < 0 {
		return fmt.Errorf(connectionAliasNotExistMsg, aliasToDel)
	}

	// Now delete the structure from the slice
	connectionsMapsSlice[aliasIndex] = connectionsMapsSlice[len(connectionsMapsSlice)-1]
	connectionsMapsSlice = connectionsMapsSlice[:len(connectionsMapsSlice)-1]

	// Finally, overwrite the token file
	return utils.WriteJsonFile(filePath, connectionsMapsSlice)
}

// quickConnectionCheck is a private function allowing to test if the connection
// specified works correctly. Only the domain name/IP can be provided, the function
// can add a port if needed.
//
// :param url: URL to check
// :type url: string
// :param port: Port on which the check needs to be done, if empty, do not specify the port
// :type port: string
// :param checkVerb: verb to call to see if the connection is good
// :type checkVerb: string
// :returns: error -- an error if something went bad
func quickConnectionCheck(url, port, checkVerb string) (bool, error) {
	fullURL := url

	// Add http prefix if missing
	if !strings.HasPrefix(url, "http://") && !strings.HasPrefix(url, "https://") {
		fullURL = "http://" + url
	}

	// Add port if missing and specified
	if port != "" && !strings.HasSuffix(fullURL, port) {
		fullURL = fullURL + ":" + port
	}

	// Now simply call the right verb
	fullURL = fullURL + "/" + checkVerb
	resp, err := http.Get(fullURL)
	if err != nil {
		return false, err
	}
	defer resp.Body.Close()

	// Check that the status is 200
	return (resp.StatusCode == 200), nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Display functions --------------------------------------------
// --------------------------------------------------------------

// displayFactoryConnections is a private function allowing to display
// a slice of factory connections in the console
//
// :param connections: tab of connections to display
// :type connections: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func displayFactoryConnections(connections interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Convert to the right type
	connects := connections.(ConnectionListParam)

	// Display result
	for index, connection := range connects.AvailableConn {
		if verbose && (index > 0) {
			fmt.Fprintln(writer)
		} else if !verbose && index == 0 {
			fmt.Fprintln(writer, connectionsListAnsHeader)
		}
		prefix := " "
		if connection.Alias == connects.CurrentConn {
			prefix = ">"
		}
		// Print this project
		_printOneConnection(connection, verbose, false, factoryConnection, writer, prefix)
	}
	return nil
}

// displayOneFactoryConnection is a private function allowing to display
// one factory connection in the console
//
// :param connection: Connection to display
// :type connection: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :returns: error -- an error if something went bad
func displayOneFactoryConnection(connection interface{}, verbose bool, writer *tabwriter.Writer) error {
	return _printOneConnection(connection, verbose, true, factoryConnection, writer, "")
}

// displayLocalConnections is a private function allowing to display
// a slice of local-builder connections in the console
//
// :param connections: tab of connections to display
// :type connections: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func displayLocalConnections(connections interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Convert to the right type
	connects := connections.(ConnectionListParam)

	// Display result
	for index, connection := range connects.AvailableConn {
		if verbose && (index > 0) {
			fmt.Fprintln(writer)
		} else if !verbose && index == 0 {
			fmt.Fprintln(writer, connectionsListAnsHeaderLocal)
		}
		prefix := " "
		if connection.Alias == connects.CurrentConn {
			prefix = ">"
		}
		// Print this project
		_printOneConnection(connection, verbose, false, localbuilderConnection, writer, prefix)
	}
	return nil
}

// displayOneLocalConnection is a private function allowing to display
// one local-builder connection in the console
//
// :param connection: Connection to display
// :type connection: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :returns: error -- an error if something went bad
func displayOneLocalConnection(connection interface{}, verbose bool, writer *tabwriter.Writer) error {
	return _printOneConnection(connection, verbose, true, localbuilderConnection, writer, "")
}

// _printOneConnection display one connection available in the host.
// It can be factory connection or a localbuilder connection
//
// :param connection: Connection to print
// :type connection: interface{}
// :param verbose: Tell if the print needs to be verbose or not
// :type verbose: bool
// :param isAlone: Tell if the object is alone or not
// :type isAlone: bool
// :param connectType: Type of connection to print
// :type connectType: ConnectionType
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func _printOneConnection(connection interface{}, verbose, isAlone bool, connectType ConnectionType, writer *tabwriter.Writer, prefix string) error {

	// Convert to the right type
	connectToPrint := connection.(rpconnect.RPCLIConnection)

	dirRootToPrint := connectToPrint.LocalBuilderRootDir
	if dirRootToPrint == "" {
		dirRootToPrint = mainNotSetMsg
	}

	if verbose {
		localbuilderURLToPrint := mainNotSetMsg
		gitpkgsToPrint := mainNotSetMsg
		gitsrcsToPrint := mainNotSetMsg

		if connectToPrint.LocalBuilderURL != "" {
			localbuilderURLToPrint = connectToPrint.LocalBuilderURL
		}
		if connectToPrint.PkgsDir != "" {
			gitpkgsToPrint = connectToPrint.PkgsDir
		}
		if connectToPrint.SourcesDir != "" {
			gitsrcsToPrint = connectToPrint.SourcesDir
		}

		serverURLField := connectionsListAnsLine1
		if connectType == localbuilderConnection {
			serverURLField = connectionsListAnsLine1Local
		}

		fmt.Fprintln(writer, connectionsListAnsLine0, connectToPrint.Alias)
		fmt.Fprintln(writer, serverURLField, connectToPrint.ServerURL)

		if connectType == localbuilderConnection {
			fmt.Fprintln(writer, connectionsListAnsLine2Local, dirRootToPrint)
		}

		// Print other fields only for factory connections
		if connectType == factoryConnection {
			fmt.Fprintln(writer, connectionsListAnsLine2, connectToPrint.AccessToken)
			fmt.Fprintln(writer, connectionsListAnsLine3, strconv.FormatBool(connectToPrint.IsAdmin))
			fmt.Fprintln(writer, connectionsListAnsLine4, localbuilderURLToPrint)
			fmt.Fprintln(writer, connectionsListAnsLine5, gitpkgsToPrint)
			fmt.Fprintln(writer, connectionsListAnsLine6, gitsrcsToPrint)
		}

	} else {
		if isAlone && connectType == factoryConnection {
			fmt.Fprintln(writer, connectionsListAnsHeader)
		} else if isAlone && connectType == localbuilderConnection {
			fmt.Fprintln(writer, connectionsListAnsHeaderLocal)
		}

		if connectType == factoryConnection {
			// Create the partial token to print
			partialToken := ""
			if len(connectToPrint.AccessToken) >= 8 {
				partialToken = connectToPrint.AccessToken[0:8] + "..."
			}
			fmt.Fprintln(writer, prefix+connectToPrint.Alias, "\t", connectToPrint.ServerURL, "\t", partialToken, "\t", strconv.FormatBool(connectToPrint.IsAdmin))
		} else if connectType == localbuilderConnection {
			fmt.Fprintln(writer, prefix+connectToPrint.Alias, "\t", connectToPrint.ServerURL, "\t", dirRootToPrint)
		}
	}

	return nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------
