/*
Copyright © 2023 IoT.bzh Company
Contact: https://www.iot.bzh/licensing

This file is part of the redpesk-cli project, the redpesk command line tool.

$RP_BEGIN_LICENSE$
Commercial License Usage
 Licensees holding valid commercial IoT.bzh licenses may use this file in
 accordance with the commercial license agreement provided with the
 Software or, alternatively, in accordance with the terms contained in
 a written agreement between you and The IoT.bzh Company. For licensing terms
 and conditions see https://www.iot.bzh/terms-conditions. For further
 information use the contact form at https://www.iot.bzh/contact.

GNU General Public License Usage
 Alternatively, this file may be used under the terms of the GNU General
 Public license version 3. This license is as published by the Free Software
 Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
 of this file. Please review the following information to ensure the GNU
 General Public License requirements will be met
 https://www.gnu.org/licenses/gpl-3.0.html.
$RP_END_LICENSE$
*/

package cmd

import (
	"testing"

	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/rpconnect"
	"github.com/stretchr/testify/assert"
)

const (
	// Connections
	ConnectDisplayNoVerboseColNb = 4
	ConnectDisplayVerboseLineNb  = 7
)

// TestConnectionsAdd is the function allowing to test if
// the command "rp-cli connections add <alias> ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestConnectionsAdd(test *testing.T) {
	// Create a new connection object for rp-cli
	connectToCreate := rpconnect.RPCLIConnection{
		Alias:     "testadd",
		ServerURL: "https://test.redpesk.bzh",
	}
	accessTokenFilePath, err := addConnectionForTest(test, &connectToCreate)

	// Delete the false file created at the beginning
	defer deleteFalseFile(accessTokenFilePath)
	if err != nil {
		test.Fatal(err)
	}
	// Check that the new value in the access token file is good
	writenConnection, err := getConnectionParams("testadd", accessTokenFile)
	if err != nil {
		test.Fatal(err)
	}
	assert.Equal(test, "testadd", writenConnection.Alias)
	assert.Equal(test, "https://test.redpesk.bzh", writenConnection.ServerURL)
	assert.Equal(test, "aaaaaaaa-bbbb-4b41-9e2e-9ced6675ea03_b336ce2c-c111-4fa8-b5f9-76a6e19724b7", writenConnection.AccessToken)
}

// TestConnectionsAddCustomLocalBConfig is the function allowing to test if
// the command "rp-cli connections add <alias> ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestConnectionsAddCustomLocalBConfig(test *testing.T) {
	// Create a new connection object for rp-cli
	connectToCreate := rpconnect.RPCLIConnection{
		Alias:           "testadd",
		ServerURL:       "https://test.redpesk.bzh",
		LocalBuilderURL: "http://rp-localbuilder:8080",
		PkgsDir:         "/home/devel/sandbox/gitpkgs",
		SourcesDir:      "/home/devel/sandbox/gitsources",
	}
	accessTokenFilePath, err := addConnectionForTest(test, &connectToCreate)
	// Delete the false file created at the beginning
	defer deleteFalseFile(accessTokenFilePath)
	if err != nil {
		test.Fatal(err)
	}
	// Check that the new value in the access token file is good
	writenConnection, err := getConnectionParams("testadd", accessTokenFile)
	if err != nil {
		test.Fatal(err)
	}
	assert.Equal(test, "testadd", writenConnection.Alias)
	assert.Equal(test, "https://test.redpesk.bzh", writenConnection.ServerURL)
	assert.Equal(test, "aaaaaaaa-bbbb-4b41-9e2e-9ced6675ea03_b336ce2c-c111-4fa8-b5f9-76a6e19724b7", writenConnection.AccessToken)
	assert.Equal(test, "http://rp-localbuilder:8080", writenConnection.LocalBuilderURL)
	assert.Equal(test, "/home/devel/sandbox/gitpkgs", writenConnection.PkgsDir)
	assert.Equal(test, "/home/devel/sandbox/gitsources", writenConnection.SourcesDir)
}

// TestConnectionsGet is the function allowing to test if
// the command "rp-cli connections get <alias>..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestConnectionsGet(test *testing.T) {
	// Create a new connection object for rp-cli
	connectToCreate := rpconnect.RPCLIConnection{
		Alias:     "testget",
		ServerURL: "https://test.redpesk.bzh",
	}
	accessTokenFilePath, err := addConnectionForTest(test, &connectToCreate)
	// Delete the false file created at the beginning
	defer deleteFalseFile(accessTokenFilePath)
	if err != nil {
		test.Fatal(err)
	}

	// Check the get return for this command
	// Not verbose mode
	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"Alias":       `^testget$`,
		"URL":         `^http(s)*://`,
		"Token_start": `[a-f0-9]{8}...`,
		"Admin":       `^false$`,
	}

	// Run the command and the tests (no verbose and verbose, since this is the same result)
	checkCmdAnsKeysRegexpCol(test, ConnectDisplayNoVerboseColNb, regexpMap, connectionsCommand, "get", "testget")

	// Verbose mode
	// Create the map containing the regexp we want to match in the answer
	regexpMap2 := map[string]string{
		"Alias":                       `^testget$`,
		"Factory URL":                 `^http(s)*://`,
		"Access token:":               `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}_[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Admin":                       `^false$`,
		"Custom localbuilder URL":     `^\[Not set\]$`,
		"Custom gitpkgs directory":    `^\[Not set\]$`,
		"Custom gitsources directory": `^\[Not set\]$`,
	}

	// Run the command and the tests
	checkCmdAnswerListKeysRegexp(test, ConnectDisplayVerboseLineNb, "", regexpMap2, connectionsCommand, "get", "testget", "-v")
}

// TestConnectionsUseAndCurrent is the function allowing to test if
// the command "rp-cli connections use" and
// "rp-cli connections current" are working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestConnectionsUseAndCurrent(test *testing.T) {
	// Create a new connection object for rp-cli
	connectToCreate := rpconnect.RPCLIConnection{
		Alias:     "testcurrent0",
		ServerURL: "https://test0.redpesk.bzh",
	}
	accessTokenFilePath, err := addConnectionForTest(test, &connectToCreate)
	// Delete the false file created at the beginning
	defer deleteFalseFile(accessTokenFilePath)
	if err != nil {
		test.Fatal(err)
	}

	// Create another new connection
	connectToCreate1 := rpconnect.RPCLIConnection{
		Alias:     "testcurrent1",
		ServerURL: "https://test1.redpesk.bzh",
	}
	_, err = addConnectionForTest(test, &connectToCreate1)
	if err != nil {
		test.Fatal(err)
	}

	// Request rp-cli to use a connection that we just add
	_, err = executeCommand(connectionsCommand, "use", "testcurrent0")
	if err != nil {
		test.Fatal(err)
	}

	// Check that the current connection returned is "testcurrent0"
	// Not verbose mode
	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"Alias":       `^testcurrent0$`,
		"URL":         `^https://test0.redpesk.bzh$`,
		"Token_start": `[a-f0-9]{8}...`,
		"Admin":       `^false$`,
	}

	// Run the command and the tests (no verbose and verbose, since this is the same result)
	checkCmdAnsKeysRegexpCol(test, ConnectDisplayNoVerboseColNb, regexpMap, connectionsCommand, "current")

	// Verbose mode
	// Create the map containing the regexp we want to match in the answer
	regexpMap2 := map[string]string{
		"Alias":                       `^testcurrent0$`,
		"Factory URL":                 `^https://test0.redpesk.bzh$`,
		"Access token:":               `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}_[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Admin":                       `^false$`,
		"Custom localbuilder URL":     `^\[Not set\]$`,
		"Custom gitpkgs directory":    `^\[Not set\]$`,
		"Custom gitsources directory": `^\[Not set\]$`,
	}

	// Run the command and the tests
	checkCmdAnswerListKeysRegexp(test, ConnectDisplayVerboseLineNb, "", regexpMap2, connectionsCommand, "current", "-v")
}

// TestConnectionsList is the function allowing to test if
// the command "rp-cli connections list..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestConnectionsList(test *testing.T) {
	// Create a new connection object for rp-cli
	connectToCreate := rpconnect.RPCLIConnection{
		Alias:     "testlist0",
		ServerURL: "https://test0.redpesk.bzh",
	}
	accessTokenFilePath, err := addConnectionForTest(test, &connectToCreate)
	// Delete the false file created at the beginning
	defer deleteFalseFile(accessTokenFilePath)
	if err != nil {
		test.Fatal(err)
	}

	// Create another new connection
	connectToCreate.Alias = "testlist1"
	connectToCreate.ServerURL = "https://test1.redpesk.bzh"
	_, err = addConnectionForTest(test, &connectToCreate)
	if err != nil {
		test.Fatal(err)
	}

	// Create another new connection
	connectToCreate.Alias = "testlist2"
	connectToCreate.ServerURL = "https://test2.redpesk.bzh"
	_, err = addConnectionForTest(test, &connectToCreate)
	if err != nil {
		test.Fatal(err)
	}

	// Create another new connection
	connectToCreate.Alias = "testlist3"
	connectToCreate.ServerURL = "https://test3.redpesk.bzh"
	_, err = addConnectionForTest(test, &connectToCreate)
	if err != nil {
		test.Fatal(err)
	}

	// Check now that the connections created are correctly listed

	// Not verbose mode
	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"Alias":       `^testlist[0-9]{1}$`,
		"URL":         `^https://test[0-9]{1}.redpesk.bzh$`,
		"Token_start": `[a-f0-9]{8}...`,
		"Admin":       `^false$`,
	}

	checkCmdAnsKeysRegexpCol(test, ConnectDisplayNoVerboseColNb, regexpMap, connectionsCommand, "list")

	// Verbose mode
	// Create the map containing the regexp we want to match in the answer
	regexpMap2 := map[string]string{
		"Alias":                       `^testlist[0-9]{1}$`,
		"Factory URL":                 `^https://test[0-9]{1}.redpesk.bzh$`,
		"Access token:":               `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}_[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Admin":                       `^false$`,
		"Custom localbuilder URL":     `^\[Not set\]$`,
		"Custom gitpkgs directory":    `^\[Not set\]$`,
		"Custom gitsources directory": `^\[Not set\]$`,
	}

	checkCmdAnswerListKeysRegexp(test, ConnectDisplayVerboseLineNb, "", regexpMap2, connectionsCommand, "list", "-v")
}

// TestConnectionsUpdateURL is the function allowing to test if
// the command "rp-cli connections update <alias>..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestConnectionsUpdateURL(test *testing.T) {
	// Create a new connection object for rp-cli
	connectToCreate := rpconnect.RPCLIConnection{
		Alias:     "testupdate",
		ServerURL: "https://test.redpesk.bzh",
	}
	accessTokenFilePath, err := addConnectionForTest(test, &connectToCreate)
	// Delete the false file created at the beginning
	defer deleteFalseFile(accessTokenFilePath)
	if err != nil {
		test.Fatal(err)
	}

	// Do a first check before updating the connection
	regexpMap := map[string]string{
		"Alias":       `^testupdate$`,
		"URL":         `^https://test.redpesk.bzh$`,
		"Token_start": `[a-f0-9]{8}...`,
		"Admin":       `^false$`,
	}

	// Run the command and the tests
	checkCmdAnsKeysRegexpCol(test, ConnectDisplayNoVerboseColNb, regexpMap, connectionsCommand, "get", "testupdate")

	// Update the connection
	_, err = executeCommand(connectionsCommand, "update", "testupdate", "-u", "https://update.redpesk.bzh")
	if err != nil {
		test.Fatal(err)
	}

	// Change the map we use
	regexpMap["URL"] = `^https://update.redpesk.bzh$`

	// Run the command and the tests
	checkCmdAnsKeysRegexpCol(test, ConnectDisplayNoVerboseColNb, regexpMap, connectionsCommand, "get", "testupdate")
}

// TestConnectionsUpdateToken is the function allowing to test if
// the command "rp-cli connections update <alias>..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestConnectionsUpdateToken(test *testing.T) {
	// Create a new connection object for rp-cli
	connectToCreate := rpconnect.RPCLIConnection{
		Alias:     "testupdatetoken",
		ServerURL: "https://test.redpesk.bzh",
	}
	accessTokenFilePath, err := addConnectionForTest(test, &connectToCreate)
	// Delete the false file created at the beginning
	defer deleteFalseFile(accessTokenFilePath)
	if err != nil {
		test.Fatal(err)
	}

	// Do a first check before updating the connection
	regexpMap := map[string]string{
		"Alias":                       `^testupdatetoken$`,
		"Factory URL":                 `^https://test.redpesk.bzh$`,
		"Access token:":               `^aaaaaaaa-bbbb-4b41-9e2e-9ced6675ea03_b336ce2c-c111-4fa8-b5f9-76a6e19724b7$`,
		"Admin":                       `^false$`,
		"Custom localbuilder URL":     `^\[Not set\]$`,
		"Custom gitpkgs directory":    `^\[Not set\]$`,
		"Custom gitsources directory": `^\[Not set\]$`,
	}

	// Run the command and the tests
	checkCmdAnswerListKeysRegexp(test, ConnectDisplayVerboseLineNb, "", regexpMap, connectionsCommand, "get", "testupdatetoken", "-v")

	// Update the connection
	_, err = executeCommand(connectionsCommand, "update", "testupdatetoken", "--token", "aaaaaaaa-bbbb-4b41-9e2e-9ced6675ea03_b336ce2c-c111-aaaa-bbbb-76a6e19724b7")
	if err != nil {
		test.Fatal(err)
	}

	// Change the map we use
	regexpMap["Access token:"] = `^aaaaaaaa-bbbb-4b41-9e2e-9ced6675ea03_b336ce2c-c111-aaaa-bbbb-76a6e19724b7$`

	// Run the command and the tests
	checkCmdAnswerListKeysRegexp(test, ConnectDisplayVerboseLineNb, "", regexpMap, connectionsCommand, "get", "testupdatetoken", "-v")
}

// TestConnectionsUpdateCustomLocalbuilder is the function allowing to test if
// the command "rp-cli connections update <alias>..." is working correctly
// when the settings for a custom localbuilder are activated
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestConnectionsUpdateCustomLocalbuilder(test *testing.T) {
	// Create a new connection object for rp-cli
	connectToCreate := rpconnect.RPCLIConnection{
		Alias:     "testupdatecustomlb",
		ServerURL: "https://test.redpesk.bzh",
	}
	accessTokenFilePath, err := addConnectionForTest(test, &connectToCreate)
	// Delete the false file created at the beginning
	defer deleteFalseFile(accessTokenFilePath)
	if err != nil {
		test.Fatal(err)
	}

	// Do a first check before updating the connection
	regexpMap := map[string]string{
		"Alias":                       `^testupdatecustomlb$`,
		"Factory URL":                 `^https://test.redpesk.bzh$`,
		"Access token:":               `^aaaaaaaa-bbbb-4b41-9e2e-9ced6675ea03_b336ce2c-c111-4fa8-b5f9-76a6e19724b7$`,
		"Admin":                       `^false$`,
		"Custom localbuilder URL":     `^\[Not set\]$`,
		"Custom gitpkgs directory":    `^\[Not set\]$`,
		"Custom gitsources directory": `^\[Not set\]$`,
	}

	// Run the command and the tests
	checkCmdAnswerListKeysRegexp(test, ConnectDisplayVerboseLineNb, "", regexpMap, connectionsCommand, "get", "testupdatecustomlb", "-v")

	// Update the connection
	_, err = executeCommand(connectionsCommand, "update", "testupdatecustomlb", "-l", "https://test.redpesk.bzh", "-p", "/home/devel/sandbox/gitpkgs", "-s", "/home/devel/sandbox/gitsources")
	if err != nil {
		test.Fatal(err)
	}

	// Change the map we use
	regexpMap["Custom localbuilder URL"] = `^https://test.redpesk.bzh$`
	regexpMap["Custom gitpkgs directory"] = `^/home/devel/sandbox/gitpkgs$`
	regexpMap["Custom gitsources directory"] = `^/home/devel/sandbox/gitsources$`

	// Run the command and the tests
	checkCmdAnswerListKeysRegexp(test, ConnectDisplayVerboseLineNb, "", regexpMap, connectionsCommand, "get", "testupdatecustomlb", "-v")
}

// TestConnectionsDelete is the function allowing to test if
// the command "rp-cli connections delete <alias>" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestConnectionsDelete(test *testing.T) {
	// Create a new connection object for rp-cli
	connectToCreate := rpconnect.RPCLIConnection{
		Alias:     "testdelete0",
		ServerURL: "https://test0.redpesk.bzh",
	}
	accessTokenFilePath, err := addConnectionForTest(test, &connectToCreate)
	// Delete the false file created at the beginning
	defer deleteFalseFile(accessTokenFilePath)
	if err != nil {
		test.Fatal(err)
	}

	// Create another new connection
	connectToCreate.Alias = "testdelete1"
	connectToCreate.ServerURL = "https://test1.redpesk.bzh"
	_, err = addConnectionForTest(test, &connectToCreate)
	if err != nil {
		test.Fatal(err)
	}

	// Request rp-cli to delete a connection that we just add
	_, err = executeCommand(connectionsCommand, "delete", "testdelete0")
	if err != nil {
		test.Fatal(err)
	}

	// Check that the connection deleted does not exist anymore
	_, err = executeCommand(connectionsCommand, "get", "testdelete0")
	assert.NotNil(test, err)
	assert.Contains(test, err.Error(), "no connection has been found with the identifier \"testdelete0\"")
}

// --------------------------------------------------------------
// Utilities functions ------------------------------------------
// --------------------------------------------------------------

// addConnectionForTest is a utility function allowing to create a
// new connection and to create a fake token file if it does not
// exist already.
// It returns a path to the fake token file created. The calling
// function needs to call 'defer deleteFalseFile(fakeFilePath)'
// in order to delete it at the end of tests
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param connection: Connection to add to rp-cli
// :type connection: *rpconnect.RPCLIConnection
// :returns: string, error -- the path to the fake token file and an error if something bad happened
func addConnectionForTest(test *testing.T, connection *rpconnect.RPCLIConnection) (string, error) {
	// Check if the token file is already existing, and create it if it is not the case
	tokenFileExists, tokenFilePath, err := fakeFileExists(accessTokenFile)
	if err != nil {
		return tokenFilePath, err
	}
	if !tokenFileExists {
		tokenFilePath, err = createFalseTokenFile([]string{})
		if err != nil {
			return tokenFilePath, err
		}
	}

	// Now add the connection object if it does not already exists
	addCmd := []string{connectionsCommand, "add", connection.Alias, "-u", connection.ServerURL, "--token", "aaaaaaaa-bbbb-4b41-9e2e-9ced6675ea03_b336ce2c-c111-4fa8-b5f9-76a6e19724b7"}

	// Now add the optional arguments if given
	if connection.LocalBuilderURL != "" {
		addCmd = append(addCmd, "-l", connection.LocalBuilderURL)
	}
	if connection.PkgsDir != "" {
		addCmd = append(addCmd, "-p", connection.PkgsDir)
	}
	if connection.SourcesDir != "" {
		addCmd = append(addCmd, "-s", connection.SourcesDir)
	}

	err = addObjectIfNotExists(test, connectionsCommand, connection.Alias, addCmd...)
	if err != nil {
		return tokenFilePath, err
	}

	return tokenFilePath, nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------
