/*
Copyright © 2021-2023 IoT.bzh Company
Contact: https://www.iot.bzh/licensing

This file is part of the redpesk-cli project, the redpesk command line tool.

$RP_BEGIN_LICENSE$
Commercial License Usage
 Licensees holding valid commercial IoT.bzh licenses may use this file in
 accordance with the commercial license agreement provided with the
 Software or, alternatively, in accordance with the terms contained in
 a written agreement between you and The IoT.bzh Company. For licensing terms
 and conditions see https://www.iot.bzh/terms-conditions. For further
 information use the contact form at https://www.iot.bzh/contact.

GNU General Public License Usage
 Alternatively, this file may be used under the terms of the GNU General
 Public license version 3. This license is as published by the Free Software
 Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
 of this file. Please review the following information to ensure the GNU
 General Public License requirements will be met
 https://www.gnu.org/licenses/gpl-3.0.html.
$RP_END_LICENSE$
*/
package cmd

import (
	"regexp"
	"strings"
	"testing"

	"github.com/gofrs/uuid"
)

const (
	// Deployments
	deploymentDisplayNoVerboseColNb    = 5
	deploymentDisplayVerboseFullLineNb = 18
)

// TestRelProjectDeployment is the function allowing to test if
// the command "rp-cli projects release deploy <prjID>..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestRelProjectDeployment(test *testing.T) {
	// Run the test
	createReleaseAndDeploy(test, "Project to deploy 0", []string{"ota-redpesk-lts-arz-1.0-x86_64"})
}

// TestDeploymentList is the function allowing to test if
// the command "rp-cli deployments list" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestDeploymentList(test *testing.T) {
	// Create the base for what we need, and do a first deployment
	createReleaseAndDeploy(test, "Project to deploy 1", []string{"ota-redpesk-lts-arz-1.0-x86_64"})
	// Do another deployment
	deployReleasedProjectforTest(test, "Project to deploy 1", []string{"ota-redpesk-lts-arz-1.0-aarch64"})
	// And another one
	deployReleasedProjectforTest(test, "Project to deploy 1", []string{"ota-redpesk-lts-arz-1.0-x86_64"})

	// Create another base project and do another deployment
	createReleaseAndDeploy(test, "Project to deploy 2", []string{"ota-redpesk-lts-arz-1.0-aarch64"})

	// Now test the "not verbose" listing
	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID":         `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Project":    `^project-to-deploy-[012]-1.0$`,
		"Status":     `^finished$`,
		"Start_time": `^[0-9]{4}(-[0-9]{2}){2} [0-9]{2}(:[0-9]{2}){2}.[0-9]+ \+[0-9]{4} (UTC|CEST|CET)$`,
		"Name":       `^\[Not set\]$`,
	}

	// Run the command and the tests (no verbose)
	checkCmdAnsKeysRegexpCol(test, deploymentDisplayNoVerboseColNb, regexpMap, deploymentsCommand, deploymentsListCmd)

	// Verbose mode
	// Create the map containing the regexp we want to match in the answer
	regexpMap2 := map[string]string{
		"ID":               `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Deployed project": `^project-to-deploy-[012]-1.0$`,
		"Name":             `^\[Not set\]$`,
		"Creation time":    `^[0-9]{4}(-[0-9]{2}){2} [0-9]{2}(:[0-9]{2}){2}.[0-9]+ \+[0-9]{4} (UTC|CEST|CET)$`,
		"Update time":      `^[0-9]{4}(-[0-9]{2}){2} [0-9]{2}(:[0-9]{2}){2}.[0-9]+ \+[0-9]{4} (UTC|CEST|CET)$`,
		"Status":           `^finished$`,
		"Start time":       `^[0-9]{4}(-[0-9]{2}){2} [0-9]{2}(:[0-9]{2}){2}.[0-9]+ \+[0-9]{4} (UTC|CEST|CET)$`,
		"End time":         `^[0-9]{4}(-[0-9]{2}){2} [0-9]{2}(:[0-9]{2}){2}.[0-9]+ \+[0-9]{4} (UTC|CEST|CET)$`,
		"Creator":          `^the_banano_$`,
	}

	// Run the command and the tests
	checkCmdAnswerListKeysRegexp(test, deploymentDisplayVerboseLightLineNb, "", regexpMap2, deploymentsCommand, deploymentsListCmd, "-v")
}

// TestDeploymentFilteredList is the function allowing to test if
// the command "rp-cli deployments list" is working correctly when
// a released project is specified with the '-p' option
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestDeploymentFilteredList(test *testing.T) {
	// Create the base for what we need, and do a first deployment
	createReleaseAndDeploy(test, "Project to deploy 3", []string{"ota-redpesk-lts-arz-1.0-x86_64"})
	// Do another deployment
	deployReleasedProjectforTest(test, "Project to deploy 3", []string{"ota-redpesk-lts-arz-1.0-aarch64"})
	// And another one
	deployReleasedProjectforTest(test, "Project to deploy 3", []string{"ota-redpesk-lts-arz-1.0-x86_64"})

	// Create another base project and do another deployment
	createReleaseAndDeploy(test, "Project to deploy 4", []string{"ota-redpesk-lts-arz-1.0-aarch64"})

	// Now test the "not verbose" listing
	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID":         `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Project":    `^project-to-deploy-4-1.0$`,
		"Status":     `^finished$`,
		"Start_time": `^[0-9]{4}(-[0-9]{2}){2} [0-9]{2}(:[0-9]{2}){2}.[0-9]+ \+[0-9]{4} (UTC|CEST|CET)$`,
		"Name":       `^\[Not set\]$`,
	}

	// Run the command and the tests (no verbose)
	checkCmdAnsKeysRegexpCol(test, deploymentDisplayNoVerboseColNb, regexpMap, deploymentsCommand, deploymentsListCmd, "-p", "project-to-deploy-4-1.0")
}

// TestDeploymentGet is the function allowing to test if
// the command "rp-cli deployments get" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestDeploymentGet(test *testing.T) {
	// Create the base for what we need, and do a first deployment
	createReleaseAndDeploy(test, "Project to deploy 5", []string{"ota-redpesk-lts-arz-1.0-x86_64"})

	// Retrieve the uuid of our deployment
	dplUUID := getFirstUUIDListed(test, deploymentsCommand, deploymentsListCmd, "-p", "project-to-deploy-5-1.0")

	// Check the content of the "not verbose" output
	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID":         `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Project":    `^project-to-deploy-5-1.0$`,
		"Status":     `^finished$`,
		"Start_time": `^[0-9]{4}(-[0-9]{2}){2} [0-9]{2}(:[0-9]{2}){2}.[0-9]+ \+[0-9]{4} (UTC|CEST|CET)$`,
		"Name":       `^\[Not set\]$`,
	}

	// Run the command and the tests (no verbose)
	checkCmdAnsKeysRegexpCol(test, deploymentDisplayNoVerboseColNb, regexpMap, deploymentsCommand, "get", dplUUID.String())

	// Now, the verbose mode
	// Create the map containing the regexp we want to match in the answer
	regexpMap2 := map[string]string{
		"ID":                             `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Deployed project":               `^project-to-deploy-5-1.0$`,
		"Name":                           `^\[Not set\]$`,
		"Creation time":                  `^[0-9]{4}(-[0-9]{2}){2} [0-9]{2}(:[0-9]{2}){2}.[0-9]+ \+[0-9]{4} (UTC|CEST|CET)$`,
		"Update time":                    `^[0-9]{4}(-[0-9]{2}){2} [0-9]{2}(:[0-9]{2}){2}.[0-9]+ \+[0-9]{4} (UTC|CEST|CET)$`,
		"Status":                         `^finished$`,
		"Start time":                     `^[0-9]{4}(-[0-9]{2}){2} [0-9]{2}(:[0-9]{2}){2}.[0-9]+ \+[0-9]{4} (UTC|CEST|CET)$`,
		"End time":                       `^[0-9]{4}(-[0-9]{2}){2} [0-9]{2}(:[0-9]{2}){2}.[0-9]+ \+[0-9]{4} (UTC|CEST|CET)$`,
		"Creator":                        `^the_banano_$`,
		"[Boards status]":                `.*`,
		"ota-redpesk-lts-arz-1.0-x86_64": `^success$`,
		"[Released apps]":                `.*`,
		"Package name":                   `[a-z0-9-]+`,
		"SHA Gitpkg":                     `[a-f0-9]{40}`,
		"Branch":                         `[a-z0-9-_\.]+`,
		"Build ID":                       `[0-9]+`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, deploymentDisplayVerboseFullLineNb, regexpMap2, deploymentsCommand, "get", dplUUID.String(), "-v")
}

// --------------------------------------------------------------
// Utilities functions ------------------------------------------
// --------------------------------------------------------------

// createReleaseAndDeploy is a function allowing to create, release a project easily.
// This project is deployed on the given boards
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param prjName: Name of the reference project
// :type prjName: string
// :param boardsTab: Tab of boards wanted to deploy the released project on
// :type boardsTab: []string
func createReleaseAndDeploy(test *testing.T, prjName string, boardsTab []string) {
	// Firstly, create and release the project
	createAndReleaseProject(test, prjName, []string{}, []string{}, false)

	// Release the project
	deployReleasedProjectforTest(test, prjName, boardsTab)
}

// deployReleasedProjectforTest is a function allowing to release an already released project.
// This project is deployed on the given boards
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param prjName: Name of the released project
// :type prjName: string
// :param boardsTab: Tab of boards wanted to deploy the released project on
// :type boardsTab: []string
func deployReleasedProjectforTest(test *testing.T, prjName string, boardsTab []string) {
	// From the name, create the slug
	prjSlug := strings.ReplaceAll(strings.ToLower(prjName), " ", "-")

	// Then create the command line allowing to deploy the project
	cmdLine := []string{prjsCommand, prjsReleasesCmd, "deploy", (prjSlug + "-1.0")}

	// Now add the boards to this command line
	for _, brd := range boardsTab {
		cmdLine = append(cmdLine, "-b", brd)
	}

	// Try to deploy the project
	checkActionCmdAns(test, cmdLine...)
}

// getFirstUUIDListed is a function allowing to run a command and get the
// first UUID that appears in the output
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param args: Command line to run to get an answer
// :type args: ...string
func getFirstUUIDListed(test *testing.T, args ...string) uuid.UUID {
	uuidResult := uuid.Nil

	// First, let's run the wanted command
	cmdOutput, err := executeCommand(args...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the regular expression for an UUID
	uuidRE := regexp.MustCompile(`[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`)

	// Get the first match
	match := uuidRE.FindString(cmdOutput)

	if match != "" {
		uuidResult = uuid.FromStringOrNil(match)
	}

	return uuidResult
}

// --------------------------------------------------------------
// --------------------------------------------------------------
