/*
Copyright © 2024 IoT.bzh Company
Contact: https://www.iot.bzh/licensing

This file is part of the redpesk-cli project, the redpesk command line tool.

$RP_BEGIN_LICENSE$
# Commercial License Usage

Licensees holding valid commercial IoT.bzh licenses may use this file in
accordance with the commercial license agreement provided with the
Software or, alternatively, in accordance with the terms contained in
a written agreement between you and The IoT.bzh Company. For licensing terms
and conditions see https://www.iot.bzh/terms-conditions. For further
information use the contact form at https://www.iot.bzh/contact.

# GNU General Public License Usage

Alternatively, this file may be used under the terms of the GNU General
Public license version 3. This license is as published by the Free Software
Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
of this file. Please review the following information to ensure the GNU
General Public License requirements will be met
https://www.gnu.org/licenses/gpl-3.0.html.

$RP_END_LICENSE$
*/
package cmd

import (
	"errors"
	"fmt"
	"strings"
	"text/tabwriter"

	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/printer"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/rpxxxapi/bckapi"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/utils"
	"git.ovh.iot/redpesk/redpesk-infra/rp-webserver/actions/rpapiv1"
	"github.com/spf13/cobra"
	"github.com/spf13/viper"
)

func initDistributionsCmd(rootCmd *cobra.Command) {

	// distroCmd represents the "distributions" command
	var distroCmd = &cobra.Command{
		Use:     distCmd,
		Aliases: []string{distCmdAlias},
		Short:   distCmdShortDesc,
	}

	// listDistrosCmd represents the "distributions list" command
	var listDistrosCmd = &cobra.Command{
		Use:     distListCmd,
		Aliases: []string{distListCmdAlias},
		Short:   distListCmdShortDesc,
		RunE:    listDistributions,
	}

	// getDistroCmd represents the "distributions get" command
	var getDistroCmd = &cobra.Command{
		Use:               distGetCmd,
		Short:             distGetCmdShortDesc,
		RunE:              getDistribution,
		ValidArgsFunction: getListExtPrjsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// pkgsDistrosCmd represents the "distributions packages" command
	var pkgsDistrosCmd = &cobra.Command{
		Use:               distPkgsCmd,
		Aliases:           []string{distPkgsCmdAlias},
		Short:             distPkgsCmdShortDesc,
		RunE:              packagesDistributions,
		ValidArgsFunction: getListExtPrjsAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// ADMIN COMMANDS FROM HERE ---------------------------

	// addDistroCmd represents the "distributions add" command
	var addDistroCmd = &cobra.Command{
		Use:     distAddCmd,
		Aliases: []string{distAddCmdAlias},
		Hidden:  !isUserAdmin,
		Short:   distAddCmdShortDesc,
		Long:    distAddCmdLongDesc,
		PreRunE: initEventProject,
		RunE:    addDistribution,
		Annotations: map[string]string{
			"__bck_ws_connect__": "TRUE"},
	}

	// updateDistroCmd represents the "distributions update" command
	var updateDistroCmd = &cobra.Command{
		Use:               distUpdateCmd,
		Hidden:            !isUserAdmin,
		Short:             distUpdateCmdShortDesc,
		Long:              distUpdateCmdLongDesc,
		PreRunE:           initEventProject,
		RunE:              updateDistribution,
		ValidArgsFunction: getListExtPrjsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// Here you will define your flags and configuration settings

	// Persistent Flags -----------------------------------------
	// ----------------------------------------------------------

	// Local Flags ----------------------------------------------
	listDistrosCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	pkgsDistrosCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	pkgsDistrosCmd.Flags().StringP(distArchFlag, distArchFlagAlias, "", distArchFlagUsage)
	pkgsDistrosCmd.Flags().StringP(distNameFlag, distNameFlagAlias, "", distNameFlagUsage)
	pkgsDistrosCmd.Flags().BoolP(distWordFlag, distWordFlagAlias, false, distWordFlagUsage)
	getDistroCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	addDistroCmd.Flags().StringP(distAddCmdVersionFlag, distAddCmdVersionFlagAlias, "", distAddCmdVersionFlagUsage)
	addDistroCmd.Flags().String(distSubVersionFlag, "", distSubVersionFlagUsage)
	addDistroCmd.Flags().StringP(distAddCmdTagFlag, distAddCmdTagFlagAlias, "", distAddCmdTagFlagUsage)
	addDistroCmd.Flags().StringP(distDescriptionFlag, distDescriptionFlagAlias, "", distDescriptionFlagUsage)
	updateDistroCmd.Flags().String(distSubVersionFlag, flagUpdateKeepCurrentValue, distSubVersionFlagUsage)
	updateDistroCmd.Flags().BoolP(distUpdateCmdRefreshFlag, distUpdateCmdRefreshFlagAlias, false, distUpdateCmdRefreshFlagUsage)
	updateDistroCmd.Flags().StringP(distDescriptionFlag, distDescriptionFlagAlias, flagUpdateKeepCurrentValue, distDescriptionFlagUsage)
	// ----------------------------------------------------------

	// Add some flags as required
	pkgsDistrosCmd.MarkFlagRequired(distArchFlag)

	// Add some flags as hidden
	// N.A

	// Specify dynamic flag completion
	pkgsDistrosCmd.RegisterFlagCompletionFunc(distArchFlag, getListArchAsTab)

	// Distributions subcommands
	// Add the group commands "distributions" as subcommand of rootCmd
	rootCmd.AddCommand(distroCmd)
	// Add the command "list" as a subcommand of distroCmd
	distroCmd.AddCommand(listDistrosCmd)
	// Add the command "get" as a subcommand of distroCmd
	distroCmd.AddCommand(getDistroCmd)
	// Add the command "packages" as a subcommand of distroCmd
	distroCmd.AddCommand(pkgsDistrosCmd)
	// Add the command "add" as a subcommand of distroCmd
	distroCmd.AddCommand(addDistroCmd)
	// Add the command "update" as a subcommand of distroCmd
	distroCmd.AddCommand(updateDistroCmd)
	// ---------------------------------------------------------
}

// --------------------------------------------------------------
// Completion functions -----------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Command functions --------------------------------------------
// --------------------------------------------------------------

// listDistributions gets the list of all the available distributions on
// the Redpesk backend
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func listDistributions(cmd *cobra.Command, args []string) error {

	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Prepare the function needed by the printer to get the distributions
	getDistrosCB := func(pageNb, perPage int) (interface{}, printer.RPPrinterPaging, error) {
		// Get distributions list
		distros := bckapi.RPCliDistributions{}
		if err := bckAPI.RPDistrosGetByPage(&distros, pageNb, perPage); err != nil {
			return distros, printer.RPPrinterPaging{}, err
		}
		// Get the paging interface from the structure retrieved
		pagingStruct, _ := printer.GetPagingFromInterface(&distros)

		return distros, pagingStruct, nil
	}

	// Prepare the structure to manage paging
	pagingStruct := printer.RPPagingToolsStruct{
		Verbose:       verbose,
		VerboseLineNb: distDisplayVerboseLineNb,
		GetObjCB:      getDistrosCB,
		PrintFancyCB:  printDistros,
	}

	return RPPrinter.InteractivelyPrintPages(&pagingStruct, viper.GetBool(NoHeaderFlag))
}

// getDistribution gets the properties of a distributions on
// the Redpesk factory
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func getDistribution(cmd *cobra.Command, args []string) error {
	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Now get the project from the args
	project, err := genericGetPrjFromArgs(args, false)
	if err != nil {
		return err
	}

	// Get information for this project
	tmpDistros := bckapi.RPCliDistributions{
		Distros: []bckapi.RPCliDistribution{{}},
	}
	if err := bckAPI.RPDistroGet(project.ID, &tmpDistros.Distros[0]); err != nil {
		return err
	}

	// Print the result to the user
	err = RPPrinter.CustomPrintObject(tmpDistros, verbose, printDistros)
	if err != nil {
		return err
	}

	return nil
}

// packagesDistributions gets the packages list within a distributions on
// the Redpesk backend
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func packagesDistributions(cmd *cobra.Command, args []string) error {
	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Get the architecture flag
	arch, err := cmd.Flags().GetString(distArchFlag)
	if err != nil {
		logFlagError(distArchFlag)
		return err
	}

	// Get the name flag
	nameFilter, err := cmd.Flags().GetString(distNameFlag)
	if err != nil {
		logFlagError(distNameFlag)
		return err
	}

	// Get the word flag
	wholeWord, err := cmd.Flags().GetBool(distWordFlag)
	if err != nil {
		logFlagError(distWordFlag)
		return err
	}

	// Firstly check that the architecture given is available in the factory
	availableArches, err := getTabOfArchitectures()
	if err != nil {
		return err
	}
	if !utils.ContainsString(availableArches, arch) {
		errMsg := fmt.Sprintf(imgPrepareArchNotExisting, arch)
		return errors.New(errMsg)
	}

	// Now get the project from the args
	project, err := genericGetPrjFromArgs(args, false)
	if err != nil {
		return err
	}

	// Now that we are sure the architecture is existing, let ask the backend for the packages
	nameFilterSubStr := ""
	if !wholeWord {
		nameFilterSubStr = nameFilter
		nameFilter = ""
	}

	// Prepare the function needed by the printer to get the packages
	getPkgsCB := func(pageNb, perPage int) (interface{}, printer.RPPrinterPaging, error) {
		// Get packages list
		pkgsInfo := rpapiv1.RPRepoInfoRpm{}

		if err := bckAPI.RPPackagesGetByPage(project.ID, arch, nameFilter, nameFilterSubStr, &pkgsInfo, pageNb, perPage); err != nil {
			return pkgsInfo, printer.RPPrinterPaging{}, err
		}

		// Get the paging interface from the structure retrieved
		pagingStruct, _ := printer.GetPagingFromInterface(&pkgsInfo)

		return pkgsInfo, pagingStruct, nil
	}

	// Prepare the structure to manage paging
	pagingStruct := printer.RPPagingToolsStruct{
		Verbose:       verbose,
		VerboseLineNb: distPkgsDisplayVerboseLineNb,
		GetObjCB:      getPkgsCB,
		PrintFancyCB:  printPackages,
	}

	return RPPrinter.InteractivelyPrintPages(&pagingStruct, viper.GetBool(NoHeaderFlag))
}

// addDistribution allows an administrator to import a distribution
// existing in Koji on the remote factory
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func addDistribution(cmd *cobra.Command, args []string) error {
	RPPrinter.CustomPrintLn(distAddCmdAnsLine0)

	// Get the common info for distribution import/update
	commonInfo, err := getAddUpdateCommonInfo(cmd, args)
	if err != nil {
		return err
	}

	// Get the value of the version flag
	distroVersion, err := cmd.Flags().GetString(distAddCmdVersionFlag)
	if err != nil {
		logFlagError(distAddCmdVersionFlag)
		return err
	}

	// Get the tag flag
	distroTag, err := cmd.Flags().GetString(distAddCmdTagFlag)
	if err != nil {
		logFlagError(distAddCmdTagFlag)
		return err
	}

	// Request the factory to import the distribution
	RPPrinter.CustomPrintF(distAddCmdAnsLine1, commonInfo["distro-id"])
	prjReq := rpapiv1.RPProjectFull{RPProject: rpapiv1.RPProject{Description: commonInfo["description"]}}
	prjAns := rpapiv1.RPProjectFull{}
	if err := bckAPI.ProjectMngtImport(commonInfo["distro-id"], distroVersion, commonInfo["sub-version"], distroTag, &prjReq, &prjAns); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	// Display something to the user if in raw mode
	RPPrinter.CustomPrintObject(prjAns, false, displayNothing)

	// Check from viper the non-blocking flag to see if we wait
	if viper.GetBool(nonBlockingFlag) {
		return nil
	}

	RPPrinter.CustomPrintF(distAddCmdAnsLine2, commonInfo["distro-id"]+" "+distroVersion)
	RPPrinter.ForceFlush()

	return waitForStableProject(prjAns.ID, rpapiv1.EVTProjectStateChange)
}

// updateDistribution allows an administrator to update a distribution
// existing in the remote factory
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func updateDistribution(cmd *cobra.Command, args []string) error {
	RPPrinter.CustomPrintLn(distUpdateCmdAnsLine0)

	// Get the common info for distribution import/update
	commonInfo, err := getAddUpdateCommonInfo(cmd, args)
	if err != nil {
		return err
	}

	// Get the "--refresh" flag value
	refreshNeeded, err := cmd.Flags().GetBool(distUpdateCmdRefreshFlag)
	if err != nil {
		logFlagError(distUpdateCmdRefreshFlag)
		return err
	}

	// Get the project
	distro, err := getProjectFromIdentifier(commonInfo["distro-id"], false)
	if err != nil {
		return err
	}

	// Change the different values if the user changed it
	updateNeeded := false
	if (commonInfo["sub-version"] != flagUpdateKeepCurrentValue) ||
		(commonInfo["description"] != flagUpdateKeepCurrentValue) {
		updateNeeded = true
	}

	// Ask refresh if needed
	if refreshNeeded {
		prjAns := rpapiv1.RPProjectFull{}
		RPPrinter.CustomPrintF(distUpdateCmdAnsLine1, commonInfo["distro-id"])
		RPPrinter.ForceFlush()
		if err := bckAPI.ProjectMngtRefresh(distro.ID.String(), &prjAns); err != nil {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
			return err
		}

		// Display something to the user if in raw mode
		RPPrinter.CustomPrintObject(prjAns, false, displayNothing)

		// Check from viper the non-blocking flag to see if we wait
		if !viper.GetBool(nonBlockingFlag) {
			err = waitForStableProject(prjAns.ID, rpapiv1.EVTProjectStateChange)
			if err != nil {
				RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
				return err
			}
			RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
		}
	}

	// Update only if needed
	if updateNeeded {
		// Get the root command
		rpCliRootCmd := cmd.Root()

		// Find the projects update command from that
		cmdLine := []string{prjsCommand, "update", distro.Slug}
		prjUpdateCmd, inArgs, err := rpCliRootCmd.Find(cmdLine)
		if err != nil {
			return err
		}

		// Set the different flags needed
		if commonInfo["sub-version"] != flagUpdateKeepCurrentValue {
			if err = prjUpdateCmd.Flags().Set(distSubVersionFlag, commonInfo["sub-version"]); err != nil {
				return err
			}
		}
		if commonInfo["description"] != flagUpdateKeepCurrentValue {
			if err = prjUpdateCmd.Flags().Set(distDescriptionFlag, commonInfo["description"]); err != nil {
				return err
			}
		}

		RPPrinter.CustomPrintLn()
		RPPrinter.ForceFlush()

		// Now call the update command: first the PreRunE, then RunE
		if err = prjUpdateCmd.PreRunE(prjUpdateCmd, inArgs); err != nil {
			return err
		}
		if err = prjUpdateCmd.RunE(prjUpdateCmd, inArgs); err != nil {
			return err
		}

	}

	return nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Utilities functions ------------------------------------------
// --------------------------------------------------------------

// getAddUpdateCommonInfo is a utility function allowing to get the info
// needed by the commands 'distributions add' and 'distributions update'
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: map[string]string, error -- the map of info returned and an error if something went bad
func getAddUpdateCommonInfo(cmd *cobra.Command, args []string) (map[string]string, error) {
	// Initialize the error we will use in this function
	var err error

	// First initialize the map that we will return
	resultMap := map[string]string{
		"distro-id":   "",
		"sub-version": "",
		"description": "",
	}

	// Get the value of the distro-ID (first args)
	resultMap["distro-id"], err = getFirstArg(args, fmt.Sprintf(objectIDNotSetMsg, "distribution"), fmt.Sprintf(objectIDRequiredMsg, "distribution"))
	if err != nil {
		return resultMap, err
	}

	// Get the value of the sub-version flag
	resultMap["sub-version"], err = cmd.Flags().GetString(distSubVersionFlag)
	if err != nil {
		logFlagError(distSubVersionFlag)
		return resultMap, err
	}

	// Get the value of the description flag
	resultMap["description"], err = cmd.Flags().GetString(distDescriptionFlag)
	if err != nil {
		logFlagError(distDescriptionFlag)
		return resultMap, err
	}

	return resultMap, nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Display functions --------------------------------------------
// --------------------------------------------------------------

// printDistros display the list of the current distributions available
// in the current Redpesk instance
//
// :param distros: List of distros to display
// :type distros: interface{}
// :param verbose: Tell if the print needs to be verbose or not
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func printDistros(distros interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Convert to the right type
	dists := distros.(bckapi.RPCliDistributions)

	if !verbose {
		// Print the header for distributions
		manageHeaderPrint(distListAnsHeader, writer)
	}

	// Display each architecture
	for index, dist := range dists.Distros {

		// Prepare the architectures and the RPMs number strings
		arches := []string{}
		rpmsNb := []string{}
		for _, arch := range dist.Architectures {
			if arch.Set {
				// Architecture string
				arches = append(arches, arch.ID)
				// RPMs number string
				if tmpRPMNb, ok := dist.NbRPMsPerArch[arch.ID]; ok {
					rpmsNb = append(rpmsNb, fmt.Sprintf("%d(%s)", tmpRPMNb, arch.ID))
				}
			}
		}

		if verbose {
			fmt.Fprintln(writer, distListAnsLine1, dist.ID.String())
			fmt.Fprintln(writer, distListAnsLine2, dist.Slug)
			fmt.Fprintln(writer, distListAnsLine3, dist.Name)
			fmt.Fprintln(writer, distListAnsLine4, dist.Version)
			fmt.Fprintln(writer, distListAnsLine5, dist.SubVersion)
			fmt.Fprintln(writer, distListAnsLine6, dist.Description)
			fmt.Fprintln(writer, distListAnsLine7, strings.Join(arches, ","))
			fmt.Fprintln(writer, distListAnsLine8, strings.Join(rpmsNb, ","))
			if index < (len(dists.Distros) - 1) {
				fmt.Fprintln(writer)
			}
		} else {
			versionToPrint := dist.Version
			if dist.SubVersion != "" {
				versionToPrint += " (" + dist.SubVersion + ")"
			}
			fmt.Fprintln(writer, dist.ID.String(), "\t", dist.Slug, "\t", versionToPrint, "\t", strings.Join(arches, ","), "\t", strings.Join(rpmsNb, ","))
		}
	}

	return nil
}

// printPackages display a list of packages available in a distribution
//
// :param packages: List of packages to display
// :type packages: interface{}
// :param verbose: Tell if the print needs to be verbose or not
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func printPackages(packages interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Convert to the right type
	pkgs := packages.(rpapiv1.RPRepoInfoRpm)

	if !verbose {
		// Print the header for distributions
		manageHeaderPrint(distPkgsAnsHeader, writer)
	}

	// Display each architecture
	for _, pkg := range pkgs.Res {

		if verbose {
			fmt.Fprintln(writer, distPkgsAnsLine1, pkg.Name)
			fmt.Fprintln(writer, distPkgsAnsLine2, pkg.Version)
			fmt.Fprintln(writer, distPkgsAnsLine3, pkg.Release)
			fmt.Fprintln(writer, distPkgsAnsLine4, pkg.Nvra)
			fmt.Fprintln(writer)
		} else {
			fmt.Fprintln(writer, pkg.Name, "\t", pkg.Version, "\t", pkg.Release, "\t", pkg.Nvra)
		}
	}

	return nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------
