/*
Copyright © 2024 IoT.bzh Company
Contact: https://www.iot.bzh/licensing

This file is part of the redpesk-cli project, the redpesk command line tool.

$RP_BEGIN_LICENSE$
# Commercial License Usage

Licensees holding valid commercial IoT.bzh licenses may use this file in
accordance with the commercial license agreement provided with the
Software or, alternatively, in accordance with the terms contained in
a written agreement between you and The IoT.bzh Company. For licensing terms
and conditions see https://www.iot.bzh/terms-conditions. For further
information use the contact form at https://www.iot.bzh/contact.

# GNU General Public License Usage

Alternatively, this file may be used under the terms of the GNU General
Public license version 3. This license is as published by the Free Software
Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
of this file. Please review the following information to ensure the GNU
General Public License requirements will be met
https://www.gnu.org/licenses/gpl-3.0.html.

$RP_END_LICENSE$
*/
package cmd

import "testing"

const (
	// Distribution
	distroDisplayNoVerboseColNb = 5
	distroDisplayVerboseLineNb  = 8

	// Package
	distroPkgDisplayNoVerboseColNb = 4
	distroPkgDisplayVerboseLineNb  = 4
)

// TestDistrosList is the function allowing to test if
// the command "rp-cli distributions list" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestDistrosList(test *testing.T) {

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID":            `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Slug":          `^[a-z0-9-]+$`,
		"Version":       `^([0-9\.]+)|([a-z]+ \([0-9\.]+\))$`,
		"Architectures": `.+`,
		"RPMs_number":   `.+`,
	}

	// Run the command and the tests (no verbose)
	checkCmdAnsKeysRegexpCol(test, distroDisplayNoVerboseColNb, regexpMap, distCmd, distListCmd)

	regexpMap["Name"] = `^[A-Za-z0-9- ]+$`
	regexpMap["Version"] = `^[a-z0-9.-]+$`
	regexpMap["Description"] = `.*`

	// Run the command and the tests (verbose)
	checkCmdAnswerListKeysRegexp(test, distroDisplayVerboseLineNb, "", regexpMap, distCmd, distListCmd, "-v")
}

// TestDistrosGet is the function allowing to test if
// the command "rp-cli distributions get" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestDistrosGet(test *testing.T) {

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID":            `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Slug":          `^redpesk-devel-batz$`,
		"Version":       `^1\.0$`,
		"Architectures": `^(aarch64|x86_64),(aarch64|x86_64)$`,
		"RPMs_number":   `^[0-9]+\((aarch64|x86_64)\),[0-9]+\((aarch64|x86_64)\)$`,
	}

	// Run the command and the tests (no verbose)
	checkCmdAnsKeysRegexpCol(test, distroDisplayNoVerboseColNb, regexpMap, distCmd, "get", "redpesk-devel-batz")

	regexpMap["Name"] = `^redpesk devel batz$`
	regexpMap["Version"] = `^1\.0$`
	regexpMap["SubVersion"] = `^$`
	regexpMap["Description"] = `.*`

	// Run the command and the tests (verbose)
	checkCmdAnswerListKeysRegexp(test, distroDisplayVerboseLineNb, "", regexpMap, distCmd, "get", "redpesk-devel-batz", "-v")

	// Create the map containing the regexp we want to match in the answer
	regexpMap = map[string]string{
		"ID":            `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Slug":          `^redpesk-zephyr-lts-3-7$`,
		"Version":       `^lts \(3\.7\.0\)$`,
		"Architectures": `^(aarch64|x86_64),(aarch64|x86_64)$`,
		"RPMs_number":   `^[0-9]+\((aarch64|x86_64)\),[0-9]+\((aarch64|x86_64)\)$`,
	}

	// Run the command and the tests (no verbose)
	checkCmdAnsKeysRegexpCol(test, distroDisplayNoVerboseColNb, regexpMap, distCmd, "get", "redpesk-zephyr-lts-3-7")

	regexpMap["Name"] = `^redpesk Zephyr$`
	regexpMap["Version"] = `^lts$`
	regexpMap["SubVersion"] = `^3\.7\.0$`
	regexpMap["Description"] = `Zephyr.*`

	// Run the command and the tests (verbose)
	checkCmdAnswerListKeysRegexp(test, distroDisplayVerboseLineNb, "", regexpMap, distCmd, "get", "redpesk-zephyr-lts-3-7", "-v")
}

// TestDistrosGetFromName is the function allowing to test if
// the command "rp-cli distributions get" is working correctly
// when a name is passed instead of a slug
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestDistrosGetFromName(test *testing.T) {

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID":            `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Slug":          `^redpesk-devel-batz$`,
		"Name":          `^redpesk devel batz$`,
		"Version":       `^1\.0$`,
		"SubVersion":    `^$`,
		"Description":   `.*`,
		"Architectures": `^(aarch64|x86_64),(aarch64|x86_64)$`,
		"RPMs_number":   `^[0-9]+\((aarch64|x86_64)\),[0-9]+\((aarch64|x86_64)\)$`,
	}

	// Run the command and the tests (verbose)
	checkCmdAnswerListKeysRegexp(test, distroDisplayVerboseLineNb, "", regexpMap, distCmd, "get", "redpesk devel batz", "-v")
}

// TestDistrosPackages is the function allowing to test if
// the command "rp-cli distributions packages" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestDistrosPackages(test *testing.T) {

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"Name":    `^[a-z0-9-.]+$`,
		"Version": `^1\.0$`,
		"Release": `^27.apps.rpbatz$`,
		"NVRA":    `^[a-z0-9._-]+$`,
	}

	// Run the command and the tests (no verbose)
	checkCmdAnsKeysRegexpCol(test, distroPkgDisplayNoVerboseColNb, regexpMap, distCmd, "packages", "redpesk-devel-batz-update", "-a", "x86_64")

	// Run the command and the tests (verbose)
	checkCmdAnswerListKeysRegexp(test, distroPkgDisplayVerboseLineNb, "", regexpMap, distCmd, "packages", "redpesk-devel-batz-update", "-a", "aarch64", "-v")

	// Modify the regexp map in order to test if the filtering is working correctly
	regexpMap["Name"] = `^gawk.*`
	regexpMap["NVRA"] = `^gawk.*-1.0-27.apps.rpbatz-(aarch64|x86_64)$`

	// Run the command and the tests (with filter)
	checkCmdAnsKeysRegexpCol(test, distroPkgDisplayNoVerboseColNb, regexpMap, distCmd, "packages", "redpesk-devel-batz-update", "-a", "x86_64", "--name", "gawk")

	// Modify the regexp map in order to test if the filtering is working correctly
	regexpMap["Name"] = `^autoconf$`
	regexpMap["NVRA"] = `^autoconf-1.0-27.apps.rpbatz-(aarch64|x86_64)$`

	// Run the command and the tests (with filter)
	checkCmdAnsKeysRegexpCol(test, distroPkgDisplayVerboseLineNb, regexpMap, distCmd, "packages", "redpesk-devel-batz-update", "-a", "x86_64", "--name", "autoconf", "-w")
}

// TestDistrosAdd is the function allowing to test if
// the command "rp-cli distributions add" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestDistrosAdd(test *testing.T) {
	// Firstly, add a distribution if it does not exist
	err := addObjectIfNotExists(test, distCmd, "my-own-distribution", distCmd, "add", "my-own-distribution", "-v", "1.0", "-d", "This is my own distribution!")
	if err != nil {
		test.Fatal(err)
	}

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID":            `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Slug":          `^my-own-distribution$`,
		"Name":          `^my-own-distribution$`,
		"Version":       `^1\.0$`,
		"SubVersion":    `^$`,
		"Description":   `.*`,
		"Architectures": `^(aarch64|x86_64),(aarch64|x86_64)$`,
		"RPMs_number":   `^[0-9]+\((aarch64|x86_64)\),[0-9]+\((aarch64|x86_64)\)$`,
	}

	// Run the command and the tests (verbose)
	checkCmdAnswerListKeysRegexp(test, distroDisplayVerboseLineNb, "", regexpMap, distCmd, "get", "my-own-distribution", "-v")
}

// TestDistrosUpdate is the function allowing to test if
// the command "rp-cli distributions update" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestDistrosUpdate(test *testing.T) {
	// Firstly, add a distribution if it does not exist
	err := addObjectIfNotExists(test, distCmd, "my-own-distribution-to-update", distCmd, "add", "my-own-distribution-to-update", "-v", "1.0", "--sub-version", "175", "-d", "This is my own distribution!")
	if err != nil {
		test.Fatal(err)
	}

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID":            `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Slug":          `^my-own-distribution-to-update$`,
		"Name":          `^my-own-distribution-to-update$`,
		"Version":       `^1\.0$`,
		"SubVersion":    `^175$`,
		"Description":   `^This is my own distribution!$`,
		"Architectures": `^(aarch64|x86_64),(aarch64|x86_64)$`,
		"RPMs_number":   `^[0-9]+\((aarch64|x86_64)\),[0-9]+\((aarch64|x86_64)\)$`,
	}

	// Check the values
	checkCmdAnswerListKeysRegexp(test, distroDisplayVerboseLineNb, "", regexpMap, distCmd, "get", "my-own-distribution-to-update", "-v")

	// Run the update command and check that all is [OK]
	checkActionCmdAns(test, distCmd, "update", "my-own-distribution-to-update", "-d", "This is my own distribution to update!", "--sub-version", "85", "--refresh")

	// Change the regexp to match what we will update
	regexpMap["Version"] = `^1\.0-refreshed$`
	regexpMap["SubVersion"] = `^85$`
	regexpMap["Description"] = `^This is my own distribution to update!$`

	// Check the values
	checkCmdAnswerListKeysRegexp(test, distroDisplayVerboseLineNb, "", regexpMap, distCmd, "get", "my-own-distribution-to-update", "-v")
}
