/*
Copyright © 2021-2023 IoT.bzh Company
Contact: https://www.iot.bzh/licensing

This file is part of the redpesk-cli project, the redpesk command line tool.

$RP_BEGIN_LICENSE$
Commercial License Usage
 Licensees holding valid commercial IoT.bzh licenses may use this file in
 accordance with the commercial license agreement provided with the
 Software or, alternatively, in accordance with the terms contained in
 a written agreement between you and The IoT.bzh Company. For licensing terms
 and conditions see https://www.iot.bzh/terms-conditions. For further
 information use the contact form at https://www.iot.bzh/contact.

GNU General Public License Usage
 Alternatively, this file may be used under the terms of the GNU General
 Public license version 3. This license is as published by the Free Software
 Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
 of this file. Please review the following information to ensure the GNU
 General Public License requirements will be met
 https://www.gnu.org/licenses/gpl-3.0.html.
$RP_END_LICENSE$
*/

package cmd

import (
	"encoding/json"
	"fmt"
	"os"
	"path"
	"regexp"
	"strconv"
	"strings"
	"text/tabwriter"

	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/printer"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/rpxxxapi/ladapi"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/utils"
	"git.ovh.iot/redpesk/redpesk-infra/rp-webserver/actions/rpapiv1"
	"github.com/gofrs/uuid"
	"github.com/pkg/errors"
	"github.com/spf13/cobra"
	"github.com/spf13/viper"
	"golang.org/x/term"
)

// -------------------------------------------------
// Types definition --------------------------------
// -------------------------------------------------

// -------------------------------------------------
// -------------------------------------------------

// -------------------------------------------------
// Global variable (inside this package) -----------
// -------------------------------------------------

// -------------------------------------------------
// -------------------------------------------------

// --------------------------------------------------------------
// Initialization functions -------------------------------------
// --------------------------------------------------------------

func initImagesCmd(rootCmd *cobra.Command) {

	// imagesCmd represents the images command
	var imagesCmd = &cobra.Command{
		Use:     imagesCmdName,
		Aliases: []string{imageCmdAlias},
		Short:   imageCmdShortUsage,
	}

	// -------------------------------------------------
	// Images sub-command ------------------------------
	// -------------------------------------------------

	// listImgCmd represents the "images list" command
	var listImgCmd = &cobra.Command{
		Use:     imagesListCmd,
		Aliases: []string{imagesListCmdAlias},
		Short:   imagesListCmdShortUsage,
		RunE:    listImages,
		Annotations: map[string]string{
			"__flag_need_bck_connect__": "TRUE",
		},
	}

	// getImgCmd represents the "images get" command
	var getImgCmd = &cobra.Command{
		Use:               imagesGetCmd,
		Short:             imagesGetCmdShortUsage,
		RunE:              getImage,
		ValidArgsFunction: getListImagesAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// addImgCmd represents the "images add" command
	var addImgCmd = &cobra.Command{
		Use:     imagesAddCmd,
		Aliases: []string{imagesAddAlias},
		Short:   imagesAddCmdShortUsage,
		PreRunE: initEventImages,
		RunE:    addImage,
		Annotations: map[string]string{
			"__flag_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":        "TRUE"},
	}

	// updateImgCmd represents the "images update" command
	var updateImgCmd = &cobra.Command{
		Use:               imagesUpdateCmd,
		Aliases:           []string{imagesUpdateAlias},
		Short:             imagesUpdateCmdShortUsage,
		SilenceUsage:      true,
		PreRunE:           initEventImages,
		RunE:              updateImage,
		ValidArgsFunction: getListImagesAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// delImgCmd represents the "images delete" command
	var delImgCmd = &cobra.Command{
		Use:               imagesDelCmd,
		Aliases:           []string{imagesDelAlias},
		Short:             imagesDelCmdShortUsage,
		PreRunE:           initEventImages,
		RunE:              deleteImage,
		ValidArgsFunction: getListImagesAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// localBuildImgCmd represents the "images localbuild" command
	var localBuildImgCmd = &cobra.Command{
		Use:               imageLocalBuildCmd,
		Aliases:           []string{imageLocalBuildCmdAlias},
		Short:             imageLocalBuildCmdShortUsage,
		Long:              imageLocalBuildCmdLongUsage,
		PreRunE:           initEventsRPLAD,
		RunE:              localBuildImage,
		ValidArgsFunction: getListImagesAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__lad_connect__":          "TRUE"},
	}

	// buildImgCmd represents the "images build" command
	var buildImgCmd = &cobra.Command{
		Use:               fmt.Sprintf(genericBuildCmd, "img-ID"),
		Aliases:           []string{genericBuildAlias},
		Short:             fmt.Sprintf(genericBuildCmdShortUsage, "image"),
		Long:              fmt.Sprintf(genericBuildCmdLongUsage, "image", "image", "images"),
		PreRunE:           initEventBuildAndImage,
		RunE:              buildImage,
		ValidArgsFunction: getListImagesAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// buildsCmd represents the "images builds" command
	var buildsCmd = &cobra.Command{
		Use:   genericBuildsCmd,
		Short: fmt.Sprintf(genericBuildsCmdShortUsage, "image"),
	}

	// buildsListImagesCmd represents the "images builds list" command
	var buildsListImagesCmd = &cobra.Command{
		Use:               fmt.Sprintf(genericSubBuildsListCmd, "img-ID"),
		Aliases:           []string{genericSubBuildsListAlias},
		Short:             fmt.Sprintf(genericSubBuildsListCmdShortUsage, "image"),
		RunE:              buildsListImage,
		ValidArgsFunction: getListImagesAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// buildsGetImagesCmd represents the "images builds get" command
	var buildsGetImagesCmd = &cobra.Command{
		Use:               fmt.Sprintf(genericBuildsGetCmd, "img-ID"),
		Short:             genericBuildsGetCmdShortUsage,
		Long:              fmt.Sprintf(genericBuildsGetCmdLongUsage, "image", "image", "img"),
		RunE:              imgBuildsGetAndDisplay,
		ValidArgsFunction: getListImagesAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// buildsCancelImagesCmd represents the "applications builds cancel" command
	var buildsCancelImagesCmd = &cobra.Command{
		Use:               fmt.Sprintf(genericBuildsCancelCmd, "img-ID"),
		Short:             genericBuildsCancelCmdShortUsage,
		Long:              fmt.Sprintf(genericBuildsCancelCmdLongUsage, "image", "image", "img"),
		RunE:              buildCancelImage,
		ValidArgsFunction: getListImagesAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// buildsLogsImgsCmd represents the "images builds logs" command
	var buildsLogsImgsCmd = &cobra.Command{
		Use:               fmt.Sprintf(genericBuildLogsCmd, "img-ID"),
		Short:             fmt.Sprintf(genericBuildLogsCmdShortUsage, "image"),
		Long:              fmt.Sprintf(genericBuildLogsCmdLongUsage, "image", "image", "image", "img"),
		RunE:              buildGetLogsImage,
		ValidArgsFunction: getListImagesAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// testImgCmd represents the "images test" command
	var testImgCmd = &cobra.Command{
		Use:               fmt.Sprintf(genericTestCmd, "img-ID"),
		Aliases:           []string{genericTestAlias},
		Short:             fmt.Sprintf(genericTestCmdShortUsage, "image"),
		Long:              fmt.Sprintf(genericTestCmdLongUsage, "image", "image", "image", "--force-img"),
		PreRunE:           initEventTests,
		RunE:              testGeneric,
		ValidArgsFunction: getListImagesAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE",
			"__read_object_command__":  "TRUE"}, // Needed to get the list of applications
	}

	// testsCmd represents the "images tests" command
	var testsCmd = &cobra.Command{
		Use:   genericTestsCmd,
		Short: fmt.Sprintf(genericTestsCmdShortUsage, "image"),
	}

	// testsListImgsCmd represents the "images tests list" command
	var testsListImgsCmd = &cobra.Command{
		Use:               fmt.Sprintf(genericTestsListCmd, "img-ID"),
		Aliases:           []string{genericTestsListAlias},
		Short:             fmt.Sprintf(genericTestsListCmdShortUsage, "image"),
		RunE:              testsListGeneric,
		ValidArgsFunction: getListImagesAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// testsGetImgsCmd represents the "images tests get" command
	var testsGetImgsCmd = &cobra.Command{
		Use:               fmt.Sprintf(genericTestsGetCmd, "img-ID"),
		Short:             genericTestsGetCmdShortUsage,
		Long:              fmt.Sprintf(genericTestsGetCmdLongUsage, "image", "image", "--force-img"),
		RunE:              testsGetGeneric,
		ValidArgsFunction: getListImagesAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// testsLogsImgsCmd represents the "images tests logs" command
	var testsLogsImgsCmd = &cobra.Command{
		Use:               fmt.Sprintf(genericTestLogsCmd, "img-ID"),
		Short:             fmt.Sprintf(genericTestLogsCmdShortUsage, "image"),
		Long:              fmt.Sprintf(genericTestLogsCmdLongUsage, "image", "image", "image", "--force-img"),
		RunE:              testGetLogsGeneric,
		ValidArgsFunction: getListImagesAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// testCancelImgsCmd represents the "images tests cancel" command
	var testsCancelImgsCmd = &cobra.Command{
		Use:               fmt.Sprintf(genericTestCancelCmd, "img-ID"),
		Short:             genericTestCancelCmdShortUsage,
		Long:              fmt.Sprintf(genericTestCancelCmdLongUsage, "image", "image", "--force-img"),
		RunE:              testCancelGeneric,
		ValidArgsFunction: getListImagesAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// testsStatsImgCmd represents the "images tests stats" command
	var testsStatsImgsCmd = &cobra.Command{
		Use:               fmt.Sprintf(genericTestsStatsCmd, "img-ID"),
		Aliases:           []string{genericTestsStatsAlias},
		Short:             fmt.Sprintf(genericTestsStatsCmdShortUsage, "image"),
		SilenceUsage:      true,
		RunE:              testStatsImage,
		ValidArgsFunction: getListImagesAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	var releaseImagesCmd = &cobra.Command{
		Use:               imageReleaseCmd,
		Aliases:           []string{imageReleaseCmdAlias},
		Short:             imageReleaseCmdShortUsage,
		Long:              imageReleaseCmdLongUsage,
		PreRunE:           initEventReleaseImage,
		RunE:              releaseImage,
		ValidArgsFunction: getListImagesAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// releasesCmd represents the "images releases" commands group
	var releasesCmd = &cobra.Command{
		Use:          imageReleasesCmd,
		Short:        imageReleasesCmdShortUsage,
		SilenceUsage: true,
	}

	// -------------------------------------------------
	// -------------------------------------------------

	// -------------------------------------------------
	// Image's releases sub-commands -----------------
	// -------------------------------------------------

	// listReleaseCmd represents the "projects releases list" command
	var listReleaseCmd = &cobra.Command{
		Use:          imagesListCmd,
		Aliases:      []string{imagesListCmdAlias},
		Short:        imageListReleasedCmdShortUsage,
		SilenceUsage: true,
		RunE:         listReleasedImages,
	}

	// getReleaseCmd represents the "images releases get" command
	var getReleaseCmd = &cobra.Command{
		Use:               imagesGetCmd,
		Short:             imageGetReleasedCmdShortUsage,
		SilenceUsage:      true,
		RunE:              getReleasedImage,
		ValidArgsFunction: getListReleasedImagesAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// updateReleaseCmd represents the "images releases update" command
	var updateReleaseCmd = &cobra.Command{
		Use:               imagesUpdateCmd,
		Aliases:           []string{imagesUpdateAlias},
		Short:             imageUpdateReleasedCmdShortUsage,
		SilenceUsage:      true,
		RunE:              updateReleasedImage,
		ValidArgsFunction: getListReleasedImagesAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// deleteReleaseCmd represents the "images releases delete" command
	var deleteReleaseCmd = &cobra.Command{
		Use:               imagesDelCmd,
		Aliases:           []string{imagesDelAlias},
		Short:             imageDeleteReleasedCmdShortUsage,
		SilenceUsage:      true,
		RunE:              deleteReleasedImage,
		ValidArgsFunction: getListReleasedImagesAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}
	// -------------------------------------------------
	// -------------------------------------------------

	// Here you will define your flags and configuration settings.

	// Persistent Flags -----------------------------------------
	// ----------------------------------------------------------

	// Local Flags ----------------------------------------------
	// List
	listImgCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	listImgCmd.Flags().StringSlice(imagesListCmdTypeFlag, []string{"standard", "external"}, imagesListCmdTypeFlagUsage)
	listImgCmd.Flags().StringP(imagesCmdPrjFlag, imagesCmdPrjFlagAlias, imagesCmdPrjFlagDefault, imagesListCmdPrjFlagUsage)
	// Get
	getImgCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	// Add
	addImgCmd.Flags().StringP(imagesCmdNameFlag, imagesCmdNameFlagAlias, "", imagesCmdNameFlagUsage)
	addImgCmd.Flags().StringP(imagesCmdAliasFlag, imagesCmdAliasFlagAlias, "", imagesCmdAliasFlagUsage)
	addImgCmd.Flags().StringP(imagesCmdDistroFlag, imagesCmdDistroFlagAlias, "", imagesCmdDistroFlagUsage)
	addImgCmd.Flags().String(imagesCmdArchFlag, "", imagesCmdArchFlagUsage)
	addImgCmd.Flags().StringP(imagesCmdSecurityFlag, imagesCmdSecurityFlagAlias, imagesCmdSecurityFlagDef, imagesCmdSecurityFlagUsage)
	addImgCmd.Flags().String(imagesCmdKickstartURLFlag, "", imagesCmdKickstartURLFlagUsage)
	addImgCmd.Flags().StringP(imagesCmdKickstartBranchFlag, imagesCmdKickstartBranchFlagAlias, "", imagesCmdKickstartBranchFlagUsage)
	addImgCmd.Flags().String(imagesCmdKickstartPathFlag, ".", imagesCmdKickstartPathFlagUsage)
	addImgCmd.Flags().StringP(imagesCmdKickstartNameFlag, imagesCmdKickstartNameFlagAlias, "", imagesCmdKickstartNameFlagUsage)
	addImgCmd.Flags().StringSliceP(imagesCmdAddRepoFlag, imagesCmdAddRepoFlagAlias, []string{}, imagesCmdAddRepoFlagUsage)
	addImgCmd.Flags().Bool(imagesCmdNoAutoRepoFlag, false, imagesCmdNoAutoRepoFlagUsage)
	addImgCmd.Flags().StringP(imagesCmdPrjFlag, imagesCmdPrjFlagAlias, "", imagesCmdPrjFlagUsage)
	addImgCmd.Flags().StringP(imagesCmdBoardModelFlag, imagesCmdBoardModelFlagAlias, imagesCmdBoardModelFlagDef, imagesCmdBoardModelFlagUsage)
	addImgCmd.Flags().BoolP(imagesCmdForceFlag, imagesCmdForceFlagAlias, false, imagesCmdForceFlagUsage)
	// Update
	updateImgCmd.Flags().StringP(imagesCmdNameFlag, imagesCmdNameFlagAlias, flagUpdateKeepCurrentValue, imagesCmdNameFlagUsage)
	updateImgCmd.Flags().StringP(imagesCmdDistroFlag, imagesCmdDistroFlagAlias, flagUpdateKeepCurrentValue, imagesCmdDistroFlagUsage)
	updateImgCmd.Flags().String(imagesCmdArchFlag, flagUpdateKeepCurrentValue, imagesCmdArchFlagUsage)
	updateImgCmd.Flags().StringP(imagesCmdSecurityFlag, imagesCmdSecurityFlagAlias, flagUpdateKeepCurrentValue, imagesCmdSecurityFlagUsage)
	updateImgCmd.Flags().String(imagesCmdKickstartURLFlag, flagUpdateKeepCurrentValue, imagesCmdKickstartURLFlagUsage)
	updateImgCmd.Flags().StringP(imagesCmdKickstartBranchFlag, imagesCmdKickstartBranchFlagAlias, flagUpdateKeepCurrentValue, imagesCmdKickstartBranchFlagUsage)
	updateImgCmd.Flags().String(imagesCmdKickstartPathFlag, flagUpdateKeepCurrentValue, imagesCmdKickstartPathFlagUsage)
	updateImgCmd.Flags().StringP(imagesCmdKickstartNameFlag, imagesCmdKickstartNameFlagAlias, flagUpdateKeepCurrentValue, imagesCmdKickstartNameFlagUsage)
	updateImgCmd.Flags().StringSliceP(imagesCmdAddRepoFlag, imagesCmdAddRepoFlagAlias, []string{flagUpdateKeepCurrentValue}, imagesCmdAddRepoFlagUsage)
	updateImgCmd.Flags().Bool(imageUpdateCmdResetReposFlag, false, imageUpdateCmdResetReposFlagUsage)
	updateImgCmd.Flags().Bool(imagesCmdNoAutoRepoFlag, false, imagesCmdNoAutoRepoFlagUsage)
	updateImgCmd.Flags().StringP(imagesCmdPrjFlag, imagesCmdPrjFlagAlias, flagUpdateKeepCurrentValue, imagesCmdPrjFlagUsage)
	updateImgCmd.Flags().StringP(imagesCmdBoardModelFlag, imagesCmdBoardModelFlagAlias, flagUpdateKeepCurrentValue, imagesCmdBoardModelFlagUsage)
	updateImgCmd.Flags().BoolP(imagesCmdForceFlag, imagesCmdForceFlagAlias, false, imagesCmdForceFlagUsage)
	updateImgCmd.Flags().StringP(imagesCmdDlURLFlag, imagesCmdDlURLFlagAlias, "", imagesCmdDlURLFlagUsage)
	// Localbuild
	localBuildImgCmd.Flags().String(imagesLocalBuildCmdLocalKickFlag, "", imagesLocalBuildCmdLocalKickFlagUsage)
	localBuildImgCmd.Flags().IntP(imagesLocalBuildCmdSizeFlag, imagesLocalBuildCmdSizeFlagAlias, -1, imagesLocalBuildCmdSizeFlagUsage)
	// Build
	buildImgCmd.Flags().StringP(genericBuildCmdCommentFlag, genericBuildCmdCommentFlagAlias, "", genericBuildCmdCommentFlagUsage)
	// Builds list
	buildsListImagesCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	// Builds get
	buildsGetImagesCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	buildsGetImagesCmd.Flags().BoolP(imgsForceImgFlag, imgsForceImgFlagAlias, false, imgsForceImgFlagUsage)
	// Builds logs
	buildsLogsImgsCmd.Flags().BoolP(imgsForceImgFlag, imgsForceImgFlagAlias, false, imgsForceImgFlagUsage)
	buildsLogsImgsCmd.Flags().StringP(genericBuildLogsDirFlag, genericBuildLogsDirFlagAlias, genericBuildLogsDirFlagDefault, genericBuildLogsDirFlagUsage)
	buildsLogsImgsCmd.Flags().BoolP(genericBuildZipFlag, genericBuildZipFlagAlias, false, genericBuildZipFlagUsage)
	// Builds cancel
	buildsCancelImagesCmd.Flags().BoolP(imgsForceImgFlag, imgsForceImgFlagAlias, false, imgsForceImgFlagUsage)
	// Test
	testImgCmd.Flags().BoolP(imgsForceImgFlag, imgsForceImgFlagAlias, false, imgsForceImgFlagUsage)
	testImgCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	testImgCmd.Flags().StringP(genericTestCmdCommentFlag, genericTestCmdCommentFlagAlias, "", genericTestCmdCommentFlagUsage)
	testImgCmd.Flags().StringP(imgsTestCmdAppFlag, imgsTestCmdAppFlagAlias, "", imgsTestCmdAppFlagUsage)
	testImgCmd.Flags().String(imgsTestCmdAppProjectFlag, "", imgsTestCmdAppProjectFlagUsage)
	testImgCmd.Flags().StringP(genericTestCmdBoardModelFlag, genericTestCmdBoardModelFlagAlias, "", genericTestCmdBoardModelFlagUsage)
	testImgCmd.Flags().Int(genericTestCmdTimeOutFlag, 5, genericTestCmdTimeOutFlagUsage)
	testImgCmd.Flags().Bool(genericTestCmdProgressFlag, genericTestCmdProgressFlagDefault, genericTestCmdProgressFlagUsage)
	// Tests list
	testsListImgsCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	// Tests get
	testsGetImgsCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	testsGetImgsCmd.Flags().BoolP(imgsForceImgFlag, imgsForceImgFlagAlias, false, imgsForceImgFlagUsage)
	// Tests logs
	testsLogsImgsCmd.Flags().BoolP(imgsForceImgFlag, imgsForceImgFlagAlias, false, imgsForceImgFlagUsage)
	testsLogsImgsCmd.Flags().StringP(genericBuildLogsDirFlag, genericBuildLogsDirFlagAlias, genericBuildLogsDirFlagDefault, genericBuildLogsDirFlagUsage)
	testsLogsImgsCmd.Flags().BoolP(genericBuildZipFlag, genericBuildZipFlagAlias, false, genericBuildZipFlagUsage)
	// Tests cancel
	testsCancelImgsCmd.Flags().BoolP(imgsForceImgFlag, imgsForceImgFlagAlias, false, imgsForceImgFlagUsage)
	// Tests stats
	testsStatsImgsCmd.Flags().Int(genericTestsStatsCmdLastFlag, 0, genericTestsStatsCmdLastFlagUsage)
	testsStatsImgsCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	// Release
	releaseImagesCmd.Flags().StringP(imageReleaseCmdNameFlag, imageReleaseCmdNameFlagAlias, "", imageReleaseCmdNameFlagUsage)
	releaseImagesCmd.Flags().String(imageReleaseCmdDistroFlag, "", imageReleaseCmdDistroFlagUsage)
	releaseImagesCmd.Flags().StringP(imageReleaseCmdVersionFlag, imageReleaseCmdVersionFlagAlias, "", imageReleaseCmdVersionFlagUsage)
	releaseImagesCmd.Flags().StringSlice(imageReleaseCmdTeamFlag, []string{}, imageReleaseCmdTeamFlagUsage)
	releaseImagesCmd.Flags().StringSliceP(imageReleaseCmdUserFlag, imageReleaseCmdUserFlagAlias, []string{}, imageReleaseCmdUserFlagUsage)
	releaseImagesCmd.Flags().BoolP(imageReleaseCmdAllUsersFlag, imageReleaseCmdAllUsersFlagAlias, false, imageReleaseCmdAllUsersFlagUsage)
	// List the released images
	listReleaseCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	// Get a released image
	getReleaseCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	getReleaseCmd.Flags().BoolP(imageGetReleaseCmdAudienceFlag, imageGetReleaseCmdAudienceFlagAlias, false, imageGetReleaseCmdAudienceFlagUsage)
	// Update a released image
	updateReleaseCmd.Flags().StringP(imageReleaseCmdNameFlag, imageReleaseCmdNameFlagAlias, flagUpdateKeepCurrentValue, imageReleaseCmdNameFlagUsage)
	updateReleaseCmd.Flags().StringSlice(imageReleaseCmdTeamFlag, []string{flagUpdateKeepCurrentValue}, imageReleaseCmdTeamFlagUsage)
	updateReleaseCmd.Flags().StringSliceP(imageReleaseCmdUserFlag, imageReleaseCmdUserFlagAlias, []string{flagUpdateKeepCurrentValue}, imageReleaseCmdUserFlagUsage)
	updateReleaseCmd.Flags().StringP(imageReleaseCmdAllUsersFlag, imageReleaseCmdAllUsersFlagAlias, flagUpdateKeepCurrentValue, imageReleaseCmdAllUsersFlagUsage)
	updateReleaseCmd.Flags().Bool(imageUpdateReleaseCmdResetUsersFlag, false, imageUpdateReleaseCmdResetUsersFlagUsage)
	updateReleaseCmd.Flags().Bool(imageUpdateReleaseCmdResetTeamsFlag, false, imageUpdateReleaseCmdResetTeamsFlagUsage)
	// ----------------------------------------------------------

	// Add some flags as required
	// Add
	addImgCmd.MarkFlagRequired(imagesCmdNameFlag)
	addImgCmd.MarkFlagRequired(imagesCmdDistroFlag)
	addImgCmd.MarkFlagRequired(imagesCmdArchFlag)
	// Release
	releaseImagesCmd.MarkFlagRequired(imageReleaseCmdDistroFlag)
	releaseImagesCmd.MarkFlagRequired(imageReleaseCmdVersionFlag)

	// Add some flags as hidden
	addImgCmd.Flags().MarkHidden(imagesCmdBoardModelFlag)
	addImgCmd.Flags().MarkHidden(imagesCmdSecurityFlag)
	updateImgCmd.Flags().MarkHidden(imagesCmdBoardModelFlag)
	updateImgCmd.Flags().MarkHidden(imagesCmdSecurityFlag)
	updateImgCmd.Flags().MarkHidden(imagesCmdDlURLFlag)

	// Specify dynamic flag completion
	// List
	listImgCmd.RegisterFlagCompletionFunc(imagesListCmdTypeFlag, getListTypeAsTab)
	listImgCmd.RegisterFlagCompletionFunc(imagesCmdPrjFlag, getFullListPrjsAsTab)

	// Add
	addImgCmd.RegisterFlagCompletionFunc(imagesCmdArchFlag, getListArchAsTab)
	addImgCmd.RegisterFlagCompletionFunc(imagesCmdDistroFlag, getListDistAsTab)
	addImgCmd.RegisterFlagCompletionFunc(imagesCmdPrjFlag, getListPrjsAsTab)
	addImgCmd.RegisterFlagCompletionFunc(imagesCmdSecurityFlag, getSecuModelsAsTab)
	// Update
	updateImgCmd.RegisterFlagCompletionFunc(imagesCmdArchFlag, getListArchAsTab)
	updateImgCmd.RegisterFlagCompletionFunc(imagesCmdDistroFlag, getListDistAsTab)
	updateImgCmd.RegisterFlagCompletionFunc(imagesCmdPrjFlag, getListPrjsAsTab)
	updateImgCmd.RegisterFlagCompletionFunc(imagesCmdSecurityFlag, getSecuModelsAsTab)
	// Release
	releaseImagesCmd.RegisterFlagCompletionFunc(imageReleaseCmdDistroFlag, getListDistAsTab)
	releaseImagesCmd.RegisterFlagCompletionFunc(imageReleaseCmdTeamFlag, getListTeamsAsTab)
	releaseImagesCmd.RegisterFlagCompletionFunc(imageReleaseCmdUserFlag, getListUserAsTab)
	// Release update
	updateReleaseCmd.RegisterFlagCompletionFunc(imageReleaseCmdTeamFlag, getListTeamsAsTab)
	updateReleaseCmd.RegisterFlagCompletionFunc(imageReleaseCmdUserFlag, getListUserAsTab)
	// Test
	testImgCmd.RegisterFlagCompletionFunc(imgsTestCmdAppFlag, getListAppsAsTab)
	testImgCmd.RegisterFlagCompletionFunc(imgsTestCmdAppProjectFlag, getListPrjsAsTab)
	testImgCmd.RegisterFlagCompletionFunc(genericTestCmdBoardModelFlag, getListRTMBoardModelsAsTab)

	// Images subcommands ------------------------------------
	// Add the group commands "images" as subcommand of rootCmd
	rootCmd.AddCommand(imagesCmd)
	// Add the command "list" as a subcommand of imagesCmd
	imagesCmd.AddCommand(listImgCmd)
	// Add the command "get" as a subcommand of imagesCmd
	imagesCmd.AddCommand(getImgCmd)
	// Add the command "add" as a subcommand of imagesCmd
	imagesCmd.AddCommand(addImgCmd)
	// Add the command "del" as a subcommand of imagesCmd
	imagesCmd.AddCommand(delImgCmd)
	// Add the command "update" as a subcommand of imagesCmd
	imagesCmd.AddCommand(updateImgCmd)
	// Add the command "localbuild" as a subcommand of imagesCmd
	imagesCmd.AddCommand(localBuildImgCmd)
	// Add the command "build" as a subcommand of imagesCmd
	imagesCmd.AddCommand(buildImgCmd)
	// Add the command "builds" as a subcommand of imagesCmd
	imagesCmd.AddCommand(buildsCmd)
	// Add the command "test" as a subcommand of imagesCmd
	imagesCmd.AddCommand(testImgCmd)
	// Add the command "tests" as a subcommand of imagesCmd
	imagesCmd.AddCommand(testsCmd)
	// Add the command "release" as a subcommand of imagesCmd
	imagesCmd.AddCommand(releaseImagesCmd)
	// Add the command "releases" as a subcommand of imagesCmd
	imagesCmd.AddCommand(releasesCmd)
	// -------------------------------------------------------

	// Images builds subcommands ------------------------------
	buildsCmd.AddCommand(buildsListImagesCmd)
	buildsCmd.AddCommand(buildsGetImagesCmd)
	buildsCmd.AddCommand(buildsLogsImgsCmd)
	buildsCmd.AddCommand(buildsCancelImagesCmd)
	// -------------------------------------------------------

	// Images tests subcommands ------------------------------
	testsCmd.AddCommand(testsListImgsCmd)
	testsCmd.AddCommand(testsGetImgsCmd)
	testsCmd.AddCommand(testsLogsImgsCmd)
	testsCmd.AddCommand(testsCancelImgsCmd)
	testsCmd.AddCommand(testsStatsImgsCmd)
	// -------------------------------------------------------

	// Images releases subcommands ---------------------------
	releasesCmd.AddCommand(listReleaseCmd)
	releasesCmd.AddCommand(getReleaseCmd)
	releasesCmd.AddCommand(updateReleaseCmd)
	releasesCmd.AddCommand(deleteReleaseCmd)
	// -------------------------------------------------------
}

// initEventImages inits everything that is needed for the images event
// related
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func initEventImages(cmd *cobra.Command, args []string) error {

	// Init the events list
	eventsList := []string{
		rpapiv1.EvtImageAdd,
		rpapiv1.EvtImageDelete,
		rpapiv1.EvtImageStateChange,
	}

	return CLIRegisterToMultipleEvents(eventsList)
}

// initEventBuildAndImage inits everything that is needed for the build image
// command
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func initEventBuildAndImage(cmd *cobra.Command, args []string) error {

	// First to the initialization for the App
	err := initEventImages(cmd, args)
	if err != nil {
		return err
	}

	// Init the events list
	eventsList := []string{
		rpapiv1.EvtImageBuildInfo,
		rpapiv1.EvtImageBuildSubTaskInfo,
	}

	return CLIRegisterToMultipleEvents(eventsList)
}

// initEventReleaseImage inits everything that is needed for the image release
// command to run
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func initEventReleaseImage(cmd *cobra.Command, args []string) error {
	// First init the image events
	err := initEventImages(cmd, args)
	if err != nil {
		return err
	}
	return CLIRegisterToMultipleEvents([]string{rpapiv1.EVTReleaseImageStateChange})
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Completion functions -----------------------------------------
// --------------------------------------------------------------

// getListImagesAsTab gets all the images available in the Redpesk
// instance and return it as an array of strings
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getListImagesAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	imagesList := []string{}
	typesList := []string{"standard", "external"}

	// Get images list
	images := rpapiv1.RPImages{}
	if err := bckAPI.ImagesListGetOfTypes(&images, "", typesList, ""); err != nil {
		return imagesList, cobra.ShellCompDirectiveNoFileComp
	}

	// We got correctly the images list, fill the imagesList tab
	for _, image := range images.Images {
		imagesList = append(imagesList, image.Slug)
	}

	return imagesList, cobra.ShellCompDirectiveNoFileComp
}

// getSecuModelsAsTab returns the static security models available:
// smack and selinux
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getSecuModelsAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	return []string{"smack", "selinux"}, cobra.ShellCompDirectiveNoFileComp
}

// getListReleasedImagesAsTab gets all the released images available in the Redpesk
// instance and return it as an array of strings
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getListReleasedImagesAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	imagesList := []string{}

	// Get the release images
	tmpList := &rpapiv1.RPReleaseImages{}
	// Add a smart filter to avoid asking for all the images
	filter := fmt.Sprintf("slug~=%s*", toComplete)
	// Request the released images using the filter
	if err := bckAPI.ListReleasedImages(tmpList, filter); err != nil {
		return imagesList, cobra.ShellCompDirectiveNoFileComp
	}

	// We got correctly the images list, fill the imagesList tab
	for _, image := range tmpList.ReleaseImages {
		imagesList = append(imagesList, image.Slug)
	}

	return imagesList, cobra.ShellCompDirectiveNoFileComp
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Command functions --------------------------------------------
// --------------------------------------------------------------

// listImages gets all the images available in the Redpesk
// instance and prints it
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func listImages(cmd *cobra.Command, args []string) error {

	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Get the project flag
	projectID, err := cmd.Flags().GetString(appsCmdPrjFlag)
	if err != nil {
		logFlagError(appsCmdPrjFlag)
		return err
	}

	// Get the type flag
	typesSlice, err := cmd.Flags().GetStringSlice(imagesListCmdTypeFlag)
	if err != nil {
		logFlagError(imagesListCmdTypeFlag)
		return err
	}

	// Convert "iotbzh" to "external", and remove duplicated values
	utils.ReplaceInStringSlice(&typesSlice, "iotbzh", "external")
	typesSlice = utils.RemoveDuplicateInStrSlice(typesSlice)

	var prjUUIDStr string
	if projectID != "" {
		// Get the project
		project, err := getProjectFromIdentifier(projectID, false)
		if err != nil {
			return err
		}
		prjUUIDStr = project.ID.String()
	} else {
		prjUUIDStr = ""
	}

	// Prepare the function needed by the printer to get the images
	getImagesCB := func(pageNb, perPage int) (interface{}, printer.RPPrinterPaging, error) {
		// Get images list
		images := rpapiv1.RPImages{}
		if err := bckAPI.ImagesListByTypesPerPage(&images, prjUUIDStr, typesSlice, pageNb, perPage); err != nil {
			return images, printer.RPPrinterPaging{}, err
		}
		// Get the paging interface from the structure retrieved
		pagingStruct, _ := printer.GetPagingFromInterface(&images)

		return images, pagingStruct, nil
	}

	// Prepare the structure to manage paging
	pagingStruct := printer.RPPagingToolsStruct{
		Verbose:       verbose,
		VerboseLineNb: imgDisplayVerboseLightLineNb,
		GetObjCB:      getImagesCB,
		PrintFancyCB:  displayImages,
	}

	return RPPrinter.InteractivelyPrintPages(&pagingStruct, viper.GetBool(NoHeaderFlag))
}

// getImage gets the image information available in the Redpesk
// instance and prints it
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func getImage(cmd *cobra.Command, args []string) error {

	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Get the image given by the user
	img, err := genericGetImgFromArgs(args, true)
	if err != nil {
		return err
	}

	// Print the result to the user
	err = RPPrinter.CustomPrintObject(img, verbose, displayOneImage)
	if err != nil {
		return err
	}

	return nil
}

// addImage creates an image as requested from the user
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func addImage(cmd *cobra.Command, args []string) error {

	// Display the first line to the user
	RPPrinter.CustomPrintLn(imagesAddCmdAnsLine0)

	// Create the structure to fill
	img := rpapiv1.RPImageFull{}

	// Fill the structure with the right values (given in command line)
	if err := prepareImageStruct(cmd, args, &img, true, false); err != nil {
		return err
	}

	// Log the image creation
	Log.Infof(imgCreationInfoMsg, img.Name)
	Log.Debugf(imgCreationDebugMsg, img)

	// Tell the user that the image is currently being created
	RPPrinter.CustomPrintF(imagesAddCmdAnsLine1, img.Name)

	// Request the creation of this image to the backend
	imgAns := rpapiv1.RPImageFull{}
	if err := bckAPI.ImageCreate(&img, &imgAns); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	// Display something to the user if in raw mode
	RPPrinter.CustomPrintObject(imgAns, false, displayNothing)

	// Create the sentence to display when the object will be added
	slugSentence := fmt.Sprintf(objectSlugDisplayMsg, "image", imgAns.Slug)

	// Check from viper the non-blocking flag to see if we wait
	nonBlocking := viper.GetBool(nonBlockingFlag)
	if nonBlocking {
		// The user do not want blocking functions
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
		// Display the slug of the created object
		RPPrinter.CustomPrintLn(slugSentence)
		return nil
	}

	// Block until we receive a corresponding websocket event
	for {
		rEvent := <-cmdEventChannel

		// Check that the received event is for a image addition
		if rEvent.Type == rpapiv1.EvtImageAdd {
			// Function handling the events
			// DecodeImage needs to be done!
			rImage, err := rEvent.DecodeImage()
			if err != nil {
				return err
			}

			if utils.UUIDEqual(rImage.ID, imgAns.ID) {
				// Display OK if not in raw mode
				RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
				// Display something to the user if in raw mode
				RPPrinter.CustomPrintObject(rImage, false, displayNothing)
				// Display the slug of the created object
				RPPrinter.CustomPrintLn(slugSentence)
				break
			}
		}

		// Check if we get an error event during the operation
		if (rEvent.Type == rpapiv1.EVTError) || (rEvent.Type == rpapiv1.EVTSessionExpired) {
			// Function handling the events
			err := handleErrorEvent(imgAns.ID, rEvent, true)
			if err != nil {
				return err
			}
		}
	}

	return nil
}

// updateImage update an image as requested from the user
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func updateImage(cmd *cobra.Command, args []string) error {

	RPPrinter.CustomPrintLn(imagesUpdateCmdAnsLine0)

	// Get the image given by the user
	img, err := genericGetImgFromArgs(args, true)
	if err != nil {
		return err
	}

	// Fill the structure with the right values (given in command line)
	if err := prepareImageStruct(cmd, args, &img, false, false); err != nil {
		return err
	}

	// Log the update
	Log.Infof(imgUpdateInfoMsg, img.Name)
	Log.Debugf(imgUpdateDebugMsg, img)

	// Tell the user that the image is currently being updated
	RPPrinter.CustomPrintF(imagesUpdateCmdAnsLine2, img.Name)

	// Request the update of this update to the backend
	imgAns := rpapiv1.RPImageFull{}
	if err := bckAPI.ImageUpdate(img.ID.String(), &img, &imgAns); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	// Display something to the user if in raw mode
	RPPrinter.CustomPrintObject(imgAns, false, displayNothing)

	// Check from viper the non-blocking flag to see if we wait
	nonBlocking := viper.GetBool(nonBlockingFlag)
	if nonBlocking {
		// The user do not want blocking functions
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
		return nil
	}

	// Block until we receive a corresponding websocket event
	for {
		rEvent := <-cmdEventChannel

		// Check that the received event is for a image update
		if rEvent.Type == rpapiv1.EvtImageStateChange {
			// Function handling the events
			rImage, err := rEvent.DecodeImage()
			if err != nil {
				return err
			}

			if rImage.Slug == imgAns.Slug {
				RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
				// Display something to the user if in raw mode
				RPPrinter.CustomPrintObject(rImage, false, displayNothing)
				break
			}
		}

		// Check if we get an error event during the operation
		if (rEvent.Type == rpapiv1.EVTError) || (rEvent.Type == rpapiv1.EVTSessionExpired) {
			// Function handling the events
			err = handleErrorEvent(imgAns.ID, rEvent, true)
			if err != nil {
				return err
			}
		}
	}

	return nil
}

// deleteImage deletes an existing image as requested from the user
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func deleteImage(cmd *cobra.Command, args []string) error {

	RPPrinter.CustomPrintLn(imagesDelCmdAnsLine0)

	// Get the image given by the user
	image, err := genericGetImgFromArgs(args, false)
	if err != nil {
		return err
	}

	// Tell the user that the image is currently being deleted
	RPPrinter.CustomPrintF(imagesDelCmdAnsLine1, image.ID.String())

	// Ask the backend for deletion
	imgAns := rpapiv1.RPImageFull{}
	if err := bckAPI.ImageDelete(image.ID.String(), &imgAns); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	// Display something to the user if in raw mode
	RPPrinter.CustomPrintObject(imgAns, false, displayNothing)

	// Check from viper the non-blocking flag to see if we wait
	nonBlocking := viper.GetBool(nonBlockingFlag)
	if nonBlocking {
		// The user do not want blocking functions
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
		return nil
	}

	// Block until we receive a corresponding websocket event
	for {
		rEvent := <-cmdEventChannel

		// Check that the received event is for an image deletion
		if rEvent.Type == rpapiv1.EvtImageDelete {
			// Function handling the events
			rImage, err := rEvent.DecodeImage()
			if err != nil {
				return err
			}

			if rImage.Slug == imgAns.Slug {
				// Display OK if not in raw mode
				RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
				// Display something to the user if in raw mode
				RPPrinter.CustomPrintObject(rImage, false, displayNothing)
				break
			}
		}

		// Check if we get an error event during the operation
		if (rEvent.Type == rpapiv1.EVTError) || (rEvent.Type == rpapiv1.EVTSessionExpired) {
			// Function handling the events
			err = handleErrorEvent(imgAns.ID, rEvent, true)
			if err != nil {
				return err
			}
		}
	}

	return nil
}

// localBuildImage run the local build of an image as requested from the user
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func localBuildImage(cmd *cobra.Command, args []string) error {

	RPPrinter.CustomPrintLn(imagesLocalBuildCmdAnsLine0)

	// Get the image given by the user
	image, err := genericGetImgFromArgs(args, false)
	if err != nil {
		return err
	}

	// Get the "local-kickstart" flag (optional)
	imgLocalKick, err := cmd.Flags().GetString(imagesLocalBuildCmdLocalKickFlag)
	if err != nil {
		logFlagError(imagesLocalBuildCmdLocalKickFlag)
		return err
	}

	// Get the "size" flag (optional)
	imgSize, err := cmd.Flags().GetInt(imagesLocalBuildCmdSizeFlag)
	if err != nil {
		logFlagError(imagesLocalBuildCmdSizeFlag)
		return err
	}

	// Check that all the needed arguments are in the structure
	RPPrinter.CustomPrint(imagesLocalBuildCmdAnsLine1)

	fullKickstart, err := checkImageParametersForBuild(&image.RPImage, imgLocalKick)
	if err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	// Check and prepare the directory for image build
	RPPrinter.CustomPrint(imagesLocalBuildCmdAnsLine2)

	resultingDir, err := setUpLocalImageBuildDir(&image.RPImage, fullKickstart)
	if err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	// Print info to the user about the directory set-up
	kickstartName := ""
	RPPrinter.CustomPrintLn()
	RPPrinter.CustomPrintF(imagesLocalBuildCmdAnsLine3, resultingDir)
	if fullKickstart != "" {
		kickstartName = path.Base(imgLocalKick)
		RPPrinter.CustomColorPrintF(printer.ColorCyan, imagesLocalBuildCmdAnsLine4, kickstartName, path.Join(resultingDir, kickstartName))
	}

	// Run the "localimagebuild" command inside the local-builder
	RPPrinter.CustomColorPrintLn(printer.ColorPurple, imagesLocalBuildCmdAnsLogs1)
	RPPrinter.ForceFlush()
	errToReturn := runLocalImageBuild(&image.RPImage, resultingDir, kickstartName, imgSize)
	RPPrinter.CustomColorPrintLn(printer.ColorPurple, imagesLocalBuildCmdAnsLogs2)

	// If the build went well, display where the image is
	if errToReturn == nil {
		RPPrinter.CustomColorPrintF(printer.ColorGreen, imagesLocalBuildCmdAnsLine5, resultingDir)
		RPPrinter.CustomPrintLn()
	} else {
		// If a problem appeared, return an error
		RPPrinter.CustomColorPrintLn(printer.ColorRed, imagesLocalBuildCmdAnsLine6)
	}

	// Remove the file ".dsk" located in the local builder under '/home/devel/.oz/images/'
	cmdToRun := fmt.Sprintf("rm -f /home/devel/.oz/images/%s.dsk", image.Slug)
	_, err = execInLocalBuilderStatic(cmdToRun, "")
	if err != nil {
		errMsg := fmt.Sprintf(imgDeleteLocalDskFileError, err.Error())
		Log.Error(errMsg)
	}

	return errToReturn
}

// buildImage run the build of an image as requested from the user
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func buildImage(cmd *cobra.Command, args []string) error {

	RPPrinter.CustomPrintLn(fmt.Sprintf(genericBuildCmdAnsLine0, "Image"))

	// Get the image given by the user
	img, err := genericGetImgFromArgs(args, false)
	if err != nil {
		return err
	}

	// Get the additional information (comment) to save with the build
	comment, err := cmd.Flags().GetString(genericBuildCmdCommentFlag)
	if err != nil {
		logFlagError(genericBuildCmdCommentFlag)
		return err
	}

	// Tell the user that we start the build of the image
	RPPrinter.CustomPrintF(genericBuildCmdAnsLine1, img.Slug, "image")
	RPPrinter.ForceFlush()

	// Ask the backend to start the build
	buildsAns := rpapiv1.RPBuilds{}
	if err := bckAPI.ImageBuild(img.ID.String(), comment, &buildsAns); err != nil {
		return err
	}

	// always one build should be started
	if len(buildsAns.Builds) < 1 {
		return errors.Errorf("no image build has been started")
	}
	buildAns := buildsAns.Builds[0]

	// Display something to the user if in raw mode
	RPPrinter.CustomPrintObject(buildAns, false, displayNothing)

	// Check from viper the non-blocking flag to see if we wait
	nonBlocking := viper.GetBool(nonBlockingFlag)
	if nonBlocking {
		// The user do not want blocking functions
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
		return nil
	}

	// Create a tab that will contains the results we want to display at the end
	resultsTab := []string{}

	// Block until we receive a corresponding websocket event
	for {
		rEvent := <-cmdEventChannel

		// Check that the received event is for application build
		if rEvent.Type == rpapiv1.EvtImageBuildInfo || rEvent.Type == rpapiv1.EvtImageBuildSubTaskInfo {

			// Function handling the events data
			rBuild, err := rEvent.DecodeBuild()
			if err != nil {
				return err
			}

			// Check if we are monitoring this build ID
			if rBuild.ID == buildAns.ID {
				// Print the build in the output
				RPPrinter.CustomPrintObject(rBuild, true, printOneBuildLive)

				// Check if the build is done
				if rBuild.Status == rpapiv1.RPBuildTaskStatusDone {
					// Add a result line for this build
					resultsTab = append(resultsTab, getLiveBuildLine(rBuild, genericBuildLiveLabelSuccess))
					break
				}

				if rBuild.Status == rpapiv1.RPBuildTaskStatusCanceled {
					// Add a result line for this build
					resultsTab = append(resultsTab, getLiveBuildLine(rBuild, genericBuildLiveLabelCanceled))
					// Display the comment if there is one
					if rBuild.Comment != "" {
						resultsTab = append(resultsTab, fmt.Sprintf("%s Cancellation comment: '%s'", genericBuildLiveLabelCanceled, rBuild.Comment))
					}
					break
				}

				// Check if the build is failed
				if rBuild.Status == "failed" {
					// Add a result line for this build
					resultsTab = append(resultsTab, getLiveBuildLine(rBuild, genericBuildLiveLabelFailed))

					// Save output of error fault code and string if there are some
					if rBuild.Task.Error.FaultString != "" {
						resultsTab = append(resultsTab, fmt.Sprintf("%s %s%d", genericBuildLiveLabelFailed, genericBuildCmdVerboseErrorFaultCode, rBuild.Task.Error.FaultCode))
						resultsTab = append(resultsTab, fmt.Sprintf("%s %s'%s'", genericBuildLiveLabelFailed, genericBuildCmdVerboseErrorFaultMsg, rBuild.Task.Error.FaultString))
					}
					break
				}
			}
		}

		// Check if we get an error event during the operation
		if (rEvent.Type == rpapiv1.EVTError) || (rEvent.Type == rpapiv1.EVTSessionExpired) {
			// Function handling the events
			err = handleErrorEvent(img.ID, rEvent, false)
			if err != nil {
				return err
			}
		}
	}

	return printLiveBuildResults(resultsTab, imgBuildFailedMsg)
}

// buildsListImage gets all the builds available for the given image
// and prints it.
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func buildsListImage(cmd *cobra.Command, args []string) error {
	return buildsList(cmd, args, "image")
}

// buildGetLogsImage allows to retrieve logs from an image build.
// If the build ID is not specified, the logs of the latest build are retrieved
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func buildGetLogsImage(cmd *cobra.Command, args []string) error {

	RPPrinter.CustomPrintLn(genericBuildLogsCmdAnsLine0)

	// Retrieve the build ID from the args given by the user
	buildID, err := getBuildIDFromArgs(cmd, args, imgsForceImgFlag, "image", getImgUUIDCallBack)
	if err != nil {
		return err
	}

	return buildGetLogs(cmd, buildID, "image")
}

// buildCancelImage allows to cancel a build.
// If the build ID is not specified, the latest build is cancelled
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func buildCancelImage(cmd *cobra.Command, args []string) error {

	// Retrieve the build ID from the args given by the user
	buildID, err := getBuildIDFromArgs(cmd, args, imgsForceImgFlag, "image", getImgUUIDCallBack)
	if err != nil {
		return err
	}

	return buildCancel(buildID)
}

// testStatsImage display the tests statistics for the given
// image
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func testStatsImage(cmd *cobra.Command, args []string) error {
	// Get the image given by the user
	img, err := genericGetImgFromArgs(args, false)
	if err != nil {
		return err
	}

	// Now request and display builds stats for this project
	return getAndDisplayTestsStats(cmd, img.ParentProjectID, uuid.Nil, img.ID)
}

// releaseImage allows to ask the factory to release an image.
// a build ID can be specified. If not, the latest build ID
// is used.
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func releaseImage(cmd *cobra.Command, args []string) error {
	// Print the header
	RPPrinter.CustomPrintLn(imageReleaseCmdAnsLine0)

	// Get the image given by the user
	img, err := genericGetImgFromArgs(args, false)
	if err != nil {
		return err
	}

	// Get the distribution
	distro, err := cmd.Flags().GetString(imageReleaseCmdDistroFlag)
	if err != nil {
		logFlagError(imageReleaseCmdDistroFlag)
		return err
	}

	// Get the version
	version, err := cmd.Flags().GetString(imageReleaseCmdVersionFlag)
	if err != nil {
		logFlagError(imageReleaseCmdVersionFlag)
		return err
	}

	// Check that we can find the distribution in projects
	RPPrinter.CustomPrint(imageReleaseCmdAnsLine1)
	prjDistro, err := getProjectFromIdentifier(distro, false)
	if err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	// Prepare the structure to send to the backend
	imgRelease := rpapiv1.RPReleaseImage{
		DistributionID: prjDistro.ID,
		Version:        version,
	}

	// Fill the image release structure with other value
	err = prepareRelImageStruct(cmd, &imgRelease, true)
	if err != nil {
		return err
	}

	// Prepare the answer we will receive
	ansImage := rpapiv1.RPReleaseImageFull{}

	// Send the request to the backend to release the project
	RPPrinter.CustomPrintF(imageReleaseCmdAnsLine2, img.Slug)
	if err = bckAPI.ImageRelease(img.ID.String(), &imgRelease, &ansImage); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	// Display something to the user if in raw mode
	RPPrinter.CustomPrintObject(ansImage, false, displayNothing)

	// Check from viper the non-blocking flag to see if we wait
	nonBlocking := viper.GetBool(nonBlockingFlag)
	if nonBlocking {
		return nil
	}

	RPPrinter.CustomPrintF(imageReleaseCmdAnsLine3, img.Slug)
	RPPrinter.ForceFlush()

	// Block until we receive a corresponding websocket event
	refImageStable := false
	relImageStable := false
	for {
		rEvent := <-cmdEventChannel

		// Check that the received event is for a image state change
		if rEvent.Type == rpapiv1.EvtImageStateChange {

			// Function handling the events
			rImage, err := rEvent.DecodeImage()
			if err != nil {
				RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
				return err
			}

			// Check if this is one of the project we monitoring and if there are stable
			if utils.UUIDEqual(rImage.ID, img.ID) && (rImage.State == rpapiv1.RPImageStableState) {
				refImageStable = true
				// Display something to the user if in raw mode
				RPPrinter.CustomPrintObject(rImage, false, displayNothing)
			}

		} else if rEvent.Type == rpapiv1.EVTReleaseImageStateChange {
			// Function handling the events
			rImageRel, err := rEvent.DecodeReleaseImage()
			if err != nil {
				RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
				return err
			}

			// Check if it is our release project
			if utils.UUIDEqual(rImageRel.ID, ansImage.ID) && (rImageRel.State == rpapiv1.RPImageStableState) {
				relImageStable = true
				// Display something to the user if in raw mode
				RPPrinter.CustomPrintObject(rImageRel, false, displayNothing)
			}

		} else if (rEvent.Type == rpapiv1.EVTError) || (rEvent.Type == rpapiv1.EVTSessionExpired) {
			// Function handling the error, try with the reference image
			err = handleErrorEvent(img.ID, rEvent, true)
			if err != nil {
				return err
			}

			// Function handling the error, try with the released image
			err = handleErrorEvent(ansImage.ID, rEvent, true)
			if err != nil {
				return err
			}
		}

		// Check if both images are stable
		if refImageStable && relImageStable {
			// The release is done
			RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
			break
		}
	}

	// Manage the audience for the released image
	infoMessage := fmt.Sprintf(imageReleaseCmdAnsLine4, img.Slug)
	err = manageAudienceUpdate(cmd, ansImage.ID, true, infoMessage, bckAPI.GetReleasedImageAudience, bckAPI.UpdateReleasedImageAudience)

	return err
}

// listReleasedImages list the released images from the factory
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func listReleasedImages(cmd *cobra.Command, args []string) error {
	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Prepare the function needed by the printer to get Released Projects
	getReleasedImgsCB := func(pageNb, perPage int) (interface{}, printer.RPPrinterPaging, error) {
		// Get release images list
		relImgs := rpapiv1.RPReleaseImages{}
		if err := bckAPI.ListReleasedImagesPerPage(&relImgs, pageNb, perPage); err != nil {
			return relImgs.ReleaseImages, printer.RPPrinterPaging{}, err
		}
		// Get the paging interface from the structure retrieved
		pagingStruct, _ := printer.GetPagingFromInterface(&relImgs)

		return relImgs.ReleaseImages, pagingStruct, nil
	}

	// Prepare the structure to manage paging
	pagingStruct := printer.RPPagingToolsStruct{
		Verbose:       verbose,
		VerboseLineNb: imgRelDisplayVerboseLightLineNb,
		GetObjCB:      getReleasedImgsCB,
		PrintFancyCB:  displayRelImages,
	}

	return RPPrinter.InteractivelyPrintPages(&pagingStruct, viper.GetBool(NoHeaderFlag))
}

// getReleasedImage display one released image from the factory
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func getReleasedImage(cmd *cobra.Command, args []string) error {
	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Get the audience flag
	audience, err := cmd.Flags().GetBool(imageGetReleaseCmdAudienceFlag)
	if err != nil {
		logFlagError(imageGetReleaseCmdAudienceFlag)
		return err
	}

	// Get the full released image
	relImg, err := genericGetRelImgFromArgs(args, true)
	if err != nil {
		return err
	}

	// Print the result to the user
	err = RPPrinter.CustomPrintObject(relImg, verbose, displayOneRelImage)
	if err != nil {
		return err
	}

	// Check if we need to display the audience
	if !audience {
		return nil
	}

	// Retrieve the audience for this released project
	relImgAudience := rpapiv1.RPObjectRoles{}
	err = bckAPI.GetReleasedImageAudience(relImg.ID.String(), &relImgAudience)
	if err != nil {
		return err
	}

	// Print the result to the user
	err = RPPrinter.CustomPrintObject(relImgAudience, verbose, printRelAudience)

	return err
}

// updateReleasedImage update a released image settings
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func updateReleasedImage(cmd *cobra.Command, args []string) error {
	// Print the header
	RPPrinter.CustomPrintLn(imageReleaseUpdateCmdAnsLine0)

	// Get the release image that the user wants to update
	relImg, err := genericGetRelImgFromArgs(args, false)
	if err != nil {
		return err
	}

	oldName := relImg.Name

	// Fill the image release structure with other value
	err = prepareRelImageStruct(cmd, &relImg.RPReleaseImage, false)
	if err != nil {
		return err
	}

	// Print a warning to say that we cannot update the name of the released-image
	if oldName != relImg.Name {
		Log.Warn("The released image name cannot be changed for the moment. It is not implemented by the backend.")
	}

	// // Prepare the answer we will receive
	// ansImage := &rpapiv1.RPReleaseImageFull{}

	// // Send the request to the backend to update the released image
	// RPPrinter.CustomPrintF(imageReleaseUpdateCmdAnsLine1, imgSlug)
	// if err = bckAPI.UpdateReleasedImage(imgUUID.String(), &currReleasedImg, ansImage); err != nil {
	// 	RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
	// 	return err
	// }
	// RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	// // Display something to the user if in raw mode
	// RPPrinter.CustomPrintObject(ansImage, false, displayNothing)

	// No need to wait for event for the moment, no async operation

	// Manage the audience for the released project
	infoMessage := fmt.Sprintf(imageReleaseUpdateCmdAnsLine2, relImg.Slug)
	err = manageAudienceUpdate(cmd, relImg.ID, false, infoMessage, bckAPI.GetReleasedImageAudience, bckAPI.UpdateReleasedImageAudience)

	return err
}

// deleteReleasedImage delete a released image from the factory
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if a problem has occurred
func deleteReleasedImage(cmd *cobra.Command, args []string) error {
	// Print the header
	RPPrinter.CustomPrintLn(imageDeleteReleaseCmdAnsLine0)

	// Get the release image that the user wants to delete
	relImg, err := genericGetRelImgFromArgs(args, false)
	if err != nil {
		return err
	}

	// Request to delete the released project
	RPPrinter.CustomPrintF(imageDeleteReleaseCmdAnsLine1, relImg.Slug)
	ansImage := &rpapiv1.RPReleaseImageFull{}
	if err = bckAPI.DeleteReleasedImage(relImg.ID.String(), ansImage); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	// Display something to the user if in raw mode
	RPPrinter.CustomPrintObject(ansImage, false, displayNothing)

	return nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Utilities functions ------------------------------------------
// --------------------------------------------------------------

// getImgUUIDCallBack is a function specifically designed to be call by the
// `getBuildIDFromArgs()` function to retrieve the UUID of the build parent.
// Most of the time, it is better to call the `genericGetImgFromArgs()`
// since you get the app structure.
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: uuid.UUID, error -- the image UUID and an error if something went wrong
func getImgUUIDCallBack(cmd *cobra.Command, args []string) (uuid.UUID, error) {
	img, err := genericGetImgFromArgs(args, false)
	if err != nil {
		return uuid.Nil, err
	}
	return img.ID, nil
}

// genericGetImgFromArgs allows a function to retrieve the
// image targeted by the user.
//
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param getFullImg: if set, the caller request to have a truly full app (so a GET request on /images/<img_uuid>) will be done
// :type getFullImg: bool
// :returns: (rpapiv1.RPProjectFull, error) -- the project and an error if something went bad
func genericGetImgFromArgs(args []string, getFullImg bool) (rpapiv1.RPImageFull, error) {
	// Get the image ID and return an error if there is nothing
	imageID, err := getFirstArg(args, imageIDNotSetMsg, imageIDRequiredMsg)
	if err != nil {
		return rpapiv1.RPImageFull{}, err
	}

	return getImgFromIdentifier(imageID, getFullImg)
}

// getImgFromIdentifier is a private function allowing to retrieve an image from a given identifier.
// This identifier can be the UUID itself as a string representation or the slug of the image
//
// :param imgIdentifier: the identifier of the image
// :type imgIdentifier: string
// :param getFullImg: if set, the caller request to have a truly full app (so a GET request on /images/<img_uuid>) will be done
// :type getFullImg: bool
// :returns: (rpapiv1.RPImageFull, error) -- the image and an error if something went bad
func getImgFromIdentifier(imgIdentifier string, getFullImg bool) (rpapiv1.RPImageFull, error) {
	// First try to see if the imgIdentifier is a valid UUID
	isValidUUID := utils.IsValidUUID(imgIdentifier)

	// If it is the case, call directly the backend with this UUID to see if it works
	if isValidUUID {
		image := rpapiv1.RPImageFull{}
		if err := bckAPI.ImageGet(imgIdentifier, &image); err == nil {
			// If everything went well, this means that the identifier was truly the UUID of the image!
			// In the other case keep going and try with slug...
			return image, nil
		}
	}

	// So, try with slug now (using filter feature)
	tmpList := &rpapiv1.RPImages{}
	// Create a slice of string to use as filter
	filterValues := []string{
		fmt.Sprintf("slug=%s", imgIdentifier),
		fmt.Sprintf("name=%s", imgIdentifier),
	}

	// Try each filter until finding the right one
	for index, filter := range filterValues {
		if err := bckAPI.ImagesListGetOfTypes(tmpList, "", []string{"standard", "external"}, filter); err != nil {
			return rpapiv1.RPImageFull{}, err
		}

		// If we got one match, bingo
		if len(tmpList.Images) == 1 {
			break
		}

		// If we got no match, try the other filter
		if (len(tmpList.Images) < 1) && (runMode != "testing") {
			Log.Warnf(imgIdentifierNotFoundOnTryX, imgIdentifier, index, filter)
			continue
		}

		// If we got too many matches
		if (len(tmpList.Images) > 1) && (runMode != "testing") {
			Log.Warnf(tooManyImgIDFoundOnTryX, index, filter)
			continue
		}

	}

	if len(tmpList.Images) != 1 {
		// Error: no project found with this identifier
		return rpapiv1.RPImageFull{}, fmt.Errorf(imgIdentifierNotFound, imgIdentifier)
	}

	// If the user wants the real full image, do another request
	if getFullImg {
		imgToReturn := rpapiv1.RPImageFull{}
		err := bckAPI.ImageGet(tmpList.Images[0].ID.String(), &imgToReturn)
		return imgToReturn, err
	}

	// The user does not need a full image: just set the light one inside an "empty" shell
	imgToReturn := rpapiv1.RPImageFull{
		RPImage: tmpList.Images[0],
	}
	return imgToReturn, nil
}

// genericGetRelImgFromArgs allows a function to retrieve the
// released image targeted by the user.
//
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param getFullRelImg: if set, the caller request to have a truly full released image (so a GET request on /release-images/<img_uuid>) will be done in all case
// :type getFullRelImg: bool
// :returns: (rpapiv1.RPReleaseProjectFull, error) -- the released project and an error if something went bad
func genericGetRelImgFromArgs(args []string, getFullRelImg bool) (rpapiv1.RPReleaseImageFull, error) {
	// Get the image ID and return an error if there is nothing
	relImgID, err := getFirstArg(args, imageIDNotSetMsg, imageIDRequiredMsg)
	if err != nil {
		return rpapiv1.RPReleaseImageFull{}, err
	}

	return getRelImageFromIdentifier(relImgID, getFullRelImg)
}

// getRelImageFromIdentifier is a private function allowing to retrieve
// a released image from a given identifier.
// This identifier can be the UUID itself as a string representation or
// the slug of the released image
//
// :param identifier: the identifier of the released image
// :type identifier: string
// :param getFullRelImg: if set, the caller request to have a truly full released image (so a GET request on /release-images/<img_uuid>) will be done in all case
// :type getFullRelImg: bool
// :returns: (rpapiv1.RPReleaseImageFull, error) -- the released image and an error if something went bads
func getRelImageFromIdentifier(identifier string, getFullRelImg bool) (rpapiv1.RPReleaseImageFull, error) {
	// Check if the identifier is really looking like an UUID
	isValidUUID := utils.IsValidUUID(identifier)

	// If it is the case, call directly the backend with this UUID to see if it works
	if isValidUUID {
		relImg := rpapiv1.RPReleaseImageFull{}
		if err := bckAPI.GetReleasedImage(identifier, &relImg); err == nil {
			// If everything went well, this means that the identifier was truly the UUID of the released project!
			// In the other case keep going and try with slug...
			return relImg, nil
		}
	}

	// So, try with slug now (using filter feature)
	tmpList := &rpapiv1.RPReleaseImages{}
	// Create a slice of string to use as filter
	filterValues := []string{
		fmt.Sprintf("slug=%s", identifier),
		fmt.Sprintf("name=%s", identifier),
	}

	// Try each filter until finding the right one
	for index, filter := range filterValues {
		if err := bckAPI.ListReleasedImages(tmpList, filter); err != nil {
			return rpapiv1.RPReleaseImageFull{}, err
		}

		// If we got one match, bingo
		if len(tmpList.ReleaseImages) == 1 {
			break
		}

		// If we got no match, try the other filter
		if (len(tmpList.ReleaseImages) < 1) && (runMode != "testing") {
			Log.Warnf(imgIdentifierNotFoundOnTryX, identifier, index, filter)
			continue
		}

		// If we got too many matches
		if (len(tmpList.ReleaseImages) > 1) && (runMode != "testing") {
			Log.Warnf(tooManyImgIDFoundOnTryX, index, filter)
			continue
		}

	}

	if len(tmpList.ReleaseImages) != 1 {
		// Error: no image found with this identifier
		return rpapiv1.RPReleaseImageFull{}, fmt.Errorf(imgIdentifierNotFound, identifier)
	}

	// If the user wants the real full released image, do another request
	if getFullRelImg {
		imgToReturn := rpapiv1.RPReleaseImageFull{}
		err := bckAPI.GetReleasedImage(tmpList.ReleaseImages[0].ID.String(), &imgToReturn)
		return imgToReturn, err
	}

	// The user does not need a full image: just set the light one inside an "empty" shell
	imgToReturn := rpapiv1.RPReleaseImageFull{
		RPReleaseImage: tmpList.ReleaseImages[0],
	}
	return imgToReturn, nil
}

// getFullGitpkgsDirPath is a private function allowing to easily get
// the full path of the gitpkgs/ directory on the host
//
// :returns: string, error -- the full path to gitpkgs and an error if something went wrong
func getFullGitpkgsDirPath() (string, error) {
	// Get the  gitpkgs/ directory in the host
	tmpPkgsDir, err := getCustomBuildsDirs(localPkgCmdPackageFlag)
	if err != nil {
		return "", err
	}

	// Check that the directory exists
	pkgsDir, err := utils.SanitizePath(tmpPkgsDir)
	if err != nil {
		return "", err
	}
	if _, err = os.Stat(pkgsDir); os.IsNotExist(err) {
		return "", err
	}
	return pkgsDir, nil
}

// prepareImageStruct prepare an image structure to send to the backend
// to create or update an image
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param img: Image structure to fill
// :type img: *rpapiv1.RPImageFull
// :param isCreation: Indicate if we are creating the image
// :type isCreation: bool
// :param isImport: Indicate if we are importing the image
// :type isImport: bool
// :returns: error -- an error if a problem has occurred
func prepareImageStruct(cmd *cobra.Command, args []string, img *rpapiv1.RPImageFull, isCreation, isImport bool) error {

	// Get the name of the image wanted by the user (mandatory)
	imgName, err := cmd.Flags().GetString(imagesCmdNameFlag)
	if err != nil {
		logFlagError(imagesCmdNameFlag)
		return err
	}

	if isCreation && imgName == "" {
		Log.Error(imgNameFlagNotSetShortMsg)
		return errors.New(imgNameFlagNotSetLongMsg)
	}

	var imgAlias string
	if isCreation {
		// Get the alias of the image wanted by the user (optional)
		imgAlias, err = cmd.Flags().GetString(imagesCmdAliasFlag)
		if err != nil {
			logFlagError(imagesCmdAliasFlag)
			return err
		}
	}

	// Get the name of the distribution wanted by the user (mandatory)
	imgDistro, err := cmd.Flags().GetString(imagesCmdDistroFlag)
	if err != nil {
		logFlagError(imagesCmdDistroFlag)
		return err
	}

	imgDistroPrj := rpapiv1.RPProjectFull{RPProject: rpapiv1.RPProject{ID: uuid.Nil}}
	if imgDistro != flagUpdateKeepCurrentValue {
		// If the user gives something, try to get the corresponding project
		imgDistroPrj, err = getProjectFromIdentifier(imgDistro, false)
		if err != nil {
			return err
		}
	}

	// Get the architecture wanted by the user (mandatory)
	imgArch, err := cmd.Flags().GetString(imagesCmdArchFlag)
	if err != nil {
		logFlagError(imagesCmdArchFlag)
		return err
	}

	// Get the security model wanted by the user (optional)
	imgSecu, err := cmd.Flags().GetString(imagesCmdSecurityFlag)
	if err != nil {
		logFlagError(imagesCmdSecurityFlag)
		return err
	}

	// Get the board model wanted by the user (optional)
	imgBoardModel, err := cmd.Flags().GetString(imagesCmdBoardModelFlag)
	if err != nil {
		logFlagError(imagesCmdBoardModelFlag)
		return err
	}

	// Get the kickstart URL wanted by the user (optional)
	imgKickstartURL, err := cmd.Flags().GetString(imagesCmdKickstartURLFlag)
	if err != nil {
		logFlagError(imagesCmdKickstartURLFlag)
		return err
	}

	// Get the kickstart filename wanted by the user (optional)
	imgKickstartFile, err := cmd.Flags().GetString(imagesCmdKickstartNameFlag)
	if err != nil {
		logFlagError(imagesCmdKickstartNameFlag)
		return err
	}

	// Get the download url (optional)
	imgDownloadUrl := ""
	if isImport || !isCreation {
		imgDownloadUrl, err = cmd.Flags().GetString(imagesCmdDlURLFlag)
		if err != nil {
			logFlagError(imagesCmdDlURLFlag)
			return err
		}
	}

	var imgKickstartBranch string
	var imgKickstartPath string
	if !isImport {
		// Get the kickstart branch wanted by the user (optional)
		imgKickstartBranch, err = cmd.Flags().GetString(imagesCmdKickstartBranchFlag)
		if err != nil {
			logFlagError(imagesCmdKickstartBranchFlag)
			return err
		}

		// Get the kickstart path wanted by the user (optional)
		imgKickstartPath, err = cmd.Flags().GetString(imagesCmdKickstartPathFlag)
		if err != nil {
			logFlagError(imagesCmdKickstartPathFlag)
			return err
		}
	}

	// Get the additional repos wanted by the user (optional)
	imgAdditionalRepos, err := cmd.Flags().GetStringSlice(imagesCmdAddRepoFlag)
	if err != nil {
		logFlagError(imagesCmdAddRepoFlag)
		return err
	}

	noAutoRepos := false
	if !isImport {
		// Get the flag allowing the user to tell rp-cli to not set the automatic additional repos
		noAutoRepos, err = cmd.Flags().GetBool(imagesCmdNoAutoRepoFlag)
		if err != nil {
			logFlagError(imagesCmdNoAutoRepoFlag)
			return err
		}
	}

	// Get the projectID wanted by the user (optional)
	imgProjectId, err := cmd.Flags().GetString(imagesCmdPrjFlag)
	if err != nil {
		logFlagError(imagesCmdPrjFlag)
		return err
	}

	imgProjectUUID := uuid.Nil
	if !isCreation && (imgProjectId == flagUpdateKeepCurrentValue) {
		// Retrieve the UUID from the current image
		imgProjectUUID = img.ParentProjectID
	} else if imgProjectId != "" {
		// If a project ID is set, retrieve its UUID
		tmpPrj, err := getProjectFromIdentifier(imgProjectId, false)
		if err != nil {
			return err
		}
		imgProjectUUID = tmpPrj.ID
	}

	if isCreation {
		// Fill the structure with all the info
		img.Name = imgName
		img.Slug = imgAlias
		img.DistributionID = imgDistroPrj.ID
		img.DistributionSlug = imgDistroPrj.Slug
		img.Architecture = imgArch
		img.SecurityModel = imgSecu
		img.ModelBoard = imgBoardModel
		if isImport {
			img.KickURL = imgKickstartURL
		} else if imgKickstartURL != "" {
			img.KickURL = fmt.Sprintf("git+%s?%s#%s", imgKickstartURL, imgKickstartPath, imgKickstartBranch)
		}
		img.Kickstart = imgKickstartFile
		if isImport {
			img.URL = imgDownloadUrl
		}

	} else {
		if imgName != flagUpdateKeepCurrentValue {
			img.Name = imgName
		}
		if imgDistro != flagUpdateKeepCurrentValue {
			img.DistributionID = imgDistroPrj.ID
			img.DistributionSlug = imgDistroPrj.Slug
		}
		if imgArch != flagUpdateKeepCurrentValue {
			img.Architecture = imgArch
		}
		if imgSecu != flagUpdateKeepCurrentValue {
			img.SecurityModel = imgSecu
		}
		if (imgKickstartURL != flagUpdateKeepCurrentValue) || (imgKickstartPath != flagUpdateKeepCurrentValue) || (imgKickstartBranch != flagUpdateKeepCurrentValue) {
			// If we are here, the user wants to change something in the URL, we have to retrieve the different part from
			// the current URL in order to "reconstruct" it

			// Define the regexp we will use
			ksURLRegexp := `^(git\+)(https*://.*)\?(.*)#(.*)`
			regularExp := regexp.MustCompile(ksURLRegexp)

			// Now, we will use the different match to retrieve the different part
			matchGroups := regularExp.FindStringSubmatch(img.KickURL)
			if len(matchGroups) != 5 {
				errMsg := fmt.Sprintf(imgKsUrlDoNotMatchRegExp, img.KickURL, ksURLRegexp)
				return errors.New(errMsg)
			}

			tmpURL := matchGroups[2]
			tmpPath := matchGroups[3]
			tmpBranch := matchGroups[4]

			if imgKickstartURL != flagUpdateKeepCurrentValue {
				tmpURL = imgKickstartURL
			}
			if imgKickstartPath != flagUpdateKeepCurrentValue {
				tmpPath = imgKickstartPath
			}
			if imgKickstartBranch != flagUpdateKeepCurrentValue {
				tmpBranch = imgKickstartBranch
			}

			// Finally re-assemble the whole URL
			img.KickURL = fmt.Sprintf("git+%s?%s#%s", tmpURL, tmpPath, tmpBranch)
		}
		if imgKickstartFile != flagUpdateKeepCurrentValue {
			img.Kickstart = imgKickstartFile
		}
		if imgBoardModel != flagUpdateKeepCurrentValue {
			img.ModelBoard = imgBoardModel
		}
		if imgDownloadUrl != flagUpdateKeepCurrentValue {
			img.URL = imgDownloadUrl
		}
	}

	// Additional repositories management
	if !isCreation {
		// Get the '--reset-repo' flag
		resetRepo, err := cmd.Flags().GetBool(imageUpdateCmdResetReposFlag)
		if err != nil {
			logFlagError(imageUpdateCmdResetReposFlag)
			return err
		}

		// Reset the repository if requested by user
		if resetRepo && utils.ContainsString(imgAdditionalRepos, flagUpdateKeepCurrentValue) {
			imgAdditionalRepos = []string{}
		}
	}

	err = additionalReposManagement(isCreation, noAutoRepos, imgProjectUUID, imgAdditionalRepos, img)
	if err != nil {
		return err
	}

	// Get the '--force' flag
	forceFlag, err := cmd.Flags().GetBool(imagesCmdForceFlag)
	if err != nil {
		logFlagError(imagesCmdForceFlag)
		return err
	}

	// Check that the given repositories match with the image architecture (if '--force' not set)
	if !forceFlag {
		RPPrinter.CustomPrint(imagesAddCmdAnsLine2)

		// Do the check
		warnings, err := checkAdditionalReposArch(img)

		// According to the results, display correctly what needs to be print
		if err != nil {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		} else {
			RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
		}

		// Print all the warning messages
		for _, warn := range warnings {
			Log.Warn(warn)
		}

		// Finally print the error if there is one
		if err != nil {
			Log.Error((err.Error()))
			return err
		}
	}

	// Set other parameters
	img.ParentProjectID = imgProjectUUID

	return nil
}

// additionalReposManagement is a function allowing to manage the case of additional repositories
// in the images, since it can be a bit touchy
//
// :param isCreation: Indicate if we are creating the image
// :type isCreation: bool
// :param noAutoRepo: Flag forcing to not set automatic repos
// :type noAutoRepo: bool
// :param projectUUID: UUID of the project concerned
// :type projectUUID: uuid.UUID
// :param userAdditionalRepos: additional repositories given by the user
// :type userAdditionalRepos: []string
// :param img: Image structure to fill
// :type img: *rpapiv1.RPImageFull
// :returns: string, error -- the repository's URL and an error if a problem has occurred
func additionalReposManagement(isCreation, noAutoRepo bool, projectUUID uuid.UUID, userAdditionalRepos []string, img *rpapiv1.RPImageFull) error {

	if isCreation || (!utils.ContainsString(userAdditionalRepos, flagUpdateKeepCurrentValue)) {

		// Create the slice
		img.AdditionalRepos = []string{}

		// Possible correction of the length
		if !utils.UUIDEqual(projectUUID, uuid.Nil) && !noAutoRepo {
			// Retrieve all the repositories available for this project thanks to the "install_repo.sh" file
			repoURLs, err := getAllProjectRepositories(projectUUID, img.DistributionID, img.Architecture)
			if err != nil {
				return err
			}

			// Add all the repos to the additional repos field
			img.AdditionalRepos = append(img.AdditionalRepos, repoURLs...)
		}

		// Finally add the user repositories
		img.AdditionalRepos = append(img.AdditionalRepos, userAdditionalRepos...)
	}

	return nil
}

// checkAdditionalReposArch is a function allowing to check that the additional repositories furnished
// by the user are correctly matching the architecture of the image.
//
// :param img: Image structure to check
// :type img: *rpapiv1.RPImageFull
// :returns: []string, error -- warning messages to display and an error if the check went bad
func checkAdditionalReposArch(img *rpapiv1.RPImageFull) ([]string, error) {
	// Initialize the warnings array
	warnings := []string{}

	// Get the list of every available architecture
	archStr, err := getTabOfArchitectures()
	if err != nil {
		return warnings, err
	}

	// First, check that the architecture of the image exists
	archExists := utils.ContainsString(archStr, img.Architecture)
	if !archExists {
		errMsg := fmt.Sprintf(imgPrepareArchNotExisting, img.Architecture)
		return warnings, errors.New(errMsg)
	}

	// Delete the image architecture from the arch slice
	utils.DeleteStringFromSlice(&archStr, img.Architecture)

	// Then check that for each additional repository, we have the right architecture in it
	// Return an error if there is another architecture
	// Add a warning if no architecture is found
	for _, repo := range img.AdditionalRepos {

		// For this repos check if another architecture is seen
		otherArchInRepos := ""
		for _, arch := range archStr {
			if strings.Contains(repo, arch) {
				otherArchInRepos = arch
				break
			}
		}

		if otherArchInRepos != "" {
			errMsg := fmt.Sprintf(imgPrepareAnotherArchInRepos, otherArchInRepos, repo)
			return warnings, errors.New(errMsg)
		}

		// Check that the image architecture is in the additional repository
		if !strings.Contains(repo, img.Architecture) {
			// Add a warning in the warning slice
			wrngMsg := fmt.Sprintf(imgPrepareArchNotInRepoMsg, img.Architecture, repo)
			warnings = append(warnings, wrngMsg)
		}
	}

	return warnings, nil
}

// prepareRelImageStruct prepare the rpapiv1.RPReleaseImage structure needed to release an image.
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param relImg: Release Image structure to fill
// :type relImg: *rpapiv1.RPReleaseImage
// :param isCreation: Indicate if we are creating the project
// :type isCreation: bool
// :returns: error -- an error if a problem has occurred
func prepareRelImageStruct(cmd *cobra.Command, relImg *rpapiv1.RPReleaseImage, isCreation bool) error {

	// First, take care of name and description
	// ----------------------------------------

	// Name of the project release
	relImgName, err := cmd.Flags().GetString(imageReleaseCmdNameFlag)
	if err != nil {
		logFlagError(imageReleaseCmdNameFlag)
		return err
	}

	// Then fill the information we got in the expected structure
	// ----------------------------------------------------------

	if isCreation {
		// Name and description management
		relImg.Name = relImgName
	} else {
		if relImgName != flagUpdateKeepCurrentValue {
			relImg.Name = relImgName
		}
	}

	return nil
}

// getAllProjectRepositories is a function allowing to get all the project repositories (parent and distros, etc.) from
// the project UUID, the distribution and the architecture for the repository
//
// :param projectUUID: UUID of the project concerned
// :type projectUUID: uuid.UUID
// :param distroUUID: UUID of the project concerned
// :type distroUUID: uuid.UUID
// :param architecture: Architecture wanted for the repository
// :type architecture: string
// :returns: []string, error -- the list of repositories' URL and an error if a problem has occurred
func getAllProjectRepositories(projectUUID uuid.UUID, distroUUID uuid.UUID, architecture string) ([]string, error) {
	result := []string{}

	// Firstly, get the project structure
	prj := rpapiv1.RPProjectFull{}
	if err := bckAPI.ProjectGet(projectUUID.String(), &prj); err != nil {
		return result, err
	}

	// Get the repository that is interesting us
	repoURL := ""
	for _, currRepo := range prj.Repos {
		if utils.UUIDEqual(currRepo.Distro.ID, distroUUID) {
			repoURL = strings.ReplaceAll(currRepo.Build, "$basearch", "")
			break
		}
	}

	// If not repo has been found, return an error
	if repoURL == "" {
		errMsg := fmt.Sprintf(imgDistroNotFoundInPrjErr, distroUUID.String(), prj.Slug)
		return result, errors.New(errMsg)
	}

	// Get the token to access this repo
	jwtToken := prj.ProjectAccessTokens[0].JWT
	installRepoURLWithToken := fmt.Sprintf("%s/install_repo.sh?token=%s", repoURL, jwtToken)

	// Download the file in /tmp to extract the basURL afterwards
	err := utils.HTTPDwnldFileInDir("/tmp", installRepoURLWithToken)
	if err != nil {
		return result, err
	}

	return utils.ExtractBaseURLvalues("/tmp/install_repo.sh", architecture)
}

// checkImageParametersForBuild is a private function allowing to check
// if every parameters needed to build locally an image is contains inside
// the retrieved RPImage and an eventual local kickstart file
//
// :param image: Pointer to the image structure that should contain the parameters
// :type image: *rpapiv1.RPImage
// :param localKickstart: Kickstart file flag eventually furnished by the user
// :type localKickstart: string
// :returns: string, error -- the full path to the kickstart and an error if something went bad
func checkImageParametersForBuild(image *rpapiv1.RPImage, localKickstart string) (string, error) {
	var err error

	if image.Slug == "" || image.DistributionID == uuid.Nil ||
		image.Architecture == "" || image.InstallTreeURL == "" {
		return "", errors.New(imgBuildMissArguments1)
	}

	if localKickstart == "" && (image.KickURL == "" || image.Kickstart == "") {
		return "", errors.New(imgBuildMissArguments2)
	}

	fullKickstartFile := ""
	if localKickstart != "" {
		// Check that the local kickstart file exists
		fullKickstartFile, err = utils.SanitizePath(localKickstart)
		if err != nil {
			return "", err
		}

		if _, err := os.Stat(fullKickstartFile); os.IsNotExist(err) {
			return "", errors.New(imgBuildNoLocalKickFileError)
		}
	}

	return fullKickstartFile, nil
}

// setUpLocalImageBuildDir is a private function allowing to set-up the directory
// that will be used to build locally the image
//
// :param image: Pointer to the image structure that contains the needed parameters
// :type image: *rpapiv1.RPImage
// :param localKickstart: Kickstart file eventually furnished by the user
// :type localKickstart: string
// :returns: string, error -- the full path to the directory and an error if something went bad
func setUpLocalImageBuildDir(image *rpapiv1.RPImage, localKickstart string) (string, error) {

	// Get the  gitpkgs/ directory in the host
	pkgsDir, err := getFullGitpkgsDirPath()
	if err != nil {
		return "", err
	}

	// Construct the directory name for the image
	imageBuildDir := path.Join(pkgsDir, ("image_localbuild_" + image.Slug))

	// Create the directory if it does not exist and overwrite it if it does
	err = utils.CreateDirIfNotExists(imageBuildDir, true)
	if err != nil {
		return "", err
	}

	// If a local kickstart has been given, copy it inside the new directory
	if localKickstart != "" {
		// Get only the name of the file
		kickstartName := path.Base(localKickstart)
		// Construct the full path to the copy of the kickstart
		kickstartInImgDir := path.Join(imageBuildDir, kickstartName)

		// Copy the kickstart in the directory made for image
		err = utils.CopyFile(localKickstart, kickstartInImgDir)
		if err != nil {
			return "", err
		}
	}

	return imageBuildDir, nil
}

// runLocalImageBuild is a private function allowing to start the build of an image
// inside the local builder from rp-cli, when requested by the user
//
// :param image: Pointer to the image structure that contains the needed parameters
// :type image: *rpapiv1.RPImage
// :param workingDir: Working directory, set-up to build the image
// :type workingDir: string
// :param kickstartFile: Kickstart file eventually furnished by the user
// :type kickstartFile: string
// :param imageSize: Image size wanted by the user (in gigabytes)
// :type imageSize: int
// :returns: error -- the error resulting from the build (nil if OK)
func runLocalImageBuild(image *rpapiv1.RPImage, workingDir, localKickstart string, imageSize int) error {

	// Get the  gitpkgs/ directory in the host
	pkgsDir, err := getFullGitpkgsDirPath()
	if err != nil {
		return err
	}

	// First, map the directory gitpkgs/ from host to the one in local-builder
	localBuilderWorkingDir := strings.ReplaceAll(workingDir, pkgsDir, getLclBldGitpkgsDir())

	// Construct the version we want from what the user gives us, if it is possible
	// For the moment we will force the version to one since it is just "local"
	finalVersion := "1"

	// Construct the command line to run in the local-builder
	cmdLine := "iotpkg -v localimagebuild"
	// Add the name
	cmdLine = strings.Join([]string{cmdLine, "--name", image.Slug}, " ")
	// Add the distribution name
	cmdLine = strings.Join([]string{cmdLine, "--distname", "Redpesk"}, " ")
	// Add the distribution version
	cmdLine = strings.Join([]string{cmdLine, "--distver", finalVersion}, " ")
	// Add the image architecture
	cmdLine = strings.Join([]string{cmdLine, "--arch", image.Architecture}, " ")
	// Add the install tree for the image
	cmdLine = strings.Join([]string{cmdLine, "--inst_tree", image.InstallTreeURL}, " ")
	// Add the kickstart file parameters according if it is local or not
	if image.KickURL != "" {
		// Fix #2881: it is a workaround for iotpkg to work systematically
		iotpkgKsURL := strings.ReplaceAll(image.KickURL, "?#", "?.#")

		cmdLine = strings.Join([]string{cmdLine, "--ksurl", iotpkgKsURL}, " ")
	}
	if localKickstart == "" {
		cmdLine = strings.Join([]string{cmdLine, "--kickstart", image.Kickstart}, " ")
	} else {
		kickstartInLocalBuilder := path.Join(localBuilderWorkingDir, localKickstart)
		cmdLine = strings.Join([]string{cmdLine, "--kickstart", kickstartInLocalBuilder}, " ")
	}
	// Add the potential additional repos
	for _, repo := range image.AdditionalRepos {
		if repoArg := strings.Trim(repo, " "); repoArg != "" {
			cmdLine = strings.Join([]string{cmdLine, "--repo", repoArg}, " ")
		}
	}
	// Add the size of the image as flag, if set by the user
	if imageSize > 0 {
		cmdLine = strings.Join([]string{cmdLine, "--size", strconv.Itoa(imageSize)}, " ")
	}

	// Get the size of the current terminal
	termWidth, termHeight, err := term.GetSize(int(os.Stdin.Fd()))
	if err != nil {
		return err
	}

	// Request local-builder to run the command
	// Request body
	execRequest := ladapi.RPLADExecCmdStruct{
		RPLADMainCmdParams: ladapi.RPLADMainCmdParams{
			RPLADWinSize: ladapi.RPLADWinSize{
				WinSizeRows: termHeight,
				WinSizeCol:  termWidth,
			},
			WorkingDir: localBuilderWorkingDir,
		},
		CmdLine: cmdLine,
	}
	// Answer body template
	procTicket := ladapi.RPLADRunnerTicket{}
	err = localBAPI.ExecPost(&execRequest, &procTicket)
	if err != nil {
		return err
	}

	err = listenStdoutErrAndIn(procTicket)

	return err
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Display functions --------------------------------------------
// --------------------------------------------------------------

// displayImages is a private function allowing to display
// a slice of images in the console
//
// :param images: tab of images to display
// :type images: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func displayImages(images interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Convert to the right type
	imgs := images.(rpapiv1.RPImages)

	// Display result
	first := true

	for _, image := range imgs.Images {
		if verbose {
			if !first {
				fmt.Fprintln(writer)
			}
			// Print this image
			_printOneImage(image, verbose, false, writer)

		} else {
			if first {
				manageHeaderPrint(imagesListCmdHeader, writer)
			}
			// Print this image
			_printOneImage(image, verbose, false, writer)
		}
		first = false
	}
	return nil
}

// displayOneImage is a private function allowing to display
// one image in the console
//
// :param img: Image to display
// :type img: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :returns: error -- an error if something went bad
func displayOneImage(img interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Print the image
	_printOneImage(img, verbose, true, writer)
	return nil
}

// _printOneImage is a private function allowing to print one image in the console
//
// :param img: image to print
// :type img: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param isAlone: True if there is only one image to display in the context where this function is called
// :type isAlone: bool
// :param writer: Writer to used to print the image information
// :type writer: *tabwriter.Writer
func _printOneImage(img interface{}, verbose bool, isAlone bool, writer *tabwriter.Writer) {

	image := rpapiv1.RPImageFull{}

	// Cast to the full type in all case but use "full" parameters as indicated by the user
	// We have to pass through the Marshal/Unmarshal tricks to not have problem during conversion from
	// RPImage -> RPImageFull at runtime
	dataBytes, _ := json.Marshal(img)
	json.Unmarshal(dataBytes, &image)

	// Create the string for type
	typeToDisplay := image.Type
	if typeToDisplay == "external" {
		typeToDisplay = "iotbzh (external)"
	}

	if verbose {
		// Create the string containing the additional repositories
		addiRepos := ""
		if len(image.AdditionalRepos) > 0 {
			for _, repo := range image.AdditionalRepos {
				if addiRepos != "" {
					addiRepos += ", "
				}
				addiRepos += repo
			}
		}

		if addiRepos == "" {
			addiRepos = mainNotSetMsg
		}

		downloadURL := mainNotSetMsg
		if image.URL != "" {
			downloadURL = image.URL
		}

		kickURL := mainNotSetMsg
		if image.KickURL != "" {
			kickURL = image.KickURL
		}

		kickstartName := mainNotSetMsg
		if image.Kickstart != "" {
			kickstartName = image.Kickstart
		}

		installTree := mainNotSetMsg
		if image.InstallTreeURL != "" {
			installTree = image.InstallTreeURL
		}

		boardModel := mainNotSetMsg
		if image.ModelBoard != "" {
			boardModel = image.ModelBoard
		}

		// Construct the access token string if full data
		accessTokensStr := ""
		if image.FullData {
			for index, token := range image.ImageAccessTokens {
				if index > 0 {
					accessTokensStr += ", "
				}
				accessTokensStr += token.JWT
			}
		}

		parentProjectIDStr := mainNotSetMsg
		if image.ParentProjectID != uuid.Nil {
			parentProjectIDStr = image.ParentProjectID.String()
		}

		parentProjectSlug := mainNotSetMsg
		if image.ParentProjectSlug != "" {
			parentProjectSlug = image.ParentProjectSlug
		}

		// Print the result on stdout
		fmt.Fprintln(writer, imagesListCmdVerboseLine0, image.ID.String())
		fmt.Fprintln(writer, imagesListCmdVerboseLine1, image.Name)
		fmt.Fprintln(writer, imagesListCmdVerboseLine2, image.Slug)
		fmt.Fprintln(writer, imagesListCmdVerboseLine7, typeToDisplay)
		fmt.Fprintln(writer, imagesListCmdVerboseLine3, image.DistributionSlug)
		fmt.Fprintln(writer, imagesListCmdVerboseLine5, image.Architecture)
		fmt.Fprintln(writer, imagesListCmdVerboseLine15, boardModel)
		fmt.Fprintln(writer, imagesListCmdVerboseLine6, image.SecurityModel)
		fmt.Fprintln(writer, imagesListCmdVerboseLine8, downloadURL)
		// If we have an access token, print it here
		if accessTokensStr != "" {
			fmt.Fprintln(writer, imagesListCmdVerboseLine19, accessTokensStr)
		}
		fmt.Fprintln(writer, imagesListCmdVerboseLine9, kickURL)
		fmt.Fprintln(writer, imagesListCmdVerboseLine10, kickstartName)
		fmt.Fprintln(writer, imagesListCmdVerboseLine11, installTree)
		fmt.Fprintln(writer, imagesListCmdVerboseLine12, addiRepos)
		fmt.Fprintln(writer, imagesListCmdVerboseLine13, parentProjectIDStr)
		fmt.Fprintln(writer, imagesListCmdVerboseLine20, parentProjectSlug)
		fmt.Fprintln(writer, imagesListCmdVerboseLine14, strconv.FormatBool(image.Disabled))
	} else {
		if isAlone {
			fmt.Fprintln(writer, imagesListCmdHeader)
		}
		fmt.Fprintln(writer, image.ID.String(), "\t", image.Slug, "\t", image.Name, "\t", image.Architecture, "\t", typeToDisplay, "\t", strconv.FormatBool(image.Disabled))
	}
}

// displayRelImages is a private function allowing to display
// a slice of released images in the console
//
// :param images: tab of images to display
// :type images: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func displayRelImages(images interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Convert to the right type
	relImgs := images.([]rpapiv1.RPReleaseImage)

	// Display result
	for index, image := range relImgs {
		if verbose && (index > 0) {
			fmt.Fprintln(writer)
			fmt.Fprintln(writer)
		} else if !verbose && index == 0 {
			manageHeaderPrint(imageListReleasedCmdHeader, writer)
		}
		// Print this image
		_printOneImageReleased(image, verbose, false, writer)
	}
	return nil
}

// displayOneRelImage is a private function allowing to display
// one released image in the console
//
// :param img: Image to display
// :type img: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :returns: error -- an error if something went bad
func displayOneRelImage(img interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Print the project
	_printOneImageReleased(img, verbose, true, writer)
	return nil
}

// _printOneImageReleased is a private function allowing to print one released image in the console
//
// :param imgRel: image to print (rpapiv1.RPReleaseImageFull or rpapiv1.RPReleaseImage)
// :type imgRel: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param isAlone: True if there is only one project to display in the context where this function is called
// :type isAlone: bool
// :param writer: Writer to used to print the project information
// :type writer: *tabwriter.Writer
func _printOneImageReleased(imgRel interface{}, verbose, isAlone bool, writer *tabwriter.Writer) {
	releasedImg := rpapiv1.RPReleaseImageFull{}

	// Cast to the full type in all case but use "full" parameters as indicated by the user
	// We have to pass through the Marshal/Unmarshal tricks to not have problem during conversion from
	// RPReleaseImage -> RPReleaseImageFull at runtime
	dataBytes, _ := json.Marshal(imgRel)
	json.Unmarshal(dataBytes, &releasedImg)

	if verbose {
		// Display the released image in verbose mode

		// Display the release project ID details
		fmt.Fprintln(writer, imagesListCmdVerboseLine0, releasedImg.ID.String())
		fmt.Fprintln(writer, imagesListCmdVerboseLine1, releasedImg.Name)
		fmt.Fprintln(writer, imagesListCmdVerboseLine2, releasedImg.Slug)

		// Then, display the project reference
		fmt.Fprintln(writer, imageGetReleasedCmdVerboseLine0, releasedImg.ImageReferenceID.String())
		fmt.Fprintln(writer, imageGetReleasedCmdVerboseLine1, releasedImg.ImageReferenceSlug)
		fmt.Fprintln(writer, imageGetReleasedCmdVerboseLine2, releasedImg.DistributionID.String())
		fmt.Fprintln(writer, imagesListCmdVerboseLine5, releasedImg.Architecture)

		// After, display more details
		fmt.Fprintln(writer, imagesListCmdVerboseLine4, releasedImg.Version)
		fmt.Fprintln(writer, imagesListCmdVerboseLine16, releasedImg.CreatedAt.String())
		fmt.Fprintln(writer, imagesListCmdVerboseLine17, releasedImg.UpdatedAt.String())
		fmt.Fprintln(writer, imageGetReleasedCmdVerboseLine7, releasedImg.URL)
		if releasedImg.FullData {
			// Construct the access token string if full data
			accessTokensStr := ""
			for index, token := range releasedImg.ImageAccessTokens {
				if index > 0 {
					accessTokensStr += ", "
				}
				accessTokensStr += token.JWT
			}
			// If we have an access token, print it here
			if accessTokensStr != "" {
				fmt.Fprintln(writer, imagesListCmdVerboseLine19, accessTokensStr)
			}
		}
		fmt.Fprintln(writer, imageGetReleasedCmdVerboseLine8, strconv.FormatBool(releasedImg.ReadOnly))

		if releasedImg.FullData {
			fmt.Fprintln(writer)
			fmt.Fprintln(writer, prjsGetReleasedCmdVerboseLine9)
			fmt.Fprintln(writer)
			// Create the json string
			jsonObj, _ := json.MarshalIndent(releasedImg.ReleaseInfo, "", "    ")
			// Print the pretty json string
			fmt.Fprintln(writer, string(jsonObj))
		}

	} else {
		// Display essential info
		if isAlone {
			fmt.Fprintln(writer, imageListReleasedCmdHeader)
		}
		fmt.Fprintln(writer, releasedImg.ID.String(), "\t", releasedImg.Slug, "\t", releasedImg.Version, "\t", releasedImg.ImageReferenceSlug)
	}
}

// --------------------------------------------------------------
// --------------------------------------------------------------
