/*
Copyright © 2021-2023 IoT.bzh Company
Contact: https://www.iot.bzh/licensing

This file is part of the redpesk-cli project, the redpesk command line tool.

$RP_BEGIN_LICENSE$
Commercial License Usage
 Licensees holding valid commercial IoT.bzh licenses may use this file in
 accordance with the commercial license agreement provided with the
 Software or, alternatively, in accordance with the terms contained in
 a written agreement between you and The IoT.bzh Company. For licensing terms
 and conditions see https://www.iot.bzh/terms-conditions. For further
 information use the contact form at https://www.iot.bzh/contact.

GNU General Public License Usage
 Alternatively, this file may be used under the terms of the GNU General
 Public license version 3. This license is as published by the Free Software
 Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
 of this file. Please review the following information to ensure the GNU
 General Public License requirements will be met
 https://www.gnu.org/licenses/gpl-3.0.html.
$RP_END_LICENSE$
*/

package cmd

import (
	"fmt"
	"path"
	"strings"
	"testing"
	"time"

	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/rpbckmock"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/rpladmock"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/utils"
	"git.ovh.iot/redpesk/redpesk-infra/rp-webserver/actions/rpapiv1"
	"github.com/stretchr/testify/assert"
)

const (
	// Images
	imgDisplayNoVerboseColNb    = 6
	imgDisplayVerboseFullLineNb = 17

	// Images released
	imgRelDisplayNoVerboseColNb            = 4
	imgRelDisplayVerboseFullLineNb         = 20
	imgRelDisplayVerboseFullAudienceLineNb = imgRelDisplayVerboseFullLineNb + 6

	// Application tests stats
	imgTestsStatsNoVerboseLineNb = 6
	imgTestsStatsVerboseColNb    = 3
)

// TestImagesAddNoBlock is the function allowing to test if
// the command "rp-cli images add --nonblocking ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestImagesAddNoBlock(test *testing.T) {

	// Create the command
	cmdToRun := getMainAddImgCmd("Test Image x86_64", "test-img-x86-64", "x86_64", "redpesk-devel-arz")

	// Add the non-blocking flag
	cmdToRun = append(cmdToRun, "--nonblocking")

	// Create the image if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, imagesCmdName, "test-img-x86-64", cmdToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// The image has been created, check it is now in backend

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID":           `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Slug":         `^test-img-x86-64$`,
		"Name":         `^Test Image x86_64$`,
		"Architecture": `^x86_64$`,
		"Type":         `^standard$`,
		"Disabled":     `^false$`,
	}

	// Run the command and the tests (no verbose and verbose, since this is the same result)
	checkCmdAnsKeysRegexpCol(test, imgDisplayNoVerboseColNb, regexpMap, imagesCmdName, "get", "test-img-x86-64")
}

// TestImagesAddBlock is the function allowing to test if
// the command "rp-cli images add ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestImagesAddBlock(test *testing.T) {
	// Create the command
	cmdToRun := getMainAddImgCmd("Test Image aarch64", "test-img-aarch64", "aarch64", "redpesk-devel-arz")

	// Create the image if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, imagesCmdName, "test-img-aarch64", cmdToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// The image has been created, check it is now in backend

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID":           `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Slug":         `^test-img-aarch64$`,
		"Name":         `^Test Image aarch64$`,
		"Architecture": `^aarch64$`,
		"Type":         `^standard$`,
		"Disabled":     `^false$`,
	}

	// Run the command and the tests (no verbose and verbose, since this is the same result)
	checkCmdAnsKeysRegexpCol(test, imgDisplayNoVerboseColNb, regexpMap, imagesCmdName, "get", "test-img-aarch64")
}

// TestImagesAddBlockFullParams is the function allowing to test if
// the command "rp-cli images add ..." is working correctly, when all the
// parameter possible are set
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestImagesAddBlockFullParams(test *testing.T) {
	// Create the command
	cmdToRun := getMainAddImgCmd("Test Image x86_64 Selinux", "test-img-x86-64-selinux", "x86_64", "redpesk-devel-arz",
		"-k", "Redpesk-core-x86_64-selinux.ks",
		"--ks-url", "http://git.ovh.iot/redpesk/redpesk-infra/imager-configuration.git",
		"--ks-branch", "02d883f4",
		"--ks-path", "kickstarts/RedPesk",
		"-r", "https://download.redpesk.bzh/redpesk-devel/releases/arz/Everything/x86_64/os/",
		"-r", "https://download.redpesk.bzh/redpesk-devel/releases/arz/AGL/x86_64/os/")

	// Create the image if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, imagesCmdName, "test-img-x86-64-selinux", cmdToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// The image has been created, check it is now in backend

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID":           `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Slug":         `^test-img-x86-64-selinux$`,
		"Name":         `^Test Image x86_64 Selinux$`,
		"Architecture": `^x86_64$`,
		"Type":         `^standard$`,
		"Disabled":     `^false$`,
	}

	// Run the command and the tests (no verbose and verbose, since this is the same result)
	checkCmdAnsKeysRegexpCol(test, imgDisplayNoVerboseColNb, regexpMap, imagesCmdName, "get", "test-img-x86-64-selinux")
}

// TestImagesAddWithProject is the function allowing to test if
// the command "rp-cli images add ..." is working correctly, when a project
// is linked with it
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestImagesAddWithProject(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Test Project 4 Pierre", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "test-project-4-pierre", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the command to add the image
	cmdToRun2 := getMainAddImgCmd("Test Image pour Pierre", "test-img-pierre", "aarch64", "redpesk-devel-arz",
		"-p", "test-project-4-pierre")

	// Create the image if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, imagesCmdName, "test-img-pierre", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// The image has been created, check it is now in backend

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":                 `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":                `^Test Image pour Pierre$`,
		"Slug":                `^test-img-pierre$`,
		"Type":                `^standard$`,
		"Distribution":        `^redpesk-devel-arz$`,
		"Architecture":        `^aarch64$`,
		"Board model":         `^generic$`,
		"Security model":      `^smack$`,
		"Download URL":        `^\[Not set\]$`,
		"Access token(s)":     `(_*[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}){2}`,
		"Kickstart URL":       `^git\+https*://`,
		"Kickstart name":      `^Redpesk-minimal-x86_64-smack.ks$`,
		"Install tree URL":    `^https://download.redpesk.bzh/redpesk-lts/arz-1.0/imager-os/aarch64/$`,
		"Additional repos":    `^(https?://.+, )(https?://.+)$`,
		"Parent project ID":   `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Parent project slug": `^test-project-4-pierre$`,
		"Disabled":            `^false$`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, imgDisplayVerboseFullLineNb, regexpMap, imagesCmdName, "get", "test-img-pierre", "-v")
}

// TestImagesAddWithBoardModel is the function allowing to test if
// the command "rp-cli images add ..." is working correctly, when a board model
// is specified
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestImagesAddWithBoardModel(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Test Project 4 Elisa", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "test-project-4-elisa", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the command to add the image
	cmdToRun2 := getMainAddImgCmd("Test Image pour Elisa", "test-img-elisa", "aarch64", "redpesk-devel-arz",
		"-p", "test-project-4-elisa", "--model-board", "raspberry-pi-4")

	// Create the image if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, imagesCmdName, "test-img-elisa", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// The image has been created, check it is now in backend

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":                 `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":                `^Test Image pour Elisa$`,
		"Slug":                `^test-img-elisa$`,
		"Type":                `^standard$`,
		"Distribution":        `^redpesk-devel-arz$`,
		"Architecture":        `^aarch64$`,
		"Board model":         `^raspberry-pi-4$`,
		"Security model":      `^smack$`,
		"Download URL":        `^\[Not set\]$`,
		"Access token(s)":     `(_*[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}){2}`,
		"Kickstart URL":       `^git\+https*://`,
		"Kickstart name":      `^Redpesk-minimal-x86_64-smack.ks$`,
		"Install tree URL":    `^https://download.redpesk.bzh/redpesk-lts/arz-1.0/imager-os/aarch64/$`,
		"Additional repos":    `^(https?://.+, )(https?://.+)$`,
		"Parent project ID":   `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Parent project slug": `^test-project-4-elisa$`,
		"Disabled":            `^false$`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, imgDisplayVerboseFullLineNb, regexpMap, imagesCmdName, "get", "test-img-elisa", "-v")
}

// TestImagesAddCheckAdditionalRepos is the function allowing to test if
// the command "rp-cli images add ..." is working correctly, especially the feature
// allowing to check that the additional repository is corresponding to the image
// architecture
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestImagesAddCheckAdditionalRepos(test *testing.T) {
	// Create the command
	cmdToRun := getMainAddImgCmd("Test Check Image aarch64", "test-check-img-aarch64", "aarch64", "redpesk-devel-arz",
		"-r", "https://download.redpesk.bzh/redpesk-devel/releases/arz/Everything/x86_64/os/")

	// Check that rp-cli is given us an error
	_, err := executeCommand(cmdToRun...)

	// Test that we have an error
	assert.Error(test, err)
	assert.Contains(test, err.Error(), "another architecture (\"x86_64\") has been detected in the following additional repository")

	// Create another command where the additional repository does not have an architecture
	cmdToRun = getMainAddImgCmd("Test Check Image aarch64", "test-check-img-aarch64", "aarch64", "redpesk-devel-arz",
		"-r", "https://download.redpesk.bzh/redpesk-devel/releases/arz/Everything/os/", "-L", "warning")

	// Check that rp-cli accept but output a warning
	cmdOutput := checkActionCmdAns(test, cmdToRun...)
	assert.Contains(test, cmdOutput, "warning")
	assert.Contains(test, cmdOutput, "the following repository does not contain the \\\"aarch64\\\" image architecture:")

	// Check that it is possible to force the creation of a possible problematic image
	cmdToRun = getMainAddImgCmd("Test Check Image aarch64 Force", "test-check-img-aarch64-force", "aarch64", "redpesk-devel-arz",
		"-r", "https://download.redpesk.bzh/redpesk-devel/releases/arz/Everything/x86_64/os/", "--force")

	checkActionCmdAns(test, cmdToRun...)
}

// TestImagesAddCheckNoAutoRepos is the function allowing to test if
// the command "rp-cli images add ..." is working correctly, when a project
// is linked with it and the user tries to use the '--no-auto-repos' flag
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestImagesAddCheckNoAutoRepos(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Test Project 4 Advanced users", "redpesk-lts-arz")
	err := addObjectIfNotExists(test, prjsCommand, "test-project-4-advanced-users", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the command to add the image
	cmdToRun2 := getMainAddImgCmd("Test Image Advanced Users", "test-img-advanced-users", "aarch64", "redpesk-lts-arz",
		"-p", "test-project-4-advanced-users", "--no-auto-repos", "-r", "http://my-only-repos.fr")

	// Create the image if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, imagesCmdName, "test-img-advanced-users", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// The image has been created, check it is now in backend

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":                 `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":                `^Test Image Advanced Users$`,
		"Slug":                `^test-img-advanced-users$`,
		"Type":                `^standard$`,
		"Distribution":        `^redpesk-lts-arz$`,
		"Architecture":        `^aarch64$`,
		"Board model":         `^generic$`,
		"Security model":      `^smack$`,
		"Download URL":        `^\[Not set\]$`,
		"Access token(s)":     `(_*[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}){2}`,
		"Kickstart URL":       `^git\+https*://`,
		"Kickstart name":      `^Redpesk-minimal-x86_64-smack.ks$`,
		"Install tree URL":    `^https://download.redpesk.bzh/redpesk-lts/arz-1.0/imager-os/aarch64/$`,
		"Additional repos":    `^http://my-only-repos.fr$`,
		"Parent project ID":   `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Parent project slug": `^test-project-4-advanced-users$`,
		"Disabled":            `^false$`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, imgDisplayVerboseFullLineNb, regexpMap, imagesCmdName, "get", "test-img-advanced-users", "-v")
}

// TestImagesList is the function allowing to test if
// the command "rp-cli images list ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestImagesList(test *testing.T) {

	// Create the commands
	cmdToRun1 := getMainAddImgCmd("Imx8 Image aarch64", "imx8-img-aarch64", "aarch64", "redpesk-devel-arz")
	cmdToRun2 := getMainAddImgCmd("Generic Image aarch64", "generic-img-aarch64", "aarch64", "redpesk-devel-arz")
	cmdToRun3 := getMainAddImgCmd("Generic Image x86_64", "generic-img-x86-64", "x86_64", "redpesk-devel-arz")

	// Create the images if they do not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, imagesCmdName, "imx8-img-aarch64", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, imagesCmdName, "generic-img-aarch64", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, imagesCmdName, "generic-img-x86-64", cmdToRun3...)
	if err != nil {
		test.Fatal(err)
	}

	// The images have been created, check it is now in backend

	// Not verbose mode
	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID":           `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Slug":         `^[a-z0-9-]+$`,
		"Name":         `.+`,
		"Architecture": `(x86_64|aarch64)`,
		"Type":         `^(iotbzh|standard)`,
		"Disabled":     `^false$`,
	}

	// Run the command and the tests (no verbose and verbose, since this is the same result)
	checkCmdAnsKeysRegexpCol(test, imgDisplayNoVerboseColNb, regexpMap, imagesCmdName, "list")

	// Verbose mode
	// Create the map containing the regexp we want to match in the answer
	regexpMap2 := map[string]string{
		"ID:":                 `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":                `.+`,
		"Slug":                `^[a-z0-9-]+$`,
		"Type":                `^(iotbzh|standard)`,
		"Distribution":        `^redpesk-(devel|lts)-arz$`,
		"Architecture":        `^(x86_64|aarch64)$`,
		"Board model":         `^(generic|\[Not set\]|raspberry-pi-4)$`,
		"Security model":      `^(smack|selinux)$`,
		"Download URL":        `.+`,
		"Kickstart URL":       `.+`,
		"Kickstart name":      `.+`,
		"Install tree URL":    `.+`,
		"Additional repos":    `.+`,
		"Parent project ID":   `([a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}|\[Not set\])`,
		"Parent project slug": `.+`,
		"Disabled":            `^false$`,
	}

	// Run the command and the tests
	checkCmdAnswerListKeysRegexp(test, imgDisplayVerboseLightLineNb, "", regexpMap2, imagesCmdName, "list", "-v")
}

// TestImagesListType specified is the function allowing to test if
// the command "rp-cli images list ..." is working correctly when a
// type is specified in the command line
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestImagesListType(test *testing.T) {

	// Create the commands
	cmdToRun1 := getMainAddImgCmd("Imx8 Image aarch64", "imx8-img-aarch64", "aarch64", "redpesk-devel-arz")
	cmdToRun2 := getMainAddImgCmd("Generic Image aarch64", "generic-img-aarch64", "aarch64", "redpesk-devel-arz")
	cmdToRun3 := getMainAddImgCmd("Generic Image x86_64", "generic-img-x86-64", "x86_64", "redpesk-devel-arz")

	// Create the images if they do not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, imagesCmdName, "imx8-img-aarch64", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, imagesCmdName, "generic-img-aarch64", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, imagesCmdName, "generic-img-x86-64", cmdToRun3...)
	if err != nil {
		test.Fatal(err)
	}

	// The images have been created, check it is now in backend

	// Not verbose mode
	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID":           `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Slug":         `^[a-z0-9-]+$`,
		"Name":         `.+`,
		"Architecture": `(x86_64|aarch64)`,
		"Type":         `^iotbzh`,
		"Disabled":     `^false$`,
	}

	// Run the command and the tests (no verbose and verbose, since this is the same result)
	checkCmdAnsKeysRegexpCol(test, imgDisplayNoVerboseColNb, regexpMap, imagesCmdName, "list", "--type", "iotbzh")

	// Verbose mode
	// Create the map containing the regexp we want to match in the answer
	regexpMap2 := map[string]string{
		"ID:":                 `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":                `.+`,
		"Slug":                `^[a-z0-9-]+$`,
		"Type":                `^standard$`,
		"Distribution":        `^redpesk-(devel|lts)-arz$`,
		"Architecture":        `^(x86_64|aarch64)$`,
		"Board model":         `^(generic|raspberry-pi-4)$`,
		"Security model":      `^(smack|selinux)$`,
		"Download URL":        `.+`,
		"Kickstart URL":       `.+`,
		"Kickstart name":      `.+`,
		"Install tree URL":    `.+`,
		"Additional repos":    `.+`,
		"Parent project ID":   `([a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}|\[Not set\])`,
		"Parent project slug": `.+`,
		"Disabled":            `^false$`,
	}

	// Run the command and the tests
	checkCmdAnswerListKeysRegexp(test, imgDisplayVerboseLightLineNb, "", regexpMap2, imagesCmdName, "list", "-v", "--type", "standard")
}

// TestImagesGet is the function allowing to test if
// the command "rp-cli images get ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestImagesGet(test *testing.T) {
	// Firstly, creates a test image if it does not exists
	cmdToRun1 := getMainAddImgCmd("Imx6 Image aarch64", "imx6-img-aarch64", "aarch64", "redpesk-devel-arz")

	err := addObjectIfNotExists(test, imagesCmdName, "imx6-img-aarch64", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Verbose only, since the no verbose option is tested in almost all other tests

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":                 `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":                `^Imx6 Image aarch64$`,
		"Slug":                `^imx6-img-aarch64$`,
		"Type":                `^standard$`,
		"Distribution":        `^redpesk-devel-arz$`,
		"Architecture":        `^aarch64$`,
		"Board model":         `^generic$`,
		"Security model":      `^smack$`,
		"Download URL":        `^\[Not set\]$`,
		"Access token(s)":     `(_*[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}){2}`,
		"Kickstart URL":       `^git\+https*://`,
		"Kickstart name":      `^Redpesk-minimal-x86_64-smack.ks$`,
		"Install tree URL":    `^https://download.redpesk.bzh/redpesk-lts/arz-1.0/imager-os/aarch64/$`,
		"Additional repos":    `^\[Not set\]$`,
		"Parent project ID":   `^\[Not set\]$`,
		"Parent project slug": `^\[Not set\]$`,
		"Disabled":            `^false$`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, imgDisplayVerboseFullLineNb, regexpMap, imagesCmdName, "get", "imx6-img-aarch64", "-v")

	// Try to get from name
	checkCmdAnswerKeysRegexp(test, imgDisplayVerboseFullLineNb, regexpMap, imagesCmdName, "get", "Imx6 Image aarch64", "-v")
}

// TestImagesUpdateBasic is the function allowing to test if
// the command "rp-cli images update ..." is working correctly
// when few basic parameters are updated
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestImagesUpdateBasic(test *testing.T) {
	// Firstly, creates a test image if it does not exists
	cmdToRun1 := getMainAddImgCmd("Solidrun Image aarch64", "solidrun-img-aarch64", "aarch64", "redpesk-devel-arz")

	err := addObjectIfNotExists(test, imagesCmdName, "solidrun-img-aarch64", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Update the image with new arguments
	updateCmd := getMainUpdateImgCmd("solidrun-img-aarch64", "New Solidrun Image aarch64", "", "redpesk-brehat")

	// Run the update command and check that all is [OK]
	checkActionCmdAns(test, updateCmd...)

	// Check that the object has been updated

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":                 `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":                `^New Solidrun Image aarch64$`,
		"Slug":                `^solidrun-img-aarch64$`,
		"Type":                `^standard$`,
		"Distribution":        `^redpesk-brehat$`,
		"Architecture":        `^aarch64$`,
		"Board model":         `^generic$`,
		"Security model":      `^smack$`,
		"Download URL":        `^\[Not set\]$`,
		"Access token(s)":     `(_*[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}){2}`,
		"Kickstart URL":       `^git\+https*://`,
		"Kickstart name":      `^Redpesk-minimal-x86_64-smack.ks$`,
		"Install tree URL":    `^https://download.redpesk.bzh/redpesk-lts/arz-1.0/imager-os/aarch64/$`,
		"Additional repos":    `^\[Not set\]$`,
		"Parent project ID":   `^\[Not set\]$`,
		"Parent project slug": `^\[Not set\]$`,
		"Disabled":            `^false$`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, imgDisplayVerboseFullLineNb, regexpMap, imagesCmdName, "get", "solidrun-img-aarch64", "-v")
}

// TestImagesUpdateFull is the function allowing to test if
// the command "rp-cli images update ..." is working correctly
// when few full parameters are updated
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestImagesUpdateFull(test *testing.T) {
	// Firstly, creates a test image if it does not exists
	cmdToRun1 := getMainAddImgCmd("Upsquare Image x86_64", "upsquare-img-x86_64", "x86_64", "redpesk-devel-arz")

	err := addObjectIfNotExists(test, imagesCmdName, "upsquare-img-x86_64", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Update the image with new arguments
	updateCmd := getMainUpdateImgCmd("upsquare-img-x86_64", "", "", "redpesk-cezembre",
		"-k", "Redpesk-core-x86_64-smack.ks",
		"--ks-url", "http://git.ovh.iot/redpesk/redpesk-infra/imager-configuration.git",
		"--ks-branch", "02d883f4",
		"--ks-path", "kickstarts/RedPesk",
		"-r", "https://download.redpesk.bzh/redpesk-devel/releases/cezembre/Everything/x86_64/os/",
		"-r", "https://download.redpesk.bzh/redpesk-devel/releases/cezembre/AGL/x86_64/os/",
		"--reset-repo")

	// Run the update command and check that all is [OK]
	checkActionCmdAns(test, updateCmd...)

	// Check that the object has been updated

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":                 `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":                `^Upsquare Image x86_64$`,
		"Slug":                `^upsquare-img-x86_64$`,
		"Type":                `^standard$`,
		"Distribution":        `^redpesk-cezembre$`,
		"Architecture":        `^x86_64$`,
		"Board model":         `^generic$`,
		"Security model":      `^smack$`,
		"Download URL":        `^\[Not set\]$`,
		"Access token(s)":     `(_*[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}){2}`,
		"Kickstart URL":       `^git\+http://git\.ovh\.iot/redpesk/redpesk-infra/imager-configuration\.git\?kickstarts/RedPesk#02d883f4$`,
		"Kickstart name":      `^Redpesk-core-x86_64-smack.ks$`,
		"Install tree URL":    `^https://download.redpesk.bzh/redpesk-lts/arz-1.0/imager-os/x86_64/$`,
		"Additional repos":    `^http(s)*://`,
		"Parent project ID":   `^\[Not set\]$`,
		"Parent project slug": `^\[Not set\]$`,
		"Disabled":            `^false$`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, imgDisplayVerboseFullLineNb, regexpMap, imagesCmdName, "get", "upsquare-img-x86_64", "-v")
}

// TestImagesUpdateWithProject is the function allowing to test if
// the command "rp-cli images update ..." is working correctly, when a project
// is linked with it
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestImagesUpdateWithProject(test *testing.T) {
	// Firstly, creates a test project if it does not exists
	cmdToRun1 := getMainAddPrjCmd("Test Project for Pierre", "redpesk-devel-arz")
	err := addObjectIfNotExists(test, prjsCommand, "test-project-for-pierre", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the command to add the image
	cmdToRun2 := getMainAddImgCmd("Test Image for Pierre", "test-img-for-pierre", "aarch64", "redpesk-devel-arz",
		"-p", "test-project-for-pierre")

	// Create the image if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, imagesCmdName, "test-img-for-pierre", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}

	// The image has been created, check it is now in backend

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":                 `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":                `^Test Image for Pierre$`,
		"Slug":                `^test-img-for-pierre$`,
		"Type":                `^standard$`,
		"Distribution":        `^redpesk-devel-arz$`,
		"Architecture":        `^aarch64$`,
		"Board model":         `^generic$`,
		"Security model":      `^smack$`,
		"Download URL":        `^\[Not set\]$`,
		"Access token(s)":     `(_*[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}){2}`,
		"Kickstart URL":       `^git\+https*://`,
		"Kickstart name":      `^Redpesk-minimal-x86_64-smack.ks$`,
		"Install tree URL":    `^https://download.redpesk.bzh/redpesk-lts/arz-1.0/imager-os/aarch64/$`,
		"Additional repos":    `^(https?://.+, )(https?://.+)$`,
		"Parent project ID":   `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Parent project slug": `^test-project-for-pierre$`,
		"Disabled":            `^false$`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, imgDisplayVerboseFullLineNb, regexpMap, imagesCmdName, "get", "test-img-for-pierre", "-v")

	// Get the update command allowing to update the repos
	updateCmd := getMainUpdateImgCmd("test-img-for-pierre", "", "", "",
		"-r", "https://download.redpesk.bzh/redpesk-devel/releases/cezembre/AGL/aarch64/os/")

	// Run the update command and check that all is [OK]
	checkActionCmdAns(test, updateCmd...)

	// Change the regexp map to take into consideration our update
	regexpMap["Additional repos"] = `^(https?://.+, )(https?://.+)$`

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, imgDisplayVerboseFullLineNb, regexpMap, imagesCmdName, "get", "test-img-for-pierre", "-v")
}

// TestImagesDelete is the function allowing to test if
// the command "rp-cli images delete ..." is working correctly
// when the image is truly deleted (no tests linked)
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestImagesDelete(test *testing.T) {
	// Firstly, creates a test image if it does not exists
	cmdToRun1 := getMainAddImgCmd("Raspberry Pi 4 Image", "rpi-4-image", "aarch64", "redpesk-devel-arz")

	err := addObjectIfNotExists(test, imagesCmdName, "rpi-4-image", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Check that the image truly exists

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID":           `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Slug":         `^rpi-4-image$`,
		"Name":         `^Raspberry Pi 4 Image$`,
		"Architecture": `^aarch64$`,
		"Type":         `^standard$`,
		"Disabled":     `^false$`,
	}

	// Run the command and the tests (no verbose)
	checkCmdAnsKeysRegexpCol(test, imgDisplayNoVerboseColNb, regexpMap, imagesCmdName, "get", "rpi-4-image")

	// Run the delete command and check that all is [OK]
	checkActionCmdAns(test, imagesCmdName, "delete", "rpi-4-image")

	// Check that the image deleted does not exist anymore
	_, err = executeCommand(imagesCmdName, "get", "rpi-4-image")
	assert.NotNil(test, err)
}

// TestImagesLocalbuild is the function allowing to test if
// the command "rp-cli images localbuild ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestImagesLocalbuild(test *testing.T) {
	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Firstly, creates an image if it does not exists
	cmdToRun1 := getMainAddImgCmd("Upsquare Image x86_64", "upsquare-img-x86_64", "x86_64", "redpesk-devel-arz",
		"-k", "Redpesk-core-x86_64-smack.ks",
		"--ks-url", "http://git.ovh.iot/redpesk/redpesk-infra/imager-configuration.git",
		"--ks-branch", "02d883f4",
		"-r", "https://download.redpesk.bzh/redpesk-devel/releases/arz/Everything/x86_64/os/")

	err := addObjectIfNotExists(test, imagesCmdName, "upsquare-img-x86_64", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Run the "images localbuild" command and check that everything is good
	runAndCheckImgBuildCmd(test, "upsquare-img-x86_64", imagesCmdName, "localbuild", "upsquare-img-x86_64")

	// Check that the image size specified to the mock is what we think it is
	assert.Equal(test, -1, rpladmock.LatestImageSize)
}

// TestImagesLocalbuildWithLocalKs is the function allowing to test if
// the command "rp-cli images localbuild ..." is working correctly when
// a local kickstart is passed to the command
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestImagesLocalbuildWithLocalKs(test *testing.T) {
	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Firstly, creates an image if it does not exists
	cmdToRun1 := getMainAddImgCmd("Raspberry Pi 3B Image", "rpi-3b-image", "aarch64", "redpesk-devel-arz",
		"-k", "Redpesk-core-aarch64-smack.ks",
		"--ks-url", "http://git.ovh.iot/redpesk/redpesk-infra/imager-configuration.git",
		"--ks-branch", "02d883f4",
		"--ks-path", "kickstarts/RedPesk",
		"-r", "https://download.redpesk.bzh/redpesk-devel/releases/arz/Everything/aarch64/os/")

	err := addObjectIfNotExists(test, imagesCmdName, "rpi-3b-image", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	localKickstartFile := "/tmp/redpesk-cli-test/Redpesk-core-aarch64-smack.ks"

	// Create false kickstart
	err = utils.WriteToFile(localKickstartFile, "THIS IS A KICKSTART FILE")
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false kickstart file at the end of the function
	defer deleteFalseFile(localKickstartFile)

	// Run the "images localbuild" command and check that everything is good
	runAndCheckImgBuildCmd(test, "rpi-3b-image", imagesCmdName, "localbuild", "rpi-3b-image", "--local-kickstart", localKickstartFile)

	// Check that the kickstart file has been correctly copied in the gitpkgs repository
	fullImageKickstartPath := path.Join(gitPackagesTestDir, ("image_localbuild_" + "rpi-3b-image"), path.Base(localKickstartFile))
	assert.FileExists(test, fullImageKickstartPath)
}

// TestImagesLocalbuildWithSize is the function allowing to test if
// the command "rp-cli images localbuild ..." is working correctly
// when the user specify a size
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestImagesLocalbuildWithSize(test *testing.T) {
	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Firstly, creates an image if it does not exists
	cmdToRun1 := getMainAddImgCmd("Upsquare Image x86_64", "upsquare-img-x86_64", "x86_64", "redpesk-devel-arz",
		"-k", "Redpesk-core-x86_64-smack.ks",
		"--ks-url", "http://git.ovh.iot/redpesk/redpesk-infra/imager-configuration.git",
		"--ks-branch", "02d883f4",
		"--ks-path", "kickstarts/RedPesk",
		"-r", "https://download.redpesk.bzh/redpesk-devel/releases/arz/Everything/x86_64/os/")

	err := addObjectIfNotExists(test, imagesCmdName, "upsquare-img-x86_64", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Run the "images localbuild" command and check that everything is good
	runAndCheckImgBuildCmd(test, "upsquare-img-x86_64", imagesCmdName, "localbuild", "upsquare-img-x86_64", "-s", "5")

	// Check that the image size specified to the mock is what we think it is
	assert.Equal(test, 5, rpladmock.LatestImageSize)
}

// TestImagesBuild is the function allowing to test if
// the command "rp-cli images build" is working correctly
// when the user specify a size
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestImagesBuild(test *testing.T) {
	// Firstly creates a project and an image inside it
	err := createProjectAndImage(test, "pone", "redpesk-far")
	if err != nil {
		test.Fatal(err)
	}

	// The image have been created, now try to build it
	outputCmd, err := executeCommand(imagesCmdName, "build", "redpesk-far")
	if err != nil {
		test.Fatal(err)
	}

	// Check the output of the build
	checkNotVerboseBuildOutput(test, outputCmd)
}

// TestImagesBuildsList is the function allowing to test if
// the command "rp-cli images builds list" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestImagesBuildsList(test *testing.T) {
	// Firstly creates a project and an image inside it
	err := createProjectAndImage(test, "wordle", "redpesk-mourn")
	if err != nil {
		test.Fatal(err)
	}

	// The image have been created, now try to build it several time
	for i := 0; i < 3; i++ {
		_, err = executeCommand(imagesCmdName, "build", "redpesk-mourn")
		if err != nil {
			test.Fatal(err)
		}
	}

	// Not verbose mode
	checkNotVerboseListBuildsOutput(test, imagesCmdName, "builds", "list", "redpesk-mourn")

	// Verbose mode
	checkVerboseListBuildsOutput(test, rpapiv1.RPBuildImage, imagesCmdName, "builds", "list", "redpesk-mourn", "-v")
}

// TestImagesBuildsGet is the function allowing to test if
// the command "rp-cli images builds get" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestImagesBuildsGet(test *testing.T) {
	// Firstly creates a project and an image inside it
	err := createProjectAndImage(test, "stromae", "redpesk-enfer")
	if err != nil {
		test.Fatal(err)
	}

	// The image have been created, now try to build it several time
	for i := 0; i < 3; i++ {
		_, err = executeCommand(imagesCmdName, "build", "redpesk-enfer", "-c", fmt.Sprintf("Comment #%d", i))
		if err != nil {
			test.Fatal(err)
		}
	}

	// Extract the latest build ID from the build list
	highestID, err := getHighestIDFromOutput("Build ID", imagesCmdName, "builds", "list", "redpesk-enfer")
	if err != nil {
		test.Fatal(err)
	}

	// Not verbose mode

	// Create the regexp map used to check (not verbose mode)
	regexpMap := map[string]string{
		"Build ID:":                        `[0-9]{4}`,
		"Status:":                          `^(done|failed)$`,
		"Subtask livemedia x86_64 status:": `^(closed|failed)$`,
	}

	// Run the "builds get" command an check the output
	checkCmdAnswerKeysRegexp(test, 4, regexpMap, imagesCmdName, "builds", "get", highestID)

	// Verbose mode

	// Create the regexp map used to check (verbose mode)
	regexpMap2 := map[string]string{
		"Build ID:":                   `[0-9]{4}`,
		"Build type:":                 `image build`,
		"Image ID:":                   `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Image slug:":                 `^redpesk-enfer$`,
		"Distribution:":               `^redpesk-lts-arz$`,
		"Built by":                    `.*`,
		"Status:":                     `^(done|failed)$`,
		"Subtask:":                    `^livemedia x86_64$`,
		"Subtask state:":              `^(closed|failed)$`,
		"Subtask method:":             `^createLiveMedia$`,
		"Subtask method description:": `.+`,
		"Comment:":                    `^Comment #[0-9]+`,
	}

	// Run the "builds get" command an check the output
	output := checkCmdAnswerKeysRegexp(test, 23, regexpMap2, imagesCmdName, "builds", "get", highestID, "-v")

	// Assert that the "tag" for the log files and image files are existing
	assert.Contains(test, output, "[Log files]")
	assert.Contains(test, output, "[Files]")
}

// TestImagesBuildsCancel is the function allowing to test if
// the command "rp-cli images builds cancel" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestImagesBuildsCancel(test *testing.T) {
	// Firstly creates a project and an image inside it
	err := createProjectAndImage(test, "claude", "tres-disco")
	if err != nil {
		test.Fatal(err)
	}

	// Say to the mocked backend that we will cancel the build
	rpbckmock.RequestNextBuildCanceled()

	// The image have been created, now try to build it
	_, err = executeCommand(imagesCmdName, "build", "tres-disco", "--nonblocking")
	if err != nil {
		test.Fatal(err)
	}

	// Right after request for the build to be canceled
	checkActionCmdAns(test, imagesCmdName, genericBuildsCmd, "cancel", "tres-disco")

	// Wait a bit for the cancel request got processed
	time.Sleep(40 * time.Millisecond)

	// Extract the latest build ID from the build list
	highestID, err := getHighestIDFromOutput("Build ID", imagesCmdName, "builds", "list", "tres-disco")
	if err != nil {
		test.Fatal(err)
	}

	// Not verbose mode

	// Create the regexp map used to check (not verbose mode)
	regexpMap := map[string]string{
		"Build ID:":                        `[0-9]{4}`,
		"Status:":                          `^canceled$`,
		"Subtask livemedia x86_64 status:": `^(closed|failed|canceled)$`,
	}

	// Run the "builds get" command to check that the build has been correctly canceled
	checkCmdAnswerKeysRegexp(test, 4, regexpMap, imagesCmdName, "builds", "get", highestID)
}

// TestImagesBuildsLogs is the function allowing to test if
// the command "rp-cli images builds logs" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestImagesBuildsLogs(test *testing.T) {
	// Firstly creates a project and an image inside it
	err := createProjectAndImage(test, "danger", "redpesk-4:30")
	if err != nil {
		test.Fatal(err)
	}

	// The image have been created, now try to build it
	_, err = executeCommand(imagesCmdName, "build", "redpesk-4:30")
	if err != nil {
		test.Fatal(err)
	}

	// Now try to download build logs from the server
	cmdOutput, err := executeCommand(imagesCmdName, "builds", "logs", "redpesk-4:30", "-d", "/tmp")
	if err != nil {
		test.Fatal(err)
	}

	// Check that rp-cli got the latest build
	latestBuildRetrieved := assert.Contains(test, cmdOutput, "The build logs for build \"")

	ourBuildID := ""
	if latestBuildRetrieved {
		// Get the buildID used
		lineWanted := utils.ExtractLineWithPrefix(cmdOutput, "The build logs for build \"")
		splitedLine := strings.Split(lineWanted, "\"")
		ourBuildID = strings.TrimSpace(splitedLine[1])
	} else {
		test.Fatal()
	}

	// Check that the output contains the needed element
	assert.Contains(test, cmdOutput, "Downloading logs for subtask \"livemedia x86_64\"...")

	// Check that the files are available
	assert.FileExists(test, path.Join("/tmp", "rp-image-builds-logs", ourBuildID, "livemedia x86_64", "root.log"))
	assert.FileExists(test, path.Join("/tmp", "rp-image-builds-logs", ourBuildID, "livemedia x86_64", "build.log"))
	assert.FileExists(test, path.Join("/tmp", "rp-image-builds-logs", ourBuildID, "livemedia x86_64", "anaconda-dbus.log"))
	assert.FileExists(test, path.Join("/tmp", "rp-image-builds-logs", ourBuildID, "livemedia x86_64", "anaconda-hawkey.log"))
	assert.FileExists(test, path.Join("/tmp", "rp-image-builds-logs", ourBuildID, "livemedia x86_64", "anaconda-anaconda.log"))
	assert.FileExists(test, path.Join("/tmp", "rp-image-builds-logs", ourBuildID, "livemedia x86_64", "Redpesk-minimal-x86_64-smack.ks"))
	assert.FileExists(test, path.Join("/tmp", "rp-image-builds-logs", ourBuildID, "livemedia x86_64", "koji-image-redpesk-my-project--redpesk-lts-arz-1.0-update-build-666.ks"))

	// Now try to download build-logs from the server but ask for zip logs
	_, err = executeCommand(imagesCmdName, "builds", "logs", "redpesk-4:30", "-d", "/tmp", "--zip")
	if err != nil {
		test.Fatal(err)
	}

	// Add to list the zip file to be deleted at teardown time
	zipName := path.Join("/tmp", "rp-image-builds-logs-"+ourBuildID+".zip")
	filesListToDel = append(filesListToDel, zipName)

	// Check that a zip file has been correctly created
	assert.FileExists(test, zipName)
}

// TestImagesTestNotVerbose is the function allowing to test if
// the command "rp-cli images test ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestImagesTestNotVerbose(test *testing.T) {

	// Firstly, check if the "tgv" application exists and if it has builds
	buildID, err := checkIfBuildsExist("redpesk-far", imagesCmdName)
	if err != nil {
		test.Fatal(err)
	}

	// Check if a buildID have been found
	if buildID == "" {
		// No build ID have been found, run a build test to have some builds to test
		TestImagesBuild(test)
	}

	// Run a test
	cmdOutput, err := executeCommand(imagesCmdName, "test", "redpesk-far")
	if err != nil {
		test.Fatal(err)
	}

	// Check the output
	checkTestOutput(test, cmdOutput, "image", false)
}

// TestImagesTestVerbose is the function allowing to test if
// the command "rp-cli images test -v ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestImagesTestVerbose(test *testing.T) {

	buildID := ""
	var err error
	for buildID == "" {
		// Firstly, check if the "redpesk-far" image exists and if it has builds
		buildID, err = checkIfBuildsExist("redpesk-far", imagesCmdName)
		if err != nil {
			test.Fatal(err)
		}
		TestImagesBuild(test)
	}

	// Run a test
	cmdOutput, err := executeCommand(imagesCmdName, "test", "redpesk-far", "-v")
	if err != nil {
		test.Fatal(err)
	}

	// Check the output
	checkTestOutput(test, cmdOutput, "image", true)

	// Re-run a test, with the build
	buildID, err = checkIfBuildsExist("redpesk-far", imagesCmdName)
	if err != nil {
		test.Fatal(err)
	}
	cmdOutput, err = executeCommand(imagesCmdName, "test", buildID)
	if err != nil {
		test.Fatal(err)
	}

	// Check the output
	checkTestOutput(test, cmdOutput, "image", false)
}

// TestImagesTestsList is the function allowing to test if
// the command "rp-cli images tests list <imgID> ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestImagesTestsList(test *testing.T) {

	// Firstly creates a project and an image inside it
	err := createProjectAndImage(test, "stromae", "redpesk-enfer")
	if err != nil {
		test.Fatal(err)
	}

	// Firstly, check if the "avion-a320" application exists and if it has builds
	buildID, err := checkIfBuildsExist("redpesk-enfer", imagesCmdName)
	if err != nil {
		test.Fatal(err)
	}

	// Check if at least one build exists
	if buildID == "" {
		// No build exists, creates one
		_, err = executeCommand(imagesCmdName, "build", "redpesk-enfer")
		if err != nil {
			test.Fatal(err)
		}
	}

	// Make sure to have at least three tests for this image
	_, err = executeCommand(imagesCmdName, "test", "redpesk-enfer")
	if err != nil {
		test.Fatal(err)
	}
	_, err = executeCommand(imagesCmdName, "test", "redpesk-enfer")
	if err != nil {
		test.Fatal(err)
	}
	_, err = executeCommand(imagesCmdName, "test", "redpesk-enfer")
	if err != nil {
		test.Fatal(err)
	}

	// Run the "tests list" command an check the output
	checkNotVerboseListTestsOutput(test, imagesCmdName, "tests", "list", "redpesk-enfer")

	// Run the "tests list -v" command an check the output
	checkVerboseListTestsOutput(test, true, imagesCmdName, "tests", "list", "redpesk-enfer", "-v")
}

// TestImagesTestsLogs is the function allowing to test if
// the command "rp-cli images tests logs <imgID> ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestImagesTestsLogs(test *testing.T) {
	// Firstly creates a project and an image inside it
	err := createProjectAndImage(test, "wordle", "redpesk-mourn")
	if err != nil {
		test.Fatal(err)
	}

	// Firstly, check if the "redpesk-mourn" image exists and if it has builds
	buildID, err := checkIfBuildsExist("redpesk-mourn", imagesCmdName)
	if err != nil {
		test.Fatal(err)
	}

	// Check if at least one build exists
	if buildID == "" {
		// No build exists, creates one
		_, err = executeCommand(imagesCmdName, "build", "redpesk-mourn")
		if err != nil {
			test.Fatal(err)
		}
	}

	// Make sure to have test for this image
	_, err = executeCommand(imagesCmdName, "test", "redpesk-mourn")
	if err != nil {
		test.Fatal(err)
	}

	// Now try to download test-logs from the server
	cmdOutput, err := executeCommand(imagesCmdName, "tests", "logs", "redpesk-mourn", "-d", "/tmp")
	if err != nil {
		test.Fatal(err)
	}

	// Check that rp-cli got the latest test
	latestTestRetrieved := assert.Contains(test, cmdOutput, "Downloading logs for test")

	ourTestID := ""
	if latestTestRetrieved {
		// Get the testID used
		lineWanted := utils.ExtractLineWithPrefix(cmdOutput, "Downloading logs for test")
		splitedLine := strings.Split(lineWanted, "\"")
		ourTestID = strings.TrimSpace(splitedLine[1])
	} else {
		test.Fatal()
	}

	// Check that the output contains the needed element
	assert.Contains(test, cmdOutput, "The test logs for test \""+ourTestID+"\" are available in:")

	// Check that both files (.tap and .zip) are available
	assert.FileExists(test, path.Join("/tmp", "rp-tests-logs", ourTestID, "tests_logs.zip"))
	assert.FileExists(test, path.Join("/tmp", "rp-tests-logs", ourTestID, "all_tests.tap"))
	assert.FileExists(test, path.Join("/tmp", "rp-tests-logs", ourTestID, "boot.log"))
	assert.FileExists(test, path.Join("/tmp", "rp-tests-logs", ourTestID, "console.log"))

	// Now try to download tests logs from the server but ask for zip logs
	_, err = executeCommand(imagesCmdName, "tests", "logs", "redpesk-mourn", "-d", "/tmp", "--zip")
	if err != nil {
		test.Fatal(err)
	}

	// Add to list the zip file to be deleted at teardown time
	zipName := path.Join("/tmp", "rp-tests-logs-"+ourTestID+".zip")
	filesListToDel = append(filesListToDel, zipName)

	// Check that a zip file has been correctly created
	assert.FileExists(test, zipName)
}

// TestImagesTestsGet is the function allowing to test if
// the command "rp-cli images tests get <buildID> ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestImagesTestsGet(test *testing.T) {

	// Firstly creates a project and an image inside it
	err := createProjectAndImage(test, "stromae", "redpesk-enfer")
	if err != nil {
		test.Fatal(err)
	}

	// Firstly, check if the "redpesk-enfer" image exists and if it has builds
	buildID, err := checkIfBuildsExist("redpesk-enfer", imagesCmdName)
	if err != nil {
		test.Fatal(err)
	}

	// Check if at least one build exists
	if buildID == "" {
		// No build exists, creates one
		_, err = executeCommand(imagesCmdName, "build", "redpesk-enfer")
		if err != nil {
			test.Fatal(err)
		}
	}

	// Make sure to have at least three tests for this image
	_, err = executeCommand(imagesCmdName, "test", "redpesk-enfer", "-c", "Test comment n°1")
	if err != nil {
		test.Fatal(err)
	}
	_, err = executeCommand(imagesCmdName, "test", "redpesk-enfer", "-c", "Test comment n°2")
	if err != nil {
		test.Fatal(err)
	}
	_, err = executeCommand(imagesCmdName, "test", "redpesk-enfer", "-c", "Test comment n°3")
	if err != nil {
		test.Fatal(err)
	}

	// Extract the latest test ID from the build list
	highestID, err := getHighestIDFromOutput("Test ID", imagesCmdName, "tests", "list", "redpesk-enfer")
	if err != nil {
		test.Fatal(err)
	}

	// Run the "tests get" command an check the output
	checkNotVerboseListTestsOutput(test, imagesCmdName, "tests", "get", highestID)

	// Create the regexp map used to check (verbose mode)
	regexpMap2 := map[string]string{
		"Test ID":                `^[0-9]+$`,
		"Start time":             `.+`,
		"Application ID":         `^\[Default image test app\]$`,
		"Build ID":               `^[0-9]{4}$`,
		"Image ID":               `^(\[Not set\]|[a-z0-9-]+)$`,
		"Board model":            `^qemu-system-x86_64$`,
		"Status":                 `^(success|failed)$`,
		"Succeeded tests number": `[0-9]+`,
		"Failed tests number":    `[0-9]+`,
		"Skipped tests number":   `[0-9]+`,
		"Total tests number":     `[0-9]+`,
		"Comment:":               `^Test comment n°[0-9]{1}`,
	}

	// Run the "tests get" command an check the output
	checkCmdAnswerKeysRegexp(test, appTestsDisplayVerboseLineNb, regexpMap2, imagesCmdName, "tests", "get", highestID, "-v")
}

// TestImagesTestsCancelOK is the function allowing to test if
// the command "rp-cli images tests cancel <imgID> ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestImagesTestsCancelOK(test *testing.T) {
	// Firstly creates a project and an image inside it
	err := createProjectAndImage(test, "wordle", "redpesk-mourn")
	if err != nil {
		test.Fatal(err)
	}

	// Firstly, check if the "redpesk-mourn" image exists and if it has builds
	buildID, err := checkIfBuildsExist("redpesk-mourn", imagesCmdName)
	if err != nil {
		test.Fatal(err)
	}

	// Check if at least one build exists
	if buildID == "" {
		// No build exists, creates one
		_, err = executeCommand(imagesCmdName, "build", "redpesk-mourn")
		if err != nil {
			test.Fatal(err)
		}
	}

	// Start a test in non-blocking mode
	_, err = executeCommand(imagesCmdName, "test", "redpesk-mourn", "--nonblocking")
	if err != nil {
		test.Fatal(err)
	}

	// Now try to cancel the test
	cmdOutput, err := executeCommand(imagesCmdName, "tests", "cancel", "redpesk-mourn")
	if err != nil {
		test.Fatal(err)
	}

	// Test that everything went OK
	checkLineEndWith(test, cmdOutput, "Cancelling test with ID", "[OK]")

	// Extract the latest test ID from the build list
	highestID, err := getHighestIDFromOutput("Test ID", imagesCmdName, "tests", "list", "redpesk-mourn")
	if err != nil {
		test.Fatal(err)
	}

	// Create the regexp map used to check (not verbose mode)
	regexpMap := map[string]string{
		"ID":     `^[0-9]+$`,
		"Status": `^aborted$`,
	}

	// Run the "tests get" command an check the output
	time.Sleep(50 * time.Millisecond)
	checkCmdAnsKeysRegexpCol(test, 2, regexpMap, imagesCmdName, "tests", "get", highestID)
}

// TestImagesTestsCancelKO is the function allowing to test if
// the command "rp-cli images test-cancel <imgID> ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestImagesTestsCancelKO(test *testing.T) {
	// Firstly creates a project and an image inside it
	err := createProjectAndImage(test, "wordle", "redpesk-mourn")
	if err != nil {
		test.Fatal(err)
	}

	// Firstly, check if the "redpesk-mourn" image exists and if it has builds
	buildID, err := checkIfBuildsExist("redpesk-mourn", imagesCmdName)
	if err != nil {
		test.Fatal(err)
	}

	// Check if at least one build exists
	if buildID == "" {
		// No build exists, creates one
		_, err = executeCommand(imagesCmdName, "build", "redpesk-mourn")
		if err != nil {
			test.Fatal(err)
		}
	}

	// Start a test in blocking mode to make sure to have finished it
	_, err = executeCommand(imagesCmdName, "test", "redpesk-mourn")
	if err != nil {
		test.Fatal(err)
	}

	// Now try to cancel the test (that is already finished)
	_, err = executeCommand(imagesCmdName, "tests", "cancel", "redpesk-mourn")
	if assert.Error(test, err) {
		assert.Contains(test, err.Error(), "cannot cancel tests when tests' status is success")
	}
}

// TestImagesTestsTimeOut is the function allowing to test if
// rp-cli is behaving well when a test is in time-out
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestImagesTestsTimeOut(test *testing.T) {
	// Firstly creates a project and an image inside it
	err := createProjectAndImage(test, "wordle", "redpesk-mourn")
	if err != nil {
		test.Fatal(err)
	}

	// Firstly, check if the "redpesk-mourn" image exists and if it has builds
	buildID, err := checkIfBuildsExist("redpesk-mourn", imagesCmdName)
	if err != nil {
		test.Fatal(err)
	}

	// Check if at least one build exists
	if buildID == "" {
		// No build exists, creates one
		_, err = executeCommand(imagesCmdName, "build", "redpesk-mourn")
		if err != nil {
			test.Fatal(err)
		}
	}

	// Request the next test to be "time-out"
	rpbckmock.RequestTimedOutTest()

	// Start a test in non-blocking mode
	_, err = executeCommand(imagesCmdName, "test", "redpesk-mourn", "-v")

	// Test that we have an error in return
	assert.EqualError(test, err, appTestsTimedOutMsg)

	// Extract the latest test ID from the build list
	highestID, err := getHighestIDFromOutput("ID", imagesCmdName, "tests", "list", "redpesk-mourn")
	if err != nil {
		test.Fatal(err)
	}

	// Create the regexp map used to check (not verbose mode)
	regexpMap := map[string]string{
		"ID":     `^[0-9]+$`,
		"Status": `^timed-out$`,
	}

	// Run the "tests get" command an check the output
	time.Sleep(50 * time.Millisecond)
	checkCmdAnsKeysRegexpCol(test, 2, regexpMap, imagesCmdName, "tests", "get", highestID)
}

// TestImagesTestsStats is the function allowing to test if
// the command "rp-cli images tests stats <imgID>" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestImagesTestsStats(test *testing.T) {
	// Firstly creates a project and an image inside it
	err := createProjectAndImage(test, "tests-stats-project", "tests-stats-image")
	if err != nil {
		test.Fatal(err)
	}

	// Firstly, check if the image exists and if it has builds
	buildID, err := checkIfBuildsExist("tests-stats-image", imagesCmdName)
	if err != nil {
		test.Fatal(err)
	}

	// Check if at least one build exists
	if buildID == "" {
		// No build exists, creates one
		_, err = executeCommand(imagesCmdName, "build", "tests-stats-image")
		if err != nil {
			test.Fatal(err)
		}
	}

	// Make sure to have at least three tests for this image
	_, err = executeCommand(imagesCmdName, "test", "tests-stats-image")
	if err != nil {
		test.Fatal(err)
	}
	_, err = executeCommand(imagesCmdName, "test", "tests-stats-image")
	if err != nil {
		test.Fatal(err)
	}
	_, err = executeCommand(imagesCmdName, "test", "tests-stats-image")
	if err != nil {
		test.Fatal(err)
	}

	// Check the content of the "not verbose" output
	regexpMap := map[string]string{
		"Image ID:":   `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Image slug:": `^tests-stats-image$`,
		"Succeeded:":  `^[0-3]+$`,
		"Failed:":     `^[0-3]+$`,
		"Cancelled:":  `^[0-3]+$`,
		"Total:":      `^3$`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, imgTestsStatsNoVerboseLineNb, regexpMap, imagesCmdName, genericTestsCmd, "stats", "tests-stats-image")

	// Now, let's go for the verbose mode
	verboseAns := checkCmdAnswerKeysRegexp(test, imgTestsStatsNoVerboseLineNb+4+3, regexpMap, imagesCmdName, genericTestsCmd, "stats", "tests-stats-image", "-v")

	// Create the regexp map we will use
	regexpMapCol := map[string]string{
		"ID":     `^[0-9]+$`,
		"Type":   `^CI_test$`,
		"Status": `^(success|failed)$`,
	}

	// Get the second part of the answer
	columnsAns := strings.Split(verboseAns, genericTestsStatsCmdAnsLine7)[1]
	// Trim it
	columnsAns = strings.TrimSpace(columnsAns)

	// Check the answer of the command
	checkObjectKeysRegexpCol(test, columnsAns, imgTestsStatsVerboseColNb, regexpMapCol)
}

// TestImagesRelease is the function allowing to test if
// the command "rp-cli image release <imgID>..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestImagesRelease(test *testing.T) {
	// Add a user if it does not exists
	cmdUserToRun := getMainAddUserCmd("the_banano", "banano@iot.bzh", "Armand", "Bénéteau", "0666989945")
	// Create the user if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, usersCommand, "the_banano", cmdUserToRun...)
	if err != nil {
		test.Fatal(err)
	}
	// Create a team if it does not exists
	cmdToRun := getMainAddTeamCmd("Bruit", "La meilleure des teams", true)
	// Create the team if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, teamsCommand, "bruit", cmdToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// Do the creation/release
	createAndReleaseImage(test, "Stupeflip Crou", "Crou anthem image", []string{"the_banano"}, []string{"bruit"}, false)
}

// TestImagesReleasesGet is the function allowing to test if
// the command "rp-cli images releases get <imgID>..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestImagesReleasesGet(test *testing.T) {
	// Do a first creation/release
	createAndReleaseImage(test, "Stupeflip Crou", "Crou anthem image 0", []string{}, []string{}, false)

	// Check the content of the "not verbose" output
	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID":        `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Slug":      `^crou-anthem-image-0-1.0$`,
		"Version":   `^1.0$`,
		"Reference": `^crou-anthem-image-0$`,
	}

	// Run the command and the tests (no verbose)
	checkCmdAnsKeysRegexpCol(test, imgRelDisplayNoVerboseColNb, regexpMap, imagesCmdName, "releases", "get", "crou-anthem-image-0-1.0")

	// Check the content of the "verbose" output
	// Create the map containing the regexp we want to match in the answer
	regexpMap2 := map[string]string{
		"ID:":                  `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":                 `^Crou anthem image 0$`,
		"Slug":                 `^crou-anthem-image-0-1.0$`,
		"Reference image ID":   `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Reference image slug": `^crou-anthem-image-0$`,
		"Distribution ID":      `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Version":              `^1.0$`,
		"Creation time":        `.+`,
		"Update time":          `.+`,
		"URL":                  `http(s)*://`,
		"Access token(s)":      `(_*[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}){2}`,
		"Read only":            `^false$`,
		"[Detailed info]":      `.*`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, imgRelDisplayVerboseFullLineNb, regexpMap2, imagesCmdName, "releases", "get", "-v", "crou-anthem-image-0-1.0")

	// Try to get with the name
	checkCmdAnswerKeysRegexp(test, imgRelDisplayVerboseFullLineNb, regexpMap2, imagesCmdName, "releases", "get", "-v", "Crou anthem image 0")
}

// TestImagesReleasesList is the function allowing to test if
// the command "rp-cli images releases list <imgID>..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestImagesReleasesList(test *testing.T) {
	// Do a first creation/release
	createAndReleaseImage(test, "Stupeflip Crou 0", "Crou anthem image 0", []string{}, []string{}, false)

	// Then release several times the image
	checkActionCmdAns(test, imagesCmdName, "release", "crou-anthem-image-0", "--distribution", "redpesk-lts-arz", "-v", "1.1")
	checkActionCmdAns(test, imagesCmdName, "release", "crou-anthem-image-0", "--distribution", "redpesk-lts-arz", "-v", "1.2")
	checkActionCmdAns(test, imagesCmdName, "release", "crou-anthem-image-0", "--distribution", "redpesk-lts-arz", "-v", "1.3")

	// Check the content of the output (no verbose)
	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID":        `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Slug":      `^crou-anthem-image-*[01]*-1.[0123]$`,
		"Version":   `^1.[0123]$`,
		"Reference": `^crou-anthem-image-*0*$`,
	}

	// Run the command and the tests (no verbose)
	checkCmdAnsKeysRegexpCol(test, imgRelDisplayNoVerboseColNb, regexpMap, imagesCmdName, "releases", "list")

	// Get the total number of released images
	relImgNb := getObjNumber(test, imagesCmdName, "releases", "list")

	// Check the content of the output (verbose)
	regexpMap2 := map[string]string{
		"ID:":                  `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":                 `^Crou anthem image *[01]*$`,
		"Slug":                 `^crou-anthem-image-*[01]*-1.[0123]$`,
		"Reference image ID":   `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Reference image slug": `^crou-anthem-image-*[01]*$`,
		"Distribution ID":      `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Version":              `^1.[0123]$`,
		"Creation time":        `.+`,
		"Update time":          `.+`,
		"URL":                  `http(s)*://`,
		"Read only":            `^false$`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, (imgRelDisplayVerboseLightLineNb*relImgNb)-2, regexpMap2, imagesCmdName, "releases", "list", "-v")
}

// TestImagesReleasesDelete is the function allowing to test if
// the command "rp-cli images releases delete <imgID>..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestImagesReleasesDelete(test *testing.T) {
	// Do a first creation/release
	createAndReleaseImage(test, "Stupeflip Crou 0", "Crou anthem image 2", []string{}, []string{}, false)

	// Delete the released image and check that it went [OK]
	checkActionCmdAns(test, imagesCmdName, "releases", "delete", "crou-anthem-image-2-1.0")
}

// TestImagesReleasesUpdate is the function allowing to test if
// the command "rp-cli image releases update <imgID>..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestImagesReleasesUpdate(test *testing.T) {
	// // Do a first creation/release
	// createAndReleaseImage(test, "Stupeflip Crou", "Crou anthem image 3", []string{}, []string{}, false)

	// // Update the released image
	// checkActionCmdAns(test, imagesCmdName, "releases", "update", "crou-anthem-image-3-1.0", "-n", "A new name")

	// // Check the content of the "verbose" output
	// // Create the map containing the regexp we want to match in the answer
	// regexpMap2 := map[string]string{
	// 	"ID:":                  `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
	// 	"Name":                 `^A new name$`,
	// 	"Slug":                 `^crou-anthem-image-3-1.0$`,
	// 	"Reference image ID":   `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
	// 	"Reference image slug": `^crou-anthem-image-3$`,
	// 	"Distribution ID":      `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
	// 	"Version":              `^1.0$`,
	// 	"Creation time":        `.+`,
	// 	"Update time":          `.+`,
	// 	"URL":                  `http(s)*://`,
	// 	"Access token(s)":      `(_*[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}){2}`,
	// 	"Read only":            `^false$`,
	// 	"[Detailed info]":      `.*`,
	// }

	// // Run the command and the tests
	// checkCmdAnswerKeysRegexp(test, imgRelDisplayVerboseFullLineNb, regexpMap2, imagesCmdName, "releases", "get", "-v", "crou-anthem-image-3-1.0")
	test.Log("Not implemented for the moment")
}

// TestImagesReleasesUpdateAudience is the function allowing to test if
// the command "rp-cli images releases update <imgID>..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestImagesReleasesUpdateAudience(test *testing.T) {
	// Add a user if it does not exists
	cmdUserToRun := getMainAddUserCmd("the_banano", "banano@iot.bzh", "Armand", "Bénéteau", "0666989945")
	// Create the user if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, usersCommand, "the_banano", cmdUserToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// Create a team if it does not exists
	cmdToRun := getMainAddTeamCmd("Private Bruit", "La meilleure des teams", false)
	// Create the team if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, teamsCommand, "private-bruit", cmdToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// Do a first creation/release
	createAndReleaseImage(test, "Stupeflip Crou", "Crou anthem image 4", []string{"the_banano"}, []string{"private-bruit"}, false)

	// Try to update the audience of the release and check that it goes right!
	checkActionCmdAns(test, imagesCmdName, "releases", "update", "crou-anthem-image-4-1.0", "--reset-teams", "--reset-users", "-a", "true")
}

// TestImagesReleasesGetAudience is the function allowing to test if
// the command "rp-cli projects releases get <imgID> -av" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestImagesReleasesGetAudience(test *testing.T) {
	// Add a user if it does not exists
	cmdUserToRun := getMainAddUserCmd("the_banano", "banano@iot.bzh", "Armand", "Bénéteau", "0666989945")
	// Create the user if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, usersCommand, "the_banano", cmdUserToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// Create a team if it does not exists
	cmdToRun := getMainAddTeamCmd("Private Bruit", "La meilleure des teams", false)
	// Create the team if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, teamsCommand, "private-bruit", cmdToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// Do a first creation/release
	createAndReleaseImage(test, "Stupeflip Crou", "Crou anthem image 5", []string{"the_banano"}, []string{"private-bruit"}, false)

	// Check the content of the "verbose" output
	// Create the map containing the regexp we want to match in the answer
	regexpMap2 := map[string]string{
		"ID:":                  `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Name":                 `^Crou anthem image 5$`,
		"Slug":                 `^crou-anthem-image-5-1.0$`,
		"Reference image ID":   `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Reference image slug": `^crou-anthem-image-5$`,
		"Distribution ID":      `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Version":              `^1.0$`,
		"Creation time":        `.+`,
		"Update time":          `.+`,
		"URL":                  `http(s)*://`,
		"Access token(s)":      `(_*[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}){2}`,
		"Read only":            `^false$`,
		"[Detailed info]":      `.*`,
		"[Audience]":           `.*`,
		"All Users":            `^no$`,
		"Team(s)":              `^private-bruit$`,
		"User(s)":              `^the_banano$`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, imgRelDisplayVerboseFullAudienceLineNb, regexpMap2, imagesCmdName, "releases", "get", "-av", "crou-anthem-image-5-1.0")
}

// --------------------------------------------------------------
// Utilities functions ------------------------------------------
// --------------------------------------------------------------

// createAndReleaseImage is a function allowing to create and release an image easily
// The calling function needs to specify the name of the parent project and of the
// reference image
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param prjName: Name of the parent project
// :type prjName: string
// :param imgName: Name of the reference project
// :type imgName: string
// :param usersTab: Tab of users wanted as audience for the release
// :type usersTab: []string
// :param teamsTab: Tab of teams wanted as audience for the release
// :type teamsTab: []string
// :param factory: True if the release is for the factory
// :type factory: bool
// :returns: error -- an error if a problem has occurred
func createAndReleaseImage(test *testing.T, prjName, imgName string, usersTab, teamsTab []string, factory bool) error {
	// First, create the project and the image
	err := createProjectAndImage(test, prjName, imgName)
	if err != nil {
		return err
	}
	imgSlug := strings.ReplaceAll(strings.ToLower(imgName), " ", "-")

	// Create the basic command line
	cmdLine := []string{imagesCmdName, "release", imgSlug, "--distribution", "redpesk-lts-arz", "-v", "1.0"}

	// Complete with potential audience
	for _, user := range usersTab {
		cmdLine = append(cmdLine, "-u", user)
	}
	for _, team := range teamsTab {
		cmdLine = append(cmdLine, "--team", team)
	}
	if factory {
		cmdLine = append(cmdLine, "-s")
	}

	// Try to release the image
	checkActionCmdAns(test, cmdLine...)

	return nil
}

// getMainAddImgCmd is an initializer function to get the mainly
// used command to create an image
//
// :param name: Name of the image to create
// :type name: string
// :param alias: Alias of the image to create
// :type alias: string
// :param arch: Architecture of the image to create
// :type arch: string
// :param distro: Distribution of the image to create
// :type distro: string
// :param distVersion: Distribution version of the image to create
// :type distVersion: string
// :param secu: Security model of the image to create
// :type secu: string
// :param args: Other args to pass as command to create the project
// :type args: ...string
// :returns: []string -- the value of the wanted constant
func getMainAddImgCmd(name, alias, arch, distro string, args ...string) []string {
	cmdToReturn := []string{
		imagesCmdName,
		imagesAddCmd,
		"-n",
		name,
		"-a",
		alias,
		"--arch",
		arch,
		"-d",
		distro,
	}

	// Add other args to the command
	cmdToReturn = append(cmdToReturn, args...)

	return cmdToReturn
}

// getMainUpdateImgCmd is an initializer function to get the mainly
// used command to update an image
//
// :param slug: Slug of the image to update
// :type slug: string
// :param name: Name of the image to update
// :type name: string
// :param arch: Architecture of the image to update
// :type arch: string
// :param distro: Distribution of the image to update
// :type distro: string
// :param distVersion: Distribution version of the image to update
// :type distVersion: string
// :param secu: Security model of the image to update
// :type secu: string
// :param args: Other args to pass as command to update the project
// :type args: ...string
// :returns: []string -- the value of the wanted constant
func getMainUpdateImgCmd(slug, name, arch, distro string, args ...string) []string {
	cmdToReturn := []string{
		imagesCmdName,
		"update",
		slug,
	}

	if name != "" {
		cmdToReturn = append(cmdToReturn, "-n", name)
	}

	if arch != "" {
		cmdToReturn = append(cmdToReturn, "--arch", arch)
	}

	if distro != "" {
		cmdToReturn = append(cmdToReturn, "-d", distro)
	}

	// Add other args to the command
	cmdToReturn = append(cmdToReturn, args...)

	return cmdToReturn
}

// runAndCheckImgBuildCmd is the function allowing to run and check the output of
// the command "rp-cli images localbuild ..."
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param slug: Slug of the image that will be built
// :type slug: string
// :param args: Args to pass as command to create the project
// :type args: ...string
func runAndCheckImgBuildCmd(test *testing.T, slug string, args ...string) {
	// Request for the image localbuild
	cmdOutput, err := execCmdInSimTerm("", args...)
	if err != nil {
		// The command returns an error
		test.Fatal(err)
	}

	// Check that every [OK]/[KO] display is at [OK]
	checkLineEndWith(test, cmdOutput, "Checking that all needed parameters are set...", "[OK]")
	checkLineEndWith(test, cmdOutput, "Checking and preparing the directory for image build...", "[OK]")

	// Check that the correct directory has been set-up
	fullImageDir := path.Join(gitPackagesTestDir, ("image_localbuild_" + slug))
	checkLineEndWith(test, cmdOutput, "\"Image build\" directory set-up:", fullImageDir)
	assert.DirExists(test, fullImageDir)

	// Check that the logs contains the fact that the packaging went well
	assert.Contains(test, cmdOutput, "Congratulations! Your image build seems to be successful!")
	assert.Contains(test, cmdOutput, "Image available in:")

	// Extract the image build logs
	imgBuildLogs := extractSubLogsWithFlag(cmdOutput, "IMAGE BUILD LOGS")

	// Log file used by mocked "redpesk-lad" to send stdout
	logFile := "rpladmock/logs/iotpkg-localimagebuild.log"

	// Check that there are some logs and that they are coherent
	assert.NotEqual(test, "", imgBuildLogs)
	packagingLogsFile := utils.GetPkgRootDir() + logFile
	md5SumImgBuildLogsFile, err := utils.CalculateMd5SumOnFile(packagingLogsFile, true)
	if err != nil {
		test.Fatal(err)
	}
	md5SumImgBuildLogsStr := utils.CalculateMd5SumOnString(imgBuildLogs)

	imgBuildLogsEqual := assert.Equal(test, md5SumImgBuildLogsFile, md5SumImgBuildLogsStr)
	if !imgBuildLogsEqual {
		fmt.Printf("Image build logs from string: \n%s\n", imgBuildLogs)
	}

	// Test that the image ".dsk" has been deleted
	assert.Contains(test, rpladmock.LatestExecCmd, "rm -f /home/devel/.oz/images/")
}

// createProjectAndImage is the function allowing to create easily an image project
// and an image inside it. The caller just have to specify the name of the project
// to create and the name of the image
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param prjName: Name of the project to create
// :type prjName: string
// :param imgName: Name of the image to create
// :type imgName: string
func createProjectAndImage(test *testing.T, prjName, imgName string) error {
	// From the name, create the slug
	prjSlug := strings.ReplaceAll(strings.ToLower(prjName), " ", "-")

	// Firstly, creates a test project of subtype "image", if it does not exists
	cmdToRun1 := getMainAddPrjCmd(prjName, "redpesk-lts-arz")
	cmdToRun1 = append(cmdToRun1, "--images")
	err := addObjectIfNotExists(test, prjsCommand, prjSlug, cmdToRun1...)
	if err != nil {
		return err
	}

	// From the name, create the slug
	imgSlug := strings.ReplaceAll(strings.ToLower(imgName), " ", "-")

	// Then create an image, if it does not exists
	cmdToRun2 := getMainAddImgCmd(imgName, imgSlug, "x86_64", "redpesk-lts-arz",
		"-k", "Redpesk-core-x86_64-smack.ks",
		"--ks-url", "http://git.ovh.iot/redpesk/redpesk-infra/imager-configuration.git",
		"--ks-branch", "02d883f4",
		"--ks-path", "kickstarts/RedPesk",
		"-p", prjSlug)

	err = addObjectIfNotExists(test, imagesCmdName, imgSlug, cmdToRun2...)
	if err != nil {
		return err
	}

	return nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------
