/*
Copyright © 2024 IoT.bzh Company
Contact: https://www.iot.bzh/licensing

This file is part of the redpesk-cli project, the redpesk command line tool.

$RP_BEGIN_LICENSE$
# Commercial License Usage

Licensees holding valid commercial IoT.bzh licenses may use this file in
accordance with the commercial license agreement provided with the
Software or, alternatively, in accordance with the terms contained in
a written agreement between you and The IoT.bzh Company. For licensing terms
and conditions see https://www.iot.bzh/terms-conditions. For further
information use the contact form at https://www.iot.bzh/contact.

# GNU General Public License Usage

Alternatively, this file may be used under the terms of the GNU General
Public license version 3. This license is as published by the Free Software
Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
of this file. Please review the following information to ensure the GNU
General Public License requirements will be met
https://www.gnu.org/licenses/gpl-3.0.html.

$RP_END_LICENSE$
*/
package cmd

import (
	"errors"
	"fmt"
	"io"
	"os"
	"path"
	"strings"
	"syscall"
	"time"

	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/printer"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/rpxxxapi/ladapi"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/signals"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/utils"
	"git.ovh.iot/redpesk/redpesk-infra/rp-webserver/actions/rpapiv1"
	"github.com/creack/goselect"
	"github.com/gofrs/uuid"
	"github.com/ianlewis/go-gitignore"
	"github.com/spf13/cobra"
	"github.com/spf13/pflag"
	"github.com/spf13/viper"
	"golang.org/x/term"
)

// -------------------------------------------------
// Global variable (inside this package) -----------
// -------------------------------------------------

// Root directory inside the localbuilder
var localBuilderRootDir = ""

// -------------------------------------------------
// -------------------------------------------------

// --------------------------------------------------------------
// Initialization functions -------------------------------------
// --------------------------------------------------------------

func initLocalNewCmd(rootCmd *cobra.Command) {

	// localCmd represents the local command
	var localNewCmd = &cobra.Command{
		Use:     localCommand,
		Aliases: []string{"local-v2"},
		Short:   localCmdShortDesc,
	}

	// localExecCmd represents the "exec" command for local
	var localExecCmd = &cobra.Command{
		Use:          localExecCmdName,
		Short:        localExecCmdShortUsage,
		Long:         localExecCmdLongUsage,
		SilenceUsage: true,
		PreRunE:      initEventsRPLAD,
		RunE:         execCmdLine,
		Annotations: map[string]string{
			"__bck_connect__": "FALSE", // No need for backend connection for this command
			"__lad_connect__": "TRUE"},
	}

	// localPkgCmd represents the "local package" command
	var localPkgCmd = &cobra.Command{
		Use:     localPkgCmdName,
		Aliases: []string{localPkgCmdAlias},
		Short:   localPkgCmdShortUsage,
	}

	// localPkgInitCmd represents the "local package init" command
	var localPkgInitCmd = &cobra.Command{
		Use:               localPkgInitCmdName,
		Short:             localPkgInitCmdShortUsage,
		Long:              localPkgInitCmdLongUsage,
		SilenceUsage:      true,
		PreRunE:           initEventsRPLAD,
		RunE:              packageInit,
		ValidArgsFunction: getListAppsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__":  "TRUE",
			"__flag_need_bck_connect__": "TRUE",
			"__lad_connect__":           "TRUE"},
	}

	// localPkgFetchCmd represents the "local package fetch" command
	var localPkgFetchCmd = &cobra.Command{
		Use:               localFetchCmd,
		Short:             localFetchCmdShortDesc,
		Long:              localFetchCmdLongDesc,
		SilenceUsage:      true,
		RunE:              packageFetch,
		ValidArgsFunction: getListAppsAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
		},
	}

	// localPkgStatusCmd represents the "local package status" command
	var localPkgStatusCmd = &cobra.Command{
		Use:               localStatusCmd,
		Short:             localStatusCmdShortDesc,
		Long:              localStatusCmdLongDesc,
		SilenceUsage:      true,
		RunE:              packageStatus,
		ValidArgsFunction: getListAppsFromBldCfgAsTab,
		Annotations: map[string]string{
			"__bck_connect__": "FALSE",
		},
	}

	// localPkgDiffCmd represents the "local package diff" command
	var localPkgDiffCmd = &cobra.Command{
		Use:               localDiffCmd,
		Short:             localDiffCmdShortDesc,
		Long:              localDiffCmdLongDesc,
		SilenceUsage:      true,
		RunE:              packageDiff,
		ValidArgsFunction: getListAppsFromBldCfgAsTab,
	}

	// localPkgCommitCmd represents the "local package commit" command
	var localPkgCommitCmd = &cobra.Command{
		Use:               localCommitCmd,
		Short:             localCommitCmdShortDesc,
		Long:              localCommitCmdLongDesc,
		SilenceUsage:      true,
		PreRunE:           initEventApp,
		RunE:              packageCommit,
		ValidArgsFunction: getListAppsFromBldCfgAsTab,
		Annotations: map[string]string{
			"__bck_ws_connect__": "TRUE"},
	}

	// localPkgBuildCmd represents the "local package build" command
	var localPkgBuildCmd = &cobra.Command{
		Use:                localBuildCmd,
		Short:              localBuildCmdShortUsage,
		Long:               localBuildCmdLongUsage,
		SilenceUsage:       true,
		FParseErrWhitelist: cobra.FParseErrWhitelist{UnknownFlags: true},
		PreRunE:            initEventsRPLAD,
		RunE:               startBuild,
		ValidArgsFunction:  getListAppsFromBldCfgAsTab,
		Annotations: map[string]string{
			"__bck_connect__": "FALSE",
			"__lad_connect__": "TRUE"},
	}

	// localPkgServiceCmd represents the "local package service" command
	var localPkgServiceCmd = &cobra.Command{
		Use:                localServiceCmd,
		Short:              localServiceCmdShortDesc,
		SilenceUsage:       true,
		FParseErrWhitelist: cobra.FParseErrWhitelist{UnknownFlags: true},
		PreRunE:            initEventsRPLAD,
		RunE:               packageService,
		ValidArgsFunction:  getListAppsFromBldCfgAsTab,
		Annotations: map[string]string{
			"__bck_connect__": "FALSE",
			"__lad_connect__": "TRUE"},
	}

	// localPkgMetadataCmd represents the "local package metadata" command
	var localPkgMetadataCmd = &cobra.Command{
		Use:                localMetadataCmd,
		Short:              localMetadataCmdShortDesc,
		SilenceUsage:       true,
		FParseErrWhitelist: cobra.FParseErrWhitelist{UnknownFlags: true},
		PreRunE:            initEventsRPLAD,
		RunE:               packageMetadata,
		ValidArgsFunction:  getListAppsFromBldCfgAsTab,
		Annotations: map[string]string{
			"__bck_connect__": "FALSE",
			"__lad_connect__": "TRUE"},
	}

	// localPkgShellCmd represents the "local package shell" command
	var localPkgShellCmd = &cobra.Command{
		Use:               localShellCmd,
		Short:             localShellCmdShortUsage,
		Long:              localShellCmdLongUsage,
		SilenceUsage:      true,
		PreRunE:           initEventsRPLAD,
		RunE:              startShell,
		ValidArgsFunction: getListAppsFromBldCfgAsTab,
		Annotations: map[string]string{
			"__bck_connect__": "FALSE",
			"__lad_connect__": "TRUE"},
	}

	// localPkgCleanCmd represents the "local package clean" command
	var localPkgCleanCmd = &cobra.Command{
		Use:               localPkgCleanCmdName,
		Short:             localPkgCleanCmdShortUsage,
		Long:              localPkgCleanCmdLongUsage,
		SilenceUsage:      true,
		PreRunE:           initEventsRPLAD,
		RunE:              packageClean,
		ValidArgsFunction: getListAppsFromBldCfgAsTab,
		Annotations: map[string]string{
			"__bck_connect__": "FALSE",
			"__lad_connect__": "TRUE"},
	}

	// localCheckSetUpCmd represents the "local check-setup" command
	var localCheckSetUpCmd = &cobra.Command{
		Use:          localCheckSetupCmd,
		Short:        localCheckSetupCmdShortUsage,
		SilenceUsage: true,
		PreRunE:      initEventsRPLAD,
		RunE:         localCheckSetup,
		Annotations:  map[string]string{"__lad_connect__": "TRUE"},
	}

	// connectionsCmd represents the "local connections" command
	var connectionsCmd = &cobra.Command{
		Use:   connectionsCommand,
		Short: localConnectionsCmdShortDesc,
	}

	// connectionsListCmd represents the "local connections list" command
	var connectionsListCmd = &cobra.Command{
		Use:         connectionsListCmd,
		Short:       fmt.Sprintf(connectionsListCmdShortDesc, "local-builders"),
		Long:        fmt.Sprintf(connectionsListCmdLongDesc, "local-builders"),
		RunE:        listConnections,
		Annotations: map[string]string{"__bck_connect__": "FALSE"},
	}

	// getCmd represents the get command
	var connectionsGetCmd = &cobra.Command{
		Use:               connectionsGetCmd,
		Short:             connectionsGetCmdShortUsage,
		RunE:              getConnection,
		ValidArgsFunction: getLocalBuilderAliases,
		Annotations:       map[string]string{"__bck_connect__": "FALSE"},
	}

	// connectionsCurrentCmd represents the current command
	var connectionsCurrentCmd = &cobra.Command{
		Use:         connectionsCurrentCmd,
		Short:       connectionsCurrentCmdShortUsage,
		RunE:        currentConnection,
		Annotations: map[string]string{"__bck_connect__": "FALSE"},
	}

	// addCmd represents the get command
	var connectionsAddCmd = &cobra.Command{
		Use:         connectionsAddCmd,
		Short:       connectionsAddShortDesc,
		Long:        fmt.Sprintf(connectionsAddLongDesc, "local-builders"),
		RunE:        addConnection,
		Annotations: map[string]string{"__bck_connect__": "FALSE"},
	}

	// updateCmd represents the update command
	var connectionsUpdateCmd = &cobra.Command{
		Use:               connectionsUpdateCmd,
		Short:             connectionsUpdateShortDesc,
		RunE:              updateConnection,
		ValidArgsFunction: getLocalBuilderAliases,
		Annotations:       map[string]string{"__bck_connect__": "FALSE"},
	}

	// deleteCmd represents the delete command
	var connectionsDeleteCmd = &cobra.Command{
		Use:               connectionsDeleteCmd,
		Short:             fmt.Sprintf(connectionsDeleteCmdShortUsage, "local-builder"),
		RunE:              deleteConnection,
		ValidArgsFunction: getLocalBuilderAliases,
		Annotations:       map[string]string{"__bck_connect__": "FALSE"},
	}

	// useCmd represents the use command
	var connectionsUseCmd = &cobra.Command{
		Use:               connectionsUseCmd,
		Short:             fmt.Sprintf(connectionsUseCmdShortDesc, "local-builder"),
		RunE:              useConnection,
		ValidArgsFunction: getLocalBuilderAliases,
		Annotations:       map[string]string{"__bck_connect__": "FALSE"},
	}

	// Here you will define your flags and configuration settings.

	// Persistent Flags -----------------------------------------
	localPkgCmd.PersistentFlags().StringP(localPkgCmdSourceFlag, localPkgCmdSourceFlagAlias, "~/my_rp_builder_dir/gitsources", localPkgCmdSourceFlagUsage)
	localPkgCmd.PersistentFlags().StringP(localPkgCmdPackageFlag, localPkgCmdPackageFlagAlias, "~/my_rp_builder_dir/gitpkgs", localPkgCmdPackageFlagUsage)
	// ----------------------------------------------------------

	// Bind the cobra flags to viper
	viper.BindPFlag(localPkgCmdSourceFlag, localPkgCmd.PersistentFlags().Lookup(localPkgCmdSourceFlag))   // Corresponding environment var: RP_SOURCESDIR
	viper.BindPFlag(localPkgCmdPackageFlag, localPkgCmd.PersistentFlags().Lookup(localPkgCmdPackageFlag)) // Corresponding environment var: RP_PKGSDIR
	// ----------------------------------------------------------

	// Local Flags ----------------------------------------------
	// Exec
	localExecCmd.Flags().StringP(localExecCmdWorkingDirFlag, localExecCmdWorkingDirFlagAlias, "", localExecCmdWorkingDirFlagUsage)
	// Package init
	localPkgInitCmd.Flags().StringP(localProjectFlag, localProjectFlagAlias, "", localProjectFlagUsage)
	localPkgInitCmd.Flags().StringP(localPkgInitBranchFlag, localPkgInitBranchFlagAlias, "", localPkgInitBranchFlagUsage)
	localPkgInitCmd.Flags().BoolP(localPkgInitForceFlag, localPkgInitForceFlagAlias, false, localPkgInitForceFlagUsage)
	// Package fetch
	localPkgFetchCmd.Flags().StringP(localProjectFlag, localProjectFlagAlias, "", localProjectFlagUsage)
	localPkgFetchCmd.Flags().BoolP(localFetchCmdGitpkgFlag, localFetchCmdGitpkgFlagAlias, false, localFetchCmdGitpkgFlagUsage)
	localPkgFetchCmd.Flags().BoolP(localFetchCmdMockCfgFlag, localFetchCmdMockCfgFlagAlias, false, localFetchCmdMockCfgFlagUsage)
	localPkgFetchCmd.Flags().StringP(localPkgInitBranchFlag, localPkgInitBranchFlagAlias, "", localPkgInitBranchFlagUsage)
	localPkgFetchCmd.Flags().BoolP(localFetchCmdForceFlag, localFetchCmdForceFlagAlias, false, localFetchCmdForceFlagUsage)
	// Package status
	localPkgStatusCmd.Flags().StringP(localProjectFlag, localProjectFlagAlias, "", localProjectFlagUsage)
	// Package diff
	localPkgDiffCmd.Flags().StringP(localProjectFlag, localProjectFlagAlias, "", localProjectFlagUsage)
	localPkgDiffCmd.Flags().Bool(localDiffCmdNoColorFlag, false, localDiffCmdNoColorFlagUsage)
	// Package commit
	localPkgCommitCmd.Flags().StringP(localProjectFlag, localProjectFlagAlias, "", localProjectFlagUsage)
	// Package build
	localPkgBuildCmd.Flags().StringP(localProjectFlag, localProjectFlagAlias, "", localProjectFlagUsage)
	localPkgBuildCmd.Flags().StringP(localBuildArchFlag, localBuildArchFlagAlias, localBuildArchFlagDefault, localBuildArchFlagUsage)
	localPkgBuildCmd.Flags().StringP(localBuildDistroFlag, localBuildDistroFlagAlias, "", localBuildDistroFlagUsage)
	localPkgBuildCmd.Flags().BoolP(localBuildOfflineFlag, localBuildOfflineFlagAlias, false, localBuildOfflineFlagUsage)
	// Package service
	localPkgServiceCmd.Flags().StringP(localProjectFlag, localProjectFlagAlias, "", localProjectFlagUsage)
	// Package metadata
	localPkgMetadataCmd.Flags().StringP(localProjectFlag, localProjectFlagAlias, "", localProjectFlagUsage)
	// Package Shell
	localPkgShellCmd.Flags().StringP(localProjectFlag, localProjectFlagAlias, "", localProjectFlagUsage)
	localPkgShellCmd.Flags().StringP(localBuildArchFlag, localBuildArchFlagAlias, localBuildArchFlagDefault, localBuildArchFlagUsage)
	localPkgShellCmd.Flags().StringP(localBuildDistroFlag, localBuildDistroFlagAlias, "", localBuildDistroFlagUsage)
	localPkgShellCmd.Flags().BoolP(localBuildOfflineFlag, localBuildOfflineFlagAlias, false, localBuildOfflineFlagUsage)
	// Package clean
	localPkgCleanCmd.Flags().BoolP(localPkgCleanInitFlag, localPkgCleanInitFlagAlias, false, localPkgCleanInitFlagUsage)
	localPkgCleanCmd.Flags().StringP(localProjectFlag, localProjectFlagAlias, "", localProjectFlagUsage)
	localPkgCleanCmd.Flags().BoolP(localPkgCleanForceFlag, localPkgCleanForceFlagAlias, false, localPkgCleanForceFlagUsage)
	localPkgCleanCmd.Flags().BoolP(localCleanNoMockFlag, localCleanNoMockFlagAlias, false, localCleanNoMockFlagUsage)
	// Connections list
	connectionsListCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	// Connections get
	connectionsGetCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	// Connections current
	connectionsCurrentCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	// Connections add
	connectionsAddCmd.Flags().StringP(connectionsAddIPFlag, connectionsAddIPFlagAlias, "", connectionsAddIPFlagUsage)
	connectionsAddCmd.Flags().StringP(connectionsLocalBldRootDirFlag, connectionsLocalBldRootDirFlagAlias, "", connectionsLocalBldRootDirFlagUsage)
	// Connections update
	connectionsUpdateCmd.Flags().StringP(connectionsUpdateIPFlag, connectionsUpdateIPFlagAlias, flagUpdateKeepCurrentValue, connectionsUpdateIPFlagUsage)
	connectionsUpdateCmd.Flags().StringP(connectionsLocalBldRootDirFlag, connectionsLocalBldRootDirFlagAlias, flagUpdateKeepCurrentValue, connectionsLocalBldRootDirFlagUsage)
	// ----------------------------------------------------------

	// Add some flags as required

	// Add some flags as hidden

	// Specify dynamic flag completion
	localPkgInitCmd.RegisterFlagCompletionFunc(localProjectFlag, getListPrjsAsTab)
	localPkgFetchCmd.RegisterFlagCompletionFunc(localProjectFlag, getListPrjsAsTab)
	localPkgStatusCmd.RegisterFlagCompletionFunc(localProjectFlag, getListPrjsFromBldCfgAsTab)
	localPkgDiffCmd.RegisterFlagCompletionFunc(localProjectFlag, getListPrjsFromBldCfgAsTab)
	localPkgCommitCmd.RegisterFlagCompletionFunc(localProjectFlag, getListPrjsFromBldCfgAsTab)
	localPkgBuildCmd.RegisterFlagCompletionFunc(localProjectFlag, getListPrjsFromBldCfgAsTab)
	localPkgBuildCmd.RegisterFlagCompletionFunc(localBuildArchFlag, getListArchFromBldCfgAsTab)
	localPkgBuildCmd.RegisterFlagCompletionFunc(localBuildDistroFlag, getListDistroFromBldCfgAsTab)
	localPkgServiceCmd.RegisterFlagCompletionFunc(localProjectFlag, getListPrjsFromBldCfgAsTab)
	localPkgMetadataCmd.RegisterFlagCompletionFunc(localProjectFlag, getListPrjsFromBldCfgAsTab)
	localPkgShellCmd.RegisterFlagCompletionFunc(localProjectFlag, getListPrjsFromBldCfgAsTab)
	localPkgShellCmd.RegisterFlagCompletionFunc(localBuildArchFlag, getListArchFromBldCfgAsTab)
	localPkgShellCmd.RegisterFlagCompletionFunc(localBuildDistroFlag, getListDistroFromBldCfgAsTab)
	localPkgCleanCmd.RegisterFlagCompletionFunc(localProjectFlag, getListPrjsFromBldCfgAsTab)

	// Add the command "local" as a subcommand of rp-cli
	rootCmd.AddCommand(localNewCmd)
	// Add the command "exec" as a subcommand of localCmd
	localNewCmd.AddCommand(localExecCmd)
	// Add the command "package" as a subcommand of localCmd
	localNewCmd.AddCommand(localPkgCmd)
	// Add the command "connections" as a subcommand of localCmd
	localNewCmd.AddCommand(connectionsCmd)
	// Add the command "init" as a subcommand of localPkgCmd
	localPkgCmd.AddCommand(localPkgInitCmd)
	// Add the command "fetch" as a subcommand of localPkgCmd
	localPkgCmd.AddCommand(localPkgFetchCmd)
	// Add the command "status" as a subcommand of localPkgCmd
	localPkgCmd.AddCommand(localPkgStatusCmd)
	// Add the command "diff" as a subcommand of localPkgCmd
	localPkgCmd.AddCommand(localPkgDiffCmd)
	// Add the command "commit" as a subcommand of localPkgCmd
	localPkgCmd.AddCommand(localPkgCommitCmd)
	// Add the command "build" as a subcommand of localPkgCmd
	localPkgCmd.AddCommand(localPkgBuildCmd)
	// Add the command "service" as a subcommand of localPkgCmd
	localPkgCmd.AddCommand(localPkgServiceCmd)
	// Add the command "metadata" as a subcommand of localPkgCmd
	localPkgCmd.AddCommand(localPkgMetadataCmd)
	// Add the command "shell" as a subcommand of localPkgCmd
	localPkgCmd.AddCommand(localPkgShellCmd)
	// Add the command "clean" as a subcommand of localPkgCmd
	localPkgCmd.AddCommand(localPkgCleanCmd)
	// Add the command "check-setup" as a subcommand of localCmd
	localNewCmd.AddCommand(localCheckSetUpCmd)
	// Add the command "list" as a subcommand of connectionsCmd
	connectionsCmd.AddCommand(connectionsListCmd)
	// Add the command "get" as a subcommand of connectionsCmd
	connectionsCmd.AddCommand(connectionsGetCmd)
	// Add the command "current" as a subcommand of connectionsCmd
	connectionsCmd.AddCommand(connectionsCurrentCmd)
	// Add the command "add" as a subcommand of connectionsCmd
	connectionsCmd.AddCommand(connectionsAddCmd)
	// Add the command "update" as a subcommand of connectionsCmd
	connectionsCmd.AddCommand(connectionsUpdateCmd)
	// Add the command "delete" as a subcommand of connectionsCmd
	connectionsCmd.AddCommand(connectionsDeleteCmd)
	// Add the command "use" as a subcommand of connectionsCmd
	connectionsCmd.AddCommand(connectionsUseCmd)
}

// initEventsRPLAD inits everything that is needed to communicate
// with the Redpesk LAD service through the websocket
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func initEventsRPLAD(cmd *cobra.Command, args []string) error {

	// Subscribe to all the events in relation with local-builder
	err := localBConnection.RPEventRegister(localStdOutEvt)
	if err != nil {
		return err
	}
	err = localBConnection.RPEventRegister(localStdErrEvt)
	if err != nil {
		return err
	}
	err = localBConnection.RPEventRegister(localProcEndEvt)
	if err != nil {
		return err
	}

	// Save the call-backs needed when some events are received
	// from the local builder
	err = localBConnection.WSocket.OnEvent(localStdOutEvt, CLIReceiveLocalBEvent)
	if err != nil {
		return err
	}
	err = localBConnection.WSocket.OnEvent(localStdErrEvt, CLIReceiveLocalBEvent)
	if err != nil {
		return err
	}
	err = localBConnection.WSocket.OnEvent(localProcEndEvt, CLIReceiveLocalBEvent)
	if err != nil {
		return err
	}

	// Find the root directory to use for the current local-builder
	localBldRootExists := false
	if localBuilderRootDir == "" {
		dirsToCheck := []string{"/home/devel/rp-build", "/home/devel", "/root"}
		for _, directory := range dirsToCheck {
			localBuilderRootDir = directory
			_, err := execInLocalBuilderStatic("ls "+directory, "/tmp")
			if err == nil {
				localBldRootExists = true
				break
			}
		}
	} else {
		// Check that the given directory exists in the localbuilder
		_, err = execInLocalBuilderStatic("ls "+localBuilderRootDir, "/tmp")
		if err == nil {
			localBldRootExists = true
		}
	}
	if !localBldRootExists {
		return fmt.Errorf(rootBuildDirDoesntExist, localBuilderRootDir)
	}
	Log.Infoln("Local-builder internal build root directory:", "'"+localBuilderRootDir+"'")

	return nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Completion functions -----------------------------------------
// --------------------------------------------------------------

// getListObjectFromBldCfgAsTab gets all the object slugs
// corresponding to the local build configurations found in the
// host, it can either returns appSlug (index = 1) or
// projectSlug (index = 0)
//
// :param index: index of the name in the key value
// :type index: int
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getListObjectFromBldCfgAsTab(index int) ([]string, cobra.ShellCompDirective) {
	appsList := []string{}

	allCfg, err := retrieveAllLocalBldConfigs()
	if err != nil {
		return appsList, cobra.ShellCompDirectiveNoFileComp
	}

	for key := range allCfg {
		splitKey := strings.Split(key, "/")
		if (len(splitKey) > 1) && !utils.ContainsString(appsList, splitKey[index]) {
			appsList = append(appsList, splitKey[index])
		}
	}

	return appsList, cobra.ShellCompDirectiveNoFileComp
}

// getListAppsFromBldCfgAsTab gets all the applications slugs
// corresponding to the local build configurations found in the
// host
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getListAppsFromBldCfgAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	return getListObjectFromBldCfgAsTab(1)
}

// getListPrjsFromBldCfgAsTab gets all the project slugs
// corresponding to the local build configurations found in the
// host
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getListPrjsFromBldCfgAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	return getListObjectFromBldCfgAsTab(0)
}

// getListDistroFromBldCfgAsTab gets all the distributions
// from the local build configurations found in the host
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getListDistroFromBldCfgAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	distroList := []string{}

	allCfg, err := retrieveAllLocalBldConfigs()
	if err != nil {
		return distroList, cobra.ShellCompDirectiveNoFileComp
	}

	for _, config := range allCfg {
		distroList = append(distroList, config.DistrosList...)
	}

	return utils.RemoveDuplicateInStrSlice(distroList), cobra.ShellCompDirectiveNoFileComp
}

// getListArchFromBldCfgAsTab gets all the architectures
// from the local build configurations found in the host
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getListArchFromBldCfgAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	archList := []string{}

	allCfg, err := retrieveAllLocalBldConfigs()
	if err != nil {
		return archList, cobra.ShellCompDirectiveNoFileComp
	}

	for _, config := range allCfg {
		archList = append(archList, config.ArchsList...)
	}

	return utils.RemoveDuplicateInStrSlice(archList), cobra.ShellCompDirectiveNoFileComp
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Command functions --------------------------------------------
// --------------------------------------------------------------

// execCmdLine runs the command line inside the local-builder
// configured on the host
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func execCmdLine(cmd *cobra.Command, args []string) error {

	// Get the working directory to use (if set by user)
	workingDir, err := cmd.Flags().GetString(localExecCmdWorkingDirFlag)
	if err != nil {
		logFlagError(localExecCmdWorkingDirFlag)
		return err
	}

	// If the user did not set any working directory, check if we are not inside
	// a gitsources/ or gitpkgs/ directory or sub-directory
	if workingDir == "" {
		// Get the "pkgsdir" value in config file
		pkgsDir := ""
		pkgsDir, err = manuallyGetPkgDirFlag()
		if err != nil {
			return err
		}
		if pkgsDir != "" {
			pkgsDir, err = utils.SanitizePath(pkgsDir)
			if err != nil {
				return err
			}
		}

		// Get the "sourcesdir" value in config file
		srcDir := ""
		srcDir, err = manuallyGetSrcDirFlag()
		if err != nil {
			return err
		}
		if srcDir != "" {
			srcDir, err = utils.SanitizePath(srcDir)
			if err != nil {
				return err
			}
		}

		// Get the current working directory
		currentDir, err := os.Getwd()
		if err != nil {
			return err
		}

		// Check if the user is in gitpkgs/ dir
		isInPkgsDir := strings.HasPrefix(currentDir, pkgsDir)

		// Check if the user is in gitsources/ dir
		isInSrcsDir := strings.HasPrefix(currentDir, srcDir)

		if isInPkgsDir && (pkgsDir != "") {
			// Replace the right directory in the workingDir to pass
			workingDir = strings.ReplaceAll(currentDir, pkgsDir, getLclBldGitpkgsDir())
		} else if isInSrcsDir && (srcDir != "") {
			// Replace the right directory in the workingDir to pass
			workingDir = strings.ReplaceAll(currentDir, srcDir, getLclBldGitsourcesDir())
		}
	}

	// Get the command line passed by the user
	cmdLine, err := getCmdLine(args)
	if err != nil {
		return err
	}

	// Get the size of the current terminal
	termWidth, termHeight, err := term.GetSize(int(os.Stdin.Fd()))
	if err != nil {
		return err
	}

	// Request local-builder to run the command
	// Request body
	execRequest := ladapi.RPLADExecCmdStruct{
		RPLADMainCmdParams: ladapi.RPLADMainCmdParams{
			RPLADWinSize: ladapi.RPLADWinSize{
				WinSizeRows: termHeight,
				WinSizeCol:  termWidth,
			},
			WorkingDir: workingDir,
		},
		CmdLine: cmdLine,
	}
	// Answer body template
	procTicket := ladapi.RPLADRunnerTicket{}
	err = localBAPI.ExecPost(&execRequest, &procTicket)
	if err != nil {
		return err
	}

	err = listenStdoutErrAndIn(procTicket)

	return err
}

// packageInit initializes everything that is needed to package
// an application inside the local-builder
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func packageInit(cmd *cobra.Command, args []string) error {
	RPPrinter.CustomPrintLn(localPkgInitCmdAnsLine0)

	// Before anything check that 'git' and 'git-lfs' are installed in the system
	_, err := utils.RunCmdInDirectory("/tmp/", "git", "--version")
	if err != nil {
		return errors.New(gitPkgNotInstalled)
	}

	// Get the "force" flag value
	forceFlag, err := cmd.Flags().GetBool(localPkgInitForceFlag)
	if err != nil {
		logFlagError(localPkgInitForceFlag)
		return err
	}

	// Get the first main infos we need
	fullConfiguration, appStructure, err := prePackagingRoutine(cmd, args)
	if err != nil {
		return err
	}

	// Display something like "Retrieving packaging configurations" here
	RPPrinter.CustomPrintF(localPkgInitCmdAnsLine3, appStructure.Slug)

	// Get the branch flag
	fullConfiguration.Branch, err = cmd.Flags().GetString(localPkgInitBranchFlag)
	if err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	// Display something like "Setting up the local-builder" here
	RPPrinter.CustomPrintLn(localPkgInitCmdAnsLine4)

	// Set-up the local builder
	err = localBuilderSetUp(&appStructure, &fullConfiguration, forceFlag)
	if err != nil {
		return err
	}

	return nil
}

// packageStatus shows the status of files in a local working directory
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func packageStatus(cmd *cobra.Command, args []string) error {
	RPPrinter.CustomPrintLn(localStatusCmdAnsLine0)

	// Get full context -----------------------------------
	appStructure, _, rootGitpkgDir, err := getLocalContext(cmd, args, false, true, false)
	if err != nil {
		return err
	}
	// ----------------------------------------------------

	fullGitpkgDir := path.Join(rootGitpkgDir, appStructure.Parent.ProjectSlug, appStructure.PackageName)

	// Get the 'local_build_config.json' file to print the branch used
	localCfg, _, err := getLocalBuildConfig(fullGitpkgDir, false)
	if err != nil {
		return err
	}
	// ----------------------------------------------------

	// Print the branch used
	if localCfg.Branch != "" {
		RPPrinter.CustomPrintLn()
		RPPrinter.CustomPrintF(localStatusCmdAnsLine2, localCfg.Branch)
	}

	RPPrinter.CustomPrintLn()
	RPPrinter.CustomPrintLn(localStatusCmdAnsLine1)

	// Finally, check the status
	_, _, err = checkGitpkgsStatus(fullGitpkgDir, true, false)

	RPPrinter.CustomPrintLn()

	return err

}

// packageDiff shows the diff of files in a local working directory
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func packageDiff(cmd *cobra.Command, args []string) error {
	// Check that diff binary is install
	_, err := utils.RunCmdInDirectory("/tmp/", "diff", "--version")
	if err != nil {
		return errors.New(diffUtilsPkgNotInstalled)
	}

	// Get full context -----------------------------------
	appStructure, _, rootGitpkgDir, err := getLocalContext(cmd, args, false, false, true)
	if err != nil {
		return err
	}
	// ----------------------------------------------------
	fullGitpkgDir := path.Join(rootGitpkgDir, appStructure.Parent.ProjectSlug, appStructure.PackageName)

	// Get status details ---------------------------------
	statusBool, modifDetails, err := checkGitpkgsStatus(fullGitpkgDir, false, false)
	if err != nil {
		return err
	}
	// ----------------------------------------------------

	// If status true: nothing to display (because no modification)
	if statusBool {
		return nil
	}

	// Handle the case where we need to switch branch
	previousBranch, err := handleBranchSwitch(fullGitpkgDir, appStructure.ID)
	if err != nil {
		return err
	}
	if previousBranch != "" {
		// Defer switching back to previous branch
		defer switchToExistingBranch(appStructure.ID, previousBranch)
	}
	// ----------------------------------------------------

	// Create a temporary directory -----------------------
	originalDirectory := path.Join(fullGitpkgDir, localCfgSubDirectory, localOriginFilesSubDirectory)
	err = utils.CreateDirIfNotExists(originalDirectory, true)
	if err != nil {
		return err
	}
	defer os.RemoveAll(originalDirectory)
	// ----------------------------------------------------

	// Get the "no-color flag"
	noColor, err := cmd.Flags().GetBool(localDiffCmdNoColorFlag)
	if err != nil {
		logFlagError(localDiffCmdNoColorFlag)
		return err
	}

	// Get the list of files that the user may have given to rp-cli
	userFiles := []string{}
	if len(args) > 1 {
		userFiles = args[1:]
	}

	// If the user did not specify anything: save all the different files as user specified files
	if len(userFiles) < 1 {
		userFiles = append(userFiles, modifDetails.Added...)
		userFiles = append(userFiles, modifDetails.Modified...)
		userFiles = append(userFiles, modifDetails.Deleted...)
	}

	// Now, for each user file, display the diff
	for index, userFile := range userFiles {

		// Check if the file is either added, modified, deleted or none of it
		fileAdded, fileModified, fileDeleted := false, false, false
		if fileAdded = utils.ContainsString(modifDetails.Added, userFile); !fileAdded {
			if fileModified = utils.ContainsString(modifDetails.Modified, userFile); !fileModified {
				fileDeleted = utils.ContainsString(modifDetails.Deleted, userFile)
			}
		}

		// Check that the file is at the root of the directory, ignore in the other case
		if path.Dir(userFile) != "." {
			Log.Warnf(ignoreFilesInSubDir, userFile)
			continue
		}

		// According to the case, adapt the diff behavior
		first_arg, second_arg := "", ""
		switch {
		case fileAdded:
			first_arg = "/dev/null"
			second_arg = userFile
		case fileDeleted:
			// Download the original file
			err = downloadGitPkgsAppFiles(appStructure.ID, originalDirectory, []string{userFile}, false)
			if err != nil {
				return err
			}
			first_arg = path.Join(localCfgSubDirectory, localOriginFilesSubDirectory, userFile)
			second_arg = "/dev/null"
		case fileModified:
			// Download the original file
			err = downloadGitPkgsAppFiles(appStructure.ID, originalDirectory, []string{userFile}, false)
			if err != nil {
				return err
			}
			first_arg = path.Join(localCfgSubDirectory, localOriginFilesSubDirectory, userFile)
			second_arg = userFile
		default:
			continue
		}

		// Create 'diff' color flag according to what the user want
		diffColorFlag := "--color=always"
		if noColor {
			diffColorFlag = "--color=never"
		}

		// Now call the 'diff' command
		diffArgs := []string{diffColorFlag, "-u", first_arg, second_arg}
		Log.Infof("diff command executed: 'diff %s'\n", strings.Join(diffArgs, " "))

		diffOutput, err := utils.RunCmdInDirectory(fullGitpkgDir, "diff", diffArgs...)
		if (err != nil) && (err.Error() != "exit status 1") {
			// Treat as error uniquely if the status is different than 1
			// Indeed, for diff, a '1' status simply means that there are some differences
			return err
		}

		// Print the output
		fmt.Fprint(RPPrinter.Writer, diffOutput)

		if index < (len(userFiles) - 1) {
			fmt.Fprintln(RPPrinter.Writer)
		}
	}

	return nil
}

// packageFetch fetches the app content and the local-builder configuration
// in order to be able to build it locally
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func packageFetch(cmd *cobra.Command, args []string) error {
	RPPrinter.CustomPrintLn(localFetchCmdAnsLine0)

	// Get the value of the different flags ---------------
	gitpkgFlag, err := cmd.Flags().GetBool(localFetchCmdGitpkgFlag)
	if err != nil {
		logFlagError(localFetchCmdGitpkgFlag)
		return err
	}
	localBuilderFlag, err := cmd.Flags().GetBool(localFetchCmdMockCfgFlag)
	if err != nil {
		logFlagError(localFetchCmdMockCfgFlag)
		return err
	}
	forceFlag, err := cmd.Flags().GetBool(localFetchCmdForceFlag)
	if err != nil {
		logFlagError(localFetchCmdForceFlag)
		return err
	}
	getEverything := !gitpkgFlag && !localBuilderFlag

	// Get the branch flag
	branch, err := cmd.Flags().GetString(localPkgInitBranchFlag)
	if err != nil {
		return err
	}
	// ----------------------------------------------------

	// Get full context -----------------------------------
	appStructure, rootSrcDir, rootGitpkgDir, err := getLocalContext(cmd, args, false, false, false)
	if err != nil {
		return err
	}
	// ----------------------------------------------------

	localConfig := RPLocalBuildConfig{
		SourcesDir:  rootSrcDir,
		PackagesDir: rootGitpkgDir,
		Branch:      branch,
		AppPkgName:  appStructure.PackageName,
		AppSlug:     appStructure.Slug,
		ProjectSlug: appStructure.Parent.ProjectSlug,
	}

	// Now do the work!
	return fetchGitpkgAndMockFiles(&localConfig,
		&appStructure,
		(gitpkgFlag || getEverything),
		(localBuilderFlag || getEverything),
		forceFlag)
}

// packageCommit uploads files that have been modified/added/deleted in
// the working directory, to the remote gitpkg in the redpesk factory
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func packageCommit(cmd *cobra.Command, args []string) error {
	RPPrinter.CustomPrintLn(localCommitCmdAnsLine0)

	// Get full context -----------------------------------
	appStructure, _, rootGitpkgDir, err := getLocalContext(cmd, args, false, false, false)
	if err != nil {
		return err
	}
	// ----------------------------------------------------

	fullGitpkgDir := path.Join(rootGitpkgDir, appStructure.Parent.ProjectSlug, appStructure.PackageName)

	RPPrinter.CustomPrintLn()
	RPPrinter.CustomPrintLn(localStatusCmdAnsLine1)

	// Get status details ---------------------------------
	statusBool, modifDetails, err := checkGitpkgsStatus(fullGitpkgDir, true, false)
	if err != nil {
		return err
	}
	// ----------------------------------------------------
	RPPrinter.CustomPrintLn()

	// If the status is true, we have nothing to commit!
	if statusBool {
		RPPrinter.CustomPrintLn(localCommitNothingToCommit)
		return nil
	}

	// Get the list of files that the user may have given to rp-cli
	userFiles := []string{}
	if len(args) > 1 {
		userFiles = args[1:]
	}

	// If we reached here, this means that we have some files to upload/remove

	// Handle the case where we need to switch branch
	previousBranch, err := handleBranchSwitch(fullGitpkgDir, appStructure.ID)
	if err != nil {
		return err
	}
	if previousBranch != "" {
		// Defer switching back to previous branch
		defer switchToExistingBranch(appStructure.ID, previousBranch)
	}

	// Upload all added/modified files
	filesToUpload := append(modifDetails.Modified, modifDetails.Added...)
	filesTabToUpload := []string{}
	for _, tmpFile := range filesToUpload {

		// Filter the files to upload
		if (len(userFiles) > 0) && !utils.ContainsString(userFiles, tmpFile) {
			continue
		}

		// Check if there are files inside a new directory and return an error if it is the case
		if path.Dir(tmpFile) != "." {
			return fmt.Errorf(filesMustBeAtRootOfGitpkg, path.Base(tmpFile))
		}

		// Now add the current file to the directory to upload
		filesTabToUpload = append(filesTabToUpload, path.Join(fullGitpkgDir, tmpFile))
	}

	err = uploadFileUtility(filesTabToUpload, appStructure.ID, false)
	if err != nil {
		return err
	}

	// Remove all removed files
	for _, tmpFile := range modifDetails.Deleted {

		// Filter the files to delete
		if (len(userFiles) > 0) && !utils.ContainsString(userFiles, tmpFile) {
			continue
		}

		RPPrinter.CustomPrintF(localCommitRemoval, tmpFile)

		// Now remove the current file
		err := bckAPI.ApplicationRemoveFile(appStructure.ID.String(), tmpFile)
		if err != nil {
			Log.Error("error when removing ", tmpFile, ": ", err)
			return err
		}
	}

	// Update the md5sum file
	return updateOrCreateMD5SumFile(fullGitpkgDir, userFiles)
}

// startBuild builds locally the application passed as argument
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func startBuild(cmd *cobra.Command, args []string) error {
	// Before anything, handle the case where the user is asking for help
	helpAsked, err := handleIotpkgSubCmdHelp(cmd, originalArgs, "build")
	if err != nil {
		return err
	}
	if helpAsked {
		// Reset the global variable (mainly for tests)
		originalArgs = []string{}
		return nil
	}

	RPPrinter.CustomPrintLn(localBuildRunCmdAnsLine0)

	// Parse rp-cli flags and extract iotpkg args ---------
	iotpkgArgs, err := getIotpkgArgs(cmd)
	if err != nil {
		return err
	}
	Log.Debugf("[LOCAL] Iotpkg args: %+v\n", iotpkgArgs)
	// ----------------------------------------------------

	// Get local context -----------------------------------
	appStructure, _, fullGitPkgDir, err := getLocalContext(cmd, cmd.Flags().Args(), true, true, false)
	if err != nil {
		return err
	}
	// ----------------------------------------------------

	// Get the build context (arch and distro) ------------
	archToBld, distroToBld, err := getLocalBuildContext(cmd, &appStructure)
	if err != nil {
		return err
	}
	// ----------------------------------------------------

	// Delete the previous'result' directory if it exists -
	fullResultDir := path.Join(fullGitPkgDir, getResultsDirName(fullGitPkgDir))
	if err = os.RemoveAll(fullResultDir); err != nil {
		return err
	}
	// ----------------------------------------------------

	// Get the offline flag -------------------------------
	offline, err := cmd.Flags().GetBool(localBuildOfflineFlag)
	if err != nil {
		logFlagError(localBuildOfflineFlag)
		return err
	}
	// ----------------------------------------------------

	RPPrinter.CustomColorPrintLn(printer.ColorPurple, localLocalBldLogsTag1)
	RPPrinter.ForceFlush()

	// Get the size of the current terminal
	termWidth, termHeight, err := term.GetSize(int(os.Stdin.Fd()))
	if err != nil {
		return err
	}

	// Create the gitpkg directory path from the local-builder POV
	localBldGitpkgDir := path.Join(getLclBldGitpkgsDir(), appStructure.Parent.ProjectSlug, appStructure.PackageName)

	// Add the flag "--repofile" if it is not already set
	if !utils.ContainsString(iotpkgArgs, "--repofile") && !utils.ContainsString(iotpkgArgs, "-r") {
		iotpkgArgs = append(iotpkgArgs, "--repofile", path.Join(localBldGitpkgDir, localCfgSubDirectory, appStructure.Parent.ProjectSlug+".json"))
	}

	// Add the distro and the architecture flag
	iotpkgArgs = append(iotpkgArgs, "--arch", archToBld, "--distro", distroToBld)

	// If the user specify the offline flags, add the corresponding flag to `iotpkg`
	if offline {
		iotpkgArgs = append(iotpkgArgs, "--offline")
	}

	buildRequest := ladapi.RPLADIotpkgCmdStruct{
		RPLADMainCmdParams: ladapi.RPLADMainCmdParams{
			RPLADWinSize: ladapi.RPLADWinSize{
				WinSizeRows: termHeight,
				WinSizeCol:  termWidth,
			},
			WorkingDir: localBldGitpkgDir,
		},
		IotpkgFlags: []string{"-v"},
		SubcmdFlags: iotpkgArgs,
	}

	procTicket := ladapi.RPLADRunnerTicket{}
	err = localBAPI.IotpkgRPBuildPost(&buildRequest, &procTicket)
	if err != nil {
		return err
	}

	err = listenStdoutErrAndIn(procTicket)
	RPPrinter.CustomColorPrintLn(printer.ColorPurple, localLocalBldLogsTag2)

	// Construct the directory containing the results
	resultsPath := fmt.Sprintf("%s/results_%s/", fullGitPkgDir, appStructure.PackageName)

	// Display the info to the user
	RPPrinter.CustomColorPrintLn(printer.ColorCyan, localBuildRunCmdAnsResultIn, resultsPath)
	RPPrinter.CustomPrintLn()

	if err == nil {
		// Congrats the user if the packaging went well!
		RPPrinter.CustomColorPrintLn(printer.ColorGreen, localBuildRunCmdAnsResultGood)
	} else {
		RPPrinter.CustomColorPrintLn(printer.ColorRed, localBuildRunCmdAnsResultBad)
	}

	return err
}

// startShell start a shell in the mock in the local-builder
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func startShell(cmd *cobra.Command, args []string) error {
	RPPrinter.CustomPrintLn(localShellCmdAnsLine0)

	// Get local context -----------------------------------
	appStructure, _, _, err := getLocalContext(cmd, cmd.Flags().Args(), true, true, false)
	if err != nil {
		return err
	}
	// ----------------------------------------------------

	// Get the build context (arch and distro) ------------
	archToBld, distroToBld, err := getLocalBuildContext(cmd, &appStructure)
	if err != nil {
		return err
	}
	// ----------------------------------------------------

	// Get the offline flag -------------------------------
	offline, err := cmd.Flags().GetBool(localBuildOfflineFlag)
	if err != nil {
		logFlagError(localBuildOfflineFlag)
		return err
	}
	// ----------------------------------------------------

	RPPrinter.CustomColorPrintLn(printer.ColorPurple, localLocalBldLogsTag1)
	RPPrinter.ForceFlush()

	// Get the size of the current terminal
	termWidth, termHeight, err := term.GetSize(int(os.Stdin.Fd()))
	if err != nil {
		return err
	}

	// Create the gitpkg directory path from the local-builder POV
	localBldGitpkgDir := path.Join(getLclBldGitpkgsDir(), appStructure.Parent.ProjectSlug, appStructure.PackageName)

	// Add the flag "--repofile" in the flags, also the distro, the architecture flag and finally, the shell flag
	iotpkgArgs := []string{
		"--repofile", path.Join(localBldGitpkgDir, localCfgSubDirectory, appStructure.Parent.ProjectSlug+".json"),
		"--arch", archToBld,
		"--distro", distroToBld,
		"--shell",
	}

	// If the user specify the offline flags, add the corresponding flag to `iotpkg`
	if offline {
		iotpkgArgs = append(iotpkgArgs, "--offline")
	}

	buildRequest := ladapi.RPLADIotpkgCmdStruct{
		RPLADMainCmdParams: ladapi.RPLADMainCmdParams{
			RPLADWinSize: ladapi.RPLADWinSize{
				WinSizeRows: termHeight,
				WinSizeCol:  termWidth,
			},
			WorkingDir: localBldGitpkgDir,
		},
		IotpkgFlags: []string{"-v"},
		SubcmdFlags: iotpkgArgs,
	}

	procTicket := ladapi.RPLADRunnerTicket{}
	err = localBAPI.IotpkgRPBuildPost(&buildRequest, &procTicket)
	if err != nil {
		return err
	}

	err = listenStdoutErrAndIn(procTicket)
	RPPrinter.CustomColorPrintLn(printer.ColorPurple, localLocalBldLogsTag2)

	return err
}

// packageClean cleans the repositories used to build locally an
// application
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func packageClean(cmd *cobra.Command, args []string) error {
	RPPrinter.CustomPrintLn(localPkgCleanCmdAnsLine0)
	RPPrinter.CustomPrintLn()

	// Initialize the configuration structure that we may use
	fullConfiguration := RPLocalBuildConfig{}

	// Before anything check that 'git' is installed in the system
	_, err := utils.RunCmdInDirectory("/tmp/", "git", "--version")
	if err != nil {
		return errors.New(gitPkgNotInstalled)
	}

	// Get the init flag
	initFlag, err := cmd.Flags().GetBool(localPkgCleanInitFlag)
	if err != nil {
		logFlagError(localPkgCleanInitFlag)
		return err
	}

	// Get the force flag
	forceFlag, err := cmd.Flags().GetBool(localPkgCleanForceFlag)
	if err != nil {
		logFlagError(localPkgCleanForceFlag)
		return err
	}

	// Try to find the paths of the directories to delete
	srcDir, pkgsDir, err := findPathsToDelete(cmd, args)
	if err != nil {
		return err
	}

	if initFlag {
		// Retrieve the configuration file from the already existing pkgs directory
		configFilePath := path.Join(pkgsDir, localCfgSubDirectory, localBldCfgFile)
		err = utils.ReadJsonFile(configFilePath, &fullConfiguration)
		if err != nil {
			return err
		}
	}

	if !forceFlag && srcDir != "" {
		// Check if there is modification in source directory
		gitStatusOutput, err := utils.RunCmdInDirectory(srcDir, "git", "status")
		if err != nil {
			Log.Warnf(noGitSourceFoundWarn, gitStatusOutput)
		} else if !strings.Contains(gitStatusOutput, "nothing to commit, working tree clean") {
			// Ask the user if he really wants to delete the repository
			answerLow := rpYesOrNoQuestion(localPkgCleanCmdAnsLine1)
			if answerLow == "n" {
				// The user does not want to delete the directory, stop here
				return nil
			}
		}
	}

	// Delete the mock if requested
	noDelMock, err := cmd.Flags().GetBool(localCleanNoMockFlag)
	if err != nil {
		return err
	}
	if !noDelMock {
		iotpkgFullCfgDir := path.Join(pkgsDir, localIotpkgSubDirectory)

		// Check if the .iotpkg/ directory exists
		_, err := os.Stat(iotpkgFullCfgDir)
		if !os.IsNotExist(err) {
			// Get the name of the mock(s) to delete (go look into .iotpkg/ directory inside the gitpkgs/ directory)
			dirsAndFilesMap, err := utils.FindAllFilesAndDirsRecursively(iotpkgFullCfgDir)
			if err != nil {
				return err
			}

			// For each mock name found, delete the mock and the corresponding ".cfg" file
			for _, filepath := range dirsAndFilesMap["files"] {
				tmpFilename := path.Base(filepath)

				// Find the real mock name within the .cfg file
				mockName, err := getMockNameFromCfgFile(filepath)
				if err != nil {
					Log.Errorf("error while getting the mock name for '%s': %+v", tmpFilename, err)
				}

				// Check that we got a name
				if mockName == "" {
					Log.Warnf("No mock found for %s", tmpFilename)
					continue
				}

				// User communication
				RPPrinter.CustomPrintF("Deleting mock '%s'...\n", mockName)

				// Ask redpesk LAD to delete the mock
				err = handleIotpkgCleanMock(mockName)
				if err != nil {
					Log.Errorf("error while cleaning the mock named '%s': %+v", mockName, err)
					// Does not stop on error at this level
				}

				// Now, delete the corresponding .cfg file
				tmpPkgsDir, err := getCustomBuildsDirs(localPkgCmdPackageFlag)
				if err != nil {
					return err
				}
				// Try '/home/my-user/my_rp_builder_dir/gitpkgs/'
				if err = utils.RemoveFileIfExists(path.Join(tmpPkgsDir, tmpFilename)); err != nil {
					return err
				}
				// Try '/home/my-user/my_rp_builder_dir/.iotpkgroot/'
				if err = utils.RemoveFileIfExists(path.Join(tmpPkgsDir, "../.iotpkgroot", tmpFilename)); err != nil {
					return err
				}
			}
		}
	}

	// Delete the directories
	err = os.RemoveAll(srcDir)
	if err != nil {
		return err
	}
	err = os.RemoveAll(pkgsDir)
	if err != nil {
		return err
	}

	// If the parent directory is empty, delete it as well!
	parentSrcDir := path.Dir(srcDir)
	err = utils.RemoveDirIfEmpty(parentSrcDir)
	if err != nil {
		return err
	}
	parentPkgDir := path.Dir(pkgsDir)
	err = utils.RemoveDirIfEmpty(parentPkgDir)
	if err != nil {
		return err
	}

	if initFlag {
		// Run the local package init command using the fields retrieved in the configuration file

		// Get the root command
		rpCliRootCmd := cmd.Root()

		// Find the local package init command from that
		cmdLine := []string{localCommand, "package", "init", path.Base(srcDir)}
		localPkgInitCmd, inArgs, err := rpCliRootCmd.Find(cmdLine)
		if err != nil {
			return err
		}

		projectSlug := path.Base(path.Dir(srcDir))

		// Set the different flags needed
		localPkgInitCmd.Flags().Set(localPkgCmdPackageFlag, fullConfiguration.PackagesDir)
		localPkgInitCmd.Flags().Set(localPkgCmdSourceFlag, fullConfiguration.SourcesDir)
		localPkgInitCmd.Flags().Set(localProjectFlag, projectSlug)

		// As log info, print the command line we are about to play
		// Log.Info(initAfterCleanInfoMsg, strings.Join(cmdLine, " "))

		RPPrinter.CustomPrintLn()
		RPPrinter.ForceFlush()

		// Now call the command
		err = localPkgInitCmd.PreRunE(localPkgInitCmd, inArgs)
		if err != nil {
			return err
		}

		err = localPkgInitCmd.RunE(localPkgInitCmd, inArgs)
		if err != nil {
			return err
		}
	}

	RPPrinter.CustomColorPrintLn(printer.ColorGreen, localPkgCleanCmdAnsLine2)

	return nil
}

// packageService allow to set app service value using `iotpkg` inside
// the local-builder
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func packageService(cmd *cobra.Command, args []string) error {

	// Before anything, handle the case where the user is asking for help
	helpAsked, err := handleIotpkgSubCmdHelp(cmd, originalArgs, "service")
	if err != nil {
		return err
	}
	if helpAsked {
		// Reset the global variable (mainly for tests)
		originalArgs = []string{}
		return nil
	}

	RPPrinter.CustomPrintLn(localServiceCmdAnsLine0)

	// Parse rp-cli flags and extract iotpkg args ---------
	iotpkgArgs, err := getIotpkgArgs(cmd)
	if err != nil {
		return err
	}
	Log.Debugf("[LOCAL] Iotpkg args: %+v\n", iotpkgArgs)
	// ----------------------------------------------------

	// Get full context -----------------------------------
	appStructure, _, _, err := getLocalContext(cmd, cmd.Flags().Args(), true, true, false)
	if err != nil {
		return err
	}
	// ----------------------------------------------------

	RPPrinter.CustomColorPrintLn(printer.ColorPurple, localLocalBldLogsTag1)
	RPPrinter.ForceFlush()

	// Get the size of the current terminal
	termWidth, termHeight, err := term.GetSize(int(os.Stdin.Fd()))
	if err != nil {
		return err
	}

	serviceRequest := ladapi.RPLADIotpkgCmdStruct{
		RPLADMainCmdParams: ladapi.RPLADMainCmdParams{
			RPLADWinSize: ladapi.RPLADWinSize{
				WinSizeRows: termHeight,
				WinSizeCol:  termWidth,
			},
			WorkingDir: path.Join(getLclBldGitpkgsDir(), appStructure.Parent.ProjectSlug, appStructure.PackageName),
		},
		IotpkgFlags: []string{"-v"},
		SubcmdFlags: iotpkgArgs,
	}

	procTicket := ladapi.RPLADRunnerTicket{}
	err = localBAPI.IotpkgServicePost(&serviceRequest, &procTicket)
	if err != nil {
		return err
	}

	err = listenStdoutErrAndIn(procTicket)
	RPPrinter.CustomColorPrintLn(printer.ColorPurple, localLocalBldLogsTag2)

	return err
}

// packageMetadata allow to set app metadata value using `iotpkg` inside
// the local-builder
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func packageMetadata(cmd *cobra.Command, args []string) error {

	// Before anything, handle the case where the user is asking for help
	helpAsked, err := handleIotpkgSubCmdHelp(cmd, originalArgs, "metadata")
	if err != nil {
		return err
	}
	if helpAsked {
		return nil
	}

	RPPrinter.CustomPrintLn(localMetadataCmdAnsLine0)

	// Parse rp-cli flags and extract iotpkg args ---------
	iotpkgArgs, err := getIotpkgArgs(cmd)
	if err != nil {
		return err
	}
	Log.Debugf("[LOCAL] Iotpkg args: %+v\n", iotpkgArgs)
	// ----------------------------------------------------

	// Get full context -----------------------------------
	appStructure, _, _, err := getLocalContext(cmd, cmd.Flags().Args(), true, true, false)
	if err != nil {
		return err
	}
	// ----------------------------------------------------

	RPPrinter.CustomColorPrintLn(printer.ColorPurple, localLocalBldLogsTag1)
	RPPrinter.ForceFlush()

	// Get the size of the current terminal
	termWidth, termHeight, err := term.GetSize(int(os.Stdin.Fd()))
	if err != nil {
		return err
	}

	metadataRequest := ladapi.RPLADIotpkgCmdStruct{
		RPLADMainCmdParams: ladapi.RPLADMainCmdParams{
			RPLADWinSize: ladapi.RPLADWinSize{
				WinSizeRows: termHeight,
				WinSizeCol:  termWidth,
			},
			WorkingDir: path.Join(getLclBldGitpkgsDir(), appStructure.Parent.ProjectSlug, appStructure.PackageName),
		},
		IotpkgFlags: []string{"-v"},
		SubcmdFlags: iotpkgArgs,
	}

	procTicket := ladapi.RPLADRunnerTicket{}
	err = localBAPI.IotpkgMetadataPost(&metadataRequest, &procTicket)
	if err != nil {
		return err
	}

	err = listenStdoutErrAndIn(procTicket)
	RPPrinter.CustomColorPrintLn(printer.ColorPurple, localLocalBldLogsTag2)

	return err
}

// localCheckSetup check that everything needed to local build
// is correctly set-up on the host
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func localCheckSetup(cmd *cobra.Command, args []string) error {
	RPPrinter.CustomPrintLn(localCheckSetupCmdAnsLine0)

	defer RPPrinter.ForceFlush()

	// Check factory connection ---------------------------
	RPPrinter.CustomPrint(localCheckSetupCmdAnsLine1)
	cfg := rpapiv1.RPAPIConfig{}
	if err := bckAPI.RPConfigGet(&cfg); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
	// ----------------------------------------------------

	// Check local-builder connection ---------------------
	RPPrinter.CustomPrint(localCheckSetupCmdAnsLine2)
	_, err := localBAPI.GetVersion()
	if err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
	// ----------------------------------------------------

	// Check mount-points ---------------------------------
	RPPrinter.CustomPrint(localCheckSetupCmdAnsLine3)
	err = checkLocalBuilderMountPoints()
	if err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
	// ----------------------------------------------------

	// Check QEMU set-up ----------------------------------
	RPPrinter.CustomPrint(localCheckSetupCmdAnsLine4)
	err = checkQemuSetUp()
	if err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
	// ----------------------------------------------------

	return nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Utilities functions ------------------------------------------
// --------------------------------------------------------------

// getCustomBuildsDirs is the function that looks for the different builds directories inside
// viper environment but also in the connection parameters if the user set-up a custom path.
// The priority is given to the custom value
//
// :param flagName: Flag to look for
// :type flagName: string
// :returns: string, error -- path to the directory and an error if something went wrong
func getCustomBuildsDirs(flagName string) (string, error) {
	// First thing to do, check if the flag is defined in the connection file for the current server alias
	// Get the alias of the remote connection
	serverAlias := viper.GetString(serverAliasFlag)

	// Get the connection parameters
	connecParams, err := getConnectionParams(serverAlias, accessTokenFile)
	if err != nil {
		return "", err
	}

	// Look for the right field in the connection
	strToConsider := ""
	if flagName == localPkgCmdSourceFlag {
		strToConsider = connecParams.SourcesDir
	} else if flagName == localPkgCmdPackageFlag {
		strToConsider = connecParams.PkgsDir
	}

	// If we still have an empty string, let's get the default value
	if strToConsider == "" {
		strToConsider = viper.GetString(flagName)
	}

	// Sanitize the path and return
	return utils.SanitizePath(strToConsider)
}

// getAndChecksBuildDirs get the gitsources/ and gitpkgs/
// directories on the host sides. It checks them as well
//
// :param dirSuffix: Additional dir to add at the end of directory to check (if not empty)
// :type dirSuffix: string
// :returns: string, string, error -- sources dir, pkgs dir and an error if something went wrong
func getAndChecksBuildDirs(dirSuffix string) (string, string, error) {
	// Try to get the sources directory from everywhere we can
	tmpSrcDir, err := getCustomBuildsDirs(localPkgCmdSourceFlag)
	if err != nil {
		return "", "", err
	}

	// Try to get the packages directory from Viper
	tmpPkgsDir, err := getCustomBuildsDirs(localPkgCmdPackageFlag)
	if err != nil {
		return "", "", err
	}

	// Add suffix if set in parameter
	checkSourceDir := true
	if dirSuffix != "" {
		tmpSrcDir = path.Join(tmpSrcDir, dirSuffix)
		tmpPkgsDir = path.Join(tmpPkgsDir, dirSuffix)
		// If a suffix has been set, we are at running time!
		// No need to check for source directory (it is possible that it is not existing)
		checkSourceDir = false
	}

	// SOURCES DIRECTORY --------------
	var SrcDir string

	// Sanitize path (~ and ./, ../, etc.)
	SrcDir, err = utils.SanitizePath(tmpSrcDir)
	if err != nil {
		return "", "", err
	}

	// Check that the resulting directory exists
	if _, err := os.Stat(SrcDir); checkSourceDir && os.IsNotExist(err) {
		return "", "", err
	}
	// --------------------------------

	// PACKAGES DIRECTORY -------------
	var PkgsDir string

	// Sanitize path (~ and ./, ../, etc.)
	PkgsDir, err = utils.SanitizePath(tmpPkgsDir)
	if err != nil {
		return "", "", err
	}

	// Check that the resulting directory exists
	if _, err := os.Stat(PkgsDir); os.IsNotExist(err) {
		return "", "", err
	}
	// --------------------------------

	return SrcDir, PkgsDir, nil
}

// findPathsToDelete tries to look into the local packaging directories in order to find
// one that match the application slug and optionally the project slug given in arguments
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: string, string, error -- sources dir, pkgs dir and an error if something went wrong
func findPathsToDelete(cmd *cobra.Command, args []string) (string, string, error) {
	// Try to get the sources directory from everywhere we can
	tmpSrcDir, err := getCustomBuildsDirs(localPkgCmdSourceFlag)
	if err != nil {
		return "", "", err
	}

	// Try to get the packages directory from Viper
	tmpPkgsDir, err := getCustomBuildsDirs(localPkgCmdPackageFlag)
	if err != nil {
		return "", "", err
	}

	// From command line get the application details
	appStructure, err := getLocalApp(cmd, args, true)
	if err != nil {
		return "", "", err
	}

	// Now construct the paths to the right directories
	srcDirectory := path.Join(tmpSrcDir, appStructure.Parent.ProjectSlug, appStructure.PackageName)
	pkgDirectory := path.Join(tmpPkgsDir, appStructure.Parent.ProjectSlug, appStructure.PackageName)

	// Check that the source directory exists
	_, err = os.Stat(srcDirectory)
	if os.IsNotExist(err) {
		Log.Warnf(localCleanNoSourcesDirWarn, appStructure.Slug)
		srcDirectory = ""
	}

	// Check that the gitpkg directory exists
	_, err = os.Stat(pkgDirectory)
	if os.IsNotExist(err) {
		return "", "", fmt.Errorf(packageCleanDirNotFound, pkgDirectory)
	}

	return srcDirectory, pkgDirectory, nil
}

// listenStdoutErrAndIn allows to listen simultaneously stdin from the
// user and stdout/err from the remote running process.
// Basically, this function is really handly to call afetr sending a
// command to run on the remote server. This command could be interactive
// or not.
//
// :param procTicket: The ticket received from the server and identifying the
// remote process
// :type procTicket: ladapi.RPLADRunnerTicket
// :returns: error -- an error if something went wrong, nil in the other case
func listenStdoutErrAndIn(procTicket ladapi.RPLADRunnerTicket) error {

	// Setup terminal (raw mode to handle escape and control keys)
	if (!term.IsTerminal(int(os.Stdin.Fd())) || !term.IsTerminal(int(os.Stdout.Fd()))) &&
		(runMode != "testing") {
		return errors.New(stdShouldBeTerminal)
	}
	oldState, err := term.MakeRaw(int(os.Stdin.Fd()))
	if err != nil {
		return err
	}
	defer term.Restore(int(os.Stdin.Fd()), oldState)

	// Start the process listening on stdin
	exitChannel := liveListenStdin(procTicket.ID, oldState)

	// Handle signals
	err = signals.OnSignals(func(sig os.Signal) {
		Log.Debugf("Send signal %v", sig)

		if signals.IsWinResizeSignal(sig) {
			// Request Redpesk LAD to resize the terminal size
			terminalResizeWin(procTicket.ID)

		} else if signals.IsInterruptSignal(sig) {
			stdinEvt := RPLocalBStdinEvent{
				ID:    procTicket.ID,
				Chars: "\x04\n",
			}
			localBConnection.WSocket.EmitEvent(localStdInEvt, &stdinEvt)

		} else {
			// Send the signal to the server (and the process)
			terminalSendSignal(procTicket.ID, sig)
		}
	})
	if err != nil {
		return err
	}

	// Block until we receive all the websocket events (until localProcEndEvt)
	procRunning := true
	var errToReturn error
	for procRunning {
		select {
		// We received an event from the remote server
		case localEvent := <-cmdEventLocalBChannel:
			// Check that the received event is for our process
			if localEvent.ID == procTicket.ID {
				switch localEvent.Type {
				case localStdOutEvt:
					// Print to the os.Stdout if the line is not empty
					if len(localEvent.Line) > 0 {
						os.Stdout.Write([]byte(localEvent.Line))
					}
				case localStdErrEvt:
					// Print to the os.Stderr if the line is not empty
					if len(localEvent.Line) > 0 {
						os.Stderr.Write([]byte(localEvent.Line))
					}
				case localProcEndEvt:
					// This is the end of the terminal
					errorMsg := fmt.Sprintf(endOfProc, localEvent.ID, localEvent.Status)
					Log.Debugln(errorMsg)

					// End this blocking code
					procRunning = false

					// Return an error if the terminal end badly
					if localEvent.Status != 0 {
						errToReturn = errors.New(errorMsg)
					} else {
						errToReturn = nil
					}
				}
			}

		// End of terminal (high chance for error, see liveListenStdin())
		case res := <-exitChannel:
			// End this blocking code
			procRunning = false
			errToReturn = nil

			if res.Error != nil {
				Log.Errorf(execErrorMsg, res.Error)
				errToReturn = res.Error
			}

		// Disconnection from the websocket, since everything go through the
		// socket, stop the command
		case <-localBConnection.WSocket.DisconnectionChan:
			// End this blocking code
			procRunning = false
			Log.Error(webSocketDisconnectionErr)
			errToReturn = errors.New(webSocketDisconnectionErr)
		}
	}

	return errToReturn
}

// liveListenStdin is a function to call as a coroutine in order to
// listen on the stdin of the terminal. It is used to send stdin event
// to the local-builder server.
//
// :param id: ID of the process to send event to
// :type id: string
// :param oldState: Previous state of the terminal
// :type oldState: *term.State
// :returns: chan RPLocalExitResult -- the channel to use to stop the process
func liveListenStdin(id string, oldState *term.State) chan RPLocalExitResult {
	exitChannel := make(chan RPLocalExitResult)

	go func(id string, oldState *term.State) {
		// Declaration of exposeFd type
		type exposeFd interface {
			Fd() uintptr
		}

		buff := make([]byte, 128)
		rdfs := &goselect.FDSet{}
		reader := io.ReadCloser(os.Stdin)
		defer reader.Close()

		for {
			rdfs.Zero()
			rdfs.Set(reader.(exposeFd).Fd())
			err := goselect.Select(int(os.Stdin.Fd())+1, rdfs, nil, nil, 50*time.Millisecond)
			if err != nil {
				if err == syscall.EINTR {
					continue
				} else {
					term.Restore(int(os.Stdin.Fd()), oldState)
					exitChannel <- RPLocalExitResult{err, 3}
					return
				}
			}

			// Protection to avoid out of range error seen during tests
			if (runMode == "testing") && (reader.(exposeFd).Fd() > goselect.NFDBITS*16) {
				term.Restore(int(os.Stdin.Fd()), oldState)
				exitChannel <- RPLocalExitResult{err, 7}
				return
			}

			if rdfs.IsSet(reader.(exposeFd).Fd()) {
				size, err := reader.Read(buff)

				if err != nil {

					Log.Debugf("Read error %v; err %v", size, err)
					if err == io.EOF {

						// CTRL-D exited scanner, so send it explicitly
						stdinEvt := RPLocalBStdinEvent{
							ID:    id,
							Chars: "\x04\n",
						}
						err := localBConnection.WSocket.EmitEvent(localStdInEvt, &stdinEvt)

						// If a problem occurs, restore the old terminal
						if err != nil {
							term.Restore(int(os.Stdin.Fd()), oldState)
							exitChannel <- RPLocalExitResult{err, 4}
							return
						}

						time.Sleep(time.Millisecond * 100)
						continue

					} else {
						term.Restore(int(os.Stdin.Fd()), oldState)
						exitChannel <- RPLocalExitResult{err, 5}
						return
					}
				}

				if size <= 0 {
					continue
				}

				// Nominal way: send the input to the server
				data := buff[:size]
				stdinEvt := RPLocalBStdinEvent{
					ID:    id,
					Chars: string(data),
				}
				err = localBConnection.WSocket.EmitEvent(localStdInEvt, &stdinEvt)

				// If a problem occurs, restore the old terminal
				if err != nil {
					term.Restore(int(os.Stdin.Fd()), oldState)
					exitChannel <- RPLocalExitResult{err, 6}
					return
				}
			}
		}

	}(id, oldState)

	return exitChannel
}

// getLclBldGitpkgsDir returns the current value for the local-builder
// 'gitpkgs/' directory. It takes into account the root directory to be used
// inside the local-builder (indeed, this can be customized now)
//
// :returns: string -- the path to the 'gitpkgs' directory
func getLclBldGitpkgsDir() string {
	if localBuilderRootDir == "" {
		Log.Errorln("The local-builder root directory does not seem to have been initialized!")
	}
	return path.Join(localBuilderRootDir, "gitpkgs")
}

// getLclBldGitsourcesDir returns the current value for the local-builder
// 'gitsources/' directory. It takes into account the root directory to be used
// inside the local-builder (indeed, this can be customized now)
//
// :returns: string -- the path to the 'gitsources' directory
func getLclBldGitsourcesDir() string {
	if localBuilderRootDir == "" {
		Log.Errorln("The local-builder root directory does not seem to have been initialized!")
	}
	return path.Join(localBuilderRootDir, "gitsources")
}

// createLocalBuilderWorkingDirs is the function that manage the creation of the directories
// used to build inside the local-builder
//
// :param rootSrcDir: Root directory containing the source directories
// :type rootSrcDir: string
// :param rootGitpkgDir: Root directory containing the gitpkg directories
// :type rootGitpkgDir: string
// :param prjSlug: Project slug
// :type prjSlug: string
// :param appPkgName: Application package name
// :type appPkgName: string
// :returns: error -- an error if something went wrong
func createLocalBuilderWorkingDirs(rootSrcDir, rootGitpkgDir, prjSlug, appPkgName string) error {

	// Create the path to the directories that are interesting us ------------------
	// Project directories
	ourSourcesPrjDir := path.Join(rootSrcDir, prjSlug)
	ourPackagesPrjDir := path.Join(rootGitpkgDir, prjSlug)
	// App directories
	ourSourcesDir := path.Join(ourSourcesPrjDir, appPkgName)
	ourPackagesDir := path.Join(ourPackagesPrjDir, appPkgName)
	// rp-cli configuration directories
	ourCfgPkgDir := path.Join(ourPackagesDir, localCfgSubDirectory)
	// -----------------------------------------------------------------------------

	// Create the project level directories if there are not existing
	err := utils.CreateDirIfNotExists(ourSourcesPrjDir, false)
	if err != nil {
		return err
	}
	err = utils.CreateDirIfNotExists(ourPackagesPrjDir, false)
	if err != nil {
		return err
	}

	// Now create the app slug sub-directories
	err = utils.CreateDirIfNotExists(ourSourcesDir, false)
	if err != nil {
		return err
	}
	err = utils.CreateDirIfNotExists(ourPackagesDir, false)
	if err != nil {
		return err
	}

	// Finally the configuration sub-directory
	err = utils.CreateDirIfNotExists(ourCfgPkgDir, false)
	if err != nil {
		return err
	}

	return nil
}

// getCmdLine get the command line furnished by the user. If no
// command is passed or if it is do not respect the format, it
// returns an error
//
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: string, error -- the command line and an error if something went wrong
func getCmdLine(args []string) (string, error) {
	// Initialize the result
	resultCmd := ""

	// Check that the command is at least 1 args long ('--' is not counted)
	if len(args) < 1 {
		return resultCmd, errors.New(cmdToShortError)
	}

	return strings.Join(args, " "), nil
}

// checkGitpkgsStatus is the function that check if some modification have been done on the
// gitpkgs files
//
// :param fullGitpkgDir: Full path to the directory containing the files to check
// :type fullGitpkgDir: string
// :param displayRes: Ask to display details of difference
// :type displayRes: bool
// :param quiet: If true, the function does not return an error if the md5sum file does not exist
// :type quiet: bool
// :returns: bool, RPLocalGitpkgsModification, error -- true if the status is OK, false in the other case, the details of modification and an error if something went wrong
func checkGitpkgsStatus(fullGitpkgDir string, displayRes, quiet bool) (bool, RPLocalGitpkgsModification, error) {
	status := true
	statusDetails := RPLocalGitpkgsModification{
		Added:    []string{},
		Modified: []string{},
		Deleted:  []string{},
	}

	// Firstly check if the md5sum file exist
	md5sumFile := path.Join(fullGitpkgDir, localCfgSubDirectory, localMD5SUMFile)

	// Return an error if the md5sum file does not exist
	_, err := os.Stat(md5sumFile)
	if os.IsNotExist(err) && !quiet {
		return false, statusDetails, errors.New(md5sumFileDoesNotExist)
	} else if os.IsNotExist(err) && quiet {
		// Return true if the function is quiet and that the md5sum does not exist
		return true, statusDetails, nil
	}

	// Read the JSON file
	md5sumMap := make(map[string]string)
	err = utils.ReadJsonFile(md5sumFile, &md5sumMap)
	if err != nil {
		return false, statusDetails, err
	}

	// Get all the file currently in the gitpkg "repo"
	currentLocalFiles, err := getLocalGitpkgsFileOrDirsList(fullGitpkgDir, true)
	if err != nil {
		return false, statusDetails, err
	}
	filesList := filterNonGitpkgsFiles(currentLocalFiles["files"], fullGitpkgDir)

	// Now for each file, calculate the md5sum and compare with what we got in the file
	for filename, filemd5sum := range md5sumMap {

		// Create the path for the current file
		currFilePath := path.Join(fullGitpkgDir, filename)

		// Delete the file from our "current local" list
		utils.DeleteStringFromSlice(&filesList, currFilePath)

		// Check that the file still exist
		_, err = os.Stat(currFilePath)
		if os.IsNotExist(err) {
			// The file does not exist anymore
			status = false
			statusDetails.Deleted = append(statusDetails.Deleted, filename)
			if displayRes {
				RPPrinter.CustomColorPrintF(printer.ColorRed, localStatusDelFileLine, filename)
			}
			continue
		}

		// If the file exists, check the md5sum
		localMd5sum, err := utils.CalculateMd5SumOnFile(currFilePath, true)
		if err != nil {
			return false, statusDetails, err
		}
		if localMd5sum != filemd5sum {
			// The file has been modified
			status = false
			statusDetails.Modified = append(statusDetails.Modified, filename)
			if displayRes {
				RPPrinter.CustomColorPrintF(printer.ColorRed, localStatusModifiedFileLine, filename)
			}
		}
	}

	// Create the directory prefix to delete from file path
	dirPrefix := fullGitpkgDir
	if !strings.HasSuffix(dirPrefix, "/") {
		dirPrefix += "/"
	}

	// Now add every file that stayed in our "current local" list, as an "added" file
	for _, tmpAddFile := range filesList {
		status = false
		// Change the file path to convert to the local gitpkg path
		addFile := strings.ReplaceAll(tmpAddFile, dirPrefix, "")
		// Add it to the "added" files list
		statusDetails.Added = append(statusDetails.Added, addFile)
		if displayRes {
			RPPrinter.CustomColorPrintF(printer.ColorRed, localStatusAddedFileLine, addFile)
		}
	}

	if status && displayRes {
		RPPrinter.CustomColorPrintLn(printer.ColorGreen, localStatusNoChanges)
	}

	return status, statusDetails, nil
}

// getLocalGitpkgsFileOrDirsList retrieve the list of all the files/dirs currently present in the gitpkg directory given.
// It filters out the files that can be found in the gitpkg directory but that are not part of gitpkg (like ./rp-cli content)
//
// :param fullGitpkgDir: Full path to the directory containing the gitpkg files
// :type fullGitpkgDir: string
// :param filesOnly: Tell the function to not list the directories
// :type filesOnly: bool
// :returns: map[string][]string error -- lists of files and subdirectories and an error if something went wrong
func getLocalGitpkgsFileOrDirsList(fullGitpkgDir string, filesOnly bool) (map[string][]string, error) {
	// Get the list of all files in the gitpkg directory
	tmpList, err := utils.FindAllFilesAndDirsRecursively(fullGitpkgDir)
	if err != nil {
		return tmpList, err
	}

	// Now filter out the .rp-cli/ lines
	actualMap := make(map[string][]string)
	if !filesOnly {
		actualMap["dirs"] = []string{}
		for _, tmpDirName := range tmpList["dirs"] {
			if !strings.Contains(tmpDirName, localCfgSubDirectory) && (tmpDirName != fullGitpkgDir) {
				actualMap["dirs"] = append(actualMap["dirs"], tmpDirName)
			}
		}
	}
	actualMap["files"] = []string{}
	for _, tmpFileName := range tmpList["files"] {
		if !strings.Contains(tmpFileName, localCfgSubDirectory) {
			actualMap["files"] = append(actualMap["files"], tmpFileName)
		}
	}

	return actualMap, nil
}

// getLocalBuildConfig retrieves the local-builder configuration in the specified path.
// The function adds the .rp-cli/ suffix if it is not included.
// Additionally, it can output an error on the "local_build_config.json" file not existing if asked
//
// :param packageDir:Path to the directory containing the configuration file (with or without '.rp-cli')
// :type packageDir: string
// :param errIfNotExist: If set, the function returns an error when "local_build_config.json" does not exist
// :type errIfNotExist: bool
// :returns: RPLocalBuildConfig, string error -- the local-builder configuration, the path to the config file and an error if something went wrong
func getLocalBuildConfig(packageDir string, errIfNotExist bool) (RPLocalBuildConfig, string, error) {
	// Create the return value
	localConfig := RPLocalBuildConfig{}

	// Work on the package directory given
	fullCfgPath := packageDir
	if !strings.Contains(".rp-cli", path.Base(packageDir)) {
		fullCfgPath = path.Join(packageDir, localCfgSubDirectory)
	}

	// Create the config file path
	localBldConfFile := path.Join(fullCfgPath, localBldCfgFile)

	// Check if file exists
	_, err := os.Stat(localBldConfFile)
	if err != nil && !os.IsNotExist(err) {
		return localConfig, localBldConfFile, err
	} else if err == nil {
		Log.Debugln("[FETCH] Retrieving local config from:", localBldConfFile)
		err = utils.ReadJsonFile(localBldConfFile, &localConfig)
		if err != nil {
			return localConfig, localBldConfFile, err
		}
	} else if os.IsNotExist(err) {
		Log.Debugln("[FETCH] 'local_build_config.json' does not exist, nothing to retrieve")
		if errIfNotExist {
			return localConfig, localBldConfFile, err
		}
	}
	return localConfig, localBldConfFile, nil
}

// updateOrCreateMD5SumFile is the function that takes as input the gitpkg directory and creates or update a
// json file containing all the md5sum of the files in the .rp-cli/ subdirectory
//
// :param fullGitpkgDir: Full path to the directory containing the files to check
// :type fullGitpkgDir: string
// :param fileNamesToUpdate: List of files to update in the file (if empty, the file is just re-created)
// :type fileNamesToUpdate: []string
// :returns: error -- an error if something went wrong
func updateOrCreateMD5SumFile(fullGitpkgDir string, fileNamesToUpdate []string) error {

	// Create the object that will store the previous values if necessary
	oldMd5sumMap := make(map[string]string)
	// Path to the md5sum file
	md5sumFile := path.Join(fullGitpkgDir, localCfgSubDirectory, localMD5SUMFile)

	// Check if the file exists (only if the user gave some names, in the other case -> full update)
	if _, err := os.Stat(md5sumFile); err == nil && len(fileNamesToUpdate) > 0 {
		// If the file exists, get the params already existing inside
		err := utils.ReadJsonFile(md5sumFile, &oldMd5sumMap)
		if err != nil {
			return err
		}
	}

	// Get the list of files to check
	objectsInDir, err := getLocalGitpkgsFileOrDirsList(fullGitpkgDir, true)
	if err != nil {
		return err
	}
	// Filters the files that we don't want
	filesOfInterest := filterNonGitpkgsFiles(objectsInDir["files"], fullGitpkgDir)

	// Create the directory prefix
	dirPrefix := fullGitpkgDir
	if !strings.HasSuffix(dirPrefix, "/") {
		dirPrefix += "/"
	}

	// Rework the filenames to update to add the directory prefix
	fullFileNamesToUpdate := []string{}
	for _, fileToUpdate := range fileNamesToUpdate {
		fullFileName := path.Join(dirPrefix, fileToUpdate)
		fullFileNamesToUpdate = append(fullFileNamesToUpdate, fullFileName)

		// Check if the filename is in the files list found on the disk
		// If it is not the case, this means that the modification on the file is a deletion
		if !utils.ContainsString(filesOfInterest, fullFileName) {
			// Delete from oldMd5sumMap
			delete(oldMd5sumMap, fileToUpdate)
		}
	}

	// Iterate over files to calculate their md5sum
	for _, file := range filesOfInterest {
		// Filter the files to update if the user gave some
		if utils.ContainsString(fullFileNamesToUpdate, file) || (len(fullFileNamesToUpdate) == 0) {
			tmpMd5sum, err := utils.CalculateMd5SumOnFile(file, true)
			if err != nil {
				return err
			}
			// Get rid of the directory prefix and save the md5sum in the map
			tmpFileName := strings.ReplaceAll(file, dirPrefix, "")
			// Save in map
			oldMd5sumMap[tmpFileName] = tmpMd5sum
		}
	}

	// Write in md5sum file
	return utils.WriteJsonFile(md5sumFile, oldMd5sumMap)
}

// localBuilderSetUp fills the RPLocalBuildConfig structure with the values left
// to set-up fully the local-builder to build
//
// :param app: Structure containing all needed data for application
// :type app: *rpapiv1.RPApplicationFull
// :param config: Configuration to set for the local builder
// :type config: *RPLocalBuildConfig
// :param overwrite: Overwrite any existing directory
// :type overwrite: bool
// :returns: error -- an error if something went wrong
func localBuilderSetUp(app *rpapiv1.RPApplicationFull, localConfig *RPLocalBuildConfig, overwrite bool) error {

	// --------------------------------------------------------------
	// -----             '<app_gitpkgs>/' directory             -----
	// --------------------------------------------------------------
	err := fetchGitpkgAndMockFiles(localConfig, app, true, true, overwrite)
	if err != nil {
		return err
	}

	// --------------------------------------------------------------
	// -----           '<app_gitsources>/' directory            -----
	// --------------------------------------------------------------
	// Clone the sources inside gitsources/
	err = cloneSourcesRepo(app, localConfig, overwrite)
	if err != nil {
		return err
	}

	return nil
}

// getLocalApp parses the user command line and retrieve the
// application UUID and slug. It also permits to check that the corresponding
// project is created and if the application exists in Redpesk
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param offline: Ask the function to stay offline: i.e. no request to the redpesk factory or internet
// :type offline: bool
// :returns: rpapiv1.RPApplicationFull, error -- the application and an error if something went wrong
func getLocalApp(cmd *cobra.Command, args []string, offline bool) (rpapiv1.RPApplicationFull, error) {
	var err error
	appSlug, prjSlug := "", ""
	argsToPass := append([]string{}, args...)

	// Check if the args contains an application slug (If a slug is given, do nothing)
	if len(args) < 1 {
		// If a slug is not given, try to get it from the current path
		appSlug, prjSlug, err = getAppAndPrjSlugFromLocalBldConfig()
		if err != nil {
			return rpapiv1.RPApplicationFull{}, err
		}

		// Check that we've got something
		if appSlug == "" || prjSlug == "" {
			return rpapiv1.RPApplicationFull{}, errors.New(noAppSpecifiedAndNotInDir)
		}

		// Rework the args to pass
		argsToPass = append([]string{appSlug}, argsToPass...)

		// Add the project slug with the right flag
		cmd.Flags().Set(appsCmdPrjFlag, prjSlug)
	}

	if offline {
		// Retrieve all the local-build configs available
		allLocalCfgs, err := retrieveAllLocalBldConfigs()
		if err != nil {
			return rpapiv1.RPApplicationFull{}, err
		}

		if appSlug == "" {
			// Retrieve appSlug from command line
			appSlug, err = getFirstArg(args, fmt.Sprintf(objectIDNotSetMsg, "application"), fmt.Sprintf(objectIDRequiredMsg, "application"))
			if err != nil {
				return rpapiv1.RPApplicationFull{}, err
			}
		}

		// Try to find the right config with only the appSlug
		var cfgResult RPLocalBuildConfig
		goodKeys := []string{}
		for key := range allLocalCfgs {
			if strings.HasSuffix(key, "/"+appSlug) {
				goodKeys = append(goodKeys, key)
			}
		}
		// Check that we have only one key
		if len(goodKeys) == 1 {
			cfgResult = allLocalCfgs[goodKeys[0]]
		} else if len(goodKeys) < 1 {
			return rpapiv1.RPApplicationFull{}, fmt.Errorf(appIdentifierNotFoundLocally, appSlug)
		} else {
			// Retrieve the project flag to have more info
			prjSlug, err = cmd.Flags().GetString(appsCmdPrjFlag)
			if err != nil {
				logFlagError(appsCmdPrjFlag)
				return rpapiv1.RPApplicationFull{}, err
			}

			// Try to get the right configuration
			var ok bool
			if cfgResult, ok = allLocalCfgs[prjSlug+"/"+appSlug]; !ok {
				return rpapiv1.RPApplicationFull{}, errors.New(noLocalCfgForThisAppAndPrj)
			}
		}

		// Here, we have the right configuration, so now creates the application from it
		resultApp := rpapiv1.RPApplicationFull{
			RPApplication: rpapiv1.RPApplication{
				Slug:        cfgResult.AppSlug,
				PackageName: cfgResult.AppPkgName,
				Parent: rpapiv1.RPAppParent{
					ProjectSlug: cfgResult.ProjectSlug,
				},
			},
		}
		return resultApp, nil
	} else {
		// Check if the project and the application exists, and get it
		return genericGetAppFromArgs(cmd, argsToPass, true)
	}
}

// createRepoInheritanceFile retrieves the repo inheritance file from
// the backend and write it in the given target directory
//
// :param appParent: Parent of the application that the user wants to build
// :type appParent: rpapiv1.RPAppParent
// :param targetDir: Directory where the caller wants to create the file
// :type targetDir: string
// :returns: error -- an error if a problem occurred
func createRepoInheritanceFile(appParent rpapiv1.RPAppParent, targetDir string) error {
	var err error

	repoInheritanceFile := path.Join(targetDir, (appParent.ProjectSlug + ".json"))

	// Get the connection parameters
	serverAlias := viper.GetString(serverAliasFlag)
	connecParams, err := getConnectionParams(serverAlias, accessTokenFile)
	if err != nil {
		return err
	}

	return bckAPI.ProjectGetReposInheritance(appParent.ProjectID.String(), repoInheritanceFile, connecParams.ServerURL)
}

// getLocalContext retrieves the structures necessary for several local subcommands
// These include:
//   - The full application structure
//   - Path to the main source directory (or full path if getFullDir is True)
//   - Path to the main gitpkg directory (or full path if getFullDir is True)
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param getFullDir: Ask the function to get and check the full directories path and not just the root
// :type getFullDir: bool
// :param offline: Ask the function to stay offline: i.e. no request to the redpesk factory or internet
// :type offline: bool
// :returns: rpapiv1.RPApplicationFull, string, string, error -- app, rootSrcDir, rootGitpkgDir and an error if a problem occurred
func getLocalContext(cmd *cobra.Command, args []string, getFullDir, offline, quiet bool) (rpapiv1.RPApplicationFull, string, string, error) {
	// Get applications details ---------------------------
	if !quiet {
		// Display info
		RPPrinter.CustomPrint(localFetchCmdAnsLine1)
	}
	// Now, retrieve the app ID
	appStructure, err := getLocalApp(cmd, args, offline)
	if err != nil {
		if !quiet {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		}
		return rpapiv1.RPApplicationFull{}, "", "", err
	}
	if !quiet {
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
	}
	// ----------------------------------------------------

	// Fill a directory suffix if the caller wants to check the full directory
	dirSuffix := ""
	if getFullDir {
		dirSuffix = path.Join(appStructure.Parent.ProjectSlug, appStructure.PackageName)
	}

	// Check the directories to use for the local-builder -
	if !quiet {
		// Display info
		RPPrinter.CustomPrint(localFetchCmdAnsLine2)
	}
	// Check that the local builder directories exist
	rootSrcDir, rootGitpkgDir, err := getAndChecksBuildDirs(dirSuffix)
	if err != nil {
		if !quiet {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		}
		return rpapiv1.RPApplicationFull{}, "", "", err
	}
	if !quiet {
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
	}
	// ----------------------------------------------------

	return appStructure, rootSrcDir, rootGitpkgDir, nil
}

// getLocalBuildContext retrieves mandatory info to start the build locally:
//   - The target architecture
//   - The target distribution
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param app: Pointer to the full structure of the app to build
// :type app: *rpapiv1.RPApplicationFull
// :returns: string, string, error -- architecture, distribution and an error if a problem occurred
func getLocalBuildContext(cmd *cobra.Command, app *rpapiv1.RPApplicationFull) (string, string, error) {
	archResult, distroResult := "", ""

	// Firstly, get the values given by the user
	// Architecture
	userArch, err := cmd.Flags().GetString(localBuildArchFlag)
	if err != nil {
		logFlagError(localBuildArchFlag)
		return archResult, distroResult, err
	}
	// Distribution
	userDistribution, err := cmd.Flags().GetString(localBuildDistroFlag)
	if err != nil {
		logFlagError(localBuildArchFlag)
		return archResult, distroResult, err
	}

	// Retrieve the local configuration directly from appSlug and projectSlug
	localConfig, err := retrieveOneLocalBldConfig(app.Parent.ProjectSlug, app.PackageName)
	if err != nil {
		return archResult, distroResult, err
	}

	// Check that the distribution given is available for this application
	// 1st case: no distro specified by the user and more than one distro available
	if (userDistribution == "") && len(localConfig.DistrosList) > 1 {
		errMsg := fmt.Sprintf(needToSpecifyDistroErr, app.Parent.ProjectSlug)
		errMsg += " (supported distributions: " + strings.Join(localConfig.DistrosList, ", ") + ")"
		return archResult, distroResult, errors.New(errMsg)
	}

	// 2nd case: distro specified by the user, but not available
	if (userDistribution != "") && !utils.ContainsString(localConfig.DistrosList, userDistribution) {
		errMsg := fmt.Sprintf(needToChangeDistroErr, userDistribution, app.Parent.ProjectSlug)
		errMsg += " (supported distributions: " + strings.Join(localConfig.DistrosList, ", ") + ")"
		return archResult, distroResult, errors.New(errMsg)
	}

	// 3rd case: no distribution specified, but only 1 distro available
	if (userDistribution == "") && len(localConfig.DistrosList) == 1 {
		distroResult = localConfig.DistrosList[0]
	} else {
		// 4th case: distribution specified, and in available distros
		distroResult = userDistribution
	}

	// Check that the architecture given is available for this project
	// Check if the architecture is available for this project
	if !utils.ContainsString(localConfig.ArchsList, userArch) {
		errMsg := fmt.Sprintf(needToChangeArchErr, userArch, app.Parent.ProjectSlug)
		errMsg += " (supported architectures: " + strings.Join(localConfig.ArchsList, ", ") + ")"
		return archResult, distroResult, errors.New(errMsg)
	} else {
		archResult = userArch
	}

	return archResult, distroResult, nil
}

// fetchGitpkgAndMockFiles can fetch gitpkgs files and the mock configuration files from the redpesk factory,
// and to put those files at the right location
//
// :param localConfig: Configuration to use for the local builder
// :type localConfig: *RPLocalBuildConfig
// :param appStructure: Pointer to the application structure
// :type appStructure: *rpapiv1.RPApplicationFull
// :param getGitPkgFiles: Boolean saying that the pkg files need to be retrieved
// :type getGitPkgFiles: bool
// :param getMockFiles: Boolean saying that the mock files need to be retrieved
// :type getMockFiles: bool
// :param force: Ask the function to overwrite the gitpkg directory anyway
// :type force: bool
// :returns: error -- an error if a problem occurred
func fetchGitpkgAndMockFiles(localConfig *RPLocalBuildConfig, appStructure *rpapiv1.RPApplicationFull, getGitPkgFiles, getMockFiles, force bool) error {

	// Create local directories for the app if necessary --
	// Display info
	RPPrinter.CustomPrint(localFetchCmdAnsLine3)
	// Create the needed repositories if necessary
	err := createLocalBuilderWorkingDirs(localConfig.SourcesDir, localConfig.PackagesDir, appStructure.Parent.ProjectSlug, appStructure.PackageName)
	if err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
	// ----------------------------------------------------

	// Check gitpkg status for the wanted repo ------------
	fullGitpkgDir := path.Join(localConfig.PackagesDir, appStructure.Parent.ProjectSlug, appStructure.PackageName)
	if !force && getGitPkgFiles {
		// Display info
		RPPrinter.CustomPrint(localFetchCmdAnsLine4)
		// If the user does not wan't to force the fetch, check if the md5sum has changed (if some files were already present)
		filesStatus, _, err := checkGitpkgsStatus(fullGitpkgDir, false, true)
		if err != nil {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
			return err
		}
		if !filesStatus {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
			RPPrinter.ForceFlush()
			return errors.New(cannotFetchWithChanges)
		}
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
	}
	// ----------------------------------------------------

	// Create the needed variables
	ourCfgPkgDir := path.Join(fullGitpkgDir, localCfgSubDirectory)

	// Fetch the 'gitpkg' files ---------------------------
	if getGitPkgFiles {

		// Try to get the local_builder_configuration.json file if it exists for two reasons:
		//     - Eventually get the branch passed the first time
		//     - Save the configuration if the fetch is done between a "init" and a "build"
		branchConfiguration, localBldConfFile, err := getLocalBuildConfig(ourCfgPkgDir, false)
		if err != nil {
			return err
		}

		// Get the current Active branch
		activeBranch, err := getActiveBranch(appStructure.ID)
		if err != nil {
			return err
		}

		// Now, decide which branch needs to be used, the order of priority is (from more priority to less)
		//    1. User choice through flag
		//    2. Previous user choice through 'local_build_config.json' file
		//    3. Current active branch if both other options are empty
		branchToFetch := ""
		switch {
		case localConfig.Branch != "":
			branchToFetch = localConfig.Branch
		case branchConfiguration.Branch != "":
			branchToFetch = branchConfiguration.Branch
		}
		if branchToFetch != "" {
			// Switch to specified branch if the user gave one
			err = switchToExistingBranch(appStructure.ID, branchToFetch)
			if err != nil {
				return err
			}

			// Defer switching back to current branch
			defer switchToExistingBranch(appStructure.ID, activeBranch)

		} else {
			branchToFetch = activeBranch
		}

		// Save the branch used in the configuration structure
		// It will be later saved in the 'local_build_config.json' file
		branchConfiguration.Branch = branchToFetch

		// Display info
		RPPrinter.CustomPrintF(localFetchCmdAnsLine5, branchToFetch)

		// Get the list of all files and directories
		fileDirList, err := getLocalGitpkgsFileOrDirsList(fullGitpkgDir, false)
		if err != nil {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
			return err
		}

		// Delete all the files and then all directories
		for _, tmpFile := range fileDirList["files"] {
			err := os.Remove(tmpFile)
			if err != nil {
				RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
				return err
			}
		}
		for _, tmpDir := range fileDirList["dirs"] {
			err := os.Remove(tmpDir)
			if err != nil {
				RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
				return err
			}
		}

		// Download the full gitpkg in the gitpkg/ directory
		files := []string{}
		content := rpapiv1.RPFileList{}
		// Ask for the content (the list of files)
		err = bckAPI.ApplicationGetListFiles(appStructure.ID.String(), &content)
		if err != nil {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
			return err
		}
		for _, item := range content.ListFiles {
			files = append(files, item.Name)
		}
		err = downloadGitPkgsAppFiles(appStructure.ID, fullGitpkgDir, files, false)
		if err != nil {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
			return err
		}

		// Create the md5sum for each file
		err = updateOrCreateMD5SumFile(fullGitpkgDir, []string{})
		if err != nil {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
			return err
		}

		// Write a file in the package directory to save the configuration
		localConfig.Branch = branchConfiguration.Branch
		err = utils.WriteJsonFile(localBldConfFile, *localConfig)
		if err != nil {
			return err
		}

		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
	}
	// ----------------------------------------------------

	// Fetch the 'mockconfig' file ------------------------
	if getMockFiles {
		// Retrieve all the build configuration files
		RPPrinter.CustomPrint(localFetchCmdAnsLine6)
		err = createRepoInheritanceFile(appStructure.Parent, ourCfgPkgDir)
		if err != nil {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
			RPPrinter.ForceFlush()
			return err
		}
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
	}
	// ----------------------------------------------------

	return nil
}

// prePackagingRoutine is a common "routine" between `package init` and
// `package build`. It allows to retrieve all the information that are common
// to `run` and `init` that have been set by the user
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: RPLocalBuildConfig, RPApplicationFull, error -- build config, the local app structure
// and an error if something went wrong
func prePackagingRoutine(cmd *cobra.Command, args []string) (RPLocalBuildConfig, rpapiv1.RPApplicationFull, error) {
	// Initialize the main structures we need to return
	fullConfiguration := RPLocalBuildConfig{}

	// Get all the context needed for the packaging
	appStructure, tmpSrcsDir, tmpPkgsDir, err := getLocalContext(cmd, args, false, false, false)
	if err != nil {
		return fullConfiguration, appStructure, err
	}

	// Save the packages and sources directories
	fullConfiguration.SourcesDir = tmpSrcsDir
	fullConfiguration.PackagesDir = tmpPkgsDir

	// Save the identities of the application and the parent project
	fullConfiguration.AppPkgName = appStructure.PackageName
	fullConfiguration.AppSlug = appStructure.Slug
	fullConfiguration.ProjectSlug = appStructure.Parent.ProjectSlug

	// Save the distributions in the configuration file
	for _, distro := range appStructure.Distributions {
		if distro.Set && distro.Slug != "" {
			fullConfiguration.DistrosList = append(fullConfiguration.DistrosList, distro.Slug)
		}
	}

	// Retrieve the parent project in order to save the available architecture
	project := rpapiv1.RPProjectFull{}
	if err := bckAPI.ProjectGet(appStructure.Parent.ProjectID.String(), &project); err != nil {
		return fullConfiguration, appStructure, err
	}
	for _, arch := range project.Architectures {
		if arch.Set && arch.ID != "" {
			fullConfiguration.ArchsList = append(fullConfiguration.ArchsList, arch.ID)
		}
	}

	// Display the path to both directories here
	RPPrinter.CustomPrintLn()
	RPPrinter.CustomPrintF(localPkgInitCmdAnsLine1, fullConfiguration.SourcesDir)
	RPPrinter.CustomPrintF(localPkgInitCmdAnsLine2, fullConfiguration.PackagesDir)
	RPPrinter.CustomPrintLn()

	return fullConfiguration, appStructure, nil
}

// cloneSourcesRepo can be called to clone the sources GIT repository at the
// right place. It also manages the status check to any already existing directory
//
// :param app: Structure containing all needed data for application
// :type app: *rpapiv1.RPApplicationFull
// :param config: Configuration to set for the local builder
// :type config: *RPLocalBuildConfig
// :param force: Ask the function to overwrite the gitsource directory anyway
// :type force: bool
// :returns: error -- an error if a problem occurred
func cloneSourcesRepo(app *rpapiv1.RPApplicationFull, localConfig *RPLocalBuildConfig, force bool) error {

	// Check gitsources status for the wanted repo --------
	fullGitSourcesDir := path.Join(localConfig.SourcesDir, app.Parent.ProjectSlug, app.PackageName)
	// Check if directory exists and not empty
	isDir, _ := utils.IsDirectory(fullGitSourcesDir)
	isDirEmpty, _ := utils.IsDirectoryEmpty(fullGitSourcesDir)
	if !force && isDir && !isDirEmpty {
		// Display info
		RPPrinter.CustomPrint(localPkgInitAnsLine0)
		// If the user does not wan't to force the clone, check if some modifications have been done in the repo
		gitStatusOutput, err := utils.RunCmdInDirectory(fullGitSourcesDir, "git", "status")
		if err != nil {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
			RPPrinter.ForceFlush()
			Log.Warnf(noGitSourceFoundWarn, gitStatusOutput)
		} else if !strings.Contains(gitStatusOutput, "nothing to commit, working tree clean") {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
			RPPrinter.ForceFlush()
			return errors.New(cannotCloneWithChanges)
		}
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
	}
	// ----------------------------------------------------

	// Remove the previous source directory if it exists --
	if isDir && !isDirEmpty {
		// Display info
		RPPrinter.CustomPrint(localPkgInitAnsLine1)
		// Remove the directory
		err := os.RemoveAll(fullGitSourcesDir)
		if err != nil {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
			RPPrinter.ForceFlush()
			return err
		}
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
	}
	// ----------------------------------------------------

	// Finally, clone the source repository ---------------
	// Use `iotpkg init` command
	RPPrinter.CustomPrintLn(localPkgInitCmdAnsLine5)
	RPPrinter.CustomColorPrintLn(printer.ColorPurple, localLocalBldLogsTag1)
	RPPrinter.ForceFlush()

	// Get the size of the current terminal
	termWidth, termHeight, err := term.GetSize(int(os.Stdin.Fd()))
	if err != nil {
		return err
	}

	initRequest := ladapi.RPLADIotpkgCmdStruct{
		RPLADMainCmdParams: ladapi.RPLADMainCmdParams{
			RPLADWinSize: ladapi.RPLADWinSize{
				WinSizeRows: termHeight,
				WinSizeCol:  termWidth,
			},
			WorkingDir: path.Join(getLclBldGitpkgsDir(), app.Parent.ProjectSlug, app.PackageName),
		},
		IotpkgFlags: []string{"-v"},
		SubcmdFlags: []string{"--sourcedir", path.Join("../../../gitsources", app.Parent.ProjectSlug, app.PackageName)},
	}

	procTicket := ladapi.RPLADRunnerTicket{}
	err = localBAPI.IotpkgInitPost(&initRequest, &procTicket)
	if err != nil {
		return err
	}

	err = listenStdoutErrAndIn(procTicket)
	RPPrinter.CustomColorPrintLn(printer.ColorPurple, localLocalBldLogsTag2)
	// ----------------------------------------------------

	return err
}

// handleIotpkgSubCmdHelp is used to handle the help requests from the user on
// the following commands
//
// - rp-cli local package service
// - rp-cli local package metadata
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param originalUsrArgs: Original arguments given by the user before being filtered
// :type originalUsrArgs: []string
// :param subCmdName: Name of the sub-command to handle
// :type subCmdName: string
// :returns: bool, error -- true if the user asked for help and an error if something went wrong
func handleIotpkgSubCmdHelp(cmd *cobra.Command, originalUsrArgs []string, subCmdName string) (bool, error) {
	// Firstly, check if something has been detected earlier
	if len(originalUsrArgs) == 0 {
		// No help needed
		return false, nil
	}

	ourArgs := originalUsrArgs[0:]

	// Since iotpkg does not like "help", convert every "help" to "--help"
	utils.ReplaceInStringSlice(&ourArgs, "help", "--help")

	// Get the elements that are interesting us
	ourIndex := utils.IndexOf(ourArgs, subCmdName)

	// Get rid of the first two args, to keep only the ones compatible with iotpkg
	iotpkgArgs := ourArgs[ourIndex:]
	subCommandArgs := iotpkgArgs[1:]
	// If the user gave an app name, get rid of it
	if len(iotpkgArgs[1:]) > 2 {
		subCommandArgs = iotpkgArgs[2:]
	}

	// Get the usage from rp-cli
	RPPrinter.CustomPrint(cmd.UsageString())

	RPPrinter.CustomColorPrintLn(printer.ColorPurple, localLocalBldLogsTag1)
	RPPrinter.ForceFlush()
	// Get the size of the current terminal
	termWidth, termHeight, err := term.GetSize(int(os.Stdin.Fd()))
	if err != nil {
		return true, err
	}

	helpRequest := ladapi.RPLADIotpkgCmdStruct{
		RPLADMainCmdParams: ladapi.RPLADMainCmdParams{
			RPLADWinSize: ladapi.RPLADWinSize{
				WinSizeRows: termHeight,
				WinSizeCol:  termWidth,
			},
			WorkingDir: "/tmp/",
		},
		SubcmdFlags: subCommandArgs,
	}

	procTicket := ladapi.RPLADRunnerTicket{}
	if iotpkgArgs[0] == "service" {
		err = localBAPI.IotpkgServicePost(&helpRequest, &procTicket)
	} else if iotpkgArgs[0] == "metadata" {
		err = localBAPI.IotpkgMetadataPost(&helpRequest, &procTicket)
	} else if iotpkgArgs[0] == "build" {
		err = localBAPI.IotpkgRPBuildPost(&helpRequest, &procTicket)
	}
	if err != nil {
		return true, err
	}

	err = listenStdoutErrAndIn(procTicket)
	RPPrinter.CustomColorPrintLn(printer.ColorPurple, localLocalBldLogsTag2)

	return true, err
}

// getMockNameFromCfgFile is used to extract the true mock name from the configuration
// file used to construct it
//
// :param pathToCfgFile: Path to the iotpkg file to extract the mock name from
// :type pathToCfgFile: string
// :returns: string, error -- name of the corresponding mock and an error if something went wrong
func getMockNameFromCfgFile(pathToCfgFile string) (string, error) {
	// Read the file
	fileContent, err := os.ReadFile(pathToCfgFile)
	if err != nil {
		return "", err
	}

	// Extract the line that is interesting us
	mockNameLine := utils.ExtractLineWithPrefix(string(fileContent), "config_opts['root']")
	if mockNameLine == "" {
		return "", nil
	}

	// Now extract the mock name
	splitLine := strings.Split(mockNameLine, "=")
	if len(splitLine) < 2 {
		return "", nil
	}
	mockName := strings.ReplaceAll(strings.TrimSpace(splitLine[1]), "'", "")

	return mockName, nil
}

// handleIotpkgCleanMock is used to request iotpkg to clean a given mock
//
// :param mockName: Name of the mock to clean
// :type mockName: string
// :returns: bool, error -- true if the user asked for help and an error if something went wrong
func handleIotpkgCleanMock(mockName string) error {
	// Local-builder tag
	RPPrinter.CustomColorPrintLn(printer.ColorPurple, localLocalBldLogsTag1)
	RPPrinter.ForceFlush()

	// Get the size of the current terminal
	termWidth, termHeight, err := term.GetSize(int(os.Stdin.Fd()))
	if err != nil {
		return err
	}

	cleanRequest := ladapi.RPLADIotpkgCmdStruct{
		RPLADMainCmdParams: ladapi.RPLADMainCmdParams{
			RPLADWinSize: ladapi.RPLADWinSize{
				WinSizeRows: termHeight,
				WinSizeCol:  termWidth,
			},
			WorkingDir: "/tmp",
		},
		IotpkgFlags: []string{"-v"},
		SubcmdFlags: []string{"-n", mockName},
	}

	procTicket := ladapi.RPLADRunnerTicket{}
	err = localBAPI.IotpkgCleanMockPost(&cleanRequest, &procTicket)
	if err != nil {
		return err
	}

	err = listenStdoutErrAndIn(procTicket)
	RPPrinter.CustomColorPrintLn(printer.ColorPurple, localLocalBldLogsTag2)

	return err
}

// getIotpkgArgs is a function allowing to filter out the flags and values
// used by rp-cli. In order to work, the following parameter must have been set this way:
//
//	FParseErrWhitelist: cobra.FParseErrWhitelist{UnknownFlags: true},
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :returns: []string, error -- the arguments that need to be passed to iotpkg and an error if a problem appeared
func getIotpkgArgs(cmd *cobra.Command) ([]string, error) {
	returnSlice, itemToDel := []string{}, []string{}

	// Firstly get the args given by the user (get rid of 'rp-cli' name)
	origArgs := os.Args[1:]

	// Retrieve the command path that will obviously appears in the full command-line args
	cmdPath := strings.Split(cmd.CommandPath(), " ")

	// Add the cmd path to the items to delete (but without the rp-cli name though, which is in first position)
	itemToDel = append(itemToDel, cmdPath[1:]...)

	// Now visit all flags
	filterOutRPCliFlags := func(flag *pflag.Flag) {
		// Add the name in our slice if the flag has been passed by the user
		if flag.Changed {
			Log.Debugf("[LOCAL] Filtering out %s flag\n", flag.Name)
			fullFlag := fmt.Sprintf("--%s", flag.Name)
			shortFlag := fmt.Sprintf("-%s", flag.Shorthand)

			// Boolean case - no additional value to delete
			if flag.NoOptDefVal != "" {
				// Delete what exists in the command line
				itemToDel = append(itemToDel, fullFlag, shortFlag)
			} else {
				// Non-boolean case, we need to get rid of the flag and the value!
				// Check if the user use the flag assignment with '='
				fullFlagWithEq := fmt.Sprintf("%s=%s", fullFlag, flag.Value.String())
				shortFlagWithEq := fmt.Sprintf("%s=%s", shortFlag, flag.Value.String())
				if utils.ContainsString(origArgs, fullFlagWithEq) {
					itemToDel = append(itemToDel, fullFlagWithEq)
				} else if utils.ContainsString(origArgs, fullFlagWithEq) {
					itemToDel = append(itemToDel, shortFlagWithEq)
				} else {
					itemToDel = append(itemToDel, fullFlag, shortFlag, flag.Value.String())
				}
			}
		}
	}
	cmd.Flags().VisitAll(filterOutRPCliFlags)

	Log.Debugf("[LOCAL] Items to delete from the original args: %+v\n", itemToDel)

	// Now create arguments for iotpkg
	for _, item := range origArgs {
		if !utils.ContainsString(itemToDel, item) {
			returnSlice = append(returnSlice, item)
		}
	}

	// Get rid of the application slug before returning, only if is has been set (and therefore, only if not a flag, i.e. starting with "-")
	if (len(returnSlice) < 1) || (strings.HasPrefix(returnSlice[0], "-")) {
		return returnSlice, nil
	} else {
		return returnSlice[1:], nil
	}
}

// checkLocalBuilderMountPoints is a function that test if the mount-points from
// the host to the local-builder is correctly done (directories and UID/GID mapping)
//
// :returns: error -- an error if a problem appeared
func checkLocalBuilderMountPoints() error {
	// Retrieve the local-builder directories
	rootSrcDir, rootGitpkgDir, err := getAndChecksBuildDirs("")
	if err != nil {
		return err
	}

	// Create a temporary files in gitsources/ and in gitpkg/ directories
	srcFile, err := os.CreateTemp(rootSrcDir, "")
	if err != nil {
		return err
	}
	defer os.Remove(srcFile.Name())
	srcFileInLocalBuilder := path.Join(getLclBldGitsourcesDir(), path.Base(srcFile.Name()))

	pkgFile, err := os.CreateTemp(rootGitpkgDir, "")
	if err != nil {
		return err
	}
	defer os.Remove(pkgFile.Name())
	pkgFileInLocalBuilder := path.Join(getLclBldGitpkgsDir(), path.Base(pkgFile.Name()))

	// Some logs
	Log.Debugf("Creating temporary files in local-builder: %s and %s\n", srcFileInLocalBuilder, pkgFileInLocalBuilder)

	// Check that the files are correctly seen inside the local-builder
	cmdOutput, err := execInLocalBuilderStatic("ls -al "+srcFileInLocalBuilder, "/tmp")
	if err != nil {
		return err
	}

	// Some logs
	Log.Debugf("'ls' result for %s: %s", srcFileInLocalBuilder, cmdOutput)

	if !strings.Contains(cmdOutput, "devel mock") && !strings.Contains(cmdOutput, "devel devel") {
		return fmt.Errorf(pbOnMountPoints, getLclBldGitsourcesDir())
	}

	cmdOutput, err = execInLocalBuilderStatic("ls -al "+pkgFileInLocalBuilder, "/tmp")
	if err != nil {
		return err
	}

	// Some logs
	Log.Debugf("'ls' result for %s: %s", pkgFileInLocalBuilder, cmdOutput)

	if !strings.Contains(cmdOutput, "devel mock") && !strings.Contains(cmdOutput, "devel devel") {
		return fmt.Errorf(pbOnMountPoints, getLclBldGitpkgsDir())
	}

	return nil
}

// execInLocalBuilderStatic executes a command inside the local-builder and returns
// the output of this command. Nothing is interactive here, the user does not see anything
//
// :param cmdLine: Command line to run in the local-builder
// :type cmdLine: string
// :param workingDir: Working directory where the command needs to be run
// :type workingDir: string
// :returns: error -- an error if a problem appeared
func execInLocalBuilderStatic(cmdLine, workingDir string) (string, error) {

	// Request body
	execRequest := ladapi.RPLADExecCmdStruct{
		RPLADMainCmdParams: ladapi.RPLADMainCmdParams{
			RPLADWinSize: ladapi.RPLADWinSize{
				WinSizeRows: 50,
				WinSizeCol:  50,
			},
			WorkingDir: workingDir,
		},
		CmdLine: cmdLine,
	}

	// Answer body template
	procTicket := ladapi.RPLADRunnerTicket{}
	err := localBAPI.ExecPost(&execRequest, &procTicket)
	if err != nil {
		return "", err
	}

	// Get the output from the redpesk-lad websocket
	// Block until we receive all the websocket events (until localProcEndEvt)
	procRunning := true
	cmdOutput := ""
	var errToReturn error
	for procRunning {
		select {
		// We received an event from the remote server
		case localEvent := <-cmdEventLocalBChannel:
			// Check that the received event is for our process
			if localEvent.ID == procTicket.ID {
				switch localEvent.Type {
				case localStdOutEvt:
					cmdOutput += localEvent.Line
				case localStdErrEvt:
					cmdOutput += localEvent.Line
				case localProcEndEvt:
					// This is the end of the terminal
					errorMsg := fmt.Sprintf(endOfProc, localEvent.ID, localEvent.Status)
					Log.Debugln(errorMsg)
					// End this blocking code
					procRunning = false
					// Return an error if the terminal end badly
					if localEvent.Status != 0 {
						errToReturn = errors.New(errorMsg)
					} else {
						errToReturn = nil
					}
				}
			}
		// Disconnection from the websocket, since everything go through the
		// socket, stop the command
		case <-localBConnection.WSocket.DisconnectionChan:
			// End this blocking code
			procRunning = false
			Log.Error(webSocketDisconnectionErr)
			errToReturn = errors.New(webSocketDisconnectionErr)
		}
	}

	Log.Debugf("Command run in local-builder: '%s' in directory '%s'\n", cmdLine, workingDir)
	Log.Debugf("Command return: '%s'\n", cmdOutput)

	return cmdOutput, errToReturn
}

// terminalSendSignal allows to send a signal to a process running remotely
// in the container where redpesk LAD is running
//
// :param id: ID of the process to send the signal to
// :type id: string
// :param signal: Signal to send to the process
// :type signal: os.Signal
func terminalSendSignal(id string, signal os.Signal) {
	// Initialize the structure to send to the server
	sigStructure := ladapi.RPLADSignalSend{
		ID:     id,
		Signal: signal.String(),
	}

	// Send the request to the server
	localBAPI.TerminalSendSignalPost(&sigStructure)
}

// terminalResizeWin allows to resize the window of the terminal
// open in the remote server (redpesk LAD)
//
// :param id: ID of the process to resize
// :type id: string
func terminalResizeWin(id string) {
	// Get the current terminal size
	termWidth, termHeight, err := term.GetSize(int(os.Stdin.Fd()))
	if err != nil {
		Log.Errorf(getTermSizeError, err)
	}

	// Initialize the structure to send to the server
	winSizeStructure := ladapi.RPLADWinResize{
		ID: id,
		RPLADWinSize: ladapi.RPLADWinSize{
			WinSizeRows: termHeight,
			WinSizeCol:  termWidth,
		},
	}

	// Send the request to the server
	localBAPI.TerminalSizePut(&winSizeStructure)
}

// checkLocalBuilderMountPoints is a function that test if the qemu
// is correctly set-up on the host machine
//
// :returns: error -- an error if a problem appeared
func checkQemuSetUp() error {
	// Detect the linux distribution
	linuxDistro := utils.GetLinuxDistro()
	if linuxDistro == "ubuntu" {
		// Check partially the content of /var/lib/binfmts/qemu-aarch64
		binFmtAarch64Content, err := os.ReadFile(getQemuAarch64FilePath("/var/lib/binfmts/qemu-aarch64"))
		if err != nil {
			return err
		}
		if !strings.Contains(string(binFmtAarch64Content), varLibBinfmtsQemuAarch64) {
			Log.Error(varLibBinfmtsFileNotGood)
			Log.Errorf("Try to fix this by running the following command:\n%s\n", fixVarLibBinfmtsQemuAarch64file)
			return errors.New(varLibBinfmtsFileNotGood)
		}
	}

	// Check the content of /proc/sys/fs/binfmt_misc/qemu-aarch64
	binFmtMiscAarch64Content, err := os.ReadFile(getQemuAarch64FilePath("/proc/sys/fs/binfmt_misc/qemu-aarch64"))
	if err != nil {
		return err
	}
	if !strings.Contains(string(binFmtMiscAarch64Content), procSysFsBinfmtMiscQemuAarch64V0) &&
		!strings.Contains(string(binFmtMiscAarch64Content), procSysFsBinfmtMiscQemuAarch64V1) &&
		!strings.Contains(string(binFmtMiscAarch64Content), procSysFsBinfmtMiscQemuAarch64V2) {
		return errors.New(procSysFsBinfmtsFileNotGood)
	}
	return nil
}

// filterNonGitpkgsFiles is a function that delete the files that we are not considering
// inside a gitpkg directory when the user asks for the status or to commit
//
// :param filesList: input list of files to filter
// :type filesList: []string
// :param fullGitpkgDir: Full path to the directory containing the files
// :type fullGitpkgDir: string
// :returns: []string -- the filtered list of files
func filterNonGitpkgsFiles(filesList []string, fullGitpkgDir string) []string {
	filteredFiles := []string{}

	// Create a list of prefixes that we will used to filter the files
	prefixToFilter := []string{".redpesk/autogen::", "gitsource", ".iotpkg/", getResultsDirName(fullGitpkgDir)}

	// Create the path to delete from the file name
	dirPrefix := fullGitpkgDir
	if !strings.HasSuffix(dirPrefix, "/") {
		dirPrefix += "/"
	}

	// Check if we got a '.gitignore' file
	var ignore gitignore.GitIgnore = nil
	gitignorePath := path.Join(fullGitpkgDir, ".gitignore")
	if _, err := os.Stat(gitignorePath); err == nil {
		ignore, err = gitignore.NewFromFile(gitignorePath)
		if err != nil {
			Log.Errorf("error when creating ignore object from %s", gitignorePath)
		}
	}

	// Now filter the files having this prefixes
	for _, file := range filesList {
		tmpFile := strings.ReplaceAll(file, dirPrefix, "")
		currentFileHasPrefix := false

		// Check if the current file is in the .gitignore list (if the file exists)
		if ignore != nil && ignore.Ignore(file) {
			continue
		}

		// Loop over the prefixes to ignore
		for _, prefix := range prefixToFilter {
			if strings.HasPrefix(tmpFile, prefix) {
				currentFileHasPrefix = true
				break
			}
		}

		// If the file does not have a prefix that we ignore, add it to the final list
		if !currentFileHasPrefix {
			filteredFiles = append(filteredFiles, file)
		}
	}

	return filteredFiles
}

// getQemuAarch64FilePath is here uniquely for compatibility with
// CI test to redirect to the mock files
//
// :returns: error -- an error if a problem appeared
func getQemuAarch64FilePath(initialPath string) string {
	if runMode == "testing" && strings.Contains(initialPath, "/proc/sys/fs") {
		return utils.GetPkgRootDir() + "rpladmock/binfmt_qemu/proc_sys_fs_binfmt_misc_qemu-aarch64"
	} else if runMode == "testing" && strings.Contains(initialPath, "/var/lib/binfmts") {
		return utils.GetPkgRootDir() + "rpladmock/binfmt_qemu/var_lib_binfmts_qemu-aarch64"
	} else {
		return initialPath
	}
}

// getResultsDirName is a function that returns the name of the directory
// that will contains the results of a build, when local-building
//
// :param fullGitpkgDir: Full path to the directory containing the files
// :type fullGitpkgDir: string
// :returns: string -- name of the results directory
func getResultsDirName(fullGitpkgDir string) string {
	// Extract the application package name from the fullGitpkgDir path
	appPkgName := path.Base(fullGitpkgDir)
	return ("results_" + appPkgName)
}

// getAppAndPrjSlugFromLocalBldConfig retrieves the application and project
// slugs from the local-builder configuration file if the current working
// directory is either the host 'gitpkgs/' or 'gitsources/' directory.
// If it is not the case, empty strings are returned.
//
// :returns: string, string, error -- application slug, project slug and an error if something bad happened
func getAppAndPrjSlugFromLocalBldConfig() (string, string, error) {
	// First retrieve the current working directory
	cwd, err := os.Getwd()
	if err != nil {
		return "", "", err
	}

	// Get the local-build directories
	gitSrcDir, gitPkgDir, err := getAndChecksBuildDirs("")
	if err != nil {
		return "", "", err
	}

	// Check if we are in either a valid 'gitpkgs/' or 'gitsources/' directory (for an app)
	if !strings.HasPrefix(cwd, gitSrcDir) && !strings.HasPrefix(cwd, gitPkgDir) {
		// We are not, return empty strings
		return "", "", nil
	}

	// Check that we are in two subdirectories inside either 'gitpkgs/' or 'gitsources/'
	subDirectories := strings.Replace(cwd, gitSrcDir, "", 1)
	subDirectories = strings.Replace(subDirectories, gitPkgDir, "", 1)
	matched, err := path.Match("/*/*", subDirectories)
	if err != nil {
		return "", "", err
	}
	if !matched {
		// We are not, return empty strings
		return "", "", nil
	}

	// Finally, check if we are in the sources or in the gitpkgs
	// If we are in the sources retrieve the path to the gitpkgs
	dirToUse := cwd
	if strings.HasPrefix(cwd, gitSrcDir) {
		dirToUse = strings.Replace(cwd, "gitsources", "gitpkgs", 1)
	}

	// Create the path to the configuration file
	cfgFile := path.Join(dirToUse, localCfgSubDirectory, localBldCfgFile)

	Log.Debugf("Retrieving application slug from '%s'", cfgFile)

	localBldCfg := RPLocalBuildConfig{}

	// Get the info from the file
	err = utils.ReadJsonFile(cfgFile, &localBldCfg)
	if err != nil {
		return "", "", err
	}

	// Finally, returns the value
	return localBldCfg.AppSlug, localBldCfg.ProjectSlug, nil
}

// retrieveAllLocalBldConfigs retrieves all the configurations available
// locally to build. This permits to rp-cli to not need every time to
// connect to the redpesk factory in order to just retrieve some
// basic info like app slug, project slug, etc.
//
// The return object is a map of all the configurations, index with
// the following string: "prj-slug/app-slug"
//
// :returns: map[string]RPLocalBuildConfig, error -- a map containing all the configs and an error if something bad happened
func retrieveAllLocalBldConfigs() (map[string]RPLocalBuildConfig, error) {
	resultMap := make(map[string]RPLocalBuildConfig)

	// Get the local-build directories
	_, gitPkgDir, err := getAndChecksBuildDirs("")
	if err != nil {
		return resultMap, err
	}

	// Retrieve all the files and directories
	filesAndDirMap, err := utils.FindAllFilesAndDirsRecursively(gitPkgDir)
	if err != nil {
		return resultMap, err
	}

	// Now in the files list, get every 'local_build_config.json' file in order to fill our resultMap
	for _, filePath := range filesAndDirMap["files"] {
		if strings.HasSuffix(filePath, localBldCfgFile) {
			tmpConfig := RPLocalBuildConfig{}

			// Read the configuration
			err = utils.ReadJsonFile(filePath, &tmpConfig)
			if err != nil {
				return resultMap, err
			}

			// Add it in our resultMap
			resultMap[tmpConfig.ProjectSlug+"/"+tmpConfig.AppSlug] = tmpConfig
		}
	}
	return resultMap, nil
}

// retrieveOneLocalBldConfig retrieves the local configuration corresponding to
// the project slug and the application package name given as parameter
//
// :param prjSlug: Project slug
// :type prjSlug: string
// :param appPkgName: Application package name
// :type appPkgName: string
// :returns: RPLocalBuildConfig, error -- the local configuration and an error if something bad happened
func retrieveOneLocalBldConfig(prjSlug, appPkgName string) (RPLocalBuildConfig, error) {
	resultCfg := RPLocalBuildConfig{}

	// Get the local-build directories
	_, gitPkgDir, err := getAndChecksBuildDirs("")
	if err != nil {
		return resultCfg, err
	}

	// Creates the path to the configuration
	localCfgPath := path.Join(gitPkgDir, prjSlug, appPkgName, ".rp-cli", localBldCfgFile)

	// Read the configuration json file
	err = utils.ReadJsonFile(localCfgPath, &resultCfg)
	return resultCfg, err
}

// handleBranchSwitch can be used by functions that needs to work with a different
// branch than the one possibly used as active. So this function retrieves the branch
// to work with, switch to it and returns the original branch. This is the responsability
// of the caller to call the "defer" method like that:
//
//	'defer switchToExistingBranch(appStructure.ID, prevBranch)'
//
// # If the function return an empty string an no error this means that we are already on the right branch
//
// :param fullGitpkgDir: Full path to the gitpkg directory containing the local-build configuration
// :type fullGitpkgDir: string
// :param appID: UUID of the application
// :type appID: uuid.UUID
// :returns: RPLocalBuildConfig, error -- the local configuration and an error if something bad happened
func handleBranchSwitch(fullGitpkgDir string, appID uuid.UUID) (string, error) {
	// Get the branch used to fetch the gitpkgs files
	branchConfiguration, _, err := getLocalBuildConfig(fullGitpkgDir, false)
	if err != nil {
		return "", err
	}

	// Get the active branch to see if we need to change the branch
	activeBranch, err := getActiveBranch(appID)
	if err != nil {
		return "", err
	}

	// Switch to specified branch if needed
	if activeBranch != branchConfiguration.Branch {
		err = switchToExistingBranch(appID, branchConfiguration.Branch)
		if err != nil {
			return "", err
		}

		return activeBranch, nil
	}
	return "", nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Display functions --------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// --------------------------------------------------------------
