/*
Copyright © 2020-2023 IoT.bzh Company
Contact: https://www.iot.bzh/licensing

This file is part of the redpesk-cli project, the redpesk command line tool.

$RP_BEGIN_LICENSE$
# Commercial License Usage

Licensees holding valid commercial IoT.bzh licenses may use this file in
accordance with the commercial license agreement provided with the
Software or, alternatively, in accordance with the terms contained in
a written agreement between you and The IoT.bzh Company. For licensing terms
and conditions see https://www.iot.bzh/terms-conditions. For further
information use the contact form at https://www.iot.bzh/contact.

# GNU General Public License Usage

Alternatively, this file may be used under the terms of the GNU General
Public license version 3. This license is as published by the Free Software
Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
of this file. Please review the following information to ensure the GNU
General Public License requirements will be met
https://www.gnu.org/licenses/gpl-3.0.html.

$RP_END_LICENSE$
*/
package cmd

import (
	"encoding/json"
	"fmt"
	"os"
	"path"
	"strings"
	"testing"
	"time"

	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/rpconnect"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/rpladmock"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/utils"
	"git.ovh.iot/redpesk/redpesk-infra/rp-golang-common/rputils"
	homedir "github.com/mitchellh/go-homedir"
	"github.com/stretchr/testify/assert"
)

// Constants
const (
	// Directories used for tests
	gitSourcesTestDir  = "/tmp/redpesk-cli-test/gitsources"
	gitPackagesTestDir = "/tmp/redpesk-cli-test/gitpkgs"
	cfgDirPathTestDir  = "/tmp/redpesk-cli-test/.redpesk/rp-cli/"

	package4TestsHelloAlias     = "my-helloworld-binding"
	package4TestsHelloName      = "My Helloworld binding"
	package4TestsHelloURL       = "http://git.ovh.iot/redpesk/redpesk-samples/helloworld-binding.git"
	package4TestsHelloPathInSrc = "conf.d/packaging/helloworld-binding.spec"

	albatrosPoem = `L’albatros

Souvent, pour s’amuser, les hommes d’équipage
Prennent des albatros, vastes oiseaux des mers,
Qui suivent, indolents compagnons de voyage,
Le navire glissant sur les gouffres amers.

A peine les ont-ils déposés sur les planches,
Que ces rois de l’azur, maladroits et honteux,
Laissent piteusement leurs grandes ailes blanches
Comme des avirons traîner à côté d’eux.

Ce voyageur ailé, comme il est gauche et veule !
Lui, naguère si beau, qu’il est comique et laid !
L’un agace son bec avec un brûle-gueule,
L’autre mime, en boitant, l’infirme qui volait !

Le Poète est semblable au prince des nuées
Qui hante la tempête et se rit de l’archer ;
Exilé sur le sol au milieu des huées,
Ses ailes de géant l’empêchent de marcher.

Charles Baudelaire`

	// Connections
	localConnectDisplayNoVerboseColNb = 3
	localConnectDisplayVerboseLineNb  = 3
)

// TestOnBoardLocalBuilder is the function allowing to test if
// the command "rp-cli onboard" is working correctly for the
// local-builder connection part
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestOnBoardLocalBuilder(test *testing.T) {

	// We will have to fake the stdin in this situation (because this command is interactive)
	// See https://eli.thegreenplace.net/2020/faking-stdin-and-stdout-in-go/ for more info

	// Create a false local_builder configuration file
	cfgFilePath, err := createFalseLocalConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Get the home directory
	home, err := homedir.Dir()
	if err != nil {
		test.Fatal(err)
	}
	gitsourcesPath := path.Join(home, "my_rp_builder_dir", "gitsources")

	// Check if the directory exists
	defBuildDirExists := rputils.Exists(gitsourcesPath)

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(cfgFilePath)

	// Create a pipe
	readPipe, writePipe, err := os.Pipe()
	if err != nil {
		test.Fatal(err)
	}

	// Save the original *os.File
	origStdin := os.Stdin

	// Set os.Stdin with our fake one
	os.Stdin = readPipe

	// Start a go routine that will fill what the command 'rp-cli onboard' wants
	go func() {
		// Wait a bit, in order for the command to be executed
		time.Sleep(100 * time.Millisecond)

		// Say that we do not want to configure the backend connection
		writePipe.Write([]byte("n\n"))

		// Wait a bit, in order for the command to be executed
		time.Sleep(100 * time.Millisecond)

		// Say that we want to configure the local-builder connection
		writePipe.Write([]byte("Y\n"))

		// Wait a bit, in order for the command to be executed
		time.Sleep(100 * time.Millisecond)

		// Write in the pipe the localbuilder IP
		writePipe.Write([]byte("http://0.0.0.0:8112\n"))

		// Wait a bit, in order for the command to be executed
		time.Sleep(100 * time.Millisecond)

		// Write in the pipe that we don't want to modify the IP
		writePipe.Write([]byte("n\n"))

		if defBuildDirExists {
			time.Sleep(100 * time.Millisecond)
			// Write in the pipe that we don't want to modify the IP
			writePipe.Write([]byte("n\n"))
		}

		// Write in the pipe
		time.Sleep(100 * time.Millisecond)
		writePipe.Write([]byte("/tmp/my_rp_builder_dir/gitsources\n"))

		// Wait a bit, in order for the command to be executed
		time.Sleep(100 * time.Millisecond)

		// Say that we want to create the directory
		writePipe.Write([]byte("Y\n"))

		if defBuildDirExists {
			time.Sleep(100 * time.Millisecond)
			// Write in the pipe that we don't want to modify the IP
			writePipe.Write([]byte("n\n"))
		}

		// Wait a bit, in order for the command to be executed
		time.Sleep(100 * time.Millisecond)

		// Write in the pipe
		writePipe.Write([]byte("/tmp/my_rp_builder_dir/gitpkgs\n"))

		// Wait a bit, in order for the command to be executed
		time.Sleep(100 * time.Millisecond)

		// Say that we want to create the directory
		writePipe.Write([]byte("Y\n"))
	}()

	// Execute the onboard command
	cmdOutput, err := executeCommand(rootOnBoardCmd)

	// Restore Stdin
	os.Stdin = origStdin

	if err != nil {
		test.Fatal(err)
	}

	// Remove the directories that have been created
	defer os.RemoveAll("/tmp/my_rp_builder_dir")

	// Check that the new value in the local builder configuration file is good
	writenConnection, err := getConnectionParams(localBuilderAlias, localBuilderFile)
	if err != nil {
		test.Fatal(err)
	}
	assert.Equal(test, "http://0.0.0.0:8112", writenConnection.ServerURL)

	// Check that in cmdOutput, an [OK] appears
	checkLineEndWith(test, cmdOutput, "> Modification of the local builder file...", "[OK]")

	// Check that the directories have been created
	assert.DirExists(test, "/tmp/my_rp_builder_dir/gitsources")
	assert.DirExists(test, "/tmp/my_rp_builder_dir/gitpkgs")
}

// TestLocalConnectionsAdd is the function allowing to test if
// the command "rp-cli local connections add" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalConnectionsAdd(test *testing.T) {
	// Create a new local-builder connection object
	connectToCreate := rpconnect.RPCLIConnection{
		Alias:     "localbuilder-testadd",
		ServerURL: "10.231.232.5",
	}
	localBuilderFilePath, err := addLocalConnectionForTest(test, &connectToCreate)

	// Delete the false file created at the beginning
	defer deleteFalseFile(localBuilderFilePath)
	if err != nil {
		test.Fatal(err)
	}

	// Check that the connection has been correctly added
	regexpMap := map[string]string{
		"Alias": `^localbuilder-testadd$`,
		"IP":    `^10\.231\.232\.5$`,
	}

	// Run the command and the tests
	checkCmdAnsKeysRegexpCol(test, localConnectDisplayNoVerboseColNb, regexpMap, localCommand, connectionsCommand, "get", "localbuilder-testadd")
}

// TestLocalConnectionsGet is the function allowing to test if
// the command "rp-cli local connections get" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalConnectionsGet(test *testing.T) {
	// Create a new local-builder connection object
	connectToCreate := rpconnect.RPCLIConnection{
		Alias:     "localbuilder-testget",
		ServerURL: "10.231.232.6",
	}
	localBuilderFilePath, err := addLocalConnectionForTest(test, &connectToCreate)

	// Delete the false file created at the beginning
	defer deleteFalseFile(localBuilderFilePath)
	if err != nil {
		test.Fatal(err)
	}

	// Check that the connection has been correctly added
	regexpMap := map[string]string{
		"Alias": `^localbuilder-testget$`,
		"IP":    `^10\.231\.232\.6$`,
	}

	// Run the command and the tests
	checkCmdAnsKeysRegexpCol(test, localConnectDisplayNoVerboseColNb, regexpMap, localCommand, connectionsCommand, "get", "localbuilder-testget")

	// Modify the map in order to test the verbose mode
	regexpMap["Local-builder IP"] = `^10\.231\.232\.6$`

	// Run the command and the tests
	checkCmdAnswerListKeysRegexp(test, localConnectDisplayVerboseLineNb, "", regexpMap, localCommand, connectionsCommand, "get", "localbuilder-testget", "-v")
}

// TestLocalConnectionsUseAndCurrent is the function allowing to test if
// the commands "rp-cli local connections use" and
// "rp-cli local connections current" are working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalConnectionsUseAndCurrent(test *testing.T) {
	// Create a new local-builder connection object
	connectToCreate := rpconnect.RPCLIConnection{
		Alias:     "localbuilder-test-use-current-0",
		ServerURL: "10.231.232.7",
	}
	localBuilderFilePath, err := addLocalConnectionForTest(test, &connectToCreate)

	// Delete the false file created at the beginning
	defer deleteFalseFile(localBuilderFilePath)
	if err != nil {
		test.Fatal(err)
	}

	// Create another connection
	connectToCreate = rpconnect.RPCLIConnection{
		Alias:     "localbuilder-test-use-current-1",
		ServerURL: "10.231.232.8",
	}
	_, err = addLocalConnectionForTest(test, &connectToCreate)
	if err != nil {
		test.Fatal(err)
	}

	// Request rp-cli to use a connection that we just add
	_, err = executeCommand(localCommand, connectionsCommand, "use", "localbuilder-test-use-current-0")
	if err != nil {
		test.Fatal(err)
	}

	// Check that the current connection returned is "localbuilder-test-use-current-0"
	regexpMap := map[string]string{
		"Alias": `^localbuilder-test-use-current-0$`,
		"IP":    `^10\.231\.232\.7$`,
	}

	// Run the command and the tests
	checkCmdAnsKeysRegexpCol(test, localConnectDisplayNoVerboseColNb, regexpMap, localCommand, connectionsCommand, "current")

	// Verbose mode
	regexpMap["Local-builder IP"] = `^10\.231\.232\.7$`
	checkCmdAnswerListKeysRegexp(test, localConnectDisplayVerboseLineNb, "", regexpMap, localCommand, connectionsCommand, "current", "-v")
}

// TestLocalConnectionsList is the function allowing to test if
// the command "rp-cli local connections list" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalConnectionsList(test *testing.T) {
	// Create a new local-builder connection object
	connectToCreate := rpconnect.RPCLIConnection{
		Alias:     "localbuilder-test-list-0",
		ServerURL: "10.231.233.1",
	}
	localBuilderFilePath, err := addLocalConnectionForTest(test, &connectToCreate)

	// Delete the false file created at the beginning
	defer deleteFalseFile(localBuilderFilePath)
	if err != nil {
		test.Fatal(err)
	}

	// Create other connections
	for i := 1; i <= 3; i++ {
		connectToCreate.Alias = fmt.Sprintf("localbuilder-test-list-%d", i)
		_, err = addLocalConnectionForTest(test, &connectToCreate)
		if err != nil {
			test.Fatal(err)
		}
	}

	// Check now that the connections created are correctly listed

	// Not verbose mode
	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"Alias": `^localbuilder-test-list-[0-9]{1}$`,
		"IP":    `^10\.231\.233\.1$`,
	}
	checkCmdAnsKeysRegexpCol(test, localConnectDisplayNoVerboseColNb, regexpMap, localCommand, connectionsCommand, "list")

	// Verbose mode
	regexpMap["Local-builder IP"] = `^10\.231\.233\.1$`
	checkCmdAnswerListKeysRegexp(test, localConnectDisplayVerboseLineNb, "", regexpMap, localCommand, connectionsCommand, "list", "-v")
}

// TestLocalConnectionsUpdate is the function allowing to test if
// the command "rp-cli local connections update" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalConnectionsUpdate(test *testing.T) {
	// Create a new local-builder connection object
	connectToCreate := rpconnect.RPCLIConnection{
		Alias:     "localbuilder-test-update",
		ServerURL: "10.231.234.1",
	}
	localBuilderFilePath, err := addLocalConnectionForTest(test, &connectToCreate)

	// Delete the false file created at the beginning
	defer deleteFalseFile(localBuilderFilePath)
	if err != nil {
		test.Fatal(err)
	}

	// Run the update command and check that all is [OK]
	checkActionCmdAns(test, localCommand, connectionsCommand, "update", "localbuilder-test-update", "-i", "10.231.234.2")

	// Check that the update command correctly worked
	regexpMap := map[string]string{
		"Alias": `^localbuilder-test-update$`,
		"IP":    `^10\.231\.234\.2$`,
	}

	// Run the command and the tests
	checkCmdAnsKeysRegexpCol(test, localConnectDisplayNoVerboseColNb, regexpMap, localCommand, connectionsCommand, "get", "localbuilder-test-update")
}

// TestLocalExecCmdLs is the function allowing to test if
// the command "rp-cli local exec" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalExecCmdLs(test *testing.T) {
	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Create a directory in our gitsources/ used for test
	testDir := gitSourcesTestDir + "/test-local-exec-dir"
	err := utils.CreateDirIfNotExists(testDir, true)
	if err != nil {
		test.Fatal(err)
	}

	// Create few tests files
	utils.WriteToFile(testDir+"/test_01.log", "THIS IS A FIRST FILE!!!")
	utils.WriteToFile(testDir+"/test_iot_bzh.log", "THIS IS SPARTA!!!")
	utils.WriteToFile(testDir+"/redpesk_test_lorient.log", "TOUS A LA TAVERNE!!!\n LOVE")
	utils.WriteToFile(testDir+"/linux_i_love_you.log", "THIS HURTS HERE, RIGHT IN MY MEOW MEOW")

	// Run the "ls -al" command
	cmdOutput, err := execCmdInSimTerm("", localCommand, "exec", "-w", testDir, "--", "ls", "-ahl")
	if err != nil {
		// The command returns an error
		test.Fatal(err)
	}

	// Check that the returned output contains all the file created
	assert.Contains(test, cmdOutput, "test_01.log")
	assert.Contains(test, cmdOutput, "test_iot_bzh.log")
	assert.Contains(test, cmdOutput, "redpesk_test_lorient.log")
	assert.Contains(test, cmdOutput, "linux_i_love_you.log")
}

// TestLocalExecCmdSmartWDLink is the function allowing to test if
// the command "rp-cli local exec" is working correctly when launched
// from a gitpkgs/ or gitsources/ directory
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalExecCmdSmartWDLink(test *testing.T) {
	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a directory in our gitsources/ used for test
	testDir := gitSourcesTestDir + "/test-local-exec-smart-link"
	err = utils.CreateDirIfNotExists(testDir, true)
	if err != nil {
		test.Fatal(err)
	}

	// Create few tests files
	utils.WriteToFile(testDir+"/redpesk_test_lorient.log", "TOUS A LA TAVERNE!!!\n LOVE")
	utils.WriteToFile(testDir+"/linux_i_love_you.log", "THIS HURTS HERE, RIGHT IN MY MEOW MEOW")

	// Change the current directory from Golang point of view
	err = os.Chdir(testDir)
	if err != nil {
		test.Fatal(err)
	}

	// Run the "ls -al" command
	cmdOutput, err := execCmdInSimTerm("", localCommand, "exec", "--", "ls", "-ahl")
	if err != nil {
		// The command returns an error
		test.Fatal(err)
	}

	// Check that the returned output contains all the file created
	assert.Contains(test, cmdOutput, "redpesk_test_lorient.log")
	assert.Contains(test, cmdOutput, "linux_i_love_you.log")
}

// TestLocalExecCmdEchoStdin is the function allowing to test if
// the command "rp-cli local exec" is working correctly with a bit
// of interactivity using the stdin event
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalExecCmdEchoStdin(test *testing.T) {
	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Create a directory in our gitsources/ used for test
	testDir := gitSourcesTestDir + "/test-local-exec-dir"
	err := utils.CreateDirIfNotExists(testDir, true)
	if err != nil {
		test.Fatal(err)
	}

	// Define the stdin to pass to the command
	stdinString := albatrosPoem

	// Calculate the md5sum on the stdin test
	stdinMd5Sum := utils.CalculateMd5SumOnString(stdinString)

	// Define the variables that we will use for test retries...
	testCounter := 0
	testOK := false
	md5SumCreatedFile := ""
	for !testOK && testCounter < 3 {
		// Write some logs
		test.Logf("[INFO] Attempt n°%d to validate stdin test", testCounter)

		// Run the "cat" command in an interactive way
		_, err = execCmdInSimTerm(stdinString, localCommand, "exec", "-w", testDir, "--", "sleep", "5")
		if err != nil {
			// The command returns an error
			test.Fatal(err)
		}

		// Check that the file has been created and contains what we write in it
		filesTab, err := utils.FindFilesWithExtension(testDir, ".log")
		assert.Nil(test, err)
		assert.Len(test, filesTab, 1)
		assert.Regexp(test, "test_file_.*.log$", filesTab[0])

		// Calculate the md5sum on the created file
		md5SumCreatedFile, err = utils.CalculateMd5SumOnFile(filesTab[0], true)
		if err != nil {
			test.Fatal(err)
		}

		if md5SumCreatedFile == stdinMd5Sum {
			testOK = true
		}

		// Increase the counter
		testCounter += 1

		// Remove the file
		err = os.Remove(filesTab[0])
		if err != nil {
			test.Fatal(err)
		}
	}

	// Test that the md5sums are the same
	assert.Equal(test, stdinMd5Sum, md5SumCreatedFile)
}

// TestLocalPackageFetch is the function allowing to test if
// the command "rp-cli local package fetch" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalPackageFetch(test *testing.T) {
	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project and an application for our tests
	err = createFullPrjAppsBuildsTests(test, "local-fetch-project", "redpesk-lts-arz", "chickenpox", 1, 0, 0)
	if err != nil {
		test.Fatal(err)
	}

	// Ask rp-cli to fetch all the gitpkg files of the newly created app
	cmdOutput := checkActionCmdAns(test, localCommand, localPkgCmdName, "fetch", "chickenpox-0")

	// Check various things on the output string
	assert.Contains(test, cmdOutput, "Retrieve application details")
	assert.Contains(test, cmdOutput, "Check local-builder configuration")
	assert.Contains(test, cmdOutput, "Create directories if needed")
	assert.Contains(test, cmdOutput, "Check 'gitpkg' repository status")
	assert.Contains(test, cmdOutput, "Fetch the 'gitpkg' repository from branch 'main'")
	assert.Contains(test, cmdOutput, "Fetch local-builder configuration")

	// Check that the right files exist now
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-fetch-project/chickenpox-0/chickenpox-0.spec")
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-fetch-project/chickenpox-0/chickenpox-0-1.0.0.tar.gz")
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-fetch-project/chickenpox-0/.redpesk/autogen::service.json")
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-fetch-project/chickenpox-0/.rp-cli/local-fetch-project.json")
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-fetch-project/chickenpox-0/.rp-cli/md5sum.json")
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-fetch-project/chickenpox-0/.rp-cli/"+localBldCfgFile)
}

// TestLocalPackageFetchTestFlags is the function allowing to test if
// the command "rp-cli local package fetch" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalPackageFetchTestFlags(test *testing.T) {
	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project and applications for our tests
	err = createFullPrjAppsBuildsTests(test, "local-fetch-flags-project", "redpesk-lts-arz", "chickenpox-flags", 2, 0, 0)
	if err != nil {
		test.Fatal(err)
	}

	// Gitpkg flag --------------------
	cmdOutput := checkActionCmdAns(test, localCommand, localPkgCmdName, "fetch", "chickenpox-flags-0", "-g")
	// Check various things on the output string
	assert.Contains(test, cmdOutput, "Check 'gitpkg' repository status")
	assert.Contains(test, cmdOutput, "Fetch the 'gitpkg' repository")
	assert.NotContains(test, cmdOutput, "Fetch local-builder configuration")
	// Check that the right files exist now
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-fetch-flags-project/chickenpox-flags-0/chickenpox-flags-0.spec")
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-fetch-flags-project/chickenpox-flags-0/chickenpox-flags-0-1.0.0.tar.gz")
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-fetch-flags-project/chickenpox-flags-0/.redpesk/autogen::service.json")
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-fetch-flags-project/chickenpox-flags-0/.rp-cli/md5sum.json")
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-fetch-flags-project/chickenpox-flags-0/.rp-cli/"+localBldCfgFile)
	assert.NoFileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-fetch-flags-project/chickenpox-flags-0/.rp-cli/local-fetch-flags-project.json")
	// --------------------------------

	// Localbuilder flag --------------
	cmdOutput = checkActionCmdAns(test, localCommand, localPkgCmdName, "fetch", "chickenpox-flags-1", "-c")
	// Check various things on the output string
	assert.NotContains(test, cmdOutput, "Check 'gitpkg' repository status")
	assert.NotContains(test, cmdOutput, "Fetch the 'gitpkg' repository")
	assert.Contains(test, cmdOutput, "Fetch local-builder configuration")
	// Check that the right files exist now
	assert.NoFileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-fetch-flags-project/chickenpox-flags-1/chickenpox-flags-1.spec")
	assert.NoFileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-fetch-flags-project/chickenpox-flags-1/chickenpox-flags-1-1.0.0.tar.gz")
	assert.NoFileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-fetch-flags-project/chickenpox-flags-1/.redpesk/autogen::service.json")
	assert.NoFileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-fetch-flags-project/chickenpox-flags-1/.rp-cli/md5sum.json")
	assert.NoFileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-fetch-flags-project/chickenpox-flags-1/.rp-cli/"+localBldCfgFile)
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-fetch-flags-project/chickenpox-flags-1/.rp-cli/local-fetch-flags-project.json")
	// --------------------------------

	// Force flag ---------------------
	cmdOutput = checkActionCmdAns(test, localCommand, localPkgCmdName, "fetch", "chickenpox-flags-0", "-f")
	// Check various things on the output string
	assert.NotContains(test, cmdOutput, "Check 'gitpkg' repository status")
	assert.Contains(test, cmdOutput, "Fetch the 'gitpkg' repository")
	assert.Contains(test, cmdOutput, "Fetch local-builder configuration")
	// --------------------------------
}

// TestLocalPackageFetchBranchFlag is the function allowing to test if
// the command "rp-cli local package fetch" is working correctly when
// the "--branch" flag is used
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalPackageFetchBranchFlag(test *testing.T) {
	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project and an application for our tests
	err = createFullPrjAppsBuildsTests(test, "local-fetch-branch-project", "redpesk-lts-arz", "wednesday", 1, 0, 0)
	if err != nil {
		test.Fatal(err)
	}

	// Create a new branch on this app
	_, err = createBranchOnApp(test, "local-fetch-branch-project", "wednesday-0", "main", "my-branch", false, false, []string{})
	if err != nil {
		test.Fatal(err)
	}

	// Ask rp-cli to fetch all the gitpkg files of the newly created app, on a given
	cmdOutput := checkActionCmdAns(test, localCommand, localPkgCmdName, "fetch", "wednesday-0", "--project", "local-fetch-branch-project", "-b", "my-branch")

	// Check various things on the output string
	assert.Contains(test, cmdOutput, "Retrieve application details")
	assert.Contains(test, cmdOutput, "Check local-builder configuration")
	assert.Contains(test, cmdOutput, "Create directories if needed")
	assert.Contains(test, cmdOutput, "Check 'gitpkg' repository status")
	assert.Contains(test, cmdOutput, "Fetch the 'gitpkg' repository from branch 'my-branch'")
	assert.Contains(test, cmdOutput, "Fetch local-builder configuration")

	// Check that the right files exist now
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-fetch-branch-project/wednesday-0/wednesday-0.spec")
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-fetch-branch-project/wednesday-0/wednesday-0-1.0.0.tar.gz")
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-fetch-branch-project/wednesday-0/.redpesk/autogen::service.json")
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-fetch-branch-project/wednesday-0/.rp-cli/local-fetch-branch-project.json")
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-fetch-branch-project/wednesday-0/.rp-cli/md5sum.json")
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-fetch-branch-project/wednesday-0/.rp-cli/"+localBldCfgFile)

	// Check that the 'local_build_config.json' contains the right branch
	pathToPkgDir := path.Join("/tmp/redpesk-cli-test/gitpkgs", "local-fetch-branch-project", "wednesday-0")
	localBldCfg, _, err := getLocalBuildConfig(pathToPkgDir, true)
	if err != nil {
		test.Fatal(err)
	}
	assert.Equal(test, "my-branch", localBldCfg.Branch)
}

// TestLocalPackageStatus is the function allowing to test if
// the command "rp-cli local package status" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalPackageStatus(test *testing.T) {
	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project and an application for our tests
	err = createFullPrjAppsBuildsTests(test, "local-status-project", "redpesk-lts-arz", "chickenpox-status", 1, 0, 0)
	if err != nil {
		test.Fatal(err)
	}

	// Ask rp-cli to fetch all the files of the newly created app
	checkActionCmdAns(test, localCommand, localPkgCmdName, "fetch", "chickenpox-status-0")

	// Try a first time to check the status of the "chickenpox-status-0" app
	outputCmd, err := executeCommand(localCommand, localPkgCmdName, "status", "chickenpox-status-0")
	if err != nil {
		test.Fatal(err)
	}

	// Check that the status is good
	assert.Contains(test, outputCmd, localStatusNoChanges)

	// Now, modify a file, delete another and add another one
	err = os.WriteFile("/tmp/redpesk-cli-test/gitpkgs/local-status-project/chickenpox-status-0/chickenpox-status-0.spec", []byte("MODIFIED FILE"), 0600)
	if err != nil {
		test.Fatal(err)
	}
	err = os.Remove("/tmp/redpesk-cli-test/gitpkgs/local-status-project/chickenpox-status-0/chickenpox-status-0-1.0.0.tar.gz")
	if err != nil {
		test.Fatal(err)
	}
	err = os.WriteFile("/tmp/redpesk-cli-test/gitpkgs/local-status-project/chickenpox-status-0/new-file.txt", []byte("NEW FILE"), 0600)
	if err != nil {
		test.Fatal(err)
	}

	// Try another time to check the status of the "chickenpox-status-0" app
	outputCmd, err = executeCommand(localCommand, localPkgCmdName, "status", "chickenpox-status-0")
	if err != nil {
		test.Fatal(err)
	}

	// Check that the status is good
	assert.Contains(test, outputCmd, "added:    new-file.txt")
	assert.Contains(test, outputCmd, "deleted:  chickenpox-status-0-1.0.0.tar.gz")
	assert.Contains(test, outputCmd, "modified: chickenpox-status-0.spec")
}

// TestLocalPackageStatusGitIgnore is the function allowing to test if
// the command "rp-cli local package status" is working correctly,
// when there is a .gitignore file in the repository
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalPackageStatusGitIgnore(test *testing.T) {

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project and an application for our tests
	err = createFullPrjAppsBuildsTests(test, "local-status-gitignore-project", "redpesk-lts-arz", "la-vie-status", 1, 0, 0)
	if err != nil {
		test.Fatal(err)
	}

	// Ask rp-cli to fetch all the files of the newly created app
	checkActionCmdAns(test, localCommand, localPkgCmdName, "fetch", "la-vie-status-0", "--project", "local-status-gitignore-project")

	// Try a first time to check the status of the "la-vie-status-0" app
	outputCmd, err := executeCommand(localCommand, localPkgCmdName, "status", "la-vie-status-0", "--project", "local-status-gitignore-project")
	if err != nil {
		test.Fatal(err)
	}

	// Check that the status is good
	assert.Contains(test, outputCmd, localStatusNoChanges)

	// Add a .gitignore file
	err = os.WriteFile("/tmp/redpesk-cli-test/gitpkgs/local-status-gitignore-project/la-vie-status-0/.gitignore", []byte("debug*"), 0600)
	if err != nil {
		test.Fatal(err)
	}

	// Add a file that normally should be ignore by the status command
	err = os.WriteFile("/tmp/redpesk-cli-test/gitpkgs/local-status-gitignore-project/la-vie-status-0/debug_for_test", []byte("I hope that I will be ignored"), 0600)
	if err != nil {
		test.Fatal(err)
	}

	// Try another time to check the status of the "la-vie-status-0" app
	outputCmd, err = executeCommand(localCommand, localPkgCmdName, "status", "la-vie-status-0", "--project", "local-status-gitignore-project")
	if err != nil {
		test.Fatal(err)
	}

	// Check that the status is good
	assert.Contains(test, outputCmd, "added:    .gitignore")
	assert.NotContains(test, outputCmd, "debug_for_test")
}

// TestLocalPackageDIff is the function allowing to test if
// the command "rp-cli local package diff" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalPackageDIff(test *testing.T) {
	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project and an application for our tests
	err = createFullPrjAppsBuildsTests(test, "local-diff-project", "redpesk-lts-arz", "grand-soleil-diff", 1, 0, 0)
	if err != nil {
		test.Fatal(err)
	}

	// Ask rp-cli to fetch all the files of the newly created app
	checkActionCmdAns(test, localCommand, localPkgCmdName, "fetch", "grand-soleil-diff-0")

	// Try a first time to check the diff of the "grand-soleil-diff-0" app
	outputCmd, err := executeCommand(localCommand, localPkgCmdName, "diff", "grand-soleil-diff-0")
	if err != nil {
		test.Fatal(err)
	}

	// Check that we got an empty string
	assert.Empty(test, outputCmd)

	// Now, modify a file, delete another and add another one
	err = os.WriteFile("/tmp/redpesk-cli-test/gitpkgs/local-diff-project/grand-soleil-diff-0/grand-soleil-diff-0.spec", []byte("MODIFIED FILE"), 0600)
	if err != nil {
		test.Fatal(err)
	}
	err = os.Remove("/tmp/redpesk-cli-test/gitpkgs/local-diff-project/grand-soleil-diff-0/grand-soleil-diff-0-1.0.0.tar.gz")
	if err != nil {
		test.Fatal(err)
	}
	err = os.WriteFile("/tmp/redpesk-cli-test/gitpkgs/local-diff-project/grand-soleil-diff-0/new-file.txt", []byte("NEW FILE"), 0600)
	if err != nil {
		test.Fatal(err)
	}

	// Now retrieve the diff given by rp-cli
	outputCmd, err = executeCommand(localCommand, localPkgCmdName, "diff", "grand-soleil-diff-0", "--no-color")
	if err != nil {
		test.Fatal(err)
	}

	// Check that the diff is working correctly
	assert.Contains(test, outputCmd, "--- /dev/null")
	assert.Contains(test, outputCmd, "+++ new-file.txt")
	assert.Contains(test, outputCmd, "+NEW FILE")
	assert.Contains(test, outputCmd, "+MODIFIED FILE")
	assert.Contains(test, outputCmd, "--- .rp-cli/origin-files/grand-soleil-diff-0-1.0.0.tar.gz")
	assert.Contains(test, outputCmd, "+++ /dev/null")
}

// TestLocalPackageCommit is the function allowing to test if
// the command "rp-cli local package commit" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalPackageCommit(test *testing.T) {
	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project and applications for our tests
	err = createFullPrjAppsBuildsTests(test, "local-commit-project", "redpesk-lts-arz", "chickenpox-commit", 1, 0, 0)
	if err != nil {
		test.Fatal(err)
	}

	// Ask rp-cli to fetch all the files of the newly created app
	checkActionCmdAns(test, localCommand, localPkgCmdName, "fetch", "chickenpox-commit-0")

	// Try to commit immediately
	outputCmd, err := executeCommand(localCommand, localPkgCmdName, "commit", "chickenpox-commit-0")
	if err != nil {
		test.Fatal(err)
	}

	// Check that we have nothing to commit
	assert.Contains(test, outputCmd, localCommitNothingToCommit)

	// Now, modify a file, delete another and add another one
	err = os.WriteFile("/tmp/redpesk-cli-test/gitpkgs/local-commit-project/chickenpox-commit-0/chickenpox-commit-0.spec", []byte("MODIFIED FILE"), 0600)
	if err != nil {
		test.Fatal(err)
	}
	err = os.Remove("/tmp/redpesk-cli-test/gitpkgs/local-commit-project/chickenpox-commit-0/chickenpox-commit-0-1.0.0.tar.gz")
	if err != nil {
		test.Fatal(err)
	}
	err = os.WriteFile("/tmp/redpesk-cli-test/gitpkgs/local-commit-project/chickenpox-commit-0/new-file.txt", []byte("NEW FILE"), 0600)
	if err != nil {
		test.Fatal(err)
	}

	// Now try to commit just one file
	outputCmd, err = executeCommand(localCommand, localPkgCmdName, "commit", "chickenpox-commit-0", "new-file.txt")
	if err != nil {
		test.Fatal(err)
	}

	// Check that the file has been committed
	assert.Contains(test, outputCmd, "Uploading new-file.txt")
	assert.NotContains(test, outputCmd, "Upload of 'chickenpox-commit-0.spec'")
	assert.NotContains(test, outputCmd, "Removal of 'chickenpox-commit-0-1.0.0.tar.gz'")

	// Then, commit the other files
	outputCmd, err = executeCommand(localCommand, localPkgCmdName, "commit", "chickenpox-commit-0")
	if err != nil {
		test.Fatal(err)
	}

	// Check that the files has been committed
	assert.NotContains(test, outputCmd, "Uploading new-file.txt")
	assert.Contains(test, outputCmd, "Uploading chickenpox-commit-0.spec")
	assert.Contains(test, outputCmd, "Removal of 'chickenpox-commit-0-1.0.0.tar.gz'")
}

// TestLocalPackageCommitOnBranch is the function allowing to test if
// the command "rp-cli local package commit" is working correctly when a
// branch has been specified by the user at fetch time
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalPackageCommitOnBranch(test *testing.T) {
	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project and an application for our tests
	err = createFullPrjAppsBuildsTests(test, "local-commit-branch-project", "redpesk-lts-arz", "friday", 1, 0, 0)
	if err != nil {
		test.Fatal(err)
	}

	// Create a new branch on this app
	_, err = createBranchOnApp(test, "local-commit-branch-project", "friday-0", "main", "my-branch", false, false, []string{})
	if err != nil {
		test.Fatal(err)
	}

	// Ask rp-cli to fetch all the gitpkg files of the newly created app, on a given
	checkActionCmdAns(test, localCommand, localPkgCmdName, "fetch", "friday-0", "--project", "local-commit-branch-project", "-b", "my-branch")

	// Now, modify a file
	err = os.WriteFile("/tmp/redpesk-cli-test/gitpkgs/local-commit-branch-project/friday-0/friday-0.spec", []byte("MODIFIED FILE"), 0600)
	if err != nil {
		test.Fatal(err)
	}

	// Then, commit the other files
	outputCmd, err := executeCommand(localCommand, localPkgCmdName, "commit", "friday-0")
	if err != nil {
		test.Fatal(err)
	}

	// Check that the file has been committed
	assert.Contains(test, outputCmd, "Uploading friday-0.spec")
}

// TestLocalPackageInit is the function allowing to test if
// the command "rp-cli local package init" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalPackageInit(test *testing.T) {
	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project and application for our tests
	err = createFullPrjAppsBuildsTests(test, "local-init-project", "redpesk-lts-arz", "lily-pad", 1, 0, 0)
	if err != nil {
		test.Fatal(err)
	}

	// Ask rp-cli to run all the initialization to build locally this app
	cmdOutput, err := execCmdInSimTerm("", localCommand, localPkgCmdName, "init", "lily-pad-0", "--project", "local-init-project", "-b", "main")
	if err != nil {
		test.Fatal(err)
	}
	cmdOutput = strings.TrimSpace(cmdOutput)

	// Check the output of the command
	checkPkgInitCmd(test, cmdOutput, "lily-pad-0", "main")

	// Check that the right files have been set
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-init-project/lily-pad-0/lily-pad-0.spec")
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-init-project/lily-pad-0/lily-pad-0-1.0.0.tar.gz")
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-init-project/lily-pad-0/.redpesk/autogen::service.json")
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-init-project/lily-pad-0/.rp-cli/local-init-project.json")
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-init-project/lily-pad-0/.rp-cli/md5sum.json")
	assert.FileExists(test, "/tmp/redpesk-cli-test/gitpkgs/local-init-project/lily-pad-0/.rp-cli/"+localBldCfgFile)

	// Check that the 'local_build_config.json' contains the right values
	pathToPkgDir := path.Join("/tmp/redpesk-cli-test/gitpkgs", "local-init-project", "lily-pad-0")
	localBldCfg, _, err := getLocalBuildConfig(pathToPkgDir, true)
	if err != nil {
		test.Fatal(err)
	}
	assert.Equal(test, "/tmp/redpesk-cli-test/gitsources", localBldCfg.SourcesDir)
	assert.Equal(test, "/tmp/redpesk-cli-test/gitpkgs", localBldCfg.PackagesDir)
	assert.Equal(test, "main", localBldCfg.Branch)
}

// TestLocalPackageInitSrcDirAlreadyExists is the function allowing to test if
// the command "rp-cli local package init" is working correctly when
// the sources directory already exists before initialization
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalPackageInitSrcDirAlreadyExists(test *testing.T) {
	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project and application for our tests
	err = createFullPrjAppsBuildsTests(test, "local-init-dir-exists-project", "redpesk-lts-arz", "dir-exists-app", 1, 0, 0)
	if err != nil {
		test.Fatal(err)
	}

	// Create empty dir corresponding to project and app
	srcDir := path.Join("/tmp/redpesk-cli-test/gitsources", "local-init-dir-exists-project", "dir-exists-app-0")
	utils.CreateDirIfNotExists(srcDir, true)
	// Create a fake file
	utils.WriteToFile(path.Join(srcDir, "fake_file"), "Tchou Tchou!")

	// Ask rp-cli to run all the initialization to build locally this app
	cmdOutput, err := execCmdInSimTerm("", localCommand, localPkgCmdName, "init", "dir-exists-app-0", "--project", "local-init-dir-exists-project")
	if err != nil {
		test.Fatal(err)
	}
	cmdOutput = strings.TrimSpace(cmdOutput)

	// Check the output of the command
	checkPkgInitCmd(test, cmdOutput, "dir-exists-app-0", "main")

	// Check that we got a warning
	assert.Contains(test, cmdOutput, "The source directory may not be a git repository")
}

// TestLocalPackageService is the function allowing to test if
// the command "rp-cli local package service" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalPackageService(test *testing.T) {
	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project and application for our tests
	err = createFullPrjAppsBuildsTests(test, "local-service-project", "redpesk-lts-arz", "service-app", 1, 0, 0)
	if err != nil {
		test.Fatal(err)
	}

	// Ask rp-cli to run all the initialization to build locally this app
	_, err = execCmdInSimTerm("", localCommand, localPkgCmdName, "init", "service-app-0", "--project", "local-service-project")
	if err != nil {
		test.Fatal(err)
	}

	// Ask to run a specific service
	cmdOutput, err := execCmdInSimTerm("", localCommand, localPkgCmdName, "service", "service-app-0", "--project", "local-service-project", "excludearch", "--arches", "[x86_64]")
	if err != nil {
		test.Fatal(err)
	}

	// Check that every [OK]/[KO] display is at [OK]
	checkLineEndWith(test, cmdOutput, "Retrieve application details...", "[OK]")
	checkLineEndWith(test, cmdOutput, "Check local-builder configuration...", "[OK]")

	// Check the local-builder logs tags
	assert.Contains(test, cmdOutput, localLocalBldLogsTag1)
	assert.Contains(test, cmdOutput, localLocalBldLogsTag2)
}

// TestLocalPackageServiceHelpFlag is the function allowing to test if
// the command "rp-cli local package service --help" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalPackageServiceHelpFlag(test *testing.T) {
	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Ask to run a specific service
	cmdOutput, err := execCmdInSimTerm("", localCommand, localPkgCmdName, "service", "--help")
	if err != nil {
		test.Fatal(err)
	}

	// Check the local-builder logs tags
	assert.Contains(test, cmdOutput, localLocalBldLogsTag1)
	assert.Contains(test, cmdOutput, localLocalBldLogsTag2)

	// Extract the logs from mocked repdesk-lad
	localBuilderLogs := extractSubLogsWithFlag(cmdOutput, "LOCAL-BUILDER LOGS")

	// Check that the first line and the last line of the help command is here
	assert.Contains(test, localBuilderLogs, "usage: iotpkg service [-h] [--no-service]")
	assert.Contains(test, localBuilderLogs, "go_dependency       Fetch go dependencies")
}

// TestLocalPackageMetadata is the function allowing to test if
// the command "rp-cli local package metadata" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalPackageMetadata(test *testing.T) {
	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project and application for our tests
	err = createFullPrjAppsBuildsTests(test, "local-metadata-project", "redpesk-lts-arz", "metadata-app", 1, 0, 0)
	if err != nil {
		test.Fatal(err)
	}

	// Ask rp-cli to run all the initialization to build locally this app
	_, err = execCmdInSimTerm("", localCommand, localPkgCmdName, "init", "metadata-app-0", "--project", "local-metadata-project")
	if err != nil {
		test.Fatal(err)
	}

	// Ask to run a specific service
	cmdOutput, err := execCmdInSimTerm("", localCommand, localPkgCmdName, "metadata", "metadata-app-0", "--project", "local-metadata-project", "--revision", "1.0")
	if err != nil {
		test.Fatal(err)
	}

	// Check that every [OK]/[KO] display is at [OK]
	checkLineEndWith(test, cmdOutput, "Retrieve application details...", "[OK]")
	checkLineEndWith(test, cmdOutput, "Check local-builder configuration...", "[OK]")

	// Check the local-builder logs tags
	assert.Contains(test, cmdOutput, localLocalBldLogsTag1)
	assert.Contains(test, cmdOutput, localLocalBldLogsTag2)
}

// TestLocalPackageBuild is the function allowing to test if
// the command "rp-cli local package build" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalPackageBuild(test *testing.T) {
	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project and application for our tests
	err = createFullPrjAppsBuildsTests(test, "local-build-project", "redpesk-lts-arz", "build-app", 1, 0, 0)
	if err != nil {
		test.Fatal(err)
	}

	// Ask rp-cli to run all the initialization to build locally this app
	_, err = execCmdInSimTerm("", localCommand, localPkgCmdName, "init", "build-app-0", "--project", "local-build-project")
	if err != nil {
		test.Fatal(err)
	}

	// Ask to build the application
	cmdOutput, err := execCmdInSimTerm("", localCommand, localPkgCmdName, "build", "build-app-0", "--project", "local-build-project", "--arch", "x86_64", "--distro", "redpesk-lts-arz")
	if err != nil {
		test.Fatal(err)
	}

	// Check that every [OK]/[KO] display is at [OK]
	checkLineEndWith(test, cmdOutput, "Retrieve application details...", "[OK]")
	checkLineEndWith(test, cmdOutput, "Check local-builder configuration...", "[OK]")

	// Check that everything went well
	assert.Contains(test, cmdOutput, localBuildRunCmdAnsResultGood)
	assert.Contains(test, cmdOutput, localBuildRunCmdAnsResultIn)
}

// TestLocalPackageBuildOffline is the function allowing to test if
// the command "rp-cli local package build" is working correctly
// when the flag "--offset" is given
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalPackageBuildOffline(test *testing.T) {
	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project and application for our tests
	err = createFullPrjAppsBuildsTests(test, "local-build-offline-project", "redpesk-lts-arz", "build-offline-app", 1, 0, 0)
	if err != nil {
		test.Fatal(err)
	}

	// Ask rp-cli to run all the initialization to build locally this app
	_, err = execCmdInSimTerm("", localCommand, localPkgCmdName, "init", "build-offline-app-0", "--project", "local-build-offline-project")
	if err != nil {
		test.Fatal(err)
	}

	// Ask to build the application
	cmdOutput, err := execCmdInSimTerm("", localCommand, localPkgCmdName, "build", "build-offline-app-0", "--project", "local-build-offline-project", "--arch", "x86_64", "--offline")
	if err != nil {
		test.Fatal(err)
	}

	// Check that every [OK]/[KO] display is at [OK]
	checkLineEndWith(test, cmdOutput, "Retrieve application details...", "[OK]")
	checkLineEndWith(test, cmdOutput, "Check local-builder configuration...", "[OK]")

	// Check that everything went well
	assert.Contains(test, cmdOutput, localBuildRunCmdAnsResultGood)
	assert.Contains(test, cmdOutput, localBuildRunCmdAnsResultIn)

	// Check that the offline flag has been correctly passed
	assert.True(test, rpladmock.LatestBuildOffline)
}

// TestLocalPackageBuildWithoutAppSlug is the function allowing to test if
// the command "rp-cli local package build" is working correctly, when the
// user has not given any application slug in the command line
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalPackageBuildWithoutAppSlug(test *testing.T) {
	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project and application for our tests
	err = createFullPrjAppsBuildsTests(test, "local-build-project", "redpesk-lts-arz", "build-app-no-slug", 1, 0, 0)
	if err != nil {
		test.Fatal(err)
	}

	// Ask rp-cli to run all the initialization to build locally this app
	_, err = execCmdInSimTerm("", localCommand, localPkgCmdName, "init", "build-app-no-slug-0", "--project", "local-build-project")
	if err != nil {
		test.Fatal(err)
	}

	// Now, goes into the right directory
	// Save current directory
	oldCwd, err := os.Getwd()
	if err != nil {
		test.Fatal(err)
	}
	// Defer changing the directory to the current one
	defer os.Chdir(oldCwd)
	err = os.Chdir("/tmp/redpesk-cli-test/gitpkgs/local-build-project/build-app-no-slug-0")
	if err != nil {
		test.Fatal(err)
	}

	// Ask to build the application
	cmdOutput, err := execCmdInSimTerm("", localCommand, localPkgCmdName, "build", "--arch", "x86_64", "--distro", "redpesk-lts-arz")
	if err != nil {
		test.Fatal(err)
	}

	// Check that every [OK]/[KO] display is at [OK]
	checkLineEndWith(test, cmdOutput, "Retrieve application details...", "[OK]")
	checkLineEndWith(test, cmdOutput, "Check local-builder configuration...", "[OK]")

	// Check that everything went well
	assert.Contains(test, cmdOutput, localBuildRunCmdAnsResultGood)
	assert.Contains(test, cmdOutput, localBuildRunCmdAnsResultIn)

	// Try the same thing but in the 'gitsources/' directory
	err = os.Chdir("/tmp/redpesk-cli-test/gitsources/local-build-project/build-app-no-slug-0")
	if err != nil {
		test.Fatal(err)
	}

	// Ask to build the application
	cmdOutput, err = execCmdInSimTerm("", localCommand, localPkgCmdName, "build", "--arch", "aarch64")
	if err != nil {
		test.Fatal(err)
	}

	// Check that every [OK]/[KO] display is at [OK]
	checkLineEndWith(test, cmdOutput, "Retrieve application details...", "[OK]")
	checkLineEndWith(test, cmdOutput, "Check local-builder configuration...", "[OK]")

	// Check that everything went well
	assert.Contains(test, cmdOutput, localBuildRunCmdAnsResultGood)
	assert.Contains(test, cmdOutput, localBuildRunCmdAnsResultIn)
}

// TestLocalPackageShell is the function allowing to test if
// the command "rp-cli local package shell" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalPackageShell(test *testing.T) {
	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project and application for our tests
	err = createFullPrjAppsBuildsTests(test, "local-shell-project", "redpesk-lts-arz", "shell-app", 1, 0, 0)
	if err != nil {
		test.Fatal(err)
	}

	// Ask rp-cli to run all the initialization to build locally this app
	_, err = execCmdInSimTerm("", localCommand, localPkgCmdName, "init", "shell-app-0", "--project", "local-shell-project")
	if err != nil {
		test.Fatal(err)
	}

	// Ask to get a shell
	cmdOutput, err := execCmdInSimTerm("", localCommand, localPkgCmdName, "shell", "shell-app-0", "--project", "local-shell-project", "--arch", "aarch64", "--distro", "redpesk-lts-arz")
	if err != nil {
		test.Fatal(err)
	}

	// Check that every [OK]/[KO] display is at [OK]
	checkLineEndWith(test, cmdOutput, "Retrieve application details...", "[OK]")
	checkLineEndWith(test, cmdOutput, "Check local-builder configuration...", "[OK]")

	// Check that we got a pseudo prompt
	assert.Contains(test, cmdOutput, "devel@rp-localbuilder")
}

// TestLocalPackageShellOffline is the function allowing to test if
// the command "rp-cli local package shell" is working correctly
// when the flag "--offset" is given
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalPackageShellOffline(test *testing.T) {
	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project and application for our tests
	err = createFullPrjAppsBuildsTests(test, "local-shell-project", "redpesk-lts-arz", "shell-app", 1, 0, 0)
	if err != nil {
		test.Fatal(err)
	}

	// Ask rp-cli to run all the initialization to build locally this app
	_, err = execCmdInSimTerm("", localCommand, localPkgCmdName, "init", "shell-app-0", "--project", "local-shell-project")
	if err != nil {
		test.Fatal(err)
	}

	// Ask to get a shell
	cmdOutput, err := execCmdInSimTerm("", localCommand, localPkgCmdName, "shell", "shell-app-0", "--project", "local-shell-project", "--arch", "aarch64", "--offline")
	if err != nil {
		test.Fatal(err)
	}

	// Check that every [OK]/[KO] display is at [OK]
	checkLineEndWith(test, cmdOutput, "Retrieve application details...", "[OK]")
	checkLineEndWith(test, cmdOutput, "Check local-builder configuration...", "[OK]")

	// Check that we got a pseudo prompt
	assert.Contains(test, cmdOutput, "devel@rp-localbuilder")

	// Check that the offline flag has been correctly passed
	assert.True(test, rpladmock.LatestBuildOffline)
}

// TestLocalPackageClean is the function allowing to test if
// the command "rp-cli local package clean" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalPackageClean(test *testing.T) {
	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project and application for our tests
	err = createFullPrjAppsBuildsTests(test, "local-clean-project", "redpesk-lts-arz", "clean-app", 1, 0, 0)
	if err != nil {
		test.Fatal(err)
	}

	// Ask rp-cli to run all the initialization to build locally this app
	_, err = execCmdInSimTerm("", localCommand, localPkgCmdName, "init", "clean-app-0", "--project", "local-clean-project")
	if err != nil {
		test.Fatal(err)
	}

	// Ask to build the application
	_, err = execCmdInSimTerm("", localCommand, localPkgCmdName, "build", "clean-app-0", "--project", "local-clean-project")
	if err != nil {
		test.Fatal(err)
	}

	// Check that the right directories exist
	assert.DirExists(test, path.Join(gitPackagesTestDir, "local-clean-project", "clean-app-0", ".iotpkg"))
	assert.FileExists(test, path.Join(gitPackagesTestDir, "local-clean-project", "clean-app-0", ".iotpkg", "clean-app-0"+"-redpesk-distro-aarch64-cross.cfg"))
	assert.FileExists(test, path.Join(gitPackagesTestDir, "clean-app-0"+"-redpesk-distro-aarch64-cross.cfg"))

	// Try to clean
	cmdOutput, err := execCmdInSimTerm("", localCommand, localPkgCmdName, "clean", "clean-app-0")
	if err != nil {
		test.Fatal(err)
	}

	// Check the logs
	assert.Contains(test, cmdOutput, localPkgCleanCmdAnsLine2)
	assert.Contains(test, cmdOutput, "Start: clean chroot")
	assert.Contains(test, cmdOutput, "Finish: clean chroot")

	// Check that the directories have been deleted
	assert.NoDirExists(test, path.Join(gitPackagesTestDir, "local-clean-project"))
	assert.NoFileExists(test, path.Join(gitPackagesTestDir, "clean-app-0"+"-redpesk-distro-aarch64-cross.cfg"))
}

// TestLocalPackageCleanInit is the function allowing to test if
// the command "rp-cli local package clean --init" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalPackageCleanInit(test *testing.T) {
	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project and application for our tests
	err = createFullPrjAppsBuildsTests(test, "local-clean-project-init", "redpesk-lts-arz", "clean-app-init", 1, 0, 0)
	if err != nil {
		test.Fatal(err)
	}

	// Ask rp-cli to run all the initialization to build locally this app
	_, err = execCmdInSimTerm("", localCommand, localPkgCmdName, "init", "clean-app-init-0", "--project", "local-clean-project-init")
	if err != nil {
		test.Fatal(err)
	}

	// Ask to build the application
	_, err = execCmdInSimTerm("", localCommand, localPkgCmdName, "build", "clean-app-init-0", "--project", "local-clean-project-init")
	if err != nil {
		test.Fatal(err)
	}

	// Check that the right directories exist
	assert.DirExists(test, path.Join(gitPackagesTestDir, "local-clean-project-init", "clean-app-init-0", ".iotpkg"))
	assert.FileExists(test, path.Join(gitPackagesTestDir, "local-clean-project-init", "clean-app-init-0", ".iotpkg", "clean-app-init-0"+"-redpesk-distro-aarch64-cross.cfg"))
	assert.FileExists(test, path.Join(gitPackagesTestDir, "clean-app-init-0"+"-redpesk-distro-aarch64-cross.cfg"))

	// Try to clean
	cmdOutput, err := execCmdInSimTerm("", localCommand, localPkgCmdName, "clean", "clean-app-init-0", "--init")
	if err != nil {
		test.Fatal(err)
	}

	// Check the logs
	assert.Contains(test, cmdOutput, localPkgCleanCmdAnsLine2)
	assert.Contains(test, cmdOutput, "Start: clean chroot")
	assert.Contains(test, cmdOutput, "Finish: clean chroot")

	// Extract the "local package init" part
	outputTab := strings.Split(cmdOutput, localPkgInitCmdAnsLine0)
	if assert.Equal(test, 2, len(outputTab)) {
		checkPkgInitCmd(test, outputTab[1], "clean-app-init-0", "main")
	}
}

// TestLocalPkgCleanSrcModified is the function allowing to test if
// the command "rp-cli local package clean" is working correctly
// even when the sources have been modified
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalPkgCleanSrcModified(test *testing.T) {
	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project and an application for our tests
	err = createFullPrjAppsBuildsTests(test, "local-pkg-clean-src-prj-0", "redpesk-lts-arz", "local-pkg-clean-src-app", 1, 0, 0)
	if err != nil {
		test.Fatal(err)
	}

	// Ask rp-cli to run all the initialization to build locally this app
	_, err = execCmdInSimTerm("", localCommand, localPkgCmdName, "init", "local-pkg-clean-src-app-0", "--project", "local-pkg-clean-src-prj-0")
	if err != nil {
		test.Fatal(err)
	}

	// Now clone a simple project in one of the created source repository
	prjSrcDirectory := path.Join(gitSourcesTestDir, "local-pkg-clean-src-prj-0")
	_, err = utils.RunCmdInDirectory(prjSrcDirectory, "git", "clone", package4TestsHelloURL, "local-pkg-clean-src-app-0")
	if err != nil {
		test.Fatal(err)
	}

	// Modify something in the just cloned repo
	appSrcDirectory := path.Join(prjSrcDirectory, "local-pkg-clean-src-app-0")
	_, err = utils.RunCmdInDirectory(appSrcDirectory, "touch", "modification.txt")
	if err != nil {
		test.Fatal(err)
	}

	// Now try to delete the directory

	// Create a pipe
	readPipe, writePipe, err := os.Pipe()
	if err != nil {
		test.Fatal(err)
	}

	// Save the original *os.File
	origStdin := os.Stdin

	// Set os.Stdin with our fake one
	os.Stdin = readPipe

	// Start a go routine that will answer the request of the command rp-cli local package clean
	go func() {
		// Wait a bit, in order for the command to be executed
		time.Sleep(100 * time.Millisecond)

		// Say that we still want to delete the directory
		writePipe.Write([]byte("y\n"))
	}()

	// Run the command
	output, err := executeCommand(localCommand, localPkgCmdName, "clean", "local-pkg-clean-src-app-0")

	// Restore Stdin
	os.Stdin = origStdin

	if err != nil {
		test.Fatal(err)
	}

	// Check that the output contains a request from rp-cli
	assert.Contains(test, output, localPkgCleanCmdAnsLine1)
	assert.Contains(test, output, localPkgCleanCmdAnsLine2)

	// Check that the directory does not exist anymore
	assert.NoDirExists(test, path.Join(gitPackagesTestDir, "local-pkg-clean-src-prj-0", "local-pkg-clean-src-app-0"))

	// Empty test "gitpkgs/" and "gitsources/" directories
	err = emptyLocalTestDirs()
	if err != nil {
		test.Fatal(err)
	}
}

// TestLocalPkgCleanForce is the function allowing to test if
// the command "rp-cli local package clean -f" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalPkgCleanForce(test *testing.T) {
	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a project and an application for our tests
	err = createFullPrjAppsBuildsTests(test, "local-pkg-clean-force-prj-0", "redpesk-lts-arz", "local-pkg-clean-force-app", 1, 0, 0)
	if err != nil {
		test.Fatal(err)
	}

	// Ask rp-cli to run all the initialization to build locally this app
	_, err = execCmdInSimTerm("", localCommand, localPkgCmdName, "init", "local-pkg-clean-force-app-0", "--project", "local-pkg-clean-force-prj-0")
	if err != nil {
		test.Fatal(err)
	}

	// Now clone a simple project in one of the created source repository
	prjSrcDirectory := path.Join(gitSourcesTestDir, "local-pkg-clean-force-prj-0")
	_, err = utils.RunCmdInDirectory(prjSrcDirectory, "git", "clone", package4TestsHelloURL, "local-pkg-clean-force-app-0")
	if err != nil {
		test.Fatal(err)
	}

	// Modify something in the just cloned repo
	appSrcDirectory := path.Join(prjSrcDirectory, "local-pkg-clean-force-app-0")
	_, err = utils.RunCmdInDirectory(appSrcDirectory, "touch", "modification.txt")
	if err != nil {
		test.Fatal(err)
	}

	// Now try to delete the directory

	// Run the command
	output, err := executeCommand(localCommand, localPkgCmdName, "clean", "local-pkg-clean-force-app-0", "--force")
	if err != nil {
		test.Fatal(err)
	}

	// Check that the output contains a request from rp-cli
	assert.Contains(test, output, localPkgCleanCmdAnsLine2)

	// Check that the directory does not exist anymore
	assert.NoDirExists(test, path.Join(gitPackagesTestDir, "local-pkg-clean-force-prj-0", "local-pkg-clean-force-app-0"))

	// Empty test "gitpkgs/" and "gitsources/" directories
	err = emptyLocalTestDirs()
	if err != nil {
		test.Fatal(err)
	}
}

// TestLocalCheckSetup is the function allowing to test if
// the command "rp-cli local check-setup" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestLocalCheckSetup(test *testing.T) {
	// Set-up what is needed for redpesk-lad
	setupRPLADTests()
	// Defer the teardown function for redpesk-lad
	defer teardownRPLADTests()

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Run the update command and check that all is [OK]
	checkActionCmdAns(test, localCommand, localCheckSetupCmd)
}

// --------------------------------------------------------------
// Utilities functions ------------------------------------------
// --------------------------------------------------------------

// emptyLocalTestDirs allows to empty the directories necessary
// to test the "local" subcommand (gitpkgs and gitsources)
//
// :returns: error -- an error if a problem has occurred
func emptyLocalTestDirs() error {
	// Empty the gitsources directory
	err := utils.RemoveDirContents(gitSourcesTestDir)
	if err != nil {
		return err
	}

	// Empty the gitpkgs directory
	err = utils.RemoveDirContents(gitPackagesTestDir)
	if err != nil {
		return err
	}

	return nil
}

// createFalseLocalBuilderFile allows to create a false local-builder file while
// saving the current one thanks to a backup
//
// :returns: string, error -- The path to the file and an error if something went wrong
func createFalseLocalBuilderFile() (string, error) {
	// Get path to the configuration directory
	configDirPath, err := getConfigDir()
	if err != nil {
		return "", err
	}
	localBuilderFilePath := path.Join(configDirPath, localBuilderFile)

	// Convert the connections slice to json string
	connectsJsonBytes, err := json.MarshalIndent([]rpconnect.RPCLIConnection{}, "", "  ")
	if err != nil {
		return "", err
	}

	err = createFalseFile(localBuilderFilePath, string(connectsJsonBytes))
	if err != nil {
		return "", err
	}

	return localBuilderFilePath, nil
}

// addLocalConnectionForTest is a utility function allowing to create a
// new connection and to create a fake local-builder file if it does not
// exist already.
// It returns a path to the fake file created. The calling
// function needs to call 'defer deleteFalseFile(fakeFilePath)'
// in order to delete it at the end of tests
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param connection: Connection to add to rp-cli
// :type connection: *rpconnect.RPCLIConnection
// :returns: string, error -- the path to the fake token file and an error if something bad happened
func addLocalConnectionForTest(test *testing.T, connection *rpconnect.RPCLIConnection) (string, error) {
	// Check if the local-builder file is already existing, and create it if it is not the case
	lbFileExists, lbFilePath, err := fakeFileExists(localBuilderFile)
	if err != nil {
		return lbFilePath, err
	}
	if !lbFileExists {
		lbFilePath, err = createFalseLocalBuilderFile()
		if err != nil {
			return lbFilePath, err
		}
	}

	// Now add the connection object if it does not already exists
	addCmd := []string{localCommand, connectionsCommand, "add", connection.Alias, "-i", connection.ServerURL}

	err = addObjectIfNotExists(test, localCommand+" "+connectionsCommand, connection.Alias, addCmd...)
	if err != nil {
		return lbFilePath, err
	}

	return lbFilePath, nil
}

// checkPkgInitCmd is the function allowing to check the output of
// the command "rp-cli local package init ..."
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param cmdOutput : Output from the rp-cli local package init
// :type cmdOutput: string
func checkPkgInitCmd(test *testing.T, cmdOutput, appName, branchName string) {

	// Check that the directories used are the one we used for tests
	checkLineEndWith(test, cmdOutput, "\"gitsources/\" directory used:", gitSourcesTestDir)
	checkLineEndWith(test, cmdOutput, "\"gitpkgs/\" directory used:", gitPackagesTestDir)

	// Check that every [OK]/[KO] display is at [OK]
	checkLineEndWith(test, cmdOutput, "Retrieve application details...", "[OK]")
	checkLineEndWith(test, cmdOutput, "Check local-builder configuration...", "[OK]")

	checkLineEndWith(test, cmdOutput, fmt.Sprintf(localPkgInitCmdAnsLine3, appName), "[OK]")
	checkLineEndWith(test, cmdOutput, "Create directories if needed", "[OK]")
	checkLineEndWith(test, cmdOutput, "Check 'gitpkg' repository status", "[OK]")
	checkLineEndWith(test, cmdOutput, fmt.Sprintf("Fetch the 'gitpkg' repository from branch '%s'", branchName), "[OK]")
	checkLineEndWith(test, cmdOutput, "Fetch local-builder configuration", "[OK]")

	// Extract the clone logs
	cloneLogs := extractSubLogsWithFlag(cmdOutput, "LOCAL-BUILDER LOGS")

	// Check that there are some clone logs and that they look coherent
	assert.NotEqual(test, "", cloneLogs)
	assert.Contains(test, cloneLogs, "Successfull shallow cloned")
}

// createFakeBuildConfig is the function allowing to create a fake
// build configuration at the right location from the configuration
// structure
//
// :param config: Configuration to write
// :type config: RPLocalBuildConfig
// :returns: error -- an error if something went wrong, nil in the other case
func createFakeBuildConfig(config RPLocalBuildConfig) error {
	// Create the path to the directory to create
	configDirPath := path.Join(gitPackagesTestDir, config.ProjectSlug, config.AppPkgName, localCfgSubDirectory)

	// Create the directory
	if err := utils.CreateDirIfNotExists(configDirPath, true); err != nil {
		return err
	}

	// Now, creates the file that will contains the configuration
	return utils.WriteJsonFile(path.Join(configDirPath, localBldCfgFile), config)
}
