/*
Copyright © 2020-2023 IoT.bzh Company
Contact: https://www.iot.bzh/licensing

This file is part of the redpesk-cli project, the redpesk command line tool.

$RP_BEGIN_LICENSE$
Commercial License Usage

Licensees holding valid commercial IoT.bzh licenses may use this file in
accordance with the commercial license agreement provided with the
Software or, alternatively, in accordance with the terms contained in
a written agreement between you and The IoT.bzh Company. For licensing terms
and conditions see https://www.iot.bzh/terms-conditions. For further
information use the contact form at https://www.iot.bzh/contact.

# GNU General Public License Usage

Alternatively, this file may be used under the terms of the GNU General
Public license version 3. This license is as published by the Free Software
Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
of this file. Please review the following information to ensure the GNU
General Public License requirements will be met
https://www.gnu.org/licenses/gpl-3.0.html.

$RP_END_LICENSE$
*/
package cmd

import (
	"fmt"
	"io"
	"os"
	"path"
	"path/filepath"
	"regexp"
	"strings"
	"syscall"
	"testing"
	"time"

	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/rpbckmock"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/rpladmock"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/utils"
	"github.com/stretchr/testify/assert"

	expect "github.com/Netflix/go-expect"
)

// Variable containing the mocked backend
var MockedRPBackend *rpbckmock.RPMockedBackend

// Variable containing the mocked redpesk-lad server
var MockedRPLAD *rpladmock.RPMockedLAD

// Variable allowing to know if we are running with a true
// backend or a mocked one
var bckIsReal = false

// Simulated console used to run command that normally needs to be run
// in a real terminal
var simulatedConsole *expect.Console

// List of files to delete at teardown
var filesListToDel = []string{}

// TestMain is the main test function invocated when running
// "go test" it allows especially to have setup and teardown
// functions
//
// :param cmd: Pointer to M testing object
// :type m: *testing.M
func TestMain(m *testing.M) {
	if err := setupTests(); err != nil {
		os.Exit(1)
	} else {
		code := m.Run()
		tearDownTests(code)
		os.Exit(code)
	}
}

// setupTests inits everything that is needed for the tests
// run for rp-cli
//
// :returns: error -- an error if a problem has occurred
func setupTests() error {
	var err error

	// Tell the application that we are running in test mode
	runMode = "testing"

	// Get the Environment variable named BCK_URL
	testsBckURL = os.Getenv("BCK_URL")

	// Make sure the use specific directory for testing to avoid to overwrite user config
	cfgDirPath = cfgDirPathTestDir

	// Create test "gitpkgs/" and "gitsources/" directories if there are not existing
	err = createLocalTestDirs()
	if err != nil {
		return err
	}

	// Create a configuration directory if not existing
	configDirs, err := getConfigDir()
	if err != nil {
		return err
	}
	err = utils.CreateDirIfNotExists(configDirs, false)
	if err != nil {
		return err
	}

	// Create an export directory if not existing
	err = utils.CreateDirIfNotExists(exportImportProjectsDirectory, false)
	if err != nil {
		return err
	}

	if testsBckURL == "" {
		// Start the mocked Redpesk backend
		MockedRPBackend, err = rpbckmock.StartMockedRPBackend(Log)
		if err != nil {
			return err
		}
	} else {
		bckIsReal = true
	}

	// The redpesk-lad server is not started here, it is started uniquely when needed
	// (Attempt to correct IO Wait problems seen in Jenkins)

	return nil
}

// tearDownTests terminates everything that is needed to terminate
// before ending the tests
//
// :returns: error -- an error if a problem has occurred
func tearDownTests(code int) error {
	if testsBckURL == "" {
		// Stop the mocked Redpesk backend
		if err := MockedRPBackend.ShutdownServer(); err != nil {
			return err
		}
	}

	// Keep logs and temporary files to allow analysis when tests fail
	if code == 0 {

		// Remove "redpesk-cli-test" directory
		err := os.RemoveAll("/tmp/redpesk-cli-test")
		if err != nil {
			return err
		}

		// Remove the rp-tests-logs directory
		err = os.RemoveAll(testsLogsDirectory)
		if err != nil {
			return err
		}

		// Remove the rp-application-builds-logs directory
		err = os.RemoveAll(buildsLogsDirectory)
		if err != nil {
			return err
		}

		// Remove the rp-cli-export-project directory
		err = os.RemoveAll(exportImportProjectsDirectory)
		if err != nil {
			return err
		}

		// Remove all files that needs to be deleted
		for _, file := range filesListToDel {
			os.Remove(file)
		}
	}

	// Close the simulated console if running
	if simulatedConsole != nil {
		simulatedConsole.Close()
		simulatedConsole = nil
	}

	return nil
}

// setupRPLADTests inits everything that is needed when the tests
// run needs the redpesk-lad mock to be started
func setupRPLADTests() {
	// Start the mocked redpesk-lad server
	MockedRPLAD, _ = rpladmock.StartMockedRPLad(Log)
}

// teardownRPLADTests stops everything that was needed when the tests
// run needs the redpesk-lad mock to be started
func teardownRPLADTests() {
	// Stop the mocked redpesk-lad server
	MockedRPLAD.ShutdownServer()
}

// TestOnBoardBackend is the function allowing to test if
// the command "rp-cli onboard" is working correctly for the
// backend part
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestOnBoardBackend(test *testing.T) {

	// We will have to fake the stdin in this situation (because this command is interactive)
	// See https://eli.thegreenplace.net/2020/faking-stdin-and-stdout-in-go/ for more info

	// Create a false token file
	tokenFilePath, err := createFalseTokenFile(nil)
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false token file at the end of the function
	defer deleteFalseFile(tokenFilePath)

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Create a pipe
	readPipe, writePipe, err := os.Pipe()
	if err != nil {
		test.Fatal(err)
	}

	// Save the original *os.File
	origStdin := os.Stdin

	// Set os.Stdin with our fake one
	os.Stdin = readPipe

	// Start a go routine that will fill what the command 'rp-cli onboard' wants
	go func() {
		// Wait a bit, in order for the command to be executed
		time.Sleep(250 * time.Millisecond)

		// Say that we want to configure the backend connection
		writePipe.Write([]byte("Y\n"))

		// Wait a bit, in order for the command to be executed
		time.Sleep(250 * time.Millisecond)

		// Write in the pipe
		writePipe.Write([]byte("onBoardAlias\n"))

		// Wait a bit, in order for the command to be executed
		time.Sleep(250 * time.Millisecond)

		// Write in the pipe
		writePipe.Write([]byte("http://0.0.0.0:8111\n"))

		// Wait a bit, in order for the command to be executed
		time.Sleep(250 * time.Millisecond)

		// Now write an access token
		writePipe.Write([]byte("aaaaaaaa-bbbb-4b41-9e2e-9ced6675ea03_b336ce2c-c111-4fa8-b5f9-76a6e19724b8\n"))

		// Wait a bit, in order for the command to be executed
		time.Sleep(250 * time.Millisecond)

		// Say that we dont want to configure the local-builder connection
		// FIXME
		writePipe.Write([]byte("n\n"))
	}()

	// Execute the onboard command
	_, err = executeCommand(rootOnBoardCmd)

	// Restore Stdin
	os.Stdin = origStdin

	if err != nil {
		test.Fatal(err)
	}

	// Check that the new value in the access token file is good
	writenConnection, err := getConnectionParams("onBoardAlias", accessTokenFile)
	if err != nil {
		test.Fatal(err)
	}
	assert.Equal(test, "onBoardAlias", writenConnection.Alias)
	assert.Equal(test, "http://0.0.0.0:8111", writenConnection.ServerURL)
	assert.Equal(test, "aaaaaaaa-bbbb-4b41-9e2e-9ced6675ea03_b336ce2c-c111-4fa8-b5f9-76a6e19724b8", writenConnection.AccessToken)

	// Check that the configuration file has been correctly updated
	configFileBytes, err := os.ReadFile(configFilePath)
	if err != nil {
		test.Fatal(err)
	}
	configFileContent := string(configFileBytes)
	assert.Contains(test, configFileContent, "onBoardAlias")
}

// TestInteractTokenReq is the function allowing to test if
// the feature allowing to request to the user interactively
// a token + URL if nothing is set works correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestInteractTokenReq(test *testing.T) {

	// Test not available for true backend since we change the token used...
	if bckIsReal {
		test.Log("[WARNING] It is not possible to test interaction token with a true backend...")
		return
	}

	// Create a false token file
	accessTokenFilePath, err := createFalseTokenFile(nil)
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false file created at the beginning
	defer deleteFalseFile(accessTokenFilePath)

	// Tell the rest of the code we want to test the interactive requesting
	testInteractReq = true

	// Create a pipe
	readPipe, writePipe, err := os.Pipe()
	if err != nil {
		test.Fatal(err)
	}

	// Save the original *os.File
	origStdin := os.Stdin

	// Set os.Stdin with our fake one
	os.Stdin = readPipe

	// Start a go routine that will fill what the command 'rp-cli onboard' wants
	go func() {
		// Wait a bit, in order for the command to be executed
		time.Sleep(100 * time.Millisecond)

		// Write in the pipe
		writePipe.Write([]byte("http://0.0.0.0:8111\n"))

		// Wait a bit, in order for the command to be executed
		time.Sleep(100 * time.Millisecond)

		// Now write an access token
		writePipe.Write([]byte("aaaaaaaa-bbbb-4b41-9e2e-9ced6675ea03_b336ce2c-c111-4fa8-b5f9-76a6e19724b8\n"))
	}()

	// Run the `misc arch` command
	_, err = executeCommand(miscCommand, miscArchCmd)

	// Wait for the end of the test
	time.Sleep(300 * time.Millisecond)

	// Restore Stdin
	os.Stdin = origStdin

	// Tell the rest of the code that we have finished the interactive requesting test
	testInteractReq = false

	if err != nil {
		test.Fatal(err)
	}

	// Check that the new value in the access token file is good
	writenConnection, err := getConnectionParams("tests-BCK", accessTokenFile)
	if err != nil {
		test.Fatal(err)
	}
	assert.Equal(test, "tests-BCK", writenConnection.Alias)
	assert.Equal(test, "http://0.0.0.0:8111", writenConnection.ServerURL)
	assert.Equal(test, "aaaaaaaa-bbbb-4b41-9e2e-9ced6675ea03_b336ce2c-c111-4fa8-b5f9-76a6e19724b8", writenConnection.AccessToken)
}

// TestPanicManagementDevMode is the function allowing to test if
// rp-cli reacts as it needs to do when a panic pops in code during
// development mode
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestPanicManagementDevMode(test *testing.T) {
	// Set logs to be written in our test buffer
	Log.Out = testingOutputBuffer

	// Reset the buffer
	testingOutputBuffer.Reset()

	// Chaos is coming...
	makeAppCrashWithRecover()
	// Hopefully, it has been recovered...

	// Read from the string buffer
	out, err := io.ReadAll(testingOutputBuffer)
	if err != nil {
		test.Fatal(err)
	}

	// Check the output
	outStr := string(out)
	assert.Contains(test, outStr, runtimeErrorMsg)
	assert.Contains(test, outStr, "index out of range")
	assert.Contains(test, outStr, "Stacktrace")
	assert.Contains(test, outStr, "level=error")

	// Set the logs to be written in Stderr again
	Log.Out = os.Stderr
}

// TestPanicManagementProdMode is the function allowing to test if
// rp-cli reacts as it needs to do when a panic pops in code during
// production mode
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestPanicManagementProdMode(test *testing.T) {
	// Set production mode
	BuildMode = "production"

	// Set logs to be written in our test buffer
	Log.Out = testingOutputBuffer

	// Reset the buffer
	testingOutputBuffer.Reset()

	// Chaos is coming...
	makeAppCrashWithRecover()
	// Hopefully, it has been recovered...

	// Read from the string buffer
	out, err := io.ReadAll(testingOutputBuffer)
	if err != nil {
		test.Fatal(err)
	}

	// Check the output
	outStr := string(out)
	assert.Contains(test, outStr, runtimeErrorMsg)
	assert.Contains(test, outStr, "a crash report file, named rp-cli.crash, has been created inside")
	assert.Contains(test, outStr, "level=error")

	// Check that a file named "rp-cli.crash" exists
	// Get the configuration directory
	configFileDirPath, err := getConfigDir()
	if err != nil {
		test.Fatal(err)
	}
	crashFilePath := path.Join(configFileDirPath, "rp-cli.crash")
	if assert.FileExists(test, crashFilePath) {

		// Defer the deletion of the file
		defer os.Remove(crashFilePath)

		// Since the file exists, read it
		crashFileBytes, err := os.ReadFile(crashFilePath)
		if err != nil {
			test.Fatal(err)
		}
		crashFileContent := string(crashFileBytes)

		// Check that the file is really a crash report
		assert.Contains(test, crashFileContent, "error: runtime error: index out of range")
		assert.Contains(test, crashFileContent, "Stacktrace:")
	}

	// Set the logs to be written in Stderr again
	Log.Out = os.Stderr

	// Set development mode back
	BuildMode = "development"
}

// TestSignalMonitoring is the function allowing to test if
// rp-cli reacts as it needs to do when a signal is received from
// the OS (typically, when a user does CTRL+C or CTRL+D)
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestSignalMonitoring(test *testing.T) {
	// Request the main code to enable signal monitoring
	signalTestBeingRun = true

	// Save the PID
	ourPID := syscall.Getpid()

	// Create a pipe
	readPipe, writePipe, err := os.Pipe()
	if err != nil {
		test.Fatal(err)
	}

	// Save the original *os.File
	origStdin := os.Stdin

	// Set os.Stdin with our fake one
	os.Stdin = readPipe

	// Start a go routine that will send a interrupt signal to the application
	go func() {
		// Execute the onboard command
		executeCommand(rootOnBoardCmd, "-L", "debug")
	}()

	// Wait a bit, in order for the command to be executed
	time.Sleep(250 * time.Millisecond)

	// Send the singint signal
	syscall.Kill(ourPID, syscall.SIGINT)

	// Wait a bit
	time.Sleep(250 * time.Millisecond)

	// Read from the string buffer
	out, err := io.ReadAll(testingOutputBuffer)
	if err != nil {
		test.Fatal(err)
	}

	// Send 2 "no" just in case
	writePipe.Write([]byte("n\n"))
	time.Sleep(50 * time.Millisecond)
	writePipe.Write([]byte("n\n"))
	time.Sleep(50 * time.Millisecond)

	// Restore Stdin
	os.Stdin = origStdin

	// Convert bytes slice to string
	strOutput := string(out)

	fmt.Println(strOutput)

	// Check that the signal has been detected
	assert.Contains(test, strOutput, "level=debug msg=\"Detect signal interrupt\"")
	assert.Contains(test, strOutput, "level=debug msg=\"Send signal interrupt (from main)\"")
	assert.Contains(test, strOutput, "level=debug msg=\"End rp-cli with signal 4\"")
	assert.Contains(test, strOutput, "level=info msg=\"Ending signals monitoring...\"")

	// Reset the variable
	signalTestBeingRun = false
}

// TestHandlerErrorEvent is the function allowing to test if
// the function handleErrorEvent() is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestHandlerErrorEvent(test *testing.T) {
	// Request the mocked import to fail at blocking time
	rpbckmock.RequestErrorEvent()

	// Try to import a project
	_, err := executeCommand("admin", "import-project", "-i", "redpesk-test-handle-error", "-v", "1.0", "-d", "description of imported distro")

	// Test that we have the correct error message
	assert.ErrorContains(test, err, "Status: failed - Message: This is a mocked error for test")
}

// TestPrinterInteractive the function allowing to test if
// the feature allowing to interactively print pages of objects works correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestPrinterInteractive(test *testing.T) {

	// In order to test that, we will create a lot of teams!
	for i := 0; i < 70; i++ {
		// Create the command
		teamName := fmt.Sprintf("Interactive team %d", i)
		teamDescription := fmt.Sprintf("Équipe n°%d pour le test printer interactif", i)
		cmdToRun := getMainAddTeamCmd(teamName, teamDescription, false)
		// Create the team
		checkActionCmdAns(test, cmdToRun...)
	}

	// Change the nopager option
	NoPagerTest = false
	defer func() { NoPagerTest = true }()

	// Run the command in a simulated terminal
	cmdsString := "n\nn\nq"
	cmdOutput, err := execCmdInSimTerm(cmdsString, teamsCommand, "list", "-v")
	if err != nil {
		test.Fatal(err)
	}

	// Check that in the output we correctly find the indication that the code work well
	assert.Regexp(test, `Page: [0-9]+/[0-9]+ - Navigation: (\(N\)ext|\(P\)revious)/(\(Q\)uit|\(P\)revious)`, cmdOutput)
}

// --------------------------------------------------------------
// Utilities functions ------------------------------------------
// --------------------------------------------------------------

// executeCommand allows the testing part of the code to run
// the rp-cli commands
//
// :param args: Args to pass to the commands (sub-command, flags, etc.)
// :type args: ...string
// :returns: string, error -- command result and an error if a problem has occurred
func executeCommand(args ...string) (string, error) {

	// Get the rp-cli command
	rootCmd := RootCmd()

	// Reset the buffer
	testingOutputBuffer.Reset()

	// Set the output buffer on the test command
	rootCmd.SetOut(testingOutputBuffer)
	rootCmd.SetErr(testingOutputBuffer)

	// Set the arguments to the root command
	rootCmd.SetArgs(args)

	// Execute the command
	err := rootCmd.Execute()

	// End properly what needs to be ended (RPPrinter, and the different connections)
	endProperlyCLI(false, 0)

	if err != nil && !utils.ContainsString(args, "--noerror") {
		return "", err
	}

	// Read from the string buffer
	out, err := io.ReadAll(testingOutputBuffer)
	if err != nil {
		return "", err
	}

	return string(out), err
}

// execCmdInSimTerm allows the testing part of the code to run
// the rp-cli commands but in a simulated terminal.
// This function is necessary when `rp-cli` code use the "terminal" functions,
// like in `rp-cli local package` for example
// Moreover, the user can supply stdin to be send to the executed process
//
// :param stdin: STDIN to pass to the simulated console
// :type stdin: string
// :param args: Args to pass to the commands (sub-command, flags, etc.)
// :type args: ...string
// :returns: string, error -- command result and an error if a problem has occurred
func execCmdInSimTerm(stdin string, args ...string) (string, error) {
	var err error

	// Reset the buffer
	testingOutputBuffer.Reset()

	// Create the simulated terminal if necessary
	if simulatedConsole == nil {
		simulatedConsole, err = expect.NewConsole(expect.WithStdout(testingOutputBuffer))
		if err != nil {
			return "", err
		}
		go func() {
			simulatedConsole.ExpectEOF()
		}()
	}

	// Get the rp-cli command
	rootCmd := RootCmd()

	// Save the original *os.File
	origStdin := os.Stdin
	origStdout := os.Stdout

	// Set os.Stdin and os.Stdout with our fake one
	os.Stdin = simulatedConsole.Tty()
	os.Stdout = simulatedConsole.Tty()

	// Set the output buffer on the test command
	rootCmd.SetOut(simulatedConsole.Tty())
	rootCmd.SetErr(simulatedConsole.Tty())

	// Function allowing to rework some arguments if needed
	reworkedArgs := reworksArgsIfNeeded(args)
	if len(reworkedArgs) > 0 {
		rootCmd.SetArgs(reworkedArgs)
	} else {
		// Set the arguments to the root command
		rootCmd.SetArgs(args)
	}

	if stdin != "" {
		go func() {
			// Wait a bit before sending the lines on stdin
			time.Sleep(time.Duration(100) * time.Millisecond)

			// Split the stdin into all the lines
			linesSlice := strings.Split(stdin, "\n")
			for _, line := range linesSlice {
				time.Sleep(time.Duration(50) * time.Millisecond)
				simulatedConsole.SendLine(line)
			}
		}()
	}

	// Execute the command
	err = rootCmd.Execute()

	// End properly what needs to be ended (RPPrinter, and the different connections)
	endProperlyCLI(false, 0)

	// Restore Stdin and Stdout
	os.Stdin = origStdin
	os.Stdout = origStdout

	if err != nil {
		return "", err
	}

	// Read from the string buffer
	out, err := io.ReadAll(testingOutputBuffer)
	if err != nil {
		return "", err
	}

	// Close the simulated console if running
	if simulatedConsole != nil {
		simulatedConsole.Close()
		simulatedConsole = nil
	}

	return string(out), err
}

// checkLineEndWith allows to check that, in a given text, there is a line
// that starts with the given prefix and that also finish with the given
// suffix
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param fullText: Full text where the line is hidden
// :type fullText: string
// :param prefix: Prefix of the line to look for
// :type prefix: string
// :param suffix: Suffix that the line needs to contain
// :type suffix: string
func checkLineEndWith(test *testing.T, fullText, prefix, suffix string) {
	// Get the line we want
	lineWanted := utils.ExtractLineWithPrefix(fullText, prefix)

	// Check that the line has been found
	assert.NotEqual(test, "", lineWanted)

	// Check that this line ends with what the caller wants
	assert.True(test, strings.HasSuffix(lineWanted, suffix))
}

// extractSubLogsWithFlag is the function allowing to extract some sublogs
// separated from the main logs with flags looking like:
//
// - Start flag: "CLONE LOGS >>>"
// - End flag: "<<< CLONE LOGS"
//
// :param fullLogs: Full logs where there are the sublogs to extract
// :type fullLogs: string
// :param flag: Flag to look for in order to extract the sublogs
// :type flag: string
// :returns: string -- the sublogs wanted
func extractSubLogsWithFlag(fullLogs, flag string) string {
	// Construct the flags (begin and end) from the flag given by the caller
	startFlag := flag + " >>>"
	endFlag := "<<< " + flag

	// First, check that both flags are inside
	if !strings.Contains(fullLogs, startFlag) || !strings.Contains(fullLogs, endFlag) {
		return ""
	}

	// Split the full logs
	tmpSlice := strings.Split(fullLogs, startFlag)
	if len(tmpSlice) != 2 {
		return ""
	}

	// Split the first substring
	tmpSlice2 := strings.Split(tmpSlice[1], endFlag)
	if len(tmpSlice2) != 2 {
		return ""
	}

	finalSubLogs := strings.TrimSpace(tmpSlice2[0])

	return finalSubLogs
}

// checkCmdAnswerKeysRegexp is a generic function allowing to run
// a rp-cli command, to get the answer and to assert, in the answer,
// if the specified value for each key match a given regular expression
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param lineNb: Number of line to have in the answer
// :type lineNb: int
// :param regExpMap: Map containing the key and the corresponding regexp to match on the value
// :type regExpMap: map[string]string
// :param args: Args to pass as command (sub-command, flags, etc.)
// :type args: ...string
// :returns: string -- command result
func checkCmdAnswerKeysRegexp(test *testing.T, lineNb int, regExpMap map[string]string, args ...string) string {
	// First, let's run the wanted command
	cmdOutput, err := executeCommand(args...)
	if err != nil {
		test.Log("A problem occurred while executing the following command: ", args)
		test.Fatal(err)
	}
	cmdOutputTrimmed := strings.TrimSpace(cmdOutput)

	// Check the object obtained
	checkObjectKeysRegexp(test, cmdOutputTrimmed, lineNb, regExpMap, true)

	return cmdOutputTrimmed
}

// checkObjectKeysRegexp is a generic function allowing to check that
// an object returned by rp-cli after a "get" or "list" command run matches
// the regexp given in parameter
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param lineNb: Number of line to have in the object
// :type lineNb: int
// :param objToCheck: Object as a suit of line to check
// :type objToCheck: string
// :param regExpMap: Map containing the key and the corresponding regexp to match on the value
// :type regExpMap: map[string]string
// :param strictCheck: True if the number of lines needs to be checked as well as that each key is in the output
// :type strictCheck: bool
func checkObjectKeysRegexp(test *testing.T, objToCheck string, lineNb int, regExpMap map[string]string, strictCheck bool) {

	finalResult := true

	// Split the answer by line
	outputLines := strings.Split(objToCheck, "\n")

	if strictCheck {
		// Check that all the keys are inside the output
		for key := range regExpMap {
			finalResult = finalResult && assert.Contains(test, objToCheck, key)
		}

		// Check the number of lines
		finalResult = finalResult && assert.Equal(test, lineNb, len(outputLines))
	}

	// Check the answer line by line
	for _, line := range outputLines {

		// Get rid of space, tab, new line at the end and at the beginning of the line
		lineNoSpace := strings.TrimSpace(line)

		// Try to split the line in two (key, value)
		lineKeyValue := strings.SplitN(lineNoSpace, ":", 2)

		lineValue := ""
		if len(lineKeyValue) > 1 {
			lineValue = strings.TrimSpace(lineKeyValue[1])
		}

		for key, regExp := range regExpMap {
			// Check if the current line start with the current key
			if strings.HasPrefix(lineNoSpace, key) {
				// Check if the line match the corresponding regexp
				finalResult = (finalResult && assert.Regexp(test, regexp.MustCompile(regExp), lineValue))
			}
		}
	}

	// One of the line did not match the regular expression
	if !finalResult {
		test.Logf("Not matching object:\n\n%s\n", objToCheck)
	}
}

// checkCmdAnswerListKeysRegexp is a generic function allowing to run
// a rp-cli command, to get the answer and to assert, in the answer,
// if the specified value for each key match a given regular expression
// The answer received from the command line can be a list of object,
// separated by an empty line
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param lineNb: Number of line to have in the answer, for each object
// :type lineNb: int
// :param objSep: Separator used between each object in the output
// :type objSep: string
// :param regExpMap: Map containing the key and the corresponding regexp to match on the value
// :type regExpMap: map[string]string
// :param args: Args to pass as command (sub-command, flags, etc.)
// :type args: ...string
func checkCmdAnswerListKeysRegexp(test *testing.T, lineNb int, objSep string, regExpMap map[string]string, args ...string) {

	// First, let's run the wanted command
	cmdOutput, err := executeCommand(args...)
	if err != nil {
		test.Log("A problem occurred while executing the following command: ", args)
		test.Fatal(err)
	}
	cmdOutputTrimmed := strings.TrimSpace(cmdOutput)

	// Split the answer line by line
	outputLines := strings.Split(cmdOutputTrimmed, "\n")

	// Accumulate the line until we have a full object
	currentObj := ""
	for _, line := range outputLines {
		if line == objSep {
			// Get rid of the last new line
			trimmedObj := strings.TrimSpace(currentObj)
			// This is a line separator, check the current object
			checkObjectKeysRegexp(test, trimmedObj, lineNb, regExpMap, true)
			// Reset the current object
			currentObj = ""
		} else {
			// Add the line in the current object to process it later
			currentObj += (line + "\n")
		}
	}

	// Run the checks on the last object
	if currentObj != "" {
		// Get rid of the last new line
		trimmedObj := strings.TrimSpace(currentObj)
		// This is a line separator, check the current object
		checkObjectKeysRegexp(test, trimmedObj, lineNb, regExpMap, true)
	}
}

// getEachColumnFromLine allows to get the value of each column
// from a line containing tabs, space, etc.
//
// :param line: Line to process
// :type line: string
// :returns: []string -- Each column of the line
func getEachColumnFromLine(line string, colNamesWanted ...string) []string {

	var resultSlice []string

	// Get rid of space, tab, new line at the end and at the beginning of the line
	trimmedLine := strings.TrimSpace(line)

	// Check if there are double space in the line and turn it to tab if this is the case
	// We have to do that in order to comply with the tab writer used by rp-cli
	// I know this is not that clean though
	if strings.Contains(trimmedLine, "  ") {
		trimmedLine = strings.ReplaceAll(trimmedLine, "  ", "\t")
	}

	// Split the line with each column
	tmpColumnsNames := utils.DeleteEmptyStringFromSlice(strings.Split(trimmedLine, "\t"))

	// Trim each item of the list
	trimmedColNames := []string{}
	for _, item := range tmpColumnsNames {
		trimmedColNames = append(trimmedColNames, strings.TrimSpace(item))
	}

	// If possible, check that we match the result we want
	if (len(colNamesWanted) > 0) && (len(tmpColumnsNames) < len(colNamesWanted)) {
		// Find the index where we have a problem
		for _, item := range trimmedColNames {
			if !utils.ContainsString(colNamesWanted, item) {
				resultSlice = append(resultSlice, strings.Split(item, " ")...)
			} else {
				resultSlice = append(resultSlice, item)
			}
		}

		// We've got the index
	} else {
		resultSlice = append(resultSlice, trimmedColNames...)
	}

	return resultSlice
}

// checkObjectKeysRegexpCol is a generic function allowing to check that
// an object returned by rp-cli after a "get" or "list" command run matches
// the regexp given in parameter.
// The answer of the command needs to be a "column type". By example:
//
// ID					 					 Slug			 		 Name
// 651b517f-4640-49e5-8dd7-7ad3a1b43a82 	 Helloworld binding 	 helloworld-binding
// ebebaac6-f54d-4172-970d-8513eb6dba46 	 Python3 pytest tap 	 python3-pytest-tap
// 3b3db877-1fd5-4433-9ddf-661c98dc6ab2 	 Python3 tap.py 	 	 python3-tap-py
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param objToCheck: Object to check
// :type objToCheck: string
// :param ColNb: Number of Column to have in the object
// :type ColNb: int
// :param regExpMap: Map containing the key and the corresponding regexp to match on the value
// :type regExpMap: map[string]string
func checkObjectKeysRegexpCol(test *testing.T, objToCheck string, ColNb int, regExpMap map[string]string) {
	// Split the answer by line
	outputLines := strings.Split(objToCheck, "\n")

	// Check that the first line contains all the keys that needs to be there
	colNameWanted := []string{}
	for key := range regExpMap {
		isKeyInHeader := assert.Contains(test, outputLines[0], key)
		// Stop checking things if the key is not in header
		if !isKeyInHeader {
			return
		}
		colNameWanted = append(colNameWanted, key)
	}

	// Get each column from the first line
	columnsNames := getEachColumnFromLine(outputLines[0], colNameWanted...)

	// Check the number of columns
	assert.Equal(test, ColNb, len(columnsNames))

	// Check the answer line by line
	for _, line := range outputLines[1:] {

		// Get each column of the current line
		columnsValues := getEachColumnFromLine(line)

		for index, key := range columnsNames {
			// Check that for this line we have enough values (to avoid panic)
			if len(columnsValues) == len(columnsNames) {
				// Check if the value match the corresponding regexp
				assert.Regexp(test, regexp.MustCompile(regExpMap[key]), columnsValues[index])
			} else {
				// Print a warning
				test.Log("[WARNING] The following line has not been checked because of missing columns: ", columnsValues)
				test.Log("[WARNING] Wanted columns: ", colNameWanted)
			}
		}
	}
}

// checkCmdAnsKeysRegexpCol is a generic function allowing to run
// a rp-cli command, to get the answer and to assert, in the answer,
// if the specified value for each key match a given regular expression.
// The answer of the command needs to be a "column type". By example:
//
// ID					 					 Slug			 		 Name
// 651b517f-4640-49e5-8dd7-7ad3a1b43a82 	 Helloworld binding 	 helloworld-binding
// ebebaac6-f54d-4172-970d-8513eb6dba46 	 Python3 pytest tap 	 python3-pytest-tap
// 3b3db877-1fd5-4433-9ddf-661c98dc6ab2 	 Python3 tap.py 	 	 python3-tap-py
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param columnNb: Number of column to have in the answer
// :type columnNb: int
// :param regExpMap: Map containing the key and the corresponding regexp to match on the value
// :type regExpMap: map[string]string
// :param args: Args to pass as command (sub-command, flags, etc.)
// :type args: ...string
func checkCmdAnsKeysRegexpCol(test *testing.T, columnNb int, regExpMap map[string]string, args ...string) {

	// First, let's run the wanted command
	cmdOutput, err := executeCommand(args...)
	if err != nil {
		test.Log("A problem occurred while executing the following command: ", args)
		test.Fatal(err)
	}
	cmdOutputTrimmed := strings.TrimSpace(cmdOutput)

	// Check the answer of the command
	checkObjectKeysRegexpCol(test, cmdOutputTrimmed, columnNb, regExpMap)
}

// checkActionCmdAns is a generic function allowing to run
// a rp-cli command, to get the answer and to assert, in the answer,
// if every action is OK for rp-cli.
// The answer of the command needs to be like the following example:
//
// -- Saving access token for Redpesk authentication --
// Creation of the token file...	[OK]
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param args: Args to pass as command (sub-command, flags, etc.)
// :type args: ...string
// :returns: string -- Output of the command
func checkActionCmdAns(test *testing.T, args ...string) string {

	// First, let's run the wanted command
	cmdOutput, err := executeCommand(args...)
	if err != nil {
		test.Log("A problem occurred while executing the following command: ", args)
		test.Fatal(err)
	}
	cmdOutputTrimmed := strings.TrimSpace(cmdOutput)

	// Split the answer by line
	outputLines := strings.Split(cmdOutputTrimmed, "\n")

	// Check the answer line by line
	for _, line := range outputLines {

		if !strings.HasPrefix(line, "--") &&
			!strings.HasPrefix(line, "time=") &&
			!strings.HasPrefix(line, "Slug of the created") &&
			line != "" {
			assert.Contains(test, line, "[OK]")
		}
	}

	return cmdOutputTrimmed
}

// getObjNumber is a generic function allowing to get the number
// of object from a command given by the user (most of the time, a list command)
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param args: Args to pass as command (sub-command, flags, etc.)
// :type args: ...string
// :returns: string -- Output of the command
func getObjNumber(test *testing.T, args ...string) int {

	// First, let's run the wanted command
	cmdOutput, err := executeCommand(args...)
	if err != nil {
		test.Log("A problem occurred while executing the following command: ", args)
		test.Fatal(err)
	}
	cmdOutputTrimmed := strings.TrimSpace(cmdOutput)

	// Split the answer by line
	outputLines := strings.Split(cmdOutputTrimmed, "\n")

	return (len(outputLines) - 1)
}

// addObjectIfNotExists is the function allowing to add an object
// in the current backend if it does not already exists
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param mainSubCmd: Main subcommand to use to get the object
// :type mainSubCmd: string
// :param objSlug: Slug of the object to create if it does not
// already exists
// :type objSlug: string
// :param args: Args to pass as command to create the project
// :type args: ...string
func addObjectIfNotExists(test *testing.T, mainSubCmd string, objSlug string, args ...string) error {
	// Try to split the sub command (in case where this is a sub sub-command)
	mainSubCmds := strings.Split(mainSubCmd, " ")
	// Add "get" and the slug in the string slice
	mainSubCmds = append(mainSubCmds, "get", objSlug)
	// Try to get the project to see if it exists
	_, err := executeCommand(mainSubCmds...)
	if err != nil {
		// The command returns an error, we need to create the project
		checkActionCmdAns(test, args...)
	}
	return nil
}

// createFalseFile allows to backup an eventual existing file,
// and to create a new one with a given content. This is useful
// for our test suite
//
// :param filePath: Path to the false file
// :type filePath: string
// :param falseFileContent: Content of the false file to create
// :type falseFileContent: string
// :returns: error -- An error if something went wrong, Nil in the other case
func createFalseFile(filePath string, falseFileContent string) error {

	// Check if a file exists with the given path
	if _, err := os.Stat(filePath); err == nil {
		// The file exists, create a back-up of this file
		currentFilename := filepath.Base(filePath)
		fileDir := filepath.Dir(filePath)
		backupFilename := "backup_" + currentFilename
		err := utils.CopyFile(filePath, fileDir+"/"+backupFilename)
		if err != nil {
			return err
		}

		// Remove the file
		err = os.Remove(filePath)
		if err != nil {
			return err
		}

	}

	// Create the false file with the given content
	falseFile, err := os.Create(filePath)
	if err != nil {
		return err
	}
	defer falseFile.Close()
	_, err = falseFile.WriteString(falseFileContent)
	if err != nil {
		return err
	}

	return nil
}

// deleteFalseFile allows to delete a false file created before
// with the createFalseFile() function. It restore the backup if one
// is existing.
//
// :param filePath: Path to the false file
// :type filePath: string
// :returns: error -- An error if something went wrong, Nil in the other case
func deleteFalseFile(filePath string) error {

	// Guess the backup file name
	currentFilename := filepath.Base(filePath)
	fileDir := filepath.Dir(filePath)
	backupFilename := "backup_" + currentFilename

	// Delete the current false file
	err := os.Remove(filePath)
	if err != nil {
		return err
	}

	// Check if a backup file exists
	if _, err := os.Stat(fileDir + "/" + backupFilename); err == nil {
		// A backup file exists, set-it back as main file
		err := utils.CopyFile(fileDir+"/"+backupFilename, filePath)
		if err != nil {
			return err
		}

		// Delete the backup file
		err = os.Remove(fileDir + "/" + backupFilename)
		if err != nil {
			return err
		}
	}

	return nil
}

// createFalseTokenFile allows to create a false token file while
// saving the current one thanks to a backup
//
// :param aliases: Slice of aliases to set in the file
// :type aliases: []string
// :returns: string, error -- The path to the token file and an error if something went wrong, Nil in the other case
func createFalseTokenFile(aliases []string) (string, error) {
	// Get path to the configuration directory
	configDirPath, err := getConfigDir()
	if err != nil {
		return "", err
	}
	accessTokenFilePath := configDirPath + "access_token.json"

	// If the slice is nil, force its value
	if aliases == nil {
		aliases = []string{"mytestalias"}
	}

	// Generate the token file content
	falseToken := "["
	for index, alias := range aliases {
		falseToken += fmt.Sprintf(fakeConnectionTemplate, alias, "https://test-my-app.redpesk.bzh", "aaaaaaaa", "false")
		if index < (len(aliases) - 1) {
			falseToken += ",\n"
		}
	}
	falseToken += "\n]\n"

	err = createFalseFile(accessTokenFilePath, falseToken)
	if err != nil {
		return "", err
	}

	return accessTokenFilePath, nil
}

// falseFileExists allows to check if a fake token file already exists
// for tests
//
// :returns: bool, string, error -- True if a fake file already exists, false in the other case, the token file path and an error
func fakeFileExists(filename string) (bool, string, error) {
	// Get path to the configuration directory
	configDirPath, err := getConfigDir()
	if err != nil {
		return false, "", err
	}
	filePath := path.Join(configDirPath, filename)

	// Check if a file exists with the given path
	if _, err := os.Stat(filePath); err == nil {
		return true, filePath, nil
	} else {
		return false, filePath, nil
	}
}

// createFalseConfigFile allows to create a false configuration file while
// saving the current one thanks to a backup
//
// :returns: string, error -- The path to the token file and an error if something went wrong, Nil in the other case
func createFalseConfigFile() (string, error) {
	// Get path to the configuration file
	configFileDirPath, err := getConfigDir()
	if err != nil {
		return "", err
	}
	configFilePath := configFileDirPath + "rp-cli-config.json"

	// Create our false configuration file and backup the other one
	falseCnt := `{
	"serveralias": "mytestalias",
	"pkgsdir": "/tmp/redpesk-cli-test/gitpkgs",
	"sourcesdir": "/tmp/redpesk-cli-test/gitsources"
	}`
	err = createFalseFile(configFilePath, falseCnt)
	if err != nil {
		return "", err
	}

	return configFilePath, nil
}

// createFalseLocalConfigFile allows to create a false local-builder
// configuration file while saving the current one thanks to a backup
//
// :returns: string, error -- The path to the token file and an error if something went wrong, Nil in the other case
func createFalseLocalConfigFile() (string, error) {
	// Get path to the configuration file
	configDirPath, err := getConfigDir()
	if err != nil {
		return "", err
	}
	configFilePath := configDirPath + localBuilderFile

	// Create our false configuration file and backup the other one
	falseCnt := `[{
	"alias": "local_builder",
	"serverurl": "http://test-rp-cli-lad.bzh:8080",
	"token": "",
	"is_admin": false
	}]`
	err = createFalseFile(configFilePath, falseCnt)
	if err != nil {
		return "", err
	}

	return configFilePath, nil
}

// createLocalTestDirs allows to create the directories necessary
// to test the "local" subcommand (gitpkgs and gitsources)
//
// :returns: error -- an error if a problem has occurred
func createLocalTestDirs() error {
	// Create the gitsources directory
	err := utils.CreateDirIfNotExists(gitSourcesTestDir, true)
	if err != nil {
		return err
	}

	// Create the gitpkgs directory
	err = utils.CreateDirIfNotExists(gitPackagesTestDir, true)
	if err != nil {
		return err
	}

	// Create the config directory for tests
	err = utils.CreateDirIfNotExists(cfgDirPathTestDir, true)
	if err != nil {
		return err
	}

	return nil
}

// makeAppCrashWithRecover is a function allowing to crash the app
// and check that the recover works
//
// :returns: error -- an error if a problem has occurred
func makeAppCrashWithRecover() string {
	// Enable the recovering
	defer gracefulRecoverFromPanic(false)

	// Do a huge mess in here!
	anInnocentSlice := []string{"one", "two"}
	hugeMistake := anInnocentSlice[3]

	return hugeMistake
}

// --------------------------------------------------------------
// --------------------------------------------------------------
