/*
Copyright © 2020-2023 IoT.bzh Company
Contact: https://www.iot.bzh/licensing

This file is part of the redpesk-cli project, the redpesk command line tool.

$RP_BEGIN_LICENSE$
# Commercial License Usage

Licensees holding valid commercial IoT.bzh licenses may use this file in
accordance with the commercial license agreement provided with the
Software or, alternatively, in accordance with the terms contained in
a written agreement between you and The IoT.bzh Company. For licensing terms
and conditions see https://www.iot.bzh/terms-conditions. For further
information use the contact form at https://www.iot.bzh/contact.

# GNU General Public License Usage

Alternatively, this file may be used under the terms of the GNU General
Public license version 3. This license is as published by the Free Software
Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
of this file. Please review the following information to ensure the GNU
General Public License requirements will be met
https://www.gnu.org/licenses/gpl-3.0.html.

$RP_END_LICENSE$
*/
package cmd

import (
	"errors"
	"fmt"
	"os"
	"regexp"
	"strings"
	"text/tabwriter"
	"time"

	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/printer"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/rpxxxapi/bckapi"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/utils"
	"git.ovh.iot/redpesk/redpesk-infra/rp-webserver/actions/rpapiv1"
	"github.com/gofrs/uuid"
	"github.com/spf13/cobra"
	"github.com/spf13/pflag"
	"github.com/spf13/viper"
)

func initMiscCmd(rootCmd *cobra.Command) {

	// miscCmd represents the misc command
	var miscCmd = &cobra.Command{
		Use:   miscCommand,
		Short: miscCmdShortDesc,
	}

	// statusCmd represents the status command
	var statusCmd = &cobra.Command{
		Use:     miscStatusCmd,
		Aliases: []string{miscStatusCmdAlias},
		Short:   miscStatusCmdShortDesc,
		RunE:    rpStatus,
	}

	// versionCmd represents the version command
	var versionCmd = &cobra.Command{
		Use:     miscVersionCmd,
		Aliases: []string{miscVersionCmdAlias},
		Short:   miscVersionCmdShortDesc,
		RunE:    rpVersion,
	}

	var freshCmd = &cobra.Command{
		Use:          "one-beer",
		SilenceUsage: true,
		Hidden:       true,
		RunE:         beerCmdHandler,
	}

	// archCmd represents the arch command
	var archCmd = &cobra.Command{
		Use:   miscArchCmd,
		Short: miscArchCmdShortDesc,
		Long:  miscArchCmdLongDesc,
		RunE:  rpArch,
	}

	// configCmd represents the configuration command
	var configCmd = &cobra.Command{
		Use:         miscConfigCmd,
		Aliases:     []string{miscConfigCmdAlias},
		Short:       miscConfigCmdShortDesc,
		RunE:        rpCliConfiguration,
		Annotations: map[string]string{"__bck_connect__": "FALSE"}, // No need for backend connection for this command
	}

	// configGetCmd represents the "configuration get" command
	var configGetCmd = &cobra.Command{
		Use:               miscConfigGetCmd,
		Short:             miscConfigGetCmdShortUsage,
		Long:              miscConfigGetCmdLongUsage,
		RunE:              rpCliConfigurationGet,
		ValidArgsFunction: getPersistentFlagsAsTab,
		Annotations:       map[string]string{"__bck_connect__": "FALSE"}, // No need for backend connection for this command
	}

	// configSetCmd represents the "configuration get" command
	var configSetCmd = &cobra.Command{
		Use:               miscConfigSetCmd,
		Short:             miscConfigSetCmdShortUsage,
		Long:              miscConfigSetCmdLongUsage,
		RunE:              rpCliConfigurationSet,
		ValidArgsFunction: getPersistentFlagsAsTab,
		Annotations:       map[string]string{"__bck_connect__": "FALSE"}, // No need for backend connection for this command
	}

	// cmdsTreeCmd represents the cmd-tree command
	var cmdsTreeCmd = &cobra.Command{
		Use:         miscCmdTreeCmd,
		Short:       miscCmdTreeCmdShortDesc,
		RunE:        rpCmdsTreeOutput,
		Annotations: map[string]string{"__bck_connect__": "FALSE"}, // No need for backend connection for this command
	}

	// cmdSearchCmd represents the cmd-search command
	var cmdSearchCmd = &cobra.Command{
		Use:         miscCmdSearchCmd,
		Short:       miscCmdSearchCmdShortDesc,
		Long:        miscCmdSearchCmdLongDesc,
		RunE:        rpCmdSearch,
		Annotations: map[string]string{"__bck_connect__": "FALSE"}, // No need for backend connection for this command
	}

	// Here you will define your flags and configuration settings.

	// Persistent Flags -----------------------------------------
	// ----------------------------------------------------------

	// Local Flags ----------------------------------------------
	versionCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	archCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	configCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	statusCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	configSetCmd.Flags().String(miscConfigSetValueFlag, "", miscConfigSetValueFlagUsage)
	cmdSearchCmd.Flags().BoolP(miscCmdSearchRegexpFlag, miscCmdSearchRegexpFlagAlias, false, miscCmdSearchRegexpFlagUsage)
	// ----------------------------------------------------------

	// Add some flags as required -------------------------------
	// ----------------------------------------------------------

	// Add the group commands "misc" as subcommand of rootCmd
	rootCmd.AddCommand(miscCmd)
	// Add the command "status" as a subcommand of miscCmd
	miscCmd.AddCommand(statusCmd)
	// Add the command "version" as a subcommand of miscCmd
	miscCmd.AddCommand(versionCmd)
	// Add the command "one-beer" as a subcommand of miscCmd
	miscCmd.AddCommand(freshCmd)
	// Add the command "arch" as a subcommand of miscCmd
	miscCmd.AddCommand(archCmd)
	// Add the command "cmd-tree" as a subcommand of miscCmd
	miscCmd.AddCommand(cmdsTreeCmd)
	// Add the command "cmd-search" as a subcommand of miscCmd
	miscCmd.AddCommand(cmdSearchCmd)
	// Add the command "configuration" as a subcommand of miscCmd
	miscCmd.AddCommand(configCmd)
	// Add the command "configuration get" as a subcommand of configCmd
	configCmd.AddCommand(configGetCmd)
	// Add the command "configuration set" as a subcommand of configCmd
	configCmd.AddCommand(configSetCmd)
}

// --------------------------------------------------------------
// Completion functions -----------------------------------------
// --------------------------------------------------------------

// getPersistentFlagsAsTab gets all the name of the persistent flags
// as an array
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getPersistentFlagsAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {

	// Get the full list of persistent flags
	flagNameTab := []string{}
	getAllPersistentFlags(cmd.Root(), &flagNameTab)

	return flagNameTab, cobra.ShellCompDirectiveNoFileComp
}

// getServicesNamesAsTab gets all the name of the services available
// on the redpesk factory as an array
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getServicesNamesAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	var err error
	servicesList := []string{}

	// Get the list of every available services
	services := bckapi.RPBCKServices{}
	if err = bckAPI.RPServicesGet(&services); err != nil {
		return servicesList, cobra.ShellCompDirectiveNoFileComp
	}

	// Fill the servicesList tab
	for name := range services {
		// Add the slug of the application in the suggestions
		servicesList = append(servicesList, (name + "="))
	}

	return servicesList, cobra.ShellCompDirectiveNoFileComp
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// rpStatus gets the configuration of the Redpesk server
// and prints it to the standard output
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: pointer to the RPAPIInterface object
func rpStatus(cmd *cobra.Command, args []string) error {

	cfg := rpapiv1.RPAPIConfig{}
	if err := bckAPI.RPConfigGet(&cfg); err != nil {
		return err
	}

	// Print the result to the user
	err := RPPrinter.CustomPrintObject(cfg, true, printStatus)

	return err
}

// rpVersion gets the version of the Redpesk server
// and prints it to the standard output
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: pointer to the RPAPIInterface object
func rpVersion(cmd *cobra.Command, args []string) error {

	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Get the version
	ver := rpapiv1.RPVersion{}
	if err = bckAPI.RPVersionGet(&ver, false); err != nil {
		return err
	}

	// Print the result to the user
	err = RPPrinter.CustomPrintObject(ver, verbose, printVersion)

	return err
}

// rpArch gets the list of all the available architectures on
// the Redpesk backend (x86_64, aarch64, etc.)
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func rpArch(cmd *cobra.Command, args []string) error {

	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Get the list of every available architecture
	archis := []rpapiv1.RPArch{}
	if err = bckAPI.RPArchGet(&archis, false); err != nil {
		return err
	}

	// Print the result to the user
	err = RPPrinter.CustomPrintObject(archis, verbose, printArches)

	return err
}

// rpCmdsTreeOutput get the full list of commands available in rp-cli and print it
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func rpCmdsTreeOutput(cmd *cobra.Command, args []string) error {

	// Firstly, retrieve the root command
	rootCmd := cmd.Root()

	// Function needed to print the result for one command
	printForOneCmd := func(oneCmd *cobra.Command, isLast bool, deepnessTabs []string) string {
		if oneCmd.Hidden {
			return ""
		}
		symbolToPrint := "├── "
		if isLast {
			symbolToPrint = "└── "
		}
		if len(deepnessTabs) == 0 {
			symbolToPrint = ""
		}
		valueToReturn := strings.Split(oneCmd.Use, " ")
		return (symbolToPrint + valueToReturn[0] + "\t -- " + oneCmd.Short)
	}

	// Recursion function needed
	var cmdsRecursion func(*cobra.Command, bool, []string) string

	cmdsRecursion = func(localCmd *cobra.Command, isLast bool, deepnessTabs []string) string {

		childrenCmds := localCmd.Commands()

		// Stop case
		if len(childrenCmds) == 0 {
			return printForOneCmd(localCmd, isLast, deepnessTabs)
		}

		// Other case
		outputStr := (printForOneCmd(localCmd, isLast, deepnessTabs))
		for index, tmpCmd := range childrenCmds {
			if !tmpCmd.Hidden {

				// Creation of tab string for the current command
				tabStr := ""
				for _, deepTab := range deepnessTabs {
					tabStr += deepTab
				}

				// Add a new value in the deepness tab according to the situation
				newDeepnessTab := []string{}
				newDeepnessTab = append(newDeepnessTab, deepnessTabs...)
				if index == len(childrenCmds)-1 {
					newDeepnessTab = append(newDeepnessTab, "    ")
				} else {
					newDeepnessTab = append(newDeepnessTab, "│   ")
				}

				outputStr += ("\n" + tabStr + cmdsRecursion(tmpCmd, (index == len(childrenCmds)-1), newDeepnessTab))
			}
		}
		return outputStr
	}

	// Get the string result
	resultOutput := cmdsRecursion(rootCmd, false, []string{})

	// Try to print the result
	RPPrinter.CustomPrintLn(resultOutput)

	return nil
}

// rpCmdSearch allows to run a research inside the commands tree
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func rpCmdSearch(cmd *cobra.Command, args []string) error {

	// Firstly, retrieve the root command
	rootCmd := cmd.Root()

	// Function needed to print the result for one command
	printOneCmd := func(oneCmd *cobra.Command, parentCmds []string) string {
		// Initialize a temp tab
		tabToPrint := []string{}
		tabToPrint = append(tabToPrint, parentCmds...)
		tabToPrint = append(tabToPrint, "--", oneCmd.Short)
		return strings.Join(tabToPrint, " ")
	}

	// Recursion function needed to construct the full commands string
	var cmdsRecursion func(*cobra.Command, []string) string
	cmdsRecursion = func(localCmd *cobra.Command, parentCmds []string) string {
		// Get the children commands
		childrenCmds := localCmd.Commands()

		// Add the current command name in the parent command
		cmdName := strings.Split(localCmd.Use, " ")[0]
		newParentCmd := []string{}
		newParentCmd = append(newParentCmd, parentCmds...)
		newParentCmd = append(newParentCmd, cmdName)

		// Stop case
		if len(childrenCmds) == 0 {
			return printOneCmd(localCmd, newParentCmd)
		}

		// Other case
		outputStr := printOneCmd(localCmd, newParentCmd)
		for _, tmpCmd := range childrenCmds {
			if !tmpCmd.Hidden {
				outputStr += ("\n" + cmdsRecursion(tmpCmd, newParentCmd))
			}
		}
		return outputStr
	}

	// Get the string result containing every command available in 'rp-cli'
	resultOutput := cmdsRecursion(rootCmd, []string{})

	// Try to get a pattern if the user gave it
	searchPat, err := getFirstArg(args, "", "a new error")
	if err != nil {
		searchPat = ""
	}

	if searchPat == "" {
		// Print the full list of commands
		RPPrinter.CustomPrintLn(resultOutput)
	}

	// Try to get the regexp flag
	isRegexp, err := cmd.Flags().GetBool(miscCmdSearchRegexpFlag)
	if err != nil {
		logFlagError(miscCmdSearchRegexpFlag)
		return err
	}

	allLinesInSlice := strings.Split(resultOutput, "\n")
	linesToPrint := []string{}
	if isRegexp {
		// The user gave us a regexp analyse each line with this regexp
		regularExp := regexp.MustCompile(searchPat)
		for _, line := range allLinesInSlice {
			if regularExp.MatchString(line) {
				linesToPrint = append(linesToPrint, line)
			}
		}
	} else {
		// The user only gave us a pattern to search
		for _, line := range allLinesInSlice {
			lineLowCase := strings.ToLower(line)
			if strings.Contains(lineLowCase, searchPat) {
				linesToPrint = append(linesToPrint, line)
			}
		}
	}

	// Try to print the result
	resToPrint := strings.Join(linesToPrint, "\n")
	RPPrinter.CustomPrintLn(resToPrint)

	return nil
}

// rpCliConfiguration gets all the configuration parameters of rp-cli
// and displays it to the user
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func rpCliConfiguration(cmd *cobra.Command, args []string) error {

	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Get the path to the configuration file used
	configFilePath := viper.ConfigFileUsed()
	if configFilePath == "" {
		configFilePath = miscConfigNoFileMsg
	}

	// Print the configuration file path
	RPPrinter.CustomPrintLn(configFileInfoPrefix, configFilePath)

	// If verbose is enabled, print the content of the file
	if verbose && (configFilePath != miscConfigNoFileMsg) {
		content, err := os.ReadFile(configFilePath)
		if err != nil {
			Log.Errorln(pbWhileReadCfgFileMsg, configFilePath)
			return err
		}
		contentStr := string(content)
		// Print the file content
		RPPrinter.CustomPrintLn()
		RPPrinter.CustomPrintLn(miscConfigAnsLine0)
		RPPrinter.CustomPrintLn(contentStr)
		RPPrinter.CustomPrintLn()
	}

	// Get the current connection used for backend and print it
	if configFilePath != miscConfigNoFileMsg {
		// Create the struct where the the config file content will be stored as json
		configJSON := make(map[string]interface{})

		// Read the json value of the configuration
		err = utils.ReadJsonFile(configFilePath, &configJSON)
		if err != nil {
			Log.Errorln(pbWhileReadCfgFileMsg, configFilePath)
			return err
		}

		// Get the serveralias value and print it
		if currConnection, ok := configJSON[serverAliasFlag]; ok {
			currentBckConLine := miscConfigAnsLine4 + "\t" + fmt.Sprintf("%v", currConnection)
			RPPrinter.CustomPrintLn(currentBckConLine)
		}
	}

	// Get the list of available connections for rp-cli
	availableConnect, err := getAllConnectionsParams(accessTokenFile)
	if err != nil {
		return err
	}
	// Print the list of available alias
	aliasString := ""
	// If there are no connection, simply print "[None]"
	if len(availableConnect) == 0 {
		aliasString = "\t" + miscConfigNoConnectMsg
	} else {
		aliasString = "\t"
	}
	// Add each connection
	for index, connection := range availableConnect {
		if index == (len(availableConnect) - 1) {
			aliasString += connection.Alias + "\n"
		} else {
			aliasString += connection.Alias + ", "
		}
	}
	// Finally, print the result
	RPPrinter.CustomPrint(miscConfigAnsLine2)
	RPPrinter.CustomPrint(" " + aliasString)
	if verbose {
		RPPrinter.CustomPrintLn()
	}

	// Get the local builder connection
	localBConnection, err := getConnectionParams(localBuilderAlias, localBuilderFile)
	if err != nil {
		return err
	}

	localServer := ""
	if localBConnection.ServerURL == "" {
		localServer = "[NONE]"
	} else {
		localServer = localBConnection.ServerURL
	}

	// Print the local builder URL info
	RPPrinter.CustomPrint(miscConfigAnsLine3)
	RPPrinter.CustomPrint("\t" + localServer + "\n")
	if verbose {
		RPPrinter.CustomPrintLn()
	}

	// Get all the environment variables that may influence rp-cli behavior
	flagNameTab := []string{}
	getAllPersistentFlags(cmd.Root(), &flagNameTab)

	// Print each environment variables that may interest us
	RPPrinter.CustomPrintLn(miscConfigAnsLine1)
	alreadyPrintedFlag := []string{}
	for _, flag := range flagNameTab {

		if !utils.ContainsString(alreadyPrintedFlag, flag) {
			// Construct the environment variable name from flag
			fullFlag := envPrefix + "_" + strings.ToUpper(flag)

			// Get the value of the corresponding environment variable
			flagValue := os.Getenv(fullFlag)
			if flagValue == "" {
				flagValue = mainNotSetMsg
			}

			// Print the value of the environment variable
			RPPrinter.CustomPrintLn("  " + fullFlag + ":\t" + flagValue)

			// Add the flag in the tab of already printed flag
			alreadyPrintedFlag = append(alreadyPrintedFlag, flag)
		}
	}

	return nil
}

// rpCliConfigurationGet gets the value of a specific persistent
// parameter used by rp-cli
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func rpCliConfigurationGet(cmd *cobra.Command, args []string) error {

	paramValue := mainNotSetMsg

	// Get the "name" of the persistent flag to display
	flagName, err := getFirstArg(args, flagIDNotSetMsg, flagIDRequiredMsg)
	if err != nil {
		return err
	}

	// Get the full list of persistent flags
	flagNameTab := []string{}
	getAllPersistentFlags(cmd.Root(), &flagNameTab)

	// Check that the flag given by the user exists
	doesFlagExists := utils.ContainsString(flagNameTab, flagName)
	if !doesFlagExists {
		return errors.New(paramNotExistMsg)
	}

	// Read the configuration file in order to get the parameter value
	// Get the path to the configuration file used
	configFilePath := viper.ConfigFileUsed()
	if configFilePath != "" {
		// Read the configuration file and get the parameter that is interesting us
		configJSON := make(map[string]interface{})
		err = utils.ReadJsonFile(configFilePath, &configJSON)
		if err != nil {
			Log.Errorln(pbWhileReadCfgFileMsg, configFilePath)
			return err
		}

		if tmpParamValue, ok := configJSON[flagName]; ok {
			paramValue = fmt.Sprintf("%v", tmpParamValue)
		}
	}

	// If the parameter exists in the file print its value,
	// in the other case, just return "[Not set]"
	RPPrinter.CustomPrintLn(paramValue)

	return nil
}

// rpCliConfigurationSet set the value of a specific persistent
// parameter used by rp-cli
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func rpCliConfigurationSet(cmd *cobra.Command, args []string) error {

	// Get the "name" of the persistent flag to display
	flagName, err := getFirstArg(args, flagIDNotSetMsg, flagIDRequiredMsg)
	if err != nil {
		return err
	}

	// Get the full list of persistent flags
	flagNameTab := []string{}
	getAllPersistentFlags(cmd.Root(), &flagNameTab)

	// Check that the flag given by the user exists
	doesFlagExists := utils.ContainsString(flagNameTab, flagName)
	if !doesFlagExists {
		return errors.New(paramNotExistMsg)
	}

	// Get the value the user wants to set
	flagValue, err := cmd.Flags().GetString(miscConfigSetValueFlag)
	if err != nil {
		logFlagError(miscConfigSetValueFlag)
		return err
	}
	if flagValue == "" {
		return errors.New(noNewValueForFlagMsg)
	}

	// Update the configuration file
	return initOrUpdateDefCfgFile(flagName, flagValue)
}

// --------------------------------------------------------------
// Utilities functions ------------------------------------------
// --------------------------------------------------------------

// getDistroUUID is a private function allowing to retrieve the
// UUID of a distribution from a given identifier. This identifier can
// be the UUID itself as a string representation or the Slug of the
// distribution
//
// :param identifier: the identifier of the distribution
// :type identifier: string
// :returns: (uuid.UUID, error) -- the UUID of the distribution and an error if something went bad
func getDistroUUID(identifier string) (uuid.UUID, error) {
	uuidIndex := -1

	// Get the list of the project existing on the Redpesk instance
	distros := bckapi.RPCliDistributions{}
	if err := bckAPI.RPDistrosGet(&distros, false); err != nil {
		return uuid.Nil, err
	}

	// Goes along the project to find the right UUID
	for index, dist := range distros.Distros {
		// Check if the UUID or the Slug is the one we are looking for
		if (dist.ID.String() == identifier) || (dist.Slug == identifier) {
			uuidIndex = index
			break
		}
	}

	if uuidIndex == -1 {
		return uuid.Nil, fmt.Errorf(distIdentifierNotFound, identifier)
	}

	return distros.Distros[uuidIndex].ID, nil
}

// getTabOfDistrosUUID is a private function allowing to get a tab
// of UUID as string from a tab of identifiers. This identifier can
// be the UUID itself as a string representation or the Slug of the
// distribution
//
// :param identifiers: tab of identifiers of the distribution to convert
// :type identifiers: []string
// :returns: ([]string, error) -- the tab of UUIDs (as string) of
// the distribution and an error if something went bad
func getTabOfDistrosUUID(identifiers []string) ([]string, error) {
	// Initialize the result variable
	uuidTab := []string{}

	// Go through the tab pass in parameter
	for _, id := range identifiers {
		tmpUUID, err := getDistroUUID(id)
		if err != nil {
			return uuidTab, err
		}
		// Add the result in the tab
		uuidTab = append(uuidTab, tmpUUID.String())
	}

	return uuidTab, nil
}

// getAllPersistentFlags is a private function allowing to get all the
// persistent flags defined inside the rp-cli utilities. The cmd parameter
// needs to be the earliest parent we want (the root command for example)
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param flagsList: Pointer to the list of flags to complete
// :type flagsList: *[]string
func getAllPersistentFlags(cmd *cobra.Command, flagsList *[]string) {
	// Get all the persistent flags for the current command
	cmd.PersistentFlags().VisitAll(func(flag *pflag.Flag) {
		// Add the name in our slice
		*flagsList = append(*flagsList, flag.Name)
	})

	// Get all the children commands
	childrenCmd := cmd.Commands()

	// For each children command, get the persistent flags
	for _, childCmd := range childrenCmd {
		getAllPersistentFlags(childCmd, flagsList)
	}
}

// getTabOfArchitectures is a private function allowing to get a tab
// of architectures available on the redpesk factory.
//
// :returns: ([]string, error) -- the tab of architectures (as string) and an error if something went bad
func getTabOfArchitectures() ([]string, error) {
	// Get the list of every available architecture
	archis := []rpapiv1.RPArch{}
	archStr := []string{}
	if err := bckAPI.RPArchGet(&archis, false); err != nil {
		return []string{}, err
	}
	for _, arch := range archis {
		archStr = append(archStr, arch.ID)
	}
	return archStr, nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Display functions --------------------------------------------
// --------------------------------------------------------------

// printStatus display the status of the current Redpesk instance
//
// :param status: Status to display in the console
// :type status: interface{}
// :param verbose: Tell if the print needs to be verbose or not
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func printStatus(status interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Get the server alias of the current connection
	serverAlias := viper.GetString(serverAliasFlag)

	// Convert to the right type
	rpStatus := status.(rpapiv1.RPAPIConfig)

	// Display the information
	fmt.Fprintln(writer, miscStatusAnsLine0)
	fmt.Fprintln(writer, miscStatusAnsLine0_1, serverAlias)
	fmt.Fprintln(writer, miscStatusAnsLine1, rpStatus.ServerUID)
	fmt.Fprintln(writer, miscStatusAnsLine2, rpStatus.URL)
	fmt.Fprintln(writer, miscStatusAnsLine3, rpStatus.ConnRetry)

	return nil
}

// printVersion display the version of the current Redpesk instance
//
// :param version: Version to display in the console
// :type version: interface{}
// :param verbose: Tell if the print needs to be verbose or not
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func printVersion(version interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Convert to the right type
	rpVersion := version.(rpapiv1.RPVersion)

	// Write the version in console and respect the verbose option
	fmt.Fprintln(writer, miscVersionAnsLine0)
	fmt.Fprintln(writer, miscVersionAnsLine1, rpVersion.ID)
	v := rpVersion.Version
	if verbose {
		v += " (" + rpVersion.VersionGitTag + ")"
	}
	fmt.Fprintln(writer, miscVersionAnsLine2, v)
	if verbose {
		fmt.Fprintln(writer, miscVersionAnsLine3, rpVersion.APIVersion)
	}

	return nil
}

// printArches display the list of the current arches available
// in the current Redpesk instance
//
// :param arches: List of arches to display
// :type arches: interface{}
// :param verbose: Tell if the print needs to be verbose or not
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func printArches(arches interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Convert to the right type
	archis := arches.([]rpapiv1.RPArch)

	if verbose {
		// Display each architecture
		fmt.Fprintln(writer, miscArchAnsLine0)
		for _, arch := range archis {
			fmt.Fprintln(writer, arch.ID)
		}
	} else {
		for index, arch := range archis {
			fmt.Fprint(writer, arch.ID)
			if index < (len(archis) - 1) {
				fmt.Fprint(writer, ",")
			}
		}
		fmt.Fprintln(writer)
	}

	return nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Banano's Touch -----------------------------------------------
// --------------------------------------------------------------

// beerCmdHandler is a refreshing command
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func beerCmdHandler(cmd *cobra.Command, args []string) error {

	RPPrinter.CustomColorPrintLn(printer.ColorCyan, "--The freshest beer around, YEC'HED MAT!--")

	// Split the string by line
	toDisplay := strings.Split(beerASCIIArt, "\n")
	for _, line := range toDisplay {
		RPPrinter.CustomColorPrintLn(printer.ColorYellow, line)
		time.Sleep(50 * time.Millisecond)
		RPPrinter.ForceFlush()
	}
	RPPrinter.CustomPrintLn()

	return nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------
