/*
Copyright © 2020-2023 IoT.bzh Company
Contact: https://www.iot.bzh/licensing

This file is part of the redpesk-cli project, the redpesk command line tool.

$RP_BEGIN_LICENSE$
Commercial License Usage

	Licensees holding valid commercial IoT.bzh licenses may use this file in
	accordance with the commercial license agreement provided with the
	Software or, alternatively, in accordance with the terms contained in
	a written agreement between you and The IoT.bzh Company. For licensing terms
	and conditions see https://www.iot.bzh/terms-conditions. For further
	information use the contact form at https://www.iot.bzh/contact.

GNU General Public License Usage

	Alternatively, this file may be used under the terms of the GNU General
	Public license version 3. This license is as published by the Free Software
	Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
	of this file. Please review the following information to ensure the GNU
	General Public License requirements will be met
	https://www.gnu.org/licenses/gpl-3.0.html.

$RP_END_LICENSE$
*/
package cmd

import (
	"fmt"
	"os"
	"regexp"
	"strings"
	"testing"

	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/rpbckmock"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/utils"
	"github.com/stretchr/testify/assert"
)

// TestMiscVersion is the function allowing to test if
// the command "rp-cli misc version" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestMiscVersion(test *testing.T) {

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		miscVersionAnsLine0: `.*`,
		"ID:":               `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Version:":          `([0-9]+.){2}[0-9]+`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, 3, regexpMap, miscCommand, miscVersionCmd)
}

// TestMiscVersionVerbose is the function allowing to test if
// the command "rp-cli misc version -v" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestMiscVersionVerbose(test *testing.T) {

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		miscVersionAnsLine0: `.*`,
		"ID:":               `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Version:":          `([0-9]+.){2}[0-9]+ \([0-9]+-[a-z0-9]{9}\)`,
		"API Version:":      `[0-9]+$`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, 4, regexpMap, miscCommand, miscVersionCmd, "-v")
}

// TestMiscStatus is the function allowing to test if
// the command "rp-cli misc status" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestMiscStatus(test *testing.T) {

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		miscVersionAnsLine0: `.*`,
		"Alias:":            `.*`,
		"ID:":               `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"URL:":              `http(s)*://`,
		"Connection retry:": `[0-9]+$`,
	}

	// Run the command and the tests (no verbose and verbose, since this is the same result)
	checkCmdAnswerKeysRegexp(test, 5, regexpMap, miscCommand, miscStatusCmd)
	checkCmdAnswerKeysRegexp(test, 5, regexpMap, miscCommand, miscStatusCmd, "-v")
}

// TestMiscStatusExpiredToken is the function allowing to test if
// the command "rp-cli misc status" is returning the right error
// message when the user tries a request with an expired token
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestMiscStatusExpiredToken(test *testing.T) {
	// Create a false token file
	tokenFilePath, err := createFalseTokenFile(nil)
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false token file at the end of the function
	defer deleteFalseFile(tokenFilePath)

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false configuration file at the end of the function
	defer deleteFalseFile(configFilePath)

	// Request the mocked backend to simulate an "unauthorized" error on next request
	rpbckmock.RequestUnauthError("Invalid username or password")

	// Execute the `rp-cli misc status` command
	cmdOutput, err := executeCommand(miscCommand, miscStatusCmd)

	// Test that the output of the command is empty
	assert.Empty(test, cmdOutput)

	// Test that the right error happened
	if assert.Error(test, err) {
		assert.Contains(test, err.Error(), "connection is expired, please generate a new one")
	}
}

// TestMiscArches is the function allowing to test if
// the command "rp-cli misc arch" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestMiscArches(test *testing.T) {

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"x86_64,aarch64": `.*`,
	}

	// Run the command and the tests (no verbose and verbose, since this is the same result)
	checkCmdAnswerKeysRegexp(test, 1, regexpMap, miscCommand, miscArchCmd)
}

// TestMiscArchesVerbose is the function allowing to test if
// the command "rp-cli misc arch -v" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestMiscArchesVerbose(test *testing.T) {

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		miscArchAnsLine0: `.*`,
		"x86_64":         `.*`,
		"aarch64":        `.*`,
	}

	// Run the command and the tests (no verbose and verbose, since this is the same result)
	checkCmdAnswerKeysRegexp(test, 3, regexpMap, miscCommand, miscArchCmd, "-v")
}

// TestMiscConfiguration is the function allowing to test if
// the command "rp-cli misc configuration" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestMiscConfiguration(test *testing.T) {

	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false file and restore the backup
	defer deleteFalseFile(configFilePath)

	// Create a false token file
	accessTokenFilePath, err := createFalseTokenFile(nil)
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false file created at the beginning
	defer deleteFalseFile(accessTokenFilePath)

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		configFileInfoPrefix:    `.+/rp-cli-config.json`,
		miscConfigAnsLine2:      `^mytestalias$`,
		miscConfigAnsLine1:      `.*`,
		miscConfigAnsLine3:      `.*`,
		miscConfigAnsLine4:      `^mytestalias$`,
		"RP_CONFIG:":            `[Not set]`,
		"RP_LOCALBUILDERALIAS:": `[Not set]`,
		"RP_LOGFILENAME:":       `[Not set]`,
		"RP_LOGLEVEL:":          `[Not set]`,
		"RP_NONBLOCKING:":       `[Not set]`,
		"RP_NOPAGER:":           `[Not set]`,
		"RP_RAWOUTPUT:":         `[Not set]`,
		"RP_SERVERALIAS:":       `[Not set]`,
		"RP_TIMESTAMP:":         `[Not set]`,
		"RP_PKGSDIR:":           `[Not set]`,
		"RP_SOURCESDIR:":        `[Not set]`,
		"RP_NOERROR:":           `[Not set]`,
		"RP_NOHEADER:":          `[Not set]`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, 18, regexpMap, miscCommand, miscConfigCmd)

	// Set two of the environment variables and check if it has changed in the configuration answer
	os.Setenv("RP_NONBLOCKING", "1")
	os.Setenv("RP_TIMESTAMP", "1")

	// Create the map containing the regexp we want to match in the answer
	regexpMap1 := map[string]string{
		configFileInfoPrefix:    `.+/rp-cli-config.json`,
		miscConfigAnsLine2:      `^mytestalias$`,
		miscConfigAnsLine1:      `.*`,
		miscConfigAnsLine3:      `.*`,
		miscConfigAnsLine4:      `^mytestalias$`,
		"RP_CONFIG:":            `[Not set]`,
		"RP_LOCALBUILDERALIAS:": `[Not set]`,
		"RP_LOGFILENAME:":       `[Not set]`,
		"RP_LOGLEVEL:":          `[Not set]`,
		"RP_NONBLOCKING:":       `1`,
		"RP_NOPAGER:":           `[Not set]`,
		"RP_RAWOUTPUT:":         `[Not set]`,
		"RP_SERVERALIAS:":       `[Not set]`,
		"RP_TIMESTAMP:":         `1`,
		"RP_PKGSDIR:":           `[Not set]`,
		"RP_SOURCESDIR:":        `[Not set]`,
		"RP_NOERROR:":           `[Not set]`,
		"RP_NOHEADER:":          `[Not set]`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, 18, regexpMap1, miscCommand, miscConfigCmd)

	// Unset the environment variable
	os.Unsetenv("RP_NONBLOCKING")
	os.Unsetenv("RP_TIMESTAMP")
}

// TestMiscConfigurationVerbose is the function allowing to test if
// the command "rp-cli misc configuration --verbose" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestMiscConfigurationVerbose(test *testing.T) {

	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false file and restore the backup
	defer deleteFalseFile(configFilePath)

	// Create a false token file
	accessTokenFilePath, err := createFalseTokenFile(nil)
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false file created at the beginning
	defer deleteFalseFile(accessTokenFilePath)

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		configFileInfoPrefix:    `.+/rp-cli-config.json`,
		miscConfigAnsLine0:      `.*`,
		"{":                     `.*`,
		"\"serveralias\"":       `mytestalias`,
		"\"pkgsdir\"":           `/tmp/redpesk-cli-test/gitpkgs`,
		"\"sourcesdir\"":        `/tmp/redpesk-cli-test/gitsources`,
		"}":                     `.*`,
		miscConfigAnsLine2:      `.*`,
		miscConfigAnsLine4:      `^mytestalias$`,
		miscConfigAnsLine1:      `.*`,
		miscConfigAnsLine3:      `.*`,
		"RP_CONFIG:":            `[Not set]`,
		"RP_LOCALBUILDERALIAS:": `[Not set]`,
		"RP_LOGFILENAME:":       `[Not set]`,
		"RP_LOGLEVEL:":          `[Not set]`,
		"RP_NONBLOCKING:":       `[Not set]`,
		"RP_NOPAGER:":           `[Not set]`,
		"RP_RAWOUTPUT:":         `[Not set]`,
		"RP_SERVERALIAS:":       `[Not set]`,
		"RP_TIMESTAMP:":         `[Not set]`,
		"RP_PKGSDIR:":           `[Not set]`,
		"RP_SOURCESDIR:":        `[Not set]`,
		"RP_NOERROR:":           `[Not set]`,
		"RP_NOHEADER:":          `[Not set]`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, 28, regexpMap, miscCommand, miscConfigCmd, "--verbose")
}

// TestOneBeerCmd is the function allowing to test if
// the command "rp-cli misc one-beer" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestOneBeerCmd(test *testing.T) {
	executeCommand("misc", "one-beer")
}

// TestMiscConfigurationGet is the function allowing to test if
// the command "rp-cli misc configuration get" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestMiscConfigurationGet(test *testing.T) {

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false file and restore the backup
	defer deleteFalseFile(configFilePath)

	// First, let's run the wanted command
	cmdOutput, err := executeCommand(miscCommand, miscConfigCmd, "get", "serveralias")
	if err != nil {
		test.Fatal(err)
	}
	cmdOutputTrimmed := strings.TrimSpace(cmdOutput)

	// Check that we have the test alias as an answer
	assert.Equal(test, "mytestalias", cmdOutputTrimmed)

	// Run another get
	cmdOutput, err = executeCommand(miscCommand, miscConfigCmd, "get", "pkgsdir")
	if err != nil {
		test.Fatal(err)
	}
	cmdOutputTrimmed = strings.TrimSpace(cmdOutput)

	// Check that we have the test alias as an answer
	assert.Equal(test, "/tmp/redpesk-cli-test/gitpkgs", cmdOutputTrimmed)
}

// TestMiscConfigurationSet is the function allowing to test if
// the command "rp-cli misc configuration set" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestMiscConfigurationSet(test *testing.T) {

	// Create a false configuration file
	configFilePath, err := createFalseConfigFile()
	if err != nil {
		test.Fatal(err)
	}

	// Delete the false file and restore the backup
	defer deleteFalseFile(configFilePath)

	// First, let's run the wanted command
	_, err = executeCommand(miscCommand, miscConfigCmd, "set", "pkgsdir", "--value", "/tmp")
	if err != nil {
		test.Fatal(err)
	}

	// Run the get on the pkgsdir
	cmdOutput, err := executeCommand(miscCommand, miscConfigCmd, "get", "pkgsdir")
	if err != nil {
		test.Fatal(err)
	}
	cmdOutputTrimmed := strings.TrimSpace(cmdOutput)

	// Check that we have the test alias as an answer
	assert.Equal(test, "/tmp", cmdOutputTrimmed)
}

// TestMiscCmdTreeCmd is the function allowing to test if
// the command "rp-cli misc cmd-tree" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestMiscCmdTreeCmd(test *testing.T) {
	// Run the cmd-tree command
	output, err := executeCommand("misc", "cmd-tree")

	// Check that the error is nil
	errIsNil := assert.Nil(test, err)
	if !errIsNil {
		test.Fatal(err)
	}

	// Check that the output contains few things
	assert.Contains(test, output, "rp-cli")
	assert.Contains(test, output, "applications")
	assert.Contains(test, output, "projects")
	assert.Contains(test, output, "images")
}

// TestMiscCmdSearchSimple is the function allowing to test if
// the command "rp-cli misc cmd-search" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestMiscCmdSearchSimple(test *testing.T) {
	// Run the cmd-tree command
	output, err := executeCommand("misc", "cmd-search")

	// Check that the error is nil
	errIsNil := assert.Nil(test, err)
	if !errIsNil {
		test.Fatal(err)
	}

	// Check that the output contains few things
	assert.Contains(test, output, "rp-cli applications")
	assert.Contains(test, output, "rp-cli projects")
	assert.Contains(test, output, "rp-cli images")
}

// TestMiscCmdSearchContains is the function allowing to test if
// the command "rp-cli misc cmd-search <pattern>" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestMiscCmdSearchContains(test *testing.T) {
	// Run the cmd-tree command
	output, err := executeCommand("misc", "cmd-search", "releases")

	// Check that the error is nil
	errIsNil := assert.Nil(test, err)
	if !errIsNil {
		test.Fatal(err)
	}

	trimmedOutput := strings.TrimSpace(output)

	// Check that each line contains "releases"
	lineSlice := strings.Split(trimmedOutput, "\n")
	for _, line := range lineSlice {
		assert.Contains(test, line, "releases")
	}
}

// TestMiscCmdSearchRegexp is the function allowing to test if
// the command "rp-cli misc cmd-search <pattern> --regexp" is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestMiscCmdSearchRegexp(test *testing.T) {
	// Run the cmd-tree command
	output, err := executeCommand("misc", "cmd-search", "^rp-cli projects", "--regexp")

	// Check that the error is nil
	errIsNil := assert.Nil(test, err)
	if !errIsNil {
		test.Fatal(err)
	}

	trimmedOutput := strings.TrimSpace(output)

	// Check that each line match the regexp "^rp-cli projects"
	lineSlice := strings.Split(trimmedOutput, "\n")
	regularExp := regexp.MustCompile(`^rp-cli projects`)
	for _, line := range lineSlice {
		assert.True(test, regularExp.MatchString(line), fmt.Sprintf("%s does not match %s", line, "^rp-cli projects"))
	}
}

// --------------------------------------------------------------
// Utilities functions ------------------------------------------
// --------------------------------------------------------------

// checkServicesParamsTypesValues is a function allowing to check, from a rp-cli output
// that a parameter has the right type and eventually value
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param cmdOutput: Output comming from rp-cli command line
// :type cmdOutput: string
// :param paramName: Name of the parameter to check
// :type paramName: string
// :param typeRegexp: Regexp that must match with the parameter type
// :type typeRegexp: string
// :param valueRegexp: Regexp that must match with the parameter value (can be empty if no check needs top be done)
// :type valueRegexp: string
// :returns: bool -- True if the assertions are good, False in the other case
func checkServicesParamsTypesValues(test *testing.T, cmdOutput, paramName, typeRegexp, valueRegexp string) bool {
	retValue := true

	// Split the answer by line
	outputLines := strings.Split(cmdOutput, "\n")

	for _, line := range outputLines {
		// Get rid of space, tab, new line at the end and at the beginning of the line
		lineNoSpace := strings.TrimSpace(line)

		// Check if this is the line containing our parameter
		if strings.HasPrefix(lineNoSpace, paramName) {
			// This is our line! Divide it by space and delete empty strings
			tmpLineSlice := strings.Split(lineNoSpace, " ")
			tmpLineSlice = utils.DeleteEmptyStringFromSlice(tmpLineSlice)

			// Index that will allow us to know where we are in the check
			innerLineindex := 0

			if typeRegexp != "" {
				innerLineindex++
				// Check that the type is what we want
				if assert.GreaterOrEqual(test, len(tmpLineSlice), 2) {
					trimmedTypeStr := strings.TrimSpace(tmpLineSlice[innerLineindex])
					retValue = (retValue && assert.Regexp(test, regexp.MustCompile(typeRegexp), trimmedTypeStr))
				} else {
					fmt.Println("A problem appeared while checking the type of", paramName)
				}
			}

			// If requested by the caller, check also the value
			if valueRegexp != "" {
				innerLineindex++
				if assert.GreaterOrEqual(test, len(tmpLineSlice), (innerLineindex + 1)) {
					valueStr := strings.Join(tmpLineSlice[innerLineindex:], " ")
					trimmedValueStr := strings.TrimSpace(valueStr)
					retValue = (retValue && assert.Regexp(test, regexp.MustCompile(valueRegexp), trimmedValueStr))
				} else {
					fmt.Println("A problem appeared while checking the value of", paramName)
				}
			}
			// We have found the right line, end of the function
			break
		}
	}

	return retValue
}

// --------------------------------------------------------------
// --------------------------------------------------------------
