/*
Copyright © 2020-2023 IoT.bzh Company
Contact: https://www.iot.bzh/licensing

This file is part of the redpesk-cli project, the redpesk command line tool.

$RP_BEGIN_LICENSE$
# Commercial License Usage

Licensees holding valid commercial IoT.bzh licenses may use this file in
accordance with the commercial license agreement provided with the
Software or, alternatively, in accordance with the terms contained in
a written agreement between you and The IoT.bzh Company. For licensing terms
and conditions see https://www.iot.bzh/terms-conditions. For further
information use the contact form at https://www.iot.bzh/contact.

# GNU General Public License Usage

Alternatively, this file may be used under the terms of the GNU General
Public license version 3. This license is as published by the Free Software
Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
of this file. Please review the following information to ensure the GNU
General Public License requirements will be met
https://www.gnu.org/licenses/gpl-3.0.html.

$RP_END_LICENSE$
*/
package cmd

import (
	"bytes"
	"fmt"
	"os"
	"path"
	"runtime/debug"
	"strconv"
	"strings"
	"syscall"
	"text/tabwriter"

	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/printer"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/rpconnect"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/rpxxxapi/bckapi"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/rpxxxapi/ladapi"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/signals"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/utils"
	"git.ovh.iot/redpesk/redpesk-infra/rp-golang-common/rputils"
	"git.ovh.iot/redpesk/redpesk-infra/rp-webserver/actions/rpapiv1"
	"github.com/gofrs/uuid"
	gosocketio "github.com/graarh/golang-socketio"
	homedir "github.com/mitchellh/go-homedir"
	"github.com/pkg/errors"
	"github.com/sirupsen/logrus"
	"github.com/spf13/cobra"
	"github.com/spf13/pflag"
	"github.com/spf13/viper"
)

// --------------------------------------------------------------
// Public global variables --------------------------------------
// --------------------------------------------------------------

// AppName name of this application
var AppName = applicationName

// AppVersion Version of this application  (set by Makefile)
var AppVersion = "?.?.?"

// AppSubVersion is the git tag id added to version string
// Should be set by compilation -ldflags "-X main.AppSubVersion=xxx"
// (set by Makefile)
var AppSubVersion = "unknown-dev"

// AppBuildDateYear is the git tag id added to version string
// Should be set by compilation -ldflags "-X main.AppBuildDateYear=xxx"
// (set by Makefile)
var AppBuildDateYear = "2022"

// BuildMode is the mode used to build the application
// Indeed, some behavior of the application may be different
// according to the mode chosen.
// By default, this mode is "development", the other possible value
// is "production"
var BuildMode = "development"

// Log global variable that hold the logger
var Log = logrus.New()

// RPPrinter : printer interface used to display info to the user
var RPPrinter *printer.RPPrinter

// No pager variable for test purpose
var NoPagerTest = true

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Private global variables --------------------------------------
// --------------------------------------------------------------

// runMode is the mode in which the application is running.
// It can be "standard" or "testing"
var runMode = "standard"

// Global variable used in case of tests
var testsBckURL = ""

// Global variable used in case of tests
var testInteractReq = false

// Global variable used in case of tests
var testsLADURL = ""

// Global variable used in case of tests
var signalTestBeingRun = false

// Bytes buffer used by tests to get the output of rp-cli
var testingOutputBuffer = bytes.NewBufferString("")

// Authors
var appAuthors = []string{
	author1,
	author2,
}

// Variable used to store the configuration file path
var cfgFile string

// Variable used to store the possible error that happened during configuration file load
var cfgFileErr error

// Variable used to store the configuration directory path
var cfgDirPath = ""

// bckConnection: connection object to the backend
var bckConnection *rpconnect.RPCLIExtConnection

// localBConnection: connection object to the local-builder API
var localBConnection *rpconnect.RPCLIExtConnection

// API interface for requests to backend
var bckAPI *bckapi.RPBCKAPI

// API interface for requests to local-builder
var localBAPI *ladapi.RPLADAPI

// Global variable used to know if a "no-verbose" header has already been printed
var headerAlreadyPrinted bool

// Original arguments saved if we had to change the user arguments behind the back of Cobra
var originalArgs = []string{}

// Suffixes to be checked in user input
var suffixesToCheck = []string{
	"local-v2 package service",
	"local package service",
	"local-v2 package metadata",
	"local package metadata",
	"local-v2 package build",
	"local package build",
}

// Global private variable allowing to know if the user is admin on at least one stack
var isUserAdmin = false

// --------------------------------------------------------------
// --------------------------------------------------------------

// Execute adds all child commands to the root command and sets flags appropriately.
// This is called by main.main(). It only needs to happen once to the rootCmd.
func Execute() {
	// Add a defer function to catch panics
	defer gracefulRecoverFromPanic(true)

	// Create the root command to use
	rootCmd := RootCmd()

	// Function allowing to rework some arguments if needed
	reworkedArgs := reworksArgsIfNeeded([]string{})
	if len(reworkedArgs) > 0 {
		rootCmd.SetArgs(reworkedArgs)
	}

	if err := rootCmd.Execute(); err != nil {
		//fmt.Println(err)
		endProperlyCLI(true, 1)
	}
	endProperlyCLI(true, 0)
}

// RootCmd allows to create the root command for rp-cli
func RootCmd() *cobra.Command {

	// Create the main command
	mainCmd := &cobra.Command{
		Use:   AppName,
		Short: appShortDesc,
		Long:  appLongDesc,

		SilenceUsage:      true,
		Example:           appExample,
		Version:           AppVersion + " (" + AppSubVersion + ")",
		PersistentPreRunE: mainPersistentPreRunE,
	}

	// Initialize the main command
	initRootCmd(mainCmd)

	// Initialize the global variable allowing to know if the user is admin on at least one stack
	var err error
	isUserAdmin, err = isUserAdminOnAnyStack()
	if err != nil {
		// Log that we had trouble to see if the user is admin
		Log.Warn(adminCheckProblem)
	}

	// Initialize all the subcommands
	initAppCmd(mainCmd)
	initCompCmd(mainCmd)
	initMiscCmd(mainCmd)
	initDistributionsCmd(mainCmd)
	initPrjCmd(mainCmd)
	initTeamsCmd(mainCmd)
	initUsersCmd(mainCmd)
	initLocalNewCmd(mainCmd)
	initIotpkgCmd(mainCmd)
	initImagesCmd(mainCmd)
	initBoardModelsCmd(mainCmd)
	initBoardsCmd(mainCmd)
	initDeploymentsCmd(mainCmd)
	initConnectionsCmd(mainCmd)

	// Initialize the "admin" subcommand
	// (hidden if not an admin of the current factory)
	initAdminCmd(mainCmd)

	return mainCmd
}

// reworksArgsIfNeeded get the args given by the user in the command line,
// check if it match a pattern that needs to be reworked, and if so, give the new
// args instead
//
// :param args: Allow the caller to give manual args (used for tests purpose)
// :type args: []string
// :returns: []string -- new args to be used
func reworksArgsIfNeeded(args []string) []string {
	var origArgs []string

	// Firstly get the args given by the user
	if len(args) > 0 {
		origArgs = args[0:]
	} else {
		origArgs = os.Args[1:]
	}

	currentArgs := []string{}
	currentArgs = append(currentArgs, origArgs...)

	// Join all the args
	allCurrentArgs := strings.Join(currentArgs, " ")

	// Check if the args match one suffix to monitor
	for _, sufToCheck := range suffixesToCheck {
		if strings.Contains(allCurrentArgs, sufToCheck) {
			// Check if the args contain "--help"
			if utils.ContainsString(currentArgs, "--help") {
				utils.DeleteStringFromSlice(&currentArgs, "--help")
			}
			// Check if the args contain "-h"
			if utils.ContainsString(currentArgs, "-h") {
				utils.DeleteStringFromSlice(&currentArgs, "-h")
			}
			// Check if the args contain "help"
			if utils.ContainsString(currentArgs, "help") {
				utils.DeleteStringFromSlice(&currentArgs, "help")
			}
			break
		}
	}

	// If the args have been reworked, return them
	if len(currentArgs) < len(origArgs) {
		originalArgs = origArgs[0:]
		return currentArgs
	}
	return []string{}
}

func initRootCmd(rootCmd *cobra.Command) {

	cobra.OnInitialize(initConfig)

	// Add Copyright and Authors names and  inside the rootCmd description
	allAuthors := "Authors:\n  " + strings.Join(appAuthors[:], "\n  ")
	rootCmd.Long = rootCmd.Long + "\n\n" + fmt.Sprintf(appCopyright, AppBuildDateYear) + "\n\n" + allAuthors

	// onBoardCmd represents the "onboard" command
	var onBoardCmd = &cobra.Command{
		Use:         rootOnBoardCmd,
		Short:       rootOnBoardCmdDesc,
		RunE:        rpOnBoard,
		Annotations: map[string]string{"__bck_connect__": "FALSE"}, // No need for backend connection for this command
	}

	// Add the command "onboard" as subcommand of rootCmd
	rootCmd.AddCommand(onBoardCmd)

	// Persistent Flags -----------------------------------------
	rootCmd.PersistentFlags().StringVarP(&cfgFile, configFileFlag, configFileFlagAlias, "", configFileFlagUsage)
	rootCmd.PersistentFlags().StringP(logLevelFlag, logLevelFlagAlias, defaultLogLevel, logLevelFlagUsage)
	rootCmd.PersistentFlags().String(logFileFlag, logFileFlagDefault, logFileFlagUsage)
	rootCmd.PersistentFlags().String(serverAliasFlag, "default", serverAliasFlagUsage)
	rootCmd.PersistentFlags().String(localBuilderAliasFlag, localBuilderAlias, localBuilderAliasFlagUsage)
	rootCmd.PersistentFlags().BoolP(timestampFlag, timestampFlagAlias, timestampFlagDefault, timestampFlagUsage)
	rootCmd.PersistentFlags().Bool(nonBlockingFlag, nonBlockingFlagDefault, nonBlockingFlagUsage)
	rootCmd.PersistentFlags().Bool(rawOutputFlag, rawOutputFlagDefault, rawOutputFlagUsage)
	rootCmd.PersistentFlags().Bool(noErrorFlag, noErrorFlagDefault, noErrorFlagUsage)
	rootCmd.PersistentFlags().BoolP(noPagerFlag, noPagerFlagAlias, noPagerFlagDefault, noPagerFlagUsage)
	rootCmd.PersistentFlags().Bool(NoHeaderFlag, NoHeaderFlagDefault, NoHeaderFlagUsage)
	// ----------------------------------------------------------

	// Bind the cobra flags to viper
	viper.BindPFlag(configFileFlag, rootCmd.PersistentFlags().Lookup(configFileFlag))               // Corresponding environment var: RP_CONFIG
	viper.BindPFlag(logLevelFlag, rootCmd.PersistentFlags().Lookup(logLevelFlag))                   // Corresponding environment var: RP_LOGLEVEL
	viper.BindPFlag(logFileFlag, rootCmd.PersistentFlags().Lookup(logFileFlag))                     // Corresponding environment var: RP_LOGFILENAME
	viper.BindPFlag(serverAliasFlag, rootCmd.PersistentFlags().Lookup(serverAliasFlag))             // Corresponding environment var: RP_SERVERALIAS
	viper.BindPFlag(localBuilderAliasFlag, rootCmd.PersistentFlags().Lookup(localBuilderAliasFlag)) // Corresponding environment var: RP_LOCALBUILDERALIAS
	viper.BindPFlag(timestampFlag, rootCmd.PersistentFlags().Lookup(timestampFlag))                 // Corresponding environment var: RP_TIMESTAMP
	viper.BindPFlag(nonBlockingFlag, rootCmd.PersistentFlags().Lookup(nonBlockingFlag))             // Corresponding environment var: RP_NONBLOCKING
	viper.BindPFlag(rawOutputFlag, rootCmd.PersistentFlags().Lookup(rawOutputFlag))                 // Corresponding environment var: RP_RAWOUTPUT
	viper.BindPFlag(noErrorFlag, rootCmd.PersistentFlags().Lookup(noErrorFlag))                     // Corresponding environment var: RP_NOERROR
	viper.BindPFlag(noPagerFlag, rootCmd.PersistentFlags().Lookup(noPagerFlag))                     // Corresponding environment var: RP_NOPAGER
	viper.BindPFlag(NoHeaderFlag, rootCmd.PersistentFlags().Lookup(NoHeaderFlag))                   // Corresponding environment var: RP_NOHEADER

	// Local Flags ----------------------------------------------
	// ----------------------------------------------------------

	// Specify dynamic flag completion
	rootCmd.RegisterFlagCompletionFunc(serverAliasFlag, getRPFactoryAliases)
	rootCmd.RegisterFlagCompletionFunc(localBuilderAliasFlag, getLocalBuilderAliases)
}

// initConfig reads in config file and ENV variables if set.
func initConfig() {

	// Get the configuration file
	configFileHomeFullPath, err := getConfigDir()
	if err != nil {
		fmt.Println(err)
		endProperlyCLI(true, 1)
	}

	// Check if the configuration directories exists
	if _, err = os.Stat(configFileHomeFullPath + cookiesDirRelativePath); os.IsNotExist(err) {
		// Create it if it does not exist
		os.MkdirAll(configFileHomeFullPath+cookiesDirRelativePath, 0755)
	}

	if cfgFile != "" {
		// Use config file from the flag.
		viper.SetConfigFile(cfgFile)

	} else {

		// Check if the "RP_CONFIG" environment variable is set
		tmpCfgFile := os.Getenv(configFileEnvVar)

		if tmpCfgFile != "" {
			// Use config file from the variable environment.
			viper.SetConfigFile(tmpCfgFile)

		} else {
			// Search config in home directory with name "rp-cli-config" (without extension).
			viper.AddConfigPath(configFileHomeFullPath)
			viper.AddConfigPath(configFileEtcPath)
			viper.SetConfigName(configFileDefault)
		}
	}

	// Set the prefix "RP_" for environment variables used for this cli
	viper.SetEnvPrefix(envPrefix)
	// Cobra will looks, through viper for environment variable when looking for a configuration
	// For example, it will look for the env var RP_LOGLEVEL to get the loglevel flag
	viper.AutomaticEnv() // read in environment variables that match

	// If a config file is found, read it in
	cfgFileErr = viper.ReadInConfig()
}

func mainPersistentPreRunE(cmd *cobra.Command, args []string) error {
	var err error
	var msgErr string

	// Check if the configuration file load has encountered an error
	if cfgFileErr != nil {
		cfgFileErrLow := strings.ToLower(cfgFileErr.Error())
		// Check if the error is about the file not existing
		if !(strings.Contains(cfgFileErrLow, "config file") && strings.Contains(cfgFileErrLow, "not found")) {
			// The error is not about the the file not existing, but something else, return it
			return fmt.Errorf(errOnConfigFileReading, cfgFileErr.Error())
		}
	}

	// Get the log level wanted
	loglevel := viper.GetString(logLevelFlag)

	// Check if we are in raw mode or not
	rawMode := viper.GetBool(rawOutputFlag)
	// Check if we are in "no pager" mode or not
	noPager := viper.GetBool(noPagerFlag)

	// Set logger level and formatter
	if Log.Level, err = logrus.ParseLevel(loglevel); err != nil {
		msg := fmt.Sprintf(logLevelInvalidMsg, loglevel)
		return errors.New(msg)
	}
	Log.Formatter = &logrus.TextFormatter{}

	// Get the log file wanted
	logfilename := viper.GetString(logFileFlag)

	// Create log file if not stderr
	if logfilename != logFileFlagDefault {
		logFile, _ := rputils.ResolveEnvVar(logfilename)
		fdL, err := os.OpenFile(logFile, os.O_CREATE|os.O_TRUNC|os.O_WRONLY, 0666)
		if err != nil {
			msgErr = fmt.Sprintf(logFileCreationErr, logFile)
			return errors.New(msgErr)
		}
		Log.Infof(logFileInfoPrefix, logFile)
		Log.Out = fdL
	}

	Log.Infof(versionInfoMsg, AppVersion)
	Log.Debugf(envMsgPrefix, os.Environ())

	// Initialize the global variable used to manage header printing
	headerAlreadyPrinted = false

	// Start signals monitoring routine only if not in tests
	if (runMode != "testing") || signalTestBeingRun {
		signals.MonitorSignals(Log)
	}

	// Default callback to handle interrupt signal
	// Maybe be overwritten by some subcommand (eg. targets commands)
	err = signals.OnSignals(func(sig os.Signal) {
		Log.Debugf(processSignalSendPrefix, sig)
		if signals.IsInterruptSignal(sig) {
			endProperlyCLI(true, int(syscall.EINTR))
		}
	})
	if err != nil {
		msgErr = fmt.Sprintf(processSignalCallBackErr, err)
		return errors.New(msgErr)
	}

	// Initialize the printer interface
	if runMode == "testing" {
		RPPrinter, err = printer.InitRPPrinter(Log, rawMode, NoPagerTest, true, testingOutputBuffer)
		Log.Out = testingOutputBuffer
	} else {
		RPPrinter, err = printer.InitRPPrinter(Log, rawMode, noPager, false, nil)
	}
	if err != nil {
		return err
	}

	// Get both booleans needed to know if we need to connect to the backend or to the localbuilder
	backendConnectNecessary := isBCKConnectNecessary(cmd, args)
	localbuilderConnectNecessary := isLADConnectNecessary(cmd)

	// Verify if there is a "--help" or "-h" as a string flag value in the command given by user
	if isThereHelpInFlagValue(cmd) {
		cmd.RunE = func(cmd *cobra.Command, args []string) error {
			return cmd.Help()
		}
		cmd.PreRunE = func(cmd *cobra.Command, args []string) error {
			return nil
		}
		backendConnectNecessary = false
		localbuilderConnectNecessary = false
	}

	var serverAlias string
	var connectBCKParams rpconnect.RPCLIConnection
	if backendConnectNecessary || localbuilderConnectNecessary {
		// Get the alias of the remote connection
		serverAlias = viper.GetString(serverAliasFlag)
		// Get the connection parameters
		connectBCKParams, err = getConnectionParams(serverAlias, accessTokenFile)
		if err != nil {
			return err
		}
	}

	// Init the connection with backend only if it is necessary
	if backendConnectNecessary {
		// Check if we are in test mode and set the right address if it is the case
		if (runMode == "testing") && (testsBckURL == "") {
			// Set the mocked backend address
			connectBCKParams.ServerURL = "http://0.0.0.0:8111"
			connectBCKParams.Alias = "tests-BCK"
		} else if (runMode == "testing") && (testsBckURL != "") {
			// Set the real backend for tests
			connectBCKParams.ServerURL = testsBckURL
		}

		// Check if this function is called by the completion command
		var completionCall bool
		if strings.Contains(cmd.Use, "__complete") {
			completionCall = true
		} else {
			completionCall = false
		}

		// Get the configuration directory
		cfgDir, err := getConfigDir()
		if err != nil {
			return err
		}

		// Check that we have an access token
		if (connectBCKParams.AccessToken == "" && !completionCall && (runMode != "testing")) || testInteractReq {
			// There is no accessToken, request it from the user
			err = requestAccessToken(&connectBCKParams)
			if err != nil {
				return err
			}

			// Since there is a new access token, delete the cookie file if it exists
			if err = rpconnect.RPDelCookieFile(cfgDir, connectBCKParams.Alias); err != nil {
				return err
			}

		} else if connectBCKParams.AccessToken == "" && completionCall && (runMode != "testing") {
			return errors.New(noInteractiveTokenInCompletion)
		}
		Log.Debugf(accessTokenDebugMsg, connectBCKParams.AccessToken)

		// Prepare the API interface structure
		bckAPI = &bckapi.RPBCKAPI{
			HTTPCli: nil,
			Log:     Log,
		}

		// Check if we need to open the web-socket, or not
		webSocketNeeded := isWSBCKConnectNecessary(cmd, args)

		if bckConnection, err = rpconnect.RPConnInit(&connectBCKParams, cfgDir, bckAPI, webSocketNeeded, []string{rpapiv1.EVTError, rpapiv1.EVTSessionExpired}, CLIReceiveEvent); err != nil {
			// Directly call HandleExitCoder to avoid to print help (ShowAppHelp)
			// Note that this function wil never return and program will exit
			return err
		}
	}

	// Initialize the connection to the local builder only if it is necessary
	if localbuilderConnectNecessary {
		// Get the alias of the remote connection
		localBuilderServerAlias := viper.GetString(localBuilderAliasFlag)

		// Get the connection parameters
		connecParamsLB, err := getConnectionParams(localBuilderServerAlias, localBuilderFile)
		if err != nil {
			return err
		}

		// Check if there are custom localbuilder settings
		if connectBCKParams.LocalBuilderURL != "" {
			connecParamsLB.ServerURL = connectBCKParams.LocalBuilderURL
		}

		// Check that a connection has been found
		if (connecParamsLB.ServerURL == "") && (runMode != "testing") {
			return errors.New(noLocalServerURLError)
		}

		// Check that the URL has a port and force it if not
		urlPort := utils.GetURLPort(connecParamsLB.ServerURL)
		if urlPort == "" {
			connecParamsLB.ServerURL = connecParamsLB.ServerURL + ":8080"
		}

		// Check if we are in test mode and set the right address if it is the case
		if (runMode == "testing") && (testsLADURL == "") {
			// Set the mocked redpesk-lad address
			connecParamsLB.ServerURL = "http://0.0.0.0:8112"
			connecParamsLB.Alias = "tests-LAD"
		}

		// Get the root directory inside the local-builder
		if connectBCKParams.LocalBuilderRootDir != "" {
			// Save it in a global variable (located in local.go)
			localBuilderRootDir = connectBCKParams.LocalBuilderRootDir
		}

		// Get the configuration directory
		cfgDir, err := getConfigDir()
		if err != nil {
			return err
		}

		// Prepare the API interface structure
		localBAPI = &ladapi.RPLADAPI{
			HTTPCli: nil,
			Log:     Log,
		}

		// Logs
		Log.Debugf("Connection to localbuilder on: %s", connecParamsLB.ServerURL)

		if localBConnection, err = rpconnect.RPConnInit(&connecParamsLB, cfgDir, localBAPI, true, []string{}, nil); err != nil {
			// Directly call HandleExitCoder to avoid to print help (ShowAppHelp)
			// Note that this function wil never return and program will exit
			return err
		}
	}

	return nil
}

// rpOnBoard creates interactively the configuration files
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func rpOnBoard(cmd *cobra.Command, args []string) error {
	// Request if the client wants to configure the backend connection
	answerLow := rpYesOrNoQuestion(rootOnBoardAnsLineReq0)
	if (answerLow == "y") || (answerLow == "") {
		RPPrinter.CustomPrintLn()
		err := rpOnBoardBck(cmd, args)
		if err != nil {
			return err
		}
	}

	// Request for the modification of local-builder connection
	RPPrinter.CustomPrintLn()
	RPPrinter.CustomPrintLn()
	answerLow = rpYesOrNoQuestion(rootOnBoardAnsLineReq1)
	if (answerLow == "y") || (answerLow == "") {
		RPPrinter.CustomPrintLn()
		err := rpOnBoardLocal(cmd, args)
		if err != nil {
			return err
		}
	} else if answerLow != "n" {
		return errors.New(yesNoNotRespectedErrorMsg)
	}

	return nil
}

// rpOnBoardBck configures interactively the connections
// parameters with a remote backend
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func rpOnBoardBck(cmd *cobra.Command, args []string) error {
	// Print the first line for this line
	RPPrinter.CustomPrintLn(rootOnBoardAnsLine0)
	RPPrinter.CustomPrintLn()

	// Get all the available connection
	availableConnect, err := getAllConnectionsParams(accessTokenFile)
	if err != nil {
		return err
	}

	// Display the current available connections
	if len(availableConnect) <= 0 {
		RPPrinter.CustomPrintLn(rootOnBoardNoConnectAvailLine)
		RPPrinter.CustomPrintLn()
	} else {
		RPPrinter.CustomPrintLn(rootOnBoardAvailConnectLine)
		RPPrinter.CustomPrintLn()
		for _, connection := range availableConnect {
			RPPrinter.CustomPrintLn(connectionAliasLabel, connection.Alias)
			RPPrinter.CustomPrintLn(connectionURLLabel, connection.ServerURL)
			RPPrinter.CustomPrintLn(connectionTokenLabel, connection.AccessToken)
			RPPrinter.CustomPrintLn()
		}
	}

	// Request for the alias to change/create
	var alias string
	RPPrinter.CustomPrintLn(rootOnBoardAnsAliasReq1)
	RPPrinter.CustomPrintLn(rootOnBoardAnsAliasReq2)
	RPPrinter.CustomPrint(rpCliPrompt)
	RPPrinter.ForceFlush()
	// Get the input
	fmt.Scanln(&alias)
	if alias == localBuilderAlias {
		return errors.New(localBuilderAliasForbidden)
	}
	if alias == "" {
		alias = "default"
	}

	// Ask the user for the URL of the Redpesk backend server
	var serverURL string
	RPPrinter.CustomPrintLn(rootOnBoardAnsURLReq)
	RPPrinter.CustomPrint(rpCliPrompt)
	RPPrinter.ForceFlush()
	// Get the input
	fmt.Scanln(&serverURL)

	// Ask the user for the access token
	var accessToken string
	RPPrinter.CustomPrintF(createTokenMsg0, serverURL)
	RPPrinter.CustomPrintLn(createTokenMsg1)
	RPPrinter.CustomPrint(rpCliPrompt)
	RPPrinter.ForceFlush()
	// Get the input
	fmt.Scanln(&accessToken)

	// Get the configuration directory
	cfgDir, err := getConfigDir()
	if err != nil {
		return err
	}

	// Print the fact that we are deleting the previous configuration files
	RPPrinter.CustomPrint(rootOnBoardAnsLine1)

	// Delete the cookie file if it exists
	if err = rpconnect.RPDelCookieFile(cfgDir, alias); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}
	// We've deleted successfully the cookie file
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	// Create/Update the configuration file
	RPPrinter.CustomPrint(rootOnBoardAnsLine2)
	err = initOrUpdateDefCfgFile(serverAliasFlag, alias)
	if err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	// Append in the token access file or overwrite
	connectParams := rpconnect.RPCLIConnection{
		Alias:       alias,
		ServerURL:   serverURL,
		AccessToken: accessToken,
		IsAdmin:     false,
	}
	RPPrinter.CustomPrint(rootOnBoardAnsLine3)
	err = appendOrOverwriteInTokenFile(accessTokenFile, connectParams)
	if err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	return nil
}

// rpOnBoardLocal configures interactively the connections
// parameters with the local builder
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func rpOnBoardLocal(cmd *cobra.Command, args []string) error {
	// Print the first line for this line
	RPPrinter.CustomPrintLn(rootOnBoardLocalAnsLine0)
	RPPrinter.CustomPrintLn()

	// Get all the available connection
	availableConnect, err := getAllConnectionsParams(localBuilderFile)
	if err != nil {
		return err
	}

	// Display the current available connections
	if len(availableConnect) <= 0 {
		RPPrinter.CustomPrintLn(rootOnBoardNoConnectAvailLine)
		RPPrinter.CustomPrintLn()
	} else {
		RPPrinter.CustomPrintLn(rootOnBoardAvailConnectLine)
		RPPrinter.CustomPrintLn()
		for _, connection := range availableConnect {
			RPPrinter.CustomPrintLn(connectionAliasLabel, connection.Alias)
			RPPrinter.CustomPrintLn(connectionURLLabel, connection.ServerURL)
			RPPrinter.CustomPrintLn()
		}
	}

	// Get the configuration directory
	cfgDir, err := getConfigDir()
	if err != nil {
		return err
	}

	// Ask the user for the URL of the Redpesk backend server
	var localBuilderIP string
	var answerToQuestion string
	IPseemsOK := false

	for !IPseemsOK {
		RPPrinter.CustomPrintLn(rootOnBoardAnsURLLocalReq)
		RPPrinter.CustomPrint(rpCliPrompt)
		RPPrinter.ForceFlush()
		// Get the input
		fmt.Scanln(&localBuilderIP)

		// Check that the connection to local builder is good
		IPseemsOK, _ = quickConnectionCheck(localBuilderIP, "8080", "version")

		if !IPseemsOK {
			RPPrinter.CustomPrintLn()
			RPPrinter.CustomPrintF(rootOnBoardAnsFailIPCheck, localBuilderIP)
			RPPrinter.CustomPrint(rpCliPrompt)
			RPPrinter.ForceFlush()
			// Get the input
			fmt.Scanln(&answerToQuestion)
			// Convert to lowercase
			answerLow := strings.ToLower(answerToQuestion)
			// Check the answer
			if answerLow == "n" {
				// The user wants to keep his IP
				IPseemsOK = true
			}
		}
	}

	// Delete the cookie file if it exists
	if err = rpconnect.RPDelCookieFile(cfgDir, localBuilderAlias); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	// Append in the token access file or overwrite
	connectParams := rpconnect.RPCLIConnection{
		Alias:       localBuilderAlias,
		ServerURL:   localBuilderIP,
		AccessToken: "",
		IsAdmin:     false,
	}

	// Modify the local builder
	RPPrinter.CustomPrint(rootOnBoardAnsLocalLine3)
	err = appendOrOverwriteInTokenFile(localBuilderFile, connectParams)
	if err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	// Get the home directory
	home, err := homedir.Dir()
	if err != nil {
		return err
	}

	// Check the local builder mounted directories
	err = rpOnBoardLocalMountedDir(path.Join(home, "my_rp_builder_dir", "gitsources"), "sourcesdir")
	if err != nil {
		return err
	}
	err = rpOnBoardLocalMountedDir(path.Join(home, "my_rp_builder_dir", "gitpkgs"), "pkgsdir")
	if err != nil {
		return err
	}

	return nil
}

// rpOnBoardLocalMountedDir checks and configures interactively the
// mounted directory use to interact with the local builder
// :param defDirectory: Default directory to use
// :type defDirectory: string
// :param configFileKey: Key to use to save the configuration in the config file
// :type configFileKey: string
// :returns: error -- an error if something went bad
func rpOnBoardLocalMountedDir(defDirectory, configFileKey string) error {
	// Get the label to identify the directory
	dirLabel := path.Base(defDirectory)

	// Check if the default directory exists
	if rputils.Exists(defDirectory) {
		// Ask the user if we can use it
		RPPrinter.CustomPrintLn()
		question := fmt.Sprintf(rootOnBoardAnsLocalUseDefDir, defDirectory, dirLabel)
		answerLow := rpYesOrNoQuestion(question)

		// Check the answer
		if answerLow == "y" || answerLow == "" {
			// The user wants to use it in configuration, then update the configuration file
			return initOrUpdateDefCfgFile(configFileKey, defDirectory)
		}
	}

	// If we reach here, it means that either the default directory does not exist of the user want to use another one
	var dirPath string
	RPPrinter.CustomPrintLn()
	RPPrinter.CustomPrintF(rootOnBoardAnsLocalDirToUse, dirLabel, defDirectory)
	RPPrinter.CustomPrint(rpCliPrompt)
	RPPrinter.ForceFlush()
	// Get the input
	fmt.Scanln(&dirPath)

	// Fill the default value if empty
	if dirPath == "" {
		dirPath = defDirectory
	}

	// Check if the directory exists, and ask if it needs to be created
	if !rputils.Exists(dirPath) {
		RPPrinter.CustomPrintLn()
		question := fmt.Sprintf(rootOnBoardAnsLocalDirNotExist, dirPath)
		answerLow := rpYesOrNoQuestion(question)
		// Check the answer
		if answerLow == "y" || answerLow == "" {
			// The user wants to create the directory
			err := os.MkdirAll(dirPath, 0755)
			if err != nil {
				return err
			}
		}
	}

	// Finally update the configuration file
	return initOrUpdateDefCfgFile(configFileKey, dirPath)
}

// rpYesOrNoQuestion interactively ask a question that should be answered by yes or no
// :param question: Question to ask
// :type question: string
// :returns: string -- answer
func rpYesOrNoQuestion(question string) string {
	var answer string
	var answerLow string
	allowedAnswers := []string{"y", "n", ""}
	answerIsAllowed := false

	for !answerIsAllowed {
		// Ask the question to the user
		RPPrinter.CustomPrintLn(question)
		RPPrinter.CustomPrint(rpCliPrompt)
		RPPrinter.ForceFlush()

		// Get the input
		fmt.Scanln(&answer)

		// Convert to lowercase
		answerLow = strings.ToLower(answer)

		// Convert 'yes' and 'no' to 'y' and 'n'
		if len(answerLow) > 0 {
			answerLow = string(answerLow[0])
		}

		// Check that the answer is authorized
		answerIsAllowed = utils.ContainsString(allowedAnswers, answerLow)
		if !answerIsAllowed {
			RPPrinter.CustomPrintLn(yesNoNotRespectedErrorMsg)
			RPPrinter.CustomPrintLn()
		}
	}
	return answerLow
}

// getFirstArg allows to get the first argument of the command line.
// Most of the time, it is to get an object ID (user, app, project, etc.).
// It returns an error if the value is empty.
//
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param errorLog: error message to log inside the logger
// :type errorLog: string
// :param errorMsg: error message to return as an error if no value is found
// :type errorMsg: string
// :returns: string, error -- the first argument and an error if something went wrong, nil in the other case
func getFirstArg(args []string, errorLog string, errorMsg string) (string, error) {
	var arg1 string

	// Get the argument of wanted
	if len(args) <= 0 {
		arg1 = ""
	} else {
		arg1 = args[0]
	}

	// Check that a first arg exists
	if arg1 == "" {
		Log.Error(errorLog)
		return "", errors.New(errorMsg)
	}

	return arg1, nil
}

// getConfigDir allows to get the configuration directory used
// by rp-cli
//
// :returns: string, error -- the configuration and an error if something went wrong, nil in the other case
func getConfigDir() (string, error) {
	// Check if the global variable has been set already
	if cfgDirPath != "" {
		return cfgDirPath, nil
	}

	// Get the home directory
	home, err := homedir.Dir()
	if err != nil {
		return "", err
	}

	// Set the configuration directory from home
	cfgDirPath = home + configFileHomePath

	return cfgDirPath, nil
}

// initOrUpdateDefCfgFile is a function allowing to create the default configuration file
// with the given connection alias in it if it does not exist. If it already exist, it is
// only updated.
//
// :param key: Key to update/initiate the def. cfg. file with
// :type key: string
// :param value: Value associated with the key
// :type value: string
// :param overwrite: Set to true to overwrite the content without keeping what is inside
// :type overwrite: bool
// :returns: error -- an error if something went wrong
func initOrUpdateDefCfgFile(key, value string) error {
	// Initialize the configuration data map
	cfgFileData := make(map[string]interface{})

	// Try to read the current configuration file path
	configFilePath := viper.ConfigFileUsed()

	// If the file does not exist, create it
	if configFilePath == "" {

		// Get the configuration directory
		cfgDir, err := getConfigDir()
		if err != nil {
			return err
		}

		// Create the full path to the configuration file
		configFilePath = cfgDir + configFileDefault + ".json"

		// Update the info in viper in case we need it later
		viper.SetConfigFile(configFilePath)

	} else {
		// Read the configuration file and get the current parameters
		err := utils.ReadJsonFile(configFilePath, &cfgFileData)
		if err != nil {
			errMsg := fmt.Sprint(pbWhileReadCfgFileMsg, configFilePath, "-> error:", err.Error())
			return errors.New(errMsg)
		}
	}

	// Update the map with the right value
	cfgFileData[key] = value

	// Write the configuration file
	return utils.WriteJsonFile(configFilePath, cfgFileData)
}

// endProperlyCLI : End properly the cli tool
//
// :param allowExit: allow the function to run an os.Exit() if needed
// :type allowExit: bool
// :param exitCode: exit code to use to end the process
// :type exitCode: int
func endProperlyCLI(allowExit bool, exitCode int) {

	// Flush the printer if it exists
	if RPPrinter != nil {
		RPPrinter.ForceFlush()
	}

	// Close all the connections that may have been opened
	if bckConnection != nil {
		bckConnection.RPConnClose()
	}

	if localBConnection != nil {
		localBConnection.RPConnClose()
	}

	// End the signal monitoring
	signals.EndSignalMonitoring()

	if allowExit && (exitCode != 0) && !viper.GetBool(noErrorFlag) {
		// Exit with the wanted code
		Log.Debugln(endOfAppMsg, exitCode)
		// Real exit only if it is not the test being run
		if !signalTestBeingRun {
			os.Exit(exitCode)
		}
	}
}

// Channel used to know when we received a event from the backend
var cmdEventChannel = make(chan rpapiv1.EventMsg, 10)

// CLIReceiveEvent handle the reception of event from the backend
//
// :param gc: Websocket channel used
// :type gc: *gosocketio.Channel
// :param event: Event message received
// :type event: rpapiv1.EventMsg
func CLIReceiveEvent(gc *gosocketio.Channel, event rpapiv1.EventMsg) {
	// Check if this is an event saying that we have too many open WS
	if event.Type == rpapiv1.EVTSessionExpired {
		// Disable reconnection routine from websocket driver
		bckConnection.WSocket.WebSocket.ReconnectParams.ReconnectTriesNb = 0
		Log.Warnf(disableWSReconnection, rpapiv1.EVTSessionExpired)
	}
	// Send the event inside the channel
	cmdEventChannel <- event
}

// Channel used to know when we received a event from the local-builder
var cmdEventLocalBChannel = make(chan RPLocalBuilderEvent, 10)

// CLIReceiveLocalBEvent handle the reception of event from the local
// builder
//
// :param gc: Websocket channel used
// :type gc: *gosocketio.Channel
// :param event: Event message received
// :type event: RPLocalBuilderEvent
func CLIReceiveLocalBEvent(gc *gosocketio.Channel, event RPLocalBuilderEvent) {
	// Send the event inside the channel
	cmdEventLocalBChannel <- event
}

// CLIRegisterToMultipleEvents allows to subscribe to several server events
// in only one function call
//
// :param evtsList: List of events to subscribe to
// :type evtsList: []string
// :returns: error -- an error if something went wrong, nil in the other case
func CLIRegisterToMultipleEvents(evtsList []string) error {

	// If "non blocking" mode, do not init events
	if viper.GetBool(nonBlockingFlag) {
		return nil
	}

	// Subscribe to all the listed events
	for _, event := range evtsList {
		// Register to the event
		err := bckConnection.RPEventRegister(event)
		if err != nil {
			return err
		}
		// Save the callback
		err = bckConnection.WSocket.OnEvent(event, CLIReceiveEvent)
		if err != nil {
			return err
		}
	}

	return nil
}

// handleErrorEvent handle the reception of an error event from the backend
//
// :param uuidObj: UUID of the object that may be affected by an error
// :type uuidObj: uuid.UUID
// :param evtMsg: Message received as an event
// :type evtMsg: rpapiv1.EventMsg
// :param displayKO: Boolean to request to display "[KO]" message
// :type displayKO: bool
// :returns: error -- an error if something went bad
func handleErrorEvent(uuidObj uuid.UUID, evtMsg rpapiv1.EventMsg, displayKO bool) error {
	rError, err := evtMsg.DecodeError()
	if err != nil {
		if displayKO {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		}
		return err
	}
	if utils.UUIDEqual(rError.ObjectID, uuidObj) {
		// An error occurred during the handling of the object
		if displayKO {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		}
		errMsg := fmt.Sprintf("Status: %s - Message: %s", rError.RpError.Status, rError.RpError.Error)
		if rError.RpError.DetailedError != "" {
			errMsg = errMsg + fmt.Sprintf("\nDetailed error: %s", rError.RpError.DetailedError)
		}
		return errors.New(errMsg)

	} else if rError.ObjectType == rpapiv1.EvtTypeSession {
		// Websocket session error
		if displayKO {
			RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		}
		errMsg := rError.RpError.Error
		if rError.RpError.DetailedError != "" {
			errMsg = errMsg + fmt.Sprintf("\nDetailed error: %s", rError.RpError.DetailedError)
		}
		return errors.New(errMsg)
	}
	return nil
}

// displayNothing is a private function to pass to our printer when
// we want to display something in raw mode but not in user interface mode
//
// :param nothing: The interface passed by the printer
// :type nothing: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func displayNothing(nothing interface{}, verbose bool, writer *tabwriter.Writer) error {
	return nil
}

// manageHeaderPrint is a private function to manage the printing of the
// header of objects list. Indeed, in the case where the user specified "nopager" and no verbose,
// without this function, the header is printed several times (corresponding to the
// number of pages requested)
//
// :param headerToPrint: Header to print if not already done
// :type headerToPrint: string
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func manageHeaderPrint(headerToPrint string, writer *tabwriter.Writer) {
	// Check if the global "no header" flag has been set before printing any header
	noHeader := viper.GetBool(NoHeaderFlag)

	if (!headerAlreadyPrinted || !RPPrinter.NoPager) && !noHeader {
		fmt.Fprintln(writer, headerToPrint)
		headerAlreadyPrinted = true
	}
}

// genericGetBoolFlagValue is a private function allowing to retrieve
// a boolean flag value according to the situation, i.e. is it creation or update?
//
// Basically, if the mode is update, it gets the string flag and convert it to bool.
// If the mode is creation, it gets directly a boolean flag.
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param isCreation: Indicate if the object is being created
// :type isCreation: bool
// :param currentValue: Current value of the field (use only in update mode)
// :type currentValue: bool
// :param flagName: Name of the flag to retrieve
// :type flagName: string
// :param errMsg: Message to display when a problem appears while getting the flag value
// :type errMsg: string
// :returns: bool, error -- bool value of the flag and an error if something went bad
func genericGetBoolFlagValue(cmd *cobra.Command, isCreation, currentValue bool, flagName, errMsg string) (bool, error) {
	// Declare the result
	var finalValue bool
	var err error

	if isCreation {
		// Manage creation case
		finalValue, err = cmd.Flags().GetBool(flagName)
		if err != nil {
			Log.Error(errMsg)
			return finalValue, err
		}

	} else {
		// Manage update case
		tmpValue, err := cmd.Flags().GetString(flagName)
		if err != nil {
			Log.Error(errMsg)
			return finalValue, err
		}

		if tmpValue != flagUpdateKeepCurrentValue {
			finalValue, err = strconv.ParseBool(tmpValue)
			if err != nil {
				return finalValue, err
			}
		} else {
			finalValue = currentValue
		}
	}

	return finalValue, nil
}

// getAllConnectionsParams is a private function allowing to get all the connection parameters
// provided by the user
//
// :param file: Filename from which the app wants to get the connections
// :type file: string
// :returns: rpconnect.RPCLIConnections, error -- the connection parameters and an error if something went bad
func getAllConnectionsParams(file string) ([]rpconnect.RPCLIConnection, error) {
	resultParams := []rpconnect.RPCLIConnection{}

	// Get the configuration directory
	cfgDir, err := getConfigDir()
	if err != nil {
		return resultParams, err
	}

	// Check that a token file exists
	filePath := cfgDir + file
	if _, err := os.Stat(filePath); err == nil {
		// If the file exist, get the connection parameters
		err := utils.ReadJsonFile(filePath, &resultParams)
		if err != nil {
			return resultParams, err
		}
	}

	return resultParams, nil
}

// getConnectionParams is a private function allowing to get the connection parameters
// provided by the user
//
// :param alias: Connection alias for which the connections parameters are wanted
// :type alias: string
// :param file: Filename of the file containing the connections
// :type file: string
// :returns: rpconnect.RPCLIConnections, error -- the connection parameters and an error if something went bad
func getConnectionParams(alias string, file string) (rpconnect.RPCLIConnection, error) {
	// Initialize the return value
	resultParams := rpconnect.RPCLIConnection{
		Alias: alias,
	}

	// Get all the connection parameters that exist
	allConnectionParams, err := getAllConnectionsParams(file)
	if err != nil {
		return resultParams, err
	}

	// Look for the specified alias
	indexWanted := -1
	for index, params := range allConnectionParams {
		if params.Alias == alias {
			indexWanted = index
		}
	}

	// Check that we found something
	if indexWanted > -1 {
		resultParams = allConnectionParams[indexWanted]
	} else if runMode != "testing" {
		return resultParams, fmt.Errorf(connectionAliasNotExistMsg, alias)
	}

	return resultParams, nil
}

// isThereHelpInFlagValue is a function allowing to check if the user passed
// a "--help" or "-h" string as a flag value
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :returns: bool -- true if the user gave "--help" or "-h"
func isThereHelpInFlagValue(cmd *cobra.Command) bool {
	// Function checking if --help is given
	helpNeeded := false
	checkHelpFunc := func(f *pflag.Flag) {
		// Try to get value
		if (f.Value.String() == "--help") || (f.Value.String() == "-h") {
			helpNeeded = true
		}
	}

	// Now check the flag value given by the user
	cmd.Flags().Visit(checkHelpFunc)

	return helpNeeded
}

// requestAccessToken is a private function allowing to request the connection
// parameters from the user and to write it in a file
//
// :param params: Structure to fill with the right values
// :type params: *rpconnect.RPCLIConnection
// :returns: error -- an error if something went bad
func requestAccessToken(params *rpconnect.RPCLIConnection) error {

	// Write the introduction message
	RPPrinter.CustomPrintF(noConnectParamsSetMsg, params.Alias)

	// Request for URL
	RPPrinter.CustomPrintLn(rootOnBoardAnsURLReq)
	RPPrinter.CustomPrint(rpCliPrompt)
	RPPrinter.ForceFlush()
	_, err := fmt.Scanln(&params.ServerURL)
	if err != nil {
		return err
	}

	RPPrinter.CustomPrintLn()

	// Request for Token
	RPPrinter.CustomPrintF(createTokenMsg0, params.ServerURL)
	RPPrinter.CustomPrintLn(createTokenMsg1)
	RPPrinter.CustomPrint(rpCliPrompt)
	RPPrinter.ForceFlush()
	_, err = fmt.Scanln(&params.AccessToken)
	if err != nil {
		return err
	}

	// Append the results in the token file
	err = appendOrOverwriteInTokenFile(accessTokenFile, *params)
	if err != nil {
		return err
	}

	return nil
}

// isBCKConnectNecessary allows the app to know if it is necessary
// to connect to the current selected Redpesk backend
//
// :param cmd: Cobra command currently used
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: bool -- True if it is necessary, false in the other case
func isBCKConnectNecessary(cmd *cobra.Command, args []string) bool {
	result := true

	switch {
	case strings.HasPrefix(cmd.Use, "help"), strings.HasPrefix(cmd.Use, "completion"):
		// The command is an help or the completion one: no need for connection
		result = false

	case strings.HasPrefix(cmd.Use, "__complete"):
		// This is a completion call, the need of connection may vary, but by default: no need
		result = false

		// Get the root command
		tmpRootCmd := cmd.Parent()

		// Get the command object from args
		tmpChildCmd, _, err := tmpRootCmd.Find(args)
		if err != nil {
			// Print the main error message
			Log.Errorf(errWhileRetrievingChildCmd, err)
			return result
		}

		// Look for annotations in child command structure
		// "__arg_need_bck_connect__" annotation
		if connectNeeded, ok := tmpChildCmd.Annotations["__arg_need_bck_connect__"]; ok {
			if connectNeeded == "TRUE" {
				result = true
			}
		}
		// "__flag_need_bck_connect__" annotation
		if connectNeeded, ok := tmpChildCmd.Annotations["__flag_need_bck_connect__"]; ok {
			if connectNeeded == "TRUE" {
				result = true
			}
		}

	default:
		// In the default case, look for "__bck_connect__" in commands annotations
		if connectNeeded, ok := cmd.Annotations["__bck_connect__"]; ok {
			if connectNeeded == "FALSE" {
				result = false
			}
		}
	}

	return result
}

// isWSBCKConnectNecessary allows the app to know if it is necessary
// to connect to the web-socket of current selected Redpesk backend
//
// :param cmd: Cobra command currently used
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: bool -- True if it is necessary, false in the other case
func isWSBCKConnectNecessary(cmd *cobra.Command, args []string) bool {
	// By default no websocket connection
	result := false

	// Get the value of "nonblocking" flag
	nonBlocking := viper.GetBool(nonBlockingFlag)

	// In the default case, look for "__bck_ws_connect__" in commands annotations
	if connectionNeeded, ok := cmd.Annotations["__bck_ws_connect__"]; ok && !nonBlocking {
		if connectionNeeded == "TRUE" {
			result = true
		}
	}

	return result
}

// isLADConnectNecessary allows the app to know if it is necessary
// to connect to the current local builder (redpesk LAD)
//
// :param cmd: Cobra command currently used
// :type cmd: *cobra.Command
// :returns: bool -- True if it is necessary, false in the other case
func isLADConnectNecessary(cmd *cobra.Command) bool {
	result := false

	// In the default case, look for "__lad_connect__" in commands annotations
	if connectNeeded, ok := cmd.Annotations["__lad_connect__"]; ok {
		if connectNeeded == "TRUE" {
			result = true
		}
	}

	return result
}

// gracefulRecoverFromPanic allows to recover from a panic in
// a graceful way, to display the appropriate logs to the user
// in accordance with the mode the application is running
//
// :param allowExit: allow the function to run an os.Exit() if needed
// :type allowExit: bool
func gracefulRecoverFromPanic(allowExit bool) {
	if r := recover(); r != nil {

		// Exit with an error at the end of this function
		defer endProperlyCLI(allowExit, 1)

		// Print the main error message
		Log.Errorln(runtimeErrorMsg)

		// Get the error message from the panic
		var panicErrMsg error
		switch recErr := r.(type) {
		case error:
			panicErrMsg = recErr
		case string:
			panicErrMsg = errors.New(recErr)
		default:
			panicErrMsg = errors.Errorf("%v", recErr)
		}
		// Get the stacktrace of the panic
		panicStackTrace := string(debug.Stack())

		if BuildMode == "development" {
			// If the current mode is development, print directly the backtrace in console
			Log.Errorln()
			Log.Errorf(runtimeErrorDetailsMsg, panicErrMsg)
			Log.Errorln()
			Log.Errorf(runtimeErrorStackTraceMsg, panicStackTrace)

		} else if BuildMode == "production" {
			// If the current mode is production, print the backtrace in a file in the configuration directory

			// Get the configuration directory
			cfgDir, err := getConfigDir()
			if err != nil {
				Log.Errorf(err.Error())
				return
			}

			// Create the crash report
			crashFile := cfgDir + applicationName + ".crash"
			f, err := os.Create(crashFile)
			if err != nil {
				Log.Errorf(err.Error())
				return
			}
			defer f.Close()

			// Write inside the file
			fullFile := fmt.Sprintf(runtimeErrorDetailsMsg, panicErrMsg) + "\n" + fmt.Sprintf(runtimeErrorStackTraceMsg, panicStackTrace)
			_, err = f.WriteString(fullFile)
			if err != nil {
				Log.Errorf(err.Error())
				return
			}

			// Write the appropriate message to the user
			Log.Errorln()
			Log.Errorf(runtimeErrorCrashReportMsg, (applicationName + ".crash"), cfgDir)
		}
	}
}

// logFlagError is a function allowing to log an error in the command line from
// the name of the flag (we will erase a lot of constants thanks to this function)
//
// :param flagName: Name of the flag for which the error appeared
// :type flagName: string
func logFlagError(flagName string) {
	logMsg := fmt.Sprintf("%s \"--%s\" flag", pbFlagMainPrefix, flagName)
	Log.Error(logMsg)
}
