/*
Copyright © 2020-2023 IoT.bzh Company
Contact: https://www.iot.bzh/licensing

This file is part of the redpesk-cli project, the redpesk command line tool.

$RP_BEGIN_LICENSE$
Commercial License Usage

	Licensees holding valid commercial IoT.bzh licenses may use this file in
	accordance with the commercial license agreement provided with the
	Software or, alternatively, in accordance with the terms contained in
	a written agreement between you and The IoT.bzh Company. For licensing terms
	and conditions see https://www.iot.bzh/terms-conditions. For further
	information use the contact form at https://www.iot.bzh/contact.

GNU General Public License Usage

	Alternatively, this file may be used under the terms of the GNU General
	Public license version 3. This license is as published by the Free Software
	Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
	of this file. Please review the following information to ensure the GNU
	General Public License requirements will be met
	https://www.gnu.org/licenses/gpl-3.0.html.

$RP_END_LICENSE$
*/
package cmd

import (
	"strings"
	"testing"

	"github.com/stretchr/testify/assert"
)

const (
	// Teams
	teamDisplayNoVerboseColNb = 5
)

// TestTeamsAddNoBlock is the function allowing to test if
// the command "rp-cli teams add --nonblocking ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestTeamsAddNoBlock(test *testing.T) {
	// Create the command
	cmdToRun := getMainAddTeamCmd("Bruit", "La meilleure des teams", true)
	// Add the non-blocking flag
	cmdToRun = append(cmdToRun, "--nonblocking")

	// Create the team if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, teamsCommand, "bruit", cmdToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// The team has been created, check it is now in backend

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID":                `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Slug":              `^bruit$`,
		"Name":              `^Bruit$`,
		"Status":            `^(public|private)$`,
		"Number_of_members": `^[0-9]+$`,
	}

	// Run the command and the tests (no verbose and verbose, since this is the same result)
	checkCmdAnsKeysRegexpCol(test, teamDisplayNoVerboseColNb, regexpMap, teamsCommand, "get", "bruit")
}

// TestTeamsAddBlock is the function allowing to test if
// the command "rp-cli teams add ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestTeamsAddBlock(test *testing.T) {
	// Create the command
	cmdToRun := getMainAddTeamCmd("Redpesk Team", "Poisson rouge!", false)

	// Create the team if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, teamsCommand, "redpesk-team", cmdToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// The team has been created, check it is now in backend

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID":                `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Slug":              `^redpesk-team$`,
		"Name":              `^Redpesk Team$`,
		"Status":            `^(public|private)$`,
		"Number_of_members": `^[0-9]+$`,
	}

	// Run the command and the tests (no verbose and verbose, since this is the same result)
	checkCmdAnsKeysRegexpCol(test, teamDisplayNoVerboseColNb, regexpMap, teamsCommand, "get", "redpesk-team")
}

// TestTeamsGet is the function allowing to test if
// the command "rp-cli teams get ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestTeamsGet(test *testing.T) {
	// Create the command
	cmdToRun := getMainAddTeamCmd("Bruit", "La meilleure des teams", true)

	// Create the team if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, teamsCommand, "bruit", cmdToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// Verbose only, since the no verbose option is tested in almost all other tests

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":               `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Slug":              `^bruit$`,
		"Name":              `^Bruit$`,
		"Status":            `^public$`,
		"Description":       `^La`,
		"Owner":             `^[a-z0-9-_]+$`,
		"Number_of_members": `^[0-9]+$`,
	}

	// Run the "get verbose" command an get the output
	cmdOutput, err := executeCommand(teamsCommand, "get", "-v", "bruit")
	if err != nil {
		test.Fatal(err)
	}

	// Check the output of the command
	checkVerboseGetTeamOutput(test, cmdOutput, regexpMap)

	// Try the same thing but give the name to the get command
	cmdOutput, err = executeCommand(teamsCommand, "get", "-v", "Bruit")
	if err != nil {
		test.Fatal(err)
	}

	// Check the output of the command
	checkVerboseGetTeamOutput(test, cmdOutput, regexpMap)
}

// TestTeamsList is the function allowing to test if
// the command "rp-cli teams list ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestTeamsList(test *testing.T) {
	// Create the command
	cmdToRun1 := getMainAddTeamCmd("Bruit", "La meilleure des teams", true)
	cmdToRun2 := getMainAddTeamCmd("Redpesk Team", "Poisson rouge!", false)
	cmdToRun3 := getMainAddTeamCmd("Among Us", "Bouh!", true)

	// Create the teams if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, teamsCommand, "bruit", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, teamsCommand, "redpesk-team", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, teamsCommand, "among-us", cmdToRun3...)
	if err != nil {
		test.Fatal(err)
	}

	// The teams have been created, check the list, in not verbose and verbose mode

	// Not verbose mode
	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID":                `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Slug":              `^[a-z0-9-]+$`,
		"Name":              `.+`,
		"Status":            `^(public|private)$`,
		"Number_of_members": `^[0-9]+$`,
	}

	// Run the command and the tests (no verbose and verbose, since this is the same result)
	checkCmdAnsKeysRegexpCol(test, teamDisplayNoVerboseColNb, regexpMap, teamsCommand, "list")

	// Verbose mode
	// Create the map containing the regexp we want to match in the answer
	regexpMap2 := map[string]string{
		"ID:":         `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Slug":        `^[a-z0-9-]+$`,
		"Name":        `.+`,
		"Status":      `^(public|private)$`,
		"Description": `.*`,
		"Owner":       `^[a-z0-9-_]+$`,
	}

	// Check the output of the command
	checkCmdAnswerListKeysRegexp(test, teamsListVerboseLineNb, "", regexpMap2, teamsCommand, "list", "-v")
}

// TestTeamsUpdate is the function allowing to test if
// the command "rp-cli teams update ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestTeamsUpdate(test *testing.T) {
	// Create the command
	cmdToRun := getMainAddTeamCmd("Redpesk Team", "Poisson rouge!", false)

	// Create the team if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, teamsCommand, "redpesk-team", cmdToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// The team has been created
	// Create the update command
	updateCmd := getMainUpdateTeamCmd("redpesk-team", "Redpesk Team Updated", "New Poisson rouge!", true)

	// Run the update command and check that all is [OK]
	checkActionCmdAns(test, updateCmd...)

	// Check that the team has been updated

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":         `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Slug":        `^redpesk-team$`,
		"Name":        `Updated$`,
		"Status":      `^public$`,
		"Description": `^New`,
		"Owner":       `^[a-z0-9-_]+$`,
	}

	// Run the "get verbose" command an get the output
	cmdOutput, err := executeCommand(teamsCommand, "get", "redpesk-team", "--verbose")
	if err != nil {
		test.Fatal(err)
	}

	// Check the output of the command
	checkVerboseGetTeamOutput(test, cmdOutput, regexpMap)
}

// TestTeamsUpdatePrivate is the function allowing to test if
// the command "rp-cli teams update ..." is working correctly
// when the team is switched to be private
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestTeamsUpdatePrivate(test *testing.T) {
	// Create the command
	cmdToRun := getMainAddTeamCmd("Greenpesk Team", "Poisson vert!", true)

	// Create the team if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, teamsCommand, "greenpesk-team", cmdToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// The team has been created
	// Create the update command
	updateCmd := getMainUpdateTeamCmd("greenpesk-team", "", "", false)

	// Run the update command and check that all is [OK]
	checkActionCmdAns(test, updateCmd...)

	// Check that the team has been updated

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":         `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Slug":        `^greenpesk-team$`,
		"Name":        `Greenpesk Team$`,
		"Status":      `^private$`,
		"Description": `^Poisson vert!$`,
		"Owner":       `^[a-z0-9-_]+$`,
	}

	// Run the "get verbose" command an get the output
	cmdOutput, err := executeCommand(teamsCommand, "get", "greenpesk-team", "--verbose")
	if err != nil {
		test.Fatal(err)
	}

	// Check the output of the command
	checkVerboseGetTeamOutput(test, cmdOutput, regexpMap)
}

// TestTeamsInvite is the function allowing to test if
// the command "rp-cli teams invite ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestTeamsInvite(test *testing.T) {
	// Create the command
	cmdToRun := getMainAddTeamCmd("Redpesk Team", "Poisson rouge!", true)

	// Create the team if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, teamsCommand, "redpesk-team", cmdToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// TEST: Invite a user ----------------------------------------------------
	// Create a user
	cmdToRunUser := getMainAddUserCmd("tiramisu", "tiramisu@iot.bzh", "Ti", "Ramisu", "0666989945")
	// Create the user if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, usersCommand, "tiramisu", cmdToRunUser...)
	if err != nil {
		test.Fatal(err)
	}

	// Run the invite command and check that all is [OK]
	checkActionCmdAns(test, teamsCommand, "invite", "redpesk-team", "-g", "tiramisu")
	// ------------------------------------------------------------------------

	// TEST: Invite a team ----------------------------------------------------
	// Create a team
	cmdToRunTeam := getMainAddTeamCmd("Among Us", "Bouh!", false)
	// Create the team if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, teamsCommand, "among-us", cmdToRunTeam...)
	if err != nil {
		test.Fatal(err)
	}

	// Run the invite command and check that all is [OK]
	checkActionCmdAns(test, teamsCommand, "invite", "redpesk-team", "--guest", "among-us")
	// ------------------------------------------------------------------------
}

// TestTeamsDeleteMember is the function allowing to test if
// the command "rp-cli teams del-member ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestTeamsDeleteMember(test *testing.T) {

	// Test not available for true backend since the user needs to accept the invitation before we can delete him
	if bckIsReal {
		test.Log("[WARNING] It is not possible to test member deletion in true backend (the invitation needs to be accepted)...")
		return
	}

	// Create the command
	cmdToRun := getMainAddTeamCmd("Redpesk Team", "Poisson rouge!", true)

	// Create the team if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, teamsCommand, "redpesk-team", cmdToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// TEST: Invite a user ----------------------------------------------------
	// Create a user
	cmdToRunUser := getMainAddUserCmd("tiramisu", "tiramisu@iot.bzh", "Ti", "Ramisu", "0666989945")
	// Create the user if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, usersCommand, "tiramisu", cmdToRunUser...)
	if err != nil {
		test.Fatal(err)
	}

	// Run the invite command and check that all is [OK]
	checkActionCmdAns(test, teamsCommand, "invite", "redpesk-team", "-g", "tiramisu")
	// ------------------------------------------------------------------------

	// TEST: Invite a team ----------------------------------------------------
	// Create a team
	cmdToRunTeam := getMainAddTeamCmd("Among Us", "Bouh!", false)
	// Create the team if it does not exists, and check that all steps are OK
	err = addObjectIfNotExists(test, teamsCommand, "among-us", cmdToRunTeam...)
	if err != nil {
		test.Fatal(err)
	}

	// Run the invite command and check that all is [OK]
	checkActionCmdAns(test, teamsCommand, "invite", "redpesk-team", "--guest", "among-us")
	// ------------------------------------------------------------------------

	// Delete a member of the team and check that all is [OK]
	checkActionCmdAns(test, teamsCommand, "del-member", "redpesk-team", "--member", "tiramisu")
}

// --------------------------------------------------------------
// Utilities functions ------------------------------------------
// --------------------------------------------------------------

// getMainAddTeamCmd is an initializer function to get the mainly
// used command to create a team
//
// :param name: Name of the team to create
// :type name: string
// :param description: Description of the team
// :type description: string
// :param isPublic: True if the team is public, false in the other case
// :type isPublic: bool
// :returns: []string -- the value of the wanted constant
func getMainAddTeamCmd(name string, description string, isPublic bool) []string {
	cmdToReturn := []string{
		teamsCommand,
		teamAddCmdName,
		"-n",
		name,
		"-d",
		description,
	}

	if isPublic {
		cmdToReturn = append(cmdToReturn, "--public")
	}

	return cmdToReturn
}

// getMainUpdateTeamCmd is an initializer function to get the mainly
// used command to update a user
//
// :param slug: Slug of the team to update
// :type slug: string
// :param name: Name of the team to create
// :type name: string
// :param description: Description of the team
// :type description: string
// :param isPublic: True if the team is public, false in the other case
// :type isPublic: bool
// :returns: []string -- the value of the wanted constant
func getMainUpdateTeamCmd(slug string, name string, description string, isPublic bool) []string {
	cmdToReturn := []string{
		teamsCommand,
		"update",
		slug,
	}

	if name != "" {
		cmdToReturn = append(cmdToReturn, "-n", name)
	}

	if description != "" {
		cmdToReturn = append(cmdToReturn, "-d", description)
	}

	if isPublic {
		cmdToReturn = append(cmdToReturn, "--public", "true")
	} else {
		cmdToReturn = append(cmdToReturn, "--public", "false")
	}

	return cmdToReturn
}

// checkVerboseGetTeamOutput is function allowing to process the output of
// a "teams get <teamID> -v" command, check the obtained value according
// to a regex map and check the members part as well.
//
// :param test: Pointer to T testing object
// :type test: *testing.T
// :param output: Output to process
// :type output: string
// :param regexMap: Regex map used to check the answers
// :type regexMap: map[string]string
func checkVerboseGetTeamOutput(test *testing.T, output string, regexMap map[string]string) {

	// Split the answer in two and check that there are two parts
	splitAns := strings.Split(output, teamsListCmdVerboseLine6)
	isEqual := assert.Equal(test, 2, len(splitAns))

	if isEqual {
		// Get the team description part of the answer
		objDescPart := strings.TrimSpace(splitAns[0])
		// Get the members part of the answer
		membersPart := strings.TrimSpace(splitAns[1])

		// Check that the description part respect the map
		checkObjectKeysRegexp(test, objDescPart, teamsListVerboseLineNb, regexMap, true)

		// Check that the members part respect what it needs to respect
		// Create the map containing the regexp we want to match in the answer
		regexpMapCol := map[string]string{
			"Pseudo":            `^[a-z0-9-_]+$`,
			"Email":             `.+@iot.bzh$`,
			"Invitation status": `^(pending|accepted|refused)$`,
		}

		// Check that the members part of the answer is good
		checkObjectKeysRegexpCol(test, membersPart, 3, regexpMapCol)
	}
}

// --------------------------------------------------------------
// --------------------------------------------------------------
