/*
Copyright © 2020-2023 IoT.bzh Company
Contact: https://www.iot.bzh/licensing

This file is part of the redpesk-cli project, the redpesk command line tool.

$RP_BEGIN_LICENSE$
Commercial License Usage

	Licensees holding valid commercial IoT.bzh licenses may use this file in
	accordance with the commercial license agreement provided with the
	Software or, alternatively, in accordance with the terms contained in
	a written agreement between you and The IoT.bzh Company. For licensing terms
	and conditions see https://www.iot.bzh/terms-conditions. For further
	information use the contact form at https://www.iot.bzh/contact.

GNU General Public License Usage

	Alternatively, this file may be used under the terms of the GNU General
	Public license version 3. This license is as published by the Free Software
	Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
	of this file. Please review the following information to ensure the GNU
	General Public License requirements will be met
	https://www.gnu.org/licenses/gpl-3.0.html.

$RP_END_LICENSE$
*/
package cmd

import (
	"errors"
	"fmt"
	"strconv"
	"text/tabwriter"

	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/printer"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/utils"
	"git.ovh.iot/redpesk/redpesk-infra/rp-webserver/actions/rpapiv1"
	"github.com/gofrs/uuid"
	"github.com/jinzhu/copier"
	"github.com/spf13/cobra"
	"github.com/spf13/viper"
)

// -------------------------------------------------
// Global variable (inside this package) -----------
// -------------------------------------------------

// Users cached
var cachedRPUsersByUUID map[uuid.UUID]rpapiv1.RPUserFull
var cachedRPUsersByPseudo map[string]uuid.UUID

// -------------------------------------------------
// -------------------------------------------------

// --------------------------------------------------------------
// Initialization functions -------------------------------------
// --------------------------------------------------------------

func initUsersCmd(rootCmd *cobra.Command) {

	// usersCmd represents the users command
	var usersCmd = &cobra.Command{
		Use:   usersCommand,
		Short: usersCmdShortDesc,
	}

	// usersListsCmd represents the "list" command for users
	var usersListCmd = &cobra.Command{
		Use:          usersListCmdName,
		SilenceUsage: true,
		Aliases:      []string{usersListCmdAlias},
		Short:        usersListCmdShortUsage,
		RunE:         getUsersList,
	}

	// userGetCmd represents the "get" command for users
	var userGetCmd = &cobra.Command{
		Use:               userGetCmdName,
		SilenceUsage:      true,
		Short:             userGetCmdShortUsage,
		Long:              userGetCmdLongUsage,
		RunE:              getUser,
		ValidArgsFunction: getListUserAsTab,
		Annotations:       map[string]string{"__arg_need_bck_connect__": "TRUE"},
	}

	// userAddCmd represents the "add" command for users
	var userAddCmd = &cobra.Command{
		Use:          userAddCmdName,
		SilenceUsage: true,
		Short:        userAddCmdShortUsage,
		PreRunE:      initEventUser,
		RunE:         createUser,
		Annotations:  map[string]string{"__bck_ws_connect__": "TRUE"},
	}

	// userUpdateCmd represents the "update" command for users
	var userUpdateCmd = &cobra.Command{
		Use:               userUpdateCmdName,
		SilenceUsage:      true,
		Aliases:           []string{userUpdateCmdAlias},
		Short:             userUpdateCmdShortUsage,
		Long:              userUpdateCmdLongUsage,
		PreRunE:           initEventUser,
		RunE:              updateUser,
		ValidArgsFunction: getListUserAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// userDisableCmd represents the "disable" command for users
	var userDisableCmd = &cobra.Command{
		Use:               userDisableCmdName,
		SilenceUsage:      true,
		Aliases:           []string{userDisableCmdAlias},
		Short:             userDisableCmdShortUsage,
		Long:              userDisableCmdLongUsage,
		PreRunE:           initEventUser,
		RunE:              disableUser,
		ValidArgsFunction: getListUserAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// userEnableCmd represents the "enable" command for users
	var userEnableCmd = &cobra.Command{
		Use:               userEnableCmdName,
		SilenceUsage:      true,
		Aliases:           []string{userEnableCmdAlias},
		Short:             userEnableCmdShortUsage,
		Long:              userEnableCmdLongUsage,
		PreRunE:           initEventUser,
		RunE:              enableUser,
		ValidArgsFunction: getListUserAsTab,
		Annotations: map[string]string{
			"__arg_need_bck_connect__": "TRUE",
			"__bck_ws_connect__":       "TRUE"},
	}

	// Here you will define your flags and configuration settings.

	// Persistent Flags -----------------------------------------
	// ----------------------------------------------------------

	// Local Flags ----------------------------------------------
	// Define verbose flag for all the commands that use it
	usersListCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)
	userGetCmd.Flags().BoolP(mainVerboseFlag, mainVerboseFlagAlias, mainVerboseFlagDefault, mainVerboseFlagUsage)

	// Other flags needed for user creation/update
	// Creation
	userAddCmd.Flags().StringP(userAddCmdPseudoFlag, userAddCmdPseudoFlagAlias, "", userAddCmdPseudoFlagUsage)
	userAddCmd.Flags().StringP(userAddCmdMailFlag, userAddCmdMailFlagAlias, "", userAddCmdMailFlagUsage)
	userAddCmd.Flags().StringP(userAddCmdFNameFlag, userAddCmdFNameFlagAlias, "", userAddCmdFNameFlagUsage)
	userAddCmd.Flags().StringP(userAddCmdLNameFlag, userAddCmdLNameFlagAlias, "", userAddCmdLNameFlagUsage)
	userAddCmd.Flags().String(userAddCmdPhoneFlag, "", userAddCmdPhoneFlagUsage)
	// Update
	userUpdateCmd.Flags().StringP(userAddCmdMailFlag, userAddCmdMailFlagAlias, flagUpdateKeepCurrentValue, userAddCmdMailFlagUsage)
	userUpdateCmd.Flags().StringP(userAddCmdFNameFlag, userAddCmdFNameFlagAlias, flagUpdateKeepCurrentValue, userAddCmdFNameFlagUsage)
	userUpdateCmd.Flags().StringP(userAddCmdLNameFlag, userAddCmdLNameFlagAlias, flagUpdateKeepCurrentValue, userAddCmdLNameFlagUsage)
	userUpdateCmd.Flags().String(userAddCmdPhoneFlag, flagUpdateKeepCurrentValue, userAddCmdPhoneFlagUsage)
	// ----------------------------------------------------------

	// Add some flags as required
	userAddCmd.MarkFlagRequired(userAddCmdPseudoFlag)
	userAddCmd.MarkFlagRequired(userAddCmdMailFlag)

	// Add the command "users" as a subcommand of rp-cli
	rootCmd.AddCommand(usersCmd)

	// Add the command "list" as a subcommand of usersCmd
	usersCmd.AddCommand(usersListCmd)

	// Add the command "get" as a subcommand of usersCmd
	usersCmd.AddCommand(userGetCmd)

	// Add the command "add" as a subcommand of usersCmd
	usersCmd.AddCommand(userAddCmd)

	// Add the command "update" as a subcommand of usersCmd
	usersCmd.AddCommand(userUpdateCmd)

	// Add the command "disable" as a subcommand of usersCmd
	usersCmd.AddCommand(userDisableCmd)

	// Add the command "enable" as a subcommand of usersCmd
	usersCmd.AddCommand(userEnableCmd)
}

// initEventUser inits everything that is needed for the user add
// command to run
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went bad
func initEventUser(cmd *cobra.Command, args []string) error {
	// Init the events list
	eventsList := []string{
		rpapiv1.EVTUserAdd,
		rpapiv1.EVTUserDelete,
		rpapiv1.EVTUserDisable,
		rpapiv1.EVTUserStateChange,
	}

	return CLIRegisterToMultipleEvents(eventsList)
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Completion functions -----------------------------------------
// --------------------------------------------------------------

// getListUserAsTab gets all the users available in the Redpesk
// instance and return it as an array of strings
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param toComplete: String passed by the user to complete
// :type toComplete: string
// :returns: []string, cobra.ShellCompDirective -- the tab result and the shell completion directive
func getListUserAsTab(cmd *cobra.Command, args []string, toComplete string) ([]string, cobra.ShellCompDirective) {
	usersList := []string{"me"}

	// Check if we have an admin role
	isAdmin, err := isCurrentUserAdmin("user")
	if err != nil {
		return usersList, cobra.ShellCompDirectiveNoFileComp
	}

	// If the current user is not an Admin, he can only see "me"
	if !isAdmin {
		return usersList, cobra.ShellCompDirectiveNoFileComp
	}

	// Get users list
	users := rpapiv1.RPUsersFull{}
	if err := bckAPI.UsersFullGetList(&users); err != nil {
		return usersList, cobra.ShellCompDirectiveNoFileComp
	}

	// We got correctly the users list, fill the results tab
	for _, user := range users.Users {
		usersList = append(usersList, user.Pseudo)
	}

	return usersList, cobra.ShellCompDirectiveNoFileComp
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Command functions --------------------------------------------
// --------------------------------------------------------------

// getUsersList gets all the users in the Redpesk instance and prints it.
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func getUsersList(cmd *cobra.Command, args []string) error {

	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Check if we have an admin role
	isAdmin, err := isCurrentUserAdmin("user")
	if err != nil {
		return err
	}

	// Different behavior if the current user is admin or not
	if isAdmin {

		// Prepare the function needed by the printer to get the images
		getUsersCB := func(pageNb, perPage int) (interface{}, printer.RPPrinterPaging, error) {
			// Get the "full info" users list
			usersFull := rpapiv1.RPUsersFull{}
			if err := bckAPI.UsersFullGetListByPage(&usersFull, pageNb, perPage); err != nil {
				return usersFull.Users, printer.RPPrinterPaging{}, err
			}
			// Get the paging interface from the structure retrieved
			pagingStruct, _ := printer.GetPagingFromInterface(&usersFull)

			return usersFull.Users, pagingStruct, nil
		}

		// Prepare the structure to manage paging
		pagingStruct := printer.RPPagingToolsStruct{
			Verbose:       verbose,
			VerboseLineNb: userDisplayVerboseLineNb,
			GetObjCB:      getUsersCB,
			PrintFancyCB:  displayUsers,
		}

		return RPPrinter.InteractivelyPrintPages(&pagingStruct, viper.GetBool(NoHeaderFlag))

	} else {
		// The users list is authorized only for the admins!
		Log.Error(userCurrentNotAdmin)
		return errors.New(userCurrentNotAdminMsg)
	}
}

// getUser gets all the user info in the Redpesk instance and prints it.
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func getUser(cmd *cobra.Command, args []string) error {

	// Get the user ID and return an error if there is nothing
	userID, err := getFirstArg(args, userIDNotSetMsg, userIDRequiredMsg)
	if err != nil {
		return err
	}

	// Get the verbose flag
	verbose, err := cmd.Flags().GetBool(mainVerboseFlag)
	if err != nil {
		logFlagError(mainVerboseFlag)
		return err
	}

	// Get the user's structure from what the user gave us
	userToPrint := rpapiv1.RPUserFull{}
	if err = getUserStructFromAnyID(userID, &userToPrint); err != nil {
		return err
	}

	// Display the information about the obtained user
	RPPrinter.CustomPrintObject(userToPrint, verbose, displayOneUser)

	return nil
}

// prepareUserStruct prepare a user structure to send to the backend
// to create or update a user
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :param user: User full structure to fill
// :type user: *rpapiv1.RPUserFull
// :param isCreation: Indicate if we are creating the user
// :type isCreation: bool
// :returns: pointer to the RPAPIInterface object
func prepareUserStruct(cmd *cobra.Command, args []string, user *rpapiv1.RPUserFull, isCreation bool) error {

	if isCreation {
		// Get the pseudo of the user to create
		userPseudo, err := cmd.Flags().GetString(userAddCmdPseudoFlag)
		if err != nil {
			logFlagError(userAddCmdPseudoFlag)
			return err
		}
		user.Pseudo = userPseudo
	}

	// Get the email of the user to create
	userMail, err := cmd.Flags().GetString(userAddCmdMailFlag)
	if err != nil {
		logFlagError(userAddCmdMailFlag)
		return err
	}

	// Get the firstname of the user to create
	userFName, err := cmd.Flags().GetString(userAddCmdFNameFlag)
	if err != nil {
		logFlagError(userAddCmdFNameFlag)
		return err
	}

	// Get the lastname of the user to create
	userLName, err := cmd.Flags().GetString(userAddCmdLNameFlag)
	if err != nil {
		logFlagError(userAddCmdLNameFlag)
		return err
	}

	// Get the phone number of the user to create
	userPhoneNb, err := cmd.Flags().GetString(userAddCmdPhoneFlag)
	if err != nil {
		logFlagError(userAddCmdPhoneFlag)
		return err
	}

	// Fill the structure to pass as a json in the http request
	if isCreation {
		user.Email = userMail
		user.Firstname = userFName
		user.Lastname = userLName
		user.Phone = userPhoneNb
	} else {
		if userMail != flagUpdateKeepCurrentValue {
			user.Email = userMail
		}
		if userFName != flagUpdateKeepCurrentValue {
			user.Firstname = userFName
		}
		if userLName != flagUpdateKeepCurrentValue {
			user.Lastname = userLName
		}
		if userPhoneNb != flagUpdateKeepCurrentValue {
			user.Phone = userPhoneNb
		}
	}

	return nil
}

// createUser creates a user as requested from the user.
// It checks before doing so, if the user has admin rights.
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func createUser(cmd *cobra.Command, args []string) error {

	RPPrinter.CustomPrintLn(userAddCmdAnsLine0)

	// Verify that the current user has admin rights in order to create users
	isAdmin, err := isCurrentUserAdmin("user")
	if err != nil {
		return err
	}
	if !isAdmin {
		Log.Error(userCurrentNotAdmin)
		return errors.New(userCurrentNotAdminMsg)
	}

	// The user has enough rights to create a new user, let's process its request

	// Create the structure to fill
	user := rpapiv1.RPUserFull{}

	// Prepare the application structure to send to the backend
	if err = prepareUserStruct(cmd, args, &user, true); err != nil {
		return err
	}

	Log.Infof(userCreationInfoMsg, user.ID)
	Log.Debugf(userCreationDebugMsg, user)

	// Tell the user that the user is currently being created
	RPPrinter.CustomPrintF(userAddCmdAnsLine1, user.Pseudo)

	// Request the creation of this user to the backend
	userAns := rpapiv1.RPUserFull{}
	if err := bckAPI.UserCreate(&user, &userAns); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	// Display something to the user if in raw mode
	RPPrinter.CustomPrintObject(userAns, false, displayNothing)

	// Create the sentence to display when the object will be added
	slugSentence := fmt.Sprintf(objectSlugDisplayMsg, "user", userAns.Pseudo)

	// Check from viper the non-blocking flag to see if we wait
	nonBlocking := viper.GetBool(nonBlockingFlag)
	if nonBlocking {
		// The user do not want blocking functions
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
		// Display the slug of the created object
		RPPrinter.CustomPrintLn(slugSentence)
		return nil
	}

	// Block until we receive all the websocket events we need to confirm creation
	var rUser rpapiv1.RPUser
	for {
		rEvent := <-cmdEventChannel

		// Check that the received event is for a user addition
		if rEvent.Type == rpapiv1.EVTUserAdd {
			// Function handling the events
			rUser, err = rEvent.DecodeUser()
			if err != nil {
				RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
				return err
			}

			if utils.UUIDEqual(rUser.ID, userAns.ID) {
				// The user is completely created
				RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
				// Display the slug of the created object
				RPPrinter.CustomPrintLn(slugSentence)
				break
			}
		}

		// Check if we get an error event during the operation
		if (rEvent.Type == rpapiv1.EVTError) || (rEvent.Type == rpapiv1.EVTSessionExpired) {
			// Function handling the events
			err := handleErrorEvent(userAns.ID, rEvent, true)
			if err != nil {
				return err
			}
		}
	}

	return nil
}

// updateUser update a user info as requested from the user.
// It checks before doing so, if the user has admin rights.
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func updateUser(cmd *cobra.Command, args []string) error {

	RPPrinter.CustomPrintLn(userUpdCmdAnsLine0)

	// Get the user ID and return an error if there is nothing
	userID, err := getFirstArg(args, userIDNotSetMsg, userIDRequiredMsg)
	if err != nil {
		return err
	}

	if userID != "me" {
		// For the moment, only the update of "me" is available
		return errors.New(userUpdateOnlyMeMsg)
	}

	// Get the current user information
	userMe := rpapiv1.RPUserFull{}
	if err := bckAPI.UsersGetMe(&userMe); err != nil {
		return err
	}

	// Prepare the application structure to send to the backend
	if err := prepareUserStruct(cmd, args, &userMe, false); err != nil {
		return err
	}

	Log.Infof(userUpdateInfoMsg, userMe.Pseudo)
	Log.Debugf(userUpdateDebugMsg, userMe)

	// Tell the user that the user is currently being updated
	RPPrinter.CustomPrintF(userUpdCmdAnsLine1, userMe.Pseudo)

	// Request the update of this user to the backend
	userAns := rpapiv1.RPUserFull{}
	if err := bckAPI.UserUpdateMe(&userMe, &userAns); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	// Display something to the user if in raw mode
	RPPrinter.CustomPrintObject(userAns, false, displayNothing)

	// Check from viper the non-blocking flag to see if we wait
	nonBlocking := viper.GetBool(nonBlockingFlag)
	if nonBlocking {
		// The user do not want blocking functions
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
		return nil
	}

	// Block until we receive all the websocket events we need to confirm creation
	var rUser rpapiv1.RPUser
	for {
		rEvent := <-cmdEventChannel

		// Check that the received event is for a user addition
		if rEvent.Type == rpapiv1.EVTUserStateChange {
			// Function handling the events
			rUser, err = rEvent.DecodeUser()
			if err != nil {
				RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
				return err
			}

			if utils.UUIDEqual(rUser.ID, userAns.ID) {
				// The user is completely created
				RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
				// Display something to the user if in raw mode
				RPPrinter.CustomPrintObject(rUser, false, displayNothing)
				break
			}
		}

		// Check if we get an error event during the operation
		if (rEvent.Type == rpapiv1.EVTError) || (rEvent.Type == rpapiv1.EVTSessionExpired) {
			// Function handling the events
			err := handleErrorEvent(userAns.ID, rEvent, true)
			if err != nil {
				return err
			}
		}
	}

	return nil
}

// disableUser disable a given user as requested from the user.
// It checks before doing so, if the user is already disabled
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func disableUser(cmd *cobra.Command, args []string) error {

	RPPrinter.CustomPrintLn(userDsbCmdAnsLine0)

	// Get the user ID and return an error if there is nothing
	userID, err := getFirstArg(args, userIDNotSetMsg, userIDRequiredMsg)
	if err != nil {
		return err
	}

	// Get the user's structure from what the user gave us
	userToDisable := rpapiv1.RPUserFull{}
	if err = getUserStructFromAnyID(userID, &userToDisable); err != nil {
		return err
	}

	// Now, the user to disable has been found

	// Check that this user is enabled
	RPPrinter.CustomPrintF(userDsbCmdAnsLine1, userToDisable.Pseudo)
	if userToDisable.Disabled {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return errors.New(userAlreadyDisabledMsg)
	}
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	// Tell the user that we are currently requesting the backend to disable the user
	RPPrinter.CustomPrintF(userDsbCmdAnsLine2, userToDisable.Pseudo)

	// Request the disabling of this user
	userAns := rpapiv1.RPUserFull{}
	if err := bckAPI.UserDisableEnable(userToDisable.ID.String(), &userAns); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	// Display something to the user if in raw mode
	RPPrinter.CustomPrintObject(userAns, false, displayNothing)

	// Check from viper the non-blocking flag to see if we wait
	nonBlocking := viper.GetBool(nonBlockingFlag)
	if nonBlocking {
		// The user do not want blocking functions
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
		return nil
	}

	// Block until we receive all the websocket events we need to confirm creation
	var rUser rpapiv1.RPUser
	for {
		rEvent := <-cmdEventChannel

		// Check that the received event is for a user addition
		if rEvent.Type == rpapiv1.EVTUserDisable {
			// Function handling the events
			rUser, err = rEvent.DecodeUser()
			if err != nil {
				RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
				return err
			}

			if utils.UUIDEqual(rUser.ID, userAns.ID) {
				// The user is completely created
				RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
				// Display something to the user if in raw mode
				RPPrinter.CustomPrintObject(rUser, false, displayNothing)
				break
			}
		}
		// Check if we get an error event during the operation
		if (rEvent.Type == rpapiv1.EVTError) || (rEvent.Type == rpapiv1.EVTSessionExpired) {
			// Function handling the events
			err := handleErrorEvent(userAns.ID, rEvent, true)
			if err != nil {
				return err
			}
		}
	}

	return nil
}

// enableUser enable a given user as requested from the user.
// It checks before doing so, if the user is already enabled
//
// :param cmd: Pointer to the cobra command received
// :type cmd: *cobra.Command
// :param args: Arguments passed by the user to the command
// :type args: []string
// :returns: error -- an error if something went wrong, nil in the other case
func enableUser(cmd *cobra.Command, args []string) error {

	RPPrinter.CustomPrintLn(userEnbCmdAnsLine0)

	// Get the user ID and return an error if there is nothing
	userID, err := getFirstArg(args, userIDNotSetMsg, userIDRequiredMsg)
	if err != nil {
		return err
	}

	// Get the user's structure from what the user gave us
	userToEnable := rpapiv1.RPUserFull{}
	if err = getUserStructFromAnyID(userID, &userToEnable); err != nil {
		return err
	}

	// Now, the user to enable has been found

	// Check that this user is enabled
	RPPrinter.CustomPrintF(userEnbCmdAnsLine1, userToEnable.Pseudo)
	if !userToEnable.Disabled {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return errors.New(userAlreadyEnabledMsg)
	}
	RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)

	// Tell the user that we are currently requesting the backend to enable the user
	RPPrinter.CustomPrintF(userEnbCmdAnsLine2, userToEnable.Pseudo)

	// Request the disabling of this user
	userAns := rpapiv1.RPUserFull{}
	if err := bckAPI.UserDisableEnable(userToEnable.ID.String(), &userAns); err != nil {
		RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
		return err
	}

	// Display something to the user if in raw mode
	RPPrinter.CustomPrintObject(userAns, false, displayNothing)

	// Check from viper the non-blocking flag to see if we wait
	nonBlocking := viper.GetBool(nonBlockingFlag)
	if nonBlocking {
		// The user do not want blocking functions
		RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
		return nil
	}

	// Block until we receive all the websocket events we need to confirm creation
	var rUser rpapiv1.RPUser
	for {
		rEvent := <-cmdEventChannel

		// Check that the received event is for a user addition
		if rEvent.Type == rpapiv1.EVTUserDisable {
			// Function handling the events
			rUser, err = rEvent.DecodeUser()
			if err != nil {
				RPPrinter.CustomColorPrint(printer.ColorRed, globalKOLabel)
				return err
			}

			if utils.UUIDEqual(rUser.ID, userAns.ID) {
				// The user is completely created
				RPPrinter.CustomColorPrint(printer.ColorGreen, globalOKLabel)
				// Display something to the user if in raw mode
				RPPrinter.CustomPrintObject(rUser, false, displayNothing)
				break
			}
		}

		// Check if we get an error event during the operation
		if (rEvent.Type == rpapiv1.EVTError) || (rEvent.Type == rpapiv1.EVTSessionExpired) {
			// Function handling the events
			err := handleErrorEvent(userAns.ID, rEvent, true)
			if err != nil {
				return err
			}
		}
	}

	return nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Utilities functions ------------------------------------------
// --------------------------------------------------------------

// isCurrentUserAdmin is a private function allowing to know if
// the current user has administrator or super-administrator rights
//
// :param object: object on which the user needs to be admin,
// if empty any admin will be taken into account
// :type object: string
// :returns: bool, error -- True if the current user is admin and
// an error if something went wrong, nil in the other case
func isCurrentUserAdmin(object string) (bool, error) {
	// Set default return value
	resBool := false

	// Set the structure to fill
	roles := rpapiv1.RPRoles{}

	// Request the Redpesk backend
	if err := bckAPI.UserGetMyRole(&roles); err != nil {
		return resBool, err
	}

	// Check if we are admin
	for _, role := range roles.Roles {

		// If the object is empty, any admin would be ok
		objToCheck := object
		if object == "" {
			objToCheck = role.Object
		}

		if ((role.Label == "admin") && (role.Object == objToCheck)) ||
			(role.Label == "superAdmin") {
			resBool = true
			break
		}
	}

	// Return the result
	return resBool, nil
}

// getUserUUID is a private function allowing to retrieve the
// UUID of a user from a given identifier. This identifier can
// be the UUID itself as a string representation or the Pseudo of the
// user
//
// :param identifier: the identifier of the user
// :type identifier: string
// :returns: (uuid.UUID, error) -- the UUID of the user and an error if something went bad
func getUserUUID(identifier string) (uuid.UUID, error) {

	// Optimization: pass through the cache in order to avoid several requests for Teams list
	if (cachedRPUsersByUUID == nil) || (cachedRPUsersByPseudo == nil) || (runMode == "testing") {
		err := populateUsersCache()
		if err != nil {
			return uuid.Nil, err
		}
	}

	// Check if we can find the identifier in the different map
	userUUID, ok := cachedRPUsersByPseudo[identifier]
	if ok {
		// The identifier is the pseudo and we found the uuid !
		return userUUID, nil
	}

	tmpUUID, err := uuid.FromString(identifier)
	if err == nil {
		_, ok = cachedRPUsersByUUID[tmpUUID]
		if ok {
			// The identifier is the UUID and we found it !
			return tmpUUID, nil
		}
	}

	// If we reached this part of code, we did not find anything
	return uuid.Nil, fmt.Errorf(userIdentifierNotFound, identifier)
}

// populateUsersCache is a function that needs to be called when
// users will be used in the user request. The function fills the
// global variable "cachedRPUsersByUUID" and "cachedRPUsersByPseudo"
// from a request to the Redpesk backend.
//
// :returns: error -- An error if something went bad
func populateUsersCache() error {

	// Get all users list
	allUsers := rpapiv1.RPUsersFull{}
	if err := bckAPI.UsersFullGetList(&allUsers); err != nil {
		return err
	}

	// Create the map if it does not exists
	if (cachedRPUsersByUUID == nil) || (runMode == "testing") {
		cachedRPUsersByUUID = make(map[uuid.UUID]rpapiv1.RPUserFull)
	}

	// Create the map if it does not exists
	if (cachedRPUsersByPseudo == nil) || (runMode == "testing") {
		cachedRPUsersByPseudo = make(map[string]uuid.UUID)
	}

	// We got correctly the users list, fill the cache now
	for _, user := range allUsers.Users {

		// Copy each user in the cache
		tmpUser := rpapiv1.RPUserFull{}
		copier.Copy(&tmpUser, &user)
		cachedRPUsersByUUID[user.ID] = tmpUser

		// Fill the other cache
		cachedRPUsersByPseudo[user.Pseudo] = user.ID
	}

	return nil
}

// getUserPointer gets the pointer to the user for the given ID
//
// :param userID: ID of the user wanted
// :type userID: uuid.UUID
// :param userTab: Pointer to a tab users structure to fill (can be *rpapiv1.RPUsers or *rpapiv1.RPUsersFull)
// :type userTab: interface{}
// :returns: interface{}, error -- a pointer on the researched user and an error if something went wrong, nil in the other case
func getUserPointer(userID uuid.UUID, userTab interface{}) (interface{}, error) {
	userFound := false
	var userPointer interface{}

	switch typedTab := userTab.(type) {

	case *rpapiv1.RPUsersFull:
		// Get the full list of users
		if err := bckAPI.UsersFullGetList(typedTab); err != nil {
			return userPointer, err
		}

		// In the resulting list select only the user that interest us
		for index, user := range typedTab.Users {
			if utils.UUIDEqual(user.ID, userID) {
				userFound = true
				userPointer = &(typedTab.Users[index])
				break
			}
		}

	default:
		// The type is not recognized!
		return userPointer, errors.New(userTypeNotRecognized)
	}

	// Check if we found the right application
	if !userFound {
		return nil, errors.New(userNotFoundMsg)
	}

	return userPointer, nil
}

// getUserStructFromAnyID allows to get the full rpapiv1.RPUserFull structure
// from any ID that identifies a user: me, <pseudo>, <UUID>, etc.
//
// :param identifier: the identifier of the user
// :type identifier: string
// :param userBuffer: The buffer structure that the function will fill with the right info
// :type userBuffer: *rpapiv1.RPUserFull
// :returns: error -- an error if something went wrong, nil in the other case
func getUserStructFromAnyID(identifier string, userBuffer *rpapiv1.RPUserFull) error {

	if identifier == "me" {
		// The user is interested by himself

		// Get the current user information
		if err := bckAPI.UsersGetMe(userBuffer); err != nil {
			return err
		}

	} else {
		// The user wants to disable another user

		// Check that the current user is admin
		isAdmin, err := isCurrentUserAdmin("user")
		if err != nil {
			return err
		}

		if isAdmin {

			// Get the UUID of the user
			userUUID, err := getUserUUID(identifier)
			if err != nil {
				return err
			}

			// The current user is admin, we can ask for full info
			userTab := rpapiv1.RPUsersFull{}
			var userInterfacePointer interface{}

			// Get the user that is interesting us
			if userInterfacePointer, err = getUserPointer(userUUID, &userTab); err != nil {
				return err
			}

			// Save the pointer with the right type
			tmpUser := userInterfacePointer.(*rpapiv1.RPUserFull)

			// Copy the values inside the buffer
			if err = copier.Copy(userBuffer, tmpUser); err != nil {
				return err
			}

		} else {
			// Only authorized for the admins!
			Log.Error(userCurrentNotAdmin)
			return errors.New(userCurrentNotAdminMsg)
		}
	}

	return nil
}

// --------------------------------------------------------------
// --------------------------------------------------------------

// --------------------------------------------------------------
// Display functions --------------------------------------------
// --------------------------------------------------------------

// displayUsers is a private function allowing to display
// a slice of users in the console
//
// :param users: tab of users to display (can be []rpapiv1.RPUser or []rpapiv1.RPUserFull)
// :type users: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func displayUsers(users interface{}, verbose bool, writer *tabwriter.Writer) error {

	// Display result
	first := true

	// Convert to the right type
	usersTab := users.([]rpapiv1.RPUserFull)

	for _, user := range usersTab {
		if verbose {
			if !first {
				fmt.Fprintln(writer)
			}

			// Print this user
			_printOneUser(user, verbose, false, writer)

		} else {
			if first {
				manageHeaderPrint(usersListCmdHeader, writer)
			}
			// Print this project
			_printOneUser(user, verbose, false, writer)
		}
		first = false
	}

	return nil
}

// displayOneUser is a private function allowing to display
// one user in the console
//
// :param user: User structure to display (can be rpapiv1.RPUser or rpapiv1.RPUserFull)
// :type user: interface{}
// :param verbose: Verbose option given by the user
// :type verbose: bool
// :param writer: Writer to use to print the information
// :type writer: *tabwriter.Writer
// :returns: error -- an error if something went bad
func displayOneUser(user interface{}, verbose bool, writer *tabwriter.Writer) error {
	// Print the project
	_printOneUser(user, verbose, true, writer)

	return nil
}

// _printOneUser is a private function allowing to print one user in the console
//
// :param user: user to display (can be rpapiv1.RPUser or rpapiv1.RPUserFull)
// :type user: interface{}
// :param verbose: True if the user wants verbosity in the answer
// :type verbose: bool
// :param isAlone: True if there is only one user to display in the context where this function is called
// :type isAlone: bool
// :param writer: Writer to used to print the user information
// :type writer: *tabwriter.Writer
func _printOneUser(user interface{}, verbose bool, isAlone bool, writer *tabwriter.Writer) {
	// Mode administrator
	userToDisplay := user.(rpapiv1.RPUserFull)

	if verbose {
		// Mode verbose
		// Printing the result on stdout
		fmt.Fprintln(writer, usersListCmdVerboseLine0, userToDisplay.ID.String())
		fmt.Fprintln(writer, usersListCmdVerboseLine6, userToDisplay.Pseudo)
		fmt.Fprintln(writer, usersListCmdVerboseLine1, userToDisplay.Email)
		fmt.Fprintln(writer, usersListCmdVerboseLine2, userToDisplay.Firstname)
		fmt.Fprintln(writer, usersListCmdVerboseLine3, userToDisplay.Lastname)
		fmt.Fprintln(writer, usersListCmdVerboseLine4, userToDisplay.Phone)
		fmt.Fprintln(writer, usersListCmdVerboseLine5, strconv.FormatBool(userToDisplay.Disabled))

	} else {
		// Mode non-verbose
		if isAlone {
			fmt.Fprintln(writer, usersListCmdHeader)
		}
		fmt.Fprintln(writer, userToDisplay.Pseudo, "\t", userToDisplay.Firstname, "\t", userToDisplay.Lastname)
	}
}

// --------------------------------------------------------------
// --------------------------------------------------------------
