/*
Copyright © 2020-2023 IoT.bzh Company
Contact: https://www.iot.bzh/licensing

This file is part of the redpesk-cli project, the redpesk command line tool.

$RP_BEGIN_LICENSE$
Commercial License Usage
 Licensees holding valid commercial IoT.bzh licenses may use this file in
 accordance with the commercial license agreement provided with the
 Software or, alternatively, in accordance with the terms contained in
 a written agreement between you and The IoT.bzh Company. For licensing terms
 and conditions see https://www.iot.bzh/terms-conditions. For further
 information use the contact form at https://www.iot.bzh/contact.

GNU General Public License Usage
 Alternatively, this file may be used under the terms of the GNU General
 Public license version 3. This license is as published by the Free Software
 Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
 of this file. Please review the following information to ensure the GNU
 General Public License requirements will be met
 https://www.gnu.org/licenses/gpl-3.0.html.
$RP_END_LICENSE$
*/
package cmd

import (
	"testing"

	"git.ovh.iot/redpesk/redpesk-infra/rp-webserver/actions/rpapiv1"
	"github.com/stretchr/testify/assert"
)

const (
	// Users
	userDisplayNoVerboseColNb = 3
)

// TestUsersAddNoBlock is the function allowing to test if
// the command "rp-cli users add --nonblocking ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestUsersAddNoBlock(test *testing.T) {
	// Create the command
	cmdToRun := getMainAddUserCmd("the_banano", "banano@iot.bzh", "Armand", "Bénéteau", "0666989945")
	// Add the non-blocking flag
	cmdToRun = append(cmdToRun, "--nonblocking")

	// Create the user if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, usersCommand, "the_banano", cmdToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// The user has been created, check it is now in backend

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"Pseudo":    `^the_banano$`,
		"Firstname": `^Armand$`,
		"Lastname":  `^Bénéteau$`,
	}

	// Run the command and the tests (no verbose and verbose, since this is the same result)
	checkCmdAnsKeysRegexpCol(test, userDisplayNoVerboseColNb, regexpMap, usersCommand, "get", "the_banano")
}

// TestUsersAddBlock is the function allowing to test if
// the command "rp-cli users add ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestUsersAddBlock(test *testing.T) {
	// Create the command
	cmdToRun := getMainAddUserCmd("tiramisu", "tiramisu@iot.bzh", "Ti", "Ramisu", "0666989945")

	// Create the user if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, usersCommand, "tiramisu", cmdToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// The user has been created, check it is now in backend

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"Pseudo":    `^tiramisu$`,
		"Firstname": `^Ti$`,
		"Lastname":  `^Ramisu$`,
	}

	// Run the command and the tests (no verbose and verbose, since this is the same result)
	checkCmdAnsKeysRegexpCol(test, userDisplayNoVerboseColNb, regexpMap, usersCommand, "get", "tiramisu")
}

// TestUsersAddNotUserAdmin is the function allowing to test if
// the command "rp-cli users add ..." is failing correctly when the
// current user is not an "user admin"
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestUsersAddNotUserAdmin(test *testing.T) {
	// Create the command
	cmdToRun := getMainAddUserCmd("tiramisu", "tiramisu@iot.bzh", "Ti", "Ramisu", "0666989945")

	// Set-up the mock in order for the current user to not be an admin user
	dummyRoles := rpapiv1.RPRoles{
		Roles: []rpapiv1.RPRole{
			{
				Name:   "factoryMember of user",
				Label:  "factoryMember",
				Object: "user",
			},
		},
	}
	MockedRPBackend.SetCurrentUserRole(dummyRoles, false)

	// Run the command and check that an error is returned
	_, err := executeCommand(cmdToRun...)

	// Put the mock to its default value
	MockedRPBackend.SetCurrentUserRole(rpapiv1.RPRoles{}, true)

	// Test that we have an error
	assert.Error(test, err)
	assert.Contains(test, err.Error(), "you are not admin")
}

// TestUsersGet is the function allowing to test if
// the command "rp-cli users get ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestUsersGet(test *testing.T) {
	// Create the commands
	cmdToRun := getMainAddUserCmd("kravmager", "krav@iot.bzh", "Imi", "Lichtenfeld", "0666999949")

	// Create the user if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, usersCommand, "kravmager", cmdToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// Verbose only, since the no verbose option is tested in almost all other tests

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":       `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Pseudo":    `^kravmager$`,
		"Email":     `^krav@iot.bzh$`,
		"Firstname": `^Imi$`,
		"Lastname":  `^Lichtenfeld$`,
		"Phone":     `^[0-9]{10}$`,
		"Disabled":  `^(true|false)$`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, userDisplayVerboseLineNb, regexpMap, usersCommand, "get", "kravmager", "-v")
}

// TestUsersList is the function allowing to test if
// the command "rp-cli users list ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestUsersList(test *testing.T) {
	// Create the commands
	cmdToRun1 := getMainAddUserCmd("the_banano", "banano@iot.bzh", "Armand", "Bénéteau", "0666989945")
	cmdToRun2 := getMainAddUserCmd("tiramisu", "tiramisu@iot.bzh", "Ti", "Ramisu", "0666989945")
	cmdToRun3 := getMainAddUserCmd("kravmager", "krav@iot.bzh", "Imi", "Lichtenfeld", "0666999949")

	// Create the users if they do not exist, and check that all steps are OK
	err := addObjectIfNotExists(test, usersCommand, "the_banano", cmdToRun1...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, usersCommand, "tiramisu", cmdToRun2...)
	if err != nil {
		test.Fatal(err)
	}
	err = addObjectIfNotExists(test, usersCommand, "kravmager", cmdToRun3...)
	if err != nil {
		test.Fatal(err)
	}

	// The users have been created, now check the list command

	// Not verbose mode
	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"Pseudo":    `^[a-z0-9-_]+$`,
		"Firstname": `.+`,
		"Lastname":  `.+`,
	}

	// Run the command and the tests (no verbose and verbose, since this is the same result)
	checkCmdAnsKeysRegexpCol(test, userDisplayNoVerboseColNb, regexpMap, usersCommand, "list")

	// Verbose mode
	// Create the map containing the regexp we want to match in the answer
	regexpMap2 := map[string]string{
		"ID:":       `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Pseudo":    `^[a-z0-9-_]+$`,
		"Email":     `.+@iot.bzh$`,
		"Firstname": `.+`,
		"Lastname":  `.+`,
		"Phone":     `^[0-9]{10}$`,
		"Disabled":  `^(true|false)$`,
	}

	// Run the command and the tests
	checkCmdAnswerListKeysRegexp(test, userDisplayVerboseLineNb, "", regexpMap2, usersCommand, "list", "-v")
}

// TestUsersUpdate is the function allowing to test if
// the command "rp-cli users update ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestUsersUpdate(test *testing.T) {

	// Update of another user than me is not supported for the moment
	// To fix later

	// Create the commands
	cmdToRun := getMainAddUserCmd("me", "banano@iot.bzh", "Armand", "Bénéteau", "0666989945")

	// Create the user if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, usersCommand, "me", cmdToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// Create the update command
	updateCmd := getMainUpdateUserCmd("me", "banano@iot.bzh", "Banano", "Banano", "0666989945")

	// Run the update command and check that all is [OK]
	checkActionCmdAns(test, updateCmd...)

	// Check that the description has been updated

	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":       `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Pseudo":    `^me$`,
		"Email":     `^banano@iot.bzh$`,
		"Firstname": `^Banano$`,
		"Lastname":  `^Banano$`,
		"Phone":     `^[0-9]{10}$`,
		"Disabled":  `^(true|false)$`,
	}

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, userDisplayVerboseLineNb, regexpMap, usersCommand, "get", "me", "-v")

	// Now change only the email
	updateCmd = getMainUpdateUserCmd("me", "george.abitbol@iot.bzh", "", "", "")

	// Run the update command and check that all is [OK]
	checkActionCmdAns(test, updateCmd...)

	// Check that the email has been updated

	// Modify the map containing the regexp we want to match in the answer
	regexpMap["Email"] = `^george.abitbol@iot.bzh$`

	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, userDisplayVerboseLineNb, regexpMap, usersCommand, "get", "me", "-v")
}

// TestUsersEnableDisable is the function allowing to test if
// the command "rp-cli users get ..." is working correctly
//
// :param test: Pointer to T testing object
// :type test: *testing.T
func TestUsersEnableDisable(test *testing.T) {

	if bckIsReal {
		test.Log("[WARNING] Enabling and Disabling users is not supported by the true backend...")
		return
	}

	// Create the commands
	cmdToRun := getMainAddUserCmd("tiramisu", "tiramisu@iot.bzh", "Ti", "Ramisu", "0666989945")

	// Create the user if it does not exists, and check that all steps are OK
	err := addObjectIfNotExists(test, usersCommand, "tiramisu", cmdToRun...)
	if err != nil {
		test.Fatal(err)
	}

	// Check that the user is enable
	// Create the map containing the regexp we want to match in the answer
	regexpMap := map[string]string{
		"ID:":       `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Pseudo":    `^tiramisu$`,
		"Email":     `^tiramisu@iot.bzh$`,
		"Firstname": `^Ti$`,
		"Lastname":  `^Ramisu$`,
		"Phone":     `^[0-9]{10}$`,
		"Disabled":  `^false$`,
	}
	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, userDisplayVerboseLineNb, regexpMap, usersCommand, "get", "tiramisu", "-v")

	// Run the command to disable the user and check that all is [OK]
	checkActionCmdAns(test, usersCommand, "disable", "tiramisu")

	// Check that the user is disable
	// Create the map containing the regexp we want to match in the answer
	regexpMap = map[string]string{
		"ID:":       `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Pseudo":    `^tiramisu$`,
		"Email":     `^tiramisu@iot.bzh$`,
		"Firstname": `^Ti$`,
		"Lastname":  `^Ramisu$`,
		"Phone":     `^[0-9]{10}$`,
		"Disabled":  `^true$`,
	}
	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, userDisplayVerboseLineNb, regexpMap, usersCommand, "get", "tiramisu", "-v")

	// Run the command to enable the user and check that all is [OK]
	checkActionCmdAns(test, usersCommand, "enable", "tiramisu")

	// Check that the user is enable
	// Create the map containing the regexp we want to match in the answer
	regexpMap = map[string]string{
		"ID:":       `[a-f0-9]{8}(-[a-f0-9]{4}){3}-[a-f0-9]{12}`,
		"Pseudo":    `^tiramisu$`,
		"Email":     `^tiramisu@iot.bzh$`,
		"Firstname": `^Ti$`,
		"Lastname":  `^Ramisu$`,
		"Phone":     `^[0-9]{10}$`,
		"Disabled":  `^false$`,
	}
	// Run the command and the tests
	checkCmdAnswerKeysRegexp(test, userDisplayVerboseLineNb, regexpMap, usersCommand, "get", "tiramisu", "-v")
}

// --------------------------------------------------------------
// Utilities functions ------------------------------------------
// --------------------------------------------------------------

// getMainAddUserCmd is an initializer function to get the mainly
// used command to create a user
//
// :param pseudo: Pseudo of the user to create
// :type pseudo: string
// :param email: email of the user to create
// :type email: string
// :param fName: Firstname of the user to create
// :type fName: string
// :param lName: Lastname of the user to create
// :type lName: string
// :param phone: Phone number of the user to create
// :type phone: string
// :returns: []string -- the value of the wanted constant
func getMainAddUserCmd(pseudo string, email string, fName string, lName string, phone string) []string {
	return []string{
		usersCommand,
		userAddCmdName,
		"-p",
		pseudo,
		"-e",
		email,
		"-f",
		fName,
		"-n",
		lName,
		"--phone",
		phone,
	}
}

// getMainUpdateUserCmd is an initializer function to get the mainly
// used command to update a user
//
// :param pseudo: Pseudo of the user to create
// :type pseudo: string
// :param email: email of the user to create
// :type email: string
// :param fName: Firstname of the user to create
// :type fName: string
// :param lName: Lastname of the user to create
// :type lName: string
// :param phone: Phone number of the user to create
// :type phone: string
// :returns: []string -- the value of the wanted constant
func getMainUpdateUserCmd(pseudo string, email string, fName string, lName string, phone string) []string {
	cmdToReturn := []string{
		usersCommand,
		"update",
		pseudo,
	}

	if email != "" {
		cmdToReturn = append(cmdToReturn, "-e", email)
	}

	if fName != "" {
		cmdToReturn = append(cmdToReturn, "-f", fName)
	}

	if lName != "" {
		cmdToReturn = append(cmdToReturn, "-n", lName)
	}

	if phone != "" {
		cmdToReturn = append(cmdToReturn, "--phone", phone)
	}

	return cmdToReturn
}

// --------------------------------------------------------------
// --------------------------------------------------------------
