/**
 * Copyright (C) 2020-2023 IoT.bzh Company
 * Contact: https://www.iot.bzh/licensing
 *
 * This file is part of the redpesk-cli project, the redpesk command line tool.
 *
 * $RP_BEGIN_LICENSE$
 * Commercial License Usage
 *  Licensees holding valid commercial IoT.bzh licenses may use this file in
 *  accordance with the commercial license agreement provided with the
 *  Software or, alternatively, in accordance with the terms contained in
 *  a written agreement between you and The IoT.bzh Company. For licensing terms
 *  and conditions see https://www.iot.bzh/terms-conditions. For further
 *  information use the contact form at https://www.iot.bzh/contact.
 *
 * GNU General Public License Usage
 *  Alternatively, this file may be used under the terms of the GNU General
 *  Public license version 3. This license is as published by the Free Software
 *  Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
 *  of this file. Please review the following information to ensure the GNU
 *  General Public License requirements will be met
 *  https://www.gnu.org/licenses/gpl-3.0.html.
 * $RP_END_LICENSE$
 */

package printer

import (
	"encoding/json"
	"errors"
	"io"
	"os"
	"text/tabwriter"

	"github.com/sirupsen/logrus"
	"golang.org/x/term"
)

// -------------------------------------------------------------------------------
// Types definition --------------------------------------------------------------
// -------------------------------------------------------------------------------

// Printer is an interface to printthings to the user
type Printer interface {
	CustomPrint(a ...interface{}) error
	CustomColorPrint(color string, a ...interface{}) error
	CustomPrintLn(a ...interface{}) error
	CustomColorPrintLn(color string, a ...interface{}) error
	CustomPrintF(format string, a ...interface{}) error
	CustomColorPrintF(color string, format string, a ...interface{}) error
	CustomPrintObject(object interface{}, verbose bool, printFancyCB func(interface{}, bool, *tabwriter.Writer)) error
	ForceFlush()
}

// RPPrinter definition
type RPPrinter struct {
	Writer     *tabwriter.Writer
	Log        *logrus.Logger
	RawMode    bool
	NoPager    bool
	TestMode   bool
	IsTerminal bool
}

// RPPrinterPaging definition
type RPPrinterPaging struct {
	Page         int `json:"page"`
	PerPage      int `json:"per_page"`
	TotalEntries int `json:"total_entries"`
}

// RPPagingToolsStruct definition
type RPPagingToolsStruct struct {
	Verbose       bool // Gives the verbosity for printing
	VerboseLineNb int  // Number of lines consumed by 1 object when verbosity is ON
	GetObjCB      func(pageNb, perPage int) (objsToPrint interface{}, paging RPPrinterPaging, err error)
	PrintFancyCB  func(obj interface{}, verbose bool, writer *tabwriter.Writer) error
}

// -------------------------------------------------------------------------------
// -------------------------------------------------------------------------------

// -------------------------------------------------------------------------------
// Functions definition ----------------------------------------------------------
// -------------------------------------------------------------------------------

// InitRPPrinter initializes the printer interface used to display message to the
// user
//
// :param logR: Pointer to the logger used by the application
// :type logR: *logrus.Logger
// :param rawMode: True if the rawmode is activated, False in the other case
// :type rawMode: bool
// :param noPager: True if the noPager is activated, False in the other case
// :type noPager: bool
// :param testMode: True if the testmode is activated, False in the other case
// :type testMode: bool
// :param output: If different than nil, buffer to use to print
// :type output: io.Writer
// :returns: pointer to the RPAPIInterface object
func InitRPPrinter(logR *logrus.Logger, rawMode, noPager, testMode bool, output io.Writer) (*RPPrinter, error) {

	// Check that the pointer to the logger is not nil
	if logR == nil {
		return nil, errors.New(initRPPrinterError)
	}

	// Initialize the tabwriter
	tabWriter := NewTableWriter(output)

	// Initialize the RPAPIInterface
	rpPrinter := RPPrinter{
		Writer:     tabWriter,
		Log:        logR,
		RawMode:    rawMode,
		NoPager:    noPager,
		TestMode:   testMode,
		IsTerminal: term.IsTerminal(int(os.Stdout.Fd())),
	}

	return &rpPrinter, nil
}

// NewTableWriter Create a writer that inserts padding around tab-delimited
func NewTableWriter(output io.Writer) *tabwriter.Writer {
	writer := new(tabwriter.Writer)

	if output != nil {
		writer.Init(output, 0, 8, 0, '\t', 0)
	} else {
		writer.Init(os.Stdout, 0, 8, 0, '\t', 0)
	}
	return writer
}

// GetPagingFromInterface creates and fill a RPPrinterPaging object from
// any json compatible structure containing the fields page, per_page and total_entries
//
// :param object: If different than nil, buffer to use to print
// :type object: interface{}
// :returns: RPPrinterPaging object and an error if a problem occurred
func GetPagingFromInterface(object interface{}) (RPPrinterPaging, error) {
	// Initialize the return value
	pPagingObj := RPPrinterPaging{}

	// Good ol' Marshal/Unmarshal tricks!
	dataBytes, err := json.Marshal(object)
	if err != nil {
		return pPagingObj, err
	}
	err = json.Unmarshal(dataBytes, &pPagingObj)

	return pPagingObj, err
}

// -------------------------------------------------------------------------------
// -------------------------------------------------------------------------------
