/*
Copyright © 2020-2023 IoT.bzh Company
Contact: https://www.iot.bzh/licensing

This file is part of the redpesk-cli project, the redpesk command line tool.

$RP_BEGIN_LICENSE$
Commercial License Usage
 Licensees holding valid commercial IoT.bzh licenses may use this file in
 accordance with the commercial license agreement provided with the
 Software or, alternatively, in accordance with the terms contained in
 a written agreement between you and The IoT.bzh Company. For licensing terms
 and conditions see https://www.iot.bzh/terms-conditions. For further
 information use the contact form at https://www.iot.bzh/contact.

GNU General Public License Usage
 Alternatively, this file may be used under the terms of the GNU General
 Public license version 3. This license is as published by the Free Software
 Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
 of this file. Please review the following information to ensure the GNU
 General Public License requirements will be met
 https://www.gnu.org/licenses/gpl-3.0.html.
$RP_END_LICENSE$
*/

package printer

import (
	"bytes"
	"encoding/json"
	"fmt"
	"os"
	"strings"
	"text/tabwriter"

	"golang.org/x/term"
)

const (
	// --------------------------------------------------------------
	// Colors available to print ------------------------------------
	// --------------------------------------------------------------
	ColorRed    = "RED"
	ColorGreen  = "GREEN"
	ColorYellow = "YELLOW"
	ColorBlue   = "BLUE"
	ColorPurple = "PURPLE"
	ColorCyan   = "CYAN"
	ColorWhite  = "WHITE"

	ColorReset = "RESET"
	// --------------------------------------------------------------
	// --------------------------------------------------------------

	// --------------------------------------------------------------
	// Hexadecimal raw signals --------------------------------------
	// --------------------------------------------------------------
	rawSigInt  = 0x03
	rawSigQuit = 0x04
	// --------------------------------------------------------------
	// --------------------------------------------------------------
)

// Local global map containing the shell code for colors
var colorsCodes = map[string]string{
	ColorReset:  "\033[0000m",
	ColorRed:    "\033[0031m",
	ColorGreen:  "\033[0032m",
	ColorYellow: "\033[0033m",
	ColorBlue:   "\033[0034m",
	ColorPurple: "\033[0035m",
	ColorCyan:   "\033[0036m",
	ColorWhite:  "\033[0037m",
}

// Bytes slice corresponding to "Previous page" key
var PrevPageKeyBytes = []byte{27, 91, 53, 126}

// Bytes slice corresponding to "Next page" key
var NextPageKeyBytes = []byte{27, 91, 54, 126}

// -------------------------------------------------------------------------------
// Methods definition ------------------------------------------------------------
// -------------------------------------------------------------------------------

// CustomPrint allows the calling function to print something to the user if the
// current mode is not "raw mode"
//
// :param a: Several objects to print
// :type a: ...interface{}
// :returns: an error if a problem appeared, nil if not
func (p *RPPrinter) CustomPrint(a ...interface{}) error {
	// The basic print is actually a ColorPrint with reset characters in order to keep alignment
	return p.CustomColorPrint(ColorReset, a...)
}

// CustomColorPrint allows the calling function to print something to the user if the
// current mode is not "raw mode", with colors!
//
// :param color: Color to use, see colors listed at the beginning of the file
// :type color: string
// :param a: Several objects to print
// :type a: ...interface{}
// :returns: an error if a problem appeared, nil if not
func (p *RPPrinter) CustomColorPrint(color string, a ...interface{}) error {
	// Check if we are in raw mode
	if !p.RawMode {

		if !p.TestMode && p.IsTerminal {
			// Check that the color is available
			colorCode, ok := colorsCodes[color]
			if !ok {
				return fmt.Errorf(colorUnavailable, color)
			}

			// Print the color code wanted by the user and defer the reset
			fmt.Fprint(p.Writer, string(colorCode))
			defer fmt.Fprint(p.Writer, string(colorsCodes[ColorReset]))
		}

		// We are not in raw mode, we can display other kind of lines
		_, err := fmt.Fprint(p.Writer, a...)
		// Return err if an error occurred
		if err != nil {
			return err
		}
	}

	return nil
}

// CustomPrintLn allows the calling function to print something to the user if the
// current mode is not "raw mode"
//
// :param a: Several objects to print
// :type a: ...interface{}
// :returns: an error if a problem appeared, nil if not
func (p *RPPrinter) CustomPrintLn(a ...interface{}) error {
	// The basic print is actually a ColorPrint with reset characters in order to keep alignment
	return p.CustomColorPrintLn(ColorReset, a...)
}

// CustomColorPrintLn allows the calling function to print something to the user if the
// current mode is not "raw mode", with colors!
//
// :param color: Color to use, see colors listed at the beginning of the file
// :type color: string
// :param a: Several objects to print
// :type a: ...interface{}
// :returns: an error if a problem appeared, nil if not
func (p *RPPrinter) CustomColorPrintLn(color string, a ...interface{}) error {
	// Check if we are in raw mode
	if !p.RawMode {

		if !p.TestMode && p.IsTerminal {
			// Check that the color is available
			colorCode, ok := colorsCodes[color]
			if !ok {
				return fmt.Errorf(colorUnavailable, color)
			}

			// Print the color code wanted by the user and defer the reset
			fmt.Fprint(p.Writer, string(colorCode))
			defer fmt.Fprint(p.Writer, string(colorsCodes[ColorReset]))
		}

		// We are not in raw mode, we can display other kind of lines
		_, err := fmt.Fprintln(p.Writer, a...)
		// Return err if an error occurred
		if err != nil {
			return err
		}
	}

	return nil
}

// CustomPrintF allows the calling function to print something to the user if the
// current mode is not "raw mode"
//
// :param format: Formatting string
// :type format: string
// :param a: Several objects to print
// :type a: ...interface{}
// :returns: an error if a problem appeared, nil if not
func (p *RPPrinter) CustomPrintF(format string, a ...interface{}) error {
	// The basic print is actually a ColorPrint with reset characters in order to keep alignment
	return p.CustomColorPrintF(ColorReset, format, a...)
}

// CustomColorPrintF allows the calling function to print something to the user if the
// current mode is not "raw mode", with colors!
//
// :param color: Color to use, see colors listed at the beginning of the file
// :type color: string
// :param a: Several objects to print
// :type a: ...interface{}
// :returns: an error if a problem appeared, nil if not
func (p *RPPrinter) CustomColorPrintF(color string, format string, a ...interface{}) error {
	// Check if we are in raw mode
	if !p.RawMode {

		if !p.TestMode && p.IsTerminal {
			// Check that the color is available
			colorCode, ok := colorsCodes[color]
			if !ok {
				return fmt.Errorf(colorUnavailable, color)
			}

			// Print the color code wanted by the user and defer the reset
			fmt.Fprint(p.Writer, string(colorCode))
			defer fmt.Fprint(p.Writer, string(colorsCodes[ColorReset]))
		}

		// We are not in raw mode, we can display other kind of lines
		_, err := fmt.Fprintf(p.Writer, format, a...)
		// Return err if an error occurred
		if err != nil {
			return err
		}
	}

	return nil
}

// CustomPrintObject allows the calling function to print an object to the user in a
// fancy way if the current mode is not "raw mode", in raw mode in print brutally
// the object
//
// :param object: Object to print
// :type object: interface{}
// :param verbose: Boolean telling if the object needs to be output with a lot of verbs
// :type verbose: bool
// :param printFancyCB: Callback allowing to print the object in a fancy way
// :type printFancyCB: func(interface{}, bool)
// :returns: an error if a problem appeared, nil if not
func (p *RPPrinter) CustomPrintObject(object interface{}, verbose bool, printFancyCB func(interface{}, bool, *tabwriter.Writer) error) error {
	// Check if we are in raw mode
	if !p.RawMode {
		// We are not in raw mode, we can display other kind of lines
		err := printFancyCB(object, verbose, p.Writer)
		// Return err if an error occurred
		if err != nil {
			return err
		}
	} else {
		// Print the result in raw mode
		data, err := json.Marshal(object)
		if err != nil {
			return err
		}
		// Display in raw the object
		_, err = fmt.Fprintf(p.Writer, "%s\n", data)
		// Flush the writer
		p.Writer.Flush()
	}

	return nil
}

// ForceFlush allows the calling function to force the flush of the embedded writer
// of the printer
func (p *RPPrinter) ForceFlush() {
	// Flush the writer
	p.Writer.Flush()
}

// InteractivelyPrintPages prints interactively pages of object in the terminal.
// This is interactive because, thanks to the tools in parameters, the function is able
// to request the next objects to display. It results in an optimization of the number of
// requests done to a server.
//
// :param pagingTools: Pointer to the structure containing all the needed functions and variables
// :type pagingTools: RPPagingToolsStruct
// :param noHeader: Boolean saying if a header is printed or not
// :type noHeader: bool
// :returns: error -- an error if a problem appeared, nil if not
func (p *RPPrinter) InteractivelyPrintPages(pagingTools *RPPagingToolsStruct, noHeader bool) error {

	// Paging not possible if it's not an interactive terminal, so do the same thing as no-pager case, with a terminal of 50 lines height
	if !p.IsTerminal {
		p.NoPager = true
	}

	// Calculate the number of object to have by page
	perPage := 50
	if !p.NoPager {

		_, termHeight, err := term.GetSize(int(os.Stdin.Fd()))

		if err != nil {
			return err
		}
		if pagingTools.Verbose {
			perPage = termHeight / (pagingTools.VerboseLineNb + 2)
			if perPage == 0 {
				perPage = 1
			}
		} else {
			if noHeader {
				perPage = (termHeight - 2)
			} else {
				perPage = (termHeight - 3)
			}

			if perPage < 1 {
				perPage = 1
			}
		}
	}

	// Main loop
	firstLoop, waitForQ := true, true
	pageNb := 1
	for waitForQ {
		// Get the objects list
		objsToPrint, paging, err := pagingTools.GetObjCB(pageNb, perPage)
		if err != nil {
			return err
		}

		// Calculation of the total number of pages needed
		totalPages := (paging.TotalEntries / paging.PerPage)
		modulusEntries := (paging.TotalEntries % paging.PerPage)
		if modulusEntries > 0 {
			totalPages++
		}

		// Check if we reached the end of the pages
		isLastPage := (paging.Page >= totalPages) || (paging.TotalEntries == 0)

		// Print the result to the user
		err = p.CustomPrintObject(objsToPrint, pagingTools.Verbose, pagingTools.PrintFancyCB)
		if err != nil {
			return err
		}

		// Manage cases if it is the last page
		if isLastPage && (p.NoPager || firstLoop) {
			// No need to go further in these cases
			break
		} else if p.NoPager {
			// No pager wanted, continue the loop with next page
			pageNb++
			// If we are in verbose mode, don't forget to add a new line
			if pagingTools.Verbose {
				p.CustomPrintLn()
			}
			// Continue to get next page
			continue
		}

		// Adapt the footer message if it is the first page, the last page or in between
		footerMsg := ""
		if pageNb == 1 {
			footerMsg = fmt.Sprintf("Page: %d/%d - Navigation: (N)ext/(Q)uit\n", pageNb, totalPages)
		} else if isLastPage {
			footerMsg = fmt.Sprintf("Page: %d/%d - Navigation: (P)revious/(Q)uit\n", pageNb, totalPages)
		} else {
			footerMsg = fmt.Sprintf("Page: %d/%d - Navigation: (N)ext/(P)revious/(Q)uit\n", pageNb, totalPages)
		}

		// Print the footer message and wait for the user input as long as the input is not good
		nbThisPage := 0
		for {
			if pageNb == 1 || isLastPage {
				nbThisPage++
			}

			// only print footer one time (IOW don't re-print it if no page change)
			if nbThisPage < 2 {
				p.CustomColorPrint(ColorPurple, footerMsg)
				p.ForceFlush()
			}

			nextMove, err := getInput()
			if err != nil {
				return err
			}
			nextMoveLower := strings.ToLower(nextMove)
			// Convert space to next page
			if nextMoveLower == " " {
				nextMoveLower = "n"
			}
			// Convert "previous page" to previous page
			if nextMoveLower == previousPageStr {
				nextMoveLower = "p"
			}
			// Convert "next page" to next page
			if nextMoveLower == nextPageStr {
				nextMoveLower = "n"
			}

			needToBreak := true
			if (nextMoveLower == "n") && !isLastPage {
				pageNb++
			} else if (nextMoveLower == "p") && (pageNb > 1) {
				pageNb--
			} else if nextMoveLower == "q" {
				waitForQ = false
			} else {
				bytesTab := []byte(nextMove)
				if len(bytesTab) > 0 && (bytesTab[0] == rawSigQuit || bytesTab[0] == rawSigInt) {
					waitForQ = false
				} else {
					needToBreak = false
				}
			}
			if needToBreak {
				break
			}
		}
		firstLoop = false
	}
	return nil
}

// -------------------------------------------------------------------------------
// -------------------------------------------------------------------------------

// -------------------------------------------------------------------------------
// Utilities functions -----------------------------------------------------------
// -------------------------------------------------------------------------------

// getInput will read raw input from the terminal
func getInput() (string, error) {

	// Switch stdin into 'raw' mode
	oldState, err := term.MakeRaw(int(os.Stdin.Fd()))
	if err != nil {
		return "", err
	}

	// If we quit the function, restore terminal
	defer term.Restore(int(os.Stdin.Fd()), oldState)

	// Read the first input
	b := make([]byte, 1)
	_, err = os.Stdin.Read(b)
	if err != nil {
		return "", err
	}

	// Detection of "next page" and "previous page"
	if b[0] == 27 {
		// Save the first byte
		keyBytesArray := make([]byte, 4)
		keyBytesArray[0] = b[0]

		// Now, try to read 3 bytes as long as they match "previous page" or "next page"
		tmpByte := make([]byte, 1)
		for i := 1; i < 4; i++ {
			// Read the next byte
			_, err = os.Stdin.Read(tmpByte)
			if err != nil {
				return "", err
			}

			// Check that this match either "previous page" or "next page"
			if (tmpByte[0] != PrevPageKeyBytes[i]) && (tmpByte[0] != NextPageKeyBytes[i]) {
				break
			} else {
				keyBytesArray[i] = tmpByte[0]
			}
		}

		// Now, compare to "previous page" or "next page" bytes array
		if bytes.Equal(keyBytesArray, PrevPageKeyBytes) {
			return previousPageStr, nil
		} else if bytes.Equal(keyBytesArray, NextPageKeyBytes) {
			return nextPageStr, nil
		}
	}

	return string(b[0]), nil
}

// -------------------------------------------------------------------------------
// -------------------------------------------------------------------------------
