/**
 * Copyright (C) 2020-2023 IoT.bzh Company
 * Contact: https://www.iot.bzh/licensing
 *
 * This file is part of the redpesk-cli project, the redpesk command line tool.
 *
 * $RP_BEGIN_LICENSE$
 * Commercial License Usage
 *  Licensees holding valid commercial IoT.bzh licenses may use this file in
 *  accordance with the commercial license agreement provided with the
 *  Software or, alternatively, in accordance with the terms contained in
 *  a written agreement between you and The IoT.bzh Company. For licensing terms
 *  and conditions see https://www.iot.bzh/terms-conditions. For further
 *  information use the contact form at https://www.iot.bzh/contact.
 *
 * GNU General Public License Usage
 *  Alternatively, this file may be used under the terms of the GNU General
 *  Public license version 3. This license is as published by the Free Software
 *  Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
 *  of this file. Please review the following information to ensure the GNU
 *  General Public License requirements will be met
 *  https://www.gnu.org/licenses/gpl-3.0.html.
 * $RP_END_LICENSE$
 */

package rpconnect

import (
	"fmt"

	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/rpxxxapi"
	"git.ovh.iot/redpesk/redpesk-infra/rp-golang-common/rphttp"
	"git.ovh.iot/redpesk/redpesk-infra/rp-golang-common/rpwss"
	gosocketio "github.com/graarh/golang-socketio"
	"github.com/sirupsen/logrus"
)

// IOSockClient .
type IOSockClient struct {
	URL               string
	WebSocket         rpwss.WSManager
	ClientHTTP        *rphttp.HTTPClient
	Connected         bool
	APIInterface      rpxxxapi.RPCLIAPIInterface
	DisconnectionChan chan bool
}

// NewIoSocketClient Create a new IOSockClient
func NewIoSocketClient(url string, ClientHTTP *rphttp.HTTPClient, api rpxxxapi.RPCLIAPIInterface) (*IOSockClient, error) {

	var err error

	sCli := &IOSockClient{
		URL:               url,
		ClientHTTP:        ClientHTTP,
		Connected:         false,
		APIInterface:      api,
		DisconnectionChan: make(chan bool, 1),
	}

	// Get API logger
	logger := api.GetLogger()

	// Creation of the WSCookieDesc object necessary for WSManager
	cookieDesc := &rpwss.WSCookieDesc{
		CookieName:       api.GetCookieKey(),
		GetCookieValueCB: sCli.GetCookie,
	}
	// Creation of the WSCallBacks object necessary for WSManager
	callbacksObj := &rpwss.WSCallBacks{
		OnConnectCB:       sCli.WSOnConnect,
		OnDisconnectCB:    sCli.WSOnDisconnect,
		OnErrorCB:         sCli.WSOnError,
		OnReconnectCB:     sCli.WSOnReconnect,
		OnEndOfAttemptsCB: nil,
	}
	// Creation of the WSReconnectParams object necessary for WSManager
	reConnParams := &rpwss.WSReconnectParams{
		ReconnectTriesNb:     -1,
		ReconnectIntervalSec: 2,
		AutoSetCallbacks:     true,
		AutoSubscribesEvt:    true,
		SubscribeToEvtFct:    sCli.RegisterToEvent,
	}
	// Creation of the WSManager object
	sCli.WebSocket, err = rpwss.InitWSManager(url, 25, logger, cookieDesc, callbacksObj, reConnParams)
	if err != nil {
		return nil, err
	}

	return sCli, nil
}

// OnEvent record a callback on a specific message
func (c *IOSockClient) OnEvent(method string, f interface{}) (err error) {
	return c.WebSocket.RecordEventCallBack(method, f)
}

// EmitEvent send a message
func (c *IOSockClient) EmitEvent(message string, args interface{}) (err error) {
	return c.WebSocket.SendEvent(message, args)
}

// GetCookie returns a valid value for a Cookie
func (c *IOSockClient) GetCookie() (string, error) {

	_, err := c.APIInterface.GetVersion()
	if err != nil {
		return "", err
	}

	return c.ClientHTTP.GetClientID(), nil
}

// debugPrettyPrint useful method to log debug info
func (c *IOSockClient) debugPrettyPrint(fm string, data interface{}) {

	// Get API logger
	logger := c.APIInterface.GetLogger()

	if logger.Level == logrus.TraceLevel {
		logger.Tracef("%v%+v", fm, data)
	} else if logger.Level == logrus.DebugLevel {
		// limit to 500 characters
		sData := fmt.Sprintf("%+v", data)
		ln := len(sData)
		if ln > 500 {
			logger.Debugf("%v%.500q <TRUNCATED 500/%d>...", fm, sData, ln)
		} else {
			logger.Debugf("%v%s", fm, sData)
		}
	}
}

// WSOnDisconnect declaration
func (c *IOSockClient) WSOnDisconnect(gc *gosocketio.Channel) {
	c.APIInterface.GetLogger().Infof(wsOnDisconnectEventMsg)
	c.DisconnectionChan <- true
}

// WSOnConnect declaration
func (c *IOSockClient) WSOnConnect(gc *gosocketio.Channel) {
	c.APIInterface.GetLogger().Infof(wsOnConnectEventMsg)
}

// WSOnError declaration
func (c *IOSockClient) WSOnError(gc *gosocketio.Channel, err error) {
	c.APIInterface.GetLogger().Errorf(wsOnErrorEventMsg, err)
}

// WSOnReconnect declaration
func (c *IOSockClient) WSOnReconnect(oldsid string, gc *gosocketio.Channel) {
	c.APIInterface.GetLogger().Debugf(wsOnReconnectEventMsg)
}

// RegisterToEvent registers a client to an event
func (c *IOSockClient) RegisterToEvent(evt string) error {
	// Display the event
	c.debugPrettyPrint(wsEvtAdded, evt)
	// Register to it
	return c.APIInterface.RegisterToEvent(evt)
}
