/**
 * Copyright (C) 2020-2023 IoT.bzh Company
 * Contact: https://www.iot.bzh/licensing
 *
 * This file is part of the redpesk-cli project, the redpesk command line tool.
 *
 * $RP_BEGIN_LICENSE$
 * Commercial License Usage
 *  Licensees holding valid commercial IoT.bzh licenses may use this file in
 *  accordance with the commercial license agreement provided with the
 *  Software or, alternatively, in accordance with the terms contained in
 *  a written agreement between you and The IoT.bzh Company. For licensing terms
 *  and conditions see https://www.iot.bzh/terms-conditions. For further
 *  information use the contact form at https://www.iot.bzh/contact.
 *
 * GNU General Public License Usage
 *  Alternatively, this file may be used under the terms of the GNU General
 *  Public license version 3. This license is as published by the Free Software
 *  Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
 *  of this file. Please review the following information to ensure the GNU
 *  General Public License requirements will be met
 *  https://www.gnu.org/licenses/gpl-3.0.html.
 * $RP_END_LICENSE$
 */

package rpconnect

import (
	"errors"
	"os"
	"regexp"
	"strings"

	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/rpxxxapi"
	"git.ovh.iot/redpesk/redpesk-infra/rp-golang-common/rphttp"
	"github.com/sirupsen/logrus"
)

// -------------------------------------------------
// Type declaration --------------------------------
// -------------------------------------------------

// RPCLIConnection definition
type RPCLIConnection struct {
	Alias               string `json:"alias"`
	ServerURL           string `json:"serverurl"`
	AccessToken         string `json:"token"`
	IsAdmin             bool   `json:"is_admin"`
	LocalBuilderURL     string `json:"localbuilder_url"`
	PkgsDir             string `json:"pkgsdir"`
	SourcesDir          string `json:"sourcesdir"`
	LocalBuilderRootDir string `json:"localbuilder_root_dir"`
}

// RPCLIExtConnection definition
type RPCLIExtConnection struct {
	API     rpxxxapi.RPCLIAPIInterface
	WSocket *IOSockClient
}

// -------------------------------------------------
// -------------------------------------------------

// RPConnInit creates the new connection used by rp-cli
// It Initializes HTTP and WebSocket connection to the specified server
//
// :param connectParams: Connection parameters to use to connect to the remote server
// :type connectParams: *RPCLIConnection
// :param configDir: Configuration directory used by rp-cli
// :type configDir: string
// :param api: API interface to communicate with Redpesk backend
// :type api: *rpbckapi.RPAPIInterface
// :param openWS: Boolean allowing to request the opening of the WS
// :type openWS: bool
// :param eventsToRegBeforeWsOpen: list of events that need to be registered before the establishment of websocket
// :type eventsToRegBeforeWsOpen: []string
// :param eventsCB: call-back to call on reception of previous events
// :type eventsCB: interface{}
// :returns: error if something went wrong
func RPConnInit(connectParams *RPCLIConnection, configDir string, api rpxxxapi.RPCLIAPIInterface, openWS bool, eventsToRegBeforeWsOpen []string, eventsCB interface{}) (*RPCLIExtConnection, error) {
	var err error
	var fullURL string

	// Create the resulting connection
	resultConnection := RPCLIExtConnection{
		API: api,
	}

	// Set the full url
	fullURL = connectParams.ServerURL

	// Allow to only set port number
	if match, _ := regexp.MatchString("^([0-9]+)$", connectParams.ServerURL); match {
		fullURL = "http://localhost:" + connectParams.ServerURL
	}

	// Add http prefix if missing
	if !strings.HasPrefix(connectParams.ServerURL, "http://") && !strings.HasPrefix(connectParams.ServerURL, "https://") {
		fullURL = "http://" + connectParams.ServerURL
	}

	// Get the logger from the API interface
	logger := api.GetLogger()

	lvl := rphttp.HTTPLogLevelWarning
	if logger.Level == logrus.DebugLevel {
		lvl = rphttp.HTTPLogLevelDebug
	}

	// Create HTTP client
	logger.Debugln(httpClientConnectDebugMsg, fullURL)
	conf := rphttp.HTTPClientConfig{
		URLPrefix:           api.GetURLPrefix(),
		CookieURL:           fullURL,
		CookiesFilePath:     (configDir + cookiesDirRelativePath + connectParams.Alias + "_" + cookieFileName),
		HeaderClientKeyName: api.GetCookieKey(),
		CsrfDisable:         true,
		BasicAuthEnable:     false,
		ManualCookieMngt:    false,
		LogOut:              logger.Out,
		LogPrefix:           api.GetLogPrefix(),
		LogLevel:            lvl,
		RedirectAsError:     true,
	}

	// End to fill the HTTP client configuration
	api.FillAuthorizationFields(connectParams.AccessToken, &conf)

	HTTPCli, err := rphttp.HTTPNewClient(fullURL, conf)
	if err != nil {
		errmsg := err.Error()
		m, err := regexp.MatchString("Get http.?://", errmsg)
		if (m && err == nil) || strings.Contains(errmsg, "Failed to get device ID") {
			i := strings.LastIndex(errmsg, ":")
			newErr := connectPbError + fullURL
			if i > 0 {
				newErr += " (" + strings.TrimSpace(errmsg[i+1:]) + ")"
			} else {
				newErr += " (" + strings.TrimSpace(errmsg) + ")"
			}
			errmsg = newErr
		}
		return nil, errors.New(errmsg)
	}
	HTTPCli.SetLogLevel(logger.Level.String())

	// Save HTTPCli in APIInterface
	resultConnection.API.SetHTTPClient(HTTPCli)

	// Create io Websocket client
	logger.Debugln(wsConnectMsg, fullURL)

	resultConnection.WSocket, err = NewIoSocketClient(fullURL, HTTPCli, api)
	if err != nil {
		return nil, err
	}

	if openWS {
		// Call the API if there are additional things to do before to authenticate
		// ourself on the server
		err = resultConnection.API.InitiateAuthentication()
		if err != nil {
			return nil, err
		}
		// Add the custom header (authorization token)
		resultConnection.WSocket.WebSocket.AddAdditionalHeaders(map[string]string{"Authorization": ("Bearer " + connectParams.AccessToken)})

		// Connect the Websocket to the server
		err = resultConnection.WSocket.WebSocket.ConnectWebSocket(HTTPCli.GetClientID())

		if err == nil {
			// Register to some mandatory events needed AS SOON AS POSSIBLE after connecting the websocket
			// Basically this will be "error" event or "session expired" events
			for _, event := range eventsToRegBeforeWsOpen {
				err = resultConnection.RPEventRegister(event)
				if err != nil {
					return &resultConnection, err
				}
				// Save the callback
				err = resultConnection.WSocket.OnEvent(event, eventsCB)
				if err != nil {
					return &resultConnection, err
				}
			}
		}
	}

	return &resultConnection, err
}

// RPConnClose terminates connection to Redpesk agent
func (rpConnect *RPCLIExtConnection) RPConnClose() {
	// Get logger
	logger := rpConnect.API.GetLogger()

	// Log that we close the connection
	logger.Debugf(httpClientDisconnectMsg)

	// Update the cookies saved in the file if the http client has been initialized
	httpClient := rpConnect.API.GetHTTPClient()
	if httpClient != nil {
		err := httpClient.UpdateCookieFile()
		if err != nil {
			logger.Errorln(err.Error())
		}
	}

	logger.Debugf(wsClientDisconnectMsg)
	// Check that the websocket client exists
	if rpConnect.WSocket != nil && rpConnect.WSocket.WebSocket.WSClient != nil {
		// Close the websocket if needed
		if rpConnect.WSocket.WebSocket.WSClient.Connected {
			rpConnect.WSocket.WebSocket.CloseWebSocket()
		}
	}
}

// RPEventRegister registers the rp-cli client to a given event
//
// :param evtRef: String describing the event to subscribe to
// :type evtRef: string
func (rpConnect *RPCLIExtConnection) RPEventRegister(evtRef string) error {
	// Ask the registration on the websocket
	return rpConnect.WSocket.RegisterToEvent(evtRef)
}

// RPDelCookieFile deletes the current cookie file used by
// the HTTP client to load and save its cookie
//
// :param configDir: Configuration directory used by rp-cli
// :type configDir: string
// :param alias: Connection alias on which we want to delete the cookie
// :type alias: string
// :returns: error if something went wrong
func RPDelCookieFile(configDir string, alias string) error {

	// Construct the full path to the cookie file
	cookieFilePath := (configDir + cookiesDirRelativePath + alias + "_" + cookieFileName)

	// Check if the file exist and delete it if it exists
	_, err := os.Stat(cookieFilePath)

	// Check if the error is different than file not existing
	if err != nil && !os.IsNotExist(err) {
		return err
	}

	// Check that the file exists
	if err == nil {
		// Delete the file
		if err = os.Remove(cookieFilePath); err != nil {
			return err
		}
	}

	return nil
}
