/**
 * Copyright (C) 2020-2023 IoT.bzh Company
 * Contact: https://www.iot.bzh/licensing
 *
 * This file is part of the redpesk-cli project, the redpesk command line tool.
 *
 * $RP_BEGIN_LICENSE$
 * Commercial License Usage
 *  Licensees holding valid commercial IoT.bzh licenses may use this file in
 *  accordance with the commercial license agreement provided with the
 *  Software or, alternatively, in accordance with the terms contained in
 *  a written agreement between you and The IoT.bzh Company. For licensing terms
 *  and conditions see https://www.iot.bzh/terms-conditions. For further
 *  information use the contact form at https://www.iot.bzh/contact.
 *
 * GNU General Public License Usage
 *  Alternatively, this file may be used under the terms of the GNU General
 *  Public license version 3. This license is as published by the Free Software
 *  Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
 *  of this file. Please review the following information to ensure the GNU
 *  General Public License requirements will be met
 *  https://www.gnu.org/licenses/gpl-3.0.html.
 * $RP_END_LICENSE$
 */

package rpladmock

import (
	"bufio"
	"context"
	"encoding/json"
	"fmt"
	"net/http"
	"os"
	"path"
	"strconv"
	"strings"
	"time"

	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/rpxxxapi/ladapi"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/utils"
	"github.com/gorilla/mux"
	gosocketio "github.com/graarh/golang-socketio"
	"github.com/graarh/golang-socketio/transport"
	"github.com/pkg/errors"
	"github.com/sirupsen/logrus"
)

// -------------------------------------------------------------------------------
// Types definition --------------------------------------------------------------
// -------------------------------------------------------------------------------

// RPMockedLAD definition
type RPMockedLAD struct {
	http.Server
	WSServer        *gosocketio.Server
	WSServerChannel *gosocketio.Channel
	log             *logrus.Logger
}

// Type used for local builder events
type RPLADEvent struct {
	ID     string `json:"id"`
	Line   string `json:"std_line"`
	Status int    `json:"status"`
	Type   string `json:"type"`
}

// Type used to send stdin event to local builder
type RPLADStdinEvent struct {
	ID    string `json:"id"`
	Chars string `json:"stdin_chars"`
}

// -------------------------------------------------------------------------------
// -------------------------------------------------------------------------------

// -------------------------------------------------------------------------------
// Global variables --------------------------------------------------------------
// -------------------------------------------------------------------------------

// Current token used to emit event to the client
// (this is only client by client during tests)
var currentClientTicket = ""

// Current stdin received through events
var currentStdIn = ""

// Latest image size requested
var LatestImageSize = -1

// Latest exec command run
var LatestExecCmd = ""

// Global variable allowing to detect if the latest build was offline
var LatestBuildOffline = false

// -------------------------------------------------------------------------------
// -------------------------------------------------------------------------------

// -------------------------------------------------------------------------------
// Public functions definition ---------------------------------------------------
// -------------------------------------------------------------------------------

// StartMockedRPLad allows the test program to mock a Redpesk LAD
// server
//
// :param logR: Pointer to the logger used by the application
// :type logR: *logrus.Logger
// :returns: *RPMockedLAD, error -- pointer to the mocked backend and an error if a problem has occurred
func StartMockedRPLad(logR *logrus.Logger) (*RPMockedLAD, error) {
	// Initialize the mocked backend structure
	mockedLAD := &RPMockedLAD{
		Server: http.Server{
			Addr:         serverAddress,
			ReadTimeout:  5 * time.Second,
			WriteTimeout: 5 * time.Second,
		},
		log: logR,
	}

	// ------------------------------------------
	// Define the Websocket server --------------
	// ------------------------------------------
	// Get the default WebSocket Transport
	webSocketTransport := transport.GetDefaultWebsocketTransport()
	// Create the WebSocket Server
	mockedLAD.WSServer = gosocketio.NewServer(webSocketTransport)
	// Save the OnConnection callback
	mockedLAD.WSServer.On(gosocketio.OnConnection, func(c *gosocketio.Channel) {
		mockedLAD.log.Debugf(wsConnectionMsg)
		mockedLAD.WSServerChannel = c
	})
	// Save the OnDisconnection callback
	mockedLAD.WSServer.On(gosocketio.OnDisconnection, func(c *gosocketio.Channel) {
		mockedLAD.log.Debugf(wsDisconnectionMsg)
	})
	// Save the OnError callback
	mockedLAD.WSServer.On(gosocketio.OnError, func(c *gosocketio.Channel, err error) {
		mockedLAD.log.Errorf(wsErrorErr, err)
	})
	// Save the stdin event callback
	mockedLAD.WSServer.On("stdin", StdinEventCB)
	// ------------------------------------------
	// ------------------------------------------

	// ------------------------------------------
	// Define the HTTP server -------------------
	// ------------------------------------------
	// Initialize the routes that can be accessed on the server
	router := mux.NewRouter()

	// Websocket handlers
	router.Handle(wSocketRoute, mockedLAD.WSServer)
	// Version
	router.HandleFunc(versionRoute, mockedLAD.versionHandler).Methods("GET")
	// Events
	router.HandleFunc(eventSubscribeRoute, mockedLAD.eventRegisterHandler).Methods("POST")
	// exec handler
	router.HandleFunc(execRoute, mockedLAD.execHandler).Methods("POST")
	// iotpkg rpbuild handler
	router.HandleFunc(iotpkgRPBuildRoute, mockedLAD.iotpkgRPBuildHandler).Methods("POST")
	// iotpkg metadata handler
	router.HandleFunc(iotpkgMetadataRoute, mockedLAD.iotpkgMetadataHandler).Methods("POST")
	// iotpkg service handler
	router.HandleFunc(iotpkgServiceRoute, mockedLAD.iotpkgServiceHandler).Methods("POST")
	// iotpkg init handler
	router.HandleFunc(iotpkgInitRoute, mockedLAD.iotpkgInitHandler).Methods("POST")
	// iotpkg cleanmock handler
	router.HandleFunc(iotpkgCleanMockRoute, mockedLAD.iotpkgCleanMockHandler).Methods("POST")

	//set http server handler
	mockedLAD.Handler = router

	// Start the server
	go func() {
		err := mockedLAD.ListenAndServe()
		if (err != nil) && !strings.Contains(err.Error(), "Server closed") {
			mockedLAD.log.Errorf(exitServerRunErr, err)
			fmt.Print("A problem occurred in server !")
			errorMsg := fmt.Sprintf(exitServerRunErr, err)
			utils.WriteToFile("/tmp/mocked-lad.crash", errorMsg)
		}
	}()
	// ------------------------------------------
	// ------------------------------------------

	// Wait a bit for the server to be running
	time.Sleep(100 * time.Millisecond) // Wait a bit that above ListenAndServe function has started

	return mockedLAD, nil
}

// ShutdownServer is the function allowing to shutdown correctly the
// mocked server
//
// :returns: error -- an error if a problem has occurred
func (mockedLAD *RPMockedLAD) ShutdownServer() error {

	//Create shutdown context with 10 second timeout
	ctx, cancel := context.WithTimeout(context.Background(), 5*time.Second)
	defer cancel()

	err := mockedLAD.Shutdown(ctx)
	if err != nil {
		mockedLAD.log.Errorf(shutdownServerErr, err)
	}

	return err
}

// EmitEvent allows to send events to the rp-cli being in test
//
// :param method: Name of the event to send
// :type method: string
// :param args: Data to send with the event
// :type args: interface{}
// :returns: error -- an error if a problem has occurred
func (mockedLAD *RPMockedLAD) EmitEvent(method string, args interface{}) error {
	if mockedLAD.WSServerChannel == nil {
		return errors.Errorf("[MOCKED-LAD] nil WSServerChannel")
	}
	if !mockedLAD.WSServerChannel.IsAlive() {
		return errors.Errorf("[MOCKED-LAD] WSServerChannel is not alive")
	}

	mockedLAD.log.Debugf("[MOCKED-LAD] LAD Emit EVENT %s: %+v", method, args)
	return mockedLAD.WSServerChannel.Emit(method, args)
}

// StdinEventCB handle the reception of event from the “
// builder
//
// :param gc: Websocket channel used
// :type gc: *gosocketio.Channel
// :param event: Event message received
// :type event: RPLocalBuilderEvent
func StdinEventCB(gc *gosocketio.Channel, event RPLADStdinEvent) {
	// Record the current stdin
	currentStdIn = currentStdIn + event.Chars
}

// -------------------------------------------------------------------------------
// -------------------------------------------------------------------------------

// -------------------------------------------------------------------------------
// Handlers functions definition -------------------------------------------------
// -------------------------------------------------------------------------------

// versionHandler is the handler used when a call is done on the version route
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedLAD *RPMockedLAD) versionHandler(w http.ResponseWriter, req *http.Request) {
	version := map[string]string{
		"version":      "v1",
		"full_version": "0.2.0",
	}

	resJSON(w, version)
}

// eventRegisterHandler is the handler used when a call is done on the event/subscribe route
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedLAD) eventRegisterHandler(w http.ResponseWriter, req *http.Request) {
	answer := make(map[string]string)
	answer["status"] = "OK"
	resJSON(w, answer)
}

// iotpkgRPBuildHandler is the handler used when a call is done on the /iotpkg/rpbuild route
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedLAD *RPMockedLAD) iotpkgRPBuildHandler(w http.ResponseWriter, req *http.Request) {
	// Get the request body
	reqBody := &ladapi.RPLADIotpkgCmdStruct{}
	if err := reqBind(req, reqBody); err != nil {
		resJSONErrorE(w, err)
		return
	}

	// Check that a service name has been set
	if len(reqBody.SubcmdFlags) <= 0 {
		resJSONErrorE(w, errors.New("No flag has been given!"))
	}

	// Check that there is no "--project" flag here
	if utils.ContainsString(reqBody.SubcmdFlags, "--project") {
		resJSONErrorE(w, errors.New("Problem with flags given to iotpkg!"))
	}

	// Check that we have the "distro" and "arch" flag
	if !utils.ContainsString(reqBody.SubcmdFlags, "--distro") {
		resJSONErrorE(w, errors.New("No 'distro' flag given!"))
	}
	if !utils.ContainsString(reqBody.SubcmdFlags, "--arch") {
		resJSONErrorE(w, errors.New("No 'arch' flag given!"))
	}

	// Set the 'offline' global variable
	if utils.ContainsString(reqBody.SubcmdFlags, "--offline") {
		LatestBuildOffline = true
	} else {
		LatestBuildOffline = false
	}

	// Get the working dir specified by rp-cli
	workingDir := strings.ReplaceAll(reqBody.WorkingDir, "/home/devel/gitpkgs", gitPackagesTestDir)

	// Create the ".iotpkg" directory
	err := utils.CreateDirIfNotExists(path.Join(workingDir, ".iotpkg"), false)
	if err != nil {
		resJSONErrorE(w, err)
		return
	}

	// Create a completely fake cfg file
	appName := path.Base(workingDir)
	cfgFilename := appName + "-redpesk-distro-aarch64-cross.cfg"
	cfgFilePath := path.Join(workingDir, ".iotpkg", cfgFilename)
	cfgFile, err := os.Create(cfgFilePath)
	if err != nil {
		resJSONErrorE(w, err)
		return
	}
	defer cfgFile.Close()
	_, err = cfgFile.WriteString(iotpkgFileContent)
	if err != nil {
		resJSONErrorE(w, err)
		return
	}

	// Copy this file at the root of gitpkgs/
	err = utils.CopyFile(cfgFilePath, path.Join(gitPackagesTestDir, cfgFilename))
	if err != nil {
		resJSONErrorE(w, err)
		return
	}

	// Generate the current UID for the client
	currentClientTicket = utils.GenerateSecureToken(16)

	if utils.ContainsString(reqBody.SubcmdFlags, "--shell") {
		// Wait a bit and send the end event
		go mockedLAD.playFileAndEndEvent(currentClientTicket, "rpladmock/logs/iotpkg-rpbuild-shell.log", 10, 1)
	} else {
		// Wait a bit and send the end event
		go mockedLAD.playFileAndEndEvent(currentClientTicket, "rpladmock/logs/iotpkg-rpbuild.log", 10, 10)
	}

	answer := map[string]string{
		"id": currentClientTicket,
	}

	resJSON(w, answer)
}

// iotpkgMetadataHandler is the handler used when a call is done on the /iotpkg/metadata route
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedLAD *RPMockedLAD) iotpkgMetadataHandler(w http.ResponseWriter, req *http.Request) {
	// Get the request body
	reqBody := &ladapi.RPLADIotpkgCmdStruct{}
	if err := reqBind(req, reqBody); err != nil {
		resJSONErrorE(w, err)
		return
	}

	// Check that a service name has been set
	if len(reqBody.SubcmdFlags) <= 0 {
		resJSONErrorE(w, errors.New("No flag has been given!"))
	}

	// Check that there is no "--project" flag here
	if utils.ContainsString(reqBody.SubcmdFlags, "--project") {
		resJSONErrorE(w, errors.New("Problem with flags given to iotpkg!"))
	}

	// Generate the current UID for the client
	currentClientTicket = utils.GenerateSecureToken(16)

	// Wait a bit and send the end event
	go mockedLAD.waitABitAndSendEndEvent(currentClientTicket, 100)

	answer := map[string]string{
		"id": currentClientTicket,
	}

	resJSON(w, answer)
}

// iotpkgServiceHandler is the handler used when a call is done on the /iotpkg/service route
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedLAD *RPMockedLAD) iotpkgServiceHandler(w http.ResponseWriter, req *http.Request) {
	// Get the request body
	reqBody := &ladapi.RPLADIotpkgCmdStruct{}
	if err := reqBind(req, reqBody); err != nil {
		resJSONErrorE(w, err)
		return
	}

	// Check that a service name has been set
	if len(reqBody.SubcmdFlags) <= 0 {
		resJSONErrorE(w, errors.New("No service name has been given!"))
	}

	// Check that there is no "--project" flag here
	if utils.ContainsString(reqBody.SubcmdFlags, "--project") {
		resJSONErrorE(w, errors.New("Problem with flags given to iotpkg!"))
	}

	// Generate the current UID for the client
	currentClientTicket = utils.GenerateSecureToken(16)

	// If the "--help" flag is set, return the help prompt
	if utils.ContainsString(reqBody.SubcmdFlags, "--help") {
		go mockedLAD.playFileAndEndEvent(currentClientTicket, "rpladmock/logs/iotpkg-service-help.log")
	} else {
		// Wait a bit and send the end event
		go mockedLAD.waitABitAndSendEndEvent(currentClientTicket, 100)
	}

	answer := map[string]string{
		"id": currentClientTicket,
	}

	resJSON(w, answer)
}

// iotpkgInitHandler is the handler used when a call is done on the /iotpkg/init route
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedLAD *RPMockedLAD) iotpkgInitHandler(w http.ResponseWriter, req *http.Request) {
	// Get the request body
	reqBody := &ladapi.RPLADIotpkgCmdStruct{}
	if err := reqBind(req, reqBody); err != nil {
		resJSONErrorE(w, err)
		return
	}

	// Check that a service name has been set
	if len(reqBody.SubcmdFlags) <= 0 {
		resJSONErrorE(w, errors.New("No flag has been given!"))
	}

	// Check that there is no "--project" flag here
	if utils.ContainsString(reqBody.SubcmdFlags, "--project") {
		resJSONErrorE(w, errors.New("Problem with flags given to iotpkg!"))
	}

	// Generate the current UID for the client
	currentClientTicket = utils.GenerateSecureToken(16)

	// Wait a bit and send the end event
	go mockedLAD.playFileAndEndEvent(currentClientTicket, "rpladmock/logs/iotpkg-init.log")

	answer := map[string]string{
		"id": currentClientTicket,
	}

	resJSON(w, answer)
}

// iotpkgCleanMockHandler is the handler used when a call is done on the /iotpkg/cleanmock route
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedLAD *RPMockedLAD) iotpkgCleanMockHandler(w http.ResponseWriter, req *http.Request) {
	// Get the request body
	reqBody := &ladapi.RPLADIotpkgCmdStruct{}
	if err := reqBind(req, reqBody); err != nil {
		resJSONErrorE(w, err)
		return
	}

	// Check that a service name has been set
	if len(reqBody.SubcmdFlags) <= 0 {
		resJSONErrorE(w, errors.New("No flag has been given!"))
	}

	// Generate the current UID for the client
	currentClientTicket = utils.GenerateSecureToken(16)

	// Wait a bit and send the end event
	go mockedLAD.playFileAndEndEvent(currentClientTicket, "rpladmock/logs/iotpkg-cleanmock.log")

	answer := map[string]string{
		"id": currentClientTicket,
	}

	resJSON(w, answer)
}

// execHandler is the handler used when a call is done on the /exec route
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedLAD *RPMockedLAD) execHandler(w http.ResponseWriter, req *http.Request) {
	// Get the request body
	reqBody := &ladapi.RPLADExecCmdStruct{}
	if err := reqBind(req, reqBody); err != nil {
		resJSONErrorE(w, err)
		return
	}

	// Empty the stdin string
	currentStdIn = ""

	// Generate the current UID for the client
	currentClientTicket = utils.GenerateSecureToken(16)

	// Get rid of 'iotpkg' if in the command line (for test)
	cmdToRun := strings.Replace(reqBody.CmdLine, "iotpkg", "", 1)
	cmdToRun = strings.TrimSpace(cmdToRun)

	// Convert the working dir to something existing
	workingDir := reqBody.WorkingDir
	workingDir = strings.ReplaceAll(workingDir, "/home/devel/gitsources", gitSourcesTestDir)
	workingDir = strings.ReplaceAll(workingDir, "/home/devel/gitpkgs", gitPackagesTestDir)

	// Save the latest command run
	LatestExecCmd = cmdToRun

	// Test if the command is a "localimagebuild" one
	if strings.Contains(cmdToRun, "localimagebuild") {
		// Save the image size requested by the user
		LatestImageSize, _ = extractImageSizeRequested(cmdToRun)

		// Start the goroutine mocking the "iotpkg localimagebuild" command
		go mockedLAD.playFileAndEndEvent(currentClientTicket, "rpladmock/logs/iotpkg-localimagebuild.log")

	} else if strings.Contains(cmdToRun, "rm -f /home/devel/.oz/images/") {
		// In this case, this is the end of an image build, modify the command to print it somewhere
		newCmdToRun := "sleep 0.1"
		// Start the goroutine mocking the "exec" command
		go mockedLAD.playExec(currentClientTicket, workingDir, newCmdToRun)
	} else if strings.Contains(cmdToRun, "ls -al") {
		go mockedLAD.playFileAndEndEvent(currentClientTicket, "rpladmock/logs/temp-file-ls.log")
	} else {
		// Start the goroutine mocking the "exec" command
		go mockedLAD.playExec(currentClientTicket, workingDir, cmdToRun)
	}

	answer := map[string]string{
		"id": currentClientTicket,
	}

	resJSON(w, answer)
}

// -------------------------------------------------------------------------------
// -------------------------------------------------------------------------------

// -------------------------------------------------------------------------------
// Utilities functions definition ------------------------------------------------
// -------------------------------------------------------------------------------

// playFileAndEndEvent is a utility function that send the lines available in the file
// through the websocket to rp-cli in tests situation. After 'playing' the file, it
// sends a End event
//
// :param clientTicket: Current client ticket
// :type clientTicket: string
// :param localPathToFile: Local path to the file to play
// :type localPathToFile: string
// :param args: Through this arg the caller can specified the time between line sent in milliseconds and the nb of lines per blob
// :type args: int
func (mockedLAD *RPMockedLAD) playFileAndEndEvent(clientTicket, localPathToFile string, args ...int) {
	timeBetweenLineMS := 20
	nbOfLines := 2

	if len(args) == 2 {
		timeBetweenLineMS = args[0]
		nbOfLines = args[1]
	}
	// Send the logs for the archive line by line
	imgBuildLogFile := utils.GetPkgRootDir() + localPathToFile
	mockedLAD.sendFilePerLines(clientTicket, imgBuildLogFile, timeBetweenLineMS, nbOfLines)

	// Wait a bit and send the end event
	mockedLAD.waitABitAndSendEndEvent(clientTicket, 100)
}

// waitABitAndSendEndEvent does exactly what it is named after
//
// :param clientTicket: Current client ticket
// :type clientTicket: string
// :param waitTimeMs: Waiting time before sending end event
// :type waitTimeMs: int
func (mockedLAD *RPMockedLAD) waitABitAndSendEndEvent(clientTicket string, waitTimeMs int) {
	// Wait a bit for the end
	time.Sleep(time.Duration(waitTimeMs) * time.Millisecond)

	// Send the "process_ended" event
	evtData := &RPLADEvent{
		ID:     clientTicket,
		Type:   "process_ended",
		Status: 0,
	}
	mockedLAD.EmitEvent("process_ended", evtData)
}

// playExec mock the behavior of the Thread running
// the "exec" command
//
// :param clientTicket: Current client ticket
// :type clientTicket: string
// :param workDir: Working directory where to run the command
// :type workDir: string
// :param cmdLine: Command line to run
// :type cmdLine: string
func (mockedLAD *RPMockedLAD) playExec(clientTicket, workDir, cmdLine string) {

	// FOR THE MOMENT, THIS MOCK DO NOT SUPPORT INTERACTIVE COMMAND
	// LIKE VIM, LESS, ETC.

	// Divide the command line to get the first argument
	cmdAsTab := strings.Split(cmdLine, " ")

	// Run the command passed by the user
	cmdOutput, err := utils.RunCmdInDirectory(workDir, cmdAsTab[0], cmdAsTab[1:]...)
	if err != nil {
		// Send the "process_ended" event
		evtData := &RPLADEvent{
			ID:     clientTicket,
			Type:   "process_ended",
			Status: 1,
		}
		mockedLAD.EmitEvent("process_ended", evtData)
		return
	}

	// Send the logs from the output line by line
	linesOutput := strings.Split(cmdOutput, "\n")
	for _, line := range linesOutput {
		// Sleep the wanted time
		time.Sleep(time.Duration(5) * time.Millisecond)

		// Send the line(s) through event
		evtData := &map[string]string{
			"id":       clientTicket,
			"type":     "stdout",
			"std_line": (line + "\n"),
		}
		mockedLAD.EmitEvent("stdout", evtData)
	}

	// Create a file in working directory containing the stdin
	if currentStdIn != "" {
		utils.WriteToFile(workDir+"/test_file_"+clientTicket+".log", currentStdIn)
	}

	// Wait a bit for the end
	time.Sleep(100 * time.Millisecond)

	// Send the "process_ended" event
	evtData := &RPLADEvent{
		ID:     clientTicket,
		Type:   "process_ended",
		Status: 0,
	}
	mockedLAD.EmitEvent("process_ended", evtData)
}

// sendFilePerLines sends the content of a file through events to the current client
// (identified with clientTicket)
//
// :param clientTicket: Current client ticket
// :type clientTicket: string
// :param filePath: Path to the file to send
// :type filePath: string
// :param timeBtwLine: Time, in milliseconds between sending each blob of lines
// :type timeBtwLine: int
// :param nbOfLines: Number of lines per blob
// :type nbOfLines: int
func (mockedLAD *RPMockedLAD) sendFilePerLines(clientTicket, filePath string, timeBtwLine, nbOfLines int) {
	lineBuffer := ""
	lineCounter := 0
	tmpLine := ""

	// Open the file
	fileToRead, err := os.Open(filePath)
	if err != nil {
		fmt.Printf("An error occurred while opening %s!", filePath)
		return
	}
	// Defer the file closing
	defer fileToRead.Close()

	// Create the scanner on the file
	scanner := bufio.NewScanner(fileToRead)

	// Prepare the scanner to read line by line
	scanner.Split(bufio.ScanLines)

	// Iterate over each line
	for scanner.Scan() {
		// Get the current line
		tmpLine = scanner.Text()

		// Save the current line
		lineBuffer += tmpLine + "\n"
		lineCounter += 1

		// Check if the buffer is full to send it
		if lineCounter >= nbOfLines {
			// Sleep the wanted time
			time.Sleep(time.Duration(timeBtwLine) * time.Millisecond)

			// Send the line(s) through event
			evtData := &map[string]string{
				"id":       clientTicket,
				"type":     "stdout",
				"std_line": lineBuffer,
			}
			mockedLAD.EmitEvent("stdout", evtData)

			// Reset the variables
			lineBuffer = ""
			lineCounter = 0
		}
	}

	// If there is still things in the buffer, sends it
	if lineBuffer != "" {
		// Sleep the wanted time
		time.Sleep(time.Duration(timeBtwLine) * time.Millisecond)

		evtData := &map[string]string{
			"id":       clientTicket,
			"type":     "stdout",
			"std_line": lineBuffer,
		}
		mockedLAD.EmitEvent("stdout", evtData)
	}
}

// extractImageSizeRequested gets the size specified for the image in the command line,
// if no size is specified, the function return -1
//
// :param commandLine: Command line containing (possibly) the size wanted for the image
// :type commandLine: string
// :returns: int, error -- the size specified in the command, -1 if not found and an error if a problem occurred
func extractImageSizeRequested(commandLine string) (int, error) {
	returnValue := -1
	var err error

	// Check if the command line contains the "--size" flag
	if !strings.Contains(commandLine, "--size") {
		return returnValue, nil
	}

	// Split the command into arguments to retrieve the value for "--size"
	cmdArgs := strings.Split(commandLine, " ")
	for index, arg := range cmdArgs {
		if (arg == "--size") && (index < len(cmdArgs)-1) {
			returnValue, err = strconv.Atoi(cmdArgs[index+1])
			return returnValue, err
		}
	}

	// If we reached here, it means the "--size" flag exists but no size is passed
	return -1, nil
}

func reqBind(req *http.Request, v interface{}) error {
	return json.NewDecoder(req.Body).Decode(v)
}

func reqParamGet(req *http.Request, name string) string {
	params := mux.Vars(req)
	return params[name]
}

func resJSON(w http.ResponseWriter, data interface{}) error {
	w.Header().Set("Content-Type", "application/json")
	w.WriteHeader(http.StatusOK)
	return json.NewEncoder(w).Encode(data)
}

func resJSONErrorS(w http.ResponseWriter, format string, args ...interface{}) error {
	w.Header().Set("Content-Type", "application/json")
	errMsg := fmt.Sprintf(format, args...)
	e := map[string]string{
		"status": "error",
		"error":  errMsg,
		"type":   "<class xxx>",
	}
	w.WriteHeader(http.StatusInternalServerError)
	return json.NewEncoder(w).Encode(e)
}

func resJSONErrorE(w http.ResponseWriter, err error) error {
	return resJSONErrorS(w, err.Error())
}

// -------------------------------------------------------------------------------
// -------------------------------------------------------------------------------
