/**
 * Copyright (C) 2022-2023 IoT.bzh Company
 * Contact: https://www.iot.bzh/licensing
 *
 * This file is part of the redpesk-cli project, the redpesk command line tool.
 *
 * $RP_BEGIN_LICENSE$
 * Commercial License Usage
 *  Licensees holding valid commercial IoT.bzh licenses may use this file in
 *  accordance with the commercial license agreement provided with the
 *  Software or, alternatively, in accordance with the terms contained in
 *  a written agreement between you and The IoT.bzh Company. For licensing terms
 *  and conditions see https://www.iot.bzh/terms-conditions. For further
 *  information use the contact form at https://www.iot.bzh/contact.
 *
 * GNU General Public License Usage
 *  Alternatively, this file may be used under the terms of the GNU General
 *  Public license version 3. This license is as published by the Free Software
 *  Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
 *  of this file. Please review the following information to ensure the GNU
 *  General Public License requirements will be met
 *  https://www.gnu.org/licenses/gpl-3.0.html.
 * $RP_END_LICENSE$
 */

package bckapi

import (
	"fmt"

	"git.ovh.iot/redpesk/redpesk-infra/rp-webserver/actions/rpapiv1"
)

// -------------------------------------------------
// Types definition --------------------------------
// -------------------------------------------------

// RPAuditEntry definition
type RPAuditEntry struct {
	Category string            `json:"category"`
	Tools    []RPLanguageTools `json:"tools"`
}

// RPLanguageTools definition
type RPLanguageTools struct {
	Language string               `json:"language"`
	Tools    []RPLanguageToolDesc `json:"tools"`
}

// RPLanguageToolDesc definition
type RPLanguageToolDesc struct {
	Name    string `json:"name"`
	Summary string `json:"summary"`
	Version string `json:"version"`
}

// RPAuditResults definition
type RPAuditResults struct {
	Analyse     []RPAuditResult   `json:"analyze"`
	TotalReport RPAuditCodeReport `json:"report"`
}

// RPAuditResult definition
type RPAuditResult struct {
	Tool   string            `json:"tool"`
	Report RPAuditCodeReport `json:"report"`
}

// RPAuditCodeReport definition
type RPAuditCodeReport struct {
	Critical int `json:"CRITICAL"`
	High     int `json:"HIGH"`
	Info     int `json:"INFO"`
	Low      int `json:"LOW"`
	Medium   int `json:"MEDIUM"`
	Unknown  int `json:"UNKNOWN"`
	Total    int `json:"TOTAL"`
}

// -------------------------------------------------
// -------------------------------------------------

// RPAuditGet gets a full audit structure from its ID
//
// :param auditID: audit ID
// :type auditID: string
// :param audit: Pointer to the rpapiv1.RPAuditFull structure to fill
// :type audit: *rpapiv1.RPAuditFull
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) RPAuditGet(auditID string, audit *rpapiv1.RPAuditFull) error {
	// Send the appropriate GET request
	if err := rpAPIInterface.HTTPCli.Get(fmt.Sprintf(auditsIDRoute, auditID), audit); err != nil {
		return errorInterpreter(err, auditObjectType)
	}
	return nil
}

// RPAuditGetLogs download the logs of a given audit
//
// :param auditID: audit ID
// :type auditID: string
// :param targetDir: directory where the downloaded file needs to be stored
// :type targetDir: string
// :returns: (string, error) -- path to the downloaded file and an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) RPAuditGetLogs(auditID, targetDir string) (string, error) {
	return rpAPIInterface.backendFileDownload(fmt.Sprintf(auditsLogRoute, auditID), targetDir, "audit_logs.zip")
}

// RPAuditToolsList retrieve the list of tools available to audit applications
//
// :param auditTools: Pointer to the rpapiv1.AuditStruct structure to fill
// :type auditTools: *rpapiv1.AuditStruct
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) RPAuditToolsList(auditTools *rpapiv1.AuditStruct) error {
	// Send the appropriate GET request
	if err := rpAPIInterface.HTTPCli.Get(auditsListToolsRoute, auditTools); err != nil {
		return errorInterpreter(err, "")
	}
	return nil
}
