/**
 * Copyright (C) 2022-2023 IoT.bzh Company
 * Contact: https://www.iot.bzh/licensing
 *
 * This file is part of the redpesk-cli project, the redpesk command line tool.
 *
 * $RP_BEGIN_LICENSE$
 * Commercial License Usage
 *  Licensees holding valid commercial IoT.bzh licenses may use this file in
 *  accordance with the commercial license agreement provided with the
 *  Software or, alternatively, in accordance with the terms contained in
 *  a written agreement between you and The IoT.bzh Company. For licensing terms
 *  and conditions see https://www.iot.bzh/terms-conditions. For further
 *  information use the contact form at https://www.iot.bzh/contact.
 *
 * GNU General Public License Usage
 *  Alternatively, this file may be used under the terms of the GNU General
 *  Public license version 3. This license is as published by the Free Software
 *  Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
 *  of this file. Please review the following information to ensure the GNU
 *  General Public License requirements will be met
 *  https://www.gnu.org/licenses/gpl-3.0.html.
 * $RP_END_LICENSE$
 */

package bckapi

import (
	"encoding/json"
	"strconv"

	"git.ovh.iot/redpesk/redpesk-infra/rp-webserver/actions/rpapiv1"
)

// RPBoardModelsGet allows to get the list of available boards models
// on the Redpesk instance
//
// :param boards: Pointer to []RPTestBoard slice to be filled
// :type boards: *rpapiv1.RPModelBoard
// :param filter: Filter to pass to the request done to the factory
// :type filter: string
// :returns: error -- an error if something went bad
func (rpAPIInterface *RPBCKAPI) RPBoardModelsGet(boards *[]rpapiv1.RPModelBoard, filter string) error {

	addQueryParams := make(map[string]string)
	if filter != "" {
		addQueryParams["filters"] = filter
	}

	// Initialize the call-back to save current page of model boards
	savingCB := func(tmpObjects []byte) {
		tmpModelBoards := rpapiv1.RPModelBoards{}
		// Convert to the right type
		json.Unmarshal(tmpObjects, &tmpModelBoards)
		// Add the applications in the final object
		(*boards) = append((*boards), tmpModelBoards.ModelBoards...)
	}

	// Initialize the paging structure that will allow to get all the objects
	pagingMngt := &RPBCKPaging{}

	// Call the function allowing to get all the object
	err := pagingMngt.getAllObjects(rpAPIInterface, modelBoardsRoute, "", addQueryParams, savingCB)
	if err != nil {
		return errorInterpreter(err, boardModelObjectType)
	}

	return nil
}

// RPBoardModelsListByPage gets the list of board models on the factory
// but only one page of it
//
// :param boards: Pointer to structure to be filled
// :type boards: *rpapiv1.RPModelBoards
// :param page: Page number that the caller want
// :type page: int
// :param perPage: Number of item per page that the caller want
// :type perPage: int
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) RPBoardModelsListByPage(boards *rpapiv1.RPModelBoards, page, perPage int) error {

	// Initialize the query parameters object
	queryParamMap := map[string]string{
		"page":     strconv.Itoa(page),
		"per_page": strconv.Itoa(perPage),
	}

	// Now, run the get request
	if err := rpAPIInterface.HTTPCli.Get(modelBoardsRoute, boards, queryParamMap); err != nil {
		return errorInterpreter(err, boardModelObjectType)
	}

	return nil
}
