/**
 * Copyright (C) 2022-2023 IoT.bzh Company
 * Contact: https://www.iot.bzh/licensing
 *
 * This file is part of the redpesk-cli project, the redpesk command line tool.
 *
 * $RP_BEGIN_LICENSE$
 * Commercial License Usage
 *  Licensees holding valid commercial IoT.bzh licenses may use this file in
 *  accordance with the commercial license agreement provided with the
 *  Software or, alternatively, in accordance with the terms contained in
 *  a written agreement between you and The IoT.bzh Company. For licensing terms
 *  and conditions see https://www.iot.bzh/terms-conditions. For further
 *  information use the contact form at https://www.iot.bzh/contact.
 *
 * GNU General Public License Usage
 *  Alternatively, this file may be used under the terms of the GNU General
 *  Public license version 3. This license is as published by the Free Software
 *  Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
 *  of this file. Please review the following information to ensure the GNU
 *  General Public License requirements will be met
 *  https://www.gnu.org/licenses/gpl-3.0.html.
 * $RP_END_LICENSE$
 */

package bckapi

import (
	"mime"
	"path/filepath"
	"strings"

	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/utils"
	"git.ovh.iot/redpesk/redpesk-infra/rp-golang-common/rputils"
	"git.ovh.iot/redpesk/redpesk-infra/rp-webserver/actions/rpapiv1"
	"github.com/sirupsen/logrus"
)

// backendFileDownload is a function allowing to download a multipart file from the backend.
// Indeed an important thing to have in order to do that is to keep the cookies
//
// :param urlToCall: URL to call to download the file
// :type urlToCall: string
// :param targetDir: directory where the downloaded file needs to be stored
// :type targetDir: string
// :param defaultFilename: default name for the downloaded file
// :type defaultFilename: string
// :returns: string, error -- path to the downloaded file and an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) backendFileDownload(urlToCall, targetDir, defaultFilename string) (string, error) {
	// Create the empty byte array that will contain the result
	resultSlice := []byte{}

	// Send the get request and keep the http answer to retrieve the file name
	httpAns, err := rpAPIInterface.HTTPCli.HTTPGetWithRes(urlToCall, &resultSlice)
	if err != nil {
		return "", errorInterpreter(err, "")
	}

	// We get the file content, let's get the filename
	filename := defaultFilename
	headerCntDisposition := httpAns.Header.Get("Content-Disposition")
	if headerCntDisposition != "" {
		// Parse the "content-disposition" header
		_, params, err := mime.ParseMediaType(headerCntDisposition)
		if err != nil {
			return "", errorInterpreter(err, "")
		}
		// Check if a filename is specified
		tmpFilename, ok := params["filename"]
		if ok {
			// There is a filename, save it!
			filename = tmpFilename
		}
	}

	// Now, write the file
	fullFilename := filepath.Join(targetDir, filename)
	err = utils.WriteToFile(fullFilename, string(resultSlice))
	if err != nil {
		return "", errorInterpreter(err, "")
	}

	return fullFilename, nil
}

// IsFeatureSupportedByBCK allows rp-cli to know if the backend
// of the current connection support new featured implementation
//
// :returns: bool, error -- True if it is State supported, false in the other case and an error if an error occurred
func (rpAPIInterface *RPBCKAPI) IsFeatureSupportedByBCK(featureVersion, warnMsg string, logger *logrus.Logger) (bool, error) {
	var err error
	resValue := false

	// Get the version of the current backend
	ver := rpapiv1.RPVersion{}
	if err = rpAPIInterface.RPVersionGet(&ver, false); err != nil {
		return resValue, err
	}

	// Check if the backend is in "debug" mode
	if strings.Contains(ver.Version, "?.?.?") {
		if logger != nil {
			logger.Warn("The backend seems to be in debug mode, no compatibility check will be done")
		}
		resValue = true
		err = nil
	} else {
		resValue, err = rputils.CheckVersionCompatibility(ver.Version, "", featureVersion)

		if !resValue && logger != nil {
			logger.Warn(warnMsg)
		}

	}

	return resValue, err
}
