/**
 * Copyright (C) 2022-2023 IoT.bzh Company
 * Contact: https://www.iot.bzh/licensing
 *
 * This file is part of the redpesk-cli project, the redpesk command line tool.
 *
 * $RP_BEGIN_LICENSE$
 * Commercial License Usage
 *  Licensees holding valid commercial IoT.bzh licenses may use this file in
 *  accordance with the commercial license agreement provided with the
 *  Software or, alternatively, in accordance with the terms contained in
 *  a written agreement between you and The IoT.bzh Company. For licensing terms
 *  and conditions see https://www.iot.bzh/terms-conditions. For further
 *  information use the contact form at https://www.iot.bzh/contact.
 *
 * GNU General Public License Usage
 *  Alternatively, this file may be used under the terms of the GNU General
 *  Public license version 3. This license is as published by the Free Software
 *  Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
 *  of this file. Please review the following information to ensure the GNU
 *  General Public License requirements will be met
 *  https://www.gnu.org/licenses/gpl-3.0.html.
 * $RP_END_LICENSE$
 */

package bckapi

import (
	"encoding/json"
	"fmt"
	"sort"
	"strconv"

	"git.ovh.iot/redpesk/redpesk-infra/rp-webserver/actions/rpapiv1"
)

// Sort deployments by Name
type _DplByLabel []rpapiv1.RPDeployment

func (s _DplByLabel) Len() int           { return len(s) }
func (s _DplByLabel) Swap(i, j int)      { s[i], s[j] = s[j], s[i] }
func (s _DplByLabel) Less(i, j int) bool { return s[i].CreatedTS.Before(s[j].CreatedTS) }

// ListDeployments Get the list of existing deployments
//
// :param dpls: Pointer to the rpapiv1.RPDeployments structure to fill
// :type dpls: *rpapiv1.RPDeployments
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ListDeployments(dpls *rpapiv1.RPDeployments, relPrjID string) error {

	// Initialize the call-back to save current page of released projects
	savingCB := func(tmpObjects []byte) {
		tmpDpls := rpapiv1.RPDeployments{}
		// Convert to the right type
		json.Unmarshal(tmpObjects, &tmpDpls)
		// Add the deployments in the final object
		dpls.Deployments = append(dpls.Deployments, tmpDpls.Deployments...)
	}

	// Initialize the paging structure that will allow to get all the objects
	pagingMngt := &RPBCKPaging{}

	// Call the function allowing to get all the object
	routeToUse := deploymentRoute
	if relPrjID != "" {
		routeToUse = fmt.Sprintf(releasedProjectDeploymentRoute, relPrjID)
	}
	err := pagingMngt.getAllObjects(rpAPIInterface, routeToUse, "", nil, savingCB)
	if err != nil {
		return errorInterpreter(err, deploymentObjectType)
	}

	// Sort the resulting projects
	sort.Sort(_DplByLabel(dpls.Deployments))

	return nil
}

// ListDeploymentsByPage gets the list of deployments on the factory
// but only one page of it
//
// :param deployments: Pointer to structure to be filled
// :type deployments: *rpapiv1.RPDeployments
// :param relPrjID: ID of the released project associated
// :type relPrjID: string
// :param page: Page number that the caller want
// :type page: int
// :param perPage: Number of item per page that the caller want
// :type perPage: int
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ListDeploymentsByPage(deployments *rpapiv1.RPDeployments, relPrjID string, page, perPage int) error {

	// Initialize the query parameters object
	queryParamMap := map[string]string{
		"page":     strconv.Itoa(page),
		"per_page": strconv.Itoa(perPage),
	}

	// Call the function allowing to get all the object
	routeToUse := deploymentRoute
	if relPrjID != "" {
		routeToUse = fmt.Sprintf(releasedProjectDeploymentRoute, relPrjID)
	}

	// Now, run the get request
	if err := rpAPIInterface.HTTPCli.Get(routeToUse, deployments, queryParamMap); err != nil {
		return errorInterpreter(err, deploymentObjectType)
	}

	return nil
}

// GetDeployment allows to request the backend a "deployment" structure
//
// :param deploymentID: ID of the deployment to get
// :type deploymentID: string
// :param ans: rpapiv1.RPDeploymentFull structure received as an answer of the request.
// :type ans: *rpapiv1.RPDeploymentFull
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) GetDeployment(deploymentID string, ans *rpapiv1.RPDeploymentFull) error {
	// Get the released project
	if err := rpAPIInterface.HTTPCli.Get(fmt.Sprintf(deploymentIDRoute, deploymentID), ans); err != nil {
		return errorInterpreter(err, deploymentObjectType)
	}

	return nil
}
