/**
 * Copyright (C) 2024 IoT.bzh Company
 * Contact: https://www.iot.bzh/licensing
 *
 * This file is part of the redpesk-cli project, the redpesk command line tool.
 *
 * $RP_BEGIN_LICENSE$
 * Commercial License Usage
 *  Licensees holding valid commercial IoT.bzh licenses may use this file in
 *  accordance with the commercial license agreement provided with the
 *  Software or, alternatively, in accordance with the terms contained in
 *  a written agreement between you and The IoT.bzh Company. For licensing terms
 *  and conditions see https://www.iot.bzh/terms-conditions. For further
 *  information use the contact form at https://www.iot.bzh/contact.
 *
 * GNU General Public License Usage
 *  Alternatively, this file may be used under the terms of the GNU General
 *  Public license version 3. This license is as published by the Free Software
 *  Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
 *  of this file. Please review the following information to ensure the GNU
 *  General Public License requirements will be met
 *  https://www.gnu.org/licenses/gpl-3.0.html.
 * $RP_END_LICENSE$
 */

package bckapi

import (
	"encoding/json"
	"fmt"
	"strconv"

	"git.ovh.iot/redpesk/redpesk-infra/rp-webserver/actions/rpapiv1"
	"github.com/gofrs/uuid"
)

// -------------------------------------------------------------------------------
// Types definition --------------------------------------------------------------
// -------------------------------------------------------------------------------
type RPCliDistribution struct {
	ID            uuid.UUID        `json:"id"`
	Slug          string           `json:"slug"`
	Name          string           `json:"name"`
	Version       string           `json:"version"`
	SubVersion    string           `json:"sub_version"`
	Description   string           `json:"description"`
	Architectures []rpapiv1.RPArch `json:"arches"`
	NbRPMsPerArch map[string]int   `json:"nb_rpms_per_arch"`
}

type RPCliDistributions struct {
	Distros      []RPCliDistribution `json:"data"`
	Page         int                 `json:"page"`
	PerPage      int                 `json:"per_page"`
	Reverse      bool                `json:"reverse"`
	TotalEntries int                 `json:"total_entries"`
}

// -------------------------------------------------------------------------------
// -------------------------------------------------------------------------------

var cachedDistros = RPCliDistributions{}

// RPDistrosGet allows to get the list of available distributions
// on the Redpesk instance
//
// :param distros: Pointer to RPCliDistributions structure to be filled
// :type distros: *RPCliDistributions
// :param forceReq: Boolean allowing to force the HTTP request
// :type forceReq: bool
// :returns: error -- an error if something went bad
func (rpAPIInterface *RPBCKAPI) RPDistrosGet(distros *RPCliDistributions, forceReq bool) error {

	// Send the request if the cache is empty or it is forced
	if len(cachedDistros.Distros) == 0 || forceReq {

		cachedDistros = RPCliDistributions{}

		// Initialize the call-back to save current page of distros
		savingCB := func(tmpObjects []byte) {
			tmpDistros := RPCliDistributions{}
			// Convert to the right type
			json.Unmarshal(tmpObjects, &tmpDistros)
			// Add the distributions in the final object
			cachedDistros.Distros = append(cachedDistros.Distros, tmpDistros.Distros...)
		}

		// Initialize the paging structure that will allow to get all the objects
		pagingMngt := &RPBCKPaging{}

		// Call the function allowing to get all the object
		err := pagingMngt.getAllObjects(rpAPIInterface, projectRoute, "external", nil, savingCB)
		if err != nil {
			return errorInterpreter(err, "")
		}
	}

	// Copy the distributions inside the distros
	distros.Distros = append(distros.Distros, cachedDistros.Distros...)

	return nil
}

// RPDistrosGetByPage allows to get the list of available distributions
// on the Redpesk instance, by page
//
// :param distros: Pointer to RPCliDistributions structure to be filled
// :type distros: *RPCliDistributions
// :param page: Page number that the caller want
// :type page: int
// :param perPage: Number of item per page that the caller want
// :type perPage: int
// :returns: error -- an error if something went bad
func (rpAPIInterface *RPBCKAPI) RPDistrosGetByPage(distros *RPCliDistributions, page, perPage int) error {

	// Initialize the query parameters object
	queryParamMap := map[string]string{
		"page":     strconv.Itoa(page),
		"per_page": strconv.Itoa(perPage),
		"type":     "external",
	}

	// Now, run the get request
	if err := rpAPIInterface.HTTPCli.Get(projectRoute, distros, queryParamMap); err != nil {
		return errorInterpreter(err, "")
	}

	return nil
}

// RPDistroGet allows to get properties of one distribution from the redpesk factory
//
// :param distroUUID: UUID of the distribution wanted
// :type distroUUID: uuid.UUID
// :param distro: Pointer to RPCliDistribution structure to be filled
// :type distro: *RPCliDistribution
// :returns: error -- an error if something went bad
func (rpAPIInterface *RPBCKAPI) RPDistroGet(distroUUID uuid.UUID, distro *RPCliDistribution) error {
	if err := rpAPIInterface.HTTPCli.Get(fmt.Sprintf(projectUpdateRoute, distroUUID.String()), distro); err != nil {
		return errorInterpreter(err, projectObjectType)
	}
	return nil
}

// RPPackagesGetByPage allows to get the list of available packages for a given distribution
//
// :param distros: Pointer to RPCliDistributions structure to be filled
// :type distros: *RPCliDistributions
// :param page: Page number that the caller want
// :type page: int
// :param perPage: Number of item per page that the caller want
// :type perPage: int
// :returns: error -- an error if something went bad
func (rpAPIInterface *RPBCKAPI) RPPackagesGetByPage(distroUUID uuid.UUID, arch, nameFilter, nameSubStrFilter string, pkgsInfo *rpapiv1.RPRepoInfoRpm, page, perPage int) error {

	// Initialize the query parameters object
	queryParamMap := map[string]string{
		"page":     strconv.Itoa(page),
		"per_page": strconv.Itoa(perPage),
		"arch":     arch,
	}

	// Add further filters if needed
	if nameFilter != "" {
		queryParamMap["filter"] = fmt.Sprintf("{\"name\":\"%s\"}", nameFilter)
	} else if nameSubStrFilter != "" {
		queryParamMap["filter"] = fmt.Sprintf("{\"name__substr\":\"%s\"}", nameSubStrFilter)
	}

	// Now, run the get request
	if err := rpAPIInterface.HTTPCli.Get(fmt.Sprintf(projectRepoInfoRoute, distroUUID.String()), pkgsInfo, queryParamMap); err != nil {
		return errorInterpreter(err, projectObjectType)
	}

	return nil
}

// ConvertRpCliDistrosToBCKDistros does the conversion between two types:
// from RPCliDistributions to rpapiv1.RPDistros
//
// :param input: Input structure
// :type input: RPCliDistributions
// :param output: Pointer to the output structure
// :type output: *rpapiv1.RPDistros
// :returns: error -- an error if something went bad
func ConvertRpCliDistrosToBCKDistros(input RPCliDistributions, output *rpapiv1.RPDistros) error {
	// Convert to array of bytes
	dataBytes, err := json.Marshal(input)
	if err != nil {
		return err
	}
	// Convert to the right structure
	return json.Unmarshal(dataBytes, output)
}
