/**
 * Copyright (C) 2021-2023 IoT.bzh Company
 * Contact: https://www.iot.bzh/licensing
 *
 * This file is part of the redpesk-cli project, the redpesk command line tool.
 *
 * $RP_BEGIN_LICENSE$
 * Commercial License Usage
 *  Licensees holding valid commercial IoT.bzh licenses may use this file in
 *  accordance with the commercial license agreement provided with the
 *  Software or, alternatively, in accordance with the terms contained in
 *  a written agreement between you and The IoT.bzh Company. For licensing terms
 *  and conditions see https://www.iot.bzh/terms-conditions. For further
 *  information use the contact form at https://www.iot.bzh/contact.
 *
 * GNU General Public License Usage
 *  Alternatively, this file may be used under the terms of the GNU General
 *  Public license version 3. This license is as published by the Free Software
 *  Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
 *  of this file. Please review the following information to ensure the GNU
 *  General Public License requirements will be met
 *  https://www.gnu.org/licenses/gpl-3.0.html.
 * $RP_END_LICENSE$
 */

package bckapi

import (
	"encoding/json"
	"errors"
	"fmt"
	"sort"
	"strconv"

	"git.ovh.iot/redpesk/redpesk-infra/rp-webserver/actions/rpapiv1"
)

// Images tests cached
var cachedImgRPTests []rpapiv1.RPTest

// Sort images by Slug
type _ImgByLabel []rpapiv1.RPImage

func (s _ImgByLabel) Len() int           { return len(s) }
func (s _ImgByLabel) Swap(i, j int)      { s[i], s[j] = s[j], s[i] }
func (s _ImgByLabel) Less(i, j int) bool { return s[i].Slug < s[j].Slug }

// ImagesListGetOfTypes Get the list of existing images, the type
// of images to retrieve can be specified in a slice of string
//
// :param images: Pointer to the rpapiv1.RPImages structure to fill
// :type images: *rpapiv1.RPImages
// :param projectID: the project id
// :type projectID: string
// :param types: Types of images to retrieve
// :type types: []string
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ImagesListGetOfTypes(images *rpapiv1.RPImages, projectID string, types []string, filter string) error {

	// Construct the string containing all the types to retrieve, by default the value is "standard"
	typesStr := "standard"

	if len(types) > 0 {
		// Get rid of the default value
		typesStr = ""
		// Construct the type string
		for index, currentType := range types {
			typesStr += currentType
			if index < (len(types) - 1) {
				typesStr += ","
			}
		}
	}

	addQueryParams := make(map[string]string)
	if filter != "" {
		addQueryParams["filters"] = filter
	}

	// Initialize the call-back to save current page of images
	savingCB := func(tmpObjects []byte) {
		tmpImages := rpapiv1.RPImages{}
		// Convert to the right type
		json.Unmarshal(tmpObjects, &tmpImages)
		// Add the images in the final object
		images.Images = append(images.Images, tmpImages.Images...)
	}

	// Decide which route to use
	routeToCall := ""
	if projectID == "" {
		routeToCall = imagesRoute
	} else {
		routeToCall = projectRoutePrefix + projectID + imagesRoute
	}

	// Initialize the paging structure that will allow to get all the objects
	pagingMngt := &RPBCKPaging{}

	// Call the function allowing to get all the object
	err := pagingMngt.getAllObjects(rpAPIInterface, routeToCall, typesStr, addQueryParams, savingCB)
	if err != nil {
		return errorInterpreter(err, imageObjectType)
	}

	// Sort the resulting projects
	sort.Sort(_ImgByLabel(images.Images))

	return nil
}

// ImagesListByTypesPerPage Get the list of existing images for a given page.
// The type of images to retrieve can be specified in a slice of string
//
// :param images: Pointer to the rpapiv1.RPImages structure to fill
// :type images: *rpapiv1.RPImages
// :param projectID: the project id
// :type projectID: string
// :param types: Types of images to retrieve
// :type types: []string
// :param page: Page number that the caller want
// :type page: int
// :param perPage: Number of item per page that the caller want
// :type perPage: int
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ImagesListByTypesPerPage(images *rpapiv1.RPImages, projectID string, types []string, page, perPage int) error {

	// Construct the string containing all the types to retrieve, by default the value is "standard"
	typesStr := "standard"

	if len(types) > 0 {
		// Get rid of the default value
		typesStr = ""
		// Construct the type string
		for index, currentType := range types {
			typesStr += currentType
			if index < (len(types) - 1) {
				typesStr += ","
			}
		}
	}

	// Initialize the query parameters object
	queryParamMap := map[string]string{
		"page":     strconv.Itoa(page),
		"per_page": strconv.Itoa(perPage),
		"type":     typesStr,
	}

	// Decide which route to use
	routeToCall := ""
	if projectID == "" {
		routeToCall = imagesRoute
	} else {
		routeToCall = projectRoutePrefix + projectID + imagesRoute
	}

	// Now, run the get request
	if err := rpAPIInterface.HTTPCli.Get(routeToCall, &images, queryParamMap); err != nil {
		return errorInterpreter(err, imageObjectType)
	}

	return nil
}

// ImageGet Get the information available for a given image
//
// :param id: id of the image wanted
// :type id: string
// :param img: Pointer to the rpapiv1.RPImageFull structure to fill
// :type img: *rpapiv1.RPImageFull
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ImageGet(id string, img *rpapiv1.RPImageFull) error {

	if err := rpAPIInterface.HTTPCli.Get(fmt.Sprintf(imagesIDRoute, id), img); err != nil {
		return errorInterpreter(err, imageObjectType)
	}

	return nil
}

// ImageCreate asks the Redpesk backend to create a new image
//
// :param img: rpapiv1.RPImageFull structure containing the data to create the image
// :type img: *rpapiv1.RPImageFull
// :param ans: rpapiv1.RPImageFull structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPImageFull
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ImageCreate(img *rpapiv1.RPImageFull, ans *rpapiv1.RPImageFull) error {
	var tmpAns *rpapiv1.RPImageFull

	// Check that the pointer to the structure to send is not nil
	if img == nil {
		return errors.New(imageCreateNilError)
	}

	if ans == nil {
		tmpAns = &rpapiv1.RPImageFull{}
	} else {
		tmpAns = ans
	}

	if err := rpAPIInterface.HTTPCli.Post(imagesRoute, img, tmpAns); err != nil {
		return errorInterpreter(err, imageObjectType)
	}

	return nil
}

// ImageUpdate asks the Redpesk backend to update an existing image
//
// :param imageID: ID of the image to update
// :type imageID: string
// :param img: rpapiv1.RPImageFull structure containing the data to update the image
// :type img: *rpapiv1.RPImageFull
// :param ans: rpapiv1.RPImageFull structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPImageFull
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ImageUpdate(imageID string, img *rpapiv1.RPImageFull, ans *rpapiv1.RPImageFull) error {
	var tmpAns *rpapiv1.RPImageFull

	// Check that the pointer to the structure to send is not nil
	if img == nil {
		return errors.New(imageCreateNilError)
	}

	if ans == nil {
		tmpAns = &rpapiv1.RPImageFull{}
	} else {
		tmpAns = ans
	}

	if err := rpAPIInterface.HTTPCli.Put(fmt.Sprintf(imagesIDRoute, imageID), img, tmpAns); err != nil {
		return errorInterpreter(err, imageObjectType)
	}

	return nil
}

// ImageDelete asks the Redpesk backend to delete an existing image
//
// :param imageID: ID of the image to delete
// :type imageID: string
// :param ans: rpapiv1.RPImageFull structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPImageFull
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ImageDelete(imageID string, ans *rpapiv1.RPImageFull) error {
	var tmpAns *rpapiv1.RPImageFull

	if ans == nil {
		tmpAns = &rpapiv1.RPImageFull{}
	} else {
		tmpAns = ans
	}

	// Send the delete request
	if err := rpAPIInterface.HTTPCli.Delete(fmt.Sprintf(imagesIDRoute, imageID), tmpAns); err != nil {
		return errorInterpreter(err, imageObjectType)
	}

	return nil
}

// Sort released images by Slug
type _RelImgByLabel []rpapiv1.RPReleaseImage

func (s _RelImgByLabel) Len() int           { return len(s) }
func (s _RelImgByLabel) Swap(i, j int)      { s[i], s[j] = s[j], s[i] }
func (s _RelImgByLabel) Less(i, j int) bool { return s[i].Slug < s[j].Slug }

// ListReleasedImages Get the list of existing released images
//
// :param images: Pointer to the rpapiv1.RPReleaseImages structure to fill
// :type images: *rpapiv1.RPReleaseImages
// :param filter: value of the filter to send in the request
// :type filter: string
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ListReleasedImages(images *rpapiv1.RPReleaseImages, filter string) error {

	addQueryParams := make(map[string]string)
	if filter != "" {
		addQueryParams["filters"] = filter
	}

	// Initialize the call-back to save current page of released images
	savingCB := func(tmpObjects []byte) {
		tmpImages := rpapiv1.RPReleaseImages{}
		// Convert to the right type
		json.Unmarshal(tmpObjects, &tmpImages)
		// Add the projects in the final object
		images.ReleaseImages = append(images.ReleaseImages, tmpImages.ReleaseImages...)
	}

	// Initialize the paging structure that will allow to get all the objects
	pagingMngt := &RPBCKPaging{}

	// Call the function allowing to get all the object
	err := pagingMngt.getAllObjects(rpAPIInterface, releasedImageRoute, "", addQueryParams, savingCB)
	if err != nil {
		return errorInterpreter(err, releaseImageObjectType)
	}

	// Sort the resulting projects
	sort.Sort(_RelImgByLabel(images.ReleaseImages))

	return nil
}

// ListReleasedImagesPerPage Get the list of existing released images, at a given page
//
// :param images: Pointer to the rpapiv1.RPReleaseImages structure to fill
// :type images: *rpapiv1.RPReleaseImages
// :param page: Page number that the caller want
// :type page: int
// :param perPage: Number of item per page that the caller want
// :type perPage: int
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ListReleasedImagesPerPage(images *rpapiv1.RPReleaseImages, page, perPage int) error {

	// Initialize the query parameters object
	queryParamMap := map[string]string{
		"page":     strconv.Itoa(page),
		"per_page": strconv.Itoa(perPage),
	}

	// Now, run the get request
	if err := rpAPIInterface.HTTPCli.Get(releasedImageRoute, images, queryParamMap); err != nil {
		return errorInterpreter(err, releaseImageObjectType)
	}

	return nil
}

// ImageRelease asks the Redpesk backend to release an existing image.
// The image is readable and usable by all the factory members afterward.
//
// :param imageID: ID of the image to release
// :type imageID: string
// :param imgRelease: Pointer to a RPReleaseImage structure containing the details needed to release the image
// :type imgRelease: *rpapiv1.RPReleaseImage
// :param ans: rpapiv1.RPReleaseImageFull structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPReleaseImageFull
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ImageRelease(imageID string, imgRelease *rpapiv1.RPReleaseImage, ans *rpapiv1.RPReleaseImageFull) error {
	var tmpAns *rpapiv1.RPReleaseImageFull

	if ans == nil {
		tmpAns = &rpapiv1.RPReleaseImageFull{}
	} else {
		tmpAns = ans
	}

	if err := rpAPIInterface.HTTPCli.Post(fmt.Sprintf(imagesReleaseRoute, imageID), imgRelease, tmpAns); err != nil {
		return errorInterpreter(err, releaseImageObjectType)
	}

	return nil
}

// GetReleasedImage allows to request the backend the "released image" structure corresponding to a "source" image
// that has been released
//
// :param imageID: ID of the image to release
// :type imageID: string
// :param ans: rpapiv1.RPReleaseImageFull structure received as an answer of the request.
// :type ans: *rpapiv1.RPReleaseImageFull
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) GetReleasedImage(imageID string, ans *rpapiv1.RPReleaseImageFull) error {
	// Get the released project
	if err := rpAPIInterface.HTTPCli.Get(fmt.Sprintf(releasedImageIDRoute, imageID), ans); err != nil {
		return errorInterpreter(err, releaseImageObjectType)
	}
	return nil
}

// UpdateReleasedImage asks the Redpesk backend to update an existing released image.
//
// :param imageID: ID of the released image
// :type imageID: string
// :param imgRelease: rpapiv1.RPReleaseImage is a pointer to a structure containing the details needed to update the released image
// :type imgRelease: *rpapiv1.RPReleaseImage
// :param ans: rpapiv1.RPReleaseImageFull structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPReleaseImageFull
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) UpdateReleasedImage(imageID string, imgRelease *rpapiv1.RPReleaseImage, ans *rpapiv1.RPReleaseImageFull) error {
	var tmpAns *rpapiv1.RPReleaseImageFull

	if ans == nil {
		tmpAns = &rpapiv1.RPReleaseImageFull{}
	} else {
		tmpAns = ans
	}

	if err := rpAPIInterface.HTTPCli.Put(fmt.Sprintf(releasedImageIDRoute, imageID), imgRelease, tmpAns); err != nil {
		return errorInterpreter(err, releaseImageObjectType)
	}

	return nil
}

// DeleteReleasedImage asks the Redpesk backend to delete an existing released image.
//
// :param imageID: ID of the released image to delete
// :type imageID: string
// :param ans: rpapiv1.RPReleaseImageFull structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPReleaseImageFull
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) DeleteReleasedImage(imageID string, ans *rpapiv1.RPReleaseImageFull) error {
	var tmpAns *rpapiv1.RPReleaseImageFull

	if ans == nil {
		tmpAns = &rpapiv1.RPReleaseImageFull{}
	} else {
		tmpAns = ans
	}

	if err := rpAPIInterface.HTTPCli.Delete(fmt.Sprintf(releasedImageIDRoute, imageID), tmpAns); err != nil {
		return errorInterpreter(err, releaseImageObjectType)
	}

	return nil
}

// GetReleasedImageAudience asks the Redpesk backend to return the audience of a released image
//
// :param imageID: ID of the released image
// :type imageID: string
// :param ans: rpapiv1.RPObjectRoles structure received as an answer to the request.
// :type ans: *rpapiv1.RPObjectRoles
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) GetReleasedImageAudience(imageID string, ans *rpapiv1.RPObjectRoles) error {

	if err := rpAPIInterface.HTTPCli.Get(fmt.Sprintf(releasedImageAudienceRoute, imageID), ans); err != nil {
		return errorInterpreter(err, releaseImageObjectType)
	}

	// Finish to create the structure to avoid having "null pointer dereference" panic later
	if ans.IdentityRolesLists == nil {
		ans.IdentityRolesLists = make(map[string]*rpapiv1.RPIdentityRolesLists)
	}
	for _, rolesList := range ans.IdentityRolesLists {
		if rolesList.AvailableRoles == nil {
			rolesList.AvailableRoles = &([]rpapiv1.RPRole{})
		}
		if rolesList.SetRoles == nil {
			rolesList.SetRoles = &([]rpapiv1.RPRole{})
		}
	}
	// -------------------------------------------------------------------------------------

	return nil
}

// UpdateReleasedImageAudience asks the Redpesk backend to change the audience of a released image
//
// :param imageID: ID of the released image
// :type imageID: string
// :param rolesObj: rpapiv1.RPObjectRoles is a pointer to a structure containing the details about the audience wanted for the image
// :type rolesObj: *rpapiv1.RPObjectRoles
// :param ans: rpapiv1.RPObjectRoles structure received as an answer to the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPObjectRoles
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) UpdateReleasedImageAudience(imageID string, rolesObj *rpapiv1.RPObjectRoles, ans *rpapiv1.RPObjectRoles) error {
	var tmpAns *rpapiv1.RPObjectRoles

	if ans == nil {
		tmpAns = &rpapiv1.RPObjectRoles{}
	} else {
		tmpAns = ans
	}

	if err := rpAPIInterface.HTTPCli.Put(fmt.Sprintf(releasedImageAudienceRoute, imageID), rolesObj, tmpAns); err != nil {
		return errorInterpreter(err, releaseImageObjectType)
	}

	return nil
}

// ImageMngtImport asks the Redpesk backend to import a new image
//
// :param img: rpapiv1.RPImageFull structure containing the data to create the image
// :type img: *rpapiv1.RPImageFull
// :param ans: rpapiv1.RPImageFull structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPImageFull
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ImageMngtImport(img *rpapiv1.RPImageFull, ans *rpapiv1.RPImageFull) error {
	var tmpAns *rpapiv1.RPImageFull

	// Check that the pointer to the structure to send is not nil
	if img == nil {
		return errors.New(imageCreateNilError)
	}

	if ans == nil {
		tmpAns = &rpapiv1.RPImageFull{}
	} else {
		tmpAns = ans
	}

	if err := rpAPIInterface.HTTPCli.Post(imagesMngtImportRoute, img, tmpAns); err != nil {
		return errorInterpreter(err, "")
	}

	return nil
}

// ImageBuild asks the Redpesk backend to build an existing image
//
// :param imgID: ID of the image to build
// :type imgID: string
// :param comment: Comment to save with the build
// :type comment: string
// :param ans: buffer to store the structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPBuild
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ImageBuild(imgID, comment string, ans *rpapiv1.RPBuilds) error {
	var tmpAns *rpapiv1.RPBuilds

	if ans == nil {
		tmpAns = &rpapiv1.RPBuilds{}
	} else {
		tmpAns = ans
	}

	bldStartObj := rpapiv1.RPBuildStart{Comment: comment}
	// Ask the backend to start the build
	if err := rpAPIInterface.HTTPCli.Post(fmt.Sprintf(imagesBuildRoute, imgID), &bldStartObj, tmpAns); err != nil {
		return errorInterpreter(err, "")
	}

	return nil
}

// ImagesBuildsListGet gets the list of all existing builds in the given image.
//
// :param builds: Pointer to the rpapiv1.RPBuilds structure to fill
// :type builds: *rpapiv1.RPBuilds
// :param imgID: image ID
// :type imgID: string
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ImagesBuildsListGet(builds *rpapiv1.RPBuilds, imgID string) error {

	// Initialize the call-back to save current page of builds
	savingCB := func(tmpObjects []byte) {
		tmpBuilds := rpapiv1.RPBuilds{}
		// Convert to the right type
		json.Unmarshal(tmpObjects, &tmpBuilds)
		// Add the applications in the final object
		builds.Builds = append(builds.Builds, tmpBuilds.Builds...)
	}

	// Initialize the paging structure that will allow to get all the objects
	pagingMngt := &RPBCKPaging{}

	// Call the function allowing to get all the object
	err := pagingMngt.getAllObjects(rpAPIInterface, fmt.Sprintf(imagesBuildRoute, imgID), "", nil, savingCB)
	if err != nil {
		return errorInterpreter(err, "")
	}

	return nil
}

// ImagesBuildsListByPage gets the list of existing builds in the given image, by page
//
// :param builds: Pointer to the rpapiv1.RPBuilds structure to fill
// :type builds: *rpapiv1.RPBuilds
// :param imgID: image ID
// :type imgID: string
// :param page: Page number that the caller want
// :type page: int
// :param perPage: Number of item per page that the caller want
// :type perPage: int
// :param sortOpt: Optional parameter allowing to sort the list
// :type sortOpt: string
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ImagesBuildsListByPage(builds *rpapiv1.RPBuilds, imgID string, page, perPage int, sortOpt string) error {

	// Initialize the query parameters object
	queryParamMap := map[string]string{
		"page":     strconv.Itoa(page),
		"per_page": strconv.Itoa(perPage),
	}

	// Add the sorting option if specified
	if sortOpt != "" {
		queryParamMap["sort"] = sortOpt
	}

	// Now, run the get request
	if err := rpAPIInterface.HTTPCli.Get(fmt.Sprintf(imagesBuildRoute, imgID), builds, queryParamMap); err != nil {
		return errorInterpreter(err, "")
	}

	return nil
}

// ImagesTestsListGet gets the list of all existing tests for the given image.
//
// :param prjID: project ID
// :type prjID: string
// :param imgID: image ID
// :type imgID: string
// :returns: []rpapiv1.RPTest, error -- list of associated tests and an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ImagesTestsListGet(prjID, imgID string) ([]rpapiv1.RPTest, error) {
	// Call the route in order to get the tests stat for the given project
	prjStats := rpapiv1.RPTestStatsProject{}
	if err := rpAPIInterface.ProjectGetTestsStats(prjID, 0, &prjStats); err != nil {
		return []rpapiv1.RPTest{}, err
	}

	// Now find the right image
	for _, imgStats := range prjStats.ImageStats {
		if imgStats.ImageID.String() == imgID {
			return imgStats.LastCITests, nil
		}
	}

	// If we reach here, nothing has been found
	return []rpapiv1.RPTest{}, nil
}

// ImagesTestsListGetByPage gets the list of existing tests in the given image for a given page
//
// :param prjID: project ID
// :type prjID: string
// :param imgID: image ID
// :type imgID: string
// :param page: Page number that the caller want
// :type page: int
// :param perPage: Number of item per page that the caller want
// :type perPage: int
// :param forceReq: Force the request to be sent and don't use the cache
// :type forceReq: bool
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ImagesTestsListGetByPage(prjID, imgID string, page, perPage int, forceReq bool) (rpapiv1.RPTests, error) {

	// Manage a cache since there is not paging managed on the backend side.
	var err error
	if (cachedImgRPTests == nil) || forceReq {
		cachedImgRPTests, err = rpAPIInterface.ImagesTestsListGet(prjID, imgID)
		if err != nil {
			return rpapiv1.RPTests{}, err
		}
	}

	// Initialize the structure that we will return
	retStruct := rpapiv1.RPTests{
		Page:         page,
		PerPage:      perPage,
		TotalEntries: len(cachedImgRPTests),
		Tests:        []rpapiv1.RPTest{},
	}

	// Calculate the minimum and maximum index wanted by the caller
	minIndex := (page - 1) * perPage
	maxIndex := (page * perPage) - 1

	// Loop over the cached tests
	for index, test := range cachedImgRPTests {
		if (index >= minIndex) && (index <= maxIndex) {
			retStruct.Tests = append(retStruct.Tests, test)
		}
	}

	// If we reach here, nothing has been found
	return retStruct, nil
}
