/**
 * Copyright (C) 2020-2023 IoT.bzh Company
 * Contact: https://www.iot.bzh/licensing
 *
 * This file is part of the redpesk-cli project, the redpesk command line tool.
 *
 * $RP_BEGIN_LICENSE$
 * Commercial License Usage
 *  Licensees holding valid commercial IoT.bzh licenses may use this file in
 *  accordance with the commercial license agreement provided with the
 *  Software or, alternatively, in accordance with the terms contained in
 *  a written agreement between you and The IoT.bzh Company. For licensing terms
 *  and conditions see https://www.iot.bzh/terms-conditions. For further
 *  information use the contact form at https://www.iot.bzh/contact.
 *
 * GNU General Public License Usage
 *  Alternatively, this file may be used under the terms of the GNU General
 *  Public license version 3. This license is as published by the Free Software
 *  Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
 *  of this file. Please review the following information to ensure the GNU
 *  General Public License requirements will be met
 *  https://www.gnu.org/licenses/gpl-3.0.html.
 * $RP_END_LICENSE$
 */

package bckapi

import (
	"encoding/json"
	"errors"
	"fmt"
	"strconv"

	"git.ovh.iot/redpesk/redpesk-infra/rp-golang-common/rphttp"
	"git.ovh.iot/redpesk/redpesk-infra/rp-webserver/actions/rpapiv1"
	"github.com/sirupsen/logrus"
	"github.com/spf13/viper"
)

// -------------------------------------------------------------------------------
// Types definition --------------------------------------------------------------
// -------------------------------------------------------------------------------

// RPAPI definition
type RPBCKAPI struct {
	HTTPCli *rphttp.HTTPClient
	Log     *logrus.Logger
}

// RPBCKPaging definition (used to manage pagination)
type RPBCKPaging struct {
	Page         int  `json:"page"`
	PerPage      int  `json:"per_page"`
	Reverse      bool `json:"reverse"`
	TotalEntries int  `json:"total_entries"`
}

// -------------------------------------------------------------------------------
// -------------------------------------------------------------------------------

// -------------------------------------------------------------------------------
// Functions definition ----------------------------------------------------------
// -------------------------------------------------------------------------------

// GetLogger allows to get the logget used by the API
//
// :returns: *logrus.Logger -- The logger used
func (bckAPI *RPBCKAPI) GetLogger() *logrus.Logger {
	return bckAPI.Log
}

// GetURLPrefix returns the prefix to use in the API URLs
//
// :returns: string -- The prefix to use
func (bckAPI *RPBCKAPI) GetURLPrefix() string {
	return urlPrefix
}

// GetLogPrefix returns the prefix to use in the logs
//
// :returns: string -- The prefix to use
func (bckAPI *RPBCKAPI) GetLogPrefix() string {
	return bckLogPrefix
}

// GetCookieKey returns the key used to pass cookie to the
// server
//
// :returns: string -- The key to use
func (bckAPI *RPBCKAPI) GetCookieKey() string {
	return bckCookieKey
}

// SetHTTPClient save a HTTPClient pointer as its client used to
// execute HTTP requests
func (bckAPI *RPBCKAPI) SetHTTPClient(httpClient *rphttp.HTTPClient) {
	// Save the new HTTPClient
	bckAPI.HTTPCli = httpClient
}

// GetHTTPClient returns the HTTP client currently used by the API
//
// :returns: *rphttp.HTTPClient -- The HTTP client used
func (bckAPI *RPBCKAPI) GetHTTPClient() *rphttp.HTTPClient {
	return bckAPI.HTTPCli
}

// FillAuthorizationFields is called to fill the Authorization
// fields in the HTTPClientConfig. These fields will be used in the
// header of the requests sent to the server
//
// :param token: Access token furnished by the user
// :type token: string
// :param config: Configuration to complete
// :type config: *rphttp.HTTPClientConfig
func (bckAPI *RPBCKAPI) FillAuthorizationFields(token string, config *rphttp.HTTPClientConfig) {
	// Set the authorization header key
	config.HeaderAPIKeyName = authorizationHeader

	// Set the authorization value key
	config.Apikey = (bearerPrefix + token)
}

// InitiateAuthentication is called right before connecting
// the websocket. If some additional authentication is required
// by the server (additionally to the token passed in the header),
// like a verb to call or something else, this is here that
// it can be done.
//
// :returns: error -- An error if something went wrong
func (bckAPI *RPBCKAPI) InitiateAuthentication() error {
	// Get current configuration
	rpConf := rpapiv1.RPAPIConfig{}
	if err := bckAPI.RPConfigGet(&rpConf); err != nil {
		return err
	}
	return nil
}

// GetVersion returns the version of the server as a string
//
// :returns: string -- The key to use
func (bckAPI *RPBCKAPI) GetVersion() (string, error) {
	// Initialize the result
	verResult := ""

	// Structure containing the version
	versionStruct := &rpapiv1.RPVersion{}

	// Request the version to the server
	if err := bckAPI.RPVersionGet(versionStruct, true); err != nil {
		return verResult, err
	}

	// Create the version string
	verResult = fmt.Sprintf("Version: %s, Git Tag: %s", versionStruct.Version, versionStruct.VersionGitTag)

	return verResult, nil
}

// RegisterToEvent allows to register the client to an event from the server
//
// :returns: error -- an error if a problem occured, nil in the other case
func (bckAPI *RPBCKAPI) RegisterToEvent(evt string) error {
	// Create the structure used to register to an event
	eventStruct := &rpapiv1.EventRegisterArgs{
		Name: evt,
	}

	// Register to the event on the server
	res := make(map[string]interface{})
	if err := bckAPI.HTTPCli.Post(evtRegisterRoute, eventStruct, &res); err != nil {
		return err
	}

	if res["status"] != "OK" {
		errorString := fmt.Sprintf("Registration failed with error %v", res)
		return errors.New(errorString)
	}

	return nil
}

// UnregisterFromEvent allows to unregister the client from an event
//
// :returns: error -- an error if a problem occured, nil in the other case
func (bckAPI *RPBCKAPI) UnregisterFromEvent(evt string) error {
	// Create the structure used to unregister from an event
	eventStruct := &rpapiv1.EventUnRegisterArgs{
		Name: evt,
	}

	// Unregister from the event on the server
	res := make(map[string]interface{})
	if err := bckAPI.HTTPCli.Post(evtUnregisterRoute, eventStruct, &res); err != nil {
		return err
	}

	if res["status"] != "OK" {
		errorString := fmt.Sprintf("Unregistration failed with error %v", res)
		return errors.New(errorString)
	}

	return nil
}

// -------------------------------------------------------------------------------
// -------------------------------------------------------------------------------

// -------------------------------------------------------------------------------
// Private functions definition --------------------------------------------------
// -------------------------------------------------------------------------------

// updatePagingParams allows to update the paging parameters from a structure received
// in a HTTP answer
//
// :param receivedHttpStruct: Structure received from the backend that is managed using paging
// :type receivedHttpStruct: interface{}
// :returns: error -- an error if a problem occurred
func (pagingObj *RPBCKPaging) updatePagingParams(receivedHttpStruct interface{}) error {

	// In order to get uniquely the paging use Marshal/Unmarshal
	bytesData, err := json.Marshal(receivedHttpStruct)
	if err != nil {
		return err
	}

	err = json.Unmarshal(bytesData, pagingObj)
	if err != nil {
		return err
	}

	return nil
}

// updateNextQueryParams allows to update the query parameters for the next request
// allowing to get the paged objects. If no more request is needed, it return false
//
// :param queryParams: Query parameters map to update
// :type queryParams: *map[string]string
// :returns: bool, error -- false if no more request is needed an error if a problem occurred
func (pagingObj *RPBCKPaging) updateNextQueryParams(queryParams *map[string]string) (bool, error) {

	// Calculation of the total number of pages needed
	totalPages := (pagingObj.TotalEntries / pagingObj.PerPage)
	modulusEntries := (pagingObj.TotalEntries % pagingObj.PerPage)

	if modulusEntries > 0 {
		totalPages++
	}

	// Check if we need to send another request
	if (pagingObj.Page >= totalPages) || (pagingObj.TotalEntries == 0) {
		// We reached the last page!
		return false, nil
	}

	// We still have page(s) to get!
	(*queryParams)["page"] = strconv.Itoa(pagingObj.Page + 1)
	(*queryParams)["per_page"] = objectNbPerPage

	return true, nil
}

// prepareNextPageRequest allows to update the paging parameters and the query parameters used to browse all pages
// If no more request is needed, it return false
//
// :param receivedHttpStruct: Structure received from the backend that is managed using paging
// :type receivedHttpStruct: interface{}
// :param queryParams: Query parameters map to update
// :type queryParams: *map[string]string
// :returns: bool, error -- false if no more request is needed an error if a problem occurred
func (pagingObj *RPBCKPaging) prepareNextPageRequest(receivedHttpStruct interface{}, queryParams *map[string]string) (bool, error) {

	// Process the paging parameters
	err := pagingObj.updatePagingParams(receivedHttpStruct)
	if err != nil {
		return false, err
	}

	// Get the next value of query parameters
	result, err := pagingObj.updateNextQueryParams(queryParams)
	if err != nil {
		return result, err
	}

	return result, err
}

// getAllObjects is an agnostic function allowing to get all the paginated object of a given type
//
// :param bckAPIObj: Interface allowing to make HTTP requests to the backend
// :type bckAPIObj: *RPBCKAPI
// :param routeToCall: Route to call in order to get the objects
// :type routeToCall: string
// :param objTypeRequested: Type of object requested (standard, released, external)
// :type objTypeRequested: string
// :param additionalQueryParams: Additional query parameters to pass to the request
// :type additionalQueryParams: map[string]string
// :param saveObjsCB: CallBack allowing to save the current page objects
// :type saveObjsCB: func(tmpObjects []byte)
// :returns: error -- an error if a problem occurred
func (pagingObj *RPBCKPaging) getAllObjects(bckAPIObj *RPBCKAPI, routeToCall, objTypeRequested string, additionalQueryParams map[string]string, saveObjsCB func(tmpObjects []byte)) error {

	// Initialize the query parameters object
	queryParamMap := map[string]string{
		"page":     objectFirstPage,
		"per_page": objectNbPerPage,
	}
	if objTypeRequested != "" {
		queryParamMap["type"] = objTypeRequested
	}

	// Add the additional query parameters
	for queryKey, queryValue := range additionalQueryParams {
		queryParamMap[queryKey] = queryValue
	}

	// Initialize the loop condition
	anotherPageExists := true
	var err error
	for anotherPageExists {
		// Create a temporary interface
		var tmpInterface interface{}

		// Request to the redpesk backend
		if err = bckAPIObj.HTTPCli.Get(routeToCall, &tmpInterface, queryParamMap); err != nil {
			return err
		}

		// Convert the interface to byte slice
		bytesSlice, err := json.Marshal(tmpInterface)
		if err != nil {
			return err
		}

		// Call the callback to save the current page object
		if saveObjsCB != nil {
			saveObjsCB(bytesSlice)
		}

		// Prepare next request
		anotherPageExists, err = pagingObj.prepareNextPageRequest(tmpInterface, &queryParamMap)
		if err != nil {
			return err
		}
	}

	return nil
}

// errorInterpreter allows to interpret an error coming from the Redpesk backend,
// and change it to an error understandable by the user.
//
// :param inputErr: Error coming from the Redpesk backend
// :type inputErr: error
// :param objectType: Type of the object concerned by the error
// :type objectType: string
// :returns: the interpreted error
func errorInterpreter(inputErr error, objectType string) error {
	var errRes error
	var httpErr *rphttp.HTTPError
	errStatus := -1

	// Check if the error is nil
	if inputErr == nil {
		return inputErr
	}

	// Try to convert the error to HTTPError
	if errors.As(inputErr, &httpErr) {
		errStatus = httpErr.StatusCode
	}

	// Get the server alias of the current connection
	serverAlias := viper.GetString("serveralias")

	switch errStatus {
	case 401:
		// Construct the error message to let know the user which token is expired
		errMsg := fmt.Sprintf(accessTokenExpiredError, serverAlias)
		// Construct the error
		errRes = errors.New(errMsg)

	case 403:
		if objectType == "" {
			errRes = inputErr
		} else {
			// Construct the error message to let know the user what is the problem
			errMsg := fmt.Sprintf(status403ErrorOnObjError, objectType)
			// Construct the error
			errRes = errors.New(errMsg)
		}

	case 503:
		// Construct the error message to let know the user that the backend is not available
		errMsg := fmt.Sprintf(status503Error, serverAlias)
		// Construct the error
		errRes = errors.New(errMsg)

	default:
		errRes = inputErr
	}

	return errRes
}

// -------------------------------------------------------------------------------
// -------------------------------------------------------------------------------
