/**
 * Copyright (C) 2020-2023 IoT.bzh Company
 * Contact: https://www.iot.bzh/licensing
 *
 * This file is part of the redpesk-cli project, the redpesk command line tool.
 *
 * $RP_BEGIN_LICENSE$
 * Commercial License Usage
 *  Licensees holding valid commercial IoT.bzh licenses may use this file in
 *  accordance with the commercial license agreement provided with the
 *  Software or, alternatively, in accordance with the terms contained in
 *  a written agreement between you and The IoT.bzh Company. For licensing terms
 *  and conditions see https://www.iot.bzh/terms-conditions. For further
 *  information use the contact form at https://www.iot.bzh/contact.
 *
 * GNU General Public License Usage
 *  Alternatively, this file may be used under the terms of the GNU General
 *  Public license version 3. This license is as published by the Free Software
 *  Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
 *  of this file. Please review the following information to ensure the GNU
 *  General Public License requirements will be met
 *  https://www.gnu.org/licenses/gpl-3.0.html.
 * $RP_END_LICENSE$
 */

package bckapi

import (
	"git.ovh.iot/redpesk/redpesk-infra/rp-webserver/actions/rpapiv1"
	"github.com/franciscocpg/reflectme"
)

// -------------------------------------------------------------------------------
// Types definition --------------------------------------------------------------
// -------------------------------------------------------------------------------

// rpBCKServiceParam defines a parameter of a service
type RPBCKServiceParam struct {
	Description string      `json:"description"`
	Type        string      `json:"type"`
	Value       interface{} `json:"value"`
}

// rpBCKService defines one service
type RPBCKService struct {
	Summary     string                       `json:"summary"`
	Description string                       `json:"description"`
	Version     string                       `json:"version"`
	Used        bool                         `json:"used"`
	Parameters  map[string]RPBCKServiceParam `json:"parameters"`
}

type RPBCSAutoRebuildSrpmOptions map[string]bool

// rpBCKServices defines several services
type RPBCKServices map[string]*RPBCKService

// -------------------------------------------------------------------------------
// -------------------------------------------------------------------------------

var cachedVersion = rpapiv1.RPVersion{}
var cachedArches = []rpapiv1.RPArch{}

// RPVersionGet allows to get the version of Redpesk server
//
// :param ver: Pointer to RPVersion structure to be filled
// :type ver: *rpapiv1.RPVersion
// :param forceReq: Boolean allowing to force the HTTP request
// :type forceReq: bool
// :returns: error -- an error if something went bad
func (rpAPIInterface *RPBCKAPI) RPVersionGet(ver *rpapiv1.RPVersion, forceReq bool) error {

	// Send the request if the cache is empty or it is forced
	if (cachedVersion.Version == "") || forceReq {
		if err := rpAPIInterface.HTTPCli.Get(versionRoute, &cachedVersion); err != nil {
			return errorInterpreter(err, "")
		}
	}

	reflectme.Copy(&cachedVersion, ver)
	return nil
}

// RPConfigGet allows to get current Redpesk Agent config
//
// :param cfg: Pointer to RPAPIConfig structure to be filled
// :type cfg: *rpapiv1.RPAPIConfig
// :returns: error -- an error if something went bad
func (rpAPIInterface *RPBCKAPI) RPConfigGet(cfg *rpapiv1.RPAPIConfig) error {
	return errorInterpreter(rpAPIInterface.HTTPCli.Get(configRoute, cfg), "")
}

// RPArchGet  allows to get the list of available architectures
// on the Redpesk instance
//
// :param arches: Pointer to RPArches structure to be filled
// :type arches: *rpapiv1.RPArches
// :param forceReq: Boolean allowing to force the HTTP request
// :type forceReq: bool
// :returns: error -- an error if something went bad
func (rpAPIInterface *RPBCKAPI) RPArchGet(arches *[]rpapiv1.RPArch, forceReq bool) error {

	// Send the request if the cache is empty or it is forced
	if len(cachedArches) == 0 || forceReq {
		if err := rpAPIInterface.HTTPCli.Get(archRoute, &cachedArches); err != nil {
			return errorInterpreter(err, "")
		}
	}

	// Copy the arches
	(*arches) = append((*arches), cachedArches...)

	return nil
}

// RPServicesGet allows to get the list of available services
// on the Redpesk instance
//
// :param services: Pointer to RPBCKServices structure to be filled
// :type services: *RPBCKServices
// :returns: error -- an error if something went bad
func (rpAPIInterface *RPBCKAPI) RPServicesGet(services *RPBCKServices) error {
	// Get all the services from redpesk
	return errorInterpreter(rpAPIInterface.HTTPCli.Get(servicesRoute, services), "")
}
