/**
 * Copyright (C) 2020-2023 IoT.bzh Company
 * Contact: https://www.iot.bzh/licensing
 *
 * This file is part of the redpesk-cli project, the redpesk command line tool.
 *
 * $RP_BEGIN_LICENSE$
 * Commercial License Usage
 *  Licensees holding valid commercial IoT.bzh licenses may use this file in
 *  accordance with the commercial license agreement provided with the
 *  Software or, alternatively, in accordance with the terms contained in
 *  a written agreement between you and The IoT.bzh Company. For licensing terms
 *  and conditions see https://www.iot.bzh/terms-conditions. For further
 *  information use the contact form at https://www.iot.bzh/contact.
 *
 * GNU General Public License Usage
 *  Alternatively, this file may be used under the terms of the GNU General
 *  Public license version 3. This license is as published by the Free Software
 *  Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
 *  of this file. Please review the following information to ensure the GNU
 *  General Public License requirements will be met
 *  https://www.gnu.org/licenses/gpl-3.0.html.
 * $RP_END_LICENSE$
 */

package bckapi

import (
	"encoding/json"
	"errors"
	"fmt"
	"sort"
	"strconv"
	"strings"

	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/utils"
	"git.ovh.iot/redpesk/redpesk-infra/rp-webserver/actions/rpapiv1"
)

// Sort projects by Slug
type _PrjByLabel []rpapiv1.RPProject

func (s _PrjByLabel) Len() int           { return len(s) }
func (s _PrjByLabel) Swap(i, j int)      { s[i], s[j] = s[j], s[i] }
func (s _PrjByLabel) Less(i, j int) bool { return s[i].Slug < s[j].Slug }

// ProjectsListGetOfTypes Get the list of existing projects, the type
// of projects to retrieve can be specified in a slice of string
//
// :param prjs: Pointer to the rpapiv1.RPProject structure to fill
// :type prjs: *rpapiv1.RPProjects
// :param types: Types of project to retrieve
// :type types: []string
// :param filter: value of the filter to send in the request
// :type filter: string
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ProjectsListGetOfTypes(prjs *rpapiv1.RPProjects, types []string, filter string) error {

	// Construct the string containing all the types to retrieve, by default the value is "standard"
	typesStr := "standard"

	if len(types) > 0 {
		// Get rid of the default value
		typesStr = ""
		// Construct the type string
		for index, currentType := range types {
			typesStr += currentType
			if index < (len(types) - 1) {
				typesStr += ","
			}
		}
	}

	addQueryParams := make(map[string]string)
	if filter != "" {
		addQueryParams["filters"] = filter
	}

	// Initialize the call-back to save current page of projects
	savingCB := func(tmpObjects []byte) {
		tmpProjects := rpapiv1.RPProjects{}
		// Convert to the right type
		json.Unmarshal(tmpObjects, &tmpProjects)
		// Add the projects in the final object
		prjs.Projects = append(prjs.Projects, tmpProjects.Projects...)
	}

	// Initialize the paging structure that will allow to get all the objects
	pagingMngt := &RPBCKPaging{}

	// Call the function allowing to get all the object
	err := pagingMngt.getAllObjects(rpAPIInterface, projectRoute, typesStr, addQueryParams, savingCB)
	if err != nil {
		return errorInterpreter(err, projectObjectType)
	}

	// Sort the resulting projects
	sort.Sort(_PrjByLabel(prjs.Projects))

	return nil
}

// ProjectsListByTypePage Get the list of existing projects with the
// page number indicated
//
// :param prjs: Pointer to the rpapiv1.RPProject structure to fill
// :type prjs: *rpapiv1.RPProjects
// :param types: Types of project to retrieve
// :type types: []string
// :param page: Page number that the caller want
// :type page: int
// :param perPage: Number of item per page that the caller want
// :type perPage: int
// :param bookmarked: Set to true if only the bookmarked projects are wanted
// :type bookmarked: boolean
// :returns: error -- an error if a problem occurred
func (rpAPIInterface *RPBCKAPI) ProjectsListByTypePage(prjs *rpapiv1.RPProjects, types []string, page, perPage int, bookmarked bool) error {
	// Construct the string containing all the types to retrieve, by default the value is "standard"
	typesStr := "standard"

	if len(types) > 0 {
		// Get rid of the default value
		typesStr = ""
		// Construct the type string
		for index, currentType := range types {
			typesStr += currentType
			if index < (len(types) - 1) {
				typesStr += ","
			}
		}
	}

	// Initialize the query parameters object
	queryParamMap := map[string]string{
		"page":     strconv.Itoa(page),
		"per_page": strconv.Itoa(perPage),
		"type":     typesStr,
	}

	if bookmarked {
		queryParamMap["filters"] = "bookmarked==true"
	}

	// Now, run the get request
	if err := rpAPIInterface.HTTPCli.Get(projectRoute, &prjs, queryParamMap); err != nil {
		return errorInterpreter(err, projectObjectType)
	}

	return nil
}

// ProjectGet Get the information available for a given project
//
// :param id: id of the project wanted
// :type id: string
// :param prj: Pointer to the rpapiv1.RPProjectFull structure to fill
// :type prj: *rpapiv1.RPProjectFull
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ProjectGet(id string, prj *rpapiv1.RPProjectFull) error {
	urlValue := projectRoutePrefix + id

	if err := rpAPIInterface.HTTPCli.Get(urlValue, prj); err != nil {
		return errorInterpreter(err, projectObjectType)
	}

	return nil
}

// ProjectCreate asks the Redpesk backend to create a new project
//
// :param prj: rpapiv1.RPProjectFull structure containing the data to create the project
// :type prj: *rpapiv1.RPProjectFull
// :param ans: rpapiv1.RPProjectFull structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPProjectFull
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ProjectCreate(prj *rpapiv1.RPProjectFull, ans *rpapiv1.RPProjectFull) error {
	var tmpAns *rpapiv1.RPProjectFull

	// Check that the pointer to the structure to send is not nil
	if prj == nil {
		return errors.New(projectCreateNilError)
	}

	if ans == nil {
		tmpAns = &rpapiv1.RPProjectFull{}
	} else {
		tmpAns = ans
	}

	if err := rpAPIInterface.HTTPCli.Post(projectRoute, prj, tmpAns); err != nil {
		return errorInterpreter(err, projectObjectType)
	}

	return nil
}

// ProjectUpdate asks the Redpesk backend to update an existing project
//
// :param projectID: ID of the project to update
// :type projectID: string
// :param prj: rpapiv1.RPProjectFull structure containing the data to create the project
// :type prj: *rpapiv1.RPProjectFull
// :param ans: rpapiv1.RPProjectFull structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPProjectFull
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ProjectUpdate(projectID string, prj *rpapiv1.RPProjectFull, ans *rpapiv1.RPProjectFull) error {
	var tmpAns *rpapiv1.RPProjectFull

	// Check that the pointer to the structure to send is not nil
	if prj == nil {
		return errors.New(projectCreateNilError)
	}

	if ans == nil {
		tmpAns = &rpapiv1.RPProjectFull{}
	} else {
		tmpAns = ans
	}

	if err := rpAPIInterface.HTTPCli.Put(fmt.Sprintf(projectUpdateRoute, projectID), prj, tmpAns); err != nil {
		return errorInterpreter(err, projectObjectType)
	}

	return nil
}

// ProjectChangeBuildType asks the Redpesk backend to change the build type of one or several
// applications for the project
//
// :param projectID: ID of the project to update
// :type projectID: string
// :param changeBldTypeStruct: rpapiv1.RPChangeBuildType structure containing the data needed to change the build type
// :type changeBldTypeStruct: rpapiv1.RPChangeBuildType
// :param ans: rpapiv1.RPProjectFull structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPProjectFull
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ProjectChangeBuildType(projectID string, changeBldTypeStruct rpapiv1.RPChangeBuildType, ans *rpapiv1.RPProjectFull) error {
	var tmpAns *rpapiv1.RPProjectFull

	if ans == nil {
		tmpAns = &rpapiv1.RPProjectFull{}
	} else {
		tmpAns = ans
	}

	// Query parameters to add to the request
	queryParamMap := map[string]string{
		"action": "change_build_type",
	}

	if err := rpAPIInterface.HTTPCli.Put(fmt.Sprintf(projectUpdateRoute, projectID), changeBldTypeStruct, tmpAns, queryParamMap); err != nil {
		return errorInterpreter(err, projectObjectType)
	}

	return nil
}

// ProjectDelete asks the Redpesk backend to delete an existing project
//
// :param prjID: ID of the project to delete
// :type prjID: string
// :param ans: rpapiv1.RPProjectFull structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPProjectFull
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ProjectDelete(prjID string, ans *rpapiv1.RPProjectFull) error {
	var tmpAns *rpapiv1.RPProjectFull

	if ans == nil {
		tmpAns = &rpapiv1.RPProjectFull{}
	} else {
		tmpAns = ans
	}

	// Send the delete request
	if err := rpAPIInterface.HTTPCli.Delete(projectRoutePrefix+prjID, tmpAns); err != nil {
		return errorInterpreter(err, projectObjectType)
	}

	return nil
}

// ProjectBookmark asks the Redpesk backend to bookmark or reset the bookmark of an existing project
//
// :param prjID: ID of the project to bookmark or reset it
// :type prjID: string
// :param ans: rpapiv1.RPProjectBookmark structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPProjectBookmark
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ProjectBookmark(prjID string, ans *rpapiv1.RPProjectBookmark) error {
	var tmpAns *rpapiv1.RPProjectBookmark

	if ans == nil {
		tmpAns = &rpapiv1.RPProjectBookmark{}
	} else {
		tmpAns = ans
	}

	// Send the post request
	if err := rpAPIInterface.HTTPCli.Post(fmt.Sprintf(projectBookmarkRoute, prjID), nil, tmpAns); err != nil {
		return errorInterpreter(err, "")
	}

	return nil
}

// Sort released projects by Slug
type _RelPrjByLabel []rpapiv1.RPReleaseProject

func (s _RelPrjByLabel) Len() int           { return len(s) }
func (s _RelPrjByLabel) Swap(i, j int)      { s[i], s[j] = s[j], s[i] }
func (s _RelPrjByLabel) Less(i, j int) bool { return s[i].Slug < s[j].Slug }

// ListReleasedProject Get the list of existing released projects
//
// :param relPrjs: Pointer to the rpapiv1.RPReleaseProjects structure to fill
// :type relPrjs: *rpapiv1.RPReleaseProjects
// :param filter: value of the filter to send in the request
// :type filter: string
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ListReleasedProject(prjs *rpapiv1.RPReleaseProjects, filter string) error {

	addQueryParams := make(map[string]string)
	if filter != "" {
		addQueryParams["filters"] = filter
	}

	// Initialize the call-back to save current page of released projects
	savingCB := func(tmpObjects []byte) {
		tmpProjects := rpapiv1.RPReleaseProjects{}
		// Convert to the right type
		json.Unmarshal(tmpObjects, &tmpProjects)
		// Add the projects in the final object
		prjs.ReleaseProjects = append(prjs.ReleaseProjects, tmpProjects.ReleaseProjects...)
	}

	// Initialize the paging structure that will allow to get all the objects
	pagingMngt := &RPBCKPaging{}

	// Call the function allowing to get all the object
	err := pagingMngt.getAllObjects(rpAPIInterface, releasedProjectRoute, "", addQueryParams, savingCB)
	if err != nil {
		return errorInterpreter(err, releaseProjectObjectType)
	}

	// Sort the resulting projects
	sort.Sort(_RelPrjByLabel(prjs.ReleaseProjects))

	return nil
}

// ListReleasedProjectByPage Get the list of existing released projects
// but only the requested page
//
// :param relPrjs: Pointer to the rpapiv1.RPReleaseProjects structure to fill
// :type relPrjs: *rpapiv1.RPReleaseProjects
// :param page: Page number that the caller want
// :type page: int
// :param perPage: Number of item per page that the caller want
// :type perPage: int
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ListReleasedProjectByPage(prjs *rpapiv1.RPReleaseProjects, page, perPage int) error {

	// Initialize the query parameters object
	queryParamMap := map[string]string{
		"page":     strconv.Itoa(page),
		"per_page": strconv.Itoa(perPage),
	}

	// Now, run the get request
	if err := rpAPIInterface.HTTPCli.Get(releasedProjectRoute, prjs, queryParamMap); err != nil {
		return errorInterpreter(err, releaseProjectObjectType)
	}

	return nil
}

// ProjectRelease asks the Redpesk backend to release an existing project.
//
// :param projectID: ID of the project to release
// :type projectID: string
// :param prjRelease: rpapiv1.RPReleaseProject is a pointer to a structure containing the details needed to release the project
// :type prjRelease: *rpapiv1.RPReleaseProject
// :param ans: rpapiv1.RPReleaseProjectFull structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPReleaseProjectFull
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ProjectRelease(projectID string, prjRelease *rpapiv1.RPReleaseProject, ans *rpapiv1.RPReleaseProjectFull) error {
	var tmpAns *rpapiv1.RPReleaseProjectFull

	if ans == nil {
		tmpAns = &rpapiv1.RPReleaseProjectFull{}
	} else {
		tmpAns = ans
	}

	if err := rpAPIInterface.HTTPCli.Post(fmt.Sprintf(projectReleaseRoute, projectID), prjRelease, tmpAns); err != nil {
		return errorInterpreter(err, releaseProjectObjectType)
	}

	return nil
}

// GetReleasedProject allows to request the backend the "released project" structure corresponding to a "source" project
// that has been released
//
// :param projectID: ID of the project to release
// :type projectID: string
// :param ans: rpapiv1.RPReleaseProjectFull structure received as an answer of the request.
// :type ans: *rpapiv1.RPReleaseProjectFull
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) GetReleasedProject(projectID string, ans *rpapiv1.RPReleaseProjectFull) error {
	// Get the released project
	if err := rpAPIInterface.HTTPCli.Get(fmt.Sprintf(releasedProjectIDRoute, projectID), ans); err != nil {
		return errorInterpreter(err, releaseProjectObjectType)
	}

	return nil
}

// UpdateReleasedProject asks the Redpesk backend to update an existing released project.
//
// :param projectID: ID of the released project
// :type projectID: string
// :param prjRelease: rpapiv1.RPReleaseProject is a pointer to a structure containing the details needed to update the project
// :type prjRelease: *rpapiv1.RPReleaseProject
// :param ans: rpapiv1.RPReleaseProjectFull structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPReleaseProjectFull
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) UpdateReleasedProject(projectID string, prjRelease *rpapiv1.RPReleaseProject, ans *rpapiv1.RPReleaseProjectFull) error {
	var tmpAns *rpapiv1.RPReleaseProjectFull

	if ans == nil {
		tmpAns = &rpapiv1.RPReleaseProjectFull{}
	} else {
		tmpAns = ans
	}

	if err := rpAPIInterface.HTTPCli.Put(fmt.Sprintf(releasedProjectIDRoute, projectID), prjRelease, tmpAns); err != nil {
		return errorInterpreter(err, releaseProjectObjectType)
	}

	return nil
}

// DeleteReleasedProject asks the Redpesk backend to delete an existing released project.
//
// :param projectID: ID of the released project to delete
// :type projectID: string
// :param ans: rpapiv1.RPReleaseProjectFull structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPProjectFull
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) DeleteReleasedProject(projectID string, ans *rpapiv1.RPReleaseProjectFull) error {
	var tmpAns *rpapiv1.RPReleaseProjectFull

	if ans == nil {
		tmpAns = &rpapiv1.RPReleaseProjectFull{}
	} else {
		tmpAns = ans
	}

	if err := rpAPIInterface.HTTPCli.Delete(fmt.Sprintf(releasedProjectIDRoute, projectID), tmpAns); err != nil {
		return errorInterpreter(err, releaseProjectObjectType)
	}

	return nil
}

// GetReleasedProjectAudience asks the Redpesk backend to return the audience of a released project
//
// :param projectID: ID of the released project
// :type projectID: string
// :param ans: rpapiv1.RPObjectRoles structure received as an answer to the request.
// :type ans: *rpapiv1.RPObjectRoles
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) GetReleasedProjectAudience(projectID string, ans *rpapiv1.RPObjectRoles) error {

	if err := rpAPIInterface.HTTPCli.Get(fmt.Sprintf(releasedProjectAudienceRoute, projectID), ans); err != nil {
		return errorInterpreter(err, releaseProjectObjectType)
	}

	// Finish to create the structure to avoid having "null pointer dereference" panic later
	if ans.IdentityRolesLists == nil {
		ans.IdentityRolesLists = make(map[string]*rpapiv1.RPIdentityRolesLists)
	}
	for _, rolesList := range ans.IdentityRolesLists {
		if rolesList.AvailableRoles == nil {
			rolesList.AvailableRoles = &([]rpapiv1.RPRole{})
		}
		if rolesList.SetRoles == nil {
			rolesList.SetRoles = &([]rpapiv1.RPRole{})
		}
	}
	// -------------------------------------------------------------------------------------

	return nil
}

// UpdateReleasedProjectAudience asks the Redpesk backend to change the audience of a released project
//
// :param projectID: ID of the released project
// :type projectID: string
// :param rolesObj: rpapiv1.RPObjectRoles is a pointer to a structure containing the details about the audience wanted for the project
// :type rolesObj: *rpapiv1.RPObjectRoles
// :param ans: rpapiv1.RPObjectRoles structure received as an answer to the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPObjectRoles
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) UpdateReleasedProjectAudience(projectID string, rolesObj *rpapiv1.RPObjectRoles, ans *rpapiv1.RPObjectRoles) error {
	var tmpAns *rpapiv1.RPObjectRoles

	if ans == nil {
		tmpAns = &rpapiv1.RPObjectRoles{}
	} else {
		tmpAns = ans
	}

	if err := rpAPIInterface.HTTPCli.Put(fmt.Sprintf(releasedProjectAudienceRoute, projectID), rolesObj, tmpAns); err != nil {
		return errorInterpreter(err, releaseProjectObjectType)
	}

	return nil
}

// DeployReleasedProject allows to request the backend to deploy on board(s) a "released project" structure
//
// :param relProjectID: ID of the released project to deploy
// :type relProjectID: string
// :param deployStruct: List of board UUIDs where we want to deploy the project
// :type deployStruct: *rpapiv1.RPDeploy
// :param ans: rpapiv1.RPDeploymentFull structure received as an answer of the request.
// :type ans: *rpapiv1.RPDeploymentFull
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) DeployReleasedProject(relProjectID string, deployStruct *rpapiv1.RPDeploy, ans *rpapiv1.RPDeploymentFull) error {

	// Request the backend to deploy the project on the given boards
	if err := rpAPIInterface.HTTPCli.Post(fmt.Sprintf(releasedProjectDeployRoute, relProjectID), deployStruct, ans); err != nil {
		return errorInterpreter(err, releaseProjectObjectType)
	}

	return nil
}

// ProjectGetBuildConfigFile gets the configuration file needed by the local builder
// to package an application
//
// :param id: id of the project wanted
// :type id: string
// :param distro: Distribution wanted for the build config file
// :type distro: string
// :param arch: Architecture wanted for the build config file
// :type arch: string
// :param file: Pointer to the rpapiv1.RPFile structure to fill
// :type file: *rpapiv1.RPFile_Deprecated
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ProjectGetBuildConfigFile(id, distro, arch string, file *rpapiv1.RPFile_Deprecated) error {

	// Query parameters to add to teh request
	queryParamMap := map[string]string{
		"arch":   arch,
		"distro": distro,
	}

	if err := rpAPIInterface.HTTPCli.Get(fmt.Sprintf(projectBuildConfigRoute, id), file, queryParamMap); err != nil {
		return errorInterpreter(err, "")
	}

	return nil
}

// ProjectExport asks the Redpesk backend to export an existing project
//
// :param prjID: ID of the project to export
// :type prjID: string
// :param ans: rpapiv1.RPExportProjectState structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPExportProjectState
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ProjectExport(prjID string, ans *rpapiv1.RPExportProjectState) error {
	var tmpAns *rpapiv1.RPExportProjectState

	if ans == nil {
		tmpAns = &rpapiv1.RPExportProjectState{}
	} else {
		tmpAns = ans
	}

	// Send the post request
	if err := rpAPIInterface.HTTPCli.Post(fmt.Sprintf(projectExportRoute, prjID), nil, tmpAns); err != nil {
		return errorInterpreter(err, "")
	}

	return nil
}

// ProjectGetExportState asks the Redpesk backend the state of the export on the specified project
//
// :param prjID: ID of the project that has been exported
// :type prjID: string
// :param ans: rpapiv1.RPExportProjectState structure received as an answer of the request
// :type ans: *rpapiv1.RPExportProjectState
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ProjectGetExportState(prjID string, ans *rpapiv1.RPExportProjectState) error {

	// Send the post request
	if err := rpAPIInterface.HTTPCli.Get(fmt.Sprintf(projectExportRoute, prjID), ans); err != nil {
		return errorInterpreter(err, "")
	}

	return nil
}

// ProjectExportDownload downloads the archive corresponding to the "project export" from the backend
//
// :param prjID: ID of the project that has been exported
// :type prjID: string
// :param targetDir: directory where the downloaded file needs to be stored
// :type targetDir: string
// :returns: string, error -- path to the downloaded archive and an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ProjectExportDownload(prjID, targetDir string) (string, error) {
	return rpAPIInterface.backendFileDownload(fmt.Sprintf(projectExportDownloadRoute, prjID), targetDir, "project_export.zip")
}

// ProjectImportUpload uploads the archive of a project to import
//
// :param archivePath: path to the archive to upload to the backend
// :type archivePath: string
// :param ans: pointer to the RPImportProjectState structure to complete with the backend answer
// :type ans: *rpapiv1.RPImportProjectState
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ProjectImportUpload(archivePath string, ans *rpapiv1.RPImportProjectState) error {
	// Request to upload the given file and retrieve the result structure
	err := rpAPIInterface.HTTPCli.UploadFile(archivePath, projectImportUploadRoute, "upload", ans)
	if err != nil {
		return errorInterpreter(err, "")
	}
	return nil
}

// ProjectImport asks the Redpesk backend to import a project that has seen its archive
// uploaded previously
//
// :param prjID: ID of the project to import
// :type prjID: string
// :param ans: rpapiv1.RPImportProjectState structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPImportProjectState
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ProjectImport(prjID string, ans *rpapiv1.RPImportProjectState) error {
	var tmpAns *rpapiv1.RPImportProjectState

	if ans == nil {
		tmpAns = &rpapiv1.RPImportProjectState{}
	} else {
		tmpAns = ans
	}

	// Send the post request
	if err := rpAPIInterface.HTTPCli.Post(fmt.Sprintf(projectImportRoute, prjID), nil, tmpAns); err != nil {
		return errorInterpreter(err, "")
	}

	return nil
}

// ProjectGetImportState asks the Redpesk backend the state of the import on the specified project
//
// :param prjID: ID of the project that has been imported
// :type prjID: string
// :param ans: rpapiv1.RPImportProjectState structure received as an answer of the request
// :type ans: *rpapiv1.RPImportProjectState
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ProjectGetImportState(prjID string, ans *rpapiv1.RPImportProjectState) error {

	// Send the post request
	if err := rpAPIInterface.HTTPCli.Get(fmt.Sprintf(projectImportRoute, prjID), ans); err != nil {
		return errorInterpreter(err, "")
	}

	return nil
}

// ProjectMembersGet Get the list of members in a project
//
// :param prjID: ID of the project from which we want the members
// :type prjID: string
// :param members: Pointer to the rpapiv1.RPMembers structure to fill with members
// :type members: *rpapiv1.RPMembers
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ProjectMembersGet(prjID string, members *rpapiv1.RPMembers) error {

	if err := rpAPIInterface.HTTPCli.Get(fmt.Sprintf(prjMembersRoute, prjID), members); err != nil {
		return errorInterpreter(err, "")
	}

	return nil
}

// ProjectGetMembersToInvite gets the list of members that it is possible
// to invite to a project
//
// :param prjID: ID of the project in which we want to invite a member
// :type prjID: string
// :param invits: Pointer to the rpapiv1.RPMembers structure to fill with possible members
// :type invits: *rpapiv1.RPMembers
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ProjectGetMembersToInvite(prjID string, invits *rpapiv1.RPMembers) error {

	if err := rpAPIInterface.HTTPCli.Get(fmt.Sprintf(prjNewMembersRoute, prjID), invits); err != nil {
		return errorInterpreter(err, "")
	}

	return nil
}

// ProjectGetMembersRoles gets the list of available roles for members in a project
//
// :param prjID: ID of the project where we want to have the roles
// :type prjID: string
// :param roles: Pointer to the rpapiv1.RPRoleProjects structure to fill with possible roles
// :type roles: *rpapiv1.RPRoleProjects
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ProjectGetMembersRoles(prjID string, roles *rpapiv1.RPRoleProjects) error {

	if err := rpAPIInterface.HTTPCli.Get(fmt.Sprintf(prjMembersRolesRoute, prjID), roles); err != nil {
		return errorInterpreter(err, "")
	}

	return nil
}

// ProjectAddMember adds a member in a project
//
// :param prjID: ID of the project where we want to add a member
// :type prjID: string
// :param member: Pointer to the rpapiv1.RPMember structure containing the member to add
// :type member: *rpapiv1.RPMember
// :param ans: Pointer to the rpapiv1.RPMember structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPMember
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ProjectAddMember(prjID string, member *rpapiv1.RPMember, ans *rpapiv1.RPMember) error {
	var tmpAns *rpapiv1.RPMember

	// Check that the pointer to the structure to send is not nil
	if member == nil {
		return errors.New(memberAddNilError)
	}

	if ans == nil {
		tmpAns = &rpapiv1.RPMember{}
	} else {
		tmpAns = ans
	}

	if err := rpAPIInterface.HTTPCli.Post(fmt.Sprintf(prjMembersRoute, prjID), member, tmpAns); err != nil {
		return errorInterpreter(err, "")
	}

	return nil
}

// ProjectUpdateMember updates a member of a project
//
// :param prjID: ID of the targetted project
// :type prjID: string
// :param member: Pointer to the rpapiv1.RPMember structure containing the member to update
// :type member: *rpapiv1.RPMember
// :param ans: Pointer to the rpapiv1.RPMember structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPMember
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ProjectUpdateMember(prjID string, member *rpapiv1.RPMember, ans *rpapiv1.RPMember) error {
	var tmpAns *rpapiv1.RPMember

	// Check that the pointer to the structure to send is not nil
	if member == nil {
		return errors.New(memberAddNilError)
	}

	if ans == nil {
		tmpAns = &rpapiv1.RPMember{}
	} else {
		tmpAns = ans
	}

	if err := rpAPIInterface.HTTPCli.Put(fmt.Sprintf(prjMembersRoute, prjID), member, tmpAns); err != nil {
		return errorInterpreter(err, "")
	}

	return nil
}

// ProjectDeleteMember delete a member from a project
//
// :param prjID: ID of the targetted project
// :type prjID: string
// :param member: Pointer to the rpapiv1.RPMember structure containing the member to delete
// :type member: *rpapiv1.RPMember
// :param ans: Pointer to the rpapiv1.RPMember structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPMember
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ProjectDeleteMember(prjID string, member *rpapiv1.RPMember, ans *rpapiv1.RPMember) error {
	var tmpAns *rpapiv1.RPMember

	queryParamMap := map[string]string{
		"id": member.ID.String(),
	}

	// Check that the pointer to the structure to send is not nil
	if member == nil {
		return errors.New(memberAddNilError)
	}

	if ans == nil {
		tmpAns = &rpapiv1.RPMember{}
	} else {
		tmpAns = ans
	}

	if err := rpAPIInterface.HTTPCli.Delete(fmt.Sprintf(prjMembersRoute, prjID), tmpAns, queryParamMap); err != nil {
		return errorInterpreter(err, "")
	}

	return nil
}

// ProjectMngtImport import a project from Koji to the backend
//
// :param prjName: Name of the project to import from Koji
// :type prjName: string
// :param prjVersion: Version of the project to import
// :type prjVersion: string
// :param prjSubVersion: SubVersion of the project to import
// :type prjSubVersion: string
// :param ans: rpapiv1.RPProjectFull structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPProjectFull
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ProjectMngtImport(prjName, prjVersion, prjSubVersion, prjTag string, body, ans *rpapiv1.RPProjectFull) error {
	var tmpAns *rpapiv1.RPProjectFull

	queryParamMap := map[string]string{
		"name": prjName,
	}

	if prjVersion != "" {
		queryParamMap["version"] = prjVersion
	}

	if prjSubVersion != "" {
		queryParamMap["sub_version"] = prjSubVersion
	}

	if prjTag != "" {
		queryParamMap["tag"] = prjTag
	}

	if ans == nil {
		tmpAns = &rpapiv1.RPProjectFull{}
	} else {
		tmpAns = ans
	}

	if err := rpAPIInterface.HTTPCli.Post(projectMngtImportRoute, body, tmpAns, queryParamMap); err != nil {
		return errorInterpreter(err, "")
	}

	return nil
}

// ProjectMngtRefresh asks the Redpesk backend to refresh the information and
// packages of an already imported project.
//
// :param projectID: ID of the project to refresh
// :type projectID: string
// :param ans: rpapiv1.RPProjectFull structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPProjectFull
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ProjectMngtRefresh(projectID string, ans *rpapiv1.RPProjectFull) error {
	var tmpAns *rpapiv1.RPProjectFull

	if ans == nil {
		tmpAns = &rpapiv1.RPProjectFull{}
	} else {
		tmpAns = ans
	}

	if err := rpAPIInterface.HTTPCli.Put(fmt.Sprintf(projectMngtRefreshRoute, projectID), nil, tmpAns); err != nil {
		return errorInterpreter(err, "")
	}

	return nil
}

// ProjectGetBuildsStats tries to retrieve the builds stats for the given
// project and for the X last builds (if 0, all builds are used)
//
// :param projectID: ID of the project
// :type projectID: string
// :param buildsNb: Number of builds to use for the stats calculation
// :type buildsNb: int
// :param ans: rpapiv1.RPBuildStatsProject structure received as an answer of the request
// :type ans: *rpapiv1.RPBuildStatsProject
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ProjectGetBuildsStats(projectID string, buildsNb int, ans *rpapiv1.RPBuildStatsProject) error {
	// Create query argument structure
	queryParamMap := map[string]string{}
	if buildsNb > 0 {
		queryParamMap["nb-last"] = strconv.Itoa(buildsNb)
	}

	if err := rpAPIInterface.HTTPCli.Get(fmt.Sprintf(projectBuildsStatsRoute, projectID), ans, queryParamMap); err != nil {
		return errorInterpreter(err, projectObjectType)
	}

	return nil
}

// ProjectBuild asks the Redpesk backend to build all the applications or images, or only a
// part of it, of a project
//
// :param prjID: ID of the project to build
// :type prjID: string
// :param bldStartObj: Structure containing the fields needed to start the build
// :type bldStartObj: *rpapiv1.RPBuildStart
// :param ans: buffer to store the structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPBuilds
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ProjectBuild(prjID string, bldStartObj *rpapiv1.RPBuildStart, ans *rpapiv1.RPBuilds) error {
	var tmpAns *rpapiv1.RPBuilds

	if ans == nil {
		tmpAns = &rpapiv1.RPBuilds{}
	} else {
		tmpAns = ans
	}

	if err := rpAPIInterface.HTTPCli.Post(fmt.Sprintf(projectBuildsRoute, prjID), bldStartObj, tmpAns); err != nil {
		return errorInterpreter(err, projectObjectType)
	}

	return nil
}

// ProjectGetTestsStats tries to retrieve the tests stats for the given
// project and for the X last tests (if 0, all builds are used)
//
// :param projectID: ID of the project
// :type projectID: string
// :param testsNb: Number of builds to use for the stats calculation
// :type testsNb: int
// :param ans: rpapiv1.RPTestStatsProject structure received as an answer of the request
// :type ans: *rpapiv1.RPTestStatsProject
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ProjectGetTestsStats(projectID string, testsNb int, ans *rpapiv1.RPTestStatsProject) error {
	// Create query argument structure
	queryParamMap := map[string]string{}
	if testsNb > 0 {
		queryParamMap["nb-last"] = strconv.Itoa(testsNb)
	}

	if err := rpAPIInterface.HTTPCli.Get(fmt.Sprintf(projectTestsStatsRoute, projectID), ans, queryParamMap); err != nil {
		return errorInterpreter(err, projectObjectType)
	}

	return nil
}

// ProjectGetReposInheritance get the repos inheritance file from the backend, this function
// is needed by the "local" v2 command to initialize a package build
//
// :param projectID: ID of the project
// :type projectID: string
// :param fileToWrite: Path to the file to create with the answer
// :type fileToWrite: string
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) ProjectGetReposInheritance(projectID, fileToWrite, baseURL string) error {
	jsonContent := []byte{}

	// Get the URL of the current connection
	baseKbuildURL := strings.Join([]string{baseURL, "kbuild"}, "/")

	// Create query argument structure
	queryParamMap := map[string]string{
		"baseurl":    baseKbuildURL,
		"compo_type": "staging_koji",
		"with_token": "true",
		"inherit":    "true",
	}

	// Send the appropriate GET request
	if err := rpAPIInterface.HTTPCli.HTTPGet(fmt.Sprintf(projectRepoInheritanceRoute, projectID), &jsonContent, queryParamMap); err != nil {
		return errorInterpreter(err, projectObjectType)
	}

	// Write the file
	return utils.WriteToFile(fileToWrite, string(jsonContent))
}
