/**
 * Copyright (C) 2020-2023 IoT.bzh Company
 * Contact: https://www.iot.bzh/licensing
 *
 * This file is part of the redpesk-cli project, the redpesk command line tool.
 *
 * $RP_BEGIN_LICENSE$
 * Commercial License Usage
 *  Licensees holding valid commercial IoT.bzh licenses may use this file in
 *  accordance with the commercial license agreement provided with the
 *  Software or, alternatively, in accordance with the terms contained in
 *  a written agreement between you and The IoT.bzh Company. For licensing terms
 *  and conditions see https://www.iot.bzh/terms-conditions. For further
 *  information use the contact form at https://www.iot.bzh/contact.
 *
 * GNU General Public License Usage
 *  Alternatively, this file may be used under the terms of the GNU General
 *  Public license version 3. This license is as published by the Free Software
 *  Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
 *  of this file. Please review the following information to ensure the GNU
 *  General Public License requirements will be met
 *  https://www.gnu.org/licenses/gpl-3.0.html.
 * $RP_END_LICENSE$
 */

package bckapi

import (
	"encoding/json"
	"errors"
	"fmt"
	"sort"
	"strconv"

	"git.ovh.iot/redpesk/redpesk-infra/rp-webserver/actions/rpapiv1"
)

// Sort users by Slug
type _TeamByLabel []rpapiv1.RPTeam

func (u _TeamByLabel) Len() int           { return len(u) }
func (u _TeamByLabel) Swap(i, j int)      { u[i], u[j] = u[j], u[i] }
func (u _TeamByLabel) Less(i, j int) bool { return u[i].Slug < u[j].Slug }

// TeamsGetList allows to get the list of all teams available on Redpesk
//
// :param teams: Pointer to the rpapiv1.RPTeams structure to fill
// :type teams: *rpapiv1.RPTeams
// :param filter: value of the filter to send in the request
// :type filter: string
// :returns: error -- an error if something went bad
func (rpAPIInterface *RPBCKAPI) TeamsGetList(teams *rpapiv1.RPTeams, filter string) error {

	addQueryParams := make(map[string]string)
	if filter != "" {
		addQueryParams["filters"] = filter
	}

	// Initialize the call-back to save current page of teams
	savingCB := func(tmpObjects []byte) {
		tmpTeams := rpapiv1.RPTeams{}
		// Convert to the right type
		json.Unmarshal(tmpObjects, &tmpTeams)
		// Add the applications in the final object
		teams.Teams = append(teams.Teams, tmpTeams.Teams...)
	}

	// Initialize the paging structure that will allow to get all the objects
	pagingMngt := &RPBCKPaging{}

	// Call the function allowing to get all the object
	err := pagingMngt.getAllObjects(rpAPIInterface, teamsRoute, "", addQueryParams, savingCB)
	if err != nil {
		return errorInterpreter(err, teamObjectType)
	}

	sort.Sort(_TeamByLabel(teams.Teams))

	return nil
}

// TeamsGetListByPage gets the list of teams on the factory
// but only one page of it
//
// :param teams: Pointer to the rpapiv1.RPTeams structure to fill
// :type teams: *rpapiv1.RPTeams
// :param page: Page number that the caller want
// :type page: int
// :param perPage: Number of item per page that the caller want
// :type perPage: int
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) TeamsGetListByPage(teams *rpapiv1.RPTeams, page, perPage int) error {

	// Initialize the query parameters object
	queryParamMap := map[string]string{
		"page":     strconv.Itoa(page),
		"per_page": strconv.Itoa(perPage),
	}

	// Now, run the get request
	if err := rpAPIInterface.HTTPCli.Get(teamsRoute, teams, queryParamMap); err != nil {
		return errorInterpreter(err, teamObjectType)
	}

	return nil
}

// TeamGet allows to get info about a team
//
// :param teamID: ID of the team to look for
// :type teamID: string
// :param team: Pointer to the rpapiv1.RPTeam structure to fill
// :type team: *rpapiv1.RPTeam
// :returns: error -- an error if something went bad
func (rpAPIInterface *RPBCKAPI) TeamGet(teamID string, team *rpapiv1.RPTeam) error {
	// Send the appropriate GET request
	if err := rpAPIInterface.HTTPCli.Get(fmt.Sprintf(teamsIDRoute, teamID), team); err != nil {
		return errorInterpreter(err, teamObjectType)
	}

	return nil
}

// TeamCreate allows to request the backend the creation of a new team
//
// :param team: Pointer to the rpapiv1.RPTeam structure containing the data to create the team
// :type team: *rpapiv1.RPTeam
// :param ans: Pointer to the rpapiv1.RPTeam structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPTeam
// :returns: error -- an error if something went bad
func (rpAPIInterface *RPBCKAPI) TeamCreate(team *rpapiv1.RPTeam, ans *rpapiv1.RPTeam) error {

	var tmpAns *rpapiv1.RPTeam

	// Check that the pointer to the structure to send is not nil
	if team == nil {
		return errors.New(teamCreateNilError)
	}

	if ans == nil {
		tmpAns = &rpapiv1.RPTeam{}
	} else {
		tmpAns = ans
	}

	if err := rpAPIInterface.HTTPCli.Post(teamsRoute, team, tmpAns); err != nil {
		return errorInterpreter(err, teamObjectType)
	}

	return nil
}

// TeamUpdate allows to update an existing team in the Redpesk instance
//
// :param team: Pointer to the rpapiv1.RPTeam structure containing the data allowing to update the team
// :type team: *rpapiv1.RPTeam
// :param ans: Pointer to the rpapiv1.RPTeam structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPTeam
// :returns: error -- an error if something went bad
func (rpAPIInterface *RPBCKAPI) TeamUpdate(team *rpapiv1.RPTeam, ans *rpapiv1.RPTeam) error {

	var tmpAns *rpapiv1.RPTeam

	// Check that the pointer to the structure to send is not nil
	if team == nil {
		return errors.New(teamCreateNilError)
	}

	if ans == nil {
		tmpAns = &rpapiv1.RPTeam{}
	} else {
		tmpAns = ans
	}

	if err := rpAPIInterface.HTTPCli.Put(fmt.Sprintf(teamsIDRoute, team.ID.String()), team, tmpAns); err != nil {
		return errorInterpreter(err, teamObjectType)
	}

	return nil
}

// TeamGetMembersToInvite gets the list of members that it is possible
// to invite in a team
//
// :param teamID: ID of the team in which we want to invite a member
// :type teamID: string
// :param invits: Pointer to the rpapiv1.RPMembers structure to fill with possible members
// :type invits: *rpapiv1.RPMembers
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) TeamGetMembersToInvite(teamID string, invits *rpapiv1.RPMembers) error {

	if err := rpAPIInterface.HTTPCli.Get(fmt.Sprintf(teamsNewMembersListRoute, teamID), invits); err != nil {
		return errorInterpreter(err, "")
	}

	return nil
}

// TeamRequestInvitation allows to ask the backend to send an invitation to a user
//
// :param teamID: ID of the team targetted by the invitation
// :type teamID: string
// :param member: Pointer to the rpapiv1.RPMember structure to invite to the team
// :type member: *rpapiv1.RPMember
// :param ans: Pointer to the rpapiv1.RPInvitResult structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPInvitResult
// :returns: error -- an error if something went bad
func (rpAPIInterface *RPBCKAPI) TeamRequestInvitation(teamID string, member *rpapiv1.RPMember, ans *rpapiv1.RPInvitResult) error {

	var tmpAns *rpapiv1.RPInvitResult

	// Check that the pointer to the structure to send is not nil
	if member == nil {
		return errors.New(teamInviteNilError)
	}

	if ans == nil {
		tmpAns = &rpapiv1.RPInvitResult{}
	} else {
		tmpAns = ans
	}

	if err := rpAPIInterface.HTTPCli.Post(fmt.Sprintf(teamsMemberInvitRoute, teamID), member, tmpAns); err != nil {
		return errorInterpreter(err, "")
	}

	return nil
}

// TeamDeleteMember allows to delete a member from a team
//
// :param teamUUID: UUID of the team where a member needs to be deleted
// :type teamUUID: string
// :param memberUUID: UUID of the member to delete
// :type memberUUID: string
// :param ans: Pointer to the rpapiv1.RPTeam structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :returns: error -- an error if something went bad
func (rpAPIInterface *RPBCKAPI) TeamDeleteMember(teamUUID string, memberUUID string, ans *rpapiv1.RPTeam) error {

	var tmpAns *rpapiv1.RPTeam

	queryParamMap := map[string]string{
		"idmb": memberUUID,
	}

	if ans == nil {
		tmpAns = &rpapiv1.RPTeam{}
	} else {
		tmpAns = ans
	}

	if err := rpAPIInterface.HTTPCli.Delete(fmt.Sprintf(teamsMemberRoute, teamUUID), tmpAns, queryParamMap); err != nil {
		return errorInterpreter(err, "")
	}

	return nil
}
