/**
 * Copyright (C) 2023 IoT.bzh Company
 * Contact: https://www.iot.bzh/licensing
 *
 * This file is part of the redpesk-cli project, the redpesk command line tool.
 *
 * $RP_BEGIN_LICENSE$
 * Commercial License Usage
 *  Licensees holding valid commercial IoT.bzh licenses may use this file in
 *  accordance with the commercial license agreement provided with the
 *  Software or, alternatively, in accordance with the terms contained in
 *  a written agreement between you and The IoT.bzh Company. For licensing terms
 *  and conditions see https://www.iot.bzh/terms-conditions. For further
 *  information use the contact form at https://www.iot.bzh/contact.
 *
 * GNU General Public License Usage
 *  Alternatively, this file may be used under the terms of the GNU General
 *  Public license version 3. This license is as published by the Free Software
 *  Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
 *  of this file. Please review the following information to ensure the GNU
 *  General Public License requirements will be met
 *  https://www.gnu.org/licenses/gpl-3.0.html.
 * $RP_END_LICENSE$
 */

package bckapi

import (
	"fmt"

	"git.ovh.iot/redpesk/redpesk-infra/rp-webserver/actions/rpapiv1"
)

// -------------------------------------------------
// Types definition --------------------------------
// -------------------------------------------------

// -------------------------------------------------
// -------------------------------------------------

// GenericTest asks the Redpesk backend to test an existing build
//
// :param buildID: Specific build of the application to test
// :type buildID: string
// :param testReq: Body of the test request
// :type testReq: *rpapiv1.RPTestStart
// :param ans: buffer to store the structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPTest
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) GenericTest(buildID string, testReq *rpapiv1.RPTestStart, ans *rpapiv1.RPTest) error {
	var tmpAns *rpapiv1.RPTest

	if ans == nil {
		tmpAns = &rpapiv1.RPTest{}
	} else {
		tmpAns = ans
	}

	if err := rpAPIInterface.HTTPCli.Post(fmt.Sprintf(appTestRoute, buildID), testReq, tmpAns); err != nil {
		return errorInterpreter(err, appObjectType+"/"+imageObjectType)
	}

	return nil
}

// GenericTestGet gets a full test structure from its ID
//
// :param test: Pointer to the rpapiv1.RPTest structure to fill
// :type test: *rpapiv1.RPTest
// :param tstID: test ID
// :type tstID: string
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) GenericTestGet(test *rpapiv1.RPTest, tstID string) error {
	// Send the appropriate GET request
	if err := rpAPIInterface.HTTPCli.Get(fmt.Sprintf(testGetRoute, tstID), test); err != nil {
		return errorInterpreter(err, testObjectType)
	}

	return nil
}

// GenericTestGetLogsFilesList gets the list of logs files available for a test
//
// :param testID: test ID
// :type testID: string
// :returns: string, error -- the content of the file and an error if something went wrong
func (rpAPIInterface *RPBCKAPI) GenericTestGetLogsFilesList(testID string, filesList *rpapiv1.RPFileList) error {

	// Send the appropriate GET request
	if err := rpAPIInterface.HTTPCli.Get(fmt.Sprintf(testLogsFilesListRoute, testID), filesList); err != nil {
		return errorInterpreter(err, testObjectType)
	}

	return nil
}

// GenericTestGetLogsFile download the logs file from the backend and return an array of
// bytes with the content
//
// :param testID: test ID
// :type testID: string
// :param filename: Name of the file wanted
// :type filename: string
// :returns: string, error -- the content of the file and an error if something went wrong
func (rpAPIInterface *RPBCKAPI) GenericTestGetLogsFile(testID, filename string) (string, error) {
	// Create the empty byte array that will contain the result
	resultSlice := []byte{}

	// Send the appropriate GET request
	if err := rpAPIInterface.HTTPCli.HTTPGet(fmt.Sprintf(testLogsFileRoute, testID, filename), &resultSlice); err != nil {
		return "", errorInterpreter(err, testObjectType)
	}

	return string(resultSlice), nil
}

// GenericTestCancel allows to cancel a test currently running
//
// :param testID: test ID
// :type testID: string
// :param ans: buffer to store the structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPTest
// :returns: error -- an error if something went wrong
func (rpAPIInterface *RPBCKAPI) GenericTestCancel(testID string, ans *rpapiv1.RPTest) error {

	var tmpAns *rpapiv1.RPTest

	if ans == nil {
		tmpAns = &rpapiv1.RPTest{}
	} else {
		tmpAns = ans
	}

	// Send the appropriate POST request
	if err := rpAPIInterface.HTTPCli.Post(fmt.Sprintf(testCancelRoute, testID), nil, tmpAns); err != nil {
		return errorInterpreter(err, testObjectType)
	}

	return nil
}
