/**
 * Copyright (C) 2020-2023 IoT.bzh Company
 * Contact: https://www.iot.bzh/licensing
 *
 * This file is part of the redpesk-cli project, the redpesk command line tool.
 *
 * $RP_BEGIN_LICENSE$
 * Commercial License Usage
 *  Licensees holding valid commercial IoT.bzh licenses may use this file in
 *  accordance with the commercial license agreement provided with the
 *  Software or, alternatively, in accordance with the terms contained in
 *  a written agreement between you and The IoT.bzh Company. For licensing terms
 *  and conditions see https://www.iot.bzh/terms-conditions. For further
 *  information use the contact form at https://www.iot.bzh/contact.
 *
 * GNU General Public License Usage
 *  Alternatively, this file may be used under the terms of the GNU General
 *  Public license version 3. This license is as published by the Free Software
 *  Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
 *  of this file. Please review the following information to ensure the GNU
 *  General Public License requirements will be met
 *  https://www.gnu.org/licenses/gpl-3.0.html.
 * $RP_END_LICENSE$
 */

package bckapi

import (
	"encoding/json"
	"errors"
	"fmt"
	"sort"
	"strconv"

	"git.ovh.iot/redpesk/redpesk-infra/rp-webserver/actions/rpapiv1"
)

// Sort users full by Slug
type _UsrFullByLabel []rpapiv1.RPUserFull

func (u _UsrFullByLabel) Len() int           { return len(u) }
func (u _UsrFullByLabel) Swap(i, j int)      { u[i], u[j] = u[j], u[i] }
func (u _UsrFullByLabel) Less(i, j int) bool { return u[i].Pseudo < u[j].Pseudo }

// UsersFullGetList allows to get the list of all users on Redpesk
// with all their personal data (reserved for admin user)
//
// :param users: Pointer to the rpapiv1.RPUsersFull structure to fill
// :type users: *rpapiv1.RPUsersFull
// :returns: error -- an error if something went bad
func (rpAPIInterface *RPBCKAPI) UsersFullGetList(users *rpapiv1.RPUsersFull) error {

	// Initialize the call-back to save current page of usersFull
	savingCB := func(tmpObjects []byte) {
		tmpUsers := rpapiv1.RPUsersFull{}
		// Convert to the right type
		json.Unmarshal(tmpObjects, &tmpUsers)
		// Add the applications in the final object
		users.Users = append(users.Users, tmpUsers.Users...)
	}

	// Initialize the paging structure that will allow to get all the objects
	pagingMngt := &RPBCKPaging{}

	// Call the function allowing to get all the object
	err := pagingMngt.getAllObjects(rpAPIInterface, usersRoute, "", nil, savingCB)
	if err != nil {
		return errorInterpreter(err, userObjectType)
	}

	sort.Sort(_UsrFullByLabel(users.Users))

	return nil
}

// UsersFullGetListByPage gets the list of users on the factory
// but only one page of it
//
// :param users: Pointer to the rpapiv1.RPUsersFull structure to fill
// :type users: *rpapiv1.RPUsersFull
// :param page: Page number that the caller want
// :type page: int
// :param perPage: Number of item per page that the caller want
// :type perPage: int
// :returns: error -- an error if a problem has occurred
func (rpAPIInterface *RPBCKAPI) UsersFullGetListByPage(users *rpapiv1.RPUsersFull, page, perPage int) error {

	// Initialize the query parameters object
	queryParamMap := map[string]string{
		"page":     strconv.Itoa(page),
		"per_page": strconv.Itoa(perPage),
	}

	// Now, run the get request
	if err := rpAPIInterface.HTTPCli.Get(usersRoute, users, queryParamMap); err != nil {
		return errorInterpreter(err, userObjectType)
	}

	return nil
}

// UsersGetMe allows to get the information about the current user on Redpesk
//
// :param user: Pointer to the rpapiv1.RPUserFull structure to fill
// :type user: *rpapiv1.RPUserFull
// :returns: error -- an error if something went bad
func (rpAPIInterface *RPBCKAPI) UsersGetMe(user *rpapiv1.RPUserFull) error {
	// Send the appropriate GET request
	if err := rpAPIInterface.HTTPCli.Get(usersMeRoute, user); err != nil {
		return errorInterpreter(err, userObjectType)
	}

	return nil
}

// UserCreate allows to create a new user in the Redpesk instance
//
// :param user: Pointer to the rpapiv1.RPUser structure containing the data to create the user
// :type user: *rpapiv1.RPUserFull
// :param ans: Pointer to the rpapiv1.RPUser structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPUserFull
// :returns: error -- an error if something went bad
func (rpAPIInterface *RPBCKAPI) UserCreate(user *rpapiv1.RPUserFull, ans *rpapiv1.RPUserFull) error {

	var tmpAns *rpapiv1.RPUserFull

	// Check that the pointer to the structure to send is not nil
	if user == nil {
		return errors.New(userCreateNilError)
	}

	if ans == nil {
		tmpAns = &rpapiv1.RPUserFull{}
	} else {
		tmpAns = ans
	}

	if err := rpAPIInterface.HTTPCli.Post(usersRoute, user, tmpAns); err != nil {
		return errorInterpreter(err, userObjectType)
	}

	return nil
}

// UserUpdate allows to update an existing user in the Redpesk instance
//
// :param user: Pointer to the rpapiv1.RPUserFull structure containing the data allowing to update the user
// :type user: *rpapiv1.RPUserFull
// :param ans: Pointer to the rpapiv1.RPUserFull structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPUserFull
// :returns: error -- an error if something went bad
func (rpAPIInterface *RPBCKAPI) UserUpdate(user *rpapiv1.RPUserFull, ans *rpapiv1.RPUserFull) error {

	var tmpAns *rpapiv1.RPUserFull

	// Check that the pointer to the structure to send is not nil
	if user == nil {
		return errors.New(userCreateNilError)
	}

	if ans == nil {
		tmpAns = &rpapiv1.RPUserFull{}
	} else {
		tmpAns = ans
	}

	if err := rpAPIInterface.HTTPCli.Put(fmt.Sprintf(usersIDRoute, user.ID), user, tmpAns); err != nil {
		return errorInterpreter(err, userObjectType)
	}

	return nil
}

// UserUpdateMe allows to update the current user info in the Redpesk instance
//
// :param user: Pointer to the rpapiv1.RPUserFull structure containing the data to allowing to update the user
// :type user: *rpapiv1.RPUserFull
// :param ans: Pointer to the rpapiv1.RPUserFull structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPUserFull
// :returns: error -- an error if something went bad
func (rpAPIInterface *RPBCKAPI) UserUpdateMe(user *rpapiv1.RPUserFull, ans *rpapiv1.RPUserFull) error {

	var tmpAns *rpapiv1.RPUserFull

	// Check that the pointer to the structure to send is not nil
	if user == nil {
		return errors.New(userCreateNilError)
	}

	if ans == nil {
		tmpAns = &rpapiv1.RPUserFull{}
	} else {
		tmpAns = ans
	}

	if err := rpAPIInterface.HTTPCli.Put(usersMeRoute, user, tmpAns); err != nil {
		return errorInterpreter(err, userObjectType)
	}

	return nil
}

// UserGetMyRole allows to get the role of the current user
//
// :param roles: Pointer to the rpapiv1.RPRoles structure to fill
// :type roles: *rpapiv1.RPUsers
// :returns: error -- an error if something went bad
func (rpAPIInterface *RPBCKAPI) UserGetMyRole(roles *rpapiv1.RPRoles) error {

	// Send the appropriate GET request
	if err := rpAPIInterface.HTTPCli.Get(myRolesRoute, roles); err != nil {
		return errorInterpreter(err, "")
	}

	return nil
}

// UserDisableEnable allows to disbale or enable a given user, depending of its previous state
//
// :param userUUID: UUID of the user to disable or enable as a string
// :type userUUID: string
// :param ans: Pointer to the rpapiv1.RPUserFull structure received as an answer of the request. It can be nil
// if the user do not want the answer's body
// :type ans: *rpapiv1.RPUserFull
// :returns: error -- an error if something went bad
func (rpAPIInterface *RPBCKAPI) UserDisableEnable(userUUID string, ans *rpapiv1.RPUserFull) error {

	var tmpAns *rpapiv1.RPUserFull

	if ans == nil {
		tmpAns = &rpapiv1.RPUserFull{}
	} else {
		tmpAns = ans
	}

	if err := rpAPIInterface.HTTPCli.Put(fmt.Sprintf(usersEnableDisable, userUUID), nil, tmpAns); err != nil {
		return errorInterpreter(err, "")
	}

	return nil
}
