/**
 * Copyright (C) 2021-2023 IoT.bzh Company
 * Contact: https://www.iot.bzh/licensing
 *
 * This file is part of the redpesk-cli project, the redpesk command line tool.
 *
 * $RP_BEGIN_LICENSE$
 * Commercial License Usage
 *  Licensees holding valid commercial IoT.bzh licenses may use this file in
 *  accordance with the commercial license agreement provided with the
 *  Software or, alternatively, in accordance with the terms contained in
 *  a written agreement between you and The IoT.bzh Company. For licensing terms
 *  and conditions see https://www.iot.bzh/terms-conditions. For further
 *  information use the contact form at https://www.iot.bzh/contact.
 *
 * GNU General Public License Usage
 *  Alternatively, this file may be used under the terms of the GNU General
 *  Public license version 3. This license is as published by the Free Software
 *  Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
 *  of this file. Please review the following information to ensure the GNU
 *  General Public License requirements will be met
 *  https://www.gnu.org/licenses/gpl-3.0.html.
 * $RP_END_LICENSE$
 */

package ladapi

import (
	"fmt"

	"git.ovh.iot/redpesk/redpesk-infra/rp-golang-common/rphttp"
	"github.com/sirupsen/logrus"
)

// -------------------------------------------------------------------------------
// Constants definition ----------------------------------------------------------
// -------------------------------------------------------------------------------

// -------------------------------------------------------------------------------
// -------------------------------------------------------------------------------

// -------------------------------------------------------------------------------
// Types definition --------------------------------------------------------------
// -------------------------------------------------------------------------------

// RPLADAPI definition
type RPLADAPI struct {
	HTTPCli *rphttp.HTTPClient
	Log     *logrus.Logger
}

// RPLADWinSize represents the size of the pseudo terminal used in Redpesk LAD
type RPLADWinSize struct {
	WinSizeRows int `json:"win_size_rows"`
	WinSizeCol  int `json:"win_size_columns"`
}

// RPLADMainCmdParamsMainCmd is a structure containing the main parameters to run
// a command in Redpesk LAD: working directory, and the window size (rows and columns)
type RPLADMainCmdParams struct {
	RPLADWinSize
	WorkingDir string `json:"working_directory"`
}

// RPLADRunnerTicket: ticket received from Redpesk LAD when request to run a command
type RPLADRunnerTicket struct {
	ID string `json:"id"`
}

// RPLADVersion is a structure containing the version of the server
type RPLADVersion struct {
	Version     string `json:"version"`
	FullVersion string `json:"full_version"`
}

// RPLADEvent is a structure containing an event
type RPLADEvent struct {
	Event string `json:"event"`
}

// -------------------------------------------------------------------------------
// -------------------------------------------------------------------------------

// -------------------------------------------------------------------------------
// Functions definition ----------------------------------------------------------
// -------------------------------------------------------------------------------

// GetLogger allows to get the logget used by the API
//
// :returns: *logrus.Logger -- The logger used
func (ladAPI *RPLADAPI) GetLogger() *logrus.Logger {
	return ladAPI.Log
}

// GetLogPrefix returns the prefix to use in the logs
//
// :returns: string -- The prefix to use
func (bckAPI *RPLADAPI) GetLogPrefix() string {
	return ladLogPrefix
}

// GetURLPrefix returns the prefix to use in the API URLs
//
// :returns: string -- The prefix to use
func (ladAPI *RPLADAPI) GetURLPrefix() string {
	return urlPrefix
}

// GetCookieKey returns the key used to pass cookie to the
// server
//
// :returns: string -- The key to use
func (ladAPI *RPLADAPI) GetCookieKey() string {
	return ladCookieKey
}

// SetHTTPClient save a HTTPClient pointer as its client used to
// execute HTTP requests
func (ladAPI *RPLADAPI) SetHTTPClient(httpClient *rphttp.HTTPClient) {
	// Save the new HTTPClient
	ladAPI.HTTPCli = httpClient
}

// GetHTTPClient returns the HTTP client currently used by the API
//
// :returns: *rphttp.HTTPClient -- The HTTP client used
func (ladAPI *RPLADAPI) GetHTTPClient() *rphttp.HTTPClient {
	return ladAPI.HTTPCli
}

// FillAuthorizationFields is called to fill the Authorization
// fields in the HTTPClientConfig. These fields will be used in the
// header of the requests sent to the server
//
// :param token: Access token furnished by the user
// :type token: string
// :param config: Configuration to complete
// :type config: *rphttp.HTTPClientConfig
func (ladAPI *RPLADAPI) FillAuthorizationFields(token string, config *rphttp.HTTPClientConfig) {
	// No authorization needed for the moment

	// FIXME when http client will be a bit less dumb
	config.HeaderAPIKeyName = authorizationHeader
	// Set the authorization value key
	config.Apikey = "RPCLI"
}

// InitiateAuthentication is called right before connecting
// the websocket. If some additional authentication is required
// by the server (additionally to the token passed in the header),
// like a verb to call or something else, this is here that
// it can be done.
//
// :returns: error -- An error if something went wrong
func (ladAPI *RPLADAPI) InitiateAuthentication() error {
	return nil
}

// GetVersion returns the version of the server as a string
//
// :returns: string -- The key to use
func (ladAPI *RPLADAPI) GetVersion() (string, error) {
	// Initialize the result
	verResult := ""

	// Structure containing the version
	versionStruct := &RPLADVersion{}

	// Request the version to the server
	if err := ladAPI.HTTPCli.Get(versionRoute, versionStruct); err != nil {
		return verResult, err
	}

	// Create the version string
	verResult = fmt.Sprintf("Version: %s, Full version: %s", versionStruct.Version, versionStruct.FullVersion)

	return verResult, nil
}

// RegisterToEvent allows to register the client to an event from the server
//
// :returns: error -- an error if a problem occured, nil in the other case
func (ladAPI *RPLADAPI) RegisterToEvent(evt string) error {
	// Create the structure used to register to an event
	evtStruct := RPLADEvent{
		Event: evt,
	}

	// Do a request to the server
	if err := ladAPI.HTTPCli.Post(subscribeEvtRoute, evtStruct, nil); err != nil {
		return err
	}

	return nil
}

// UnregisterFromEvent allows to unregister the client from an event
//
// :returns: error -- an error if a problem occured, nil in the other case
func (ladAPI *RPLADAPI) UnregisterFromEvent(evt string) error {
	// Create the structure used to register to an event
	evtStruct := RPLADEvent{
		Event: evt,
	}

	// Do a request to the server
	if err := ladAPI.HTTPCli.Post(unsubscribeEvtRoute, evtStruct, nil); err != nil {
		return err
	}

	return nil
}

// -------------------------------------------------------------------------------
// -------------------------------------------------------------------------------

// -------------------------------------------------------------------------------
// Private functions definition --------------------------------------------------
// -------------------------------------------------------------------------------

// -------------------------------------------------------------------------------
// -------------------------------------------------------------------------------
