/**
 * Copyright (C) 2021-2023 IoT.bzh Company
 * Contact: https://www.iot.bzh/licensing
 *
 * This file is part of the redpesk-cli project, the redpesk command line tool.
 *
 * $RP_BEGIN_LICENSE$
 * Commercial License Usage
 *  Licensees holding valid commercial IoT.bzh licenses may use this file in
 *  accordance with the commercial license agreement provided with the
 *  Software or, alternatively, in accordance with the terms contained in
 *  a written agreement between you and The IoT.bzh Company. For licensing terms
 *  and conditions see https://www.iot.bzh/terms-conditions. For further
 *  information use the contact form at https://www.iot.bzh/contact.
 *
 * GNU General Public License Usage
 *  Alternatively, this file may be used under the terms of the GNU General
 *  Public license version 3. This license is as published by the Free Software
 *  Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
 *  of this file. Please review the following information to ensure the GNU
 *  General Public License requirements will be met
 *  https://www.gnu.org/licenses/gpl-3.0.html.
 * $RP_END_LICENSE$
 */

package ladapi

import (
	"errors"
	"path"
	"strings"
)

// -------------------------------------------------------------------------------
// Types definition --------------------------------------------------------------
// -------------------------------------------------------------------------------

// RPLADIotpkgCmdStruct is a structure containing all the parameters necessary to run
// an iotpkg subcommand
type RPLADIotpkgCmdStruct struct {
	RPLADMainCmdParams
	IotpkgFlags []string `json:"iotpkg_flags"`
	SubcmdFlags []string `json:"subcommand_flags"`
}

// -------------------------------------------------------------------------------
// -------------------------------------------------------------------------------

// IotpkgServicePost runs `iotpkg service` in the local builder
//
// :param cmd: Pointer to the RPLADIotpkgCmdStruct structure containing the command parameters
// :type cmd: *RPLADIotpkgCmdStruct
// :param runnerTicket: Pointer to the RPLADRunnerTicket structure to fill with the answer
// :type runnerTicket: *RPLADRunnerTicket
// :returns: error -- an error if a problem has occurred
func (ladAPI *RPLADAPI) IotpkgServicePost(cmd *RPLADIotpkgCmdStruct, runnerTicket *RPLADRunnerTicket) error {
	return ladAPI.iotpkgMainPost(iotpkgServiceRoute, cmd, runnerTicket)
}

// IotpkgMetadataPost runs `iotpkg metadata` in the local builder
//
// :param cmd: Pointer to the RPLADIotpkgCmdStruct structure containing the command parameters
// :type cmd: *RPLADIotpkgCmdStruct
// :param runnerTicket: Pointer to the RPLADRunnerTicket structure to fill with the answer
// :type runnerTicket: *RPLADRunnerTicket
// :returns: error -- an error if a problem has occurred
func (ladAPI *RPLADAPI) IotpkgMetadataPost(cmd *RPLADIotpkgCmdStruct, runnerTicket *RPLADRunnerTicket) error {
	return ladAPI.iotpkgMainPost(iotpkgMetadataRoute, cmd, runnerTicket)
}

// IotpkgInitPost runs `iotpkg init` in the local builder
//
// :param cmd: Pointer to the RPLADIotpkgCmdStruct structure containing the command parameters
// :type cmd: *RPLADIotpkgCmdStruct
// :param runnerTicket: Pointer to the RPLADRunnerTicket structure to fill with the answer
// :type runnerTicket: *RPLADRunnerTicket
// :returns: error -- an error if a problem has occurred
func (ladAPI *RPLADAPI) IotpkgInitPost(cmd *RPLADIotpkgCmdStruct, runnerTicket *RPLADRunnerTicket) error {
	return ladAPI.iotpkgMainPost(iotpkgInitRoute, cmd, runnerTicket)
}

// IotpkgRPBuildPost runs `iotpkg rpbuild` in the local builder
//
// :param cmd: Pointer to the RPLADIotpkgCmdStruct structure containing the command parameters
// :type cmd: *RPLADIotpkgCmdStruct
// :param runnerTicket: Pointer to the RPLADRunnerTicket structure to fill with the answer
// :type runnerTicket: *RPLADRunnerTicket
// :returns: error -- an error if a problem has occurred
func (ladAPI *RPLADAPI) IotpkgRPBuildPost(cmd *RPLADIotpkgCmdStruct, runnerTicket *RPLADRunnerTicket) error {
	return ladAPI.iotpkgMainPost(iotpkgRPBuildRoute, cmd, runnerTicket)
}

// IotpkgCleanMockPost runs `iotpkg cleanmock` in the local builder
//
// :param cmd: Pointer to the RPLADIotpkgCmdStruct structure containing the command parameters
// :type cmd: *RPLADIotpkgCmdStruct
// :param runnerTicket: Pointer to the RPLADRunnerTicket structure to fill with the answer
// :type runnerTicket: *RPLADRunnerTicket
// :returns: error -- an error if a problem has occurred
func (ladAPI *RPLADAPI) IotpkgCleanMockPost(cmd *RPLADIotpkgCmdStruct, runnerTicket *RPLADRunnerTicket) error {
	return ladAPI.iotpkgMainPost(iotpkgCleanMockRoute, cmd, runnerTicket)
}

// iotpkgMainPost is the main function used to post any subcommand of the iotpkg command
//
// :param subCmdURL: URL of the subcommand to run
// :type subCmdURL: string
// :param cmd: Pointer to the structure containing the command parameters
// :type cmd: *RPLADIotpkgCmdStruct
// :param runnerTicket: Pointer to the RPLADRunnerTicket structure to fill with the answer
// :type runnerTicket: *RPLADRunnerTicket
// :returns: error -- an error if a problem has occurred
func (ladAPI *RPLADAPI) iotpkgMainPost(subCmdURL string, cmd *RPLADIotpkgCmdStruct, runnerTicket *RPLADRunnerTicket) error {
	var tmpTicket *RPLADRunnerTicket

	// Check that the pointer to the structure to send is not nil
	if cmd == nil {
		return errors.New(iotpkgCmdNilError)
	}

	if runnerTicket == nil {
		tmpTicket = &RPLADRunnerTicket{}
	} else {
		tmpTicket = runnerTicket
	}

	// Logging purpose --------------------------
	iotpkgCmdLine := getIotpkgCmdLine(subCmdURL, cmd)
	ladAPI.Log.Infof("[iotpkg] command line: %s", iotpkgCmdLine)
	// ------------------------------------------

	if err := ladAPI.HTTPCli.Post(subCmdURL, cmd, tmpTicket); err != nil {
		return err
	}

	return nil
}

// getIotpkgCmdLine is a function uniquely useful to print in the logs the iotpkg command line played
// inside the local builder by redpesk-lad. Indeed this is really useful in order to know which command line
// may cause problem inside the local-builder
//
// :param subCmdURL: URL of the subcommand to run
// :type subCmdURL: string
// :param cmd: Pointer to the structure containing the command parameters
// :type cmd: *RPLADIotpkgCmdStruct
// :returns: string -- the command line run in the local-builder
func getIotpkgCmdLine(subCmdURL string, cmd *RPLADIotpkgCmdStruct) string {
	result := []string{"iotpkg"}

	// Firstly, get the subcommand played
	subCmd := path.Base(subCmdURL)

	// Add the flags given for iotpkg
	for _, value := range cmd.IotpkgFlags {
		if value != "" {
			result = append(result, value)
		}
	}

	// Add the subcommand
	result = append(result, subCmd)

	// Add the flags given for the sub-command
	for _, value := range cmd.SubcmdFlags {
		if value != "" {
			result = append(result, value)
		}
	}

	// Convert to string and return the result
	return strings.Join(result, " ")
}
