/**
 * Copyright (C) 2020-2023 IoT.bzh Company
 * Contact: https://www.iot.bzh/licensing
 *
 * This file is part of the redpesk-cli project, the redpesk command line tool.
 *
 * $RP_BEGIN_LICENSE$
 * Commercial License Usage
 *  Licensees holding valid commercial IoT.bzh licenses may use this file in
 *  accordance with the commercial license agreement provided with the
 *  Software or, alternatively, in accordance with the terms contained in
 *  a written agreement between you and The IoT.bzh Company. For licensing terms
 *  and conditions see https://www.iot.bzh/terms-conditions. For further
 *  information use the contact form at https://www.iot.bzh/contact.
 *
 * GNU General Public License Usage
 *  Alternatively, this file may be used under the terms of the GNU General
 *  Public license version 3. This license is as published by the Free Software
 *  Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
 *  of this file. Please review the following information to ensure the GNU
 *  General Public License requirements will be met
 *  https://www.gnu.org/licenses/gpl-3.0.html.
 * $RP_END_LICENSE$
 */

package signals

import (
	"fmt"
	"os"
	"os/signal"

	"github.com/sirupsen/logrus"
)

// OnSignalCB callback type for signal
type OnSignalCB func(sig os.Signal)

// MonSignals .
type MonSignals struct {
	callback          map[string][]OnSignalCB
	registeredSignals []os.Signal
}

var monSig MonSignals
var exitMonitoringChannel chan bool

// MonitorSignals Routine used to monitor signals (eg. SIGINT, SIGTERM, ...)
func MonitorSignals(log *logrus.Logger) {

	// Initialize the exit Channel
	exitMonitoringChannel = make(chan bool, 1)

	monSig.callback = make(map[string][]OnSignalCB)
	monSig.registeredSignals = GetRegisteredSignals()

	sigs := make(chan os.Signal, 1)
	signal.Notify(sigs, monSig.registeredSignals...)

	go func() {
		monitoringSignals := true
		for monitoringSignals {
			select {
			case sig := <-sigs:
				// Case 1: reception of a signal from the OS
				log.Debugf(signalDetectedMsg, sig)
				if cbArr, exist := monSig.callback[sig.String()]; exist {
					for _, cb := range cbArr {
						cb(sig)
					}
				}

			case <-exitMonitoringChannel:
				// Case 2: reception of a signal requesting to stop the monitoring
				log.Info("Ending signals monitoring...")
				monitoringSignals = false
			}
		}
		// At the end, put channel back to nil
		exitMonitoringChannel = nil
	}()
}

// EndSignalMonitoring allows to end properly the goroutine monitoring the signals
func EndSignalMonitoring() {
	if exitMonitoringChannel != nil {
		exitMonitoringChannel <- true
	}
}

// isSupportSignal Check if a signal is supported or not
func isSupportSignal(sig os.Signal) bool {
	for _, s := range monSig.registeredSignals {
		if s == sig {
			return true
		}
	}
	return false
}

// OnSignal Register a callback for a specified signal
func OnSignal(sig os.Signal, fCB OnSignalCB) error {
	if !isSupportSignal(sig) {
		return fmt.Errorf(signalNotSupportedMsg, sig)
	}
	sigStr := sig.String()
	monSig.callback[sigStr] = append(monSig.callback[sigStr], fCB)
	return nil
}

// OnSignals Register a callback for any signals
func OnSignals(fCB OnSignalCB) error {
	for _, s := range monSig.registeredSignals {
		if err := OnSignal(s, fCB); err != nil {
			return err
		}
	}
	return nil
}
