/**
 * Copyright (C) 2020-2023 IoT.bzh Company
 * Contact: https://www.iot.bzh/licensing
 *
 * This file is part of the redpesk-cli project, the redpesk command line tool.
 *
 * $RP_BEGIN_LICENSE$
 * Commercial License Usage
 *  Licensees holding valid commercial IoT.bzh licenses may use this file in
 *  accordance with the commercial license agreement provided with the
 *  Software or, alternatively, in accordance with the terms contained in
 *  a written agreement between you and The IoT.bzh Company. For licensing terms
 *  and conditions see https://www.iot.bzh/terms-conditions. For further
 *  information use the contact form at https://www.iot.bzh/contact.
 *
 * GNU General Public License Usage
 *  Alternatively, this file may be used under the terms of the GNU General
 *  Public license version 3. This license is as published by the Free Software
 *  Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
 *  of this file. Please review the following information to ensure the GNU
 *  General Public License requirements will be met
 *  https://www.gnu.org/licenses/gpl-3.0.html.
 * $RP_END_LICENSE$
 */

package utils

import (
	"archive/zip"
	"bufio"
	"bytes"
	"crypto/md5"
	"crypto/rand"
	"encoding/hex"
	"encoding/json"
	"io"
	"net/url"
	"os"
	"os/exec"
	"path"
	"path/filepath"
	"regexp"
	"runtime"
	"strings"

	"git.ovh.iot/redpesk/redpesk-infra/rp-golang-common/rphttp"
	"github.com/gofrs/uuid"
	"github.com/mitchellh/go-homedir"
	"github.com/pkg/errors"
)

const (
	noKeyInSpecFileErr       = "no line starting with \"%s\" found in the spec file \"%s\", please check it"
	tooManyArgsInSpecLineErr = "it seems to have too many arguments in processed line of the spec file: %v"
)

// ContainsString checks if a slice of strings contains a given string
//
// :param slice: Slice of string to check
// :type slice: []string
// :param str: String to look for in the slice
// :type str: string
// :returns: bool -- true if the slice contains the string, false in the other case
func ContainsString(slice []string, str string) bool {
	for _, a := range slice {
		if a == str {
			return true
		}
	}
	return false
}

// IndexOf returns the position of the element found in the slice
//
// :param data: Slice of string to check
// :type data: []string
// :param element: String to look for in the slice
// :type element: string
// :returns: int -- index of the element
func IndexOf(data []string, element string) int {
	for k, v := range data {
		if element == v {
			return k
		}
	}
	return -1 //not found.
}

// MapContainsString checks if a map of strings contains a given string as value
//
// :param mapToCheck: Map of string to check
// :type mapToCheck: map[string]string
// :param str: String to look for in the map
// :type str: string
// :returns: bool -- true if the map contains the string, false in the other case
func MapContainsString(mapToCheck map[string]string, str string) bool {
	for _, a := range mapToCheck {
		if a == str {
			return true
		}
	}
	return false
}

// DeleteStringFromSlice delete the first element equal to the passed string from the slice
//
// :param slice: Slice where an element needs to be deleted
// :type slice: *[]interface{}
// :param element: Value of the element to delete
// :type element: string
// :returns: bool -- true if the slice contains the string, false in the other case
func DeleteStringFromSlice(slice *[]string, element string) {
	indexToDelete := -1

	// Go all over the slice
	for index, value := range *slice {
		if value == element {
			indexToDelete = index
			break
		}
	}

	// Delete the element if we found it
	if indexToDelete > -1 {
		// Remove the element at index "index" from slice.
		(*slice)[indexToDelete] = (*slice)[len(*slice)-1] // Copy last element to index "index"
		(*slice) = (*slice)[:len(*slice)-1]               // Truncate slice
	}
}

// ReplaceInStringSlice allows to replace all elements with a given value
// to a new one
//
// :param slice: Slice we want to change
// :type slice: *[]string
// :param oldStr: Old value of the string
// :type oldStr: string
// :param newStr: New value of the string
// :type newStr: string
func ReplaceInStringSlice(slice *[]string, oldStr, newStr string) {
	indexFound := []int{}

	// Get the index of all the element equal to the oldStr
	for tmpIndex, item := range *slice {
		if item == oldStr {
			indexFound = append(indexFound, tmpIndex)
		}
	}

	// For each index found, change the value
	for _, index := range indexFound {
		(*slice)[index] = newStr
	}
}

// RemoveDuplicateInStrSlice allows to remove all duplicated elements
// in a string slice
//
// :param slice: Slice we want to change
// :type slice: []string
// :returns: []string -- The same slice without duplicated value
func RemoveDuplicateInStrSlice(slice []string) []string {
	// Create the variables that we will use
	tmpMap := make(map[string]int)
	result := []string{}

	// Add each element as key of map
	for _, item := range slice {
		tmpMap[item] = 1
	}

	// Now iterate ove the keys
	for strValue := range tmpMap {
		result = append(result, strValue)
	}

	return result
}

// DeleteEmptyStringFromSlice deletes all the empty string in a slice
//
// :param slice: Slice to process
// :type slice: []string
// :returns: []string -- The same slice without empty string
func DeleteEmptyStringFromSlice(slice []string) []string {

	var resultSlice []string

	// Go all over the slice
	for _, value := range slice {
		if value != "" {
			resultSlice = append(resultSlice, value)
		}
	}

	return resultSlice
}

// IsValidUUID allows to validate if a string is an uuid or not
//
// :param uuid: String to validate
// :type uuid: string
// :returns: bool -- true if the parameter is a valid UUID, false in the other case
func IsValidUUID(uuid string) bool {
	r := regexp.MustCompile("^[a-fA-F0-9]{8}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{4}-[a-fA-F0-9]{12}$")
	return r.MatchString(uuid)
}

// UUIDEqual allows to check if two uuids are equals
//
// :param u1: First uuid to compare
// :type u1: uuid.UUID
// :param u2: Second uuid to compare
// :type u2: uuid.UUID
// :returns: bool -- true if the two uuids are the same, false in the other case
func UUIDEqual(u1 uuid.UUID, u2 uuid.UUID) bool {
	return bytes.Equal(u1[:], u2[:])
}

// CopyFile allows to copy a file as its name say it
//
// :param src: Path to the file to copy
// :type src: string
// :param dst: Path to the destination
// :type dst: string
func CopyFile(src string, dst string) error {

	// Open the source file
	sourceFile, err := os.Open(src)
	if err != nil {
		return err
	}
	defer sourceFile.Close()

	// Create new file
	newFile, err := os.Create(dst)
	if err != nil {
		return err
	}
	defer newFile.Close()

	_, err = io.Copy(newFile, sourceFile)
	if err != nil {
		return err
	}

	return nil
}

// SanitizePath allows from a path, to transform any "~" to the home directory
// and any "." or ".." to the absolute path
//
// :param originalPath: Path to sanitize
// :type originalPath: string
// :returns: string, error -- the sanitized path and an error if something went wrong
func SanitizePath(originalPath string) (string, error) {
	var err error
	resPath := ""

	// Expand the '~' if present in what the user gave us
	resPath, err = homedir.Expand(originalPath)
	if err != nil {
		return resPath, err
	}

	// Get the absolute path from the given path
	resPath, err = filepath.Abs(resPath)
	if err != nil {
		return resPath, err
	}

	return resPath, nil
}

// WriteToFile will print any string of text to a file safely by
// checking for errors and syncing at the end
//
// :param filename: Path and filename to create
// :type filename: string
// :returns: error -- an error if something went wrong
func WriteToFile(filename string, data string) error {
	file, err := os.Create(filename)
	if err != nil {
		return err
	}
	defer file.Close()

	_, err = io.WriteString(file, data)
	if err != nil {
		return err
	}
	return file.Sync()
}

// RemoveFileIfExists does exactly what its name says:
// if the file exists, it is deleted. If it doesn't, no error
// is returned, if another error appeared, it is returned.
//
// :param filepath: Path to the file to delete
// :type filepath: string
// :returns: error -- an error if something went wrong
func RemoveFileIfExists(filepath string) error {
	if _, err := os.Stat(filepath); err == nil {
		// The file exists
		return os.Remove(filepath)
	} else if errors.Is(err, os.ErrNotExist) {
		return nil
	} else {
		return err
	}
}

// ExtractBaseURLvalues allows to get as a list, all the the baseURL specified inside a
// "install_repo.sh" script (generated by Koji)
//
// :param installRepoPath: Path to the "install_repo.sh" downloaded file
// :type installRepoPath: string
// :param architecture: architecture to replace "$basearch" with
// :type architecture: string
// :returns: []string, error -- the list of repos URL and an error if something went wrong
func ExtractBaseURLvalues(installRepoPath string, architecture string) ([]string, error) {
	results := []string{}
	lineBeginning := "baseurl="

	// Open the spec file
	instRepoFile, err := os.Open(installRepoPath)
	if err != nil {
		return results, err
	}
	// Defer the file closing
	defer instRepoFile.Close()

	// Create the scanner on the file
	scanner := bufio.NewScanner(instRepoFile)

	// Prepare the scanner to read line by line
	scanner.Split(bufio.ScanLines)

	// Iterate over each line until we find the line that interest us
	tmpLine := ""
	for scanner.Scan() {
		// Get the current line
		tmpLine = scanner.Text()
		// Check if this is the line that interest us
		if strings.HasPrefix(tmpLine, lineBeginning) {
			// The line interest us!
			tmpLine = strings.TrimSpace(tmpLine)
			lineTab := strings.SplitN(tmpLine, "=", 2)
			if len(lineTab) >= 2 {
				results = append(results, strings.ReplaceAll(lineTab[1], "\\$basearch", architecture))
			}
		}
	}

	return results, nil
}

// WriteJsonFile allows to write a json file from a structure
//
// :param filePath: Path to the file to create
// :type filePath: string
// :param object: Object to write in the file
// :type object: interface{}
// :returns: error -- an error if something went wrong
func WriteJsonFile(filePath string, object interface{}) error {

	// Create the file to write
	file, err := json.MarshalIndent(object, "", "\t")
	if err != nil {
		return err
	}

	// Write the file
	err = os.WriteFile(filePath, file, 0600)

	return err
}

// ReadJsonFile allows to get a structure from a json file
//
// :param filePath: Path to the file containing the struct
// :type filePath: string
// :param object: Pointer to the structure to fill
// :type object: interface{}
// :returns: error -- an error if something went wrong
func ReadJsonFile(filePath string, object interface{}) error {

	// Try to read the file
	fileContent, err := os.ReadFile(filePath)
	if err != nil {
		return err
	}

	// Try to get the values through unmarshal
	err = json.Unmarshal([]byte(fileContent), object)
	if err != nil {
		return err
	}

	return err
}

// FindFilesWithExtension allows to find the name of all the files
// from the directory it should be in and an extension
//
// :param dir: Path of the directory containing the file
// :type dir: string
// :param extension: Extension of the file to find
// :type extension: string
// :returns: error -- an error if something went wrong
func FindFilesWithExtension(dir string, extension string) ([]string, error) {
	// Initialize the result
	results := []string{}

	err := filepath.Walk(dir, func(path string, info os.FileInfo, err error) error {
		// Select only the file with the extension we want
		if filepath.Ext(path) == extension {
			results = append(results, path)
		}
		return nil
	})

	return results, err
}

// FindAllFilesAndDirsRecursively allows to get a list of all files and all sub-directories
// recursively in a directory. The function returns a map:
//
//	"dirs"  -> ["subdir1", "subdir2", etc.]
//	"files" -> ["file1", "file2", "file3", etc.]
//
// :param dir: Path of the directory containing the files and subdirectories
// :type dir: string
// :returns: map[string][]string, error -- lists of files and subdirectories and an error if something went wrong
func FindAllFilesAndDirsRecursively(dir string) (map[string][]string, error) {
	// Initialize the result
	results := make(map[string][]string)
	results["dirs"] = []string{}
	results["files"] = []string{}

	err := filepath.Walk(dir, func(path string, info os.FileInfo, err error) error {
		if err != nil {
			return err
		}

		if !info.IsDir() {
			results["files"] = append(results["files"], path)
		} else {
			results["dirs"] = append(results["dirs"], path)
		}
		return nil
	})

	return results, err
}

// RunCmdInDirectory allows to run a command given in parameter
// in a given directory. Example:
//
// RunCmdInDirectory("/tmp", "git", "clone", "http://git.ovh.iot/redtest-helloworld-api.git")
//
// :param dir: Path of the working directory
// :type dir: string
// :param name: Name of the command to run
// :type name: string
// :param args: Arguments to pass to the command
// :type args: string
// :returns: string, error -- the output of the command and an error if something went wrong
func RunCmdInDirectory(dir string, name string, args ...string) (string, error) {
	// Prepare the command object to run
	cmdToRun := exec.Command(name, args...)

	// Set the working directory for the command
	cmdToRun.Dir = dir

	// Execute the command
	output, err := cmdToRun.CombinedOutput()
	cmdToRun.Run()

	return string(output), err
}

// GetURLPort allows to get the port specified in an URL
// If no port is specified, it returns an empty string
//
// :param dir: Path of the directory containing the file
// :type dir: string
// :param extension: Extension of the file to find
// :type extension: string
// :returns: error -- an error if something went wrong
func GetURLPort(url string) string {
	// Initialize the result
	urlPort := ""

	// Split the URL with ":" as a separator
	splitStrSlice := strings.Split(url, ":")

	if len(splitStrSlice) > 2 {
		urlPort = strings.ReplaceAll(splitStrSlice[1], "/", "")
	}

	return urlPort
}

// CreateDirIfNotExists allows to create a directory if it does
// not exist. If the overwrite option is set to True , the
// existing directory is overwritten if it already exists
//
// :param path: Path of the directory to create/overwrite
// :type path: string
// :param overwrite: Overwrite option
// :type overwrite: bool
// :returns: error -- an error if something went wrong
func CreateDirIfNotExists(path string, overwrite bool) error {
	// Check if the directory exists
	_, err := os.Stat(path)

	// If the directory exists, and the overwrite option is at True:
	// delete it
	if overwrite && (err == nil) {
		err = os.RemoveAll(path)
		if err != nil {
			return err
		}
	}

	// If the directory does not exists or if we overwrite it:
	// create the directory
	if overwrite || os.IsNotExist(err) {
		// Create the directory
		err = os.MkdirAll(path, 0755)
	}

	return err
}

// RemoveDirContents allows to delete all the content of a
// repository
//
// :param dir: Path of the directory to empty
// :type dir: string
// :returns: error -- an error if something went wrong
func RemoveDirContents(dir string) error {
	d, err := os.Open(dir)
	if err != nil {
		return err
	}
	defer d.Close()
	names, err := d.Readdirnames(-1)
	if err != nil {
		return err
	}
	for _, name := range names {
		err = os.RemoveAll(filepath.Join(dir, name))
		if err != nil {
			return err
		}
	}
	return nil
}

// IsDirectoryEmpty allows to check if a directory is empty or not
//
// :param dir: Path of the directory to check
// :type dir: string
// :returns: bool, error -- a boolean saying if the directory is empty and an error if something went wrong
func IsDirectoryEmpty(dir string) (bool, error) {
	files, err := os.ReadDir(dir)
	if err != nil {
		return false, err
	}

	return (len(files) == 0), nil
}

// RemoveDirIfEmpty allows to remove a directory if it is empty
//
// :param dir: Path of the directory to check and remove
// :type dir: string
// :returns: error -- and an error if something went wrong
func RemoveDirIfEmpty(dir string) error {
	isEmpty, err := IsDirectoryEmpty(dir)
	if err != nil {
		return err
	}

	if isEmpty {
		return os.Remove(dir)
	}

	return nil
}

// IsDirectory allows to check if a path is a directory, or not
//
// :param path: Path to check
// :type path: string
// :returns: bool, error -- a boolean saying if the directory is empty and an error if something went wrong
func IsDirectory(path string) (bool, error) {
	fileInfo, err := os.Stat(path)
	if err != nil {
		return false, err
	}

	return fileInfo.IsDir(), err

}

// CreateZipFromDir allows to create a ZIP archive from
// a directory. This function walk inside the directory
// to archive every files in it.
//
// :param inputDir: Path of the directory to archive
// :type inputDir: string
// :param outputArchive: ZIP archive name
// :type outputArchive: string
// :returns: error -- an error if something went wrong
func CreateZipFromDir(inputDir, outputArchive string) error {
	file, err := os.Create(outputArchive)
	if err != nil {
		return err
	}
	defer file.Close()

	w := zip.NewWriter(file)
	defer w.Close()

	walker := func(path string, info os.FileInfo, err error) error {
		if err != nil {
			return err
		}
		if info.IsDir() {
			return nil
		}
		file, err := os.Open(path)
		if err != nil {
			return err
		}
		defer file.Close()

		// Ensure that `path` is not absolute; it should not start with "/".
		pathReworked := strings.ReplaceAll(path, inputDir, "")
		pathReworked = strings.TrimLeft(pathReworked, "/")

		f, err := w.Create(pathReworked)
		if err != nil {
			return err
		}

		_, err = io.Copy(f, file)
		if err != nil {
			return err
		}

		return nil
	}

	// Create the archive
	return filepath.Walk(inputDir, walker)
}

// HTTPDwldFileInDir allows to download a file inside a directory.
// The name of the file is retrieved from the URL
//
// :param directory: Path to the directory where the files will be downloaded
// :type directory: string
// :param url: URL to use to download the file
// :type url: string
// :returns: error -- an error if something went wrong
func HTTPDwnldFileInDir(directory string, fileURL string) error {
	// Check that the directory exists
	_, err := os.Stat(directory)
	if err != nil {
		return err
	}

	// Retrieve the name of the file
	// Get the base URL for the logs
	fileURLObj, err := url.Parse(fileURL)
	if err != nil {
		return nil
	}
	filename := path.Base(fileURLObj.Path)
	fullFilename := path.Join(directory, filename)

	// Download the file
	err = rphttp.HTTPDownloadFile(fullFilename, fileURL)

	return err
}

// GenerateSecureToken allows to generate a secure token of a given token
//
// :param length: Length of the token to generate
// :type length: int
// :returns: string -- the generated secure token
func GenerateSecureToken(length int) string {
	b := make([]byte, length)
	if _, err := rand.Read(b); err != nil {
		return ""
	}
	return hex.EncodeToString(b)
}

// GetPkgRootDir allows to get the root directory of the package
// (to use uniquely for tests or in devmode)
//
// :returns: string -- path to the pkg root directory
func GetPkgRootDir() string {
	_, b, _, _ := runtime.Caller(0)
	basepath := filepath.Dir(b)

	// Reconstruct the path to the real root of project
	result := ""
	directories := strings.Split(basepath, "/")
	for _, dir := range directories {

		// We've reached the utils directory of the project, stop it
		// OK it is a bit dangerous but it is used uniquely by the tests anyway
		if dir == "utils" {
			break
		}

		// Add this directory in the path
		result += (dir + "/")
	}

	return result
}

// ExtractLineWithPrefix allows to retrieve a line inside a string that start
// with the given prefix
//
// :returns: string -- the first line starting with the prefix
func ExtractLineWithPrefix(text, prefix string) string {
	// Convert all "\r\n" with "\n"
	tmpText := strings.ReplaceAll(text, "\r\n", "\n")

	// Convert the text into tab of each line
	textAllLines := strings.Split(tmpText, "\n")

	// Look for the line
	resultLine := ""
	for _, line := range textAllLines {
		if strings.HasPrefix(line, prefix) {
			resultLine = line
			break
		}
	}

	return resultLine
}

// CalculateMd5SumOnFile allows to calculate the MD5SUM of
// a file
//
// :param filePath: Path to the file
// :type filePath: string
// :param trim: Boolean requesting to Trim the space before doing the sum
// :type trim: bool
// :returns: string, error -- md5sum as string and an error if a problem has occurred
func CalculateMd5SumOnFile(filePath string, trim bool) (string, error) {
	// Try to read the file
	fileContent, err := os.ReadFile(filePath)
	if err != nil {
		return "", err
	}

	// Convert bytes array to string
	fileContentStr := string(fileContent)

	if trim {
		fileContentStr = strings.TrimSpace(fileContentStr)
	}

	return CalculateMd5SumOnString(fileContentStr), nil
}

// CalculateMd5SumOnString allows to calculate the MD5SUM of
// a string
//
// :param stringToHash: String we want to calculate md5sum on
// :type stringToHash: string
// :returns: string-- md5sum as string
func CalculateMd5SumOnString(stringToHash string) string {

	data := []byte(stringToHash)
	md5SumBytes := md5.Sum(data)

	return hex.EncodeToString(md5SumBytes[:])
}

// OptionNotSetConv
//
// :param value: a string value
// :type string: string
// :returns: string -- [Not set] or the string value
func OptionNotSetConv(value string) string {
	if value == "" {
		return "[Not set]"
	}
	return value
}

// OptionOffOnConv
//
// :param value: a boolean value
// :type bool: bool
// :returns: string -- OFF or ON string
func OptionOffOnConv(value bool) string {
	if value {
		return "ON"
	}
	return "OFF"
}

// OptionMandatoryOptionalConv
//
// :param set: a boolean value
// :type bool: bool
// :returns: string -- OFF|ON [Mand|Opt] string
func OptionMandatoryOptionalConv(set bool, mandatory bool) string {
	res := "OFF"
	if set {
		res = "ON"
	}
	if mandatory {
		res += " [Mand]"
	} else {
		res += " [Opt]"
	}
	return res
}

// GetLinuxDistro retrieves the linux distribution
// ID from the /etc/os-release file
//
// :returns: string -- Distribution ID of the OS
func GetLinuxDistro() string {
	// Get the content of the /etc/os-release file
	osReleaseContent, err := os.ReadFile("/etc/os-release")
	if err != nil {
		return ""
	}
	// Get the string in the text that match our ID regexp
	distroRegexp := regexp.MustCompile(`^ID=(.+)`)
	// Try to get the match
	matchGroups := distroRegexp.FindStringSubmatch(string(osReleaseContent))
	if len(matchGroups) >= 1 {
		return matchGroups[0]
	}
	return ""
}
