/**
 * Copyright (C) 2020-2023 IoT.bzh Company
 * Contact: https://www.iot.bzh/licensing
 *
 * This file is part of the redpesk-cli project, the redpesk command line tool.
 *
 * $RP_BEGIN_LICENSE$
 * Commercial License Usage
 *  Licensees holding valid commercial IoT.bzh licenses may use this file in
 *  accordance with the commercial license agreement provided with the
 *  Software or, alternatively, in accordance with the terms contained in
 *  a written agreement between you and The IoT.bzh Company. For licensing terms
 *  and conditions see https://www.iot.bzh/terms-conditions. For further
 *  information use the contact form at https://www.iot.bzh/contact.
 *
 * GNU General Public License Usage
 *  Alternatively, this file may be used under the terms of the GNU General
 *  Public license version 3. This license is as published by the Free Software
 *  Foundation and appearing in the file LICENSE.GPLv3 included in the packaging
 *  of this file. Please review the following information to ensure the GNU
 *  General Public License requirements will be met
 *  https://www.gnu.org/licenses/gpl-3.0.html.
 * $RP_END_LICENSE$
 */

package rpbckmock

import (
	"context"
	"encoding/base64"
	"encoding/json"
	"fmt"
	"io"
	"math/rand"
	"net/http"
	"os"
	"path"
	"path/filepath"
	"strconv"
	"strings"
	"sync"
	"time"

	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/rpxxxapi/bckapi"
	"git.ovh.iot/redpesk/redpesk-infra/redpesk-cli/utils"
	"git.ovh.iot/redpesk/redpesk-infra/rp-webserver/actions/rpapiv1"

	"github.com/gofrs/uuid"
	"github.com/gorilla/mux"
	gosocketio "github.com/graarh/golang-socketio"
	"github.com/graarh/golang-socketio/transport"
	"github.com/jinzhu/copier"
	"github.com/pkg/errors"
	"github.com/sirupsen/logrus"
)

// -------------------------------------------------------------------------------
// Types definition --------------------------------------------------------------
// -------------------------------------------------------------------------------

// RPMockedBackend definition
type RPMockedBackend struct {
	http.Server
	WSServer        *gosocketio.Server
	WSServerChannel *gosocketio.Channel
	log             *logrus.Logger
}

// AuditBaseStruct Audit arg entry/res structure
type AuditInputBaseCustom struct {
	Type  rpapiv1.RPAuditType `json:"type"`
	Entry AuditInputEntry     `json:"entry"` // AuditAnalyzeCodeIn AuditAnalyzeCodeResult AuditListRes AuditVersionRes
}

type AuditInputCustom struct {
	Audit []AuditInputBaseCustom `json:"audit"`
}

type AuditInputEntry struct {
	Tools []string `json:"tools"`
}

// -------------------------------------------------------------------------------
// -------------------------------------------------------------------------------

// -------------------------------------------------------------------------------
// Global variables --------------------------------------------------------------
// -------------------------------------------------------------------------------
var allProjects = []rpapiv1.RPProjectFull{
	{
		RPProject: rpapiv1.RPProject{
			ID:          uuid.FromStringOrNil(distroID33),
			Name:        distroName33,
			Slug:        distroSlug33,
			Version:     distroVersion33,
			Description: distroDesc33,
			Type:        rpapiv1.RPProjectTypeExternal,
			SupportedObjects: rpapiv1.RPSupportedObjects{
				Applications: true,
			},
			IsDistribution: true,
			Architectures: []rpapiv1.RPArch{
				{
					ID:  "aarch64",
					Set: true,
				},
				{
					ID:  "x86_64",
					Set: true,
				},
			},
			NbRPMsPerArch: map[string]int{
				"aarch64": 5,
				"x86_64":  5,
			},
		},
	},
	{
		RPProject: rpapiv1.RPProject{
			ID:      uuid.FromStringOrNil(distroIDArz),
			Name:    distroNameArz,
			Slug:    distroSlugArz,
			Version: "1.0",
			Type:    rpapiv1.RPProjectTypeExternal,
			SupportedObjects: rpapiv1.RPSupportedObjects{
				Applications: true,
			},
			IsDistribution: true,
			Architectures: []rpapiv1.RPArch{
				{
					ID:  "aarch64",
					Set: true,
				},
				{
					ID:  "x86_64",
					Set: true,
				},
			},
			NbRPMsPerArch: map[string]int{
				"aarch64": 5,
				"x86_64":  5,
			},
		},
	},
	{
		RPProject: rpapiv1.RPProject{
			ID:          uuid.FromStringOrNil(distroIDLTSArz),
			Name:        distroNameLTSArz,
			Slug:        distroSlugLTSArz,
			Version:     distroVersionLTSArz,
			Description: distroDescLTSArz,
			Type:        rpapiv1.RPProjectTypeExternal,
			SupportedObjects: rpapiv1.RPSupportedObjects{
				Applications: true,
			},
			IsDistribution: true,
			Architectures: []rpapiv1.RPArch{
				{
					ID:  "aarch64",
					Set: true,
				},
				{
					ID:  "x86_64",
					Set: true,
				},
			},
			NbRPMsPerArch: map[string]int{
				"aarch64": 5,
				"x86_64":  5,
			},
		},
	},
	{
		RPProject: rpapiv1.RPProject{
			ID:          uuid.FromStringOrNil(distroIDLTSBatz),
			Name:        distroNameLTSBatz,
			Slug:        distroSlugLTSBatz,
			Version:     distroVersionLTSBatz,
			Description: distroDescLTSBatz,
			Type:        rpapiv1.RPProjectTypeExternal,
			SupportedObjects: rpapiv1.RPSupportedObjects{
				Applications: true,
			},
			IsDistribution: true,
			Architectures: []rpapiv1.RPArch{
				{
					ID:  "aarch64",
					Set: true,
				},
				{
					ID:  "x86_64",
					Set: true,
				},
			},
			NbRPMsPerArch: map[string]int{
				"aarch64": 5,
				"x86_64":  5,
			},
		},
	},
	{
		RPProject: rpapiv1.RPProject{
			ID:          uuid.FromStringOrNil(distroIDLTSZephyr),
			Name:        distroNameLTSZephyr,
			Slug:        distroSlugLTSZephyr,
			Version:     distroVersionLTSZephyr,
			SubVersion:  distroSubVersionLTSZephyr,
			Description: distroDescLTSZephyr,
			Type:        rpapiv1.RPProjectTypeExternal,
			SupportedObjects: rpapiv1.RPSupportedObjects{
				Applications: true,
			},
			IsDistribution: true,
			Architectures: []rpapiv1.RPArch{
				{
					ID:  "aarch64",
					Set: true,
				},
				{
					ID:  "x86_64",
					Set: true,
				},
			},
			NbRPMsPerArch: map[string]int{
				"aarch64": 5,
				"x86_64":  5,
			},
		},
	},
	{
		RPProject: rpapiv1.RPProject{
			ID:          uuid.FromStringOrNil(distroIDLTSBatzUpdate),
			Name:        distroNameLTSBatzUpdate,
			Slug:        distroSlugLTSBatzUpdate,
			Version:     distroVersionLTSBatzUpdate,
			Description: distroDescLTSBatzUpdate,
			Type:        rpapiv1.RPProjectTypeExternal,
			SupportedObjects: rpapiv1.RPSupportedObjects{
				Applications: true,
			},
			IsDistribution: true,
			Architectures: []rpapiv1.RPArch{
				{
					ID:  "aarch64",
					Set: true,
				},
				{
					ID:  "x86_64",
					Set: true,
				},
			},
			NbRPMsPerArch: map[string]int{
				"aarch64": 5,
				"x86_64":  5,
			},
		},
	},
	{
		RPProject: rpapiv1.RPProject{
			ID:      uuid.FromStringOrNil("d2264369-29df-4475-b350-7d8a6df9a760"),
			Name:    "redpesk-brehat",
			Slug:    "redpesk-brehat",
			Version: "2.0",
			Type:    rpapiv1.RPProjectTypeExternal,
			SupportedObjects: rpapiv1.RPSupportedObjects{
				Applications: true,
			},
			IsDistribution: true,
			Architectures: []rpapiv1.RPArch{
				{
					ID:  "aarch64",
					Set: true,
				},
				{
					ID:  "x86_64",
					Set: true,
				},
			},
			NbRPMsPerArch: map[string]int{
				"aarch64": 5,
				"x86_64":  5,
			},
		},
	},
	{
		RPProject: rpapiv1.RPProject{
			ID:      uuid.FromStringOrNil("d2264369-29df-4475-b351-7d8a6df9a760"),
			Name:    "redpesk-cezembre",
			Slug:    "redpesk-cezembre",
			Version: "3.0",
			Type:    rpapiv1.RPProjectTypeExternal,
			SupportedObjects: rpapiv1.RPSupportedObjects{
				Applications: true,
			},
			IsDistribution: true,
			Architectures: []rpapiv1.RPArch{
				{
					ID:  "aarch64",
					Set: true,
				},
				{
					ID:  "x86_64",
					Set: true,
				},
			},
			NbRPMsPerArch: map[string]int{
				"aarch64": 5,
				"x86_64":  5,
			},
		},
	},
}

var allRPMsName = []string{
	"autoconf", "autoconf-archive", "avahi-ui", "avahi-ui-devel",
	"bind", "bind-devel", "bind-utils",
	"gawk", "gawk-all-langpacks", "gawk-devel", "gawk-doc",
	"hyphen", "hyphen-devel",
	"iproute", "ipset", "ipset-service",
	"java-1.8.0-openjdk",
	"kernel", "kernel-core", "kernel-debug",
}

var allExportedProjects = []rpapiv1.RPExportProjectState{}

var allImportedProjects = []rpapiv1.RPImportProjectState{}

var allReleasedProjects = []rpapiv1.RPReleaseProjectFull{}

var allDeployments = []rpapiv1.RPDeploymentFull{}

var allReleasedImages = []rpapiv1.RPReleaseImageFull{}

var allRelProjectsAudience = []rpapiv1.RPObjectRoles{}

var allRelImagesAudience = []rpapiv1.RPObjectRoles{}

var allUsers = rpapiv1.RPUsersFull{
	PerPage:      50,
	Page:         1,
	TotalEntries: 0,
	Reverse:      false,
}

var currentUserRoles = rpapiv1.RPRoles{
	Roles: []rpapiv1.RPRole{
		{
			Name:   defaultRoleNameCurrUser,
			Label:  defaultRoleLabelCurrUser,
			Object: defaultRoleObjectCurrUser,
		},
	},
	PerPage:      50,
	Page:         1,
	TotalEntries: 1,
	Reverse:      false,
}

var allTeams = rpapiv1.RPTeams{
	PerPage:      50,
	Page:         1,
	TotalEntries: 0,
	Reverse:      false,
}

var allMembers = rpapiv1.RPMembers{}

var allApps = []rpapiv1.RPApplicationFull{}

// Create the map that will contain the managed branches for each application
var allAppsBranches = make(map[uuid.UUID]*rpapiv1.RPBranchList)

// Builds ID counter (start at 1000)
var buildsCounter = 999

var allBuilds = []rpapiv1.RPBuildFull{}
var buildsMutex sync.Mutex

// List of builds to cancel
var buildsToCancel = []string{}

// Tests ID counter (start at 1)
var testsCounter = 0

var allTests = rpapiv1.RPTests{
	PerPage:      50,
	Page:         1,
	TotalEntries: 0,
	Reverse:      false,
}

// Audits ID counter (start at 1)
var auditsCounter = 0

// Array containing all the audits
var allAudits = []rpapiv1.RPAuditFull{}

var allImages = []rpapiv1.RPImageFull{
	{
		RPImage: rpapiv1.RPImage{
			ID:               uuid.FromStringOrNil("c091b696-657d-4e4e-8c3a-d0c5d9c43339"),
			Name:             "redpesk minimal arz x86_64 smack",
			Slug:             "redpesk-min-arz-x8664-smack",
			Type:             rpapiv1.RPImageTypeExternal,
			DistributionID:   uuid.FromStringOrNil(distroIDArz),
			DistributionSlug: distroSlugArz,
			Version:          "1.0",
			Architecture:     "x86_64",
			SecurityModel:    "smack",
			URL:              "https://download.redpesk.bzh/redpesk-devel/releases/arz-1.0/images/smack/minimal/x86_64/latest/redpesk-devel-minimal-arz-0.x86_64.raw.xz",
		},
	},
	{
		RPImage: rpapiv1.RPImage{
			ID:               uuid.FromStringOrNil("e7157a3f-2ecd-4cbe-9d04-0cc39e6980a9"),
			Name:             "redpesk minimal arz aarch64 smack",
			Slug:             "redpesk-min-arz-aarch64-smack",
			Type:             rpapiv1.RPImageTypeExternal,
			DistributionID:   uuid.FromStringOrNil(distroIDArz),
			DistributionSlug: distroSlugArz,
			Version:          "1.0",
			Architecture:     "aarch64",
			SecurityModel:    "smack",
			URL:              "https://download.redpesk.bzh/redpesk-devel/releases/arz-1.0/images/smack/minimal/aarch64/latest/redpesk-devel-minimal-arz-0.aarch64.raw.xz",
		},
	},
}

// Array containing all the board models
var allBoardModels = []rpapiv1.RPModelBoard{
	{
		ID:        uuid.FromStringOrNil("34256190-dc58-459d-bb24-bbf06bfca975"),
		Name:      "test_ota_x86_64",
		CreatedAt: time.Now(),
		UpdatedAt: time.Now(),
		Archi:     "x86_64",
		Category:  rpapiv1.RPBoardTypeOTA,
		Virtual:   false,
	},
	{
		ID:        uuid.FromStringOrNil("61161806-010e-4449-a3f0-5831656af274"),
		Name:      "test_ota_aarch64",
		CreatedAt: time.Now(),
		UpdatedAt: time.Now(),
		Archi:     "aarch64",
		Category:  rpapiv1.RPBoardTypeOTA,
		Virtual:   false,
	},
	{
		ID:        uuid.FromStringOrNil("f2418145-a15b-4b8a-8b34-f51b24d6c88d"),
		Name:      "qemu-system-x86_64",
		CreatedAt: time.Now(),
		UpdatedAt: time.Now(),
		Archi:     "x86_64",
		Category:  rpapiv1.RPBoardTypeRTM,
		Virtual:   true,
	},
	{
		ID:       uuid.FromStringOrNil("c4709f90-02d2-4bc1-a35c-a52e9eda8e16"),
		Name:     boardQemux86,
		Archi:    "x86_64",
		Category: rpapiv1.RPBoardTypeRTM,
		Virtual:  true,
	},
	{
		ID:       uuid.FromStringOrNil("62d148d3-4a82-4e1e-8c8a-41829c7b37e2"),
		Name:     boardH3,
		Archi:    "aarch64",
		Category: rpapiv1.RPBoardTypeRTM,
		Virtual:  false,
	},
	{
		ID:       uuid.FromStringOrNil("3a3f71b6-ecd7-44fd-abeb-99da1e72afa0"),
		Name:     boardM3,
		Archi:    "aarch64",
		Category: rpapiv1.RPBoardTypeRTM,
		Virtual:  false,
	},
	{
		ID:       uuid.FromStringOrNil("25f9c8c0-8b03-4117-9987-1ad3278d2db0"),
		Name:     boardUpboard,
		Archi:    "x86_64",
		Category: rpapiv1.RPBoardTypeRTM,
		Virtual:  false,
	},
	{
		ID:       uuid.FromStringOrNil("b4daf37e-72f5-4b42-b97e-29f66c090458"),
		Name:     boardImx8,
		Archi:    "aarch64",
		Category: rpapiv1.RPBoardTypeRTM,
		Virtual:  false,
	},
	{
		ID:       uuid.FromStringOrNil("b8471607-3f6d-4f49-b8f0-66732678ca3c"),
		Name:     boardRaspPi,
		Archi:    "aarch64",
		Category: rpapiv1.RPBoardTypeRTM,
		Virtual:  false,
	},
	{
		ID:       uuid.FromStringOrNil("ba0f2231-adcc-4a9b-a254-c81b3f4c6d40"),
		Name:     boardNitrogen,
		Archi:    "aarch64",
		Category: rpapiv1.RPBoardTypeRTM,
		Virtual:  false,
	},
}

// Array containing all the boards
var allBoards = []rpapiv1.RPBoard{
	{
		ID:           uuid.FromStringOrNil("208ffb8d-0d39-46d9-99ed-b2909d73ac05"),
		Name:         "ota-redpesk-lts-arz-1.0-x86_64",
		CreatedAt:    time.Now(),
		UpdatedAt:    time.Now(),
		IP:           "10.0.2.15/24",
		Off:          rpapiv1.RPBoardStatusOffLine,
		ModelID:      uuid.FromStringOrNil("34256190-dc58-459d-bb24-bbf06bfca975"),
		ModelName:    "test_ota_x86_64",
		BoardType:    rpapiv1.RPBoardTypeOTA,
		Architecture: "x86_64",
	},
	{
		ID:           uuid.FromStringOrNil("8d47aee5-d617-490a-995a-771238fdac00"),
		Name:         "ota-redpesk-lts-arz-1.0-aarch64",
		CreatedAt:    time.Now(),
		UpdatedAt:    time.Now(),
		IP:           "10.0.2.17/24",
		Off:          rpapiv1.RPBoardStatusConnected,
		ModelID:      uuid.FromStringOrNil("61161806-010e-4449-a3f0-5831656af274"),
		ModelName:    "test_ota_aarch64",
		BoardType:    rpapiv1.RPBoardTypeOTA,
		Architecture: "aarch64",
	},
	{
		ID:           uuid.FromStringOrNil("57a1a5e4-1caf-4f3b-bcff-a720b77b9dd7"),
		Name:         "qemu-system-x86_64",
		CreatedAt:    time.Now(),
		UpdatedAt:    time.Now(),
		IP:           "",
		Off:          rpapiv1.RPBoardStatusGood,
		ModelID:      uuid.FromStringOrNil("f2418145-a15b-4b8a-8b34-f51b24d6c88d"),
		ModelName:    "qemu-system-x86_64",
		BoardType:    rpapiv1.RPBoardTypeRTM,
		Architecture: "x86_64",
	},
	{
		ID:           uuid.FromStringOrNil("70890b66-5d3a-4ce3-b25a-7390b059ec1a"),
		Name:         "h3ulcb-ws3.0_25159",
		CreatedAt:    time.Now(),
		UpdatedAt:    time.Now(),
		IP:           "",
		Off:          rpapiv1.RPBoardStatusGood,
		ModelID:      uuid.FromStringOrNil("62d148d3-4a82-4e1e-8c8a-41829c7b37e2"),
		ModelName:    boardH3,
		BoardType:    rpapiv1.RPBoardTypeRTM,
		Architecture: "aarch64",
	},
	{
		ID:           uuid.FromStringOrNil("668fa7ae-a2d3-4ae2-9888-5b382599ca31"),
		Name:         "nitrogen8mn01",
		CreatedAt:    time.Now(),
		UpdatedAt:    time.Now(),
		IP:           "",
		Off:          rpapiv1.RPBoardStatusGood,
		ModelID:      uuid.FromStringOrNil("ba0f2231-adcc-4a9b-a254-c81b3f4c6d40"),
		ModelName:    boardNitrogen,
		BoardType:    rpapiv1.RPBoardTypeRTM,
		Architecture: "aarch64",
	},
	{
		ID:           uuid.FromStringOrNil("57a1a5e4-1caf-4f3b-bcff-a720b77b9dd6"),
		Name:         "qemu-redpesk01",
		CreatedAt:    time.Now(),
		UpdatedAt:    time.Now(),
		IP:           "",
		Off:          rpapiv1.RPBoardStatusGood,
		ModelID:      uuid.FromStringOrNil("c4709f90-02d2-4bc1-a35c-a52e9eda8e16"),
		ModelName:    boardQemux86,
		BoardType:    rpapiv1.RPBoardTypeRTM,
		Architecture: "x86_64",
	},
	{
		ID:           uuid.FromStringOrNil("e4458c85-c86a-41f3-8d7b-26450c0fa09f"),
		Name:         "rpi4",
		CreatedAt:    time.Now(),
		UpdatedAt:    time.Now(),
		IP:           "",
		Off:          rpapiv1.RPBoardStatusGood,
		ModelID:      uuid.FromStringOrNil("b8471607-3f6d-4f49-b8f0-66732678ca3c"),
		ModelName:    boardRaspPi,
		BoardType:    rpapiv1.RPBoardTypeRTM,
		Architecture: "aarch64",
	},
	{
		ID:           uuid.FromStringOrNil("361662c9-3616-4e48-ab48-8f388da778ab"),
		Name:         "solidrun-edge-gateways",
		CreatedAt:    time.Now(),
		UpdatedAt:    time.Now(),
		IP:           "",
		Off:          rpapiv1.RPBoardStatusMaintenance,
		ModelID:      uuid.FromStringOrNil("b4daf37e-72f5-4b42-b97e-29f66c090458"),
		ModelName:    boardImx8,
		BoardType:    rpapiv1.RPBoardTypeRTM,
		Architecture: "aarch64",
	},
	{
		ID:           uuid.FromStringOrNil("94be1133-0fbc-4be1-b61a-749cf433a4d6"),
		Name:         "upboard04",
		CreatedAt:    time.Now(),
		UpdatedAt:    time.Now(),
		IP:           "",
		Off:          rpapiv1.RPBoardStatusGood,
		ModelID:      uuid.FromStringOrNil("25f9c8c0-8b03-4117-9987-1ad3278d2db0"),
		ModelName:    boardUpboard,
		BoardType:    rpapiv1.RPBoardTypeRTM,
		Architecture: "x86_64",
	},
}

var bckVersion = versionVersion

// Error to send on the next call on /api/v1/config
var tmpErrToSend = ""

// Variable used to request project export to fail
var prjExportFails = false

// Variable used to request project import to fail
var prjImportFails = false

// Boolean allowing to request a test to be cancelled
var abortCurrentTest = false

// Boolean allowing the test module to request the next test to be timed-out
var testInTimeOut = false

// Generic waiting time before sending an event
var genericWaitingTime = (10 * time.Millisecond)

// Boolean allowing to request an error event during a blocking process
// For the moment it will be available for distro import only
var sendErrorEvent = false

// Boolean that warn the backend that the next build will be cancelled
var nextBuildCanceled = false

// Boolean asking the mocked backend to fail the next build
var nextBuildShouldFail = false

// -------------------------------------------------------------------------------
// -------------------------------------------------------------------------------

// -------------------------------------------------------------------------------
// Public functions definition ---------------------------------------------------
// -------------------------------------------------------------------------------

// StartMockedRPBackend allows the test program to mock a Redpesk backend
// server
//
// :param logR: Pointer to the logger used by the application
// :type logR: *logrus.Logger
// :returns: *RPMockedBackend, error -- pointer to the mocked backend and an error if a problem has occurred
func StartMockedRPBackend(logR *logrus.Logger) (*RPMockedBackend, error) {
	// Initialize the mocked backend structure
	mockedBck := &RPMockedBackend{
		Server: http.Server{
			Addr:         serverAddress,
			ReadTimeout:  5 * time.Second,
			WriteTimeout: 5 * time.Second,
		},
		log: logR,
	}

	// ------------------------------------------
	// Define the Websocket server --------------
	// ------------------------------------------
	// Get the default WebSocket Transport
	webSocketTransport := transport.GetDefaultWebsocketTransport()
	// Create the WebSocket Server
	mockedBck.WSServer = gosocketio.NewServer(webSocketTransport)
	// Save the OnConnection callback
	mockedBck.WSServer.On(gosocketio.OnConnection, func(c *gosocketio.Channel) {
		mockedBck.log.Debugf(wsConnectionMsg)
		mockedBck.WSServerChannel = c
	})
	// Save the OnDisconnection callback
	mockedBck.WSServer.On(gosocketio.OnDisconnection, func(c *gosocketio.Channel) {
		mockedBck.log.Debugf(wsDisconnectionMsg)
	})
	// Save the OnError callback
	mockedBck.WSServer.On(gosocketio.OnError, func(c *gosocketio.Channel, err error) {
		mockedBck.log.Errorf(wsErrorErr, err)
	})
	// ------------------------------------------
	// ------------------------------------------

	// ------------------------------------------
	// Define the HTTP server -------------------
	// ------------------------------------------
	// Initialize the routes that can be accessed on the server
	router := mux.NewRouter()

	// Register handlers
	router.Handle(wSocketRoute, mockedBck.WSServer)
	router.HandleFunc(versionRoute, mockedBck.versionHandler)
	router.HandleFunc(configRoute, mockedBck.configHandler)
	router.HandleFunc(eventsRegisterRoute, mockedBck.eventRegisterHandler).Methods("POST")
	router.HandleFunc(archRoute, mockedBck.archHandler)
	router.HandleFunc(boardModelsRoute, mockedBck.boardModelsHandler)
	// Projects routes
	router.HandleFunc(prjsRoute, mockedBck.prjsListHandler).Queries("type", "{type:[a-z,]+}", "page", "{page:[0-9]+}", "per_page", "{per_page:[0-9]+}").Methods("GET")
	router.HandleFunc(prjsRoute, mockedBck.prjsAddHandler).Methods("POST")
	router.HandleFunc(prjsIDRoute, mockedBck.prjsGetHandler).Methods("GET")
	router.HandleFunc(prjsIDRoute, mockedBck.prjsPutHandler).Methods("PUT")
	router.HandleFunc(prjsIDRoute, mockedBck.prjsDeleteHandler).Methods("DELETE")
	router.HandleFunc(prjsIDRolesRoute, mockedBck.prjsGetRolesHandler).Methods("GET")
	router.HandleFunc(prjsIDMemberRoute, mockedBck.prjsAddMemberHandler).Methods("POST")
	router.HandleFunc(prjsIDMemberRoute, mockedBck.prjsGetMembersHandler).Methods("GET")
	router.HandleFunc(prjsIDMemberRoute, mockedBck.prjsPutMemberHandler).Methods("PUT")
	router.HandleFunc(prjsIDMemberRoute, mockedBck.prjsDeleteMemberHandler).Queries("id", "{id:[0-9a-z-]+}").Methods("DELETE")
	router.HandleFunc(prjsIDNewMemberRoute, mockedBck.prjsNewMembersHandler).Methods("GET")
	router.HandleFunc(prjsIDBookmarkRoute, mockedBck.prjsPostBookmarkHandler).Methods("POST")
	router.HandleFunc(prjsIDReleaseRoute, mockedBck.prjsReleaseHandler).Methods("POST")
	router.HandleFunc(prjsIDBuildConfigRoute, mockedBck.prjsGetBuildConfigHandler).Queries("arch", "{arch:[a-z0-9-_]+}", "distro", "{distro:[a-z0-9-]+}").Methods("GET")
	router.HandleFunc(prjsMngtImportRoute, mockedBck.prjsMngtImportRoute).Queries("name", "{name:[a-z0-9-]+}", "version", "{version:[a-z0-9-.]+}").Methods("POST")
	router.HandleFunc(prjsMngtRefreshRoute, mockedBck.prjsMngtRefreshRoute).Methods("PUT")
	router.HandleFunc(prjsIDExportRoute, mockedBck.prjsPostExportRoute).Methods("POST")
	router.HandleFunc(prjsIDExportRoute, mockedBck.prjsGetExportRoute).Methods("GET")
	router.HandleFunc(prjsIDExportDownloadRoute, func(res http.ResponseWriter, req *http.Request) {
		// Add a "Content-disposition" header in the answer
		res.Header().Add("Content-Disposition", "attachment; filename=mocked_project_export.zip")
		// Serve the file
		http.ServeFile(res, req, path.Join(baseDirTests, "mocked_project_export.zip"))
	})
	router.HandleFunc("/kbuild/repos/test-project_161b813e-build/latest/install_repo.sh", func(res http.ResponseWriter, req *http.Request) {
		http.ServeFile(res, req, path.Join(baseDirTests, "install_repo.sh"))
	})
	router.HandleFunc(prjsIDBuildStatsRoute, mockedBck.prjsGetBuildsStats).Methods("GET")
	router.HandleFunc(prjsIDBuildsRoute, mockedBck.prjsStartBuilds).Methods("POST")
	router.HandleFunc(prjsIDTestStatsRoute, mockedBck.prjsGetTestsStats).Methods("GET")
	router.HandleFunc(prjsImportRoute, mockedBck.prjsUploadRoute).Methods("POST")
	router.HandleFunc(prjsIDImportRoute, mockedBck.prjsImportRoute).Methods("POST")
	router.HandleFunc(prjsIDImportRoute, mockedBck.prjsGetImportRoute).Methods("GET")
	router.HandleFunc(prjsRepoInfoRoute, mockedBck.prjsGetRepoInfo).Queries("page", "{page:[0-9]+}", "per_page", "{per_page:[0-9]+}", "arch", "{arch:[a-z0-9-_]+}").Methods("GET")
	router.HandleFunc(prjsRepoInheritanceRoute, mockedBck.prjsGetInheritance).Methods("GET")
	// Release projects routes
	router.HandleFunc(relPrjsRoute, mockedBck.relPrjsGetAllHandler).Queries("page", "{page:[0-9]+}", "per_page", "{per_page:[0-9]+}").Methods("GET")
	router.HandleFunc(relPrjsIDRoute, mockedBck.relPrjsGetHandler).Methods("GET")
	router.HandleFunc(relPrjsIDRoute, mockedBck.relPrjsPutHandler).Methods("PUT")
	router.HandleFunc(relPrjsIDRoute, mockedBck.relPrjsDeleteHandler).Methods("DELETE")
	router.HandleFunc(relPrjsIDAudienceRoute, mockedBck.relPrjsPutAudienceHandler).Methods("PUT")
	router.HandleFunc(relPrjsIDAudienceRoute, mockedBck.relPrjsGetAudienceHandler).Methods("GET")
	// Deployments routes
	router.HandleFunc(relPrjsIDDeployRoute, mockedBck.deployRelProjectHandler).Methods("POST")
	router.HandleFunc(relPrjsIDDeploymentsRoute, mockedBck.deploymentGetFromRelPrjHandler).Queries("page", "{page:[0-9]+}", "per_page", "{per_page:[0-9]+}").Methods("GET")
	router.HandleFunc(deploymentsRoute, mockedBck.deploymentGetAllHandler).Queries("page", "{page:[0-9]+}", "per_page", "{per_page:[0-9]+}").Methods("GET")
	router.HandleFunc(deploymentsIDRoute, mockedBck.deploymentGetHandler).Methods("GET")
	// Roles routes
	router.HandleFunc(rolesMeRoute, mockedBck.rolesMeHandler).Methods("GET")
	// Users routes
	router.HandleFunc(usersRoute, mockedBck.usersListHandler).Queries("page", "{page:[0-9]+}", "per_page", "{per_page:[0-9]+}").Methods("GET")
	router.HandleFunc(usersRoute, mockedBck.usersAddHandler).Methods("POST")
	router.HandleFunc(usersIDRoute, mockedBck.usersGetHandler).Methods("GET")
	router.HandleFunc(usersIDRoute, mockedBck.usersPutHandler).Methods("PUT")
	router.HandleFunc(usersMeRoute, mockedBck.usersGetMeHandler).Methods("GET")
	router.HandleFunc(usersMeRoute, mockedBck.usersPutMeHandler).Methods("PUT")
	router.HandleFunc(usersEnDisRoute, mockedBck.usersEnableDisableHandler).Methods("PUT")
	// Teams routes
	router.HandleFunc(teamsRoute, mockedBck.teamsListHandler).Queries("page", "{page:[0-9]+}", "per_page", "{per_page:[0-9]+}").Methods("GET")
	router.HandleFunc(teamsRoute, mockedBck.teamsAddHandler).Methods("POST")
	router.HandleFunc(teamsIDRoute, mockedBck.teamsGetHandler).Methods("GET")
	router.HandleFunc(teamsIDRoute, mockedBck.teamsPutHandler).Methods("PUT")
	router.HandleFunc(teamsIDRoute, mockedBck.teamsDeleteHandler).Methods("DELETE")
	router.HandleFunc(teamsNewMembersRoute, mockedBck.teamsNewMemberHandler).Methods("GET")
	router.HandleFunc(teamsInvitationRoute, mockedBck.teamsInvitationHandler).Methods("POST")
	router.HandleFunc(teamsMemberRoute, mockedBck.teamsDelMemberHandler).Queries("idmb", "{id:[0-9a-z-]+}").Methods("DELETE")
	// Applications routes
	router.HandleFunc(appsRoute, mockedBck.appsListHandler).Queries("type", "{type:[a-z,]+}", "page", "{page:[0-9]+}", "per_page", "{per_page:[0-9]+}").Methods("GET")
	router.HandleFunc(prjsIDAppRoute, mockedBck.appsGetFromPrjHandler).Methods("GET")
	router.HandleFunc(prjsIDAppRoute, mockedBck.appsCreateHandler).Methods("POST")
	router.HandleFunc(appsIDRoute, mockedBck.appsGetHandler).Methods("GET")
	router.HandleFunc(appsIDRoute, mockedBck.appsPutHandler).Methods("PUT")
	router.HandleFunc(appsIDRoute, mockedBck.appsDeleteHandler).Methods("DELETE")
	router.HandleFunc(appsIDBuildsRoute, mockedBck.buildsStartHandler).Methods("POST")
	router.HandleFunc(appsIDBuildsRoute, mockedBck.buildsGetAllHandler).Queries("page", "{page:[0-9]+}", "per_page", "{per_page:[0-9]+}").Methods("GET")
	router.HandleFunc(appsIDFileRoute, mockedBck.appsPostFileHandler).Methods("POST")
	router.HandleFunc(appsListFilesRoute, mockedBck.appsGetFilesListHandler).Methods("GET")
	router.HandleFunc(appsIDFileRoute, mockedBck.appsRemoveFileHandler).Queries("name", "{name:.*}").Methods("DELETE")
	router.HandleFunc(appsIDFileRoute, mockedBck.appsDownloadFileHandler).Queries("name", "{name:.*}").Methods("GET")
	router.HandleFunc(appsIDBookmarkRoute, mockedBck.appsPostBookmarkHandler).Methods("POST")
	router.HandleFunc(appsIDServiceRoute, mockedBck.appsGetServicesHandler).Methods("GET")
	router.HandleFunc(appsSamplesRoute, mockedBck.appsGetSamplesHandler).Methods("GET")
	router.HandleFunc(appsIDAuditsRoute, mockedBck.auditsStartHandler).Methods("POST")
	router.HandleFunc(appsIDAuditsRoute, mockedBck.auditsGetAllHandler).Queries("page", "{page:[0-9]+}", "per_page", "{per_page:[0-9]+}").Methods("GET")
	router.HandleFunc(appsIDBranchesRoute, mockedBck.branchesPostHandler).Methods("POST")
	router.HandleFunc(appsIDBranchesRoute, mockedBck.branchesGetHandler).Methods("GET")
	router.HandleFunc(appsIDBranchesRoute, mockedBck.branchesDeleteHandler).Methods("DELETE")
	router.HandleFunc(appsIDActionRoute, mockedBck.actionAppPostHandler).Queries("action", "{action:[a-z,_]+}").Methods("POST")
	// Builds routes
	router.HandleFunc(buildsIDRoute, mockedBck.buildsGetHandler).Methods("GET")
	router.HandleFunc(buildsIDTestsRoute, mockedBck.testsStartHandler).Methods("POST")
	router.HandleFunc(buildsIDCancelRoute, mockedBck.buildsCancelHandler).Methods("POST")
	router.HandleFunc(buildsIDTestsRoute, mockedBck.testsGetAllFromBuildHandler).Queries("page", "{page:[0-9]+}", "per_page", "{per_page:[0-9]+}").Methods("GET")
	// Tests routes
	router.HandleFunc(testsIDRoute, mockedBck.testsGetHandler).Methods("GET")
	router.HandleFunc(testsIDLogsListRoute, mockedBck.testsGetLogsListHandler).Methods("GET")
	router.HandleFunc(testsIDLogsRoute, mockedBck.testsGetLogsHandler).Queries("name", "{name}").Methods("GET")
	router.HandleFunc(testsIDCancel, mockedBck.testsPostCancelHandler).Methods("POST")
	// Builds logs files
	router.HandleFunc("/root.log", func(res http.ResponseWriter, req *http.Request) {
		http.ServeFile(res, req, path.Join(baseDirTests, "root.log"))
	})
	router.HandleFunc("/build.log", func(res http.ResponseWriter, req *http.Request) {
		http.ServeFile(res, req, path.Join(baseDirTests, "build.log"))
	})
	router.HandleFunc("/anaconda-dbus.log", func(res http.ResponseWriter, req *http.Request) {
		http.ServeFile(res, req, path.Join(baseDirTests, "anaconda-dbus.log"))
	})
	router.HandleFunc("/anaconda-hawkey.log", func(res http.ResponseWriter, req *http.Request) {
		http.ServeFile(res, req, path.Join(baseDirTests, "anaconda-hawkey.log"))
	})
	router.HandleFunc("/anaconda-anaconda.log", func(res http.ResponseWriter, req *http.Request) {
		http.ServeFile(res, req, path.Join(baseDirTests, "anaconda-anaconda.log"))
	})
	router.HandleFunc("/Redpesk-minimal-x86_64-smack.ks", func(res http.ResponseWriter, req *http.Request) {
		http.ServeFile(res, req, path.Join(baseDirTests, "Redpesk-minimal-x86_64-smack.ks"))
	})
	router.HandleFunc("/koji-image-redpesk-my-project--redpesk-lts-arz-1.0-update-build-666.ks", func(res http.ResponseWriter, req *http.Request) {
		http.ServeFile(res, req, path.Join(baseDirTests, "koji-image-redpesk-my-project--redpesk-lts-arz-1.0-update-build-666.ks"))
	})
	// Audits routes
	router.HandleFunc(auditsToolsListRoute, mockedBck.auditsGetToolsHandler).Methods("GET")
	router.HandleFunc(auditsIDRoute, mockedBck.auditsGetHandler).Methods("GET")
	router.HandleFunc(auditsIDLogsRoute, func(res http.ResponseWriter, req *http.Request) {
		// Add a "Content-disposition" header in the answer
		res.Header().Add("Content-Disposition", "attachment; filename=mocked_audit_logs.zip")
		// Serve the file
		http.ServeFile(res, req, path.Join(baseDirTests, "mocked_audit_logs.zip"))
	})
	// Images routes
	router.HandleFunc(imgsRoute, mockedBck.imagesListHandler).Queries("type", "{type:[a-z,]+}", "page", "{page:[0-9]+}", "per_page", "{per_page:[0-9]+}").Methods("GET")
	router.HandleFunc(prjsIDImgRoute, mockedBck.imgGetFromPrjHandler).Methods("GET")
	router.HandleFunc(imgsRoute, mockedBck.imagesAddHandler).Methods("POST")
	router.HandleFunc(imgsIDRoute, mockedBck.imagesGetHandler).Methods("GET")
	router.HandleFunc(imgsIDRoute, mockedBck.imagesPutHandler).Methods("PUT")
	router.HandleFunc(imgsIDRoute, mockedBck.imagesDeleteHandler).Methods("DELETE")
	router.HandleFunc(imgsImportRoute, mockedBck.imagesImportHandler).Methods("POST")
	router.HandleFunc(imgsBuildsRoute, mockedBck.imagesBuildHandler).Methods("POST")
	router.HandleFunc(imgsBuildsRoute, mockedBck.imagesBuildsListHandler).Queries("page", "{page:[0-9]+}", "per_page", "{per_page:[0-9]+}").Methods("GET")
	router.HandleFunc(imgsIDReleaseRoute, mockedBck.imagesReleaseHandler).Methods("POST")
	// Release image routes
	router.HandleFunc(relImgsRoute, mockedBck.relImgsGetAllHandler).Queries("page", "{page:[0-9]+}", "per_page", "{per_page:[0-9]+}").Methods("GET")
	router.HandleFunc(relImgsIDRoute, mockedBck.relImgsGetHandler).Methods("GET")
	router.HandleFunc(relImgsIDRoute, mockedBck.relImgsPutHandler).Methods("PUT")
	router.HandleFunc(relImgsIDRoute, mockedBck.relImgsDeleteHandler).Methods("DELETE")
	router.HandleFunc(relImgsIDAudienceRoute, mockedBck.relImgsPutAudienceHandler).Methods("PUT")
	router.HandleFunc(relImgsIDAudienceRoute, mockedBck.relImgsGetAudienceHandler).Methods("GET")
	// Boards routes
	router.HandleFunc(boardsRoute, mockedBck.boardsListHandler).Queries("page", "{page:[0-9]+}", "per_page", "{per_page:[0-9]+}").Methods("GET")

	//set http server handler
	mockedBck.Handler = router

	// Create the logs files that may be requested as static files
	utils.WriteToFile(path.Join(baseDirTests, "tests_logs.zip"), testsLogZIPFile)
	utils.WriteToFile(path.Join(baseDirTests, "all_tests.tap"), testsLogFile)
	utils.WriteToFile(path.Join(baseDirTests, "boot.log"), testBootLogsFile)
	utils.WriteToFile(path.Join(baseDirTests, "console.log"), testConsoleLogsFile)
	utils.WriteToFile(path.Join(baseDirTests, "root.log"), rootLogFile)
	utils.WriteToFile(path.Join(baseDirTests, "build.log"), buildLogFile)
	utils.WriteToFile(path.Join(baseDirTests, "anaconda-anaconda.log"), anacondaFakeFile)
	utils.WriteToFile(path.Join(baseDirTests, "anaconda-dbus.log"), anacondaFakeFile)
	utils.WriteToFile(path.Join(baseDirTests, "anaconda-hawkey.log"), anacondaFakeFile)
	utils.WriteToFile(path.Join(baseDirTests, "Redpesk-minimal-x86_64-smack.ks"), redpeskMinimalKSFile)
	utils.WriteToFile(path.Join(baseDirTests, "koji-image-redpesk-my-project--redpesk-lts-arz-1.0-update-build-666.ks"), redpeskFullKSFile)
	utils.WriteToFile(path.Join(baseDirTests, "mocked_project_export.zip"), releasedProjectZIPFile)
	utils.WriteToFile(path.Join(baseDirTests, "install_repo.sh"), installRepoFile)
	utils.WriteToFile(path.Join(baseDirTests, "mocked_audit_logs.zip"), auditsLogsZIPFile)

	// Start the server
	go func() {
		err := mockedBck.ListenAndServe()
		if (err != nil) && !strings.Contains(err.Error(), "Server closed") {
			mockedBck.log.Errorf(exitServerRunErr, err)
			fmt.Print("A problem occurred in server !")
			errorMsg := fmt.Sprintf(exitServerRunErr, err)
			utils.WriteToFile("/tmp/mocked-backend.crash", errorMsg)
		}
	}()
	// ------------------------------------------
	// ------------------------------------------

	// Wait a bit for the server to be running
	time.Sleep(100 * time.Millisecond) // Wait a bit that above ListenAndServe function has started

	return mockedBck, nil
}

// ShutdownServer is the function allowing to shutdown correctly the
// mocked backend server
//
// :returns: error -- an error if a problem has occurred
func (mockedBck *RPMockedBackend) ShutdownServer() error {

	//Create shutdown context with 10 second timeout
	ctx, cancel := context.WithTimeout(context.Background(), 10*time.Second)
	defer cancel()

	err := mockedBck.Shutdown(ctx)
	if err != nil {
		mockedBck.log.Errorf(shutdownServerErr, err)
	}

	return err
}

// EmitEvent allows to send events to the rp-cli being in test
//
// :param method: Name of the event to send
// :type method: string
// :param args: Data to send with the event
// :type args: interface{}
// :returns: error -- an error if a problem has occurred
func (mockedBck *RPMockedBackend) EmitEvent(method string, args interface{}) error {
	if mockedBck.WSServerChannel == nil {
		return errors.Errorf("[MOCKED-BCK] nil WSServerChannel")
	}
	if !mockedBck.WSServerChannel.IsAlive() {
		return errors.Errorf("[MOCKED-BCK] WSServerChannel is not alive")
	}

	// Create an event structure
	msg := rpapiv1.EventMsg{
		Time:          time.Now().String(),
		FromSessionID: "Session test",
		Type:          method,
		Data:          args,
	}

	mockedBck.log.Debugf("[MOCKED-BCK] Backend Emit EVENT %s: %+v", method, msg)
	return mockedBck.WSServerChannel.Emit(method, msg)
}

// RequestUnauthError is the function allowing an external process to request
// for an unauthorize error on the next call on `/api/v1/config`
//
// :param textErr: Error, as text, to return in the HTTP answer
// :type textErr: string
func RequestUnauthError(textErr string) {
	tmpErrToSend = textErr
}

// RequestPrjExportFails is the function allowing an external process to request
// for a fail during project export. This variable is reset automatically
func RequestPrjExportFails() {
	prjExportFails = true
}

// RequestPrjImportFails is the function allowing an external process to request
// for a fail during project import. This variable is reset automatically
func RequestPrjImportFails() {
	prjImportFails = true
}

// GetLatestImportedProjectUUID is the function that can be called to get the
// latest imported project UUDI
func GetLatestImportedProjectUUID() uuid.UUID {
	result := uuid.Nil
	if len(allImportedProjects) > 0 {
		result = allImportedProjects[len(allImportedProjects)-1].FuturePrjID
	}
	return result
}

// RequestTimedOutTest is the function allowing an external process to request
// to have a "timed-out" test for the next test
func RequestTimedOutTest() {
	testInTimeOut = true
}

// ChangeAppState is the function allowing to request to change the state of an app
// explicitely
func ChangeAppState(newState rpapiv1.RPAppState, appSlug string) {
	var indexToModify int
	for index, app := range allApps {
		if app.Slug == appSlug {
			indexToModify = index
			break
		}
	}

	allApps[indexToModify].State = newState
}

// RequestErrorEvent is the function allowing an external process to request
// for an error event to happen during a blocking process. This variable is reset automatically
func RequestErrorEvent() {
	sendErrorEvent = true
}

// RequestNextBuildCanceled is the function allowing an external process to warn
// the mock that the next build will be canceled. Thanks to that the mock will wait longer.
// This variable is reset automatically
func RequestNextBuildCanceled() {
	nextBuildCanceled = true
}

// RequestNextBuildFailed is the function allowing an external process to ask
// the mock to fail the next application build.
// This variable is reset automatically
func RequestNextBuildFailed() {
	nextBuildShouldFail = true
}

// -------------------------------------------------------------------------------
// -------------------------------------------------------------------------------

// -------------------------------------------------------------------------------
// Handlers functions definition -------------------------------------------------
// -------------------------------------------------------------------------------

// versionHandler is the handler used when a call is done on the version route
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) versionHandler(w http.ResponseWriter, req *http.Request) {
	version := rpapiv1.RPVersion{
		ID:            uuid.FromStringOrNil(versionBackendUUID),
		Version:       bckVersion,
		APIVersion:    mockedAPIVersion,
		VersionGitTag: versionGitTag,
		Copyright:     versionCopyright,
	}
	resJSON(w, version)
}

// configHandler is the handler used when a call is done on the config route
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) configHandler(w http.ResponseWriter, req *http.Request) {
	// Check if we want to simulate a "expired token" error
	if tmpErrToSend != "" {
		// Answering with error
		resTextUnauthError(w, errors.New(tmpErrToSend))
		// Reset the temporary error
		tmpErrToSend = ""
	} else {
		// Standard way of answering
		cfg := rpapiv1.RPAPIConfig{
			ServerUID: uuid.FromStringOrNil(configBackendUUID),
			URL:       configBackendURL,
			APIURL:    configAPIURL,
			ConnRetry: configConnectRetry,
		}
		resJSON(w, cfg)
	}
}

// eventRegisterHandler is the handler used when a call is done on the events/register route
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) eventRegisterHandler(w http.ResponseWriter, req *http.Request) {
	answer := make(map[string]string)
	answer["status"] = "OK"
	resJSON(w, answer)
}

// archHandler is the handler used when a call is done on the archis route
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) archHandler(w http.ResponseWriter, req *http.Request) {
	arches := []rpapiv1.RPArch{
		{
			ID: archID1,
		},
		{
			ID: archID2,
		},
	}
	resJSON(w, arches)
}

// boardModelsHandler is the handler used when a call is done on the tests/boards route
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) boardModelsHandler(w http.ResponseWriter, req *http.Request) {
	// Get the page number wanted
	pageNb, _ := strconv.Atoi(req.FormValue("page"))
	// Get the per_page number wanted
	perPageNb, _ := strconv.Atoi(req.FormValue("per_page"))
	// Get the potential "filters" string
	filtersStr := req.FormValue("filters")

	// Get the name filter
	nameFilter := ""
	if strings.Contains(filtersStr, "name=") {
		nameFilter = strings.ReplaceAll(filtersStr, "name=", "")
	}

	boardModels := rpapiv1.RPModelBoards{
		Page:    pageNb,
		PerPage: perPageNb,
	}

	// Calculate the minimum and maximum index wanted by the user
	minIndex := (pageNb - 1) * perPageNb
	maxIndex := (pageNb * perPageNb) - 1

	// Loop over projects and save the ones we want to send
	count := 0
	for _, boardModel := range allBoardModels {
		if (count >= minIndex) && (count <= maxIndex) && (nameFilter == "" || boardModel.Name == nameFilter) {
			boardModels.ModelBoards = append(boardModels.ModelBoards, boardModel)
		}
		count++
	}

	// Update the total entries
	boardModels.TotalEntries = len(allBoardModels)

	resJSON(w, boardModels)
}

// boardsListHandler is the handler used when someone wants the list of boards
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) boardsListHandler(w http.ResponseWriter, req *http.Request) {
	// Get the page number wanted
	pageNb, _ := strconv.Atoi(req.FormValue("page"))
	// Get the per_page number wanted
	perPageNb, _ := strconv.Atoi(req.FormValue("per_page"))
	// Get the potential "filters" string
	filtersStr := req.FormValue("filters")

	// Get the name filter
	nameFilter := ""
	if strings.Contains(filtersStr, "name=") {
		nameFilter = strings.ReplaceAll(filtersStr, "name=", "")
	}

	boards := rpapiv1.RPBoards{
		Page:    pageNb,
		PerPage: perPageNb,
	}

	// Calculate the minimum and maximum index wanted by the user
	minIndex := (pageNb - 1) * perPageNb
	maxIndex := (pageNb * perPageNb) - 1

	// Loop over projects and save the ones we want to send
	count := 0
	for _, board := range allBoards {
		if (count >= minIndex) && (count <= maxIndex) && (nameFilter == "" || board.Name == nameFilter) {
			boards.Boards = append(boards.Boards, board)
		}
		count++
	}

	// Update the total entries
	boards.TotalEntries = len(allBoards)

	resJSON(w, boards)
}

// prjsListHandler is the handler used when a call is done on the projects route
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) prjsListHandler(w http.ResponseWriter, req *http.Request) {

	// Get the type wanted by the user
	typesWanted := req.FormValue("type")
	if typesWanted == "" {
		typesWanted = "standard"
	}

	// Get the page number wanted
	pageNb, _ := strconv.Atoi(req.FormValue("page"))

	// Get the per_page number wanted
	perPageNb, _ := strconv.Atoi(req.FormValue("per_page"))

	// Get the potential "filters" string
	filtersStr := req.FormValue("filters")

	// Get the bookmark request
	bookmarkWanted := (strings.Contains(filtersStr, "bookmarked==true"))

	// Get the filters
	slugFilter := ""
	if strings.HasPrefix(filtersStr, "slug=") {
		slugFilter = strings.ReplaceAll(filtersStr, "slug=", "")
	}
	nameFilter := ""
	if strings.HasPrefix(filtersStr, "name=") {
		nameFilter = strings.ReplaceAll(filtersStr, "name=", "")
		nameFilter = strings.ReplaceAll(nameFilter, "+", " ")
	}

	// Initialize the structure that will be returned
	retProjects := rpapiv1.RPProjects{
		Page:    pageNb,
		PerPage: perPageNb,
	}

	// Calculate the minimum and maximum index wanted by the user
	minIndex := (pageNb - 1) * perPageNb
	maxIndex := (pageNb * perPageNb) - 1

	// Convert the type(s) to a slice of strings
	typesWantedSli := strings.Split(typesWanted, ",")

	// Loop over projects and save the ones we want to send
	totalPrjs := 0
	for _, project := range allProjects {
		// 'If' on the filters given by the user
		if utils.ContainsString(typesWantedSli, string(project.Type)) {
			if (!bookmarkWanted || project.Bookmarked) &&
				(slugFilter == "" || project.Slug == slugFilter) &&
				(nameFilter == "" || project.Name == nameFilter) {
				// 'If' on the object index
				if (totalPrjs >= minIndex) && (totalPrjs <= maxIndex) {
					retProjects.Projects = append(retProjects.Projects, convertProjectFullToProject(project))
				}

			}
			totalPrjs++
		}
	}

	// Update the total entries
	retProjects.TotalEntries = totalPrjs

	resJSON(w, retProjects)
}

// prjsAddHandler is the handler used when a call is done on the projects route
// with the POST method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) prjsAddHandler(w http.ResponseWriter, req *http.Request) {
	// Get the request body
	reqBody := &rpapiv1.RPProjectFull{}
	if err := reqBind(req, reqBody); err != nil {
		resJSONErrorE(w, err)
		return
	}

	// Update the creation and update time
	timeNow := time.Now()
	reqBody.CreatedAt = timeNow
	reqBody.UpdatedAt = timeNow
	reqBody.State = rpapiv1.RPPrjStableState

	// Update the repository with what is needed for the get command
	for _, distro := range reqBody.Distributions {
		if distro.Set {
			reqBody.Repos = append(reqBody.Repos, rpapiv1.RPRepo{
				Distro:  distro,
				Build:   fmt.Sprintf("http://127.0.0.1%s/kbuild/repos/test-project_161b813e-build/latest/$basearch", serverAddress),
				Release: fmt.Sprintf("http://127.0.0.1%s/kbuild/repos/test-project_161b813e-release/latest/$basearch", serverAddress),
			})
		}
	}

	// Set the UUID
	reqBody.ID, _ = uuid.NewV4()

	// Set the slug from the name
	reqBody.Slug = strings.ToLower(strings.ReplaceAll(reqBody.Name, " ", "-"))

	// Set the type of project to "standard"
	reqBody.Type = "standard"

	// Set default supported object
	if !(reqBody.SupportedObjects.Applications || reqBody.SupportedObjects.Images || reqBody.SupportedObjects.Containers) {
		reqBody.SupportedObjects.Applications = true
	}

	// Create a JWT
	jwtID, _ := uuid.NewV4()
	jwtValue1, _ := uuid.NewV4()
	jwtValue2, _ := uuid.NewV4()
	reqBody.ProjectAccessTokens = []rpapiv1.RPObjectAccessToken{
		{
			ID:  jwtID,
			JWT: jwtValue1.String() + "_" + jwtValue2.String(),
		},
	}

	// Add the projects in our list
	allProjects = append(allProjects, *reqBody)

	// Wait a bit and send the corresponding event
	go func() {
		time.Sleep(genericWaitingTime)
		mockedBck.EmitEvent(rpapiv1.EVTProjectAdd, reqBody)
	}()

	resJSON(w, reqBody)
}

// prjsGetHandler is the handler used when a call is done on the projects/{id} route
// with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) prjsGetHandler(w http.ResponseWriter, req *http.Request) {
	prjID := reqParamGet(req, "prjid")
	prjUUID := uuid.FromStringOrNil(prjID)

	for _, prj := range allProjects {
		if utils.UUIDEqual(prj.ID, prjUUID) {
			prj.FullData = true
			resJSON(w, prj)
			break
		}
	}
}

// prjsPutHandler is the handler used when a call is done on the projects/{id} route
// with the PUT method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) prjsPutHandler(w http.ResponseWriter, req *http.Request) {
	prjID := reqParamGet(req, "prjid")
	prjUUID := uuid.FromStringOrNil(prjID)

	// Get the possible action (in query parameter)
	action := req.FormValue("action")

	var indexToModify int
	for index, prj := range allProjects {
		if utils.UUIDEqual(prj.ID, prjUUID) {
			// We found the right project, save the index to modify it afterwards
			indexToModify = index
			break
		}
	}

	if action == "change_build_type" {
		// Get the request body
		reqBody := &rpapiv1.RPChangeBuildType{}
		if err := reqBind(req, reqBody); err != nil {
			resJSONErrorE(w, err)
			return
		}

		// Create a slice containing all the UUID of the app to update
		appUUIDs := []string{}
		for _, appToUpdate := range reqBody.Applications {
			appUUIDs = append(appUUIDs, appToUpdate.ApplicationID.String())
		}

		// Now change the build type for every app listed (or all if the slice is empty)
		changeAllApps := (len(appUUIDs) == 0)
		for index, currentApp := range allApps {
			if utils.UUIDEqual(currentApp.Parent.ProjectID, prjUUID) && (changeAllApps || utils.ContainsString(appUUIDs, currentApp.ID.String())) {
				allApps[index].Metadata.Data.BuildType = reqBody.BuildType
			}
		}

	} else {
		// Get the request body
		reqBody := &rpapiv1.RPProjectFull{}
		if err := reqBind(req, reqBody); err != nil {
			resJSONErrorE(w, err)
			return
		}

		// Modify the project
		allProjects[indexToModify].Name = reqBody.Name
		allProjects[indexToModify].Slug = reqBody.Slug
		allProjects[indexToModify].Description = reqBody.Description
		allProjects[indexToModify].EmulatedBuild = reqBody.EmulatedBuild
		allProjects[indexToModify].SubVersion = reqBody.SubVersion

		// Distributions
		allProjects[indexToModify].Distributions = []rpapiv1.RPDistro{}
		allProjects[indexToModify].Distributions = append(allProjects[indexToModify].Distributions, reqBody.Distributions...)

		// Architectures
		allProjects[indexToModify].Architectures = []rpapiv1.RPArch{}
		allProjects[indexToModify].Architectures = append(allProjects[indexToModify].Architectures, reqBody.Architectures...)

		// Inheritances
		allProjects[indexToModify].Inheritances = rpapiv1.RPInheritances{}
		allProjects[indexToModify].Inheritances = append(allProjects[indexToModify].Inheritances, reqBody.Inheritances...)
	}

	// Wait a bit and send the corresponding event
	go func() {
		time.Sleep(genericWaitingTime)
		mockedBck.EmitEvent(rpapiv1.EVTProjectStateChange, allProjects[indexToModify])
	}()

	resJSON(w, allProjects[indexToModify])
}

// prjsDeleteHandler is the handler used when a call is done on the projects/{id} route
// with the DELETE method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) prjsDeleteHandler(w http.ResponseWriter, req *http.Request) {
	prjID := reqParamGet(req, "prjid")
	prjUUID := uuid.FromStringOrNil(prjID)

	tmpPrjs := []rpapiv1.RPProjectFull{}
	resPrj := rpapiv1.RPProjectFull{}
	for _, prj := range allProjects {
		if !utils.UUIDEqual(prj.ID, prjUUID) {
			tmpPrjs = append(tmpPrjs, prj)
		} else {
			// Copy the project we will need to send at the end
			copier.Copy(&resPrj, &prj)
		}
	}

	// Set the new tab of projects
	allProjects = tmpPrjs

	// Wait a bit and send the corresponding event
	go func() {
		time.Sleep(genericWaitingTime)
		mockedBck.EmitEvent(rpapiv1.EVTProjectDelete, resPrj)
	}()

	resJSON(w, resPrj)
}

// prjsPostBookmarkHandler is the handler used when a call is done on the projects/{prjid}/bookmark route
// with the POST method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) prjsPostBookmarkHandler(w http.ResponseWriter, req *http.Request) {
	prjID := reqParamGet(req, "prjid")
	prjUUID := uuid.FromStringOrNil(prjID)

	indexToModify := -1
	for index, prj := range allProjects {
		if utils.UUIDEqual(prj.ID, prjUUID) {
			// We found the right project, save the index to modify it afterwards
			indexToModify = index
			break
		}
	}

	if indexToModify == -1 {
		resJSONErrorE(w, errors.New("The project does not exists"))
		return
	}

	// Modify the project
	allProjects[indexToModify].Bookmarked = !allProjects[indexToModify].Bookmarked

	// Create a UUID and a time object
	bookID, _ := uuid.NewV4()
	timeNow := time.Now()

	// Prepare the returned structure
	prjBookMark := rpapiv1.RPProjectBookmark{
		ID:          bookID,
		CreatedAt:   timeNow,
		UpdatedAt:   timeNow,
		ProjectID:   allProjects[indexToModify].ID,
		ProjectSlug: allProjects[indexToModify].Slug,
	}

	// Wait a bit and send the corresponding event
	go func() {
		time.Sleep(genericWaitingTime)
		mockedBck.EmitEvent(rpapiv1.EVTProjectStateChange, allProjects[indexToModify])
	}()

	resJSON(w, prjBookMark)
}

// prjsReleaseHandler is the handler used when a call is done on the projects/{prjid}/release route
// with the PUT method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) prjsReleaseHandler(w http.ResponseWriter, req *http.Request) {

	prjID := reqParamGet(req, "prjid")
	prjUUID := uuid.FromStringOrNil(prjID)

	prjIndex := -1
	for index, prj := range allProjects {
		if utils.UUIDEqual(prj.ID, prjUUID) {
			// We found the right project, save the index to modify it afterwards
			prjIndex = index
			break
		}
	}

	if prjIndex == -1 {
		resJSONErrorE(w, errors.New("The project does not exists"))
		return
	}

	// Get the request body
	reqBody := &rpapiv1.RPReleaseProjectFull{}
	if err := reqBind(req, reqBody); err != nil {
		resJSONErrorE(w, err)
		return
	}

	// Fill the structure to store in the "released-projects"
	reqBody.ID, _ = uuid.NewV4()
	reqBody.Slug = allProjects[prjIndex].Slug + "-" + reqBody.Version
	reqBody.Name = allProjects[prjIndex].Name
	timeNow := time.Now()
	reqBody.CreatedAt = timeNow
	reqBody.UpdatedAt = timeNow
	reqBody.Type = rpapiv1.RPProjectTypeReleased
	reqBody.State = rpapiv1.RPPrjStableState
	reqBody.ProjectReferenceID = allProjects[prjIndex].ID
	reqBody.ProjectReferenceSlug = allProjects[prjIndex].Slug

	// Create a JWT
	jwtID, _ := uuid.NewV4()
	jwtValue1, _ := uuid.NewV4()
	jwtValue2, _ := uuid.NewV4()
	reqBody.ProjectAccessTokens = []rpapiv1.RPObjectAccessToken{
		{
			ID:  jwtID,
			JWT: jwtValue1.String() + "_" + jwtValue2.String(),
		},
	}

	// Fill the applications
	for _, app := range allApps {
		if utils.UUIDEqual(app.Parent.ProjectID, reqBody.ProjectReferenceID) {
			tmpAppReleased := rpapiv1.RPAppsReleased{
				ApplicationID: app.ID,
				PackageName:   app.PackageName,
				ShaGitPkg:     "670cdb57ab4b763bcab22e13811af5ece099218bd",
				Branch:        allProjects[prjIndex].Slug + "_c7617dd3_670cdb57ab4b763bcab22e13811af5ece099218bd_redpesk",
				BuildID:       "100000",
			}
			reqBody.Apps = append(reqBody.Apps, tmpAppReleased)
		}
	}

	// Fill the full release info
	// Not necessary for tests

	// Add the released project in the array
	allReleasedProjects = append(allReleasedProjects, *reqBody)

	// Add a new audience in our audience list
	tmpAudience := rpapiv1.RPObjectRoles{
		ObjectID:           reqBody.ID,
		Stack:              false,
		IdentityRolesLists: make(map[string]*rpapiv1.RPIdentityRolesLists),
	}
	// Add all the users as available
	for _, user := range allUsers.Users {
		tmpRolesList := rpapiv1.RPIdentityRolesLists{
			ID:           user.ID,
			Type:         rpapiv1.RPMemberUserType,
			PseudoOrSlug: user.Pseudo,
			AvailableRoles: &[]rpapiv1.RPRole{
				{
					Name:   "clientRelease of release-project",
					Label:  "clientRelease",
					Object: "release-project",
				},
			},
		}
		tmpAudience.IdentityRolesLists[user.ID.String()] = &tmpRolesList
	}

	// Add all the teams as available
	for _, team := range allTeams.Teams {
		tmpRolesList := rpapiv1.RPIdentityRolesLists{
			ID:           team.ID,
			Type:         rpapiv1.RPMemberTeamType,
			PseudoOrSlug: team.Slug,
			AvailableRoles: &[]rpapiv1.RPRole{
				{
					Name:   "clientRelease of release-project",
					Label:  "clientRelease",
					Object: "release-project",
				},
			},
		}
		tmpAudience.IdentityRolesLists[team.ID.String()] = &tmpRolesList
	}

	// Add the audience in the global audience list
	allRelProjectsAudience = append(allRelProjectsAudience, tmpAudience)

	// Wait a bit to send the events
	go func(projectIndex, relProjectIndex int) {
		// Wait a bit
		time.Sleep(genericWaitingTime)

		// Send the corresponding event
		mockedBck.EmitEvent(rpapiv1.EVTProjectStateChange, allProjects[projectIndex])
		mockedBck.EmitEvent(rpapiv1.EVTReleaseProjectStateChange, allReleasedProjects[relProjectIndex])

	}(prjIndex, len(allReleasedProjects)-1)

	resJSON(w, reqBody)
}

// relPrjsGetAllHandler is the handler used when a call is done on the release-projects route
// with a GET
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) relPrjsGetAllHandler(w http.ResponseWriter, req *http.Request) {

	// Get the page number wanted
	pageNb, _ := strconv.Atoi(req.FormValue("page"))

	// Get the per_page number wanted
	perPageNb, _ := strconv.Atoi(req.FormValue("per_page"))

	// Get the potential "filters" string
	filtersStr := req.FormValue("filters")

	// Get the filters
	slugFilter := ""
	if strings.HasPrefix(filtersStr, "slug=") {
		slugFilter = strings.ReplaceAll(filtersStr, "slug=", "")
	}
	nameFilter := ""
	if strings.HasPrefix(filtersStr, "name=") {
		nameFilter = strings.ReplaceAll(filtersStr, "name=", "")
		nameFilter = strings.ReplaceAll(nameFilter, "+", " ")
	}

	// Create the object to return
	relPrjsToReturn := rpapiv1.RPReleaseProjects{
		Page:    pageNb,
		PerPage: perPageNb,
	}

	// Calculate the minimum and maximum index wanted by the user
	minIndex := (pageNb - 1) * perPageNb
	maxIndex := (pageNb * perPageNb) - 1

	// Return only the released projects wanted by the client on this request
	count := 0
	for _, project := range allReleasedProjects {
		if (count >= minIndex) && (count <= maxIndex) &&
			(slugFilter == "" || project.Slug == slugFilter) &&
			(nameFilter == "" || project.Name == nameFilter) {
			relPrjsToReturn.ReleaseProjects = append(relPrjsToReturn.ReleaseProjects, project.RPReleaseProject)
		}
		count++
	}

	resJSON(w, relPrjsToReturn)
}

// relPrjsGetHandler is the handler used when a call is done on the release-projects/{id} route
// with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) relPrjsGetHandler(w http.ResponseWriter, req *http.Request) {
	prjID := reqParamGet(req, "prjid")
	prjUUID := uuid.FromStringOrNil(prjID)

	for _, prj := range allReleasedProjects {
		if utils.UUIDEqual(prj.ID, prjUUID) {
			prj.FullData = true
			resJSON(w, prj)
			break
		}
	}
}

// relPrjsPutHandler is the handler used when a call is done on the release-projects/{id} route
// with the PUT method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) relPrjsPutHandler(w http.ResponseWriter, req *http.Request) {
	prjID := reqParamGet(req, "prjid")
	prjUUID := uuid.FromStringOrNil(prjID)

	prjIndex := -1
	for index, prj := range allReleasedProjects {
		if utils.UUIDEqual(prj.ID, prjUUID) {
			// We found the right project, save the index to modify it afterwards
			prjIndex = index
			break
		}
	}

	if prjIndex == -1 {
		resJSONErrorE(w, errors.New("The released project does not exists"))
		return
	}

	// Get the request body
	reqBody := &rpapiv1.RPReleaseProject{}
	if err := reqBind(req, reqBody); err != nil {
		resJSONErrorE(w, err)
		return
	}

	// Modify the object
	allReleasedProjects[prjIndex].Name = reqBody.Name
	allReleasedProjects[prjIndex].Description = reqBody.Description

	// Send the result
	resJSON(w, allReleasedProjects[prjIndex])
}

// relPrjsDeleteHandler is the handler used when a call is done on the release-projects/{id} route
// with the DELETE method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) relPrjsDeleteHandler(w http.ResponseWriter, req *http.Request) {
	prjID := reqParamGet(req, "prjid")
	prjUUID := uuid.FromStringOrNil(prjID)

	tmpPrjs := []rpapiv1.RPReleaseProjectFull{}
	resPrj := rpapiv1.RPReleaseProjectFull{}
	for _, prj := range allReleasedProjects {
		if !utils.UUIDEqual(prj.ID, prjUUID) {
			tmpPrjs = append(tmpPrjs, prj)
		} else {
			// Copy the project we will need to send at the end
			copier.Copy(&resPrj, &prj)
		}
	}

	// Set the new tab of released projects
	allReleasedProjects = tmpPrjs

	resJSON(w, resPrj)
}

// relPrjsPutAudienceHandler is the handler used when a call is done on the release-projects/{id}/audience route
// with the PUT method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) relPrjsPutAudienceHandler(w http.ResponseWriter, req *http.Request) {
	// Get the concerned release-project ID
	prjID := reqParamGet(req, "prjid")
	prjUUID := uuid.FromStringOrNil(prjID)

	// Get the request body
	reqBody := &rpapiv1.RPObjectRoles{}
	if err := reqBind(req, reqBody); err != nil {
		resJSONErrorE(w, err)
		return
	}

	// Find the release project to modify
	indexToModify := -1
	for index, audience := range allRelProjectsAudience {
		if utils.UUIDEqual(audience.ObjectID, prjUUID) {
			indexToModify = index
		}
	}

	if indexToModify == -1 {
		resJSONErrorE(w, errors.New("The released project audience does not exists"))
		return
	}

	// Modify this index
	allRelProjectsAudience[indexToModify] = *reqBody

	// Return the updated value
	resJSON(w, allRelProjectsAudience[indexToModify])
}

// relPrjsGetAudienceHandler is the handler used when a call is done on the release-projects/{id}/audience route
// with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) relPrjsGetAudienceHandler(w http.ResponseWriter, req *http.Request) {
	// Get the concerned release-project ID
	prjID := reqParamGet(req, "prjid")
	prjUUID := uuid.FromStringOrNil(prjID)

	// Find the release project to return
	for _, audience := range allRelProjectsAudience {
		if utils.UUIDEqual(audience.ObjectID, prjUUID) {
			resJSON(w, audience)
			return
		}
	}
	resJSONErrorE(w, errors.New("The released project audience does not exists"))
}

// deployRelProjectHandler is the handler used when a call is done on the release-projects/{prjid}/deploy route
// with the POST method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) deployRelProjectHandler(w http.ResponseWriter, req *http.Request) {

	// Get the UUID of the released project
	prjID := reqParamGet(req, "prjid")
	prjUUID := uuid.FromStringOrNil(prjID)

	// Check that the released project specified exists
	relPrjIndex := -1
	for index, prj := range allReleasedProjects {
		if utils.UUIDEqual(prj.ID, prjUUID) {
			// We found the right released project, save the index to modify it afterwards
			relPrjIndex = index
			break
		}
	}
	if relPrjIndex == -1 {
		resJSONErrorE(w, errors.New("The released project does not exists"))
		return
	}

	// Get the request body
	reqBody := &rpapiv1.RPDeploy{}
	if err := reqBind(req, reqBody); err != nil {
		resJSONErrorE(w, err)
		return
	}

	// Create the deployment structure to save
	dplUUID, _ := uuid.NewV4()
	timeNow := time.Now()
	currentDeployment := rpapiv1.RPDeploymentFull{
		RPDeployment: rpapiv1.RPDeployment{
			ID:                 dplUUID,
			Name:               "",
			CreatedAt:          timeNow,
			UpdatedAt:          timeNow,
			Status:             rpapiv1.RPDeploymentGlobalStatusPending,
			UserLauncherPseudo: "the_banano_",
			ProjectDeployed: rpapiv1.RPAppParent{
				ProjectID:   allReleasedProjects[relPrjIndex].ID,
				ProjectSlug: allReleasedProjects[relPrjIndex].Slug,
				Type:        allReleasedProjects[relPrjIndex].Type,
			},
			DeploymentsPerBoard: []rpapiv1.RPDeploymentPerBoard{},
			Apps:                []rpapiv1.RPAppsReleased{},
		},
	}

	// Now, fill the DeploymentsPerBoard part
	for _, boardUUID := range reqBody.BoardsID {
		// Look for the corresponding board
		for _, board := range allBoards {
			if utils.UUIDEqual(board.ID, boardUUID) {
				// Add the board in the deployment structure
				tmpDplPerBoard := rpapiv1.RPDeploymentPerBoard{
					Board: rpapiv1.RPBoard{
						ID:   board.ID,
						Name: board.Name,
					},
					Status: rpapiv1.RPDeploymentStatusPending,
					Logs:   "",
				}
				currentDeployment.DeploymentsPerBoard = append(currentDeployment.DeploymentsPerBoard, tmpDplPerBoard)
			}
		}
	}

	// Now, fill the application
	currentDeployment.Apps = append(currentDeployment.Apps, allReleasedProjects[relPrjIndex].Apps...)

	// Add the deployment in the array
	allDeployments = append(allDeployments, currentDeployment)

	// Wait a bit to send the events
	go func(deploymentIndex int) {

		// Create a function allowing to update every board status
		updateBoardStatus := func(newStatus rpapiv1.RPDeploymentStatus) {
			for curIndex := range allDeployments[deploymentIndex].DeploymentsPerBoard {
				allDeployments[deploymentIndex].DeploymentsPerBoard[curIndex].Status = newStatus
			}
		}

		// Update the value of Start time
		allDeployments[deploymentIndex].CreatedTS = time.Now()

		// Wait a bit
		time.Sleep(genericWaitingTime)

		// Change the global status to "in progress"
		allDeployments[deploymentIndex].Status = rpapiv1.RPDeploymentGlobalStatusInProgress

		// Send the corresponding event
		mockedBck.EmitEvent(rpapiv1.EVTDeployReleasedProjectStateChange, allDeployments[deploymentIndex])

		// Wait a bit
		time.Sleep(2 * genericWaitingTime)

		// Update each board status to downloading
		updateBoardStatus(rpapiv1.RPDeploymentStatusDownloading)

		// Send the corresponding event
		mockedBck.EmitEvent(rpapiv1.EVTDeployReleasedProjectStateChange, allDeployments[deploymentIndex])

		// Wait a bit
		time.Sleep(3 * genericWaitingTime)

		// Update each board status to installing
		updateBoardStatus(rpapiv1.RPDeploymentStatusInstalling)

		// Send the corresponding event
		mockedBck.EmitEvent(rpapiv1.EVTDeployReleasedProjectStateChange, allDeployments[deploymentIndex])

		// Wait a bit
		time.Sleep(1 * genericWaitingTime)

		// Update each board status to installing
		updateBoardStatus(rpapiv1.RPDeploymentStatusSuccess)

		// Change the global status to "finished"
		allDeployments[deploymentIndex].Status = rpapiv1.RPDeploymentGlobalStatusFinished

		// Update the value of End time
		allDeployments[deploymentIndex].FinishedTS = time.Now()

		// Send the corresponding event
		mockedBck.EmitEvent(rpapiv1.EVTDeployReleasedProjectStateChange, allDeployments[deploymentIndex])

	}(len(allDeployments) - 1)

	resJSON(w, currentDeployment)
}

// deploymentGetFromRelPrjHandler is the handler used when a call is done on the release-projects/{prjid}/deployments route
// with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) deploymentGetFromRelPrjHandler(w http.ResponseWriter, req *http.Request) {
	// Get the released project ID specified
	relPrjID := reqParamGet(req, "prjid")
	relPrjUUID := uuid.FromStringOrNil(relPrjID)

	// Get the page number wanted
	pageNb, _ := strconv.Atoi(req.FormValue("page"))

	// Get the per_page number wanted
	perPageNb, _ := strconv.Atoi(req.FormValue("per_page"))

	// Create the object to return
	deploymentsToReturn := rpapiv1.RPDeployments{
		Page:    pageNb,
		PerPage: perPageNb,
	}

	// Calculate the minimum and maximum index wanted by the user
	minIndex := (pageNb - 1) * perPageNb
	maxIndex := (pageNb * perPageNb) - 1

	// Return only the deployments wanted by the client on this request
	count := 0
	for _, dpl := range allDeployments {
		if (count >= minIndex) && (count <= maxIndex) && (utils.UUIDEqual(relPrjUUID, dpl.ProjectDeployed.ProjectID)) {
			deploymentsToReturn.Deployments = append(deploymentsToReturn.Deployments, dpl.RPDeployment)
			count++
		}
	}

	resJSON(w, deploymentsToReturn)
}

// deploymentGetAllHandler is the handler used when a call is done on the deployments route
// with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) deploymentGetAllHandler(w http.ResponseWriter, req *http.Request) {
	// Get the page number wanted
	pageNb, _ := strconv.Atoi(req.FormValue("page"))

	// Get the per_page number wanted
	perPageNb, _ := strconv.Atoi(req.FormValue("per_page"))

	// Create the object to return
	deploymentsToReturn := rpapiv1.RPDeployments{
		Page:    pageNb,
		PerPage: perPageNb,
	}

	// Calculate the minimum and maximum index wanted by the user
	minIndex := (pageNb - 1) * perPageNb
	maxIndex := (pageNb * perPageNb) - 1

	// Return only the deployments wanted by the client on this request
	count := 0
	for _, dpl := range allDeployments {
		if (count >= minIndex) && (count <= maxIndex) {
			deploymentsToReturn.Deployments = append(deploymentsToReturn.Deployments, dpl.RPDeployment)
		}
		count++
	}

	resJSON(w, deploymentsToReturn)
}

// deploymentGetHandler is the handler used when a call is done on the deployments/{dplyid} route
// with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) deploymentGetHandler(w http.ResponseWriter, req *http.Request) {
	// Get the deployment ID specified
	dplID := reqParamGet(req, "dplyid")
	dplUUID := uuid.FromStringOrNil(dplID)

	// Now look for the corresponding deployment in our array
	for _, dpl := range allDeployments {
		if utils.UUIDEqual(dpl.ID, dplUUID) {
			dpl.FullData = true
			resJSON(w, dpl)
			break
		}
	}
}

// prjsGetBuildConfigHandler is the handler used when a call is done on the projects/{prjid}/buildconfig route
// with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) prjsGetBuildConfigHandler(w http.ResponseWriter, req *http.Request) {
	prjID := reqParamGet(req, "prjid")
	prjUUID := uuid.FromStringOrNil(prjID)

	// Check that the project exists
	projectIndex := -1
	for index, prj := range allProjects {
		if utils.UUIDEqual(prj.ID, prjUUID) {
			// We found the right project
			projectIndex = index
			break
		}
	}
	if projectIndex == -1 {
		resJSONErrorE(w, errors.New("The project does not exists"))
		return
	}

	// Get the arch and the distro wanted
	archWanted := req.FormValue("arch")
	distrWanted := req.FormValue("distro")

	// Create the file object to return
	base64FileContent := base64.StdEncoding.EncodeToString([]byte(configBuildFile))
	slugIDName := allProjects[projectIndex].Slug + "_" + allProjects[projectIndex].ID.String()[:8]
	resFileObj := rpapiv1.RPFile_Deprecated{
		Name:     strings.Join([]string{string(slugIDName), distrWanted, archWanted}, "_") + ".cfg",
		Encoding: "base64",
		Content:  base64FileContent,
	}

	resJSON(w, &resFileObj)
}

// prjsMngtImportRoute is the handler used when a call is done on the projects-mngt/import route
// with the POST method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) prjsMngtImportRoute(w http.ResponseWriter, req *http.Request) {

	// Get the request body
	reqBody := &rpapiv1.RPProjectFull{}
	if err := reqBind(req, reqBody); err != nil {
		resJSONErrorE(w, err)
		return
	}

	// Get the name wanted by the user
	nameWanted := req.FormValue("name")

	// Get the version wanted by the user
	versionWanted := req.FormValue("version")

	// Get the sub-version wanted by the user
	subVersionWanted := req.FormValue("sub_version")

	// Initialized the project that will be created
	newPrj := &rpapiv1.RPProjectFull{
		RPProject: rpapiv1.RPProject{
			Description: reqBody.Description,
		},
	}

	// Update the creation and update time
	timeNow := time.Now()
	newPrj.CreatedAt = timeNow
	newPrj.UpdatedAt = timeNow
	newPrj.State = rpapiv1.RPPrjStableState

	// Set the UUID
	newPrj.ID, _ = uuid.NewV4()

	// Set the name
	newPrj.Name = nameWanted

	// Set the slug from the name
	newPrj.Slug = strings.ToLower(strings.ReplaceAll(newPrj.Name, " ", "-"))

	// Set the version and SubVersion
	newPrj.Version = versionWanted
	newPrj.SubVersion = subVersionWanted

	// Set the type of project to "external"
	newPrj.Type = "external"

	newPrj.SupportedObjects.Applications = true

	// Set the architectures of the project
	newPrj.Architectures = []rpapiv1.RPArch{
		{
			ID:        archID1,
			Set:       true,
			Mandatory: true,
		},
		{
			ID:        archID2,
			Set:       true,
			Mandatory: true,
		},
	}

	// Set the number of RPM of the project
	newPrj.NbRPMsPerArch = map[string]int{
		"aarch64": 2,
		"x86_64":  2,
	}

	// Add the projects in our list
	allProjects = append(allProjects, *newPrj)

	// Wait a bit to send the events
	go func() {
		time.Sleep(genericWaitingTime)
		mockedBck.EmitEvent(rpapiv1.EVTProjectAdd, newPrj)
		time.Sleep(genericWaitingTime)
		// Check if the user requested an error to happen
		if sendErrorEvent {
			// Send the error event
			rpErr := rpapiv1.EventError{
				ObjectType: rpapiv1.EvtTypeProject,
				ObjectID:   newPrj.ID,
				ObjectSlug: newPrj.Slug,
				RpError: rpapiv1.RPError{
					Status: "failed",
					Error:  "This is a mocked error for test",
				},
			}
			mockedBck.EmitEvent(rpapiv1.EVTError, rpErr)
			// Reset the value of the global variable
			sendErrorEvent = false
		} else {
			mockedBck.EmitEvent(rpapiv1.EVTProjectStateChange, allProjects[len(allProjects)-1])
		}
	}()

	resJSON(w, newPrj)
}

// prjsMngtRefreshRoute is the handler used when a call is done on the projects-mngt/{id}/refresh route
// with the PUT method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) prjsMngtRefreshRoute(w http.ResponseWriter, req *http.Request) {
	prjID := reqParamGet(req, "prjid")
	prjUUID := uuid.FromStringOrNil(prjID)

	var indexToModify int
	for index, prj := range allProjects {
		if utils.UUIDEqual(prj.ID, prjUUID) {
			// We found the right project, save the index to modify it afterwards
			indexToModify = index
			break
		}
	}

	// Modify the project
	allProjects[indexToModify].Version += "-refreshed"

	// Wait a bit to send the EVTProjectStateChange event
	go func(index int) {
		time.Sleep(genericWaitingTime)
		mockedBck.EmitEvent(rpapiv1.EVTProjectStateChange, allProjects[index])
	}(indexToModify)

	resJSON(w, allProjects[indexToModify])
}

// prjsPostExportRoute is the handler used when a call is done on the projects/{id}/export route
// with the POST method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) prjsPostExportRoute(w http.ResponseWriter, req *http.Request) {
	// Get the UUID of the concerned project
	prjID := reqParamGet(req, "prjid")
	prjUUID := uuid.FromStringOrNil(prjID)

	// Check that the project exists
	prjExists := false
	for _, prj := range allProjects {
		if utils.UUIDEqual(prj.ID, prjUUID) {
			// We found the project
			prjExists = true
			break
		}
	}

	if !prjExists {
		resJSONErrorE(w, errors.New("The project does not exists"))
		return
	}

	// Create the new array entry
	newExpPrj := rpapiv1.RPExportProjectState{
		ProjectID: prjUUID,
		StartDate: time.Now(),
		Status:    rpapiv1.RPStateExportProjectExporting,
	}

	// Add the project in the array
	allExportedProjects = append(allExportedProjects, newExpPrj)

	// Wait a bit to send the events
	go func() {
		// Wait a bit
		time.Sleep(genericWaitingTime)

		status := rpapiv1.RPStateExportProjectDone
		errToSet := ""
		if prjExportFails {
			status = rpapiv1.RPStateExportProjectFAILED
			errToSet = "an error occurred during project export"
			prjExportFails = false
		}

		// Change the status of the export
		allExportedProjects[len(allExportedProjects)-1].Status = status
		allExportedProjects[len(allExportedProjects)-1].Error = errToSet
		allExportedProjects[len(allExportedProjects)-1].EndDate = time.Now()

		// Send the corresponding event
		mockedBck.EmitEvent(rpapiv1.EVTProjectExport, allExportedProjects[len(allExportedProjects)-1])
	}()

	resJSON(w, newExpPrj)
}

// prjsGetExportRoute is the handler used when a call is done on the projects/{id}/export route
// with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) prjsGetExportRoute(w http.ResponseWriter, req *http.Request) {
	// Get the UUID of the concerned project
	prjID := reqParamGet(req, "prjid")
	prjUUID := uuid.FromStringOrNil(prjID)

	for _, prj := range allExportedProjects {
		if utils.UUIDEqual(prj.ProjectID, prjUUID) {
			resJSON(w, prj)
			break
		}
	}
}

// prjsUploadRoute is the handler used when a call is done on the projects/import route
// with the POST method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) prjsUploadRoute(w http.ResponseWriter, req *http.Request) {
	// Try to get the file data from the request (to see if the request is well created)
	_, _, err := req.FormFile("upload")
	if err != nil {
		resJSONErrorE(w, err)
		return
	}

	// Create the structure to return to the client
	resultStruct := rpapiv1.RPImportProjectState{
		StartDate: time.Now(),
		Status:    rpapiv1.RPStateImportProjectUploadManageInfo,
		Error:     "",
	}
	resultStruct.FuturePrjID, _ = uuid.NewV4()

	// Add the project in the array
	allImportedProjects = append(allImportedProjects, resultStruct)

	// Wait a bit to send the events
	go func() {
		// Wait a bit
		time.Sleep(genericWaitingTime)

		status := rpapiv1.RPStateImportProjectUploadManageInfoDone
		errToSet := ""
		if prjImportFails {
			status = rpapiv1.RPStateImportProjectUploadManageInfoFailed
			errToSet = "an error occurred during project upload"
			prjImportFails = false
		}

		// Change the status of the export
		allImportedProjects[len(allImportedProjects)-1].Status = status
		allImportedProjects[len(allImportedProjects)-1].Error = errToSet
		allImportedProjects[len(allImportedProjects)-1].EndDate = time.Now()

		// Send the corresponding event
		mockedBck.EmitEvent(rpapiv1.EVTProjectImport, allImportedProjects[len(allImportedProjects)-1])
	}()

	resJSON(w, resultStruct)
}

// prjsImportRoute is the handler used when a call is done on the projects/{id}/import route
// with the POST method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) prjsImportRoute(w http.ResponseWriter, req *http.Request) {
	// Get the UUID of the concerned project
	prjID := reqParamGet(req, "prjid")
	prjUUID := uuid.FromStringOrNil(prjID)

	// First, try to find if the project have already seen the archive uploaded and save the index
	ourIndex := -1
	var prjStatus rpapiv1.RPStateImportProject
	for index, prj := range allImportedProjects {
		if utils.UUIDEqual(prj.FuturePrjID, prjUUID) {
			ourIndex = index
			prjStatus = prj.Status
			break
		}
	}
	if ourIndex == -1 {
		resJSONErrorE(w, errors.New("The project does not exists"))
		return
	}

	// Then check that it has the right status
	if prjStatus != rpapiv1.RPStateImportProjectUploadManageInfoDone {
		resJSONErrorE(w, errors.New("The project to import does have the correct status to be completely imported"))
		return
	}

	// Change the status to "importing"
	allImportedProjects[ourIndex].Status = rpapiv1.RPStateImportProjectImporting

	// Wait a bit to send the events
	go func(indexToModify int) {
		// Wait a bit
		time.Sleep(genericWaitingTime)

		status := rpapiv1.RPStateImportProjectDone
		errToSet := ""
		if prjImportFails {
			status = rpapiv1.RPStateImportProjectFailed
			errToSet = "an error occurred during project import"
			prjImportFails = false
		}

		// Change the status of the export
		allImportedProjects[indexToModify].Status = status
		allImportedProjects[indexToModify].Error = errToSet
		allImportedProjects[indexToModify].EndDate = time.Now()

		// Add a new project with the same UUID
		newPrjFull := rpapiv1.RPProjectFull{
			RPProject: rpapiv1.RPProject{
				ID:    prjUUID,
				Slug:  "imported-project-" + prjUUID.String(),
				Name:  "imported-project-" + prjUUID.String(),
				State: rpapiv1.RPPrjStableState,
				Type:  "standard",
			},
		}
		allProjects = append(allProjects, newPrjFull)

		// Send the corresponding event
		mockedBck.EmitEvent(rpapiv1.EVTProjectImport, allImportedProjects[indexToModify])

	}(ourIndex)

	resJSON(w, allImportedProjects[ourIndex])
}

// prjsGetImportRoute is the handler used when a call is done on the projects/{id}/import route
// with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) prjsGetImportRoute(w http.ResponseWriter, req *http.Request) {
	// Get the UUID of the concerned project
	prjID := reqParamGet(req, "prjid")
	prjUUID := uuid.FromStringOrNil(prjID)

	for _, prj := range allImportedProjects {
		if utils.UUIDEqual(prj.FuturePrjID, prjUUID) {
			resJSON(w, prj)
			break
		}
	}
}

// prjsGetRepoInfo is the handler used when a call is done on the projects/{id}/repo_info route
// with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) prjsGetRepoInfo(w http.ResponseWriter, req *http.Request) {
	// Get the page number wanted
	pageNb, _ := strconv.Atoi(req.FormValue("page"))
	// Get the per_page number wanted
	perPageNb, _ := strconv.Atoi(req.FormValue("per_page"))
	// Get the architecture wanted
	archWanted := req.FormValue("arch")

	// Get the filter request
	filterValue := req.FormValue("filter")
	nameFilter := ""
	nameSubStrFilter := ""
	if strings.Contains(filterValue, "\"name") {
		// Convert json string to map
		jsonMap := map[string]string{}
		err := json.Unmarshal([]byte(filterValue), &jsonMap)
		if err != nil {
			resJSONErrorE(w, err)
			return
		}
		if tmpFilter, ok := jsonMap["name"]; ok {
			nameFilter = tmpFilter
		}
		if tmpFilter, ok := jsonMap["name__substr"]; ok {
			nameSubStrFilter = tmpFilter
		}
	}

	// Initialize the structure that will be returned
	repoInfoRpm := rpapiv1.RPRepoInfoRpm{
		Page:    pageNb,
		PerPage: perPageNb,
	}

	// Calculate the minimum and maximum index wanted by the user
	minIndex := (pageNb - 1) * perPageNb
	maxIndex := (pageNb * perPageNb) - 1

	// Loop over packages and save the ones we want to send
	count := 0
	for _, rpm := range allRPMsName {
		// Create the corresponding object dynamically
		dynRPMInfo := rpapiv1.RPRPMMetadata{
			Name:        rpm,
			Description: "",
			Version:     "1.0",
			Release:     "27.apps.rpbatz",
			Nvra:        rpm + "-1.0-27.apps.rpbatz-" + archWanted,
		}

		if (count >= minIndex) && (count <= maxIndex) &&
			(((nameFilter != "") && (rpm == nameFilter)) ||
				((nameSubStrFilter != "") && strings.Contains(rpm, nameSubStrFilter)) ||
				((nameFilter == "") && (nameSubStrFilter == ""))) {
			repoInfoRpm.Res = append(repoInfoRpm.Res, dynRPMInfo)
		}
		count++
	}

	// Update the total entries
	repoInfoRpm.TotalEntries = len(repoInfoRpm.Res)

	resJSON(w, repoInfoRpm)
}

// prjsGetInheritance is the handler used when a call is done on the projects/{id}/repos_inheritances route
// with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) prjsGetInheritance(w http.ResponseWriter, req *http.Request) {
	// Get the baseurl request
	baseURLValue := req.FormValue("baseurl")

	// Create the string containing the result
	responseBodyStr := fmt.Sprintf(repoInheritanceFile, baseURLValue)

	// Now write the answer
	w.Header().Set("Content-Type", "application/json")
	w.WriteHeader(http.StatusOK)
	w.Write([]byte(responseBodyStr))
}

// prjsStartBuilds is the handler used when a call is done on the projects/{id}/builds route
// with the POST method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) prjsStartBuilds(w http.ResponseWriter, req *http.Request) {
	// Get the UUID of the concerned project
	prjID := reqParamGet(req, "prjid")
	prjUUID := uuid.FromStringOrNil(prjID)

	// Get the start build object
	startBuildObj := &rpapiv1.RPBuildStart{}
	if err := reqBind(req, startBuildObj); err != nil {
		resJSONErrorE(w, err)
		return
	}

	// Get the build type
	buildType := startBuildObj.BuildType

	// Create the list of app/image UUID to build
	UUIDToBuild := []uuid.UUID{}

	if buildType == rpapiv1.RPBuildApp {
		// Check that each appUUID given by the user is correct (truly a project app)
		for _, appID := range startBuildObj.Applications {
			for _, app := range allApps {
				if utils.UUIDEqual(app.ID, appID) && !utils.UUIDEqual(app.Parent.ProjectID, prjUUID) {
					resJSONErrorE(w, errors.New("the specified application is not part of the right project"))
					return
				} else if utils.UUIDEqual(app.ID, appID) && utils.UUIDEqual(app.Parent.ProjectID, prjUUID) {
					UUIDToBuild = append(UUIDToBuild, app.ID)
					break
				}
			}
		}

		// If nothing given by the user, get all the project apps
		if len(UUIDToBuild) <= 0 {
			for _, app := range allApps {
				if utils.UUIDEqual(app.Parent.ProjectID, prjUUID) {
					UUIDToBuild = append(UUIDToBuild, app.ID)
				}
			}
		}
	} else if buildType == rpapiv1.RPBuildImage {
		// Check that each imgUUID given by the user is correct (truly a project image)
		for _, imgID := range startBuildObj.Images {
			for _, img := range allImages {
				if utils.UUIDEqual(img.ID, imgID) && !utils.UUIDEqual(img.ParentProjectID, prjUUID) {
					resJSONErrorE(w, errors.New("the specified image is not part of the right project"))
					return
				} else if utils.UUIDEqual(img.ID, imgID) && utils.UUIDEqual(img.ParentProjectID, prjUUID) {
					UUIDToBuild = append(UUIDToBuild, img.ID)
					break
				}
			}
		}

		// If nothing given by the user, get all the project images
		if len(UUIDToBuild) <= 0 {
			for _, img := range allImages {
				if utils.UUIDEqual(img.ParentProjectID, prjUUID) {
					UUIDToBuild = append(UUIDToBuild, img.ID)
				}
			}
		}
	}

	// Declare the structure that we will return
	localBuilds := rpapiv1.RPBuilds{
		Page:         1,
		PerPage:      50,
		TotalEntries: 1,
	}

	if buildType == rpapiv1.RPBuildApp {
		// Now start a build for every app
		for _, appID := range UUIDToBuild {
			for indexDistro := range startBuildObj.Distributions {
				// Start the build for one app
				currentBuild, err := mockedBck.startOneAppBuild(appID, &startBuildObj.Distributions[indexDistro], startBuildObj.Comment, string(buildType))
				if err != nil {
					resJSONErrorE(w, err)
					return
				}

				// Add the build in the list of answer builds
				localBuilds.Builds = append(localBuilds.Builds, convertBuildFullToBuild(currentBuild))
				localBuilds.TotalEntries += 1
			}
		}
	} else if buildType == rpapiv1.RPBuildImage {
		// Now, start a build for every image
		for _, imgID := range UUIDToBuild {
			// Start the build for one app
			currentBuild, err := mockedBck.startOneImgBuild(imgID, startBuildObj.Comment)
			if err != nil {
				resJSONErrorE(w, err)
				return
			}

			// Add the build in the list of answer builds
			localBuilds.Builds = append(localBuilds.Builds, convertBuildFullToBuild(currentBuild))
			localBuilds.TotalEntries += 1
		}
	}

	resJSON(w, localBuilds)
}

// prjsGetBuildsStats is the handler used when a call is done on the projects/{id}/builds/stats route
// with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) prjsGetBuildsStats(w http.ResponseWriter, req *http.Request) {
	// Get the 'nb-last' query parameter
	// nbLast, _ := strconv.Atoi(req.FormValue("nb-last"))

	// Get the UUID of the concerned project
	prjID := reqParamGet(req, "prjid")
	prjUUID := uuid.FromStringOrNil(prjID)

	result := rpapiv1.RPBuildStatsProject{
		ProjectID: prjUUID,
		AppStats:  []rpapiv1.RPBuildStatsApp{},
	}

	// First, list all the applications for this project
	for _, app := range allApps {
		// We've found an app that is part of this project
		if utils.UUIDEqual(prjUUID, app.Parent.ProjectID) {

			// Create the new app that we will add
			newApp := rpapiv1.RPBuildStatsApp{
				AppID:       app.ID,
				AppSlug:     app.Slug,
				PackageName: app.PackageName,
				LastBuilds:  []rpapiv1.RPBuild{},
			}

			// Now retrieve all the builds for this app
			for _, bld := range allBuilds {
				if utils.UUIDEqual(bld.Parent.ApplicationID, app.ID) && bld.BuildType == rpapiv1.RPBuildApp {
					// Add build in the list
					newApp.LastBuilds = append(newApp.LastBuilds, bld.RPBuild)
					// Increment total number of builds
					newApp.NbTotal += 1
					// Increment the right counter
					if bld.Status == rpapiv1.RPBuildTaskStatusDone {
						newApp.NbDone += 1
					} else if bld.Status == rpapiv1.RPBuildTaskStatusFailed {
						newApp.NbFailed += 1
					} else if bld.Status == rpapiv1.RPBuildTaskStatusCanceled {
						newApp.NbCanceled += 1
					}
				}
			}

			// Now update the different stats
			result.NbTotal += newApp.NbTotal
			result.NbDone += newApp.NbDone
			result.NbFailed += newApp.NbFailed
			result.NbCanceled += newApp.NbCanceled

			// Add the app object in the result
			result.AppStats = append(result.AppStats, newApp)
		}
	}
	resJSON(w, result)
}

// prjsGetTestsStats is the handler used when a call is done on the projects/{id}/tests/stats route
// with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) prjsGetTestsStats(w http.ResponseWriter, req *http.Request) {
	// Get the 'nb-last' query parameter
	// nbLast, _ := strconv.Atoi(req.FormValue("nb-last"))

	// Get the UUID of the concerned project
	prjID := reqParamGet(req, "prjid")
	prjUUID := uuid.FromStringOrNil(prjID)

	result := rpapiv1.RPTestStatsProject{
		ProjectID:  prjUUID,
		AppStats:   []rpapiv1.RPTestStatsApp{},
		ImageStats: []rpapiv1.RPTestStatsImg{},
	}

	// First, list all the applications for this project
	for _, app := range allApps {
		// We've found an app that is part of this project
		if utils.UUIDEqual(prjUUID, app.Parent.ProjectID) {

			// Create the new app that we will add
			newApp := rpapiv1.RPTestStatsApp{
				AppID:          app.ID,
				AppSlug:        app.Slug,
				PackageName:    app.PackageName,
				LastCITests:    []rpapiv1.RPTest{},
				LastAuditTests: []rpapiv1.RPAudit{},
			}

			// Now retrieve all the CI tests for this app
			for _, tst := range allTests.Tests {
				if utils.UUIDEqual(tst.Parent.AppID, app.ID) {
					// Add test in the list
					newApp.LastCITests = append(newApp.LastCITests, tst)
					// Increment total number of tests
					newApp.NbTotal += 1
					// Increment the right counter
					if tst.Status == rpapiv1.RPTestStatusSuccess {
						newApp.NbSucceeded += 1
					} else if tst.Status == rpapiv1.RPTestStatusFailed {
						newApp.NbFailed += 1
					} else if tst.Status == rpapiv1.RPTestStatusCancelled {
						newApp.NbCanceled += 1
					}
				}
			}

			// Now retrieve all the audit tests for this app
			for _, audit := range allAudits {
				if utils.UUIDEqual(audit.Parent.AppID, app.ID) {
					// Add test in the list
					newApp.LastAuditTests = append(newApp.LastAuditTests, audit.RPAudit)
					// Increment total number of tests
					newApp.NbTotal += 1
					// Increment the right counter
					if audit.Status == rpapiv1.RPAuditStatusSuccess {
						newApp.NbSucceeded += 1
					} else if audit.Status == rpapiv1.RPAuditStatusFailed {
						newApp.NbFailed += 1
					} else if audit.Status == rpapiv1.RPAuditStatusCancelled {
						newApp.NbCanceled += 1
					}
				}
			}

			// Now update the different stats
			result.NbTotal += newApp.NbTotal
			result.NbSucceeded += newApp.NbSucceeded
			result.NbFailed += newApp.NbFailed
			result.NbCanceled += newApp.NbCanceled

			// Add the app object in the result
			result.AppStats = append(result.AppStats, newApp)
		}
	}

	// Then, list all the images for this project
	for _, image := range allImages {
		// We've found an image that is part of this project
		if utils.UUIDEqual(prjUUID, image.ParentProjectID) {

			// Create the new app that we will add
			newImg := rpapiv1.RPTestStatsImg{
				ImageID:     image.ID,
				ImageSlug:   image.Slug,
				LastCITests: []rpapiv1.RPTest{},
			}

			// Now retrieve all the CI tests for this app
			for _, tst := range allTests.Tests {
				if utils.UUIDEqual(tst.Parent.ImageID, image.ID) {
					// Add test in the list
					newImg.LastCITests = append(newImg.LastCITests, tst)
					// Increment total number of tests
					newImg.NbTotal += 1
					// Increment the right counter
					if tst.Status == rpapiv1.RPTestStatusSuccess {
						newImg.NbSucceeded += 1
					} else if tst.Status == rpapiv1.RPTestStatusFailed {
						newImg.NbFailed += 1
					} else if tst.Status == rpapiv1.RPTestStatusCancelled {
						newImg.NbCanceled += 1
					}
				}
			}

			// Now update the different stats
			result.NbTotal += newImg.NbTotal
			result.NbSucceeded += newImg.NbSucceeded
			result.NbFailed += newImg.NbFailed
			result.NbCanceled += newImg.NbCanceled

			// Add the app object in the result
			result.ImageStats = append(result.ImageStats, newImg)
		}
	}

	resJSON(w, result)
}

// prjsGetRolesHandler is the handler used when a call is done on the projects/{id}/roles route
// with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) prjsGetRolesHandler(w http.ResponseWriter, req *http.Request) {
	// Create the result object
	result := rpapiv1.RPRoleProjects{}

	// Add the several roles
	// Manager
	result = append(result, rpapiv1.RPRoleProject{
		RoleLabel: "manager",
		RoleName:  "manager of project",
	})
	// Tester
	result = append(result, rpapiv1.RPRoleProject{
		RoleLabel: "tester",
		RoleName:  "tester of project",
	})
	// Developer
	result = append(result, rpapiv1.RPRoleProject{
		RoleLabel: "developer",
		RoleName:  "developer of project",
	})

	resJSON(w, result)
}

// prjsAddMemberHandler is the handler used when a call is done on the projects/{id}/members route
// with the POST method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) prjsAddMemberHandler(w http.ResponseWriter, req *http.Request) {
	// Get the project UUID concerned by the new member
	prjID := reqParamGet(req, "prjid")
	prjUUID := uuid.FromStringOrNil(prjID)

	// Get the member to add
	reqMember := &rpapiv1.RPMember{}
	if err := reqBind(req, reqMember); err != nil {
		resJSONErrorE(w, err)
		return
	}

	// Check that the project exists
	prjExists := false
	for _, prj := range allProjects {
		if utils.UUIDEqual(prj.ID, prjUUID) {
			prjExists = true
			break
		}
	}
	if !prjExists {
		resJSONErrorE(w, errors.New("The project does not exists"))
		return
	}

	// Add the member in the tab of all members
	allMembers = append(allMembers, *reqMember)

	// Wait a bit and send the corresponding event
	go func() {
		time.Sleep(genericWaitingTime)
		mockedBck.EmitEvent(rpapiv1.EVTMemberAdd, reqMember)
	}()

	resJSON(w, reqMember)
}

// prjsGetMembersHandler is the handler used when a call is done on the projects/{id}/members route
// with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) prjsGetMembersHandler(w http.ResponseWriter, req *http.Request) {
	// Get the project UUID concerned by the new member
	prjID := reqParamGet(req, "prjid")
	prjUUID := uuid.FromStringOrNil(prjID)

	// Create the result of the function
	resTabMember := rpapiv1.RPMembers{}

	// Go through the project members and get only the ones
	// that's interesting us
	for _, member := range allMembers {
		if utils.UUIDEqual(member.ObjectID, prjUUID) {
			resTabMember = append(resTabMember, member)
		}
	}

	resJSON(w, resTabMember)
}

// prjsPutMemberHandler is the handler used when a call is done on the projects/{id}/members route
// with the PUT method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) prjsPutMemberHandler(w http.ResponseWriter, req *http.Request) {

	// Get the member to update
	reqMember := &rpapiv1.RPMember{}
	if err := reqBind(req, reqMember); err != nil {
		resJSONErrorE(w, err)
		return
	}

	// Find the member to update
	indexToChange := -1
	for index, member := range allMembers {
		if utils.UUIDEqual(member.ID, reqMember.ID) && utils.UUIDEqual(member.ObjectID, reqMember.ObjectID) {
			indexToChange = index
			break
		}
	}

	// Check that we have found the member to change
	if indexToChange == -1 {
		resJSONErrorE(w, errors.New("The member does not exists"))
		return
	}

	// Modify the member
	allMembers[indexToChange].Roles = reqMember.Roles

	// Wait a bit and send the corresponding event
	go func() {
		time.Sleep(genericWaitingTime)
		mockedBck.EmitEvent(rpapiv1.EVTMemberStateChange, allMembers[indexToChange])
	}()

	resJSON(w, allMembers[indexToChange])
}

// prjsDeleteMemberHandler is the handler used when a call is done on the projects/{id}/members route
// with the DELETE method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) prjsDeleteMemberHandler(w http.ResponseWriter, req *http.Request) {
	// Get the project UUID concerned by the member to delete
	prjID := reqParamGet(req, "prjid")
	prjUUID := uuid.FromStringOrNil(prjID)

	// Get the ID of the member to delete
	memberID := req.FormValue("id")
	memberUUID := uuid.FromStringOrNil(memberID)

	// Go through the project members and delete the right one
	tmpMembers := rpapiv1.RPMembers{}
	resMember := rpapiv1.RPMember{}
	for _, member := range allMembers {
		if utils.UUIDEqual(member.ObjectID, prjUUID) && utils.UUIDEqual(member.ID, memberUUID) {
			// Copy the member we will need to send at the end
			copier.Copy(&resMember, &member)
		} else {
			tmpMembers = append(tmpMembers, member)
		}
	}

	// Set the new tab of member
	allMembers = tmpMembers

	// Wait a bit and send the corresponding event
	go func() {
		time.Sleep(genericWaitingTime)
		mockedBck.EmitEvent(rpapiv1.EVTMemberDelete, resMember)
	}()

	resJSON(w, resMember)
}

// prjsNewMembersHandler is the handler used when a call is done on the projects/{id}/members/newmember route
// with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) prjsNewMembersHandler(w http.ResponseWriter, req *http.Request) {
	prjID := reqParamGet(req, "prjid")
	prjUUID := uuid.FromStringOrNil(prjID)

	// Initialize the result
	membersRes := rpapiv1.RPMembers{}

	// Add all the users
	for _, user := range allUsers.Users {
		membersRes = append(membersRes, rpapiv1.RPMember{
			ID:       user.ID,
			Pseudo:   user.Pseudo,
			ObjectID: prjUUID,
			Type:     "User",
		})
	}

	// Add all the teams
	for _, team := range allTeams.Teams {
		membersRes = append(membersRes, rpapiv1.RPMember{
			ID:       team.ID,
			Pseudo:   team.Slug,
			ObjectID: prjUUID,
			Type:     "Team",
		})
	}

	resJSON(w, membersRes)
}

// SetCurrentUserRole is the function that can be used from outside
// to set the roles of the current user in order to test things
//
// :param newRoles: New roles to give to the current user
// :type newRoles: rpapiv1.RPRoles
// :param toDefault: If set to true, the roles are reset to the default values
// :type toDefault: bool
func (mockedBck *RPMockedBackend) SetCurrentUserRole(newRoles rpapiv1.RPRoles, toDefault bool) {
	// Empty the current slice
	currentUserRoles = rpapiv1.RPRoles{}

	if toDefault {
		currentUserRoles.Roles = append(currentUserRoles.Roles, rpapiv1.RPRole{
			Name:   defaultRoleNameCurrUser,
			Label:  defaultRoleLabelCurrUser,
			Object: defaultRoleObjectCurrUser,
		})
	} else {
		// Now fill it with the new roles
		currentUserRoles.Roles = append(currentUserRoles.Roles, newRoles.Roles...)
	}
}

// rolesMeHandler is the handler used when a call is done on the /roles/me route
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) rolesMeHandler(w http.ResponseWriter, req *http.Request) {
	resJSON(w, currentUserRoles)
}

// usersListHandler is the handler used when a call is done on the users route
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) usersListHandler(w http.ResponseWriter, req *http.Request) {
	// Get the page number wanted
	pageNb, _ := strconv.Atoi(req.FormValue("page"))
	// Get the per_page number wanted
	perPageNb, _ := strconv.Atoi(req.FormValue("per_page"))

	// Initialize the structure that will be returned
	retUsers := rpapiv1.RPUsersFull{
		Page:    pageNb,
		PerPage: perPageNb,
	}

	// Calculate the minimum and maximum index wanted by the user
	minIndex := (pageNb - 1) * perPageNb
	maxIndex := (pageNb * perPageNb) - 1

	// Loop over users and save the ones we want to send
	count := 0
	for _, user := range allUsers.Users {
		if (count >= minIndex) && (count <= maxIndex) {
			retUsers.Users = append(retUsers.Users, user)
		}
		count++
	}

	// Update the total entries
	retUsers.TotalEntries = len(retUsers.Users)

	resJSON(w, retUsers)
}

// usersAddHandler is the handler used when a call is done on the users route
// with the POST method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) usersAddHandler(w http.ResponseWriter, req *http.Request) {
	// Get the request body
	reqBody := &rpapiv1.RPUserFull{}
	if err := reqBind(req, reqBody); err != nil {
		resJSONErrorE(w, err)
		return
	}

	// Update the creation and update time
	timeNow := time.Now()
	reqBody.CreatedAt = timeNow
	reqBody.UpdatedAt = timeNow

	// Set the UUID
	reqBody.ID, _ = uuid.NewV4()

	// Add the user in our list
	allUsers.Users = append(allUsers.Users, *reqBody)

	// Wait a bit and send the corresponding event
	go func() {
		time.Sleep(genericWaitingTime)
		mockedBck.EmitEvent(rpapiv1.EVTUserAdd, reqBody)
	}()

	resJSON(w, reqBody)
}

// usersGetHandler is the handler used when a call is done on the users/{id} route
// with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) usersGetHandler(w http.ResponseWriter, req *http.Request) {
	usrID := reqParamGet(req, "id")
	usrUUID := uuid.FromStringOrNil(usrID)

	for _, usr := range allUsers.Users {
		if utils.UUIDEqual(usr.ID, usrUUID) {
			resJSON(w, usr)
			break
		}
	}
}

// usersGetMeHandler is the handler used when a call is done on the users/me route
// with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) usersGetMeHandler(w http.ResponseWriter, req *http.Request) {
	for _, usr := range allUsers.Users {
		if usr.Pseudo == "me" {
			resJSON(w, usr)
			break
		}
	}
}

// usersPutHandler is the handler used when a call is done on the users/{id} route
// with the PUT method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) usersPutHandler(w http.ResponseWriter, req *http.Request) {
	usrID := reqParamGet(req, "id")
	usrUUID := uuid.FromStringOrNil(usrID)

	// Get the request body
	reqBody := &rpapiv1.RPUserFull{}
	if err := reqBind(req, reqBody); err != nil {
		resJSONErrorE(w, err)
		return
	}

	var indexToModify int
	for index, usr := range allUsers.Users {
		if utils.UUIDEqual(usr.ID, usrUUID) {
			// We found the right project, save the index to modify it afterwards
			indexToModify = index
			break
		}
	}

	// Modify the project (actually modify only the email, firstname, lastname and phone)
	allUsers.Users[indexToModify].Email = reqBody.Email
	allUsers.Users[indexToModify].Firstname = reqBody.Firstname
	allUsers.Users[indexToModify].Lastname = reqBody.Lastname
	allUsers.Users[indexToModify].Phone = reqBody.Phone

	// Wait a bit and send the corresponding event
	go func() {
		time.Sleep(genericWaitingTime)
		mockedBck.EmitEvent(rpapiv1.EVTUserStateChange, allUsers.Users[indexToModify])
	}()

	resJSON(w, allUsers.Users[indexToModify])
}

// usersPutMeHandler is the handler used when a call is done on the users/me route
// with the PUT method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) usersPutMeHandler(w http.ResponseWriter, req *http.Request) {
	// Get the request body
	reqBody := &rpapiv1.RPUserFull{}
	if err := reqBind(req, reqBody); err != nil {
		resJSONErrorE(w, err)
		return
	}

	var indexToModify int
	for index, usr := range allUsers.Users {
		if usr.Pseudo == "me" {
			// We found the right project, save the index to modify it afterwards
			indexToModify = index
			break
		}
	}

	// Modify the project (actually modify only the email, firstname, lastname and phone)
	allUsers.Users[indexToModify].Email = reqBody.Email
	allUsers.Users[indexToModify].Firstname = reqBody.Firstname
	allUsers.Users[indexToModify].Lastname = reqBody.Lastname
	allUsers.Users[indexToModify].Phone = reqBody.Phone

	// Wait a bit and send the corresponding event
	go func() {
		time.Sleep(genericWaitingTime)
		mockedBck.EmitEvent(rpapiv1.EVTUserStateChange, allUsers.Users[indexToModify])
	}()

	resJSON(w, allUsers.Users[indexToModify])
}

// usersEnableDisableHandler is the handler used when a call is done on the users/enable-disable/{id} route
// with the PUT method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) usersEnableDisableHandler(w http.ResponseWriter, req *http.Request) {
	usrID := reqParamGet(req, "id")
	usrUUID := uuid.FromStringOrNil(usrID)

	var indexToModify int
	for index, usr := range allUsers.Users {
		if utils.UUIDEqual(usr.ID, usrUUID) {
			// We found the right user, save the index to modify it afterwards
			indexToModify = index
			break
		}
	}

	// Modify the project (actually modify only the email, firstname, lastname and phone)
	allUsers.Users[indexToModify].Disabled = !allUsers.Users[indexToModify].Disabled

	// Wait a bit and send the corresponding event
	go func() {
		time.Sleep(genericWaitingTime)
		mockedBck.EmitEvent(rpapiv1.EVTUserDisable, allUsers.Users[indexToModify])
	}()

	resJSON(w, allUsers.Users[indexToModify])
}

// teamsListHandler is the handler used when a call is done on the teams route
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) teamsListHandler(w http.ResponseWriter, req *http.Request) {

	// Get the page number wanted
	pageNb, _ := strconv.Atoi(req.FormValue("page"))

	// Get the per_page number wanted
	perPageNb, _ := strconv.Atoi(req.FormValue("per_page"))

	// Get the potential "filters" string
	filtersStr := req.FormValue("filters")

	// Get the filters
	slugFilter := ""
	if strings.HasPrefix(filtersStr, "slug=") {
		slugFilter = strings.ReplaceAll(filtersStr, "slug=", "")
	}
	nameFilter := ""
	if strings.HasPrefix(filtersStr, "name=") {
		nameFilter = strings.ReplaceAll(filtersStr, "name=", "")
		nameFilter = strings.ReplaceAll(nameFilter, "+", " ")
	}

	// Create the object to return
	teamToReturn := rpapiv1.RPTeams{
		Page:    pageNb,
		PerPage: perPageNb,
	}

	// Calculate the minimum and maximum index wanted by the user
	minIndex := (pageNb - 1) * perPageNb
	maxIndex := (pageNb * perPageNb) - 1

	// Return only the teams wanted by the client on this request
	count := 0
	for _, team := range allTeams.Teams {
		if (count >= minIndex) && (count <= maxIndex) &&
			(slugFilter == "" || team.Slug == slugFilter) &&
			(nameFilter == "" || team.Name == nameFilter) {
			teamToReturn.Teams = append(teamToReturn.Teams, team)
		}
		count++
	}

	teamToReturn.TotalEntries = len(allTeams.Teams)
	resJSON(w, teamToReturn)
}

// teamsAddHandler is the handler used when a call is done on the users route
// with the POST method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) teamsAddHandler(w http.ResponseWriter, req *http.Request) {
	// Get the request body
	reqBody := &rpapiv1.RPTeam{}
	if err := reqBind(req, reqBody); err != nil {
		resJSONErrorE(w, err)
		return
	}

	// Update the creation and update time
	timeNow := time.Now()
	reqBody.CreatedAt = timeNow
	reqBody.UpdatedAt = timeNow

	// Set the UUID
	reqBody.ID, _ = uuid.NewV4()

	// Set the slug from the name
	reqBody.Slug = strings.ToLower(strings.ReplaceAll(reqBody.Name, " ", "-"))

	// Set the owner and creator of the team
	reqBody.UserOwnerID = uuid.FromStringOrNil("b90c2344-15cd-40e5-870a-e3dadd33863f")
	reqBody.UserOwnerPseudo = "unit-test-owner"
	reqBody.UserCreatorID = uuid.FromStringOrNil("b90c2344-15cd-40e5-870a-e3dadd33863f")
	reqBody.UserCreatorPseudo = "unit-test-owner"

	// Set the number of members
	reqBody.NbUserMembers = 1

	// Set the only member in the team
	reqBody.Members = rpapiv1.RPTeamMembers{
		{
			ID:               uuid.FromStringOrNil("b90c2344-15cd-40e5-870a-e3dadd33863f"),
			Pseudo:           "unit-test-owner",
			Email:            "test.owner@iot.bzh",
			Firstname:        "Test",
			Lastname:         "Owner",
			InvitationStatus: "accepted",
			Type:             rpapiv1.RPUserType,
		},
	}

	// Add the user in our list
	allTeams.Teams = append(allTeams.Teams, *reqBody)

	// Wait a bit and send the corresponding event
	go func() {
		time.Sleep(genericWaitingTime)
		mockedBck.EmitEvent(rpapiv1.EVTTeamAdd, reqBody)
	}()

	resJSON(w, reqBody)
}

// teamsGetHandler is the handler used when a call is done on the teams/{id} route
// with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) teamsGetHandler(w http.ResponseWriter, req *http.Request) {
	teamID := reqParamGet(req, "teamid")
	teamUUID := uuid.FromStringOrNil(teamID)

	for _, team := range allTeams.Teams {
		if utils.UUIDEqual(team.ID, teamUUID) {
			resJSON(w, team)
			break
		}
	}
}

// teamsPutHandler is the handler used when a call is done on the teams/{id} route
// with the PUT method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) teamsPutHandler(w http.ResponseWriter, req *http.Request) {
	teamID := reqParamGet(req, "teamid")
	teamUUID := uuid.FromStringOrNil(teamID)

	// Get the request body
	reqBody := &rpapiv1.RPTeam{}
	if err := reqBind(req, reqBody); err != nil {
		resJSONErrorE(w, err)
		return
	}

	var indexToModify int
	for index, team := range allTeams.Teams {
		if utils.UUIDEqual(team.ID, teamUUID) {
			// We found the right project, save the index to modify it afterwards
			indexToModify = index
			break
		}
	}

	// Modify the team (actually modfy only the name, description and status)
	allTeams.Teams[indexToModify].Name = reqBody.Name
	allTeams.Teams[indexToModify].Description = reqBody.Description
	allTeams.Teams[indexToModify].Status = reqBody.Status

	// Wait a bit and send the corresponding event
	go func() {
		time.Sleep(genericWaitingTime)
		mockedBck.EmitEvent(rpapiv1.EVTTeamStateChange, allTeams.Teams[indexToModify])
	}()

	resJSON(w, allTeams.Teams[indexToModify])
}

// teamsDeleteHandler is the handler used when a call is done on the teams/{id} route
// with the DELETE method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) teamsDeleteHandler(w http.ResponseWriter, req *http.Request) {
	teamID := reqParamGet(req, "teamid")
	teamUUID := uuid.FromStringOrNil(teamID)

	tmpTeams := []rpapiv1.RPTeam{}
	resTeam := rpapiv1.RPTeam{}
	for _, team := range allTeams.Teams {
		if !utils.UUIDEqual(team.ID, teamUUID) {
			tmpTeams = append(tmpTeams, team)
		} else {
			// Copy the project we will need to send at the end
			copier.Copy(&resTeam, &team)
		}
	}

	// Set the new tab of teams
	allTeams.Teams = tmpTeams

	// Wait a bit and send the corresponding event
	go func() {
		time.Sleep(genericWaitingTime)
		mockedBck.EmitEvent(rpapiv1.EVTTeamDelete, resTeam)
	}()

	resJSON(w, resTeam)
}

// teamsNewMemberHandler is the handler used when a call is done on the teams/{id}/members/newmember route
// with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) teamsNewMemberHandler(w http.ResponseWriter, req *http.Request) {
	teamID := reqParamGet(req, "teamid")
	teamUUID := uuid.FromStringOrNil(teamID)

	// Initialize the result
	membersRes := rpapiv1.RPMembers{}

	// Add all the users
	for _, user := range allUsers.Users {
		membersRes = append(membersRes, rpapiv1.RPMember{
			ID:       user.ID,
			Pseudo:   user.Pseudo,
			ObjectID: teamUUID,
			Type:     "User",
		})
	}

	// Add all the teams
	for _, team := range allTeams.Teams {
		membersRes = append(membersRes, rpapiv1.RPMember{
			ID:       team.ID,
			Pseudo:   team.Slug,
			ObjectID: teamUUID,
			Type:     "Team",
		})
	}

	resJSON(w, membersRes)
}

// teamsInvitationHandler is the handler used when a call is done on the teams/{id}/members/invitation route
// with the POST method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) teamsInvitationHandler(w http.ResponseWriter, req *http.Request) {

	// Look if the team exists
	teamID := reqParamGet(req, "teamid")
	teamUUID := uuid.FromStringOrNil(teamID)

	teamIndex := -1
	for index, team := range allTeams.Teams {
		if utils.UUIDEqual(team.ID, teamUUID) {
			teamIndex = index
			break
		}
	}

	if teamIndex == -1 {
		resJSONErrorE(w, errors.New("The team does not exists"))
		return
	}

	// Get the request body
	reqGuest := &rpapiv1.RPMember{}
	if err := reqBind(req, reqGuest); err != nil {
		resJSONErrorE(w, err)
		return
	}

	// Check if the guest exists
	guestExists := false
	if reqGuest.Type == "User" {
		for _, user := range allUsers.Users {
			if utils.UUIDEqual(user.ID, reqGuest.ID) {
				guestExists = true
				break
			}
		}
	} else {
		for _, team := range allTeams.Teams {
			if utils.UUIDEqual(team.ID, reqGuest.ID) {
				guestExists = true
				break
			}
		}
	}

	if !guestExists {
		resJSONErrorE(w, errors.New("The guest does not exists"))
		return
	}

	// Check if the guest is not already in the team
	guestInTeam := false
	for _, member := range allTeams.Teams[teamIndex].Members {
		if utils.UUIDEqual(member.ID, reqGuest.ID) {
			guestInTeam = true
		}
	}

	if !guestInTeam {
		// Add the new member in the team, if it is not already in it
		teamMbr := rpapiv1.RPTeamMember{
			ID:     reqGuest.ID,
			Pseudo: reqGuest.Pseudo,
		}
		allTeams.Teams[teamIndex].Members = append(allTeams.Teams[teamIndex].Members, teamMbr)
	}

	invitAns := rpapiv1.RPInvitResult{
		Team: allTeams.Teams[teamIndex],
	}

	resJSON(w, invitAns)
}

// teamsDelMemberHandler is the handler used when a call is done on the teams/{id}/members route
// with the DELETE method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) teamsDelMemberHandler(w http.ResponseWriter, req *http.Request) {
	// Look if the team exists
	teamID := reqParamGet(req, "teamid")
	teamUUID := uuid.FromStringOrNil(teamID)

	// Get the ID of the member to delete
	memberID := req.FormValue("idmb")
	memberUUID := uuid.FromStringOrNil(memberID)

	teamIndex := -1
	memberIndexToRemove := -1
	for index, team := range allTeams.Teams {
		if utils.UUIDEqual(team.ID, teamUUID) {
			// We found the right team!
			teamIndex = index

			// Now, get the member to delete
			for memberIndex, member := range team.Members {
				if utils.UUIDEqual(memberUUID, member.ID) {
					memberIndexToRemove = memberIndex
					break
				}
			}
			break
		}
	}

	if teamIndex == -1 {
		resJSONErrorE(w, errors.New("The team does not exists"))
		return
	}

	if memberIndexToRemove == -1 {
		resJSONErrorE(w, errors.New("The member does not exists in the team"))
		return
	}

	// Delete the member in the team
	membersNb := len(allTeams.Teams[teamIndex].Members)
	allTeams.Teams[teamIndex].Members[memberIndexToRemove] = allTeams.Teams[teamIndex].Members[membersNb-1]
	allTeams.Teams[teamIndex].Members = allTeams.Teams[teamIndex].Members[:membersNb-1]

	resJSON(w, allTeams.Teams[teamIndex])
}

// appsListHandler is the handler used when a call is done on the applications route
// with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) appsListHandler(w http.ResponseWriter, req *http.Request) {

	// Get the type wanted by the user
	typesWanted := req.FormValue("type")
	if typesWanted == "" {
		typesWanted = "standard"
	}

	// Get the page number wanted
	pageNb, _ := strconv.Atoi(req.FormValue("page"))

	// Get the per_page number wanted
	perPageNb, _ := strconv.Atoi(req.FormValue("per_page"))

	// Get the potential "filters" string
	filtersStr := req.FormValue("filters")

	// Get the bookmark request
	bookmarkWanted := (strings.Contains(filtersStr, "bookmarked==true"))

	// Get the filters
	slugFilter := ""
	if strings.HasPrefix(filtersStr, "slug=") {
		slugFilter = strings.ReplaceAll(filtersStr, "slug=", "")
	}
	pkgFilter := ""
	if strings.HasPrefix(filtersStr, "package_name=") {
		pkgFilter = strings.ReplaceAll(filtersStr, "package_name=", "")
	}
	nameFilter := ""
	if strings.HasPrefix(filtersStr, "name=") {
		nameFilter = strings.ReplaceAll(filtersStr, "name=", "")
		nameFilter = strings.ReplaceAll(nameFilter, "+", " ")
	}

	// Initialize the structure that will be returned
	retApps := rpapiv1.RPApplications{
		Page:    pageNb,
		PerPage: perPageNb,
	}

	// Calculate the minimum and maximum index wanted by the user
	minIndex := (pageNb - 1) * perPageNb
	maxIndex := (pageNb * perPageNb) - 1

	// Convert the type(s) to a slice of strings
	typesWantedSli := strings.Split(typesWanted, ",")

	// Loop over apps and save the ones we want to send
	count := 0
	for _, app := range allApps {
		if utils.ContainsString(typesWantedSli, string(app.Parent.Type)) {
			if (count >= minIndex) && (count <= maxIndex) &&
				(!bookmarkWanted || app.Bookmarked) &&
				(slugFilter == "" || app.Slug == slugFilter) &&
				(pkgFilter == "" || app.PackageName == pkgFilter) &&
				(nameFilter == "" || app.Name == nameFilter) {
				retApps.Apps = append(retApps.Apps, convertAppFullToApp(app))
			}
			count++
		}
	}

	// Update the total entries
	retApps.TotalEntries = count

	resJSON(w, retApps)
}

// appsGetFromPrjHandler is the handler used when a call is done on the projects/{prjid}/applications route
// with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) appsGetFromPrjHandler(w http.ResponseWriter, req *http.Request) {
	// Get the project UUID concerned by the list of applications
	prjID := reqParamGet(req, "prjid")
	prjUUID := uuid.FromStringOrNil(prjID)

	// Get the page number wanted
	pageNb, _ := strconv.Atoi(req.FormValue("page"))

	// Get the per_page number wanted
	perPageNb, _ := strconv.Atoi(req.FormValue("per_page"))

	// Get the potential "filters" string
	filtersStr := req.FormValue("filters")

	// Get the filters
	slugFilter := ""
	if strings.HasPrefix(filtersStr, "slug=") {
		slugFilter = strings.ReplaceAll(filtersStr, "slug=", "")
	}
	pkgFilter := ""
	if strings.HasPrefix(filtersStr, "package_name=") {
		pkgFilter = strings.ReplaceAll(filtersStr, "package_name=", "")
	}
	nameFilter := ""
	if strings.HasPrefix(filtersStr, "name=") {
		nameFilter = strings.ReplaceAll(filtersStr, "name=", "")
		nameFilter = strings.ReplaceAll(nameFilter, "+", " ")
	}

	// Initialize the structure that will be returned
	retApps := rpapiv1.RPApplications{
		Page:    pageNb,
		PerPage: perPageNb,
	}

	// Calculate the minimum and maximum index wanted by the user
	minIndex := (pageNb - 1) * perPageNb
	maxIndex := (pageNb * perPageNb) - 1

	// Loop over apps and save the ones we want to send
	count := 0
	for _, app := range allApps {
		if utils.UUIDEqual(app.Parent.ProjectID, prjUUID) {
			if (count >= minIndex) && (count <= maxIndex) &&
				(slugFilter == "" || app.Slug == slugFilter) &&
				(pkgFilter == "" || app.PackageName == pkgFilter) &&
				(nameFilter == "" || app.Name == nameFilter) {
				retApps.Apps = append(retApps.Apps, convertAppFullToApp(app))
			}
			count++
		}
	}

	// Update the total entries
	retApps.TotalEntries = len(retApps.Apps)

	resJSON(w, retApps)
}

// appsCreateHandler is the handler used when a call is done on the projects/{prjid}/applications route
// with the POST method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) appsCreateHandler(w http.ResponseWriter, req *http.Request) {
	// Get the project UUID concerned by the new application
	prjID := reqParamGet(req, "prjid")
	prjUUID := uuid.FromStringOrNil(prjID)

	// Check that the project exists
	prjExists := false
	prjSlug := ""
	var prjType rpapiv1.RPProjectType
	for _, prj := range allProjects {
		if utils.UUIDEqual(prj.ID, prjUUID) {
			prjExists = true
			prjSlug = prj.Slug
			prjType = prj.Type
			break
		}
	}
	if !prjExists {
		resJSONErrorE(w, errors.New("The project does not exists"))
		return
	}

	// Get the request body
	reqBody := &rpapiv1.RPApplicationFull{}
	if err := reqBind(req, reqBody); err != nil {
		resJSONErrorE(w, err)
		return
	}

	// Update the creation and update time
	timeNow := time.Now()
	reqBody.CreatedAt = timeNow
	reqBody.UpdatedAt = timeNow
	reqBody.State = rpapiv1.RPAppStableState

	// Set the UUID
	reqBody.ID, _ = uuid.NewV4()

	// Set the slug from the name
	reqBody.Slug = strings.ToLower(strings.ReplaceAll(reqBody.Name, " ", "-"))

	// Set the project slug
	reqBody.Parent.ProjectSlug = prjSlug

	// Set the parent project type
	reqBody.Parent.Type = prjType

	// Update the revision if set to ""
	if reqBody.Metadata.Data.Revision == "" {
		reqBody.Metadata.Data.Revision = "master"
	}

	// Update the build type if not defined
	if reqBody.Metadata.Data.BuildType == "" {
		reqBody.Metadata.Data.BuildType = rpapiv1.RPAppEmulatedBuild
	}

	// set global srpm auto rebuild to true(all options true)
	reqBody.AutoRSrpmBuild = 15
	// Add the application in our list
	allApps = append(allApps, *reqBody)

	// Create the corresponding main branch
	branchToCreate := "main"
	if reqBody.PackageActiveBranch != "" {
		branchToCreate = reqBody.PackageActiveBranch
	}

	branchesList := &rpapiv1.RPBranchList{
		PackageActiveBranch: branchToCreate,
		ListBranches: []rpapiv1.RPBranchChar{
			{
				RPBranchName: rpapiv1.RPBranchName{Name: branchToCreate},
				Default:      true,
				Protected:    true,
			},
		},
	}

	// Save the created branches list in the map
	allAppsBranches[reqBody.ID] = branchesList

	// Create the corresponding Srpm build (and check dependencies)
	mockedBck.createSRPMBuild(reqBody)

	// Wait a bit and send the corresponding events
	go func() {
		time.Sleep(genericWaitingTime)
		mockedBck.EmitEvent(rpapiv1.EvtAppAdd, reqBody)
		time.Sleep(genericWaitingTime)
		mockedBck.EmitEvent(rpapiv1.EvtAppStateChange, reqBody)
	}()

	resJSON(w, reqBody)
}

// appsGetHandler is the handler used when a call is done on the application/{appid} route
// with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) appsGetHandler(w http.ResponseWriter, req *http.Request) {
	appID := reqParamGet(req, "appid")
	appUUID := uuid.FromStringOrNil(appID)

	for _, app := range allApps {
		if utils.UUIDEqual(app.ID, appUUID) {
			resJSON(w, app)
			break
		}
	}
}

// appsGetSamplesHandler is the handler used when a call is done on the applications/samples route
// with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) appsGetSamplesHandler(w http.ResponseWriter, req *http.Request) {
	samplesApps := []rpapiv1.RPSampleApplication{
		{
			Name:        "Helloworld binding",
			Description: "A binding example for AGL, implemented in a redpesk® context",
			PackageName: "helloworld-binding",
			Metadata: rpapiv1.RPApplicationMetadata{
				Data: rpapiv1.BranchData{
					SourcesURI:       "https://github.com/redpesk-samples/helloworld-binding.git",
					Revision:         "",
					SpecfileLocation: "https://raw.githubusercontent.com/redpesk-samples/helloworld-binding/master/conf.d/packaging/helloworld-binding.spec",
				},
			},
		},
	}

	// Return the samples
	resJSON(w, samplesApps)
}

// appsPutHandler is the handler used when a call is done on the application/{appid} route
// with the PUT method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) appsPutHandler(w http.ResponseWriter, req *http.Request) {
	appID := reqParamGet(req, "appid")
	appUUID := uuid.FromStringOrNil(appID)

	// Get the request body
	reqBody := &rpapiv1.RPApplicationFull{}
	if err := reqBind(req, reqBody); err != nil {
		resJSONErrorE(w, err)
		return
	}

	var indexToModify int
	for index, app := range allApps {
		if utils.UUIDEqual(app.ID, appUUID) {
			// We found the right application, save the index to modify it afterwards
			indexToModify = index
			break
		}
	}

	// Modify the project (actually modify only what can be modified)
	allApps[indexToModify].AutoTests = reqBody.AutoTests
	allApps[indexToModify].Description = reqBody.Description
	allApps[indexToModify].Metadata.Data.SpecfileLocation = reqBody.Metadata.Data.SpecfileLocation
	allApps[indexToModify].Distributions = reqBody.Distributions
	allApps[indexToModify].Name = reqBody.Name
	allApps[indexToModify].Metadata.Data.SourcesURI = reqBody.Metadata.Data.SourcesURI
	allApps[indexToModify].Metadata.Data.Revision = reqBody.Metadata.Data.Revision
	allApps[indexToModify].Metadata.Data.BuildType = reqBody.Metadata.Data.BuildType
	allApps[indexToModify].AuditsConf = reqBody.AuditsConf
	allApps[indexToModify].AutoRSrpmBuild = reqBody.AutoRSrpmBuild

	// Copy the new services from reqBody
	copier.Copy(&allApps[indexToModify].Services, &reqBody.Services)

	// Wait a bit and send the corresponding event
	go func() {
		time.Sleep(genericWaitingTime)
		mockedBck.EmitEvent(rpapiv1.EvtAppStateChange, allApps[indexToModify])
	}()

	resJSON(w, allApps[indexToModify])
}

// appsPostBookmarkHandler is the handler used when a call is done on the application/{appid}/bookmark route
// with the POST method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) appsPostBookmarkHandler(w http.ResponseWriter, req *http.Request) {
	appID := reqParamGet(req, "appid")
	appUUID := uuid.FromStringOrNil(appID)

	indexToModify := -1
	for index, app := range allApps {
		if utils.UUIDEqual(app.ID, appUUID) {
			// We found the right application, save the index to modify it afterwards
			indexToModify = index
			break
		}
	}

	if indexToModify == -1 {
		resJSONErrorE(w, errors.New("The application does not exists"))
		return
	}

	// Modify the application (actually modify only what can be modified)
	allApps[indexToModify].Bookmarked = !allApps[indexToModify].Bookmarked

	// Create a UUID and a time object
	bookID, _ := uuid.NewV4()
	timeNow := time.Now()

	// Prepare the returned structure
	appBookMark := rpapiv1.RPApplicationBookmark{
		ID:              bookID,
		CreatedAt:       timeNow,
		UpdatedAt:       timeNow,
		ApplicationID:   allApps[indexToModify].ID,
		ApplicationSlug: allApps[indexToModify].Slug,
	}

	// Wait a bit and send the corresponding event
	go func() {
		time.Sleep(genericWaitingTime)
		mockedBck.EmitEvent(rpapiv1.EvtAppStateChange, allApps[indexToModify])
	}()

	resJSON(w, appBookMark)
}

// appsGetFilesListHandler is the handler used when a call is done on the application/{appid}/listfile route
// with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) appsGetFilesListHandler(w http.ResponseWriter, req *http.Request) {
	appID := reqParamGet(req, "appid")
	appUUID := uuid.FromStringOrNil(appID)

	appSlug := ""
	for _, app := range allApps {
		if utils.UUIDEqual(app.ID, appUUID) {
			// We found the right application, save the index to modify it afterwards
			appSlug = app.Slug
			break
		}
	}

	if appSlug == "" {
		resJSONErrorE(w, errors.New("The application does not exists"))
		return
	}

	answer := rpapiv1.RPFileListDeprecated{
		ListFiles: []rpapiv1.RPFile_Deprecated{
			{
				Name:     ".redpesk/autogen::service.json",
				Encoding: "base64",
				Type:     "application/json",
				Size:     657,
			},
			{
				Name:     (appSlug + "-1.0.0.tar.gz"),
				Encoding: "base64",
				Type:     "application/gzip",
				Size:     837,
			},
			{
				Name:     (appSlug + ".spec"),
				Encoding: "base64",
				Type:     "text/plain; charset=utf-8",
				Size:     507,
			},
		},
	}

	resJSON(w, answer)
}

// appsRemoveFileHandler is the handler used when a call is done on the application/{appid}/file route
// with the DELETE method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) appsRemoveFileHandler(w http.ResponseWriter, req *http.Request) {
	appID := reqParamGet(req, "appid")
	appUUID := uuid.FromStringOrNil(appID)

	appSlug := ""
	var appIndex int
	for index, app := range allApps {
		if utils.UUIDEqual(app.ID, appUUID) {
			// We found the right application, save the index to modify it afterwards
			appSlug = app.Slug
			appIndex = index
			break
		}
	}

	if appSlug == "" {
		resJSONErrorE(w, errors.New("The application does not exists"))
		return
	}

	// Get the filename (in query parameter)
	filename := req.FormValue("name")

	if (filename != ".redpesk/autogen::service.json") && (filename != (appSlug + "-1.0.0.tar.gz")) && (filename != (appSlug + ".spec")) {
		resJSONErrorE(w, errors.New("The file does not exists for this application"))
		return
	}

	resJSON(w, allApps[appIndex])
}

// appsDownloadFileHandler is the handler used when a call is done on the application/{appid}/file route
// with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) appsDownloadFileHandler(w http.ResponseWriter, req *http.Request) {
	appID := reqParamGet(req, "appid")
	appUUID := uuid.FromStringOrNil(appID)

	appSlug := ""
	for _, app := range allApps {
		if utils.UUIDEqual(app.ID, appUUID) {
			// We found the right application, save the index to modify it afterwards
			appSlug = app.Slug
			break
		}
	}

	if appSlug == "" {
		resJSONErrorE(w, errors.New("The application does not exists"))
		return
	}

	// Get the filename (in query parameter)
	filename := req.FormValue("name")

	if (filename != ".redpesk/autogen::service.json") && (filename != (appSlug + "-1.0.0.tar.gz")) && (filename != (appSlug + ".spec")) {
		resJSONErrorE(w, errors.New("The file does not exists for this application"))
		return
	}

	// Prepare the file to send to the client
	extension := filepath.Ext(filename)
	var content string
	switch extension {
	case ".json":
		filename = "autogen::service.json"
		content = appServiceFile

	case ".spec":
		content = appSpecfile

	case ".gz":
		content = appGitpkgTarGZ

	default:
		resJSONErrorE(w, errors.New("unsupported extension "+extension))
		return
	}

	// Write the file in temporary directory to serve it afterwards
	fullPathFile := path.Join(baseDirTests, filename)
	utils.WriteToFile(fullPathFile, content)

	// Serve the file
	http.ServeFile(w, req, fullPathFile)
}

// appsGetServicesHandler is the handler used when a call is done on the application/services route
// with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) appsGetServicesHandler(w http.ResponseWriter, req *http.Request) {

	// return the default services list
	var jsonBlob = []byte(`{
		"import_gitpkg": {
			"summary": "Import a Gitpkg into redpesk environment",
			"description": "Import an external Gitpkg repository, by addressing its URL. Import the whole repository if no Git reference given, else based the imported Gitpkg on it",
			"version": "1.0.0",
			"parameters": {
			"url": {
				"description": "Gitpkg URL",
				"type": "string",
				"value": ""
			},
			"ref": {
				"description": "Git reference",
				"type": "string",
				"value": ""
			}
			}
		},
		"download_sources": {
			"summary": "Download Specfile sources",
			"description": "Download Specfile sources when addressed by URL",
			"version": "1.0.0",
			"parameters": {
			"download": {
				"description": "Download sources",
				"type": "boolean",
				"value": false
			},
			"force": {
				"description": "Overwrite sources",
				"type": "boolean",
				"value": false
			}
			}
		},
		"excludearch": {
			"summary": "Exclude architecture(s) from builds",
			"description": "Architecture(s) must be declared within redpesk project and are excluded by using Excludearch macro",
			"version": "1.0.0",
			"parameters": {
			"arches": {
				"description": "Array of architecture",
				"type": "string_list",
				"value": [
				""
				]
			}
			}
		},
		"setverrel": {
			"summary": "Automatic application version-release update",
			"description": "Increase the release each time the specfile is changed. Set the version according to the nearest Git Source tag (X.X.X) found from application revision in application sources",
			"version": "1.0.0",
			"parameters": {
			"auto": {
				"description": "Set automatically version & release",
				"type": "boolean",
				"value": false
			}
			}
		}
		}`)

	defaultServices := make(map[string]interface{})
	err := json.Unmarshal(jsonBlob, &defaultServices)
	if err != nil {
		resJSONErrorE(w, err)
		return
	}
	resJSON(w, defaultServices)
}

// appsDeleteHandler is the handler used when a call is done on the applications/{appid} route
// with the DELETE method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) appsDeleteHandler(w http.ResponseWriter, req *http.Request) {
	appID := reqParamGet(req, "appid")
	appUUID := uuid.FromStringOrNil(appID)

	tmpApps := []rpapiv1.RPApplicationFull{}
	resApp := rpapiv1.RPApplicationFull{}
	for _, app := range allApps {
		if !utils.UUIDEqual(app.ID, appUUID) {
			tmpApps = append(tmpApps, app)
		} else {
			// Copy the application we will need to send at the end
			copier.Copy(&resApp, &app)
		}
	}

	// Set the new tab of projects
	allApps = tmpApps

	// Wait a bit and send the corresponding event
	go func() {
		time.Sleep(genericWaitingTime)
		mockedBck.EmitEvent(rpapiv1.EvtAppDelete, resApp)
	}()

	resJSON(w, resApp)
}

// appsPostFileHandler is the handler used when a call is done on the applications/{appid}/file route
// with the POST method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) appsPostFileHandler(w http.ResponseWriter, req *http.Request) {
	// Get the project UUID concerned by the new application
	appID := reqParamGet(req, "appid")
	appUUID := uuid.FromStringOrNil(appID)

	// Check that the app exists
	appExists := false
	appIndex := 0
	for index, app := range allApps {
		if utils.UUIDEqual(app.ID, appUUID) {
			appExists = true
			appIndex = index
			break
		}
	}
	if !appExists {
		resJSONErrorE(w, errors.New("The application does not exists"))
		return
	}

	// UPLOAD FILE
	uploadedFile, _, err := req.FormFile("file")
	if err != nil {
		resJSONErrorE(w, err)
		return
	}
	defer uploadedFile.Close()
	defer req.MultipartForm.RemoveAll()

	destFile := "/tmp/uploaded_file"

	upFile, err := os.Create(destFile)
	if err != nil {
		resJSONErrorE(w, err)
		return
	}
	defer upFile.Close()

	// Copy fileToCopy in destinationFile
	_, err = io.Copy(upFile, uploadedFile)
	if err != nil {
		resJSONErrorE(w, err)
		return
	}

	// Wait a bit and send the corresponding event
	go func() {
		time.Sleep(genericWaitingTime)
		mockedBck.EmitEvent(rpapiv1.EvtAppStateChange, allApps[appIndex])
	}()

	resJSON(w, allApps[appIndex])
}

// buildsStartHandler is the handler used when a call is done on the applications/{appid}/applications/builds route
// with the POST method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) buildsStartHandler(w http.ResponseWriter, req *http.Request) {
	// Get the application UUID concerned by the build
	appID := reqParamGet(req, "appid")
	appUUID := uuid.FromStringOrNil(appID)

	// Get the distribution requested and check that it exists for the app
	startBuildObj := &rpapiv1.RPBuildStart{}
	if err := reqBind(req, startBuildObj); err != nil {
		resJSONErrorE(w, err)
		return
	}

	// Get eventually the build type wanted
	buildType := req.FormValue("build_type")
	if buildType == "" {
		buildType = string(rpapiv1.RPBuildApp)
	}

	// Start the build for one app
	currentBuild, err := mockedBck.startOneAppBuild(appUUID, &startBuildObj.Distributions[0], startBuildObj.Comment, buildType)
	if err != nil {
		resJSONErrorE(w, err)
	}

	localBuilds := rpapiv1.RPBuilds{
		Page:         1,
		PerPage:      50,
		TotalEntries: 1,
	}
	localBuilds.Builds = append(localBuilds.Builds, convertBuildFullToBuild(currentBuild))

	resJSON(w, localBuilds)
}

// buildsGetAllHandler is the handler used when a call is done on the applications/{appid}/builds route
// with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) buildsGetAllHandler(w http.ResponseWriter, req *http.Request) {
	// Get the project UUID concerned by the list of applications
	appID := reqParamGet(req, "appid")
	appUUID := uuid.FromStringOrNil(appID)

	// Get the page number wanted
	pageNb, _ := strconv.Atoi(req.FormValue("page"))

	// Get the per_page number wanted
	perPageNb, _ := strconv.Atoi(req.FormValue("per_page"))

	// Get eventually the build type wanted
	buildType := req.FormValue("build_type")
	if buildType == "" {
		buildType = string(rpapiv1.RPBuildApp)
	}

	// Get possibly the order wanted
	orderWanted := req.FormValue("sort")
	if orderWanted != "" && orderWanted != "desc" {
		resJSONErrorE(w, errors.New("This mock does not handle anything except descendant sorting"))
		return
	} else if orderWanted != "" && perPageNb > 1 {
		resJSONErrorE(w, errors.New("This mock does not handle sorting more than one build"))
		return
	}

	// Initialize the structure that will be returned
	retBuilds := rpapiv1.RPBuilds{
		Page:    pageNb,
		PerPage: perPageNb,
	}

	// Calculate the minimum and maximum index wanted by the user
	minIndex := (pageNb - 1) * perPageNb
	maxIndex := (pageNb * perPageNb) - 1

	// Loop over builds and save the ones we want to send
	count := 0
	highestID := 0
	highIDIndex := 0

	for index, bld := range allBuilds {
		if utils.UUIDEqual(bld.Parent.ApplicationID, appUUID) {
			// Manage descending order case
			if orderWanted != "" && orderWanted == "desc" {
				bldID, _ := strconv.Atoi(bld.ID)
				if bldID > highestID {
					highestID = bldID
					highIDIndex = index
				}

				// Main case
			} else if (count >= minIndex) && (count <= maxIndex) && (bld.BuildType == rpapiv1.RPBuildType(buildType)) {
				retBuilds.Builds = append(retBuilds.Builds, convertBuildFullToBuild(bld))
			}
			count++
		}
	}

	if orderWanted != "" && orderWanted == "desc" {
		retBuilds.Builds = append(retBuilds.Builds, convertBuildFullToBuild(allBuilds[highIDIndex]))
	}

	// Update the total entries
	retBuilds.TotalEntries = count

	resJSON(w, retBuilds)
}

// buildsGetHandler is the handler used when a call is done on the builds/{bldid} route
// with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) buildsGetHandler(w http.ResponseWriter, req *http.Request) {
	bldID := reqParamGet(req, "bldid")

	for _, bld := range allBuilds {
		if bld.ID == bldID {
			resJSON(w, bld)
			break
		}
	}
}

// buildsCancelHandler is the handler used when a call is done on the builds/{bldid}/cancel
// route with the POST method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) buildsCancelHandler(w http.ResponseWriter, req *http.Request) {
	// Look if the build exists
	bldID := reqParamGet(req, "bldid")

	for _, bld := range allBuilds {
		if bld.ID == bldID {
			// Found the test to cancel, check that it can be cancelled!
			if (bld.Status == rpapiv1.RPBuildTaskStatusDone) || (bld.Status == rpapiv1.RPBuildTaskStatusFailed) ||
				(bld.Status == rpapiv1.RPBuildTaskStatusCanceled) {
				resJSONErrorE(w, errors.New(fmt.Sprintf("cannot cancel build when build' status is %v", bld.Status)))
				return
			} else {
				buildsToCancel = append(buildsToCancel, bldID)
				resJSON(w, bld)
				break
			}
		}
	}
}

// testsStartHandler is the handler used when a call is done on the builds/{bldid}/tests route
// with the POST method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) testsStartHandler(w http.ResponseWriter, req *http.Request) {
	// Get the build ID concerned by the new application
	bldID := reqParamGet(req, "bldid")

	// Check that the build exists
	bldExists := false
	appSlug := ""
	prjSlug := ""
	imgSlug := ""
	var appUUID uuid.UUID
	var imgUUID uuid.UUID
	var prjUUID uuid.UUID
	for _, build := range allBuilds {
		if build.ID == bldID {
			bldExists = true
			appSlug = build.Parent.ApplicationSlug
			prjSlug = build.Parent.ProjectSlug
			imgSlug = build.Parent.ImageSlug
			appUUID = build.Parent.ApplicationID
			prjUUID = build.Parent.ProjectID
			imgUUID = build.Parent.ImageID
			break
		}
	}
	if !bldExists {
		resJSONErrorE(w, errors.New("The build does not exists"))
		return
	}

	// Get the request body
	reqBody := &rpapiv1.RPTestStart{}
	if err := reqBind(req, reqBody); err != nil {
		resJSONErrorE(w, err)
		return
	}

	// Get the board model and board choosen for the test
	boardUUID := uuid.Nil
	boardSlug := ""
	modelBoardUUID := uuid.Nil
	modelBoardSlug := ""

	// Define default value
	if reqBody.TargetModelID == uuid.Nil {
		reqBody.TargetModelID = uuid.FromStringOrNil("f2418145-a15b-4b8a-8b34-f51b24d6c88d")
	}

	for _, board := range allBoards {
		if utils.UUIDEqual(board.ModelID, reqBody.TargetModelID) {
			boardUUID = board.ID
			boardSlug = board.Name
			modelBoardUUID = board.ModelID
			modelBoardSlug = board.ModelName
		}
	}

	if boardSlug == "" {
		resJSONErrorE(w, errors.New("The specified board model has not been found in the \"database\""))
		return
	}

	// Get the current time
	timeNow := time.Now()

	// Increment counter
	testsCounter++

	// Create the test object
	currentTest := rpapiv1.RPTest{
		ID:             strconv.Itoa(testsCounter),
		CreatedAt:      timeNow,
		Started:        timeNow,
		Status:         rpapiv1.RPTestStatusPending,
		ImageID:        reqBody.ImageID,
		ModelBoardID:   modelBoardUUID,
		ModelBoardSlug: modelBoardSlug,
		BoardID:        boardUUID,
		BoardSlug:      boardSlug,
		Parent: rpapiv1.RPTestParent{
			BuildID:     bldID,
			AppID:       appUUID,
			AppSlug:     appSlug,
			ProjectID:   prjUUID,
			ProjectSlug: prjSlug,
			ImageID:     imgUUID,
			ImageSlug:   imgSlug,
		},
		Comment: reqBody.Comment,
	}

	// Add the build in the builds slice
	allTests.Tests = append(allTests.Tests, currentTest)

	// Start go routine to emit events like if it was building in reality
	go mockedBck.simulateSuccessTest()

	resJSON(w, currentTest)
}

// testsGetAllFromBuildHandler is the handler used when a call is done on the build/{buildid}/tests route
// with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) testsGetAllFromBuildHandler(w http.ResponseWriter, req *http.Request) {
	// Get the build ID concerned by the list of tests
	buildID := reqParamGet(req, "bldid")

	// Get the page number wanted
	pageNb, _ := strconv.Atoi(req.FormValue("page"))

	// Get the per_page number wanted
	perPageNb, _ := strconv.Atoi(req.FormValue("per_page"))

	// Initialize the structure that will be returned
	retTests := rpapiv1.RPTests{
		Page:    pageNb,
		PerPage: perPageNb,
	}

	// Calculate the minimum and maximum index wanted by the user
	minIndex := (pageNb - 1) * perPageNb
	maxIndex := (pageNb * perPageNb) - 1

	// Loop over tests and save the ones we want to send
	count := 0
	for _, test := range allTests.Tests {
		if test.Parent.BuildID == buildID {
			if (count >= minIndex) && (count <= maxIndex) {
				retTests.Tests = append(retTests.Tests, test)
			}
			count++
		}
	}

	// Update the total entries
	retTests.TotalEntries = count

	resJSON(w, retTests)
}

// testsGetHandler is the handler used when a call is done on the tests/{tstid}
// route with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) testsGetHandler(w http.ResponseWriter, req *http.Request) {
	// Look if the test exists
	testID := reqParamGet(req, "tstid")

	for _, test := range allTests.Tests {
		if test.ID == testID {
			resJSON(w, test)
			break
		}
	}
}

// testsGetLogsListHandler is the handler used when a call is done on the tests/{tstid}/listfiles
// route with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) testsGetLogsListHandler(w http.ResponseWriter, req *http.Request) {
	// Look if the test exists
	testID := reqParamGet(req, "tstid")

	for _, test := range allTests.Tests {
		if test.ID == testID {

			// Initialize the structure to return and add the files that are systematically here
			retStruct := rpapiv1.RPFileListDeprecated{
				ListFiles: []rpapiv1.RPFile_Deprecated{
					{Name: "tests_logs.zip"},
					{Name: "all_tests.tap"},
					{Name: "boot.log"},
					{Name: "console.log"},
				},
			}

			resJSON(w, retStruct)
			break
		}
	}
}

// testsGetLogsHandler is the handler used when a call is done on the tests/{tstid}/logs
// route with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) testsGetLogsHandler(w http.ResponseWriter, req *http.Request) {
	// // Look if the test exists
	// testID := reqParamGet(req, "tstid")

	// Get the file name wanted
	fileName := req.FormValue("name")

	if fileName == "" {
		resJSONErrorE(w, errors.New("The query parameter \"name\" needs to be set!"))
		return
	}

	fileToServe := path.Join(baseDirTests, fileName)

	// Check that the file exists
	if _, err := os.Stat(fileToServe); errors.Is(err, os.ErrNotExist) {
		resJSONErrorE(w, errors.New("This file does not exist"))
		return
	}

	// Serve the file
	http.ServeFile(w, req, fileToServe)
}

// testsPostCancelHandler is the handler used when a call is done on the tests/{tstid}/cancel
// route with the POST method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) testsPostCancelHandler(w http.ResponseWriter, req *http.Request) {
	// Look if the test exists
	testID := reqParamGet(req, "tstid")

	for _, test := range allTests.Tests {
		if test.ID == testID {
			// Found the test to cancel, check that it can be cancelled!
			if (test.Status == rpapiv1.RPTestStatusSuccess) || (test.Status == rpapiv1.RPTestStatusFailed) ||
				(test.Status == rpapiv1.RPTestStatusTimedOut) || (test.Status == rpapiv1.RPTestStatusAborted) {
				resJSONErrorE(w, errors.New(fmt.Sprintf("cannot cancel tests when tests' status is %v", test.Status)))
				return
			} else {
				abortCurrentTest = true
				resJSON(w, test)
				break
			}
		}
	}
}

// auditsGetToolsHandler is the handler used when a call is done on the audits/list route
// with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) auditsGetToolsHandler(w http.ResponseWriter, req *http.Request) {

	// return the default tools list
	var jsonBlob = []byte(`{
		"audit": [
			{
				"type": "code",
				"entry": {
					"category": "Source Code Analysis",
				  	"tools": [
					{
					  	"language": "Go",
					  	"tools": [
							{
						  		"name": "gosec",
						  		"summary": "Golang security checker",
						  		"version": "2.9.6"
							}
					  	]
					},
					{
					  	"language": "C/C++",
					  	"tools": [
						{
						  	"name": "flawfinder",
						  	"summary": "Examines C/C++ source code for security flaws",
						  	"version": "2.0.19"
						}
					  	]
					}
				  	]
				}
			}
		]
	}`)

	defaultAuditTools := make(map[string]interface{})
	err := json.Unmarshal(jsonBlob, &defaultAuditTools)
	if err != nil {
		resJSONErrorE(w, err)
		return
	}
	resJSON(w, defaultAuditTools)
}

// auditsStartHandler is the handler used when a call is done on the applications/{appid}/audits route
// with the POST method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) auditsStartHandler(w http.ResponseWriter, req *http.Request) {
	// Get the application UUID concerned by the build
	appID := reqParamGet(req, "appid")
	appUUID := uuid.FromStringOrNil(appID)

	// Get the structure given by the client as input
	auditsReq := &AuditInputCustom{}
	if err := reqBind(req, auditsReq); err != nil {
		resJSONErrorE(w, err)
		return
	}

	// Check that the application exists
	appExists := false
	appSlug := ""
	appPrjUUID := uuid.Nil
	appPrjSlug := ""
	for _, app := range allApps {
		if utils.UUIDEqual(app.ID, appUUID) {
			appExists = true
			appSlug = app.Slug
			appPrjUUID = app.Parent.ProjectID
			appPrjSlug = app.Parent.ProjectSlug

			break
		}
	}
	if !appExists {
		resJSONErrorE(w, errors.New("The application does not exists"))
		return
	}

	// Get the current time
	timeNow := time.Now()

	// Increment counter
	auditsCounter++

	// Create the audit object
	currentAudit := rpapiv1.RPAuditFull{
		RPAudit: rpapiv1.RPAudit{
			ID:        auditsCounter,
			CreatedAt: timeNow,
			UpdatedAt: timeNow,
			Started:   timeNow,
			Status:    rpapiv1.RPAuditStatusTesting,
			Parent: rpapiv1.RPAuditParent{
				AppID:       appUUID,
				AppSlug:     appSlug,
				ProjectID:   appPrjUUID,
				ProjectSlug: appPrjSlug,
			},
			BaseURL: fmt.Sprintf("http://localhost:8000/api/v1/audits/%d", auditsCounter),
		},
	}

	// Fill the value for the audits analysis

	infoTotal := 0
	unknownTotal := 0
	lowTotal := 0
	mediumTotal := 0
	highTotal := 0
	criticalTotal := 0

	auditAllResults := bckapi.RPAuditResults{
		Analyse:     []bckapi.RPAuditResult{},
		TotalReport: bckapi.RPAuditCodeReport{},
	}

	for _, tool := range auditsReq.Audit[0].Entry.Tools {

		// Generate randomly the value to set
		infoTmp := rand.Intn(10)
		unknownTmp := rand.Intn(10)
		lowTmp := rand.Intn(10)
		mediumTmp := rand.Intn(10)
		highTmp := rand.Intn(10)
		criticalTmp := rand.Intn(10)

		// Set these values in the analyse
		auditAllResults.Analyse = append(auditAllResults.Analyse, bckapi.RPAuditResult{
			Tool: tool,
			Report: bckapi.RPAuditCodeReport{
				Info:     infoTmp,
				Unknown:  unknownTmp,
				Low:      lowTmp,
				Medium:   mediumTmp,
				High:     highTmp,
				Critical: criticalTmp,
				Total:    (infoTmp + unknownTmp + lowTmp + mediumTmp + highTmp + criticalTmp),
			},
		})

		// Update the total results
		infoTotal += infoTmp
		unknownTotal += unknownTmp
		lowTotal += lowTmp
		mediumTotal += mediumTmp
		highTotal += highTmp
		criticalTotal += criticalTmp
	}

	// Update the overall result
	auditAllResults.TotalReport.Info = infoTotal
	auditAllResults.TotalReport.Unknown = unknownTotal
	auditAllResults.TotalReport.Low = lowTotal
	auditAllResults.TotalReport.Medium = mediumTotal
	auditAllResults.TotalReport.High = highTotal
	auditAllResults.TotalReport.Critical = criticalTotal
	auditAllResults.TotalReport.Total = infoTotal + unknownTotal + lowTotal + mediumTotal + highTotal + criticalTotal

	// Update the overall stats
	// Knowing that:
	// failed = high + medium + low + critical
	// success = info
	// skip = unknown
	// total = failed + success + skip
	currentAudit.SuccessNb = infoTotal
	currentAudit.FailedNb = highTotal + mediumTotal + lowTotal + criticalTotal
	currentAudit.SkippedNb = unknownTotal
	currentAudit.TotalNb = auditAllResults.TotalReport.Total

	// Enter the result in the audit to return
	currentAudit.Result.Audit = append(currentAudit.Result.Audit, rpapiv1.AuditBaseStruct{
		Type:  "code",
		Entry: auditAllResults,
	})

	// Add the audit in the audits slice
	allAudits = append(allAudits, currentAudit)

	// Start the go routine to emit events like if the audit was running
	go mockedBck.simulateSuccessAppAudit()

	resJSON(w, currentAudit)
}

// auditsGetHandler is the handler used when a call is done on the audits/{auditid} route
// with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) auditsGetHandler(w http.ResponseWriter, req *http.Request) {
	auditID := reqParamGet(req, "auditid")
	auditIDInt, _ := strconv.Atoi(auditID)

	for _, audit := range allAudits {
		if audit.ID == auditIDInt {
			audit.FullData = true
			resJSON(w, audit)
			break
		}
	}
}

// auditsGetAllHandler is the handler used when a call is done on the applications/{appid}/audits route
// with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) auditsGetAllHandler(w http.ResponseWriter, req *http.Request) {
	// Get the project UUID concerned by the list of applications
	appID := reqParamGet(req, "appid")
	appUUID := uuid.FromStringOrNil(appID)

	// Get the page number wanted
	pageNb, _ := strconv.Atoi(req.FormValue("page"))

	// Get the per_page number wanted
	perPageNb, _ := strconv.Atoi(req.FormValue("per_page"))

	// Get possibly the order wanted
	orderWanted := req.FormValue("sort")
	if orderWanted != "" && orderWanted != "desc" {
		resJSONErrorE(w, errors.New("This mock does not handle anything except descendant sorting"))
		return
	} else if orderWanted != "" && perPageNb > 1 {
		resJSONErrorE(w, errors.New("This mock does not handle sorting more than one build"))
		return
	}

	// Initialize the structure that will be returned
	retAudits := rpapiv1.RPAudits{
		Page:    pageNb,
		PerPage: perPageNb,
	}

	// Calculate the minimum and maximum index wanted by the user
	minIndex := (pageNb - 1) * perPageNb
	maxIndex := (pageNb * perPageNb) - 1

	// Loop over builds and save the ones we want to send
	count := 0
	highestID := 0
	highIDIndex := 0

	for index, audit := range allAudits {
		if utils.UUIDEqual(audit.Parent.AppID, appUUID) {
			// Manage descending order case
			if orderWanted != "" && orderWanted == "desc" {
				if audit.ID > highestID {
					highestID = audit.ID
					highIDIndex = index
				}

				// Main case
			} else if (count >= minIndex) && (count <= maxIndex) {
				retAudits.Audits = append(retAudits.Audits, convertAuditFullToAudit(audit))
			}
			count++
		}
	}

	if orderWanted != "" && orderWanted == "desc" {
		retAudits.Audits = append(retAudits.Audits, convertAuditFullToAudit(allAudits[highIDIndex]))
	}

	// Update the total entries
	retAudits.TotalEntries = count

	resJSON(w, retAudits)
}

// branchesPostHandler is the handler used when a call is done on the applications/{appid}/branches route
// with the POST method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) branchesPostHandler(w http.ResponseWriter, req *http.Request) {
	// Get the app UUID concerned by the request
	appID := reqParamGet(req, "appid")
	appUUID := uuid.FromStringOrNil(appID)

	// Check that the application exists
	appExists := false
	for _, app := range allApps {
		if utils.UUIDEqual(app.ID, appUUID) {
			appExists = true
			break
		}
	}
	if !appExists {
		resJSONErrorE(w, errors.New("The application does not exists"))
		return
	}

	// Get the structure given by the client as input
	branchReq := &rpapiv1.RPBranch{}
	if err := reqBind(req, branchReq); err != nil {
		resJSONErrorE(w, err)
		return
	}

	// Check if we create a new branch
	if branchReq.Source != "" {
		// Create a new branch
		newBranch := rpapiv1.RPBranchChar{
			RPBranchName:  rpapiv1.RPBranchName{Name: branchReq.Destination},
			Default:       branchReq.Default,
			Protected:     branchReq.Default,
			Distributions: branchReq.Distributions,
		}
		allAppsBranches[appUUID].ListBranches = append(allAppsBranches[appUUID].ListBranches, newBranch)
	} else if branchReq.Default {
		// Find the existing branch to set it as default
		for index, tmpBranch := range allAppsBranches[appUUID].ListBranches {
			if tmpBranch.Name == branchReq.Destination {
				allAppsBranches[appUUID].ListBranches[index].Default = true
				break
			}
		}
	}

	// If the destination branch needs to be the default one, unactivate every other default one
	if branchReq.Default {
		for index, tmpBranch := range allAppsBranches[appUUID].ListBranches {
			if tmpBranch.Name != branchReq.Destination {
				allAppsBranches[appUUID].ListBranches[index].Default = false
				break
			}
		}
	}

	// Set as active if requested
	if branchReq.Active {
		allAppsBranches[appUUID].PackageActiveBranch = branchReq.Destination
	}

	// Modify the distributions linked to this branch
	for index, tmpBranch := range allAppsBranches[appUUID].ListBranches {
		if tmpBranch.Name == branchReq.Destination {
			allAppsBranches[appUUID].ListBranches[index].Distributions = branchReq.Distributions
			break
		}
	}

	// Return OK status
	w.WriteHeader(http.StatusOK)
}

// branchesGetHandler is the handler used when a call is done on the applications/{appid}/branches route
// with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) branchesGetHandler(w http.ResponseWriter, req *http.Request) {
	// Get the app UUID concerned by the request
	appID := reqParamGet(req, "appid")
	appUUID := uuid.FromStringOrNil(appID)

	// Check that the application exists
	appExists := false
	for _, app := range allApps {
		if utils.UUIDEqual(app.ID, appUUID) {
			appExists = true
			break
		}
	}
	if !appExists {
		resJSONErrorE(w, errors.New("The application does not exists"))
		return
	}

	// Send the branches list
	resJSON(w, allAppsBranches[appUUID])
}

// branchesDeleteHandler is the handler used when a call is done on the applications/{appid}/branches route
// with the DELETE method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) branchesDeleteHandler(w http.ResponseWriter, req *http.Request) {
	// Get the app UUID concerned by the request
	appID := reqParamGet(req, "appid")
	appUUID := uuid.FromStringOrNil(appID)

	// Check that the application exists
	appExists := false
	for _, app := range allApps {
		if utils.UUIDEqual(app.ID, appUUID) {
			appExists = true
			break
		}
	}
	if !appExists {
		resJSONErrorE(w, errors.New("The application does not exists"))
		return
	}

	// Get the structure given by the client as input
	branchName := &rpapiv1.RPBranchName{}
	if err := reqBind(req, branchName); err != nil {
		resJSONErrorE(w, err)
		return
	}

	// Find the branch to delete
	indexToDel := -1
	for index, tmpBranch := range allAppsBranches[appUUID].ListBranches {
		if tmpBranch.Name == branchName.Name {
			indexToDel = index
			break
		}
	}

	if indexToDel > -1 {
		// Delete the right index
		allAppsBranches[appUUID].ListBranches[indexToDel] = allAppsBranches[appUUID].ListBranches[len(allAppsBranches[appUUID].ListBranches)-1]
		allAppsBranches[appUUID].ListBranches = allAppsBranches[appUUID].ListBranches[:len(allAppsBranches[appUUID].ListBranches)-1]
	}

	// Return OK status
	w.WriteHeader(http.StatusOK)
}

// actionAppPostHandler is the handler used when a call is done on the applications/{appid}/action route
// with the POST method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) actionAppPostHandler(w http.ResponseWriter, req *http.Request) {

	// Get the app UUID concerned by the request
	appID := reqParamGet(req, "appid")
	appUUID := uuid.FromStringOrNil(appID)

	// find application
	appStruct := rpapiv1.RPApplicationFull{}
	for _, app := range allApps {
		if utils.UUIDEqual(app.ID, appUUID) {
			// Copy the application
			copier.Copy(&appStruct, &app)
			break
		}
	}

	// Check the application exist
	if appStruct.ID == uuid.Nil {
		resJSONErrorE(w, errors.New("The app does not exists "))
		return
	}

	// Get the action (in query parameter)
	actionParam := req.FormValue("action")
	if actionParam == "" {
		resJSONErrorE(w, errors.New("Param is not set"))
		return
	}

	if actionParam == "copy" {
		// Get the structure given by the client as input
		appCopyStruct := &rpapiv1.RPApplicationCopy{}
		if err := reqBind(req, appCopyStruct); err != nil {
			resJSONErrorE(w, err)
			return
		}

		// Update the creation and update time
		timeNow := time.Now()
		appStruct.CreatedAt = timeNow
		appStruct.UpdatedAt = timeNow

		// Set the UUID
		appStruct.ID, _ = uuid.NewV4()

		// Set the project ID
		appStruct.Parent.ProjectID = appCopyStruct.Destination

		// Add the app in our list
		allApps = append(allApps, appStruct)

		// Wait a bit and send the corresponding event
		go func() {
			time.Sleep(genericWaitingTime)
			mockedBck.EmitEvent(rpapiv1.EvtAppAdd, appStruct)
		}()
	} else if actionParam == "refresh_files_package" {
		// Wait a bit and send the corresponding event
		go func() {
			time.Sleep(genericWaitingTime)
			mockedBck.EmitEvent(rpapiv1.EvtAppStateChange, appStruct)
		}()
	}

	resJSON(w, appStruct)
}

// imagesListHandler is the handler used when a call is done on the images route
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) imagesListHandler(w http.ResponseWriter, req *http.Request) {

	// Get the type wanted by the user
	typesWanted := req.FormValue("type")
	if typesWanted == "" {
		typesWanted = "standard"
	}

	// Get the page number wanted
	pageNb, _ := strconv.Atoi(req.FormValue("page"))

	// Get the per_page number wanted
	perPageNb, _ := strconv.Atoi(req.FormValue("per_page"))

	// Get the potential "filters" string
	filtersStr := req.FormValue("filters")

	// Get the filters
	slugFilter := ""
	if strings.HasPrefix(filtersStr, "slug=") {
		slugFilter = strings.ReplaceAll(filtersStr, "slug=", "")
	}
	nameFilter := ""
	if strings.HasPrefix(filtersStr, "name=") {
		nameFilter = strings.ReplaceAll(filtersStr, "name=", "")
		nameFilter = strings.ReplaceAll(nameFilter, "+", " ")
	}
	// Initialize the structure that will be returned
	retImages := rpapiv1.RPImages{
		Page:    pageNb,
		PerPage: perPageNb,
	}

	// Calculate the minimum and maximum index wanted by the user
	minIndex := (pageNb - 1) * perPageNb
	maxIndex := (pageNb * perPageNb) - 1

	// Convert the type(s) to a slice of strings
	typesWantedSli := strings.Split(typesWanted, ",")

	// Loop over images and save the ones we want to send
	count := 0
	for _, image := range allImages {
		if utils.ContainsString(typesWantedSli, string(image.Type)) {
			if (count >= minIndex) && (count <= maxIndex) &&
				(slugFilter == "" || image.Slug == slugFilter) &&
				(nameFilter == "" || image.Name == nameFilter) {
				retImages.Images = append(retImages.Images, convertImageFullToImage(image))
			}
			count++
		}
	}

	// Update the total entries
	retImages.TotalEntries = len(retImages.Images)

	resJSON(w, retImages)
}

// imgGetFromPrjHandler is the handler used when a call is done on the projects/{prjid}/images route
// with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) imgGetFromPrjHandler(w http.ResponseWriter, req *http.Request) {
	// Get the project UUID concerned by the list of applications
	prjID := reqParamGet(req, "prjid")
	prjUUID := uuid.FromStringOrNil(prjID)

	// Get the page number wanted
	pageNb, _ := strconv.Atoi(req.FormValue("page"))

	// Get the per_page number wanted
	perPageNb, _ := strconv.Atoi(req.FormValue("per_page"))

	// Initialize the structure that will be returned
	retImgs := rpapiv1.RPImages{
		Page:    pageNb,
		PerPage: perPageNb,
	}

	// Calculate the minimum and maximum index wanted by the user
	minIndex := (pageNb - 1) * perPageNb
	maxIndex := (pageNb * perPageNb) - 1

	// Loop over apps and save the ones we want to send
	count := 0
	for _, img := range allImages {
		if utils.UUIDEqual(img.ParentProjectID, prjUUID) {
			if (count >= minIndex) && (count <= maxIndex) {
				retImgs.Images = append(retImgs.Images, convertImageFullToImage(img))
			}
			count++
		}
	}

	// Update the total entries
	retImgs.TotalEntries = len(retImgs.Images)

	resJSON(w, retImgs)
}

// imagesAddHandler is the handler used when a call is done on the images route
// with the POST method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) imagesAddHandler(w http.ResponseWriter, req *http.Request) {
	// Get the request body
	reqBody := &rpapiv1.RPImageFull{}
	if err := reqBind(req, reqBody); err != nil {
		resJSONErrorE(w, err)
		return
	}

	// Set the UUID
	reqBody.ID, _ = uuid.NewV4()

	// Set the slug from the name
	if reqBody.Slug == "" {
		reqBody.Slug = strings.ToLower(strings.ReplaceAll(reqBody.Name, " ", "-"))
	}

	// Set the install tree URL
	reqBody.InstallTreeURL = fmt.Sprintf("https://download.redpesk.bzh/redpesk-lts/arz-1.0/imager-os/%s/", reqBody.Architecture)

	// Set the type of project to "standard"
	reqBody.Type = "standard"

	// Set state to stable
	reqBody.State = rpapiv1.RPImageStableState

	// Create a JWT
	jwtID, _ := uuid.NewV4()
	jwtValue1, _ := uuid.NewV4()
	jwtValue2, _ := uuid.NewV4()
	reqBody.ImageAccessTokens = []rpapiv1.RPObjectAccessToken{
		{
			ID:  jwtID,
			JWT: jwtValue1.String() + "_" + jwtValue2.String(),
		},
	}

	// Rework the kickstart url if needed
	if reqBody.KickURL == "git+?.#" || reqBody.KickURL == "" {
		reqBody.KickURL = "git+https://github.com/redpesk-infra/rp-kickstarts.git?./#arz-1.0"
	}

	// Rework the kickstart if needed
	if reqBody.Kickstart == "" {
		reqBody.Kickstart = "Redpesk-minimal-x86_64-smack.ks"
	}

	// Look for the parent project in the list
	if reqBody.ParentProjectID != uuid.Nil {
		for _, prj := range allProjects {
			if utils.UUIDEqual(prj.ID, reqBody.ParentProjectID) {
				reqBody.ParentProjectSlug = prj.Slug
				break
			}
		}
	}

	// Add the projects in our list
	allImages = append(allImages, *reqBody)

	// Wait a bit and send the corresponding event
	go func() {
		time.Sleep(genericWaitingTime)
		mockedBck.EmitEvent(rpapiv1.EvtImageAdd, reqBody)
	}()

	resJSON(w, reqBody)
}

// imagesImportHandler is the handler used when a call is done on the images-mngt/import route
// with the POST method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) imagesImportHandler(w http.ResponseWriter, req *http.Request) {
	// Get the request body
	reqBody := &rpapiv1.RPImageFull{}
	if err := reqBind(req, reqBody); err != nil {
		resJSONErrorE(w, err)
		return
	}

	// Set the UUID
	reqBody.ID, _ = uuid.NewV4()

	// Set the slug from the name
	if reqBody.Slug == "" {
		reqBody.Slug = strings.ToLower(strings.ReplaceAll(reqBody.Name, " ", "-"))
	}

	// Set the type of project to "external"
	reqBody.Type = "external"

	// Add the projects in our list
	allImages = append(allImages, *reqBody)

	// Wait a bit and send the corresponding event
	go func() {
		time.Sleep(genericWaitingTime)
		mockedBck.EmitEvent(rpapiv1.EvtImageAdd, reqBody)
	}()

	resJSON(w, reqBody)
}

// imagesGetHandler is the handler used when a call is done on the images/{id} route
// with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) imagesGetHandler(w http.ResponseWriter, req *http.Request) {
	imageID := reqParamGet(req, "imageid")
	imageUUID := uuid.FromStringOrNil(imageID)

	for _, image := range allImages {
		if utils.UUIDEqual(image.ID, imageUUID) {
			image.FullData = true
			resJSON(w, image)
			break
		}
	}
}

// imagesPutHandler is the handler used when a call is done on the images/{id} route
// with the PUT method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) imagesPutHandler(w http.ResponseWriter, req *http.Request) {
	imageID := reqParamGet(req, "imageid")
	imageUUID := uuid.FromStringOrNil(imageID)

	// Get the request body
	reqBody := &rpapiv1.RPImage{}
	if err := reqBind(req, reqBody); err != nil {
		resJSONErrorE(w, err)
		return
	}

	var indexToModify int
	for index, img := range allImages {
		if utils.UUIDEqual(img.ID, imageUUID) {
			// We found the right image, save the index to modify it afterwards
			indexToModify = index
			break
		}
	}

	// Modify the image
	allImages[indexToModify].Name = reqBody.Name
	allImages[indexToModify].Slug = reqBody.Slug
	allImages[indexToModify].DistributionID = reqBody.DistributionID
	allImages[indexToModify].DistributionSlug = reqBody.DistributionSlug
	allImages[indexToModify].Version = reqBody.Version
	allImages[indexToModify].Architecture = reqBody.Architecture
	allImages[indexToModify].SecurityModel = reqBody.SecurityModel
	allImages[indexToModify].URL = reqBody.URL
	allImages[indexToModify].KickURL = reqBody.KickURL
	allImages[indexToModify].Kickstart = reqBody.Kickstart
	allImages[indexToModify].ParentProjectID = reqBody.ParentProjectID
	allImages[indexToModify].Type = reqBody.Type
	allImages[indexToModify].InstallTreeURL = reqBody.InstallTreeURL
	allImages[indexToModify].Disabled = reqBody.Disabled
	allImages[indexToModify].AdditionalRepos = make([]string, len(reqBody.AdditionalRepos))
	copy(allImages[indexToModify].AdditionalRepos, reqBody.AdditionalRepos)

	// Wait a bit and send the corresponding event
	go func() {
		time.Sleep(genericWaitingTime)
		mockedBck.EmitEvent(rpapiv1.EvtImageStateChange, allImages[indexToModify])
	}()

	resJSON(w, allImages[indexToModify])
}

// imagesDeleteHandler is the handler used when a call is done on the images/{id} route
// with the DELETE method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) imagesDeleteHandler(w http.ResponseWriter, req *http.Request) {
	imageID := reqParamGet(req, "imageid")
	imageUUID := uuid.FromStringOrNil(imageID)

	tmpImgs := []rpapiv1.RPImageFull{}
	resImg := rpapiv1.RPImage{}
	for _, img := range allImages {
		if !utils.UUIDEqual(img.ID, imageUUID) {
			tmpImgs = append(tmpImgs, img)
		} else {
			// Copy the image we will need to send at the end
			copier.Copy(&resImg, &img)
		}
	}

	// Set the new tab of images
	allImages = tmpImgs

	// Wait a bit and send the corresponding event
	go func() {
		time.Sleep(genericWaitingTime)
		mockedBck.EmitEvent(rpapiv1.EvtImageDelete, resImg)
	}()

	resJSON(w, resImg)
}

// imagesBuildHandler is the handler used when a call is done on the images/{imageid}/builds route
// with the POST method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) imagesBuildHandler(w http.ResponseWriter, req *http.Request) {
	// Get the image UUID concerned by the build
	imageID := reqParamGet(req, "imageid")
	imageUUID := uuid.FromStringOrNil(imageID)

	// Get the build start object
	startBuildObj := &rpapiv1.RPBuildStart{}
	if err := reqBind(req, startBuildObj); err != nil {
		resJSONErrorE(w, err)
		return
	}

	// Start the image build
	currentBuild, err := mockedBck.startOneImgBuild(imageUUID, startBuildObj.Comment)
	if err != nil {
		resJSONErrorE(w, err)
		return
	}

	// since Armel 1.5.0, return struct is rpapiv1.RPBuilds (like for srpm or app build)
	res := rpapiv1.RPBuilds{
		Page:    1, // the current/started build
		PerPage: 10,
		Reverse: false,
		Builds:  []rpapiv1.RPBuild{currentBuild.RPBuild},
	}

	resJSON(w, res)
}

// imagesBuildsListHandler is the handler used when a call is done on the images/{imageid}/builds route
// with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) imagesBuildsListHandler(w http.ResponseWriter, req *http.Request) {
	// Get the image UUID concerned by the build
	imageID := reqParamGet(req, "imageid")
	imageUUID := uuid.FromStringOrNil(imageID)

	// Get the page number wanted
	pageNb, _ := strconv.Atoi(req.FormValue("page"))

	// Get the per_page number wanted
	perPageNb, _ := strconv.Atoi(req.FormValue("per_page"))

	// Get possibly the order wanted
	orderWanted := req.FormValue("sort")
	if orderWanted != "" && orderWanted != "desc" {
		resJSONErrorE(w, errors.New("This mock does not handle anything except descendant sorting"))
		return
	} else if orderWanted != "" && perPageNb > 1 {
		resJSONErrorE(w, errors.New("This mock does not handle sorting more than one build"))
		return
	}

	// Initialize the structure that will be returned
	retBuilds := rpapiv1.RPBuilds{
		Page:    pageNb,
		PerPage: perPageNb,
	}

	// Calculate the minimum and maximum index wanted by the user
	minIndex := (pageNb - 1) * perPageNb
	maxIndex := (pageNb * perPageNb) - 1

	// Loop over builds and save the ones we want to send
	count := 0
	highestID := 0
	highIDIndex := 0

	for index, bld := range allBuilds {
		if utils.UUIDEqual(bld.Parent.ImageID, imageUUID) {
			// Manage descending order case
			if orderWanted != "" && orderWanted == "desc" {
				bldID, _ := strconv.Atoi(bld.ID)
				if bldID > highestID {
					highestID = bldID
					highIDIndex = index
				}

				// Main case
			} else if (count >= minIndex) && (count <= maxIndex) {
				retBuilds.Builds = append(retBuilds.Builds, convertBuildFullToBuild(bld))
			}
			count++
		}
	}

	if orderWanted != "" && orderWanted == "desc" {
		retBuilds.Builds = append(retBuilds.Builds, convertBuildFullToBuild(allBuilds[highIDIndex]))
	}

	// Update the total entries
	retBuilds.TotalEntries = len(retBuilds.Builds)

	resJSON(w, retBuilds)
}

// imagesReleaseHandler is the handler used when a call is done on the images/{imageid}/release route
// with the POST method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) imagesReleaseHandler(w http.ResponseWriter, req *http.Request) {

	imageID := reqParamGet(req, "imageid")
	imgUUID := uuid.FromStringOrNil(imageID)

	imgIndex := -1
	for index, img := range allImages {
		if utils.UUIDEqual(img.ID, imgUUID) {
			// We found the right image, save the index to modify it afterwards
			imgIndex = index
			break
		}
	}

	if imgIndex == -1 {
		resJSONErrorE(w, errors.New("The image does not exists"))
		return
	}

	// Get the request body
	reqBody := &rpapiv1.RPReleaseImageFull{}
	if err := reqBind(req, reqBody); err != nil {
		resJSONErrorE(w, err)
		return
	}

	// Fill the structure to store in the "released-images"
	reqBody.ID, _ = uuid.NewV4()
	reqBody.Slug = allImages[imgIndex].Slug + "-" + reqBody.Version
	reqBody.Name = allImages[imgIndex].Name
	timeNow := time.Now()
	reqBody.CreatedAt = timeNow
	reqBody.UpdatedAt = timeNow
	reqBody.Type = rpapiv1.RPImageTypeReleased
	reqBody.State = rpapiv1.RPImageStableState
	reqBody.ImageReferenceID = allImages[imgIndex].ID
	reqBody.ImageReferenceSlug = allImages[imgIndex].Slug

	// Fill the full release info
	// Not necessary for tests

	// Create a JWT
	jwtID, _ := uuid.NewV4()
	jwtValue1, _ := uuid.NewV4()
	jwtValue2, _ := uuid.NewV4()
	reqBody.ImageAccessTokens = []rpapiv1.RPObjectAccessToken{
		{
			ID:  jwtID,
			JWT: jwtValue1.String() + "_" + jwtValue2.String(),
		},
	}

	// Create the URL
	reqBody.URL = fmt.Sprintf("http://localhost:8000/download/images/release/%s/", reqBody.Slug)

	// Add the released image in the array
	allReleasedImages = append(allReleasedImages, *reqBody)

	// Add a new audience in our audience list
	tmpAudience := rpapiv1.RPObjectRoles{
		ObjectID:           reqBody.ID,
		Stack:              false,
		IdentityRolesLists: make(map[string]*rpapiv1.RPIdentityRolesLists),
	}
	// Add all the users as available
	for _, user := range allUsers.Users {
		tmpRolesList := rpapiv1.RPIdentityRolesLists{
			ID:           user.ID,
			Type:         rpapiv1.RPMemberUserType,
			PseudoOrSlug: user.Pseudo,
			AvailableRoles: &[]rpapiv1.RPRole{
				{
					Name:   "clientRelease of release-image",
					Label:  "clientRelease",
					Object: "release-image",
				},
			},
		}
		tmpAudience.IdentityRolesLists[user.ID.String()] = &tmpRolesList
	}

	// Add all the teams as available
	for _, team := range allTeams.Teams {
		tmpRolesList := rpapiv1.RPIdentityRolesLists{
			ID:           team.ID,
			Type:         rpapiv1.RPMemberTeamType,
			PseudoOrSlug: team.Slug,
			AvailableRoles: &[]rpapiv1.RPRole{
				{
					Name:   "clientRelease of release-image",
					Label:  "clientRelease",
					Object: "release-image",
				},
			},
		}
		tmpAudience.IdentityRolesLists[team.ID.String()] = &tmpRolesList
	}

	// Add the audience in the global audience list
	allRelImagesAudience = append(allRelImagesAudience, tmpAudience)

	// Wait a bit to send the events
	go func(imageIndex, relImageIndex int) {
		// Wait a bit
		time.Sleep(genericWaitingTime)

		// Send the corresponding event
		mockedBck.EmitEvent(rpapiv1.EvtImageStateChange, allImages[imageIndex])
		mockedBck.EmitEvent(rpapiv1.EVTReleaseImageStateChange, allReleasedImages[relImageIndex])

	}(imgIndex, len(allReleasedImages)-1)

	resJSON(w, reqBody)
}

// relImgsGetAllHandler is the handler used when a call is done on the release-images route
// with a GET
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) relImgsGetAllHandler(w http.ResponseWriter, req *http.Request) {
	// Get the page number wanted
	pageNb, _ := strconv.Atoi(req.FormValue("page"))

	// Get the per_page number wanted
	perPageNb, _ := strconv.Atoi(req.FormValue("per_page"))

	// Get the potential "filters" string
	filtersStr := req.FormValue("filters")

	// Get the filters
	slugFilter := ""
	if strings.HasPrefix(filtersStr, "slug=") {
		slugFilter = strings.ReplaceAll(filtersStr, "slug=", "")
	}
	nameFilter := ""
	if strings.HasPrefix(filtersStr, "name=") {
		nameFilter = strings.ReplaceAll(filtersStr, "name=", "")
		nameFilter = strings.ReplaceAll(nameFilter, "+", " ")
	}

	// Create the object to return
	relImgsToReturn := rpapiv1.RPReleaseImages{
		Page:    pageNb,
		PerPage: perPageNb,
	}

	// Calculate the minimum and maximum index wanted by the user
	minIndex := (pageNb - 1) * perPageNb
	maxIndex := (pageNb * perPageNb) - 1

	// Return only the released projects wanted by the client on this request
	count := 0
	for _, image := range allReleasedImages {
		if (count >= minIndex) && (count <= maxIndex) &&
			(slugFilter == "" || image.Slug == slugFilter) &&
			(nameFilter == "" || image.Name == nameFilter) {
			relImgsToReturn.ReleaseImages = append(relImgsToReturn.ReleaseImages, image.RPReleaseImage)
		}
		count++
	}

	resJSON(w, relImgsToReturn)
}

// relImgsGetHandler is the handler used when a call is done on the release-images/{id} route
// with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) relImgsGetHandler(w http.ResponseWriter, req *http.Request) {
	imageID := reqParamGet(req, "imageid")
	imgUUID := uuid.FromStringOrNil(imageID)

	for _, img := range allReleasedImages {
		if utils.UUIDEqual(img.ID, imgUUID) {
			img.FullData = true
			resJSON(w, img)
			break
		}
	}
}

// relImgsPutHandler is the handler used when a call is done on the release-images/{id} route
// with the PUT method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) relImgsPutHandler(w http.ResponseWriter, req *http.Request) {
	imageID := reqParamGet(req, "imageid")
	imgUUID := uuid.FromStringOrNil(imageID)

	imgIndex := -1
	for index, img := range allReleasedImages {
		if utils.UUIDEqual(img.ID, imgUUID) {
			// We found the right image, save the index to modify it afterwards
			imgIndex = index
			break
		}
	}

	if imgIndex == -1 {
		resJSONErrorE(w, errors.New("The released image does not exists"))
		return
	}

	// Get the request body
	reqBody := &rpapiv1.RPReleaseImage{}
	if err := reqBind(req, reqBody); err != nil {
		resJSONErrorE(w, err)
		return
	}

	// Modify the object
	allReleasedImages[imgIndex].Name = reqBody.Name

	// Send the result
	resJSON(w, allReleasedProjects[imgIndex])
}

// relImgsDeleteHandler is the handler used when a call is done on the release-images/{id} route
// with the DELETE method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) relImgsDeleteHandler(w http.ResponseWriter, req *http.Request) {
	imageID := reqParamGet(req, "imageid")
	imgUUID := uuid.FromStringOrNil(imageID)

	tmpImgs := []rpapiv1.RPReleaseImageFull{}
	resImg := rpapiv1.RPReleaseImageFull{}
	for _, img := range allReleasedImages {
		if !utils.UUIDEqual(img.ID, imgUUID) {
			tmpImgs = append(tmpImgs, img)
		} else {
			// Copy the image we will need to send at the end
			copier.Copy(&resImg, &img)
		}
	}

	// Set the new tab of released images
	allReleasedImages = tmpImgs

	resJSON(w, resImg)
}

// relImgsPutAudienceHandler is the handler used when a call is done on the release-images/{id}/audience route
// with the PUT method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) relImgsPutAudienceHandler(w http.ResponseWriter, req *http.Request) {
	// Get the concerned release-image ID
	imageID := reqParamGet(req, "imageid")
	imgUUID := uuid.FromStringOrNil(imageID)

	// Get the request body
	reqBody := &rpapiv1.RPObjectRoles{}
	if err := reqBind(req, reqBody); err != nil {
		resJSONErrorE(w, err)
		return
	}

	// Find the release image to modify
	indexToModify := -1
	for index, audience := range allRelImagesAudience {
		if utils.UUIDEqual(audience.ObjectID, imgUUID) {
			indexToModify = index
		}
	}

	if indexToModify == -1 {
		resJSONErrorE(w, errors.New("The released image audience does not exists"))
		return
	}

	// Modify this index
	allRelImagesAudience[indexToModify] = *reqBody

	// Return the updated value
	resJSON(w, allRelImagesAudience[indexToModify])
}

// relImgsGetAudienceHandler is the handler used when a call is done on the release-images/{id}/audience route
// with the GET method
//
// :param w: Response writer
// :type w: http.ResponseWriter
// :param req: Pointer to the http request received
// :type req: *http.Request
func (mockedBck *RPMockedBackend) relImgsGetAudienceHandler(w http.ResponseWriter, req *http.Request) {
	// Get the concerned release-image ID
	imageID := reqParamGet(req, "imageid")
	imgUUID := uuid.FromStringOrNil(imageID)

	// Find the released image to return
	for _, audience := range allRelImagesAudience {
		if utils.UUIDEqual(audience.ObjectID, imgUUID) {
			resJSON(w, audience)
			return
		}
	}
	resJSONErrorE(w, errors.New("The released image audience does not exists"))
}

// -------------------------------------------------------------------------------
// -------------------------------------------------------------------------------

// -------------------------------------------------------------------------------
// Utilities functions definition ------------------------------------------------
// -------------------------------------------------------------------------------

func reqBind(req *http.Request, v interface{}) error {
	return json.NewDecoder(req.Body).Decode(v)
}

func reqParamGet(req *http.Request, name string) string {
	param, found := mux.Vars(req)[name]
	if found {
		return param
	} else {
		return ""
	}
}

func resJSON(w http.ResponseWriter, data interface{}) error {
	w.Header().Set("Content-Type", "application/json")
	w.WriteHeader(http.StatusOK)
	return json.NewEncoder(w).Encode(data)
}

func resJSONErrorS(w http.ResponseWriter, format string, args ...interface{}) error {
	w.Header().Set("Content-Type", "application/json")
	errMsg := fmt.Sprintf(format, args...)
	e := map[string]string{
		"status": "error",
		"error":  errMsg,
		"type":   "<class xxx>",
	}
	w.WriteHeader(http.StatusInternalServerError)
	return json.NewEncoder(w).Encode(e)
}

func resJSONErrorE(w http.ResponseWriter, err error) error {
	return resJSONErrorS(w, err.Error())
}

func resTextUnauthError(w http.ResponseWriter, errToSend error) error {
	// Modify the header for text/html body type
	w.Header().Set("Content-Type", "text/html; charset=utf-8")
	// Set to unauthorized status
	w.WriteHeader(http.StatusUnauthorized)
	// Write the text
	_, err := w.Write([]byte(errToSend.Error()))
	return err
}

func updateSubTaskStatus(bldToUpdate *rpapiv1.RPBuildFull, subTaskName string, newStatus rpapiv1.RPBuildTaskStatus) {
	// Look for the subtask
	for index, subtask := range bldToUpdate.Task.Subtasks {
		if subtask.Label == subTaskName {
			bldToUpdate.Task.Subtasks[index].State = newStatus
			break
		}
	}
}

func (mockedBck *RPMockedBackend) createSRPMBuild(app *rpapiv1.RPApplicationFull) {
	// Increment counter
	buildsCounter++
	srpmBuildID := buildsCounter

	// Get the current time
	timeNow := time.Now()

	// Create the build object
	srpmBuild := rpapiv1.RPBuildFull{
		RPBuild: rpapiv1.RPBuild{
			ID:               strconv.Itoa(srpmBuildID),
			BuildType:        rpapiv1.RPBuildSrpm,
			DistributionID:   app.Distributions[0].ID,
			DistributionSlug: app.Distributions[0].Slug,
			PackageName:      app.PackageName,
			Revision:         app.Metadata.Data.Revision,
			RepoURI:          app.Metadata.Data.SourcesURI,
			Status:           rpapiv1.RPBuildTaskStatusDone,
			CreatedAt:        timeNow,
			Parent: rpapiv1.RPBuildParent{
				ApplicationID:   app.ID,
				ApplicationSlug: app.Slug,
				ProjectID:       app.Parent.ProjectID,
				ProjectSlug:     app.Parent.ProjectSlug,
			},
		},
		Task: rpapiv1.RPTask{
			ID:       1000000,
			State:    rpapiv1.RPBuildTaskStatusDone,
			Subtasks: rpapiv1.RPSubTasks{},
			Error:    rpapiv1.RPBuildError{},
			BaseURL:  "http://0.0.0.0:8111",
		},
	}

	// Create the waitrepo subtask
	srpmBuild.Task.Subtasks = append(srpmBuild.Task.Subtasks, getWaitRepoSubTask(rpapiv1.RPBuildTaskStatusClosed))

	// Create the SRPM subtask
	srpmBuild.Task.Subtasks = append(srpmBuild.Task.Subtasks, getSRpmSubTask(rpapiv1.RPBuildTaskStatusClosed))

	// Create the check-deps subtask
	srpmBuild.Task.Subtasks = append(srpmBuild.Task.Subtasks, getCheckDepSubTask(rpapiv1.RPBuildTaskStatusClosed))

	// Finally add it to the allBuilds array
	allBuilds = append(allBuilds, srpmBuild)
}

func (mockedBck *RPMockedBackend) startOneAppBuild(appUUID uuid.UUID, distroToBuild *rpapiv1.RPDistro, comment, buildType string) (rpapiv1.RPBuildFull, error) {

	// Check that the project exists
	appExists := false
	appSlug := ""
	appPkgName := ""
	appRev := ""
	appRepoURL := ""
	var appPrjUUID uuid.UUID
	appPrjSlug := ""
	var appDistribID uuid.UUID
	appDistribSlug := ""
	for _, app := range allApps {
		if utils.UUIDEqual(app.ID, appUUID) {
			appExists = true
			appSlug = app.Slug
			appPkgName = app.PackageName
			appRev = app.Metadata.Data.Revision
			appRepoURL = app.Metadata.Data.SourcesURI
			appPrjUUID = app.Parent.ProjectID
			appPrjSlug = app.Parent.ProjectSlug

			isOK := false
			for _, appDistro := range app.Distributions {
				if utils.UUIDEqual(appDistro.ID, distroToBuild.ID) && distroToBuild.Set && appDistro.Set {
					isOK = true
					break
				}
			}
			if !isOK {
				errMsg := fmt.Sprintf("The application \"%s\" has a not matching distribution to build", appSlug)
				return rpapiv1.RPBuildFull{}, errors.New(errMsg)
			}

			appDistribID = distroToBuild.ID
			appDistribSlug = distroToBuild.Slug
			break
		}
	}
	if !appExists {
		return rpapiv1.RPBuildFull{}, errors.New("The application does not exists")
	}

	// Get the current time
	timeNow := time.Now()

	// Increment counter
	buildsCounter++
	buildID := buildsCounter

	// Create the build object
	currentBuild := rpapiv1.RPBuildFull{
		RPBuild: rpapiv1.RPBuild{
			ID:               strconv.Itoa(buildID),
			BuildType:        rpapiv1.RPBuildType(buildType),
			DistributionID:   appDistribID,
			DistributionSlug: appDistribSlug,
			PackageName:      appPkgName,
			Revision:         appRev,
			RepoURI:          appRepoURL,
			Status:           rpapiv1.RPBuildTaskStatusFree,
			CreatedAt:        timeNow,
			ParentType:       rpapiv1.RPBuildAppParentType,
			Parent: rpapiv1.RPBuildParent{
				BuildSrpmID:     (buildID - 1),
				ApplicationID:   appUUID,
				ApplicationSlug: appSlug,
				ProjectID:       appPrjUUID,
				ProjectSlug:     appPrjSlug,
			},
			Comment: comment,
		},
		Task: rpapiv1.RPTask{
			ID:       buildsCounter,
			State:    rpapiv1.RPBuildTaskStatusFree,
			Subtasks: rpapiv1.RPSubTasks{},
			Error:    rpapiv1.RPBuildError{},
			BaseURL:  "http://0.0.0.0:8111",
		},
	}

	// Add the build in the builds slice
	buildsMutex.Lock()
	allBuilds = append(allBuilds, currentBuild)
	bldToUpdate := &(allBuilds[len(allBuilds)-1])
	buildsMutex.Unlock()

	// Start go routine to emit events like if it was building in reality
	go mockedBck.simulateBuild(rpapiv1.RPBuildType(buildType), bldToUpdate)

	return currentBuild, nil
}

func (mockedBck *RPMockedBackend) simulateBuild(bldType rpapiv1.RPBuildType, bldPointer *rpapiv1.RPBuildFull) {

	bldToUpdate := bldPointer

	// Wait that the function has returned
	time.Sleep(10 * time.Millisecond)

	// Update the state of build
	bldToUpdate.Status = rpapiv1.RPBuildTaskStatusOpen
	bldToUpdate.Task.State = rpapiv1.RPBuildTaskStatusOpen

	// Emit the first event
	mockedBck.EmitEvent(rpapiv1.EvtAppBuildInfo, bldToUpdate)

	// Wait a bit
	time.Sleep(5 * time.Millisecond)

	// Add the subtasks in task
	if bldType == rpapiv1.RPBuildApp {
		bldToUpdate.Task.Subtasks = append(bldToUpdate.Task.Subtasks, getX8664SubTask(bldToUpdate.Task.ID))
		bldToUpdate.Task.Subtasks = append(bldToUpdate.Task.Subtasks, getAarch64SubTask(bldToUpdate.Task.ID))
	} else if bldType == rpapiv1.RPBuildSrpm {
		bldToUpdate.Task.Subtasks = append(bldToUpdate.Task.Subtasks, getWaitRepoSubTask(rpapiv1.RPBuildTaskStatusOpen))
		bldToUpdate.Task.Subtasks = append(bldToUpdate.Task.Subtasks, getSRpmSubTask(rpapiv1.RPBuildTaskStatusOpen))
		bldToUpdate.Task.Subtasks = append(bldToUpdate.Task.Subtasks, getCheckDepSubTask(rpapiv1.RPBuildTaskStatusOpen))
	} else if bldType == rpapiv1.RPBuildDeps {
		bldToUpdate.Task.Subtasks = append(bldToUpdate.Task.Subtasks, getWaitRepoSubTask(rpapiv1.RPBuildTaskStatusClosed))
		bldToUpdate.Task.Subtasks = append(bldToUpdate.Task.Subtasks, getSRpmSubTask(rpapiv1.RPBuildTaskStatusClosed))
		bldToUpdate.Task.Subtasks = append(bldToUpdate.Task.Subtasks, getCheckDepSubTask(rpapiv1.RPBuildTaskStatusOpen))
	}

	// Wait a bit and emit the fourth event
	time.Sleep(5 * time.Millisecond)
	mockedBck.EmitEvent(rpapiv1.EvtAppBuildSubTaskInfo, bldToUpdate)

	if bldType == rpapiv1.RPBuildApp {
		// Update the x86_64 subtask status to closed
		updateSubTaskStatus(bldToUpdate, "x86_64", rpapiv1.RPBuildTaskStatusClosed)
	} else {
		// Update the waitrepo subtask status to closed
		updateSubTaskStatus(bldToUpdate, "", rpapiv1.RPBuildTaskStatusClosed)
		// Update the srpm subtask status to closed
		updateSubTaskStatus(bldToUpdate, "srpm", rpapiv1.RPBuildTaskStatusClosed)
	}

	// Wait a bit and emit the fifth event
	time.Sleep(5 * time.Millisecond)
	mockedBck.EmitEvent(rpapiv1.EvtAppBuildSubTaskInfo, bldToUpdate)

	// Check if we need to fail the build
	aarch64Status := rpapiv1.RPBuildTaskStatusClosed
	finalStatus := rpapiv1.RPBuildTaskStatusDone
	if nextBuildShouldFail && (bldType == rpapiv1.RPBuildApp) {
		aarch64Status = rpapiv1.RPBuildTaskStatusFailed
		finalStatus = rpapiv1.RPBuildTaskStatusFailed
		nextBuildShouldFail = false
	}

	if bldType == rpapiv1.RPBuildApp {
		// Update the aarch64 subtask status to closed
		updateSubTaskStatus(bldToUpdate, "aarch64", aarch64Status)
	} else {
		// Update the check-deps subtask status to closed
		updateSubTaskStatus(bldToUpdate, "check-deps", rpapiv1.RPBuildTaskStatusClosed)
	}

	// Wait a bit and emit the sixth event
	time.Sleep(5 * time.Millisecond)
	mockedBck.EmitEvent(rpapiv1.EvtAppBuildSubTaskInfo, bldToUpdate)

	if nextBuildCanceled {
		// Wait until the cancel request has been sent by the test function
		for !utils.ContainsString(buildsToCancel, bldToUpdate.ID) {
			time.Sleep(5 * time.Millisecond)
		}
		nextBuildCanceled = false
	}

	// Check if the build is not part of the ones that have been canceled
	if utils.ContainsString(buildsToCancel, bldToUpdate.ID) {
		// Update the Build status to canceled
		bldToUpdate.Status = rpapiv1.RPBuildTaskStatusCanceled
		bldToUpdate.Task.State = rpapiv1.RPBuildTaskStatusCanceled
		// Delete the build ID from the slice
		utils.DeleteStringFromSlice(&buildsToCancel, bldToUpdate.ID)
	} else {
		// Update the Build status to closed
		bldToUpdate.Status = finalStatus
		bldToUpdate.Task.State = finalStatus
	}

	// Wait a bit and emit the seventh event
	time.Sleep(5 * time.Millisecond)
	mockedBck.EmitEvent(rpapiv1.EvtAppBuildInfo, bldToUpdate)
}

func (mockedBck *RPMockedBackend) startOneImgBuild(imageUUID uuid.UUID, comment string) (rpapiv1.RPBuildFull, error) {
	// Check that the image exists
	imgIndex := -1
	var imgDistribID uuid.UUID

	for index, img := range allImages {
		if utils.UUIDEqual(img.ID, imageUUID) {
			imgIndex = index
			imgDistribID = img.DistributionID
			break
		}
	}

	if imgIndex == -1 {
		return rpapiv1.RPBuildFull{}, errors.New("The image does not exists")
	}

	// Retrieve the corresponding distribution slug
	imgDistroSlug := ""
	for _, prj := range allProjects {
		if utils.UUIDEqual(prj.ID, imgDistribID) {
			imgDistroSlug = prj.Slug
		}
	}

	if imgDistroSlug == "" {
		return rpapiv1.RPBuildFull{}, errors.New("The distribution specified in image does not exists")
	}

	// Get the current time
	timeNow := time.Now()

	// Increment counter
	buildsCounter++

	// Create the build object
	currentBuild := rpapiv1.RPBuildFull{
		RPBuild: rpapiv1.RPBuild{
			ID:               strconv.Itoa(buildsCounter),
			BuildType:        rpapiv1.RPBuildImage,
			DistributionSlug: imgDistroSlug,
			DistributionID:   imgDistribID,
			Revision:         allImages[imgIndex].Version,
			Status:           rpapiv1.RPBuildTaskStatusFree,
			CreatedAt:        timeNow,
			ParentType:       rpapiv1.RPBuildImgParentType,
			Parent: rpapiv1.RPBuildParent{
				ImageID:   allImages[imgIndex].ID,
				ImageSlug: allImages[imgIndex].Slug,
			},
			Comment: comment,
		},
		Task: rpapiv1.RPTask{
			ID:         buildsCounter,
			Method:     "livemedia",
			MethodName: "livemedia",
			State:      rpapiv1.RPBuildTaskStatusFree,
			Subtasks:   rpapiv1.RPSubTasks{},
			Error:      rpapiv1.RPBuildError{},
			BaseURL:    "http://0.0.0.0:8111",
		},
	}

	// Add the build in the builds slice
	buildsMutex.Lock()
	allBuilds = append(allBuilds, currentBuild)
	bldToUpdate := &(allBuilds[len(allBuilds)-1])
	buildsMutex.Unlock()

	// Start go routine to emit events like if it was building in reality
	go mockedBck.simulateSuccessImageBuild(bldToUpdate)

	return currentBuild, nil
}

func (mockedBck *RPMockedBackend) simulateSuccessImageBuild(bldToUpdate *rpapiv1.RPBuildFull) {
	// Wait that the function has returned
	time.Sleep(10 * time.Millisecond)

	// Increment the builds ID counter
	buildsCounter++

	// Create the first subtask
	tmpSubtask := rpapiv1.RPSubTask{
		ID:    buildsCounter,
		Label: "livemedia x86_64",
		Results: rpapiv1.RPBuildResults{
			Logs: []string{"build.log", "root.log", "anaconda-dbus.log",
				"anaconda-hawkey.log", "anaconda-anaconda.log", "Redpesk-minimal-x86_64-smack.ks", "koji-image-redpesk-my-project--redpesk-lts-arz-1.0-update-build-666.ks"},
			Files: []string{"Redpesk-generic-x86_64-arz-1.0.tar.xz"},
		},
		Method:     "createLiveMedia",
		MethodName: "createLiveMedia",
		Parent:     bldToUpdate.Task.ID,
		State:      rpapiv1.RPBuildTaskStatusOpen,
	}

	// Add the subtask in task
	bldToUpdate.Task.Subtasks = append(bldToUpdate.Task.Subtasks, tmpSubtask)

	// Update the state of build
	bldToUpdate.Status = rpapiv1.RPBuildTaskStatusOpen
	bldToUpdate.Task.State = rpapiv1.RPBuildTaskStatusOpen

	// Emit the first event
	mockedBck.EmitEvent(rpapiv1.EvtImageBuildInfo, bldToUpdate)

	// Wait a bit
	time.Sleep(10 * time.Millisecond)

	// Emit the second event
	mockedBck.EmitEvent(rpapiv1.EvtImageBuildSubTaskInfo, bldToUpdate)

	// Update the livemedia x86_64 subtask status to closed
	updateSubTaskStatus(bldToUpdate, "livemedia x86_64", "closed")

	// Wait a bit and emit the third event
	time.Sleep(10 * time.Millisecond)
	mockedBck.EmitEvent(rpapiv1.EvtImageBuildSubTaskInfo, bldToUpdate)

	if nextBuildCanceled {
		// Wait until the cancel request has been sent by the test function
		for !utils.ContainsString(buildsToCancel, bldToUpdate.ID) {
			time.Sleep(2 * time.Millisecond)
		}
		nextBuildCanceled = false
	}

	// Check if the build is not part of the ones that have been canceled
	if utils.ContainsString(buildsToCancel, bldToUpdate.ID) {
		// Update the Build status to canceled
		bldToUpdate.Status = rpapiv1.RPBuildTaskStatusCanceled
		bldToUpdate.Task.State = rpapiv1.RPBuildTaskStatusCanceled
		// Delete the build ID from the slice
		utils.DeleteStringFromSlice(&buildsToCancel, bldToUpdate.ID)
	} else {
		// Update the Build status to done and the task status to "closed"
		bldToUpdate.Status = rpapiv1.RPBuildTaskStatusDone
		bldToUpdate.Task.State = "closed"
	}

	// Wait a bit and emit the seventh event
	time.Sleep(5 * time.Millisecond)
	mockedBck.EmitEvent(rpapiv1.EvtImageBuildInfo, bldToUpdate)
}

func (mockedBck *RPMockedBackend) simulateSuccessAppAudit() {
	// Get the build to update
	auditToUpdate := &(allAudits[len(allAudits)-1])

	// Wait that the function has returned
	time.Sleep(10 * time.Millisecond)

	// Update the state of build
	auditToUpdate.Status = rpapiv1.RPAuditStatusStable
	auditToUpdate.Ended = time.Now()

	// Emit the first event
	mockedBck.EmitEvent(rpapiv1.EvtAuditStateChange, auditToUpdate)
}

func (mockedBck *RPMockedBackend) simulateSuccessTest() {
	// Get the build to update
	testToUpdate := &(allTests.Tests[len(allTests.Tests)-1])

	// Wait that the function has returned
	time.Sleep(10 * time.Millisecond)

	// Emit an event
	mockedBck.EmitEvent(rpapiv1.EvtTestAdd, testToUpdate)

	// Pass test to deploying
	testToUpdate.Status = rpapiv1.RPTestStatusDeploying

	// Wait a bit and emit event
	time.Sleep(5 * time.Millisecond)
	mockedBck.EmitEvent(rpapiv1.EvtTestStateChange, testToUpdate)

	// Pass test to started
	testToUpdate.Status = rpapiv1.RPTestStatusStarted

	// Wait a bit and emit event
	time.Sleep(5 * time.Millisecond)
	mockedBck.EmitEvent(rpapiv1.EvtTestStateChange, testToUpdate)

	// Go through the logs file and emit events according to it
	logsFileLines := strings.Split(testsLogFile, "\n")
	testsRun := 0
	for index, line := range logsFileLines {

		// Check if we need to abort the test
		if abortCurrentTest {
			// Change the status of the test to "aborted"
			testToUpdate.Status = rpapiv1.RPTestStatusAborted

			// Wait a bit and emit event
			time.Sleep(5 * time.Millisecond)
			mockedBck.EmitEvent(rpapiv1.EvtTestStateChange, testToUpdate)

			// Reset the value
			abortCurrentTest = false

			break
		}

		// Get the current time
		timeNow := time.Now()

		// Check if we send as STDERR
		if strings.HasPrefix(line, "~~") {
			// This kind of line is sent as STDERR
			tmpEvtConsole := rpapiv1.RPEventTestConsole{
				ID:         testToUpdate.ID,
				Timestamp:  timeNow.Format("02-01-2006 15:04:05"),
				LineStdOut: "",
				LineStdErr: line,
			}
			time.Sleep(2 * time.Millisecond)
			mockedBck.EmitEvent(rpapiv1.EvtTestConsole, tmpEvtConsole)
		} else {
			// Send the line as STDOUT
			tmpEvtConsole := rpapiv1.RPEventTestConsole{
				ID:         testToUpdate.ID,
				Timestamp:  timeNow.Format("02-01-2006 15:04:05"),
				LineStdOut: line,
				LineStdErr: "",
			}
			time.Sleep(2 * time.Millisecond)
			mockedBck.EmitEvent(rpapiv1.EvtTestConsole, tmpEvtConsole)

			// Check if we increment the number of tests
			if strings.HasPrefix(line, "ok") {
				testsRun++
				// Send an event progress
				tmpEvtProgress := rpapiv1.RPEventTest{
					ID:                testToUpdate.ID,
					CurrentTestNb:     testsRun,
					TotalTestsNb:      10,
					CurrentTestStatus: "OK",
					CurrentTestsOKNb:  testsRun,
				}
				time.Sleep(2 * time.Millisecond)
				mockedBck.EmitEvent(rpapiv1.EvtTestProgress, tmpEvtProgress)
			}
		}

		// Check if we need to send a time-out event
		if testInTimeOut && (index > (len(logsFileLines) / 2)) {
			// Change the status of the test to "aborted"
			testToUpdate.Status = rpapiv1.RPTestStatusTimedOut

			// Wait a bit and emit event
			time.Sleep(5 * time.Millisecond)
			mockedBck.EmitEvent(rpapiv1.EvtTestStateChange, testToUpdate)

			// Reset the value
			testInTimeOut = false

			break
		}
	}

	if (testToUpdate.Status != rpapiv1.RPTestStatusAborted) &&
		(testToUpdate.Status != rpapiv1.RPTestStatusTimedOut) {
		// Pass test to testing
		testToUpdate.Status = rpapiv1.RPTestStatusTesting

		// Wait a bit and emit event
		time.Sleep(5 * time.Millisecond)
		mockedBck.EmitEvent(rpapiv1.EvtTestStateChange, testToUpdate)

		// Pass test to success
		testToUpdate.Status = rpapiv1.RPTestStatusSuccess
		testToUpdate.TotalTestsNb = 10
		testToUpdate.SucceededNb = testsRun

		// Wait a bit and emit event
		time.Sleep(5 * time.Millisecond)
		mockedBck.EmitEvent(rpapiv1.EvtTestStateChange, testToUpdate)
	}
}

func convertProjectFullToProject(prjFull rpapiv1.RPProjectFull) rpapiv1.RPProject {
	// Initialize the project structure to return
	prjToReturn := rpapiv1.RPProject{}
	// Convert the project full to slice of bytes
	dataBytes, _ := json.Marshal(prjFull)
	// Convert slice of bytes to RPProject
	json.Unmarshal(dataBytes, &prjToReturn)
	// return the created object
	return prjToReturn
}

func convertAppFullToApp(appFull rpapiv1.RPApplicationFull) rpapiv1.RPApplication {
	// Initialize the app structure to return
	appToReturn := rpapiv1.RPApplication{}
	// Convert the app full to slice of bytes
	dataBytes, _ := json.Marshal(appFull)
	// Convert slice of bytes to RPApplication
	json.Unmarshal(dataBytes, &appToReturn)
	// return the created object
	return appToReturn
}

func convertBuildFullToBuild(buildFull rpapiv1.RPBuildFull) rpapiv1.RPBuild {
	// Initialize the build structure to return
	buildToReturn := rpapiv1.RPBuild{}
	// Convert the build full to slice of bytes
	dataBytes, _ := json.Marshal(buildFull)
	// Convert slice of bytes to RPApplication
	json.Unmarshal(dataBytes, &buildToReturn)
	// return the created object
	return buildToReturn
}

func convertImageFullToImage(imgFull rpapiv1.RPImageFull) rpapiv1.RPImage {
	// Initialize the image structure to return
	imgToReturn := rpapiv1.RPImage{}
	// Convert the image full to slice of bytes
	dataBytes, _ := json.Marshal(imgFull)
	// Convert slice of bytes to RPImage
	json.Unmarshal(dataBytes, &imgToReturn)
	// return the created object
	return imgToReturn
}

func convertAuditFullToAudit(auditFull rpapiv1.RPAuditFull) rpapiv1.RPAudit {
	// Initialize the audit structure to return
	auditToReturn := rpapiv1.RPAudit{}
	// Convert the audit full to slice of bytes
	dataBytes, _ := json.Marshal(auditFull)
	// Convert slice of bytes to RPAudit
	json.Unmarshal(dataBytes, &auditToReturn)
	// return the created object
	return auditToReturn
}

func getX8664SubTask(parentTaskID int) rpapiv1.RPSubTask {
	return rpapiv1.RPSubTask{
		ID:    1000002,
		Label: "x86_64",
		Results: rpapiv1.RPBuildResults{
			Logs: []string{"root.log", "build.log"},
			RPMS: []string{
				"helloworld-binding-test-1.0.0-8.redpesk.os.samples_5cfeb3a9.rpdarz.x86_64.rpm",
				"helloworld-binding-redtest-1.0.0-8.redpesk.os.samples_5cfeb3a9.rpdarz.x86_64.rpm",
				"helloworld-binding-debuginfo-1.0.0-8.redpesk.os.samples_5cfeb3a9.rpdarz.x86_64.rpm",
				"helloworld-binding-1.0.0-8.redpesk.os.samples_5cfeb3a9.rpdarz.x86_64.rpm",
				"helloworld-binding-debugsource-1.0.0-8.redpesk.os.samples_5cfeb3a9.rpdarz.x86_64.rpm",
			},
		},
		Method:     "buildArch",
		MethodName: "Building native application",
		Parent:     parentTaskID,
		State:      rpapiv1.RPBuildTaskStatusOpen,
	}
}

func getAarch64SubTask(parentTaskID int) rpapiv1.RPSubTask {
	return rpapiv1.RPSubTask{
		ID:    1000003,
		Label: "aarch64",
		Results: rpapiv1.RPBuildResults{
			Logs: []string{"root.log", "build.log"},
			RPMS: []string{
				"helloworld-binding-test-1.0.0-8.redpesk.os.samples_5cfeb3a9.rpdarz.aarch64.rpm",
				"helloworld-binding-redtest-1.0.0-8.redpesk.os.samples_5cfeb3a9.rpdarz.aarch64.rpm",
				"helloworld-binding-debuginfo-1.0.0-8.redpesk.os.samples_5cfeb3a9.rpdarz.aarch64.rpm",
				"helloworld-binding-1.0.0-8.redpesk.os.samples_5cfeb3a9.rpdarz.aarch64.rpm",
				"helloworld-binding-debugsource-1.0.0-8.redpesk.os.samples_5cfeb3a9.rpdarz.aarch64.rpm",
			},
		},
		Method:     "buildArchCross",
		MethodName: "Cross-building application",
		Parent:     parentTaskID,
		State:      rpapiv1.RPBuildTaskStatusOpen,
	}
}

func getWaitRepoSubTask(state rpapiv1.RPBuildTaskStatus) rpapiv1.RPSubTask {
	timeNow := time.Now()
	return rpapiv1.RPSubTask{
		ID:             1000001,
		StartTime:      timeNow.String(),
		CompletionTime: timeNow.String(),
		Type:           "subtask",
		Method:         "waitrepo",
		MethodName:     "Setup build environment",
		Parent:         1000000,
		State:          state,
	}
}

func getSRpmSubTask(state rpapiv1.RPBuildTaskStatus) rpapiv1.RPSubTask {
	timeNow := time.Now()
	return rpapiv1.RPSubTask{
		ID:             1000002,
		StartTime:      timeNow.String(),
		CompletionTime: timeNow.String(),
		Type:           "subtask",
		Label:          "srpm",
		Method:         "buildSRPMFromSCM",
		MethodName:     "Packaging sources into SRPM package",
		Parent:         1000000,
		State:          state,
		Results: rpapiv1.RPBuildResults{
			Logs:  []string{"root.log", "build.log"},
			SRPMS: []string{"helloworld-binding-1.0.0-rparz.src.rpm"},
		},
	}
}

func getCheckDepSubTask(state rpapiv1.RPBuildTaskStatus) rpapiv1.RPSubTask {
	timeNow := time.Now()
	return rpapiv1.RPSubTask{
		ID:             0,
		StartTime:      timeNow.String(),
		CompletionTime: timeNow.String(),
		Type:           "subtask",
		Label:          "deps",
		Method:         "check-deps",
		MethodName:     "Check package dependencies",
		Parent:         1000000,
		State:          state,
		Results: rpapiv1.RPBuildResults{
			CheckDeps: []string{"{\"state\":\"closed\",\"status\":\"Resolvable\",\"name\":\"helloworld-binding\",\"version\":\"1.0.1+20221129+10+g561164a\"}"},
		},
	}
}

// -------------------------------------------------------------------------------
// -------------------------------------------------------------------------------
